Type: Package
Package: duckspatial
Title: R Interface to 'DuckDB' Database with Spatial Extension
Version: 1.2.1
Authors@R: c(
    person(
        "Adrián", "Cidre González", 
        email = "adrian.cidre@gmail.com", 
        role = c("aut", "cre"),
        comment = c(ORCID = "0000-0002-3310-3052")
    ),
    person(
        "Egor", "Kotov", , 
        "kotov.egor@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0001-6690-5345")
    ),
    person(
        "Rafael H. M.", "Pereira", , 
        "rafa.pereira.br@gmail.com", 
        role = "aut",
        comment = c(ORCID = "0000-0003-2125-7465")
    ))
Description: Fast and memory-efficient functions to analyze and manipulate large
             spatial datasets. It leverages the fast analytical
             capabilities of ‘DuckDB’ and its spatial extension (see
             <https://duckdb.org/docs/stable/core_extensions/spatial/overview>)
             while maintaining compatibility with R’s spatial data ecosystem to
             work with spatial vector data.
URL: https://cidree.github.io/duckspatial/,
        https://github.com/Cidree/duckspatial
BugReports: https://github.com/Cidree/duckspatial/issues
License: GPL (>= 3)
Depends: R (>= 4.1.0)
Imports: arrow, cli, DBI, dbplyr (>= 2.0.0), dplyr, duckdb (>=
        1.5.4.2), geoarrow, glue, jsonlite, lifecycle, nanoarrow,
        rlang, sf, tibble, tools, units, uuid, withr, wk
Suggests: areal, bench, duckdbfs, geojsonsf, ggplot2 (>= 3.3.1), knitr,
        lwgeom, patchwork, quadkeyr, quarto, rmarkdown, scales, terra,
        testthat (>= 3.0.0)
VignetteBuilder: quarto
Config/testthat/edition: 3
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-04 06:53:44 UTC; cidre
Author: Adrián Cidre González [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-3310-3052>),
  Egor Kotov [aut] (ORCID: <https://orcid.org/0000-0001-6690-5345>),
  Rafael H. M. Pereira [aut] (ORCID:
    <https://orcid.org/0000-0003-2125-7465>)
Maintainer: Adrián Cidre González <adrian.cidre@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-04 10:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(arrange,duckspatial_df)
S3method(arrow::as_arrow_table,duckspatial_df)
S3method(arrow::as_record_batch_reader,duckspatial_df)
S3method(as_duckspatial_df,character)
S3method(as_duckspatial_df,data.frame)
S3method(as_duckspatial_df,duckspatial_df)
S3method(as_duckspatial_df,sf)
S3method(as_duckspatial_df,tbl_duckdb_connection)
S3method(as_duckspatial_df,tbl_lazy)
S3method(collect,duckspatial_df)
S3method(compute,duckspatial_df)
S3method(count,duckspatial_df)
S3method(ddbs_crs,character)
S3method(ddbs_crs,crs)
S3method(ddbs_crs,data.frame)
S3method(ddbs_crs,default)
S3method(ddbs_crs,duckdb_connection)
S3method(ddbs_crs,duckspatial_df)
S3method(ddbs_crs,numeric)
S3method(ddbs_crs,sf)
S3method(ddbs_crs,tbl_duckdb_connection)
S3method(distinct,duckspatial_df)
S3method(dplyr_reconstruct,duckspatial_df)
S3method(filter,duckspatial_df)
S3method(full_join,duckspatial_df)
S3method(geoarrow::infer_geoarrow_schema,duckspatial_df)
S3method(glimpse,duckspatial_df)
S3method(group_by,duckspatial_df)
S3method(head,duckspatial_df)
S3method(inner_join,duckspatial_df)
S3method(left_join,duckspatial_df)
S3method(mutate,duckspatial_df)
S3method(nanoarrow::as_nanoarrow_array_stream,duckspatial_df)
S3method(print,duckspatial_df)
S3method(rename,duckspatial_df)
S3method(right_join,duckspatial_df)
S3method(select,duckspatial_df)
S3method(slice,duckspatial_df)
S3method(st_as_sf,duckspatial_df)
S3method(st_bbox,duckspatial_df)
S3method(st_crs,duckspatial_df)
S3method(st_geometry,duckspatial_df)
S3method(summarise,duckspatial_df)
S3method(ungroup,duckspatial_df)
export(as_duckspatial_df)
export(ddbs_affine)
export(ddbs_area)
export(ddbs_as_geojson)
export(ddbs_as_hexwkb)
export(ddbs_as_mvt_geom)
export(ddbs_as_points)
export(ddbs_as_text)
export(ddbs_as_wkb)
export(ddbs_azimuth)
export(ddbs_bbox)
export(ddbs_boundary)
export(ddbs_buffer)
export(ddbs_build_area)
export(ddbs_centroid)
export(ddbs_collect)
export(ddbs_combine)
export(ddbs_compute)
export(ddbs_concave_hull)
export(ddbs_contains_properly)
export(ddbs_convex_hull)
export(ddbs_covered_by)
export(ddbs_covers)
export(ddbs_create_conn)
export(ddbs_create_schema)
export(ddbs_crop)
export(ddbs_crs)
export(ddbs_difference)
export(ddbs_dimension)
export(ddbs_disjoint)
export(ddbs_distance)
export(ddbs_drivers)
export(ddbs_drop_geometry)
export(ddbs_dump)
export(ddbs_endpoint)
export(ddbs_envelope)
export(ddbs_equals)
export(ddbs_extension_info)
export(ddbs_exterior_ring)
export(ddbs_filter)
export(ddbs_flip)
export(ddbs_flip_coordinates)
export(ddbs_force_2d)
export(ddbs_force_3d)
export(ddbs_force_4d)
export(ddbs_generate_points)
export(ddbs_geom_col)
export(ddbs_geom_from_geojson)
export(ddbs_geom_from_hexewkb)
export(ddbs_geom_from_hexwkb)
export(ddbs_geom_from_text)
export(ddbs_geom_from_wkb)
export(ddbs_geometry_type)
export(ddbs_get_ngeometries)
export(ddbs_get_ninterior_rings)
export(ddbs_get_npoints)
export(ddbs_get_resources)
export(ddbs_glimpse)
export(ddbs_has_m)
export(ddbs_has_z)
export(ddbs_install)
export(ddbs_interpolate_aw)
export(ddbs_intersection)
export(ddbs_intersection_agg)
export(ddbs_intersects)
export(ddbs_intersects_extent)
export(ddbs_is_closed)
export(ddbs_is_empty)
export(ddbs_is_ring)
export(ddbs_is_simple)
export(ddbs_is_valid)
export(ddbs_is_within_distance)
export(ddbs_join)
export(ddbs_length)
export(ddbs_line_endpoint)
export(ddbs_line_interpolate)
export(ddbs_line_locate_point)
export(ddbs_line_merge)
export(ddbs_line_node)
export(ddbs_line_startpoint)
export(ddbs_line_substring)
export(ddbs_list_tables)
export(ddbs_load)
export(ddbs_locate_along)
export(ddbs_locate_between)
export(ddbs_m)
export(ddbs_make_envelope)
export(ddbs_make_line)
export(ddbs_make_polygon)
export(ddbs_make_valid)
export(ddbs_maximum_inscribed_circle)
export(ddbs_minimum_rotated_rectangle)
export(ddbs_mmax)
export(ddbs_mmin)
export(ddbs_multi)
export(ddbs_normalize)
export(ddbs_open_dataset)
export(ddbs_options)
export(ddbs_overlaps)
export(ddbs_perimeter)
export(ddbs_point)
export(ddbs_polygonize)
export(ddbs_predicate)
export(ddbs_quadkey)
export(ddbs_read_meta)
export(ddbs_read_table)
export(ddbs_read_vector)
export(ddbs_reduce_precision)
export(ddbs_register_table)
export(ddbs_register_vector)
export(ddbs_remove_repeated_points)
export(ddbs_reverse)
export(ddbs_rotate)
export(ddbs_rotate_3d)
export(ddbs_scale)
export(ddbs_set_crs)
export(ddbs_set_resources)
export(ddbs_shear)
export(ddbs_shift)
export(ddbs_shortest_line)
export(ddbs_simplify)
export(ddbs_sitrep)
export(ddbs_startpoint)
export(ddbs_stop_conn)
export(ddbs_sym_difference)
export(ddbs_touches)
export(ddbs_transform)
export(ddbs_union)
export(ddbs_union_agg)
export(ddbs_vertices)
export(ddbs_voronoi)
export(ddbs_within)
export(ddbs_within_properly)
export(ddbs_write_dataset)
export(ddbs_write_mbtiles)
export(ddbs_write_table)
export(ddbs_write_vector)
export(ddbs_x)
export(ddbs_xmax)
export(ddbs_xmin)
export(ddbs_y)
export(ddbs_ymax)
export(ddbs_ymin)
export(ddbs_z)
export(ddbs_zmax)
export(ddbs_zmin)
export(is_duckspatial_df)
importFrom(dplyr,arrange)
importFrom(dplyr,collect)
importFrom(dplyr,compute)
importFrom(dplyr,count)
importFrom(dplyr,distinct)
importFrom(dplyr,dplyr_reconstruct)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,glimpse)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,rename)
importFrom(dplyr,right_join)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(lifecycle,deprecated)
importFrom(rlang,":=")
importFrom(sf,st_as_sf)
importFrom(sf,st_bbox)
importFrom(sf,st_crs)
importFrom(sf,st_geometry)
importFrom(utils,head)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# duckspatial <a href="https://cidree.github.io/duckspatial/"><img src="man/figures/logo.png" alt="duckspatial website" align="right" height="138"/></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/duckspatial)](https://CRAN.R-project.org/package=duckspatial)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![Codecov test
coverage](https://codecov.io/gh/Cidree/duckspatial/graph/badge.svg)](https://app.codecov.io/gh/Cidree/duckspatial)
[![License: GPL
v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![Last Month
Downloads](https://cranlogs.r-pkg.org/badges/last-month/duckspatial?color=green)](https://CRAN.R-project.org/package=duckspatial)
[![Total
Downloads](https://cranlogs.r-pkg.org/badges/grand-total/duckspatial?color=green)](https://CRAN.R-project.org/package=duckspatial)
[![check](https://github.com/Cidree/duckspatial/workflows/check/badge.svg)](https://github.com/Cidree/duckspatial/actions)

<!-- badges: end -->

> As of **{duckspatial} v1.0.0**, the package uses the native CRS
> support provided by the DuckDB Spatial extension (DuckDB ≥ 1.5). The
> previous workaround of storing CRS information in a dedicated column
> has been removed, and the `crs` and `crs_column` arguments are no
> longer supported.
>
> Spatial tables are now represented using the `duckspatial_df` class, a
> lazy spatial table backed by a temporary DuckDB table. This allows
> spatial workflows to remain in DuckDB until the results are explicitly
> materialized with `ddbs_collect()`, enabling efficient processing
> without eagerly loading data into R.
>
> This release also introduces a number of new spatial functions and API
> improvements, providing a more consistent and fully native integration
> with DuckDB’s spatial capabilities.

**{duckspatial}** provides fast, memory-efficient functions for
analysing and manipulating large spatial vector datasets in R. It
bridges [DuckDB’s spatial
extension](https://duckdb.org/docs/stable/core_extensions/spatial/functions)
with R’s spatial ecosystem — in particular **{sf}** — so you can
leverage DuckDB’s analytical power without leaving your familiar R
workflow.

### How it works

Starting from v1.0.0, {duckspatial} introduces a native S3 class called
`duckspatial_df`: a `tibble`-like object with a geometry column that
lives **outside R’s memory**. Data is read and evaluated lazily (similar
to how `duckplyr` handles lazy tables) and is only loaded into R when
you explicitly materialize it (e.g. with `ddbs_collect()`).

When the first `duckspatial_df` is created (either by reading a file or
converting an `sf` object) a temporary view is registered in a default
DuckDB connection with the spatial extension enabled. All spatial
operations run inside that connection, letting DuckDB apply its own
query optimizations before any data reaches R.

### Naming conventions

All spatial functions follow the `ddbs_*()` prefix (*DuckDB Spatial*),
and their names deliberately mirror the **{sf}** package, so users
already familiar with `sf` can get started immediately.

## Installation

Install the stable release from CRAN:

``` r
# install.packages("pak")
pak::pak("duckspatial")
```

Install the latest GitHub version (more features, fewer accumulated
bugs):

``` r
pak::pak("Cidree/duckspatial")
```

Install the development version (may be unstable):

``` r
pak::pak("Cidree/duckspatial@dev")
```

## Core idea: flexible spatial workflows

A central design principle of {duckspatial} is that the same spatial
operation can be used in different ways depending on how your data is
stored and how you want to manage memory and performance.

Most functions support four complementary input/output combinations:

| Input                   | Output                  |
|-------------------------|-------------------------|
| `duckspatial_df` / `sf` | `duckspatial_df` / `sf` |
| `duckspatial_df` / `sf` | DuckDB table            |
| DuckDB table            | `duckspatial_df` / `sf` |
| DuckDB table            | DuckDB table            |

This means you can keep data inside DuckDB for as long as possible,
pulling results into R only when you need them. See the [Get Started
vignette](https://Cidree.github.io/duckspatial/articles/duckspatial.html)
for worked examples of each workflow.

## Contributing

Bug reports, feature requests, and pull requests are very welcome!

- [Raise an issue](https://github.com/Cidree/duckspatial/issues)
- [Open a pull request](https://github.com/Cidree/duckspatial/pulls)

## -----------------------------------------------------------------------------
#| include: false

# Limit threads to avoid a CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)


## -----------------------------------------------------------------------------
#| echo: false

## Internal packages
library(patchwork)

## Functions used in the vignette
ggplot_benchmark <- function(
  data, 
  log = FALSE,
  show.legend = TRUE,
  ...
  ) {

    ## Generate base plot
    ggplot(
      data = data, 
      aes(
        x     = factor(format(n, big.mark = ".")),
        y     = if (log) median else as.numeric(median),
        color = pkg,
        group = pkg
    )) +
      geom_line(
        linewidth = 0.7,
        linetype  = "dashed",
        alpha     = 0.6,
        show.legend = show.legend
      ) +
      geom_point(
        aes(size = as.numeric(mem_alloc)),
        alpha = 0.9,
        show.legend = show.legend
      ) +
      scale_size_binned(
        name   = "Memory allocated",
        labels = scales::label_bytes(),
        n.breaks = 4,
        range = c(1, 8)
      ) +
      scale_color_brewer(palette = "Set1") +
      labs(
        color    = "Package",
        x        = "Data size (rows)",
        y        = "Computation time",
        ...
      ) +
      theme_minimal(base_size = 13) +
      theme(
        # legend.position  = "top",
        panel.grid.minor = element_blank(),
        plot.title       = element_text(face = "bold")
      )
}

ggplot_assemble <- function(plot1, plot2, fun_name) {

  plot1 +
    plot2 +
    plot_annotation(
      title    = paste("Benchmark Comparison for", fun_name),
      subtitle = "Execution time vs. data size",
      theme    = theme(
        plot.title = element_text(face = "bold", hjust = .5),
        plot.subtitle = element_text(hjust = .5),
        text = element_text(size = rel(3.5))
      )
    )

}


## -----------------------------------------------------------------------------
#| message: false
#| warning: false
#| code-fold: true
#| code-summary: "Set-up"

# Load necessary packages
library(duckspatial)
library(bench)
library(dplyr)
library(sf)
library(ggplot2)
options(scipen = 999)

# Function to generate random points
make_points <- function(n_points) {
    points_df <- data.frame(
      id = 1:n_points,
      x = runif(n_points, min = -180, max = 180),
      y = runif(n_points, min = -90, max = 90),
      value = rnorm(n_points, mean = 100, sd = 15),
      category = sample(c("A", "B", "C", "D"), n_points, replace = TRUE)
  ) |>
    sf::st_as_sf(coords = c("x", "y"), crs = 4326)
}

# Generate datasets of different sizes
withr::with_seed(27, {
  points_sf_100k <- make_points(1e5)
  points_sf_1mi  <- make_points(1e6)
  points_sf_3mi  <- make_points(3e6)
})

# Generate polygons
# Create large polygon dataset (e.g., administrative regions, zones, etc.)
n_polygons <- 10000
polygons_list <- vector("list", n_polygons)

for(i in 1:n_polygons) {
  # Random center point with buffer from edges
  center_x <- runif(1, min = -170, max = 170)
  center_y <- runif(1, min = -80, max = 80)
  
  # Create simple rectangular polygons to avoid geometry issues
  width <- runif(1, min = 0.5, max = 3)
  height <- runif(1, min = 0.5, max = 3)
  
  # Create rectangle coordinates (must be closed: first point = last point)
  x_coords <- c(
    center_x - width/2,
    center_x + width/2,
    center_x + width/2,
    center_x - width/2,
    center_x - width/2  # Close the polygon
  )
  
  y_coords <- c(
    center_y - height/2,
    center_y - height/2,
    center_y + height/2,
    center_y + height/2,
    center_y - height/2  # Close the polygon
  )
  
  # Create polygon matrix
  coords <- cbind(x_coords, y_coords)
  
  # Create polygon (wrapped in list as required by st_polygon)
  polygons_list[[i]] <- st_polygon(list(coords))
}

polygons_sf <- st_sf(
  poly_id    = 1:n_polygons,
  region     = sample(c("North", "South", "East", "West"), n_polygons, replace = TRUE),
  population = sample(1000:1000000, n_polygons, replace = TRUE),
  geometry   = st_sfc(polygons_list, crs = 4326)
)


## -----------------------------------------------------------------------------
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_join"

# Helper to run the benchmark
run_join_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_join(points_sf, polygons_sf, join = "within"),
    sf          = st_join(points_sf, polygons_sf, join = st_within)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_join <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_join_benchmark
) |>
  dplyr::bind_rows()


## -----------------------------------------------------------------------------
#| echo: false
#| warning: false

# Id to store the figures, saving the older ones
id_output <- "v1.5.1"

# Generate the plots
gg_join <- ggplot_benchmark(
  data     = df_bench_join,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_join_log <- ggplot_benchmark(
  data     = df_bench_join,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_join,
  plot2 = gg_join_log,
  fun_name = "ddbs_join()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-join-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)


## -----------------------------------------------------------------------------
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_filter"

# Helper to run the benchmark
run_filter_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_filter(points_sf, polygons_sf),
    sf          = st_filter(points_sf, polygons_sf)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_filter <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_filter_benchmark
) |>
  dplyr::bind_rows()


## -----------------------------------------------------------------------------
#| echo: false
#| warning: false

# Generate the plots
gg_filter <- ggplot_benchmark(
  data     = df_bench_filter,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_filter_log <- ggplot_benchmark(
  data     = df_bench_filter,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_filter,
  plot2 = gg_filter_log,
  fun_name = "ddbs_filter()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-filter-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)


## -----------------------------------------------------------------------------
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_distance"

# Helper to run the benchmark
run_distance_benchmark <- function(n) {

  points_sf <- withr::with_seed(27, make_points(n))

  temp <- bench::mark(
    iterations  = 1,
    check       = FALSE,
    duckspatial = ddbs_distance(points_sf, points_sf),
    sf          = st_distance(points_sf, points_sf)
  )
  temp$n   <- n
  temp$pkg <- c("duckspatial", "sf")
  temp
}

df_bench_distance <- lapply(
  X   = c(1000, 5000, 10000),
  FUN = run_distance_benchmark
) |>
  dplyr::bind_rows()


## -----------------------------------------------------------------------------
#| echo: false
#| warning: false

# Generate the plots
gg_distance <- ggplot_benchmark(
  data     = df_bench_distance,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_distance_log <- ggplot_benchmark(
  data     = df_bench_distance,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_distance,
  plot2 = gg_distance_log,
  fun_name = "ddbs_distance()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-distance-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)


## -----------------------------------------------------------------------------
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_union_agg"

# Helper to run the benchmark
run_union_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_union_agg(points_sf, by = "category"),
    sf          = points_sf |> 
      group_by(category) |> 
      summarise(geometry = st_union(geometry))
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_union <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_union_benchmark
) |>
  dplyr::bind_rows()


## -----------------------------------------------------------------------------
#| echo: false
#| warning: false

# Generate the plots
gg_union_agg <- ggplot_benchmark(
  data     = df_bench_union,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_union_agg_log <- ggplot_benchmark(
  data     = df_bench_union,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_union_agg,
  plot2 = gg_union_agg_log,
  fun_name = "ddbs_union_agg()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-dissolve-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)


## -----------------------------------------------------------------------------
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_intersects"

# Helper to run the benchmark
run_predicate_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 1,
    check       = FALSE,
    duckspatial = ddbs_intersects(points_sf, polygons_sf),
    sf          = st_intersects(points_sf, polygons_sf)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_predicate <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_predicate_benchmark
) |>
  dplyr::bind_rows()


## -----------------------------------------------------------------------------
#| echo: false
#| warning: false

# Generate the plots
gg_predicate <- ggplot_benchmark(
  data     = df_bench_predicate,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_predicate_log <- ggplot_benchmark(
  data     = df_bench_predicate,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_predicate,
  plot2 = gg_predicate_log,
  fun_name = "ddbs_intersects()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-intersects-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)

## -----------------------------------------------------------------------------
#| include: false

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

# Quarto spawns a new R subprocess during R CMD build and may not pick up
# R_LIBS set by the build process; propagate it manually.
local({
  r_libs <- Sys.getenv("R_LIBS")
  if (nzchar(r_libs)) {
    paths <- unlist(strsplit(r_libs, .Platform$path.sep))
    .libPaths(unique(c(paths, .libPaths())))
  }
})


## -----------------------------------------------------------------------------
#| label: setup
#| message: false
#| warning: false
library(duckspatial)
library(sf)

# 1. Load Source Data (NC Counties)
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)

# 2. Transform to projected CRS (Albers) for accurate area calculations
nc <- st_transform(nc, 5070)

# 3. Create a Target Grid
grid <- st_make_grid(nc, n = c(10, 5)) |> st_as_sf()

# 4. Create Unique IDs (Required for interpolation)
nc$source_id <- 1:nrow(nc)
grid$target_id <- 1:nrow(grid)


## -----------------------------------------------------------------------------
# Interpolate Total Births (Extensive)
res_extensive <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  mode = "sf"
)


## -----------------------------------------------------------------------------
orig_sum <- sum(nc$BIR74)
new_sum  <- sum(res_extensive$BIR74, na.rm = TRUE)

sprintf("Original: %s | Interpolated: %s", orig_sum, round(new_sum, 1))


## -----------------------------------------------------------------------------
# Interpolate 'BIR74' treating it as an intensive variable (e.g. density assumption)
res_intensive <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  intensive = "BIR74", # Treated as density here
  weight = "sum",      # Standard behavior for intensive vars
  mode = "sf"
)


## ----fig.height=5, fig.width=7------------------------------------------------
# Combine for plotting
plot_data <- res_extensive[, "BIR74"]
names(plot_data)[1] <- "Extensive_Count"
plot_data$Intensive_Value <- res_intensive$BIR74

plot(plot_data[c("Extensive_Count", "Intensive_Value")], 
     main = "Interpolation Methods Comparison",
     border = "grey90",
     key.pos = 4)


## -----------------------------------------------------------------------------
# Return a standard data.frame/tibble without geometry
res_tbl <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74"
) |> 
  ddbs_collect(as = "tibble")

head(res_tbl)


## -----------------------------------------------------------------------------
# Create connection
conn <- ddbs_create_conn()

# Write layers to DuckDB
ddbs_write_vector(conn, nc, "nc_table", overwrite = TRUE)
ddbs_write_vector(conn, grid, "grid_table", overwrite = TRUE)


## -----------------------------------------------------------------------------
# Run interpolation and save to new table 'nc_grid_births'
ddbs_interpolate_aw(
  conn = conn,
  target = "grid_table",
  source = "nc_table",
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  name = "nc_grid_births", # <--- Writes to DB
  overwrite = TRUE
)

# Verify the table was created
DBI::dbListTables(conn)


## -----------------------------------------------------------------------------
# Read the result back from the database
final_sf <- ddbs_read_vector(conn, "nc_grid_births")

head(final_sf)


## -----------------------------------------------------------------------------
ddbs_interpolate_aw(
  conn = conn,
  target = "grid_table",
  source = "nc_table",
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  name = "nc_grid_births", # <--- Writes to DB
  overwrite = TRUE,
  mode = "tibble"
)


## -----------------------------------------------------------------------------
as_duckspatial_df("nc_grid_births", conn)


## -----------------------------------------------------------------------------
duckdb::dbDisconnect(conn)

## -----------------------------------------------------------------------------
#| include: false

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

# Quarto spawns a new R subprocess during R CMD build and may not pick up
# R_LIBS set by the build process; propagate it manually.
local({
  r_libs <- Sys.getenv("R_LIBS")
  if (nzchar(r_libs)) {
    paths <- unlist(strsplit(r_libs, .Platform$path.sep))
    .libPaths(unique(c(paths, .libPaths())))
  }
})


## -----------------------------------------------------------------------------
#| message: false
#| warning: false
library(duckspatial)
library(sf)

# polygons
countries_sf  <- sf::st_read(
    system.file("spatial/countries.geojson",  package = "duckspatial"),
    quiet = TRUE
    )

# random points
set.seed(42)
n <- 10000
points_sf <- data.frame(
  id = 1:n,
  x  = runif(n, min = -180, max = 180),
  y  = runif(n, min =  -90, max =  90)
) |>
  sf::st_as_sf(coords = c("x","y"), crs = 4326)



## ----message=FALSE------------------------------------------------------------
out_sf1 <- ddbs_join(
  x    = points_sf,
  y    = countries_sf,
  join = "within"
)

# quick peek
# mapview(out_sf1, zcol="NAME_ENGL")



## -----------------------------------------------------------------------------
# create a fresh DuckDB connection
conn <- duckspatial::ddbs_create_conn()



## ----message=FALSE------------------------------------------------------------
# write data to DuckDB
ddbs_write_table(conn, points_sf,   "points",    overwrite = TRUE)
ddbs_write_table(conn, countries_sf, "countries", overwrite = TRUE)

# spatial join inside DuckDB; result returned as sf
out_sf2 <- ddbs_join(
  conn,
  x    = "points",
  y    = "countries",
  join = "within"
)


## ----message=FALSE------------------------------------------------------------
ddbs_join(
    conn = conn,
    x = "points",
    y = "countries",
    join = "within",
    name = "points_in_countries",
    overwrite = TRUE
)

# use the result in SQL (or read back as sf later)
# DBI::dbReadTable(conn, "points_in_countries") |>
#     sf::st_as_sf(wkt = 'geometry') |> 
#     head()



## -----------------------------------------------------------------------------
duckdb::dbDisconnect(conn)

#' Checks and installs the Spatial extension
#'
#' Checks if a spatial extension is available, and installs it in a DuckDB database
#'
#' @template conn
#' @param upgrade if TRUE, it upgrades the DuckDB extension to the latest version
#' @template quiet
#' @param extension name of the extension to install, default is "spatial"
#' @param repos optional character string naming the repository to install the
#' extension from (e.g. \code{"core"}, \code{"core_nightly"}, or
#' \code{"community"}); a URL or path can also be supplied. If \code{NULL}
#' (default), the \code{core} repository is tried first, then \code{community}.
#' Switching an already-installed extension to a different repository requires
#' \code{upgrade = TRUE}. See
#' \url{https://duckdb.org/docs/stable/extensions/installing_extensions}.
#'
#' @returns TRUE (invisibly) for successful installation
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(duckdb)
#'
#' # connect to in memory database
#' conn <- duckdb::dbConnect(duckdb::duckdb())
#'
#' # install the spatial extension
#' ddbs_install(conn)
#'
#' # disconnect from db
#' duckdb::dbDisconnect(conn)
#'
#' # install the h3 community extension
#' conn <- duckdb::dbConnect(duckdb::duckdb())
#' ddbs_install(conn, extension = "h3")
#' duckdb::dbDisconnect(conn)
#' }
ddbs_install <- function(
    conn,
    upgrade = FALSE,
    quiet = FALSE,
    extension = "spatial",
    repos = NULL
) {

    # 1. Get extensions list
    ext <- DBI::dbGetQuery(conn, "SELECT * FROM duckdb_extensions();")

    # 2. Checks
    ## 2.1. Check connection
    dbConnCheck(conn)
    if (!is.null(repos)) assert_character_scalar(repos, "repos")
    ## 2.2. Check if it's installed. When already installed and not upgrading,
    ## there is nothing to do; with `upgrade = TRUE` we fall through to a
    ## (FORCE) INSTALL. DuckDB's `duckdb_extensions()` exposes no reliable
    ## "needs upgrade" flag, so we always re-install on upgrade.
    target_ext <- ext[ext$extension_name == extension, ]
    if (nrow(target_ext) == 1 && target_ext$installed && !upgrade) {
        if (isFALSE(quiet)) {
            cli::cli_alert_info(
                "{extension} extension version {.val {target_ext$extension_version}} is already installed. Use {.code upgrade = TRUE} to upgrade."
            )
        }
        return(invisible(TRUE))
    }

    # Extension cannot be upgraded if it's already loaded. It will fail
    if (isTRUE(target_ext$loaded) && isTRUE(upgrade)) {
        cli::cli_abort("{extension} is already loaded in the connection. Upgrading the version is only allowed in non-loaded connections.")
    }

    # 3. Install/upgrade extension
    install_kw <- if (upgrade) "FORCE INSTALL" else "INSTALL"

    if (!is.null(repos)) {

        ## 3a. Explicit repository chosen by the user (no fallback). Named repos
        ## (core, core_nightly, community, ...) are bare identifiers; URLs/paths
        ## must be single-quoted.
        repos_sql <- if (grepl("^[A-Za-z0-9_]+$", repos)) repos else ddbs_quote_sql_string(conn, repos)

        installed <- tryCatch({
            suppressMessages(DBI::dbExecute(conn, glue::glue("{install_kw} {extension} FROM {repos_sql};")))
            repos
        }, error = function(e) {
            cli::cli_abort(c(
                "Failed to {if (upgrade) 'upgrade' else 'install'} the {extension} extension from the {.val {repos}} repository.",
                "x" = conditionMessage(e),
                "i" = "If {extension} is already installed from a different repository, set {.code upgrade = TRUE} to switch.",
                "i" = "Check the repository name: {.url https://duckdb.org/docs/stable/extensions/installing_extensions}"
            ))
        })

    } else {

        ## 3b. Default: try core, then community, then error
        err_core <- err_community <- NULL

        installed <- tryCatch({
            suppressMessages(DBI::dbExecute(conn, glue::glue("{install_kw} {extension};")))
            "core"
        }, error = function(e) {
            err_core <<- e
            tryCatch({
                suppressMessages(DBI::dbExecute(conn, glue::glue("{install_kw} {extension} FROM community;")))
                "community"
            }, error = function(e2) {
                err_community <<- e2
                NULL
            })
        })

        if (is.null(installed)) {
            cli::cli_abort(c(
                "Failed to {if (upgrade) 'upgrade' else 'install'} the {extension} extension.",
                "x" = "core: {conditionMessage(err_core)}",
                "x" = "community: {conditionMessage(err_community)}",
                "i" = "It might not be available for this version of DuckDB, or the install location may not be writable.",
                "i" = "Check that the extension name is correct: {.url https://duckdb.org/docs/extensions/overview}"
            ))
        }
    }

    if (isFALSE(quiet)) {
        action <- if (upgrade) "upgraded" else "installed"
        repo_note <- if (!identical(installed, "core")) glue::glue(" (from {installed} repository)") else ""
        cli::cli_alert_success("{extension} extension {action}{repo_note}")
    }

    return(invisible(TRUE))
}


#' Loads the Spatial extension
#'
#' Checks if a spatial extension is installed, and loads it in a DuckDB database
#'
#' @template conn
#' @template quiet
#' @param extension name of the extension to load, default is "spatial"
#' @param create_macros if TRUE (default), it creates macros that allow
#'   some functions to be used within dplyr pipelines
#'
#' @returns TRUE (invisibly) for successful installation
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(duckdb)
#'
#' ## connect to in memory database
#' conn <- duckdb::dbConnect(duckdb::duckdb())
#'
#' ## install the spatial exntesion
#' ddbs_install(conn)
#' ddbs_load(conn)
#'
#' ## disconnect from db
#' duckdb::dbDisconnect(conn)
#' }
ddbs_load <- function(
    conn, 
    quiet = FALSE, 
    extension = "spatial",
    create_macros = TRUE
) {

    # 1. Checks

    ## 1.1. Get extensions list
    ext <- DBI::dbGetQuery(conn, "SELECT * FROM duckdb_extensions();")

    ## 1.2. Check connection
    dbConnCheck(conn)

    ## 1.3. Check if extension is installed or already loaded (bundled extensions
    ##      may show installed = FALSE but are still loadable/loaded)
    target_ext <- ext[ext$extension_name == extension, ]
    if (!isTRUE(target_ext$installed) && !isTRUE(target_ext$loaded))
        cli::cli_abort("{extension} extension is not installed, please use `ddbs_install(extension = '{extension}')`")

    
    # 2. Setup extension

    ## 2.1. Load the extension
    if (isFALSE(target_ext$loaded)) {
        tryCatch({
            suppressMessages(DBI::dbExecute(conn, glue::glue("LOAD {extension};")))
        }, error = function(e) {
            cli::cli_abort(c(
                "Failed to load the {extension} extension.",
                "x" = conditionMessage(e),
                "i" = "The binary may be corrupt or incompatible with this platform/DuckDB version.",
                "i" = "Try reinstalling with {.code ddbs_install(extension = '{extension}', upgrade = TRUE)}"
            ))
        })
    }

    ## 2.2. Activate macros
    if (isTRUE(create_macros)) create_duckspatial_macros(conn)

    ## 2.3. Message
    if (isFALSE(quiet)) {
        cli::cli_alert_success("{extension} extension loaded")
    }

}



#' Glimpse the status of a DuckDB extension
#'
#' Retrieves the row from DuckDB's \code{duckdb_extensions()} catalog for a given
#' extension (the spatial extension by default) and prints a transposed
#' \code{\link[dplyr]{glimpse}} of it: whether it is installed and loaded, its
#' version, install path, description, and so on.
#'
#' @template conn_null
#' @param extension name of the extension to inspect, default is "spatial"
#'
#' @returns A one-row \code{tibble} with the extension's metadata (invisibly).
#' Called mainly for the glimpse printed as a side effect.
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # inspect the spatial extension on the default connection
#' ddbs_extension_info()
#'
#' # or pass an explicit connection
#' conn <- ddbs_create_conn()
#' ddbs_extension_info(conn)
#' ddbs_stop_conn(conn)
#' }
ddbs_extension_info <- function(conn = NULL, extension = "spatial") {

    # 1. Resolve and validate inputs
    conn <- conn %||% ddbs_default_conn()
    dbConnCheck(conn)
    assert_character_scalar(extension, "extension")

    # 2. Query the extension catalog
    info <- DBI::dbGetQuery(
        conn,
        "SELECT * FROM duckdb_extensions() WHERE extension_name = ?",
        params = list(extension)
    )

    if (nrow(info) == 0) {
        cli::cli_abort(
            "Extension {.val {extension}} was not found in {.fn duckdb_extensions}."
        )
    }

    # 3. Glimpse and return the row invisibly
    dplyr::glimpse(tibble::as_tibble(info))

}

test_that("get_parquet_crs handles column names with dots and special characters", {
  skip_if_not_installed("arrow")
  skip_if_not_installed("duckdb")
  skip_if_not_installed("sf")

  # Create a tiny Parquet file with a dot in geometry name
  df <- data.frame(id = 1L)
  wkb <- as.raw(c(0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))
  df[["geom.1"]] <- I(list(wkb))
  
  tbl <- arrow::Table$create(df)
  
  # Fetch full PROJJSON from sf to ensure it is valid for DuckDB
  projjson_str <- sf::st_crs(3857)$projjson
  
  # Fallback to a valid complete PROJJSON for 3857 if sf returns NULL
  if (is.null(projjson_str)) {
      projjson_str <- projjson_fallback_3857
  }
  
  geo_meta <- list(
    version = "1.1.0",
    primary_column = "geom.1",
    columns = list(
      "geom.1" = list(
        encoding = "WKB",
        geometry_types = list("Point"),
        crs = jsonlite::fromJSON(projjson_str, simplifyVector = FALSE)
      )
    )
  )
  
  tbl$metadata[["geo"]] <- jsonlite::toJSON(geo_meta, auto_unbox = TRUE)
  tmp_pq <- tempfile(fileext = ".parquet")
  arrow::write_parquet(tbl, tmp_pq)
  on.exit(unlink(tmp_pq), add = TRUE)
  
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  on.exit(DBI::dbDisconnect(conn, shutdown = TRUE), add = TRUE)
  duckspatial::ddbs_install(conn, extension = "json", quiet = TRUE)
  duckspatial::ddbs_load(conn, extension = "json", quiet = TRUE)
  
  crs <- duckspatial:::get_parquet_crs(tmp_pq, conn)
  expect_false(is.null(crs))
  expect_equal(crs$epsg, 3857)
})

test_that("get_parquet_crs handles column names with spaces", {
  skip_if_not_installed("arrow")
  skip_if_not_installed("duckdb")
  skip_if_not_installed("sf")

  df <- data.frame(id = 1L)
  wkb <- as.raw(c(0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))
  df[["my geom"]] <- I(list(wkb))
  
  tbl <- arrow::Table$create(df)
  
  projjson_str <- sf::st_crs(4326)$projjson
  if (is.null(projjson_str)) {
      # Fallback to a valid complete PROJJSON for 4326
      projjson_str <- projjson_fallback_4326
  }
  
  geo_meta <- list(
    version = "1.1.0",
    primary_column = "my geom",
    columns = list(
      "my geom" = list(
        encoding = "WKB",
        geometry_types = list("Point"),
        crs = jsonlite::fromJSON(projjson_str, simplifyVector = FALSE)
      )
    )
  )
  
  tbl$metadata[["geo"]] <- jsonlite::toJSON(geo_meta, auto_unbox = TRUE)
  tmp_pq <- tempfile(fileext = ".parquet")
  arrow::write_parquet(tbl, tmp_pq)
  on.exit(unlink(tmp_pq), add = TRUE)
  
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  on.exit(DBI::dbDisconnect(conn, shutdown = TRUE), add = TRUE)
  duckspatial::ddbs_install(conn, extension = "json", quiet = TRUE)
  duckspatial::ddbs_load(conn, extension = "json", quiet = TRUE)
  
  crs <- duckspatial:::get_parquet_crs(tmp_pq, conn)
  expect_false(is.null(crs))
  expect_equal(crs$epsg, 4326)
})

# 0. Set up --------------------------------------------------------------

skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")


# 1. ddbs_install() ------------------------------------------------------

describe("ddbs_install()", {

  describe("expected behavior", {

    it("returns TRUE invisibly on install", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      result <- ddbs_install(conn, quiet = TRUE)
      expect_true(result)
    })

    it("returns TRUE invisibly when already installed", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      result <- ddbs_install(conn, quiet = TRUE)
      expect_true(result)
    })

    it("shows a message with quiet = FALSE", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      expect_message(ddbs_install(conn, quiet = FALSE))
    })

    it("shows already-installed message on repeated calls with quiet = FALSE", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      expect_message(ddbs_install(conn, quiet = FALSE), "already installed")
    })

    it("suppresses all messages with quiet = TRUE", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      expect_no_message(ddbs_install(conn, quiet = TRUE))
      expect_no_message(ddbs_install(conn, quiet = TRUE))
    })
  })

  describe("errors", {

    it("errors when extension cannot be found in any repository", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      expect_error(
        ddbs_install(conn, extension = "nonexistent_extension_xyz"),
        "Failed to install"
      )
    })

    it("errors when trying to upgrade an already-loaded extension", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      ddbs_load(conn, quiet = TRUE, create_macros = FALSE)
      expect_error(
        ddbs_install(conn, upgrade = TRUE),
        "already loaded"
      )
    })
  })

  describe("repos argument", {

    it("validates the repos type", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      expect_error(ddbs_install(conn, repos = 123))
    })

    it("installs from an explicit core repository", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ## A forced re-install does real file I/O (download + move the extension
      ## binary), which the environment can block: no network, read-only
      ## filesystem, or locked/in-use files on Windows CI. Skip in those cases.
      res <- tryCatch(
        ddbs_install(conn, extension = "spatial", repos = "core", upgrade = TRUE, quiet = TRUE),
        error = function(e) testthat::skip(paste("could not (re)install extension:", conditionMessage(e)))
      )
      expect_true(res)
    })

    it("errors clearly on an unknown repository name", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      expect_error(
        ddbs_install(conn, extension = "spatial", repos = "not_a_real_repo_xyz", upgrade = TRUE)
      )
    })
  })
})


# 2. ddbs_load() ---------------------------------------------------------

describe("ddbs_load()", {

  describe("expected behavior", {

    it("loads the spatial extension without error", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      expect_no_error(ddbs_load(conn, quiet = TRUE, create_macros = FALSE))
    })

    it("shows success message with quiet = FALSE", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      expect_message(ddbs_load(conn, quiet = FALSE, create_macros = FALSE), "loaded")
    })

    it("suppresses messages with quiet = TRUE", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      expect_no_message(ddbs_load(conn, quiet = TRUE, create_macros = FALSE))
    })

    it("can be called again on an already-loaded extension", {
      conn <- duckdb::dbConnect(duckdb::duckdb())
      on.exit(duckdb::dbDisconnect(conn), add = TRUE)
      ddbs_install(conn, quiet = TRUE)
      ddbs_load(conn, quiet = TRUE, create_macros = FALSE)
      expect_no_error(ddbs_load(conn, quiet = TRUE, create_macros = FALSE))
    })
  })


})


# 3. Other extensions ----------------------------------------------------

## Community extensions
testthat::test_that("Community extensions are installed", {
  conn <- duckdb::dbConnect(duckdb::duckdb())
  out <- tryCatch(ddbs_install(conn, extension = "h3"), error = function(e) NULL)
  if (is.null(out)) {
    testthat::pass()
  } else {
    expect_no_error(ddbs_load(conn, quiet = TRUE, create_macros = FALSE))
  }
})

## Other core extensions
testthat::test_that("Community extensions are installed", {
  conn <- duckdb::dbConnect(duckdb::duckdb())
  out <- tryCatch(ddbs_install(conn, extension = "aws"), error = function(e) NULL)
  if (is.null(out)) {
    testthat::pass()
  } else {
    expect_no_error(ddbs_load(conn, quiet = TRUE, create_macros = FALSE))
  }
})


# 4. ddbs_extension_info() -----------------------------------------------

describe("ddbs_extension_info()", {

  it("returns the extension row as a one-row tibble, invisibly", {
    conn <- ddbs_create_conn()
    on.exit(ddbs_stop_conn(conn), add = TRUE)

    res <- withVisible(ddbs_extension_info(conn))
    expect_false(res$visible)
    expect_s3_class(res$value, "tbl_df")
    expect_equal(nrow(res$value), 1L)
    expect_equal(res$value$extension_name, "spatial")
    expect_true("loaded" %in% names(res$value))
  })

  it("uses the default connection when conn = NULL", {
    out <- ddbs_extension_info()
    expect_s3_class(out, "tbl_df")
    expect_equal(out$extension_name, "spatial")
  })

  it("errors on an unknown extension", {
    conn <- duckdb::dbConnect(duckdb::duckdb())
    on.exit(duckdb::dbDisconnect(conn), add = TRUE)
    expect_error(ddbs_extension_info(conn, extension = "not_a_real_extension_xyz"))
  })

  it("validates the extension argument", {
    conn <- duckdb::dbConnect(duckdb::duckdb())
    on.exit(duckdb::dbDisconnect(conn), add = TRUE)
    expect_error(ddbs_extension_info(conn, extension = 1))
  })
})

testthat::skip_on_cran()

test_that("GeoPackage output automatically renames 'FID' to 'FID_original'", {
  conn <- ddbs_temp_conn()
  
  # Create a simple sf object with an 'FID' column
  df <- data.frame(
    FID = 1:5,
    name = letters[1:5],
    lat = 0,
    lon = 0
  )
  sf_obj <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  
  # Test 1: Writing local sf object
  # Should trigger warning and rename
  tmp_local <- tempfile(fileext = ".gpkg")
  on.exit(unlink(tmp_local), add = TRUE)
  
  expect_message(
    ddbs_write_dataset(sf_obj, tmp_local, quiet = TRUE),
    "Column 'FID' renamed to 'FID_original'"
  )
  
  # Verify local write result has 'FID_original'
  ds_local <- ddbs_open_dataset(tmp_local, conn = conn)
  cols_local <- glue::glue("DESCRIBE SELECT * FROM {attr(ds_local, 'source_table')}") |> 
    DBI::dbGetQuery(conn, statement = _)
  
  # Print names for debugging if needed
  # print(cols_local$column_name)
  
  expect_true("FID_original" %in% cols_local$column_name)
  # GDAL primary key might be 'fid', 'geom', etc. We mainly care that our data is preserved.
  
  # Test 2: Writing remote (lazy) object
  # First create a table in DuckDB with FID
  DBI::dbWriteTable(conn, "test_fid_source", df, overwrite = TRUE)
  ddbs_load(conn) # Ensure spatial
  DBI::dbExecute(conn, "
    CREATE OR REPLACE TABLE test_fid_spatial AS 
    SELECT *, ST_Point(lon, lat) as geom 
    FROM test_fid_source
  ")
  
  remote_tbl <- as_duckspatial_df("test_fid_spatial", conn)
  
  tmp_remote <- tempfile(fileext = ".gpkg")
  on.exit(unlink(tmp_remote), add = TRUE)
  
  # Provide CRS explicitly to avoid warnings and ensure correct metadata
  expect_message(
    ddbs_write_dataset(remote_tbl, tmp_remote, crs = "EPSG:4326", quiet = TRUE),
    "Column 'FID' renamed to 'FID_original'"
  )
  
  # Verify remote write result
  ds_remote <- ddbs_open_dataset(tmp_remote, conn = conn)
  cols_remote <- glue::glue("DESCRIBE SELECT * FROM {attr(ds_remote, 'source_table')}") |> 
    DBI::dbGetQuery(conn, statement = _)
    
  expect_true("FID_original" %in% cols_remote$column_name)
})

test_that("No warning for non-GeoPackage formats with FID column", {
  conn <- ddbs_temp_conn()
  
  df <- data.frame(FID = 1:5, lat = 0, lon = 0)
  sf_obj <- sf::st_as_sf(df, coords = c("lon", "lat"), crs = 4326)
  
  # GeoJSON shouldn't care about FID column
  tmp_json <- tempfile(fileext = ".geojson")
  on.exit(unlink(tmp_json), add = TRUE)
  
  expect_no_message(
    ddbs_write_dataset(sf_obj, tmp_json, quiet = TRUE)
  )
  expect_true(file.exists(tmp_json))
})

testthat::skip_on_cran()

test_that("DuckDB auto-detects spatial formats", {
  skip_if_not_installed("duckdb")
  
  # Setup connection
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)
  
  # Use existing GeoJSON
  gj_path <- system.file("spatial/countries.geojson", package = "duckspatial")
  expect_true(file.exists(gj_path))
  
  # 1. Test basic auto-detection with extension
  res_gj <- duckspatial::ddbs_open_dataset(gj_path, conn = conn)
  expect_true(inherits(res_gj, "duckspatial_df"))
  expect_gt(as.numeric(dplyr::count(res_gj) |> dplyr::collect()), 0)
  
  # 2. Test auto-detection with GPKG
  # Create a small GPKG from scratch to avoid schema issues
  tmp_gpkg <- tempfile(fileext = ".gpkg")
  
  # Simple polygon
  sf_obj <- sf::st_as_sf(data.frame(id = 1, geom = sf::st_sfc(sf::st_point(c(0,0)))), crs = 4326)
  sf::st_write(sf_obj, tmp_gpkg, quiet = TRUE)
  
  res_gpkg <- duckspatial::ddbs_open_dataset(tmp_gpkg, conn = conn)
  expect_true(inherits(res_gpkg, "duckspatial_df"))
  
  # 3. Test auto-detection WITHOUT extension (rename file)
  tmp_no_ext <- tempfile() # no extension
  file.copy(tmp_gpkg, tmp_no_ext)
  
  # This is the critical test: can GDAL/DuckDB detect format without extension?
  res_no_ext <- duckspatial::ddbs_open_dataset(tmp_no_ext, conn = conn)
  expect_true(inherits(res_no_ext, "duckspatial_df"))
  
  # 4. Test Parquet auto-detection without extension (the new path)
  tmp_pq <- tempfile(fileext = ".parquet")
  # Simple dataframe - ddbs_open_dataset just needs to be able to open it
  # We add a dummy geometry column to satisfy potential column checks if any (though lazy)
  df_pq <- data.frame(a = 1L)
  # Binary column simulating geometryWKB (just raw bytes)
  df_pq$geometry <- I(list(as.raw(c(0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00))))
  arrow::write_parquet(df_pq, tmp_pq)
  
  tmp_pq_no_ext <- tempfile()
  file.copy(tmp_pq, tmp_pq_no_ext)
  
  res_pq_no_ext <- duckspatial::ddbs_open_dataset(tmp_pq_no_ext, conn = conn)
  expect_true(inherits(res_pq_no_ext, "duckspatial_df"))
  
})

test_that("ddbs_open_dataset gives clear error for unsupported file types", {
  skip_if_not_installed("duckdb")
  
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)
  
  # Create an unsupported file (plain text)
  tmp_garbage <- tempfile(fileext = ".txt")
  writeLines("This is not spatial data", tmp_garbage)
  
  # Expect an informative error and NO CRS warnings
  expect_warning(
    expect_error(
      duckspatial::ddbs_open_dataset(tmp_garbage, conn = conn),
      regexp = "not recognized|not supported|failed",
      ignore.case = TRUE
    ),
    NA # NA means "no warnings"
  )
  
  # Test with no extension garbage file
  tmp_garbage_no_ext <- tempfile()
  writeLines("Random garbage content", tmp_garbage_no_ext)
  
  expect_warning(
    expect_error(
      duckspatial::ddbs_open_dataset(tmp_garbage_no_ext, conn = conn),
      regexp = "not recognized|not supported|failed|format",
      ignore.case = TRUE
    ),
    NA
  )
})

test_that("ddbs_open_dataset warns when file has no CRS", {
  skip_if_not_installed("duckdb")
  
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)

  # Create a shapefile without CRS
  tmp_dir <- tempdir()
  tmp_shp <- file.path(tmp_dir, "no_crs_autodetect_test.shp")
  
  # Create points without CRS
  pts <- sf::st_as_sf(
    data.frame(x = 1:3, y = 1:3, id = 1:3),
    coords = c("x", "y"),
    crs = NA
  )
  suppressWarnings(sf::st_write(pts, tmp_shp, quiet = TRUE, delete_dsn = TRUE))
  on.exit(unlink(list.files(tmp_dir, pattern = "no_crs_autodetect_test", full.names = TRUE)), add = TRUE)

  # Should warn about missing CRS
  expect_warning(
    ds <- duckspatial::ddbs_open_dataset(tmp_shp, conn = conn),
    regexp = "Could not auto-detect CRS",
    ignore.case = TRUE
  )

  # CRS should be NA
  expect_true(is.na(sf::st_crs(attr(ds, "crs"))$epsg))
})

test_that("ddbs_open_dataset ignores non-spatial columns named 'geometry'", {
  skip_if_not_installed("duckdb")
  
  conn <- tryCatch(ddbs_default_conn(), error = function(e) DBI::dbConnect(duckdb::duckdb()))
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)

  # Create a CSV with a column named "geometry" that is just text
  tmp_csv <- tempfile(fileext = ".csv")
  writeLines("id,geometry,value\n1,this is not a geometry,10", tmp_csv)
  on.exit(unlink(tmp_csv), add = TRUE)

  # Should open as a regular table, not a duckspatial_df
  ds <- duckspatial::ddbs_open_dataset(tmp_csv, conn = conn)
  
  expect_false(inherits(ds, "duckspatial_df"))
  expect_true(inherits(ds, "tbl_lazy"))
})

#' Get the boundary of geometries
#'
#' Returns the boundary of geometries as a new geometry, e.g., the edges of polygons 
#' or the start/end points of lines.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' # read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' # store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' # boundary
#' b <- ddbs_boundary(x = "argentina", conn)
#' }
ddbs_boundary <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Boundary",
        other_args = NULL
    )
    
}





#' Get the envelope (bounding box) of geometries
#'
#' Returns the minimum axis-aligned rectangle that fully contains the geometry.
#'
#' @template x
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' ST_Envelope returns the minimum bounding rectangle (MBR) of a geometry as a
#' polygon. For points and lines, this creates a rectangular polygon that
#' encompasses the geometry. For polygons, it returns the smallest rectangle
#' that contains the entire polygon.
#'
#' When \code{by_feature = FALSE}, all geometries are combined and a single envelope
#' is returned that encompasses the entire dataset.
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' # input as sf, and output as sf
#' env <- ddbs_envelope(x = argentina_ddbs, by_feature = TRUE)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' # store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' # envelope for each feature
#' env <- ddbs_envelope("argentina", conn, by_feature = TRUE)
#'
#' # single envelope for entire dataset
#' env_all <- ddbs_envelope("argentina", conn, by_feature = FALSE)
#'
#' # create a new table with envelopes
#' ddbs_envelope("argentina", conn, name = "argentina_bbox", by_feature = TRUE)
#' }
ddbs_envelope <- function(
    x,
    by_feature = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {


    # 0. Validate inputs
    assert_xy(x, "x")
    assert_logic(by_feature, "by_feature")
    assert_name(name)
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")


    # 1. Prepare inputs
  
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 2.2. Get names of the rest of the columns, or empty string if by_feature = FALSE
    x_rest <- if (isTRUE(by_feature)) {
        get_geom_name(target_conn, x_list$query_name, rest = TRUE, collapse = TRUE)
    } else {
        ""
    }

    ## 2.3. Build envelope clause based on by_feature
    if (isTRUE(by_feature)) {
        st_envelope_clause <- glue::glue("ST_Envelope({x_geom})")
    } else {
        st_envelope_clause <- glue::glue("ST_Envelope_Agg({x_geom})")
    }

    ## 2.4. Build the base query (depends on the output type - sf, duckspatial_df, table)
    base.query <- glue::glue("
        SELECT 
            {x_rest}
            {build_geom_query(st_envelope_clause, name, crs_x, mode)} as {x_geom}
        FROM 
            {x_list$query_name};
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query      = base.query,
            conn       = target_conn,
            mode       = mode,
            crs        = crs_x,
            x_geom     = x_geom
        )
    }

}





#' Get the bounding box of geometries
#'
#' Returns the minimal rectangle that encloses the geometry
#'
#' @template x
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @returns 
#' A `bbox` numeric vector with `by_feature = FALSE` 
#' A `data.frame` or `lazy tbl` when `by_feature = TRUE`
#' 
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' # option 1: passing sf objects
#' ddbs_bbox(argentina_ddbs)
#'
#' ## option 2: passing the names of tables in a duckdb db
#'
#' # creates a duckdb write sf to it
#' conn <- duckspatial::ddbs_create_conn()
#' ddbs_write_table(conn, argentina_ddbs, "argentina_tbl", overwrite = TRUE)
#'
#' output2 <- ddbs_bbox(
#'     conn = conn,
#'     x = "argentina_tbl",
#'     name = "argentina_bbox"
#' )
#'
#' DBI::dbReadTable(conn, "argentina_bbox")
#' }
ddbs_bbox <- function(
    x,
    by_feature = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_logic(by_feature, "by_feature")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
    assert_connflict(conn, xy = x, ref = "x")


    # 1. Prepare inputs
  
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 2.3 Build base query - set the extent_clause
    if (isTRUE(by_feature)) {
        st_extent_clause <- glue::glue("ST_Extent({x_geom})")
    } else {
        st_extent_clause <- glue::glue("ST_Extent_Agg({x_geom})")
    }

    base.query <- glue::glue("
        SELECT
            ST_XMin(ext) AS xmin,
            ST_YMin(ext) AS ymin,
            ST_XMax(ext) AS xmax,
            ST_YMax(ext) AS ymax
        FROM (
            SELECT {st_extent_clause} AS ext
            FROM {x_list$query_name}
        );"
    )


    # 3. Table creation if name is provided
    if (!is.null(name)) {
        return(create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        ))
    }


    # 4. Apply geospatial operation based on mode
    if (mode == "sf" | isFALSE(by_feature)) {
      
        if (isTRUE(by_feature)) {
          
            return(DBI::dbGetQuery(target_conn, base.query))
          
        } else {
          
            ## Get data as a data frame
            data_tbl <- DBI::dbGetQuery(target_conn, base.query)
        
            ## Convert to sf bbox class
            bbox_vec <- structure(
                unlist(data_tbl),
                names = c("xmin", "ymin", "xmax", "ymax"),
                class = "bbox",
                crs   = crs_x
            )      
            return(bbox_vec)
          
        }
      
    } else {

        ## Generate the query
        view_name <- ddbs_temp_table_name()
        tmp.query <- glue::glue("CREATE TEMP TABLE {view_name} AS {base.query}")

        ## Create a table, and return a pointer to that table
        DBI::dbExecute(target_conn, tmp.query)
        data_tbl <- dplyr::tbl(target_conn, view_name)
        return(data_tbl)

    }
    
}



#' Computes the minimum rotated rectangle enclosing a geometry
#'
#' Returns the smallest rectangle that fully contains the input geometry.
#' Unlike [ddbs_envelope()], the rectangle is not constrained to be axis-aligned
#' and may be rotated to minimize its area.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## without a connection
#' ddbs_minimum_rotated_rectangle(argentina_ddbs)
#' }
ddbs_minimum_rotated_rectangle <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_MinimumRotatedRectangle",
        other_args = NULL
    )

}





#' Create a rectangular polygon from bounding coordinates
#'
#' Creates a rectangular POLYGON geometry from four bounding coordinates.
#'
#' @param xmin A numeric value for the minimum X (longitude) coordinate.
#' @param ymin A numeric value for the minimum Y (latitude) coordinate.
#' @param xmax A numeric value for the maximum X (longitude) coordinate.
#' @param ymax A numeric value for the maximum Y (latitude) coordinate.
#' @param crs A character string specifying the coordinate reference system
#'   of the output geometry. Default is \code{"EPSG:4326"}.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## without storing in duckdb
#' finland_bbox_ddbs <- ddbs_make_envelope(
#'   xmin = 19.1, ymin = 59.7,
#'   xmax = 31.6, ymax = 70.1
#' )
#' }
#' @export
ddbs_make_envelope <- function(
  xmin,
  ymin,  
  xmax,
  ymax,
  crs = "EPSG:4326",
  name = NULL,
  conn = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
) {

  # 0. Validate inputs
  assert_numeric(xmin, "xmin")
  assert_numeric(xmax, "xmax")
  assert_numeric(ymin, "ymin")
  assert_numeric(ymax, "ymax")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


  # 1. Prepare inputs

  ## 1.1. Assign attributes
  mode   <- get_mode(mode, name)
  x_geom <- "geometry"

  ## 1.2. Establish connection
  if (is.null(conn)) {
    target_conn <- ddbs_default_conn()
  } else {
    target_conn <- conn
  }
  
  ## 1.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
  st_function <- glue::glue("ST_MakeEnvelope({xmin}, {ymin}, {xmax}, {ymax})")
  base.query <- glue::glue("
    SELECT {build_geom_query(st_function, name, crs, mode)} AS {x_geom};
  ")


  # 2. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs,
      x_geom = x_geom
    )
  }

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write data
duckspatial::ddbs_write_table(conn_test, countries_sf, "countries")


# 1. ddbs_boundary() -------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting a table works
## - CHECK 1.5: geometry type should be line
## - CHECK 2.1: general errors
describe("ddbs_boundary()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on all input formats", {
      output_ddbs <- ddbs_boundary(countries_ddbs)
      output_sf   <- ddbs_boundary(countries_sf)
      output_conn <- ddbs_boundary("countries", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(nrow(ddbs_collect(output_ddbs)), nrow(countries_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats", {
      output_sf_fmt <- ddbs_boundary(countries_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_boundary(countries_ddbs))
      expect_message(ddbs_boundary("countries", conn = conn_test, name = "boundary"))
      expect_message(ddbs_boundary("countries", conn = conn_test, name = "boundary", overwrite = TRUE))
      expect_true(ddbs_boundary("countries", conn = conn_test, name = "boundary2"))

      expect_no_message(ddbs_boundary(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_boundary("countries", conn = conn_test, name = "boundary", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "boundary")
      expect_equal(
        ddbs_collect(ddbs_boundary(countries_ddbs))$geometry,
        output_tbl$geometry
      )
    })

    it("produces LINESTRING / MULTILINESTRING geometry", {
      geom_type <- sf::st_geometry_type(ddbs_collect(ddbs_boundary(countries_ddbs))) |> as.character()
      expect_in(geom_type, c("LINESTRING", "MULTILINESTRING"))
    })
  })

  ### ERRORS

  describe("errors", {

    it("requires a valid connection when using table name", {
      expect_error(ddbs_boundary("countries", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_boundary(x = 999))
      expect_error(ddbs_boundary(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_boundary(countries_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_boundary(countries_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_boundary(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_boundary(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_boundary(countries_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 2. ddbs_envelope() -------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting a table works
## - CHECK 1.5: geometry type should be polygon
## - CHECK 1.6: by_feature works as expected
## - CHECK 1.7: extent should be the same as input
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_envelope()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on all input formats", {
      output_ddbs <- ddbs_envelope(countries_ddbs)
      output_sf   <- ddbs_envelope(countries_sf)
      output_conn <- ddbs_envelope("countries", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats", {
      output_sf_fmt <- ddbs_envelope(countries_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_envelope(countries_ddbs))
      expect_message(ddbs_envelope("countries", conn = conn_test, name = "envelope"))
      expect_message(ddbs_envelope("countries", conn = conn_test, name = "envelope", overwrite = TRUE))
      expect_true(ddbs_envelope("countries", conn = conn_test, name = "envelope2"))

      expect_no_message(ddbs_envelope(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_envelope("countries", conn = conn_test, name = "envelope", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "envelope")
      expect_equal(
        ddbs_collect(ddbs_envelope(countries_ddbs))$geometry,
        output_tbl$geometry
      )
    })

    it("produces POLYGON / MULTIPOLYGON geometry", {
      geom_type <- sf::st_geometry_type(ddbs_collect(ddbs_envelope(countries_ddbs))) |> as.character()
      expect_in(geom_type, c("POLYGON", "MULTIPOLYGON"))
    })

    it("respects by_feature argument", {
      bf_false <- ddbs_envelope(countries_sf, by_feature = FALSE) |> ddbs_collect()
      bf_true  <- ddbs_envelope(countries_sf, by_feature = TRUE) |> ddbs_collect()

      expect_equal(nrow(bf_false), 1)
      expect_equal(nrow(bf_true), nrow(countries_sf))
    })

    it("produces the same extent as input", {
      output <- ddbs_envelope(countries_ddbs)
      extent_output <- ddbs_bbox(output)
      extent_input  <- ddbs_bbox(countries_ddbs)

      expect_equal(extent_output, extent_input)
    })

  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("validates by_feature argument", {
      expect_error(ddbs_envelope(countries_ddbs, by_feature = "TRUE"))
    })

    it("requires a valid connection when using table name", {
      expect_error(ddbs_envelope("countries", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_envelope(x = 999))
      expect_error(ddbs_envelope(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_envelope(countries_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_envelope(countries_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_envelope(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_envelope(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_envelope(countries_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 3. ddbs_bbox() ---------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: writting a table works
## - CHECK 1.4: by_feature works as expected
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_bbox()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on all input formats", {
      output_ddbs <- ddbs_bbox(countries_ddbs)
      output_sf   <- ddbs_bbox(countries_sf)
      output_conn <- ddbs_bbox("countries", conn = conn_test)

      expect_s3_class(output_ddbs, "bbox")
      expect_equal(output_ddbs, output_sf)
      expect_equal(output_ddbs, output_conn)
    })

    it("all output formats work", {
      output_bbox    <- ddbs_bbox(countries_ddbs)
      output_tbl_db  <- ddbs_bbox(countries_ddbs, by_feature = TRUE)
      output_bbox_sf <- ddbs_bbox(countries_ddbs, mode = "sf")
      output_tbl_sf  <- ddbs_bbox(countries_ddbs, by_feature = TRUE, mode = "sf")
      
      expect_s3_class(output_bbox, "bbox")
      expect_s3_class(output_tbl_db, "tbl_duckdb_connection")
      expect_s3_class(output_bbox_sf, "bbox")
      expect_s3_class(output_tbl_sf, "data.frame")

    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_bbox(countries_ddbs))
      expect_message(ddbs_bbox("countries", conn = conn_test, name = "bbox"))
      expect_message(ddbs_bbox("countries", conn = conn_test, name = "bbox", overwrite = TRUE))
      expect_true(ddbs_bbox("countries", conn = conn_test, name = "bbox2"))

      expect_no_message(ddbs_bbox(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_bbox("countries", conn = conn_test, name = "bbox", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- DBI::dbReadTable(conn_test, "bbox")
      expect_equal(
        ddbs_bbox(countries_ddbs, mode = "sf") |> as.numeric(), 
        as.numeric(output_tbl)
      )
    })

    it("same results as sf package", {
      output_ddbs <- ddbs_bbox(countries_ddbs)
      output_sf <- ddbs_bbox(countries_sf)

      expect_equal(output_ddbs, output_sf)
    })

  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("validates by_feature argument", {
      expect_error(ddbs_bbox(countries_ddbs, by_feature = "TRUE"))
    })

    it("requires a valid connection when using table name", {
      expect_error(ddbs_bbox("countries", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_bbox(x = 999))
      expect_error(ddbs_bbox(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_bbox(countries_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_bbox(countries_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_bbox(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_bbox(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_bbox(countries_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 4. ddbs_make_envelope() ------------------------------------------------

## - CHECK 1.1: returns a single-row duckspatial_df
## - CHECK 1.2: returns different output formats
## - CHECK 1.3: messages work
## - CHECK 1.4: writing a table works
## - CHECK 1.5: produced geometry has the expected bbox
## - CHECK 1.6: geometry type is POLYGON
## - CHECK 2.1: validates coordinate arguments
## - CHECK 2.2: general errors

describe("ddbs_make_envelope()", {

  describe("expected behavior", {

    it("returns a single-row duckspatial_df", {
      output <- ddbs_make_envelope(0, 0, 10, 10)
      expect_s3_class(output, "duckspatial_df")
      expect_equal(nrow(ddbs_collect(output)), 1L)
    })

    it("returns different output formats", {
      output_sf_fmt <- ddbs_make_envelope(0, 0, 10, 10, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_make_envelope(0, 0, 10, 10))
      expect_message(ddbs_make_envelope(0, 0, 10, 10, conn = conn_test, name = "make_envelope"))
      expect_message(ddbs_make_envelope(0, 0, 10, 10, conn = conn_test, name = "make_envelope", overwrite = TRUE))
      expect_true(ddbs_make_envelope(0, 0, 10, 10, conn = conn_test, name = "make_envelope2"))

      expect_no_message(ddbs_make_envelope(0, 0, 10, 10, quiet = TRUE))
      expect_no_message(ddbs_make_envelope(0, 0, 10, 10, conn = conn_test, name = "make_envelope", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "make_envelope")
      expect_equal(
        ddbs_collect(ddbs_make_envelope(0, 0, 10, 10))$geometry,
        output_tbl$geometry
      )
    })

    it("produces a POLYGON geometry", {
      geom_type <- sf::st_geometry_type(ddbs_collect(ddbs_make_envelope(0, 0, 10, 10))) |> as.character()
      expect_equal(geom_type, "POLYGON")
    })

    it("bbox of result matches the input coordinates", {
      bbox_out <- ddbs_bbox(ddbs_make_envelope(-10, -20, 30, 40))
      expect_equal(as.numeric(bbox_out), c(-10, -20, 30, 40))
    })

  })

  describe("errors", {

    it("requires numeric coordinate arguments", {
      expect_error(ddbs_make_envelope("a", 0, 10, 10))
      expect_error(ddbs_make_envelope(0, "b", 10, 10))
      expect_error(ddbs_make_envelope(0, 0, "c", 10))
      expect_error(ddbs_make_envelope(0, 0, 10, "d"))
    })

    it("validates conn argument type", {
      expect_error(ddbs_make_envelope(0, 0, 10, 10, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_make_envelope(0, 0, 10, 10, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_make_envelope(0, 0, 10, 10, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_make_envelope(0, 0, 10, 10, conn = conn_test, name = c("a", "b")))
    })

    it("errors on unexpected arguments", {
      expect_error(ddbs_make_envelope(0, 0, 10, 10, new_column = 999))
    })

  })

})


# 5. ddbs_minimum_rotated_rectangle() ------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns different output formats
## - CHECK 1.3: messages work
## - CHECK 1.4: writing a table works
## - CHECK 1.5: geometry type is POLYGON
## - CHECK 1.6: result bbox encloses the input bbox
## - CHECK 2.1: general errors

describe("ddbs_minimum_rotated_rectangle()", {

  describe("expected behavior", {

    it("works on all input formats", {
      output_ddbs <- ddbs_minimum_rotated_rectangle(countries_ddbs)
      output_sf   <- ddbs_minimum_rotated_rectangle(countries_sf)
      output_conn <- ddbs_minimum_rotated_rectangle("countries", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(nrow(ddbs_collect(output_ddbs)), nrow(countries_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats", {
      output_sf_fmt <- ddbs_minimum_rotated_rectangle(countries_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_minimum_rotated_rectangle(countries_ddbs))
      expect_message(ddbs_minimum_rotated_rectangle("countries", conn = conn_test, name = "min_rect"))
      expect_message(ddbs_minimum_rotated_rectangle("countries", conn = conn_test, name = "min_rect", overwrite = TRUE))
      expect_true(ddbs_minimum_rotated_rectangle("countries", conn = conn_test, name = "min_rect2"))

      expect_no_message(ddbs_minimum_rotated_rectangle(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_minimum_rotated_rectangle("countries", conn = conn_test, name = "min_rect", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "min_rect")
      expect_equal(
        ddbs_collect(ddbs_minimum_rotated_rectangle(countries_ddbs))$geometry,
        output_tbl$geometry
      )
    })

    it("produces POLYGON geometry", {
      geom_type <- sf::st_geometry_type(ddbs_collect(ddbs_minimum_rotated_rectangle(argentina_ddbs))) |> as.character()
      expect_equal(geom_type, "POLYGON")
    })

    it("result bounding box encloses the input bounding box", {
      result_bbox <- ddbs_bbox(ddbs_minimum_rotated_rectangle(argentina_ddbs))
      input_bbox  <- ddbs_bbox(argentina_ddbs)

      expect_lte(result_bbox[["xmin"]], input_bbox[["xmin"]])
      expect_gte(result_bbox[["xmax"]], input_bbox[["xmax"]])
      expect_lte(result_bbox[["ymin"]], input_bbox[["ymin"]])
      expect_gte(result_bbox[["ymax"]], input_bbox[["ymax"]])
    })

  })

  describe("errors", {

    it("requires a valid connection when using table name", {
      expect_error(ddbs_minimum_rotated_rectangle("countries", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_minimum_rotated_rectangle(x = 999))
      expect_error(ddbs_minimum_rotated_rectangle(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_minimum_rotated_rectangle(countries_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_minimum_rotated_rectangle(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_minimum_rotated_rectangle(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_minimum_rotated_rectangle(countries_ddbs, conn = conn_test, name = c("a", "b")))
    })

    it("errors on unexpected arguments", {
      expect_error(ddbs_minimum_rotated_rectangle(countries_ddbs, new_column = 999))
    })

  })

})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' Perform a spatial join of two geometries
#'
#' Combines two sets of geometries based on spatial relationships, such as 
#' intersection or containment, attaching attributes from one set to the other.
#'
#' @template x
#' @template y
#' @param join A geometry predicate function. Defaults to `"intersects"`. See
#' the details for other options.
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @param distance a numeric value specifying the distance for ST_DWithin. The units
#' should be specified in meters
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @template spatial_join_predicates
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # RECOMMENDED: Efficient lazy workflow using ddbs_open_dataset
#' library(duckspatial)
#'
#' # Load data directly as lazy spatial data frames (CRS auto-detected)
#' countries <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", package = "duckspatial")
#' )
#'
#' # Create random points
#' n <- 100
#' points <- data.frame(
#'     id = 1:n,
#'     x = runif(n, min = -180, max = 180),
#'     y = runif(n, min = -90, max = 90)
#' ) |> 
#'   sf::st_as_sf(coords = c("x", "y"), crs = 4326) |>
#'   as_duckspatial_df()
#'
#' # Lazy join - computation stays in DuckDB
#' result <- ddbs_join(points, countries, join = "within")
#'
#' # Collect to sf when needed
#' result_sf <- dplyr::collect(result) |> sf::st_as_sf()
#' plot(result_sf["CNTR_NAME"])
#'
#'
#' # Alternative: using sf objects directly (legacy compatibility)
#' library(sf)
#'
#' countries_sf <- sf::st_read(system.file("spatial/countries.geojson", package = "duckspatial"))
#'
#' output <- duckspatial::ddbs_join(
#'     x = points,
#'     y = countries_sf,
#'     join = "within"
#' )
#'
#'
#' # Alternative: using table names in a duckdb connection
#' conn <- duckspatial::ddbs_create_conn()
#'
#' ddbs_write_table(conn, points, "points", overwrite = TRUE)
#' ddbs_write_table(conn, countries_sf, "countries", overwrite = TRUE)
#'
#' output2 <- ddbs_join(
#'     conn = conn,
#'     x = "points",
#'     y = "countries",
#'     join = "within"
#' )
#'
#' }
ddbs_join <- function(
    x,
    y,
    join = "intersects",
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    distance = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    # 0. Validate inputs
    assert_xy(x, "x")
    assert_xy(y, "y")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
    
    # Validate predicate early (it aborts on invalid)
    sel_pred <- get_st_predicate(join)
    
    # 1. Prepare inputs

    ## 1.1. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## 1.3. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn_x)
    crs_y    <- ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    # NOTE: Inline connection resolution logic was replaced by resolve_spatial_connections()
    # helper (defined in db_utils_not_exported.R) to maintain consistency with ddbs_join
    # and other two-input spatial functions. See tests/testthat/test-resolve_connections.R
    # for regression tests covering cross-connection scenarios.
    target_conn <- resolve_res$conn
    x           <- resolve_res$x
    y           <- resolve_res$y
    
    ## 1.4. register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_res$cleanup(), add = TRUE)   
    }
    
    ## 1.5. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)
    
    ## 1.6. Validate the CRS of x and y
    validate_xy_crs(
        crs_x = crs_x,
        crs_y = crs_y,
        conn = target_conn,
        x_list = x_list,
        y_list = y_list
    )


    # 2. Prepare the query

    ## 2.1. Get names of geometry columns (use saved sf_col_x/y from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## 2.2. Build predicate clause
    st_function <- glue::glue("v1.{x_geom}")
    
    st_predicate <- generate_predicate_clause(
        predicate = sel_pred,
        conn      = target_conn,
        x_list    = x_list,
        y_list    = y_list,
        x_geom    = x_geom,
        y_geom    = y_geom,
        distance  = distance,
        crs_x     = crs_x
    )

    ## 2.3 Build the base query
    base.query <- glue::glue("
        SELECT 
            v1.* REPLACE ({build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}),
            v2.* EXCLUDE ({y_geom})
        FROM 
            {x_list$query_name} v1
        JOIN 
            {y_list$query_name} v2
        ON 
            {st_predicate};
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs_x,
            x_geom = x_geom
        )
    }

}

# skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")


# helpers --------------------------------------------------------------

# create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

# helper function
tester <- function(x = points_sf,
                   y = countries_sf,
                   join = "intersects",
                   conn = NULL,
                   name = NULL,
                   mode = "sf",
                   overwrite = FALSE,
                   quiet = FALSE) {
    ddbs_join(
        x = x,
        y = y,
        join = join,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet
    )
}


# expected behavior --------------------------------------------------------------


testthat::test_that("expected behavior", {

    # option 1: passing sf objects
    output1 <- tester(
        x = points_sf,
        y = countries_sf,
        join = "within"
    )

    testthat::expect_true(is(output1 , 'sf'))

    # option 2: passing the names of tables in a duckdb db, returing sf
    # write sf to duckdb
    ddbs_write_table(conn_test, points_sf, "points", overwrite = TRUE)
    ddbs_write_table(conn_test, countries_sf, "countries", overwrite = TRUE)

    # spatial join
    output2 <- tester(
        conn = conn_test,
        x = "points",
        y = "countries",
        join = "within"
    )

    testthat::expect_true(is(output2 , 'sf'))

    # option 3: passing the names of tables in a duckdb db, creating new table in db
    output3 <- tester(
        conn = conn_test,
        x = "points",
        y = "countries",
        join = "within",
        name = "test_result",
        overwrite = TRUE
    )

    testthat::expect_true(output3)

    output3 <- ddbs_read_table(conn_test, "test_result")

    testthat::expect_true(is(output3 , 'sf'))

    ddbs_read_table(conn = conn_test, name = "test_result")


    # show and suppress messages
    testthat::expect_no_message( tester() )
    testthat::expect_no_message( tester(quiet = TRUE))


})


testthat::test_that("error if table already exists", {

    # write table for the 1st time
    testthat::expect_true(tester(x = "points",
                                    y = "countries",
                                    conn = conn_test,
                                    name = 'banana',
                                    overwrite = FALSE)
                             )

    # expected error if overwrite = FALSE
    testthat::expect_error(tester(x = "points",
                                    y = "countries",
                                    conn = conn_test,
                                    name = 'banana',
                                    overwrite = FALSE))

    # overwrite table
    testthat::expect_true(tester(x = "points",
                                    y = "countries",
                                    conn = conn_test,
                                    name = 'banana',
                                    overwrite = TRUE))


})

# expected errors --------------------------------------------------------------

testthat::test_that("errors with incorrect input", {

    testthat::expect_error(tester(x = 999))
    testthat::expect_error(tester(y = 999))
    testthat::expect_error(tester(join = 999))
    testthat::expect_error(tester(conn = 999))
    testthat::expect_error(tester(overwrite = 999))
    testthat::expect_error(tester(quiet = 999))

    testthat::expect_error(tester(x = "999", conn = conn_test))
    testthat::expect_error(tester(y = "999", conn = conn_test))

    testthat::expect_error(tester(conn = conn_test, name = c('banana', 'banana')))


    })



# duckspatial_df inputs --------------------------------------------------------

testthat::test_that("ddbs_join works with duckspatial_df inputs", {
  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")
  
  # Create a distinct connection for this test to avoid interference
  # Use the internal helper visible in other tests
  conn <- ddbs_temp_conn()
  
  # Load as duckspatial_df
  countries_ds <- ddbs_open_dataset(countries_path, conn = conn)
  
  # Create points as sf
  # Helper defined in test-utils.R or we just create sf manually
  points_sf <- sf::st_as_sf(
    data.frame(id = 1:10, x = 1:10, y = 1:10), 
    coords = c("x", "y"), 
    crs = 4326
  )
  
  # Register points to the same connection as a duckspatial_df for consistent testing
  ddbs_write_table(conn, points_sf, "test_points")
  points_ds <- ddbs_read_table(conn, "test_points")
  
  # 1. duckspatial_df x duckspatial_df
  # Using intersects
  result1 <- ddbs_join(points_ds, countries_ds, join = "intersects")
  expect_s3_class(result1, "duckspatial_df")
  expect_true(nrow(dplyr::collect(result1)) >= 0)
  
  # 2. duckspatial_df x sf
  # Note: Cross-connection join might occur if sf implicit connection is different, but here we expect it to work
  result2 <- ddbs_join(points_ds, countries_sf, join = "intersects")
  expect_s3_class(result2, "duckspatial_df")
  
  # 3. sf x duckspatial_df
  
  withr::with_options(list(duckspatial.mode = "sf"), {
      r_sf <- ddbs_join(points_sf, countries_ds, join = "intersects")
      expect_s3_class(r_sf, "sf")
  })
})

# predicates -------------------------------------------------------------------

testthat::test_that("ddbs_join works with different predicates", {
  # We test a few key ones to ensure parameter passing works
  # Use simple data where we know the answer
  
  # Polygon: square (0,0) to (10,10)
  p1 <- sf::st_polygon(list(matrix(c(0,0, 10,0, 10,10, 0,10, 0,0), ncol=2, byrow=TRUE)))
  poly_sf <- sf::st_sf(id = 1, geometry = sf::st_sfc(p1), crs=4326)
  
  # Points: inside (5,5), edge (0,0), outside (20,20)
  pts <- matrix(c(5,5, 0,0, 20,20), ncol=2, byrow=TRUE)
  pts_sf <- sf::st_sf(id = 1:3, geometry = sf::st_sfc(lapply(1:3, function(i) sf::st_point(pts[i,]))), crs=4326)
  
  res_within <- ddbs_join(pts_sf, poly_sf, join = "within") |> dplyr::collect()
  
  expect_true(1 %in% res_within$id)
  expect_false(3 %in% res_within$id)
  
  res_disjoint <- ddbs_join(pts_sf, poly_sf, join = "disjoint") |> dplyr::collect()
  expect_true(3 %in% res_disjoint$id)
  expect_false(1 %in% res_disjoint$id)
  
  res_intersects <- ddbs_join(pts_sf, poly_sf, join = "intersects") |> dplyr::collect()
  expect_true(1 %in% res_intersects$id)
  expect_true(2 %in% res_intersects$id)
  expect_false(3 %in% res_intersects$id)
})

# output parameters ------------------------------------------------------------

testthat::test_that("ddbs_join respects output parameter", {
  result_sf <- ddbs_join(points_sf, countries_sf, mode = "sf")
  expect_s3_class(result_sf, "sf")
  
  result_ds <- ddbs_join(points_sf, countries_sf, mode = "duckspatial")
  expect_s3_class(result_ds, "duckspatial_df")
})

# error handling ---------------------------------------------------------------

testthat::test_that("ddbs_join throws error on CRS mismatch", {
  points_3857 <- sf::st_transform(points_sf, 3857)
  
  expect_error(
    ddbs_join(points_3857, countries_sf),
    "Coordinates Reference System"
  )
})

#' Detect file format from path or URL
#'
#' @param path Character string used as path or URL
#' @return Character string: "parquet", "sf", or other extension
#' @noRd
get_file_format <- function(path) {
  # 1. Handle S3 URIs -> assume parquet default if no extension
  if (grepl("^s3://", path)) {
    # Check if it has an extension
    ext <- tools::file_ext(path)
    if (ext == "") return("parquet")
  }
  
  # 2. Get extension
  ext <- tolower(tools::file_ext(path))
  
  # 3. Check for specific extensions
  if (ext %in% c("parquet")) {
    return("parquet")
  }
  if (ext %in% c("duckdb", "db", "ddb")) {
    return("duckdb")
  }
  if (ext == "shp") {
     return("shp")
  }
  if (grepl("osm\\.pbf$", path) || ext == "osm.pbf") {
     return("osm.pbf")
  }
  
  # 4. Check for known spatial extensions -> "sf" (handled by ST_Read)
  # This list aligns with formats commonly handled by GDAL/ST_Read
  spatial_exts <- c("shp", "gpkg", "fgb", "json", "geojson", "kml", "gpx")
  if (ext %in% spatial_exts) {
    return("sf")
  }
  
  # 5. Default fallback
  if (ext == "") {
    # If no extension, we return "unknown" to trigger auto-detection logic
    return("unknown")
  }
  
  # Return the extension purely for information/fallback usage
  return(ext)
}

duckdb_file_extensions <- function() {
  c("duckdb", "db", "ddb")
}

has_duckdb_file_extension <- function(path) {
  if (!is.character(path) || length(path) != 1) {
    return(FALSE)
  }
  tolower(tools::file_ext(path)) %in% duckdb_file_extensions()
}

duckdb_file_info <- function(path) {
  result <- list(
    exists = FALSE,
    size = NA_real_,
    valid = FALSE,
    empty = FALSE
  )

  if (!is.character(path) || length(path) != 1 || grepl("^[[:alpha:]][[:alnum:].+-]*://", path)) {
    return(result)
  }

  if (!file.exists(path)) {
    return(result)
  }

  info <- file.info(path)
  result$exists <- TRUE
  result$size <- info$size
  result$empty <- isTRUE(result$size == 0)

  if (is.na(result$size) || isTRUE(info$isdir) || result$size < 12) {
    return(result)
  }

  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)
  bytes <- readBin(con, "raw", n = 12)

  result$valid <- length(bytes) >= 12 && identical(bytes[9:12], charToRaw("DUCK"))
  result
}

#' Get CRS from Parquet metadata
#' 
#' @param path Path to parquet file
#' @param conn DuckDB connection
#' @return crs object or NULL
#' @noRd
get_parquet_crs <- function(path, conn) {
  # Try to read GeoParquet metadata from KV metadata
  # We extract the raw blob and parse in R to be more robust across platforms
  
  tryCatch({
    safe_path <- DBI::dbQuoteString(conn, path)
    
    # 1. Extract the raw 'geo' metadata value
    # We use decode(key) to handle blob keys if present
    query <- glue::glue("
      SELECT value
      FROM parquet_kv_metadata({safe_path})
      WHERE decode(key) = 'geo'
      LIMIT 1
    ")    
    res <- DBI::dbGetQuery(conn, query)
    
    if (nrow(res) == 0 || is.null(res$value[[1]])) {
      return(NULL)
    }
    
    # 2. Parse JSON in R
    # DuckDB returns BLOB as a raw vector in a list
    meta_str <- rawToChar(res$value[[1]])
    meta <- jsonlite::fromJSON(meta_str, simplifyVector = FALSE)
    
    # 3. Extract CRS for the primary column
    primary_col <- meta$primary_column
    if (is.null(primary_col) || !primary_col %in% names(meta$columns)) {
       return(NULL)
    }
    
    crs_data <- meta$columns[[primary_col]]$crs
    if (is.null(crs_data)) {
      return(NULL)
    }
    
    # 4. Convert to sf CRS object
    # If it's a list (PROJJSON), try to extract EPSG or convert to string
    if (is.list(crs_data)) {
      # Try to extract EPSG code for better compatibility
      # Standard PROJJSON has id: {authority: "...", code: ...}
      authority <- crs_data$id$authority
      code <- crs_data$id$code
      
      if (!is.null(authority) && !is.null(code)) {
        return(sf::st_crs(paste0(authority, ":", code)))
      }
      
      # Handle cases where id might be a list of one element
      if (is.list(crs_data$id) && length(crs_data$id) > 0) {
        authority <- crs_data$id[[1]]$authority
        code <- crs_data$id[[1]]$code
        if (!is.null(authority) && !is.null(code)) {
          return(sf::st_crs(paste0(authority, ":", code)))
        }
      }
      
      # Fallback to JSON string for sf::st_crs
      crs_data <- jsonlite::toJSON(crs_data, auto_unbox = TRUE)
    }
    
    return(sf::st_crs(crs_data))
    
  }, error = function(e) {
    # If anything fails (file not found, bad format), warn and return NULL
    cli::cli_warn("Failed to extract GeoParquet CRS: {e$message}")
    NULL
  })
}

#' Map extension to common GDAL driver names
#' @noRd
get_driver_map <- function() {
  list(
    # Native / Common
    parquet = "parquet",
    csv = "CSV",
    
    # GDAL Spatial
    shp = "ESRI Shapefile",
    gpkg = "GPKG",
    fgb = "FlatGeoBuf",
    json = "GeoJSON",
    geojson = "GeoJSON",
    kml = "KML",
    gml = "GML",
    gpx = "GPX",
    sqlite = "SQLite",
    tab = "MapInfo File",
    mif = "MapInfo File",
    geojsonl = "GeoJSON",
    mvt = "MVT",
    dgn = "DGN",
    gdb = "OpenFileGDB", 
    gxt = "Geoconcept",
    xml = "GML"
  )
}

#' Get driver name from format/extension
#' @noRd
get_driver_name <- function(fmt) {
  map <- get_driver_map()
  if (fmt %in% names(map)) {
    return(map[[fmt]])
  }
  return(NULL)
}

#' Check if format implies spatial data
#' @noRd
is_spatial_format <- function(fmt) {
  # GDAL formats usually imply spatial data.
  spatial_formats <- c("shp", "gpkg", "fgb", "json", "geojson", "kml", "gpx")
  
  # Check if extension is one of these
  if (fmt %in% spatial_formats) return(TRUE)
  
  # Check driver map (excluding native formats which we handle separately or aren't strictly spatial)
  map <- get_driver_map()
  map$parquet <- NULL
  map$csv <- NULL
  
  driver <- get_driver_name(fmt)
  if (!is.null(driver) && driver %in% unlist(map)) return(TRUE)
  
  return(FALSE)
}

## Line operations:
## - ddbs_line_merge
## - ddbs_line_interpolate
## - ddbs_line_locate_point
## - ddbs_line_substring
## - ddbs_endpoint
## - ddbs_startpoint
## - ddbs_build_area
## - ddbs_polygonize
## - ddbs_make_polygon



#' Extract the start point of a linestring geometry
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `ddbs_startpoint()` was renamed to \code{\link{ddbs_line_startpoint}}.
#'
#' @inheritParams ddbs_endpoint_startpoint
#' @template returns_mode
#' @export
#' @keywords internal
ddbs_startpoint <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  lifecycle::deprecate_soft(
    when    = "1.1.0",
    what    = "ddbs_startpoint()",
    with    = "ddbs_line_startpoint()"
  )

  ddbs_line_startpoint(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet
  )

}




#' Extract the end point of a linestring geometry
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `ddbs_endpoint()` was renamed to \code{\link{ddbs_line_endpoint}}.
#'
#' @inheritParams ddbs_endpoint_startpoint
#' @template returns_mode
#' @export
#' @keywords internal
ddbs_endpoint <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  lifecycle::deprecate_soft(
    when    = "1.1.0",
    what    = "ddbs_endpoint()",
    with    = "ddbs_line_endpoint()"
  )

  ddbs_line_endpoint(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet
  )

}


#' Extract the start or end point of a linestring geometry
#'
#' Returns the first or last point of a LINESTRING geometry. These functions only work
#' with LINESTRING geometries (not MULTILINESTRING or other geometry types).
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @details
#' These functions wrap DuckDB Spatial's \code{ST_StartPoint} and \code{ST_EndPoint}.
#' Input geometries must be of type LINESTRING (MULTILINESTRING is not supported).
#' For each input feature, the first or last coordinate of the LINESTRING is returned
#' as a POINT geometry.
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## store in duckdb
#' ddbs_write_vector(conn, rivers_ddbs, "rivers")
#'
#' ## extract start points
#' ddbs_line_startpoint(conn = conn, "rivers")
#'
#' ## extract end points
#' ddbs_line_endpoint(conn = conn, "rivers")
#'
#' ## without using a connection
#' ddbs_line_startpoint(rivers_ddbs)
#' ddbs_line_endpoint(rivers_ddbs)
#' }
#' @name ddbs_endpoint_startpoint
#' @rdname ddbs_endpoint_startpoint
NULL



#' @rdname ddbs_endpoint_startpoint
#' @export
ddbs_line_startpoint <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Handle function-specific error
    assert_geom_type(x = x, conn = conn, geom = "LINESTRING", multi = FALSE)

    # 1. Run the template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_StartPoint",
        other_args = NULL
    )

}

#' @rdname ddbs_endpoint_startpoint
#' @export
ddbs_line_endpoint <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Handle function-specific error
    assert_geom_type(x = x, conn = conn, geom = "LINESTRING", multi = FALSE)

    # 1. Run the template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_EndPoint",
        other_args = NULL
    )

}



#' Interpolates a point or points along a line geometry
#'
#' Returns either a single point at a specified position along a line, or
#' multiple equally-spaced points along a line, depending on the value of
#' \code{intervals}. When \code{intervals = FALSE}, this wraps
#' \code{ST_LineInterpolatePoint}; when \code{intervals = TRUE}, it wraps
#' \code{ST_LineInterpolatePoints}.
#'
#' @template x
#' @param fraction a numeric value between 0 and 1. When
#' \code{intervals = FALSE}, specifies the position along the line to
#' interpolate, where \code{0} is the start and \code{1} is the end.
#' When \code{intervals = TRUE}, specifies the spacing between interpolated
#' points as a proportion of the total line length. Defaults to \code{0.5}.
#' @param intervals a logical value. If \code{FALSE} (default), returns a
#' single \code{POINT} at the position given by \code{fraction}. If
#' \code{TRUE}, returns a \code{MULTIPOINT} of equally-spaced points along
#' the line at intervals defined by \code{fraction}.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## return the midpoint of a line (default)
#' ddbs_line_interpolate(rivers_ddbs)
#'
#' ## return the point 25% along the line
#' ddbs_line_interpolate(rivers_ddbs, fraction = 0.25)
#'
#' ## return equally-spaced points every 10% of the line length
#' ddbs_line_interpolate(rivers_ddbs, fraction = 0.1, intervals = TRUE)
#'
#' ## return equally-spaced points every 50% of the line length (i.e. midpoint and end)
#' ddbs_line_interpolate(rivers_ddbs, fraction = 0.5, intervals = TRUE)
#' }
ddbs_line_interpolate <- function(
    x,
    fraction = 0.5,
    intervals = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
  
    # 0. Handle function-specific errors
    assert_numeric(fraction, "fraction")
    assert_numeric_interval(fraction, 0, 1)
    assert_logic(intervals, "intervals")
  
    # 1. Build ST_Buffer parameters string
    other_args <- glue::glue("{fraction}, {intervals}")
  
    # 2. Pass to template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_LineInterpolatePoints",
        other_args = other_args
    )

}





#' Extract a substring of a line geometry
#'
#' Returns the portion of a line between two fractional positions along its
#' length
#'
#' @template x
#' @param start a numeric value between 0 and 1. Specifies the starting
#' position along the line as a proportion of its total length, where
#' \code{0} is the beginning of the line. Defaults to \code{0}.
#' @param end a numeric value between 0 and 1. Specifies the ending
#' position along the line as a proportion of its total length, where
#' \code{1} is the end of the line. Must be greater than or equal to
#' \code{start}. Defaults to \code{0.5}.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## return the first half of each line (default)
#' ddbs_line_substring(rivers_ddbs)
#'
#' ## return the middle third of each line
#' ddbs_line_substring(rivers_ddbs, start = 0.33, end = 0.67)
#'
#' ## return the last quarter of each line
#' ddbs_line_substring(rivers_ddbs, start = 0.75, end = 1)
#' }
ddbs_line_substring <- function(
    x,
    start = 0,
    end = 0.5,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
  
    # 0. Handle function-specific errors
    assert_numeric(start, "start")
    assert_numeric(end, "end")
    assert_numeric_interval(start, 0, 1)
    assert_numeric_interval(end, 0, 1)
    if (start > end) {
        cli::cli_abort("{.arg start} must be less than or equal to {.arg end}.")
    }
  
    # 1. Build ST_Buffer parameters string
    other_args <- glue::glue("{start}, {end}")
  
    # 2. Pass to template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_LineSubstring",
        other_args = other_args
    )

}





#' Merge line geometries into a single line
#'
#' Merges a collection of line geometries that share endpoints into a single
#' \code{LINESTRING}, or \code{MULTILINESTRING} if endpoints are not shared
#'
#' @template x
#' @param preserve a logical value. If \code{TRUE} (default), line direction
#' is preserved
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(dplyr)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#' 
#' ## first, union by river name
#' rivers_union <- ddbs_union_agg(rivers_ddbs, by = "RIVER_NAME")
#'
#' ## merge lines, preserving direction
#' rivers_merged <- ddbs_line_merge(rivers_union)
#' 
#' ## check Rio Eume (union doesn't guarantee the merging)
#' rivers_union |> filter(RIVER_NAME == "Rio Eume")
#' rivers_merged |> filter(RIVER_NAME == "Rio Eume")
#' 
#' }
ddbs_line_merge <- function(
    x,
    preserve = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
  
    # 0. Handle function-specific errors
    assert_logic(preserve, "preserve")
  
    # 1. Build ST_Buffer parameters string
    other_args <- glue::glue("{preserve}")
  
    # 2. Pass to template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_LineMerge",
        other_args = other_args
    )

}




#' Node a set of linestrings
#'
#' Fully nodes a collection of line geometries, splitting them at every point
#' where they cross or touch and adding a node at each intersection. The result
#' is always a \code{MULTILINESTRING}. A polygon input is noded into its
#' boundary linework.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## node the river network (split lines at every crossing)
#' rivers_noded <- ddbs_line_node(rivers_ddbs)
#' }
ddbs_line_node <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Node",
        other_args = NULL
    )

}




#' Create a polygon from a single closed linestring
#'
#' Converts a single closed linestring geometry into a polygon. The linestring 
#' must be closed (first and last points identical). Does not work with 
#' MULTILINESTRING inputs - use [ddbs_polygonize()] or [ddbs_build_area()] instead.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @family polygon construction
#' @seealso [ddbs_polygonize()], [ddbs_build_area()]
#' @export
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina")
#'
#' ## extract exterior ring as linestring, then convert back to polygon
#' ring_ddbs <- ddbs_exterior_ring(conn = conn, "argentina")
#' ddbs_make_polygon(conn = conn, ring_ddbs, name = "argentina_poly")
#'
#' ## create polygon without using a connection
#' ddbs_make_polygon(ring_ddbs)
#' }
ddbs_make_polygon <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    # 0. Handle function-specific error
    assert_geom_type(x = x, conn = conn, geom = "LINESTRING", multi = FALSE)
    
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_MakePolygon",
        other_args = NULL
    )
}








#' Assemble polygons from multiple linestrings
#'
#' Takes a collection of linestrings or polygons and assembles them into polygons by 
#' finding all closed rings formed by the network. Returns a GEOMETRYCOLLECTION containing 
#' the resulting polygons.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @family polygon construction
#' @seealso [ddbs_make_polygon()], [ddbs_build_area()]
#' @export
ddbs_polygonize <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

  
    # 1. Prepare inputs
  
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)
  
    ## 2.2.  Build the base query (depends on the output type - sf, duckspatial_df, table)
    st_function <- glue::glue("ST_Polygonize(LIST({x_geom}))")
    base.query <- glue::glue("
      SELECT 
        {build_geom_query(st_function, name, crs_x, mode)} as {x_geom}
      FROM 
        {x_list$query_name};
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
        conn      = target_conn,
        name      = name,
        query     = base.query,
        overwrite = overwrite,
        quiet     = quiet
        )
    } else {
        ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_x,
        x_geom     = x_geom
        )
    }
    
}







#' Build polygon areas from multiple linestrings
#'
#' Constructs polygon or multipolygon geometries from a collection of linestrings, 
#' handling intersections and creating unified areas. Returns POLYGON or MULTIPOLYGON 
#' (not wrapped in a geometry collection). Requires MULTILINESTRING input - for 
#' single linestrings, use [ddbs_make_polygon()].
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @family polygon construction
#' @seealso [ddbs_make_polygon()], [ddbs_polygonize()]
#' @export
ddbs_build_area <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  
  template_unary_ops(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_BuildArea",
    other_args = NULL
  )

}




#' Locate a point along a linestring
#'
#' Returns a value between 0 and 1 representing the position of the closest
#' point on the line to the reference point, as a fraction of the line's
#' total length. 0 corresponds to the start of the line and 1 to the end.
#'
#' @template x
#' @param y The reference point. One of:
#'   \itemize{
#'     \item An \code{sf} object with exactly 1 point feature.
#'     \item A \code{duckspatial_df} with exactly 1 point feature.
#'     \item A character string naming a DuckDB table that contains exactly 1
#'       point feature (requires \code{conn}).
#'   }
#' @template new_column
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read river data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson", package = "duckspatial")
#' )
#'
#' ## define a single reference point (sf with 1 row)
#' ref_pt <- sf::st_sf(
#'   geometry = sf::st_sfc(sf::st_point(c(-8, 43)), crs = 4326)
#' )
#'
#' ## locate the fraction along each river closest to the reference point
#' ddbs_line_locate_point(rivers_ddbs, y = ref_pt)
#' }
ddbs_line_locate_point <- function(
    x,
    y,
    new_column = "line_fraction",
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Validate inputs

    ## common input checks
    assert_xy(x, "x")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
    assert_character_scalar(new_column, "new_column")

    ## y type check
    if (!inherits(y, c("sf", "duckspatial_df", "tbl_duckdb_connection")) && !is.character(y)) {
        cli::cli_abort(
            "{.arg y} must be an {.cls sf} object, a {.cls duckspatial_df}, or a character table name.",
            call = NULL
        )
    }

    ## early row-count check for sf (no DB needed)
    if (inherits(y, "sf") && nrow(y) != 1L) {
        cli::cli_abort(
            "{.arg y} must have exactly 1 feature when passed as an {.cls sf} object, not {nrow(y)}.",
            call = NULL
        )
    }

    ## connection requirements
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x, y)


    # 1. Prepare inputs

    ## 1.1. Resolve conn defaults for character inputs (mirrors binary ops pattern)
    conn_x <- if (is.character(x)) conn else NULL
    conn_y <- if (is.character(y)) conn else NULL

    ## 1.2. Normalize x
    x        <- normalize_spatial_input(x, conn_x)
    crs_x    <- ddbs_crs(x, conn_x)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)


    # 2. Resolve connections and build y_sql based on y type

    if (inherits(y, "sf")) {
        ## y is sf: use an inline WKT literal — no connection resolution for y
        y_wkt <- sf::st_as_text(sf::st_geometry(y)[[1]])

        resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
        target_conn  <- resolve_conn$conn
        x            <- resolve_conn$x
        on.exit(resolve_conn$cleanup(), add = TRUE)

        x_list <- get_query_list(x, target_conn)
        on.exit(x_list$cleanup(), add = TRUE)

        x_geom <- (sf_col_x %||% get_geom_name(target_conn, x_list$query_name))[1]
        assert_geometry_column(x_geom, x_list)

        y_sql <- glue::glue("ST_GeomFromText('{y_wkt}')")

    } else {
        ## y is duckspatial_df or character table: normalize and resolve both
        y        <- normalize_spatial_input(y, conn_y)
        sf_col_y <- attr(y, "sf_column")

        resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
        target_conn <- resolve_res$conn
        x           <- resolve_res$x
        y           <- resolve_res$y
        on.exit(resolve_res$cleanup(), add = TRUE)

        x_list <- get_query_list(x, target_conn)
        on.exit(x_list$cleanup(), add = TRUE)

        y_list <- get_query_list(y, target_conn)
        on.exit(y_list$cleanup(), add = TRUE)

        x_geom <- (sf_col_x %||% get_geom_name(target_conn, x_list$query_name))[1]
        y_geom <- (sf_col_y %||% get_geom_name(target_conn, y_list$query_name))[1]
        assert_geometry_column(x_geom, x_list)

        ## validate row count of y (after resolution so the table exists in target_conn)
        y_n <- DBI::dbGetQuery(
            target_conn,
            glue::glue("SELECT COUNT(*) AS n FROM {y_list$query_name}")
        )$n
        if (y_n != 1L) {
            cli::cli_abort(
                "{.arg y} must contain exactly 1 feature, but has {y_n} row{?s}.",
                call = NULL
            )
        }

        y_sql <- glue::glue("(SELECT {y_geom}::GEOMETRY FROM {y_list$query_name})")
    }


    # 3. Build the base query
    if (mode == "sf") {
        base.query <- glue::glue("
            SELECT ST_LineLocatePoint({x_geom}, {y_sql}) AS {new_column},
            FROM {x_list$query_name};
        ")
    } else {
        base.query <- glue::glue("
            SELECT
                * EXCLUDE {x_geom},
                ST_LineLocatePoint({x_geom}, {y_sql}) AS {new_column},
                {build_geom_query(x_geom, name, crs_x, mode)} AS {x_geom}
            FROM
                {x_list$query_name};
        ")
    }


    # 4. Table creation if name is provided, or
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query     = base.query,
            conn      = target_conn,
            mode      = mode,
            crs       = crs_x,
            x_geom    = x_geom,
            fun_group = 2,
            units     = NULL
        )
    }

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write tables
ddbs_write_table(conn_test, argentina_sf, "argentina")
ddbs_write_table(conn_test, rivers_sf,    "rivers")

## locate fixture: 3 identical horizontal lines in conn_test and a reference point table
DBI::dbExecute(conn_test, "
    CREATE TABLE locate_test AS
    SELECT 1 AS id, ST_GeomFromText('LINESTRING (0 0, 1 0)') AS geometry
    UNION ALL SELECT 2, ST_GeomFromText('LINESTRING (0 0, 1 0)')
    UNION ALL SELECT 3, ST_GeomFromText('LINESTRING (0 0, 1 0)')
")
DBI::dbExecute(conn_test, "
    CREATE TABLE ref_point AS
    SELECT ST_GeomFromText('POINT (0.3 0)') AS geometry
")

## R-side fixtures for sf / duckspatial_df tests
locate_lines_sf <- sf::st_sf(
  id = 1:3,
  geometry = sf::st_sfc(
    sf::st_linestring(matrix(c(0,0, 1,0), ncol = 2, byrow = TRUE)),
    sf::st_linestring(matrix(c(0,0, 1,0), ncol = 2, byrow = TRUE)),
    sf::st_linestring(matrix(c(0,0, 1,0), ncol = 2, byrow = TRUE))
  ),
  crs = 4326
)
locate_ddbs <- as_duckspatial_df(locate_lines_sf)

## single reference point at fraction 0.3 of the unit horizontal line
ref_point_sf   <- sf::st_sf(
  geometry = sf::st_sfc(sf::st_point(c(0.3, 0)), crs = 4326)
)
ref_point_ddbs <- as_duckspatial_df(ref_point_sf)

## prepare exterior_ring table (closed LINESTRING) needed by ddbs_make_polygon tests
ddbs_exterior_ring("argentina", conn = conn_test, name = "exterior_ring", quiet = TRUE)
ext_ring_ddbs <- ddbs_exterior_ring(argentina_ddbs)
ext_ring_sf   <- sf::st_as_sf(ext_ring_ddbs)


# 1. ddbs_make_polygon() -------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: writes table to database
## - CHECK 1.5: returns POLYGON geometry type
## - CHECK 2.1: requires closed LINESTRING input
## - CHECK 2.2: other errors
describe("ddbs_make_polygon()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_make_polygon(ext_ring_ddbs)
      output_sf   <- ddbs_make_polygon(ext_ring_sf)
      output_conn <- ddbs_make_polygon("exterior_ring", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_make_polygon(ext_ring_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_make_polygon(ext_ring_ddbs))
      expect_message(ddbs_make_polygon("exterior_ring", conn = conn_test, name = "make_polygon"))
      expect_message(ddbs_make_polygon("exterior_ring", conn = conn_test, name = "make_polygon", overwrite = TRUE))
      expect_true(ddbs_make_polygon("exterior_ring", conn = conn_test, name = "make_polygon2"))

      expect_no_message(ddbs_make_polygon(ext_ring_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_make_polygon("exterior_ring", conn = conn_test, name = "make_polygon", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_make_polygon(ext_ring_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "make_polygon")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("returns POLYGON geometry type", {
      output_ddbs <- ddbs_make_polygon(ext_ring_ddbs)
      geom_type   <- ddbs_collect(output_ddbs) |> sf::st_geometry_type() |> as.character()

      expect_equal(geom_type, "POLYGON")
    })
  })

  describe("errors", {

    it("requires linestring geometry", {
      expect_error(ddbs_make_polygon(argentina_ddbs))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_make_polygon("exterior_ring", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_make_polygon(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_make_polygon(ext_ring_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_make_polygon(ext_ring_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_make_polygon(ext_ring_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_make_polygon(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_make_polygon(ext_ring_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 2. ddbs_line_startpoint() and ddbs_line_endpoint() ---------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: returns POINT geometry type
## - CHECK 1.5: start and end points differ
## - CHECK 2.1: requires LINESTRING (not polygon)
## - CHECK 2.2: other errors
describe("ddbs_line_startpoint()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_startpoint(rivers_ddbs)
      output_sf   <- ddbs_line_startpoint(rivers_sf)
      output_conn <- ddbs_line_startpoint("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_line_startpoint(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_startpoint(rivers_ddbs))
      expect_message(ddbs_line_startpoint("rivers", conn = conn_test, name = "startpoint"))
      expect_message(ddbs_line_startpoint("rivers", conn = conn_test, name = "startpoint", overwrite = TRUE))
      expect_true(ddbs_line_startpoint("rivers", conn = conn_test, name = "startpoint2"))

      expect_no_message(ddbs_line_startpoint(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_line_startpoint("rivers", conn = conn_test, name = "startpoint", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("returns POINT geometry type", {
      output     <- ddbs_line_startpoint(rivers_ddbs, mode = "sf")
      geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

      expect_true(all(geom_types == "POINT"))
    })
  })

  describe("errors", {

    it("requires LINESTRING geometry, not polygon", {
      expect_error(ddbs_line_startpoint(argentina_ddbs))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_line_startpoint("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_startpoint(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_startpoint(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_line_startpoint(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_line_startpoint(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_startpoint(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_line_startpoint(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})


describe("ddbs_line_endpoint()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_endpoint(rivers_ddbs)
      output_sf   <- ddbs_line_endpoint(rivers_sf)
      output_conn <- ddbs_line_endpoint("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_line_endpoint(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_endpoint(rivers_ddbs))
      expect_message(ddbs_line_endpoint("rivers", conn = conn_test, name = "endpoint"))
      expect_message(ddbs_line_endpoint("rivers", conn = conn_test, name = "endpoint", overwrite = TRUE))
      expect_true(ddbs_line_endpoint("rivers", conn = conn_test, name = "endpoint2"))

      expect_no_message(ddbs_line_endpoint(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_line_endpoint("rivers", conn = conn_test, name = "endpoint", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("returns POINT geometry type", {
      output     <- ddbs_line_endpoint(rivers_ddbs, mode = "sf")
      geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

      expect_true(all(geom_types == "POINT"))
    })

    it("produces different results from ddbs_line_startpoint", {
      start_pts <- ddbs_line_startpoint(rivers_ddbs, mode = "sf")
      end_pts   <- ddbs_line_endpoint(rivers_ddbs,   mode = "sf")

      expect_false(identical(start_pts$geom, end_pts$geom))
    })
  })

  describe("errors", {

    it("requires LINESTRING geometry, not polygon", {
      expect_error(ddbs_line_endpoint(argentina_ddbs))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_line_endpoint("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_endpoint(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_endpoint(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_line_endpoint(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_line_endpoint(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_endpoint(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_line_endpoint(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 3. ddbs_line_interpolate() ---------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: intervals = FALSE returns POINT; intervals = TRUE returns MULTIPOINT
## - CHECK 1.5: different fractions produce different results
## - CHECK 2.1: fraction out of [0, 1] range
## - CHECK 2.2: non-numeric fraction
## - CHECK 2.3: non-logical intervals
## - CHECK 2.4: other errors
describe("ddbs_line_interpolate()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_interpolate(rivers_ddbs)
      output_sf   <- ddbs_line_interpolate(rivers_sf)
      output_conn <- ddbs_line_interpolate("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_line_interpolate(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_interpolate(rivers_ddbs))
      expect_message(ddbs_line_interpolate("rivers", conn = conn_test, name = "interpolate"))
      expect_message(ddbs_line_interpolate("rivers", conn = conn_test, name = "interpolate", overwrite = TRUE))
      expect_true(ddbs_line_interpolate("rivers", conn = conn_test, name = "interpolate2"))

      expect_no_message(ddbs_line_interpolate(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_line_interpolate("rivers", conn = conn_test, name = "interpolate", overwrite = TRUE, quiet = TRUE)
      )
    })

    describe("intervals parameter", {

      it("intervals = FALSE returns POINT geometry type", {
        output     <- ddbs_line_interpolate(rivers_ddbs, intervals = FALSE, mode = "sf")
        geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

        expect_true(all(geom_types == "POINT"))
      })

      it("intervals = TRUE returns MULTIPOINT geometry type", {
        output     <- ddbs_line_interpolate(rivers_ddbs, fraction = 0.25, intervals = TRUE, mode = "sf")
        geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

        expect_true(all(geom_types == "MULTIPOINT"))
      })
    })

    describe("fraction parameter", {

      it("different fractions produce different results", {
        output_25 <- ddbs_line_interpolate(rivers_ddbs, fraction = 0.25, mode = "sf")
        output_75 <- ddbs_line_interpolate(rivers_ddbs, fraction = 0.75, mode = "sf")

        expect_false(identical(output_25$geom, output_75$geom))
      })

      it("fraction = 0 returns start of line", {
        output <- ddbs_line_interpolate(rivers_ddbs, fraction = 0, mode = "sf")
        expect_s3_class(output, "sf")
      })

      it("fraction = 1 returns end of line", {
        output <- ddbs_line_interpolate(rivers_ddbs, fraction = 1, mode = "sf")
        expect_s3_class(output, "sf")
      })
    })
  })

  describe("errors", {

    it("rejects fraction below 0", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, fraction = -0.1))
    })

    it("rejects fraction above 1", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, fraction = 1.1))
    })

    it("rejects non-numeric fraction", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, fraction = "0.5"))
    })

    it("rejects non-logical intervals", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, intervals = "TRUE"))
      expect_error(ddbs_line_interpolate(rivers_ddbs, intervals = 1))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_line_interpolate("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_interpolate(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_interpolate(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_line_interpolate(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 4. ddbs_line_substring() -----------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: different start/end produce different results
## - CHECK 2.1: start > end
## - CHECK 2.2: start or end out of [0, 1] range
## - CHECK 2.3: non-numeric start / end
## - CHECK 2.4: other errors
describe("ddbs_line_substring()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_substring(rivers_ddbs)
      output_sf   <- ddbs_line_substring(rivers_sf)
      output_conn <- ddbs_line_substring("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_line_substring(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_substring(rivers_ddbs))
      expect_message(ddbs_line_substring("rivers", conn = conn_test, name = "substring"))
      expect_message(ddbs_line_substring("rivers", conn = conn_test, name = "substring", overwrite = TRUE))
      expect_true(ddbs_line_substring("rivers", conn = conn_test, name = "substring2"))

      expect_no_message(ddbs_line_substring(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_line_substring("rivers", conn = conn_test, name = "substring", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("returns LINESTRING geometry type", {
      output     <- ddbs_line_substring(rivers_ddbs, mode = "sf")
      geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

      expect_true(all(geom_types == "LINESTRING"))
    })

    describe("start and end parameters", {

      it("different start/end produce different results", {
        output_first  <- ddbs_line_substring(rivers_ddbs, start = 0,    end = 0.5, mode = "sf")
        output_second <- ddbs_line_substring(rivers_ddbs, start = 0.5,  end = 1,   mode = "sf")

        expect_false(identical(output_first$geom, output_second$geom))
      })

      it("full line (start = 0, end = 1) returns same length as input", {
        output <- ddbs_line_substring(rivers_ddbs, start = 0, end = 1, mode = "sf")
        expect_equal(nrow(output), nrow(rivers_sf))
      })
    })
  })

  describe("errors", {

    it("rejects start greater than end", {
      expect_error(ddbs_line_substring(rivers_ddbs, start = 0.8, end = 0.2))
    })

    it("rejects start below 0", {
      expect_error(ddbs_line_substring(rivers_ddbs, start = -0.1))
    })

    it("rejects end above 1", {
      expect_error(ddbs_line_substring(rivers_ddbs, end = 1.1))
    })

    it("rejects non-numeric start", {
      expect_error(ddbs_line_substring(rivers_ddbs, start = "0"))
    })

    it("rejects non-numeric end", {
      expect_error(ddbs_line_substring(rivers_ddbs, end = "1"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_line_substring("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_substring(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_substring(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_line_substring(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_line_substring(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_substring(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_line_substring(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 5. ddbs_line_merge() ---------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: preserve = TRUE / FALSE both work
## - CHECK 2.1: non-logical preserve
## - CHECK 2.2: other errors
describe("ddbs_line_merge()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_merge(rivers_ddbs)
      output_sf   <- ddbs_line_merge(rivers_sf)
      output_conn <- ddbs_line_merge("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_line_merge(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_merge(rivers_ddbs))
      expect_message(ddbs_line_merge("rivers", conn = conn_test, name = "line_merge"))
      expect_message(ddbs_line_merge("rivers", conn = conn_test, name = "line_merge", overwrite = TRUE))
      expect_true(ddbs_line_merge("rivers", conn = conn_test, name = "line_merge2"))

      expect_no_message(ddbs_line_merge(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_line_merge("rivers", conn = conn_test, name = "line_merge", overwrite = TRUE, quiet = TRUE)
      )
    })

    describe("preserve parameter", {

      it("preserve = TRUE works", {
        output <- ddbs_line_merge(rivers_ddbs, preserve = TRUE)
        expect_s3_class(output, "duckspatial_df")
      })

      it("preserve = FALSE works", {
        output <- ddbs_line_merge(rivers_ddbs, preserve = FALSE)
        expect_s3_class(output, "duckspatial_df")
      })
    })
  })

  describe("errors", {

    it("rejects non-logical preserve", {
      expect_error(ddbs_line_merge(rivers_ddbs, preserve = "TRUE"))
      expect_error(ddbs_line_merge(rivers_ddbs, preserve = 1))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_line_merge("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_merge(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_merge(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_line_merge(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_line_merge(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_merge(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_line_merge(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 6. ddbs_polygonize() ---------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 1.4: returns GEOMETRYCOLLECTION geometry type
## - CHECK 2.1: other errors
describe("ddbs_polygonize()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_polygonize(rivers_ddbs)
      output_sf   <- ddbs_polygonize(rivers_sf)
      output_conn <- ddbs_polygonize("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_polygonize(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_polygonize(rivers_ddbs))
      expect_message(ddbs_polygonize("rivers", conn = conn_test, name = "polygonize"))
      expect_message(ddbs_polygonize("rivers", conn = conn_test, name = "polygonize", overwrite = TRUE))
      expect_true(ddbs_polygonize("rivers", conn = conn_test, name = "polygonize2"))

      expect_no_message(ddbs_polygonize(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_polygonize("rivers", conn = conn_test, name = "polygonize", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("returns GEOMETRYCOLLECTION geometry type", {
      output     <- ddbs_polygonize(rivers_ddbs, mode = "sf")
      geom_types <- sf::st_geometry_type(output) |> as.character() |> unique()

      expect_true(all(geom_types == "GEOMETRYCOLLECTION"))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_polygonize("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_polygonize(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_polygonize(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_polygonize(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_polygonize(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_polygonize(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_polygonize(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 7. ddbs_build_area() ---------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns sf format
## - CHECK 1.3: messages work
## - CHECK 2.1: other errors
describe("ddbs_build_area()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_build_area(rivers_ddbs)
      output_sf   <- ddbs_build_area(rivers_sf)
      output_conn <- ddbs_build_area("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns sf format", {
      output_sf_fmt <- ddbs_build_area(rivers_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_build_area(rivers_ddbs))
      expect_message(ddbs_build_area("rivers", conn = conn_test, name = "build_area"))
      expect_message(ddbs_build_area("rivers", conn = conn_test, name = "build_area", overwrite = TRUE))
      expect_true(ddbs_build_area("rivers", conn = conn_test, name = "build_area2"))

      expect_no_message(ddbs_build_area(rivers_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_build_area("rivers", conn = conn_test, name = "build_area", overwrite = TRUE, quiet = TRUE)
      )
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_build_area("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_build_area(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_build_area(rivers_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_build_area(rivers_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_build_area(rivers_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_build_area(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_build_area(rivers_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 8. ddbs_startpoint() and ddbs_endpoint() (deprecated) -----------------

## - CHECK 1.1: ddbs_startpoint() emits deprecation warning but still works
## - CHECK 1.2: ddbs_endpoint() emits deprecation warning but still works
## - CHECK 1.3: produce the same output as their replacements
describe("deprecated ddbs_startpoint() and ddbs_endpoint()", {

  it("ddbs_startpoint() emits deprecation warning", {
    expect_warning(ddbs_startpoint(rivers_ddbs), class = "lifecycle_warning_deprecated")
  })

  it("ddbs_endpoint() emits deprecation warning", {
    expect_warning(ddbs_endpoint(rivers_ddbs), class = "lifecycle_warning_deprecated")
  })

  it("ddbs_startpoint() output matches ddbs_line_startpoint()", {
    output_dep <- suppressWarnings(ddbs_startpoint(rivers_ddbs, mode = "sf"))
    output_new <- ddbs_line_startpoint(rivers_ddbs, mode = "sf")

    expect_equal(output_dep$geom, output_new$geom)
  })

  it("ddbs_endpoint() output matches ddbs_line_endpoint()", {
    output_dep <- suppressWarnings(ddbs_endpoint(rivers_ddbs, mode = "sf"))
    output_new <- ddbs_line_endpoint(rivers_ddbs, mode = "sf")

    expect_equal(output_dep$geom, output_new$geom)
  })
})



# 9. ddbs_line_locate_point() --------------------------------------------

## y = sf (single point) tests
## - CHECK 1.1: returns duckspatial_df by default
## - CHECK 1.2: returns numeric vector with mode sf
## - CHECK 1.3: messages shown/suppressed correctly
## - CHECK 1.4: fraction = 0.3 for reference point at (0.3, 0)
## - CHECK 1.5: materializes data correctly
## - CHECK 1.6: vector and column outputs are identical
## y = duckspatial_df (single point) tests
## - CHECK 2.1: returns duckspatial_df by default
## - CHECK 2.2: returns numeric vector with mode sf
## - CHECK 2.3: messages shown/suppressed correctly
## - CHECK 2.4: fraction = 0.3 for reference point at (0.3, 0)
## - CHECK 2.5: identical results with sf and duckspatial_df y
## y = character (DuckDB table) tests
## - CHECK 3.1: returns duckspatial_df by default
## - CHECK 3.2: returns numeric vector with mode sf
## - CHECK 3.3: table is written into the database
## - CHECK 3.4: messages shown/suppressed correctly
## - CHECK 3.5: fraction = 0.3 for reference point at (0.3, 0)
## - CHECK 3.6: materializes data correctly
## - CHECK 3.7: vector and column outputs are identical
## Check that errors work
## - CHECK 4.1: overwrite = FALSE rejects existing table
## - CHECK 4.2: y with more than 1 row (sf)
## - CHECK 4.3: y with more than 1 row (duckspatial_df)
## - CHECK 4.4: invalid y type
## - CHECK 4.5: non-character new_column argument
## - CHECK 4.6: conn required when x or y is a table name
## - CHECK 4.7: invalid x type
describe("ddbs_line_locate_point()", {

  ### y = sf

  describe("expected behavior with sf y", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf, mode = "sf")
      expect_true(is.numeric(output))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_locate_point(locate_ddbs, y = ref_point_sf))
      expect_no_message(ddbs_line_locate_point(locate_ddbs, y = ref_point_sf, quiet = TRUE))
    })

    it("returns 0.3 for reference point at (0.3, 0) on a unit horizontal line", {
      result <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf, mode = "sf")
      expect_equal(result[1], 0.3, tolerance = 1e-6)
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf)
      output_sf_mat  <- output |> st_as_sf()
      output_collect <- output |> collect()
      output_ddbs    <- output |> ddbs_collect()
      expect_identical(output_sf_mat, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf_mat, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf, mode = "sf")
      output_column <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf) |> ddbs_collect()
      expect_equal(output_vec, output_column$line_fraction)
    })
  })


  ### y = duckspatial_df

  describe("expected behavior with duckspatial_df y", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_line_locate_point(locate_ddbs, y = ref_point_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_line_locate_point(locate_ddbs, y = ref_point_ddbs, mode = "sf")
      expect_true(is.numeric(output))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_locate_point(locate_ddbs, y = ref_point_ddbs))
    })

    it("returns 0.3 for reference point at (0.3, 0) on a unit horizontal line", {
      result <- ddbs_line_locate_point(locate_ddbs, y = ref_point_ddbs, mode = "sf")
      expect_equal(result[1], 0.3, tolerance = 1e-6)
    })

    it("produces identical results with sf and duckspatial_df y", {
      output_sf_y   <- ddbs_line_locate_point(locate_ddbs, y = ref_point_sf,   mode = "sf")
      output_ddbs_y <- ddbs_line_locate_point(locate_ddbs, y = ref_point_ddbs, mode = "sf")
      expect_equal(output_sf_y, output_ddbs_y)
    })
  })


  ### y = character (DuckDB table)

  describe("expected behavior with character table y", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test, mode = "sf")
      expect_true(is.numeric(output))
    })

    it("writes tables to the database", {
      output <- ddbs_line_locate_point(
        "locate_test", y = "ref_point", conn = conn_test,
        name = "ll_tbl", new_column = "frac"
      )
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(
        ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test)
      )
      expect_message(
        ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test,
          name = "ll_tbl2", new_column = "frac")
      )
      expect_no_message(
        ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test, quiet = TRUE)
      )
      expect_no_message(
        ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test,
          name = "ll_tbl3", new_column = "frac", quiet = TRUE)
      )
    })

    it("returns 0.3 for reference point at (0.3, 0) on a unit horizontal line", {
      result <- ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test, mode = "sf")
      expect_equal(result[1], 0.3, tolerance = 1e-6)
    })

    it("materializes data correctly (collect, ddbs_collect)", {
      output <- ddbs_line_locate_point("locate_test", y = "ref_point", conn = conn_test)
      output_collect <- output |> collect()
      output_ddbs    <- output |> ddbs_collect()
      expect_identical(output_collect, output_ddbs)
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- ddbs_line_locate_point(
        "locate_test", y = "ref_point", conn = conn_test, mode = "sf"
      )
      output_column <- ddbs_line_locate_point(
        "locate_test", y = "ref_point", conn = conn_test
      ) |> ddbs_collect()
      expect_equal(output_vec, output_column$line_fraction)
    })
  })


  ### errors

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_line_locate_point(
        "locate_test", y = "ref_point", conn = conn_test, name = "dup_ll_tbl", new_column = "frac"
      )
      expect_error(
        ddbs_line_locate_point(
          "locate_test", y = "ref_point", conn = conn_test, name = "dup_ll_tbl", new_column = "frac"
        )
      )
    })

    it("errors if y has more than 1 row (sf)", {
      multi_pt_sf <- sf::st_sf(
        geometry = sf::st_sfc(sf::st_point(c(0.3, 0)), sf::st_point(c(0.5, 0))),
        crs = 4326
      )
      expect_error(ddbs_line_locate_point(locate_ddbs, y = multi_pt_sf))
    })

    it("errors if y has more than 1 row (duckspatial_df)", {
      multi_pt_ddbs <- as_duckspatial_df(sf::st_sf(
        geometry = sf::st_sfc(sf::st_point(c(0.3, 0)), sf::st_point(c(0.5, 0))),
        crs = 4326
      ))
      expect_error(ddbs_line_locate_point(locate_ddbs, y = multi_pt_ddbs))
    })

    it("errors if y is an invalid type", {
      expect_error(ddbs_line_locate_point(locate_ddbs, y = 123))
      expect_error(ddbs_line_locate_point(locate_ddbs, y = TRUE))
    })

    it("errors if new_column is not a character scalar", {
      expect_error(ddbs_line_locate_point(locate_ddbs, y = ref_point_sf, new_column = 1))
    })

    it("requires conn when x or y is a table name", {
      expect_error(ddbs_line_locate_point("locate_test", y = ref_point_sf, conn = NULL))
      expect_error(ddbs_line_locate_point(locate_ddbs,   y = "ref_point",  conn = NULL))
    })

    it("errors on invalid x type", {
      expect_error(ddbs_line_locate_point(999,  y = ref_point_sf))
      expect_error(ddbs_line_locate_point(TRUE, y = ref_point_sf))
    })
  })
})



# 10. ddbs_line_node() ---------------------------------------------------

## Two crossing lines in a single MULTILINESTRING feature -> noding splits them
## at the crossing point (1,1), yielding a MULTILINESTRING of 4 parts.
cross_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(
    sf::st_multilinestring(list(
      matrix(c(0,0, 2,2), ncol = 2, byrow = TRUE),
      matrix(c(0,2, 2,0), ncol = 2, byrow = TRUE)
    )),
    crs = 4326
  )
)
cross_ddbs <- as_duckspatial_df(cross_sf)

describe("ddbs_line_node()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_line_node(rivers_ddbs)
      output_sf   <- ddbs_line_node(rivers_sf)
      output_conn <- ddbs_line_node("rivers", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf,   "duckspatial_df")
      expect_s3_class(output_conn, "duckspatial_df")
    })

    it("returns an sf object with mode = 'sf'", {
      expect_s3_class(ddbs_line_node(rivers_ddbs, mode = "sf"), "sf")
    })

    it("splits crossing lines and returns a MULTILINESTRING", {
      out <- ddbs_line_node(cross_ddbs, mode = "sf")
      expect_equal(as.character(sf::st_geometry_type(out)), "MULTILINESTRING")
      ## two crossing lines -> 4 segments after noding at (1,1)
      expect_equal(length(sf::st_geometry(out)[[1]]), 4L)
    })

    it("preserves the CRS", {
      expect_equal(sf::st_crs(ddbs_line_node(cross_ddbs, mode = "sf")), sf::st_crs(4326))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_line_node(rivers_ddbs))
      expect_message(ddbs_line_node("rivers", conn = conn_test, name = "noded"))
      expect_no_message(
        ddbs_line_node("rivers", conn = conn_test, name = "noded", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("writes a table when name is provided", {
      expect_true(ddbs_line_node("rivers", conn = conn_test, name = "noded2"))
      expect_true(DBI::dbExistsTable(conn_test, "noded2"))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_line_node("rivers", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_line_node(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_line_node(rivers_ddbs, conn = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_line_node(x = "999", conn = conn_test))
    })
  })
})



## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' Rotate geometries around their centroid
#'
#' Rotates geometries by a specified angle around their centroid (or another center), 
#' preserving their shape.
#'
#' @template x
#' @param angle a numeric value specifying the rotation angle
#' @param units character string specifying angle units: "degrees" (default) or "radians"
#' @template by_feature
#' @param center_x numeric value for the X coordinate of rotation center. If NULL,
#' rotates around the centroid of each geometry
#' @param center_y numeric value for the Y coordinate of rotation center. If NULL,
#' rotates around the centroid of each geometry
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#'
#' ## store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' ## rotate 45 degrees
#' ddbs_rotate(conn = conn, "argentina", angle = 45)
#'
#' ## rotate 90 degrees around a specific point
#' ddbs_rotate(conn = conn, "argentina", angle = 90, center_x = -64, center_y = -34)
#'
#' ## rotate without using a connection
#' ddbs_rotate(argentina_ddbs, angle = 45)
#' }
ddbs_rotate <- function(
    x,
    angle,
    units = c("degrees", "radians"),
    by_feature = FALSE,
    center_x = NULL,
    center_y = NULL,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    ## 0. Handle errors
    assert_xy(x, "x")
    assert_numeric(angle, "angle")
    units <- match.arg(units)
    assert_logic(by_feature, "by_feature")
    assert_name(name)
    assert_conn_character(conn, x)
    assert_conn_x_name(conn, x, name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

    ## validate center coordinates
    if (!is.null(center_x) && !is.numeric(center_x)) {
        cli::cli_abort("center_x must be numeric", call. = FALSE)
    }
    if (!is.null(center_y) && !is.numeric(center_y)) {
        cli::cli_abort("center_y must be numeric", call. = FALSE)
    }
    if ((!is.null(center_x) && is.null(center_y)) ||
        (is.null(center_x) && !is.null(center_y))) {
        cli::cli_abort("Both center_x and center_y must be provided together or both NULL", call. = FALSE)
    }

    ## validate by_feature and center interaction
    if (!is.null(center_x) && !by_feature) {
        cli::cli_abort("center_x and center_y cannot be used when by_feature = FALSE", call. = FALSE)
    }

    # 1. Manage connection to DB

    ## 1.1. Pre-extract attributes (CRS and geometry column name)
    ## this step should be before normalize_spatial_input()
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")

    ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names
    x <- normalize_spatial_input(x, conn)

    ## 1.3. Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)


    # 2. Manage connection to DB

    ## 2.1. Resolve connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 2.2. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 3. Prepare parameters for the query

    ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)


    ## 3.2. Convert angle to radians if needed
    if (units == "degrees") {
        angle_rad <- angle * pi / 180
    } else {
        angle_rad <- angle
    }

    ## 3.3. Calculate rotation matrix parameters
    cos_angle <- cos(angle_rad)
    sin_angle <- sin(angle_rad)

    ## 3.5. Build rotation query
    if (by_feature) {
        # Rotate each feature around its own centroid or specified center
        if (is.null(center_x)) {
            # Rotate around each geometry's centroid
            rotation_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom}, -ST_X(ST_Centroid({x_geom})), -ST_Y(ST_Centroid({x_geom}))),
                    {cos_angle}, {-sin_angle}, {sin_angle}, {cos_angle},
                    ST_X(ST_Centroid({x_geom})), ST_Y(ST_Centroid({x_geom}))
                )"
            )
        } else {
            # Rotate around specified center point
            rotation_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom}, {-center_x}, {-center_y}),
                    {cos_angle}, {-sin_angle}, {sin_angle}, {cos_angle},
                    {center_x}, {center_y}
                )"
            )
        }
    } else {
        # Rotate all features together around the dataset's overall centroid
        rotation_expr <- glue::glue(
            "ST_Affine(
                ST_Translate({x_geom},
                    -(SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                    -(SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})),
                {cos_angle}, {-sin_angle}, {sin_angle}, {cos_angle},
                (SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                (SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})
            )"
        )
    }

    ## 3.4. Build base query
    base.query <- glue::glue("
      SELECT *
      REPLACE ({build_geom_query(rotation_expr, name, crs_x, mode)} AS {x_geom})
      FROM {x_list$query_name};
    ")

  
    # 4. if name is not NULL (i.e. no SF returned)
    if (!is.null(name)) {

        ## convenient names of table and/or schema.table
        name_list <- get_query_name(name)

        ## handle overwrite
        overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

        ## create query
        tmp.query <- glue::glue("
            CREATE TABLE {name_list$query_name} AS
            {base.query}
        ")
        ## execute rotation query
        DBI::dbExecute(target_conn, tmp.query)
        feedback_query(quiet)
        return(invisible(TRUE))
    }

    # 5. Apply geospatial operation
  
    
  
    ## 5.1. Create the query based on output
    result <- ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_x,
        x_geom     = x_geom
    )

    return(result)
}





#' Rotate 3D geometries around an axis
#'
#' Rotates 3D geometries by a specified angle around the X, Y, or Z axis, 
#' preserving their shape.
#'
#' @template x
#' @param angle a numeric value specifying the rotation angle
#' @param units character string specifying angle units: "degrees" (default) or "radians"
#' @param axis character string specifying the rotation axis: "x", "y", or "z" (default = "x").
#' The geometry rotates around this axis
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(dplyr)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read 3D data
#' countries_ddbs <- ddbs_open_dataset(
#'  system.file("spatial/countries.geojson", 
#'  package = "duckspatial")
#' ) |>
#'   filter(CNTR_ID %in% c("PT", "ES", "FR", "IT"))
#'
#' ## store in duckdb
#' ddbs_write_table(conn, countries_ddbs, "countries")
#'
#' ## rotate 45 degrees around X axis (pitch)
#' ddbs_rotate_3d(conn = conn, "countries", angle = 45, axis = "x")
#'
#' ## rotate 90 degrees around Y axis (yaw)
#' ddbs_rotate_3d(conn = conn, "countries", angle = 30, axis = "y")
#'
#' ## rotate 180 degrees around Z axis (roll)
#' ddbs_rotate_3d(conn = conn, "countries", angle = 180, axis = "z")
#'
#' ## rotate without using a connection
#' ddbs_rotate_3d(countries_ddbs, angle = 45, axis = "z")
#' }
ddbs_rotate_3d <- function(
    x,
    angle,
    units = c("degrees", "radians"),
    axis = "x",
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  
    # 0. Handle function-specific errors
    assert_numeric(angle, "angle")
    units <- match.arg(units)
    assert_name(units, "units")
    assert_name(axis, "axis")
  
  
    # 1. Build ST_Rotate parameters string
    #  Convert angle to radians if needed
  
    if (units == "degrees") {
        rotate_3d_args <- angle * pi / 180
    } else {
        rotate_3d_args <- angle
    }
  
    # 2. Pass to template
    template_unary_ops(
      x = x,
      conn = conn,
      name = name,
      mode = mode,
      overwrite = overwrite,
      quiet = quiet,
      fun = glue::glue("ST_Rotate{axis}"),
      other_args = rotate_3d_args
    )

}






#' Shift geometries by X and Y offsets
#'
#' Translates geometries by specified X and Y distances, moving 
#' them without altering their shape or orientation.
#'
#' @template x
#' @param dx numeric value specifying the shift in the X direction (longitude/easting)
#' @param dy numeric value specifying the shift in the Y direction (latitude/northing)
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' ## shift 10 degrees east and 5 degrees north
#' ddbs_shift(conn = conn, "argentina", dx = 10, dy = 5)
#'
#' ## shift without using a connection
#' ddbs_shift(argentina_ddbs, dx = 10, dy = 5)
#' }
ddbs_shift <- function(
    x,
    dx = 0,
    dy = 0,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Handle function-specific errors
    assert_numeric(dx, "dx")
    assert_numeric(dy, "dy")
  
  
    # 1. Build shift expression using ST_Affine
    # Identity matrix (no rotation/scaling) with translation offsets
    shift_args <- glue::glue("1, 0, 0, 1, {dx}, {dy}")
  
  
    # 2. Pass to template
    template_unary_ops(
      x = x,
      conn = conn,
      name = name,
      mode = mode,
      overwrite = overwrite,
      quiet = quiet,
      fun = "ST_Affine",
      other_args = shift_args
    )

  
}





#' Flip geometries horizontally or vertically
#'
#' Reflects geometries across their centroid. By default, flipping is applied 
#' relative to the centroid of all geometries; if `by_feature = TRUE`, each 
#' geometry is flipped relative to its own centroid.
#'
#' @template x
#' @param direction character string specifying the flip direction: "horizontal" (default)
#' or "vertical". Horizontal flips across the Y-axis (left-right), vertical flips across
#' the X-axis (top-bottom)
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' ## flip all features together as a whole (default)
#' ddbs_flip(conn = conn, "argentina", direction = "horizontal", by_feature = FALSE)
#'
#' ## flip each feature independently
#' ddbs_flip(conn = conn, "argentina", direction = "horizontal", by_feature = TRUE)
#'
#' ## flip without using a connection
#' ddbs_flip(argentina_ddbs, direction = "horizontal")
#' }
ddbs_flip <- function(
    x,
    direction = c("horizontal", "vertical"),
    by_feature = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    ## 0. Handle errors
    assert_xy(x, "x")
    direction <- match.arg(direction)
    assert_name(direction, "direction")
    assert_logic(by_feature, "by_feature")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

  
    # 1. Manage connection to DB

    ## 1.1. Pre-extract attributes (CRS and geometry column name)
    ## this step should be before normalize_spatial_input()
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")

    ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names
    x <- normalize_spatial_input(x, conn)

    ## 1.3. Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)


    # 2. Manage connection to DB

    ## 2.1. Resolve connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 2.2. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 3. Prepare parameters for the query

    ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 3.2. Build flip expression using ST_Affine
    if (by_feature) {
        # Flip each feature around its own centroid
        if (direction == "horizontal") {
            # Flip left-right around each feature's centroid X
            flip_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom}, -ST_X(ST_Centroid({x_geom})), 0),
                    -1, 0, 0, 1,
                    ST_X(ST_Centroid({x_geom})), 0
                )"
            )
        } else {
            # Flip top-bottom around each feature's centroid Y
            flip_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom}, 0, -ST_Y(ST_Centroid({x_geom}))),
                    1, 0, 0, -1,
                    0, ST_Y(ST_Centroid({x_geom}))
                )"
            )
        }
    } else {
        # Flip all features together around the dataset's overall centroid
        # Need to calculate the centroid of all geometries combined
        if (direction == "horizontal") {
            # Flip left-right around overall centroid X
            flip_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom},
                        -(SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                        0),
                    -1, 0, 0, 1,
                    (SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                    0
                )"
            )
        } else {
            # Flip top-bottom around overall centroid Y
            flip_expr <- glue::glue(
                "ST_Affine(
                    ST_Translate({x_geom},
                        0,
                        -(SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})),
                    1, 0, 0, -1,
                    0,
                    (SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})
                )"
            )
        }
    }

    ## 3.3. Build base query
    base.query <- glue::glue("
      SELECT *
      REPLACE ({build_geom_query(flip_expr, name, crs_x, mode)} AS {x_geom})
      FROM {x_list$query_name};
    ")
  

    # 4. if name is not NULL
    if (!is.null(name)) {

        ## convenient names of table and/or schema.table
        name_list <- get_query_name(name)

        ## handle overwrite
        overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

        ## create query 
        tmp.query <- glue::glue("
            CREATE TABLE {name_list$query_name} AS
            {base.query}
        ")
        ## execute flip query
        DBI::dbExecute(target_conn, tmp.query)
        feedback_query(quiet)
        return(invisible(TRUE))
    }


    # 5. Apply geospatial operation
    result <- ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_x,
        x_geom     = x_geom
    )

    return(result)
}





#' Scale geometries by X and Y factors
#'
#' Resizes geometries by specified X and Y scale factors. By default, scaling is 
#' performed relative to the centroid of all geometries; if `by_feature = TRUE`, 
#' each geometry is scaled relative to its own centroid.
#'
#' @template x
#' @param x_scale numeric value specifying the scaling factor in the X direction (default = 1)
#' @param y_scale numeric value specifying the scaling factor in the Y direction (default = 1)
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(dplyr)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |>
#'   filter(CNTR_ID %in% c("PT", "ES", "FR", "IT"))
#'
#' ## store in duckdb
#' ddbs_write_table(conn, countries_ddbs, "countries")
#'
#' ## scale to 150% in both directions
#' ddbs_scale(conn = conn, "countries", x_scale = 1.5, y_scale = 1.5)
#'
#' ## scale to 200% horizontally, 50% vertically
#' ddbs_scale(conn = conn, "countries", x_scale = 2, y_scale = 0.5)
#'
#' ## scale all features together (default)
#' ddbs_scale(countries_ddbs, x_scale = 1.5, y_scale = 1.5, by_feature = FALSE)
#'
#' ## scale each feature independently
#' ddbs_scale(countries_ddbs, x_scale = 1.5, y_scale = 1.5, by_feature = TRUE)
#'
#' }
ddbs_scale <- function(
    x,
    x_scale = 1,
    y_scale = 1,
    by_feature = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    ## 0. Handle errors
    assert_xy(x, "x")
    assert_numeric(x_scale, "x_scale")
    assert_numeric(y_scale, "y_scale")
    assert_logic(by_feature, "by_feature")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

  
    # 1. Manage connection to DB
  
    ## 1.1. Pre-extract attributes (CRS and geometry column name)
    ## this step should be before normalize_spatial_input()
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")

    ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names
    x <- normalize_spatial_input(x, conn)

    ## 1.3. Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)


    # 2. Manage connection to DB

    ## 2.1. Resolve connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 2.2. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 3. Prepare parameters for the query

    ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 3.2. Build scale expression using ST_Scale
    if (by_feature) {
        # Scale each feature around its own centroid
        # ST_Scale scales around origin (0,0), so translate to origin, scale, translate back
        scale_expr <- glue::glue(
            "ST_Translate(
                ST_Scale(
                    ST_Translate({x_geom}, -ST_X(ST_Centroid({x_geom})), -ST_Y(ST_Centroid({x_geom}))),
                    {x_scale}, {y_scale}
                ),
                ST_X(ST_Centroid({x_geom})), ST_Y(ST_Centroid({x_geom}))
            )"
        )
    } else {
        # Scale all features together around the dataset's overall centroid
        scale_expr <- glue::glue(
            "ST_Translate(
                ST_Scale(
                    ST_Translate({x_geom},
                        -(SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                        -(SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})),
                    {x_scale}, {y_scale}
                ),
                (SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                (SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})
            )"
        )
    }

    ## 3.3. Build base query
    base.query <- glue::glue("
      SELECT *
      REPLACE ({build_geom_query(scale_expr, name, crs_x, mode)} AS {x_geom})
      FROM {x_list$query_name};
    ")

  
    # 4. if name is not NULL (i.e. no SF returned)
    if (!is.null(name)) {

        ## convenient names of table and/or schema.table
        name_list <- get_query_name(name)

        ## handle overwrite
        overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

        ## create query 
        tmp.query <- glue::glue("
            CREATE TABLE {name_list$query_name} AS
            {base.query}
        ")
        ## execute scale query
        DBI::dbExecute(target_conn, tmp.query)
        feedback_query(quiet)
        return(invisible(TRUE))
    }

    # 5. Apply geospatial operation
    result <- ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_x,
        x_geom     = x_geom
    )

    return(result)
}





#' Shear geometries
#'
#' Applies a shear transformation to geometries, shifting coordinates proportionally 
#' in the X and Y directions. By default, shearing is applied relative to the centroid 
#' of all geometries; if `by_feature = TRUE`, each geometry is sheared relative to its 
#' own centroid.
#'
#' @template x
#' @param x_shear numeric value specifying the shear factor in the X direction (default = 0).
#' For each unit in Y, X coordinates are shifted by this amount
#' @param y_shear numeric value specifying the shear factor in the Y direction (default = 0).
#' For each unit in X, Y coordinates are shifted by this amount
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(dplyr)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |>
#'   filter(CNTR_ID %in% c("PT", "ES", "FR", "IT"))
#'
#' ## store in duckdb
#' ddbs_write_table(conn, countries_ddbs, "countries")
#'
#' ## shear in X direction (creates italic-like effect)
#' ddbs_shear(conn = conn, "countries", x_shear = 0.3, y_shear = 0)
#'
#' ## shear in Y direction
#' ddbs_shear(conn = conn, "countries", x_shear = 0, y_shear = 0.3)
#'
#' ## shear in both directions
#' ddbs_shear(conn = conn, "countries", x_shear = 0.2, y_shear = 0.2)
#'
#' ## shear without using a connection
#' ddbs_shear(countries_ddbs, x_shear = 0.3, y_shear = 0)
#' }
ddbs_shear <- function(
    x,
    x_shear = 0,
    y_shear = 0,
    by_feature = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Handle errors
    assert_xy(x, "x")
    assert_numeric(x_shear, "x_shear")
    assert_numeric(y_shear, "y_shear")
    assert_logic(by_feature, "by_feature")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

  
    # 1. Manage connection to DB

    ## 1.1. Pre-extract attributes (CRS and geometry column name)
    ## this step should be before normalize_spatial_input()
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")

    ## 1.2. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names
    x <- normalize_spatial_input(x, conn)

    ## 1.3. Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)


    # 2. Manage connection to DB

    ## 2.1. Resolve connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 2.2. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 3. Prepare parameters for the query

    ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 3.2. Build shear expression using ST_Affine
    # Shear matrix: a=1, b=x_shear, d=y_shear, e=1
    if (by_feature) {
        # Shear each feature around its own centroid
        shear_expr <- glue::glue(
            "ST_Affine(
                ST_Translate({x_geom}, -ST_X(ST_Centroid({x_geom})), -ST_Y(ST_Centroid({x_geom}))),
                1, {x_shear}, {y_shear}, 1,
                ST_X(ST_Centroid({x_geom})), ST_Y(ST_Centroid({x_geom}))
            )"
        )
    } else {
        # Shear all features together around the dataset's overall centroid
        shear_expr <- glue::glue(
            "ST_Affine(
                ST_Translate({x_geom},
                    -(SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                    -(SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})),
                1, {x_shear}, {y_shear}, 1,
                (SELECT ST_X(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name}),
                (SELECT ST_Y(ST_Centroid(ST_Union_Agg({x_geom}))) FROM {x_list$query_name})
            )"
        )
    }

    ## 3.3. Build base query
    base.query <- glue::glue("
      SELECT *
      REPLACE ({build_geom_query(shear_expr, name, crs_x, mode)} AS {x_geom})
      FROM {x_list$query_name};
    ")


    # 4. if name is not NULL (i.e. no SF returned)
    if (!is.null(name)) {

        ## convenient names of table and/or schema.table
        name_list <- get_query_name(name)

        ## handle overwrite
        overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

        ## create query 
        tmp.query <- glue::glue("
            CREATE TABLE {name_list$query_name} AS
            {base.query}
        ")
        ## execute shear query
        DBI::dbExecute(target_conn, tmp.query)
        feedback_query(quiet)
        return(invisible(TRUE))
    }

    
    # 5. Apply geospatial operation
    result <- ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_x,
        x_geom     = x_geom
    )

    return(result)
}





#' Apply an affine transformation to geometries
#'
#' Applies an affine transformation to geometries using a 2x3 or 3x4 matrix.
#'
#' @template x
#' @param matrix A numeric matrix specifying the transformation:
#'   \itemize{
#'     \item 2x3: 2D transformation with rows \code{[a, b, xoff]} and \code{[d, e, yoff]}
#'     \item 3x4: 3D transformation with rows \code{[a, b, c, xoff]}, \code{[d, e, f, yoff]}, and \code{[g, h, i, zoff]}
#'   }
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## 2D translation (shift x by 1, y by 2)
#' mat_2d <- matrix(c(1, 0, 1,
#'                    0, 1, 2), nrow = 2, byrow = TRUE)
#' ddbs_affine(argentina_ddbs, mat_2d)
#'
#' ## 2D scaling (scale x by 2, y by 3)
#' mat_scale <- matrix(c(2, 0, 0,
#'                       0, 3, 0), nrow = 2, byrow = TRUE)
#' ddbs_affine(argentina_ddbs, mat_scale)
#'
#' ## 3D transformation
#' mat_3d <- matrix(c(1, 0, 0, 1,
#'                    0, 1, 0, 2,
#'                    0, 0, 1, 0), nrow = 3, byrow = TRUE)
#' ddbs_affine(argentina_ddbs, mat_3d)
#' }
ddbs_affine <- function(
    x,
    matrix,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {


  # 0. Handle function-specific errors

  ## Validate matrix input
  if (!is.matrix(matrix) || !is.numeric(matrix)) {
    cli::cli_abort("{.arg matrix} must be a numeric matrix.")
  }

  dims <- dim(matrix)
  if (!identical(dims, c(2L, 3L)) && !identical(dims, c(3L, 4L))) {
    cli::cli_abort("{.arg matrix} must be a 2x3 (2D) or 3x4 (3D) numeric matrix.")
  }


  # 1. Build ST_Affine parameters string
  if (identical(dims, c(2L, 3L))) {
    ## 2D: ST_Affine(geom, a, b, d, e, xoff, yoff)
    affine_args <- glue::glue(
      "{matrix[1,1]}, {matrix[1,2]}, {matrix[2,1]}, {matrix[2,2]}, {matrix[1,3]}, {matrix[2,3]}"
    )
  } else {
    ## 3D: ST_Affine(geom, a, b, c, d, e, f, g, h, i, xoff, yoff, zoff)
    affine_args <- glue::glue(
      "{matrix[1,1]}, {matrix[1,2]}, {matrix[1,3]}, {matrix[2,1]}, {matrix[2,2]}, 
      {matrix[2,3]}, {matrix[3,1]}, {matrix[3,2]}, {matrix[3,3]}, {matrix[1,4]}, 
      {matrix[2,4]}, {matrix[3,4]}"
    )
  }


  # 2. Pass to template
  template_unary_ops(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Affine",
    other_args = affine_args
  )

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write some data
ddbs_write_table(conn_test, argentina_sf, "argentina")
ddbs_write_table(conn_test, nc_sf, "nc")

# 1. ddbs_rotate() -------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_rotate()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_rotate(argentina_ddbs, 45)
      output_ddbs_2 <- ddbs_rotate(argentina_ddbs, 45, units = "radians")
      output_ddbs_3 <- ddbs_rotate(argentina_ddbs, 45, by_feature = TRUE)
      output_ddbs_4 <- ddbs_rotate(argentina_ddbs, 45, by_feature = TRUE, center_x = 0, center_y = 0)
      output_ddbs_5 <- ddbs_rotate(argentina_ddbs, 45, quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
      expect_s3_class(output_ddbs_5, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_rotate(argentina_ddbs, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_rotate(argentina_sf, 45)
      output_sf_2 <- ddbs_rotate(argentina_sf, 45, units = "radians")
      output_sf_3 <- ddbs_rotate(argentina_sf, 45, by_feature = TRUE)
      output_sf_4 <- ddbs_rotate(argentina_sf, 45, by_feature = TRUE, center_x = 0, center_y = 0)
      output_sf_5 <- ddbs_rotate(argentina_sf, 45, quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
      expect_s3_class(output_sf_5, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_rotate(argentina_sf, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_rotate("argentina", 45, conn = conn_test)
      output_conn_2 <- ddbs_rotate("argentina", 45, conn = conn_test, units = "radians")
      output_conn_3 <- ddbs_rotate("argentina", 45, conn = conn_test, by_feature = TRUE)
      output_conn_4 <- ddbs_rotate("argentina", 45, conn = conn_test, by_feature = TRUE, center_x = 0, center_y = 0)
      output_conn_5 <- ddbs_rotate("argentina", 45, conn = conn_test, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
      expect_s3_class(output_conn_5, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_rotate("argentina", 45, conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_rotate(argentina_ddbs, 50))
      expect_message(ddbs_rotate("argentina", 50, conn = conn_test, name = "rotated"))
      expect_message(ddbs_rotate("argentina", 50, conn = conn_test, name = "rotated", overwrite = TRUE))
      expect_true(ddbs_rotate("argentina", 50, conn = conn_test, name = "rotated2"))

      expect_no_message(ddbs_rotate(argentina_ddbs, 50, quiet = TRUE))
      expect_no_message(
        ddbs_rotate(
          "argentina",
          50,
          conn = conn_test,
          name = "rotated",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_rotate("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_rotate(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_rotate(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_rotate(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_rotate(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_rotate(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_rotate(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_rotate(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })

    it("validates required argument combinations", {
      expect_error(ddbs_rotate(argentina_ddbs))
      expect_error(ddbs_rotate(argentina_ddbs, by_feature = FALSE, center_x = 5, center_y = 5))
      expect_error(ddbs_rotate(argentina_ddbs, by_feature = TRUE, center_x = 5))
      expect_error(ddbs_rotate(argentina_ddbs, by_feature = TRUE, center_y = 5))
    })
  })
    
})




# 2. ddbs_rotate_3d ------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_rotate_3d()", {

  ### EXPECTED BEHAVIOR

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_rotate_3d(argentina_ddbs, 45)
      output_ddbs_2 <- ddbs_rotate_3d(argentina_ddbs, 45, units = "radians")
      output_ddbs_3 <- ddbs_rotate_3d(argentina_ddbs, 90, axis = "y")
      output_ddbs_4 <- ddbs_rotate_3d(argentina_ddbs, 180, axis = "z")
      output_ddbs_5 <- ddbs_rotate_3d(argentina_ddbs, 45, quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
      expect_s3_class(output_ddbs_5, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_rotate_3d(argentina_ddbs, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_rotate_3d(argentina_sf, 45)
      output_sf_2 <- ddbs_rotate_3d(argentina_sf, 45, units = "radians")
      output_sf_3 <- ddbs_rotate_3d(argentina_sf, 90, axis = "y")
      output_sf_4 <- ddbs_rotate_3d(argentina_sf, 180, axis = "z")
      output_sf_5 <- ddbs_rotate_3d(argentina_sf, 45, quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
      expect_s3_class(output_sf_5, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_rotate_3d(argentina_sf, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_rotate_3d("argentina", 45, conn = conn_test)
      output_conn_2 <- ddbs_rotate_3d("argentina", 45, conn = conn_test, units = "radians")
      output_conn_3 <- ddbs_rotate_3d("argentina", 90, conn = conn_test, axis = "y")
      output_conn_4 <- ddbs_rotate_3d("argentina", 180, conn = conn_test, axis = "z")
      output_conn_5 <- ddbs_rotate_3d("argentina", 45, conn = conn_test, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
      expect_s3_class(output_conn_5, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_rotate_3d("argentina", 45, conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_rotate_3d(argentina_ddbs, 50))
      expect_message(ddbs_rotate_3d("argentina", 50, conn = conn_test, name = "rotated_3d"))
      expect_message(ddbs_rotate_3d("argentina", 50, conn = conn_test, name = "rotated_3d", overwrite = TRUE))
      expect_true(ddbs_rotate_3d("argentina", 50, conn = conn_test, name = "rotated_3d2"))

      expect_no_message(ddbs_rotate_3d(argentina_ddbs, 50, quiet = TRUE))
      expect_no_message(
        ddbs_rotate_3d(
          "argentina",
          50,
          conn = conn_test,
          name = "rotated_3d",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_rotate_3d("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_rotate_3d(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_rotate_3d(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })

    it("validates units argument", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, units = "asdfasdf"))
    })

    it("validates axis argument", {
      expect_error(ddbs_rotate_3d(argentina_ddbs, axis = "asdfasdf"))
    })

    it("validates required argument combinations", {
      expect_error(ddbs_rotate_3d(argentina_ddbs))
    })
  })
})



# 3. ddbs_shift ------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_shift()", {

  ### EXPECTED BEHAVIOR

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_shift(argentina_ddbs, 10)
      output_ddbs_2 <- ddbs_shift(argentina_ddbs, 10, 20)
      output_ddbs_3 <- ddbs_shift(argentina_ddbs, dy = 10)
      output_ddbs_4 <- ddbs_shift(argentina_ddbs, 45, quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_shift(argentina_ddbs, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_shift(argentina_sf, 10)
      output_sf_2 <- ddbs_shift(argentina_sf, 10, 20)
      output_sf_3 <- ddbs_shift(argentina_sf, dy = 10)
      output_sf_4 <- ddbs_shift(argentina_sf, 45, quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_shift(argentina_sf, 45, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_shift("argentina", 10, conn = conn_test)
      output_conn_2 <- ddbs_shift("argentina", 10, 20, conn = conn_test)
      output_conn_3 <- ddbs_shift("argentina", dy = 10, conn = conn_test)
      output_conn_4 <- ddbs_shift("argentina", 45, conn = conn_test, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_shift("argentina", 45, conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_shift(argentina_ddbs, 50))
      expect_message(ddbs_shift("argentina", 50, conn = conn_test, name = "shift"))
      expect_message(ddbs_shift("argentina", 50, conn = conn_test, name = "shift", overwrite = TRUE))
      expect_true(ddbs_shift("argentina", 50, conn = conn_test, name = "shift2"))

      expect_no_message(ddbs_shift(argentina_ddbs, 50, quiet = TRUE))
      expect_no_message(
        ddbs_shift(
          "argentina",
          50,
          conn = conn_test,
          name = "shift",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_shift("argentina", conn = NULL))
    })

    it("validates dx argument type", {
      expect_error(ddbs_shift(argentina_ddbs, dx = "10"))
    })

    it("validates dy argument type", {
      expect_error(ddbs_shift(argentina_ddbs, dy = "banana"))
    })

    it("validates x argument type", {
      expect_error(ddbs_shift(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_shift(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_shift(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_shift(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_shift(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_shift(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_shift(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 4. ddbs_flip ------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_flip()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_flip(argentina_ddbs)
      output_ddbs_2 <- ddbs_flip(argentina_ddbs, "vertical")
      output_ddbs_3 <- ddbs_flip(nc_ddbs, by_feature = TRUE)
      output_ddbs_4 <- ddbs_flip(nc_ddbs, "vertical", by_feature = TRUE, quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_flip(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_flip(argentina_sf)
      output_sf_2 <- ddbs_flip(argentina_sf, "vertical")
      output_sf_3 <- ddbs_flip(nc_ddbs, by_feature = TRUE)
      output_sf_4 <- ddbs_flip(nc_sf, "vertical", by_feature = TRUE, quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_flip(argentina_sf, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_flip("nc", conn = conn_test)
      output_conn_2 <- ddbs_flip("nc", "vertical", conn = conn_test)
      output_conn_3 <- ddbs_flip("nc", by_feature = TRUE, conn = conn_test)
      output_conn_4 <- ddbs_flip("nc", "vertical", conn = conn_test, by_feature = TRUE, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_flip("nc", conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_flip(nc_ddbs))
      expect_message(ddbs_flip("nc", conn = conn_test, name = "flip"))
      expect_message(ddbs_flip("nc", conn = conn_test, name = "flip", overwrite = TRUE))
      expect_true(ddbs_flip("nc", conn = conn_test, name = "flip2"))

      expect_no_message(ddbs_flip(argentina_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_flip(
          "nc",
          conn = conn_test,
          name = "flip",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_flip("argentina", conn = NULL))
    })

    it("validates direction argument", {
      expect_error(ddbs_flip(argentina_ddbs, direction = "uptodown"))
    })

    it("validates by_feature argument type", {
      expect_error(ddbs_flip(argentina_ddbs, by_feature = "TRUE"))
    })

    it("validates x argument type", {
      expect_error(ddbs_flip(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_flip(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_flip(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_flip(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_flip(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_flip(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_flip(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 5. ddbs_scale ------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_scale()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_scale(argentina_ddbs)
      output_ddbs_2 <- ddbs_scale(argentina_ddbs, y_scale = -1)
      output_ddbs_3 <- ddbs_scale(nc_ddbs, by_feature = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_scale(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_scale(argentina_sf)
      output_sf_2 <- ddbs_scale(argentina_sf, y_scale = -1)
      output_sf_3 <- ddbs_scale(nc_ddbs, by_feature = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_scale(argentina_sf, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_scale("nc", conn = conn_test)
      output_conn_2 <- ddbs_scale("nc", y_scale = -1, conn = conn_test)
      output_conn_3 <- ddbs_scale("nc", by_feature = TRUE, conn = conn_test)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_scale("nc", conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_scale(nc_ddbs))
      expect_message(ddbs_scale("nc", conn = conn_test, name = "shear"))
      expect_message(ddbs_scale("nc", conn = conn_test, name = "shear", overwrite = TRUE))
      expect_true(ddbs_scale("nc", conn = conn_test, name = "shear2"))

      expect_no_message(ddbs_scale(argentina_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_scale(
          "nc",
          conn = conn_test,
          name = "shear",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_scale("argentina", conn = NULL))
    })

    it("validates x_scale argument type", {
      expect_error(ddbs_scale(argentina_ddbs, x_scale = "23"))
    })

    it("validates y_scale argument type", {
      expect_error(ddbs_scale(argentina_ddbs, y_scale = "five"))
    })

    it("validates by_feature argument type", {
      expect_error(ddbs_scale(argentina_ddbs, by_feature = "TRUE"))
    })

    it("validates x argument type", {
      expect_error(ddbs_scale(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_scale(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_scale(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_scale(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_scale(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_scale(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_scale(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 6. ddbs_shear ------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_shear()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_shear(argentina_ddbs)
      output_ddbs_2 <- ddbs_shear(argentina_ddbs, y_shear = -1)
      output_ddbs_3 <- ddbs_shear(nc_ddbs, by_feature = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_shear(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_shear(argentina_sf)
      output_sf_2 <- ddbs_shear(argentina_sf, y_shear = -1)
      output_sf_3 <- ddbs_shear(nc_ddbs, by_feature = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_shear(argentina_sf, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_shear("nc", conn = conn_test)
      output_conn_2 <- ddbs_shear("nc", y_shear = -1, conn = conn_test)
      output_conn_3 <- ddbs_shear("nc", by_feature = TRUE, conn = conn_test)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_shear("nc", conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_shear(nc_ddbs))
      expect_message(ddbs_shear("nc", conn = conn_test, name = "scale"))
      expect_message(ddbs_shear("nc", conn = conn_test, name = "scale", overwrite = TRUE))
      expect_true(ddbs_shear("nc", conn = conn_test, name = "scale2"))

      expect_no_message(ddbs_shear(argentina_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_shear(
          "nc",
          conn = conn_test,
          name = "scale",
          overwrite = TRUE,
          quiet = TRUE
        )
      )
    })
  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_shear("argentina", conn = NULL))
    })

    it("validates x_scale argument type", {
      expect_error(ddbs_shear(argentina_ddbs, x_scale = "23"))
    })

    it("validates y_shear argument type", {
      expect_error(ddbs_shear(argentina_ddbs, y_shear = "five"))
    })

    it("validates by_feature argument type", {
      expect_error(ddbs_shear(argentina_ddbs, by_feature = "TRUE"))
    })

    it("validates x argument type", {
      expect_error(ddbs_shear(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_shear(argentina_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_shear(argentina_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_shear(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_shear(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_shear(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_shear(argentina_ddbs, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



# 7. ddbs_affine -------------------------------------------------------

## - CHECK 1.1: works on ddbs
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: works on sf
## - CHECK 1.4: sf returns different outputs (duckspatial_df, sf)
## - CHECK 1.5: works on duckdb table
## - CHECK 1.6: duckdb table returns different outputs (duckspatial_df, sf)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 1.9: 2D identity matrix preserves geometry
## - CHECK 1.10: 2D translation matrix shifts geometry correctly
## - CHECK 1.11: 2D scaling matrix scales geometry correctly
## - CHECK 1.12: 3D identity matrix preserves geometry
## - CHECK 2.1: matrix must be a matrix object
## - CHECK 2.2: matrix must be numeric
## - CHECK 2.3: matrix must be 2x3 or 3x4
## - CHECK 2.4: combination of inputs / missing arguments
## - CHECK 2.5: other errors
describe("ddbs_affine()", {

  mat_identity_2d  <- matrix(c(1, 0, 0,
                                0, 1, 0), nrow = 2, byrow = TRUE)
  mat_translate_2d <- matrix(c(1, 0, 1,
                                0, 1, 2), nrow = 2, byrow = TRUE)
  mat_scale_2d     <- matrix(c(2, 0, 0,
                                0, 2, 0), nrow = 2, byrow = TRUE)
  mat_identity_3d  <- matrix(c(1, 0, 0, 0,
                                0, 1, 0, 0,
                                0, 0, 1, 0), nrow = 3, byrow = TRUE)

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_affine(argentina_ddbs, mat_identity_2d)
      output_ddbs_2 <- ddbs_affine(argentina_ddbs, mat_translate_2d)
      output_ddbs_3 <- ddbs_affine(argentina_ddbs, mat_scale_2d)
      output_ddbs_4 <- ddbs_affine(argentina_ddbs, mat_identity_3d)
      output_ddbs_5 <- ddbs_affine(argentina_ddbs, mat_identity_2d, quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
      expect_s3_class(output_ddbs_5, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_affine(argentina_ddbs, mat_identity_2d, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_affine(argentina_sf, mat_identity_2d)
      output_sf_2 <- ddbs_affine(argentina_sf, mat_translate_2d)
      output_sf_3 <- ddbs_affine(argentina_sf, mat_scale_2d)
      output_sf_4 <- ddbs_affine(argentina_sf, mat_identity_3d)
      output_sf_5 <- ddbs_affine(argentina_sf, mat_identity_2d, quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
      expect_s3_class(output_sf_5, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_affine(argentina_sf, mat_identity_2d, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_affine("argentina", mat_identity_2d, conn = conn_test)
      output_conn_2 <- ddbs_affine("argentina", mat_translate_2d, conn = conn_test)
      output_conn_3 <- ddbs_affine("argentina", mat_scale_2d, conn = conn_test)
      output_conn_4 <- ddbs_affine("argentina", mat_identity_3d, conn = conn_test)
      output_conn_5 <- ddbs_affine("argentina", mat_identity_2d, conn = conn_test, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
      expect_s3_class(output_conn_5, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_affine("argentina", mat_identity_2d, conn = conn_test, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_affine(argentina_ddbs, mat_identity_2d))
      expect_message(ddbs_affine("argentina", mat_identity_2d, conn = conn_test, name = "affine"))
      expect_message(ddbs_affine("argentina", mat_identity_2d, conn = conn_test, name = "affine", overwrite = TRUE))
      expect_true(ddbs_affine("argentina", mat_identity_2d, conn = conn_test, name = "affine2"))

      expect_no_message(ddbs_affine(argentina_ddbs, mat_identity_2d, quiet = TRUE))
      expect_no_message(
        ddbs_affine(
          "argentina",
          mat_identity_2d,
          conn     = conn_test,
          name     = "affine",
          overwrite = TRUE,
          quiet    = TRUE
        )
      )
    })

    describe("2D matrix transformations", {

      it("identity matrix preserves geometry", {
        original <- ddbs_collect(argentina_ddbs)
        result   <- ddbs_affine(argentina_ddbs, mat_identity_2d) |> ddbs_collect()
        expect_equal(original$geometry, result$geometry)
      })

      it("translation matrix shifts coordinates by xoff and yoff", {
        original <- argentina_sf
        result   <- ddbs_affine(argentina_sf, mat_translate_2d, mode = "sf")

        coords_orig   <- sf::st_coordinates(original)[, c("X", "Y")]
        coords_result <- sf::st_coordinates(result)[, c("X", "Y")]

        expect_equal(coords_result[, "X"], coords_orig[, "X"] + 1, tolerance = 1e-6)
        expect_equal(coords_result[, "Y"], coords_orig[, "Y"] + 2, tolerance = 1e-6)
      })

      it("scaling matrix multiplies coordinates", {
        original <- argentina_sf
        result   <- ddbs_affine(argentina_sf, mat_scale_2d, mode = "sf")

        coords_orig   <- sf::st_coordinates(original)[, c("X", "Y")]
        coords_result <- sf::st_coordinates(result)[, c("X", "Y")]

        expect_equal(coords_result[, "X"], coords_orig[, "X"] * 2, tolerance = 1e-6)
        expect_equal(coords_result[, "Y"], coords_orig[, "Y"] * 2, tolerance = 1e-6)
      })
    })

    describe("3D matrix transformations", {

      it("3D identity matrix preserves geometry", {
        original <- ddbs_collect(argentina_ddbs)
        result   <- ddbs_affine(argentina_ddbs, mat_identity_3d) |> ddbs_collect()
        expect_equal(original$geometry, result$geometry)
      })

      it("3D translation matrix shifts X and Y coordinates", {
        mat_translate_3d <- matrix(c(1, 0, 0, 5,
                                     0, 1, 0, 3,
                                     0, 0, 1, 0), nrow = 3, byrow = TRUE)
        original <- argentina_sf
        result   <- ddbs_affine(argentina_sf, mat_translate_3d, mode = "sf")

        coords_orig   <- sf::st_coordinates(original)[, c("X", "Y")]
        coords_result <- sf::st_coordinates(result)[, c("X", "Y")]

        expect_equal(coords_result[, "X"], coords_orig[, "X"] + 5, tolerance = 1e-6)
        expect_equal(coords_result[, "Y"], coords_orig[, "Y"] + 3, tolerance = 1e-6)
      })
    })
  })

  ### EXPECTED ERRORS ---------------------------------------------------

  describe("errors", {

    describe("matrix argument validation", {

      it("rejects non-matrix input", {
        expect_error(ddbs_affine(argentina_ddbs, matrix = c(1, 0, 0, 0, 1, 0)))
        expect_error(ddbs_affine(argentina_ddbs, matrix = list(1, 0, 0, 0, 1, 0)))
        expect_error(ddbs_affine(argentina_ddbs, matrix = "1, 0, 0, 0, 1, 0"))
      })

      it("rejects non-numeric matrix", {
        expect_error(
          ddbs_affine(argentina_ddbs, matrix = matrix(c("a", "b", "c", "d", "e", "f"), nrow = 2))
        )
      })

      it("rejects wrong matrix dimensions", {
        expect_error(ddbs_affine(argentina_ddbs, matrix = matrix(1:4, nrow = 2)))
        expect_error(ddbs_affine(argentina_ddbs, matrix = matrix(1:6, nrow = 3)))
        expect_error(ddbs_affine(argentina_ddbs, matrix = matrix(1:9, nrow = 3)))
        expect_error(ddbs_affine(argentina_ddbs, matrix = matrix(1:12, nrow = 4)))
      })
    })

    it("requires connection when using table names", {
      expect_error(ddbs_affine("argentina", mat_identity_2d, conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_affine(x = 999, matrix = mat_identity_2d))
    })

    it("validates conn argument type", {
      expect_error(ddbs_affine(argentina_ddbs, mat_identity_2d, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_affine(argentina_ddbs, mat_identity_2d, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_affine(argentina_ddbs, mat_identity_2d, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_affine(x = "999", mat_identity_2d, conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_affine(argentina_ddbs, mat_identity_2d, conn = conn_test, name = c("banana", "banana")))
    })
  })
})



## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' Calculate geometric measurements
#'
#' Compute area, length, perimeter, or distance of geometries with automatic
#' method selection based on the coordinate reference system (CRS).
#'
#' @name ddbs_measure_funs
#' @rdname ddbs_measure_funs
#' @aliases ddbs_area ddbs_length ddbs_perimeter ddbs_distance ddbs_azimuth
#'
#' @param x Input geometry (sf object, duckspatial_df, or table name in DuckDB)
#' @param y Second input geometry for distance and azimuth calculations (sf object, duckspatial_df, or table name)
#' @param dist_type Character. Distance type to be calculated. By default it uses
#'   the best option for the input CRS (see details).
#' @param unit Character. Output unit for \code{ddbs_azimuth}: \code{"radians"}
#'   (default) or \code{"degrees"}.
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @param id_x Character; optional name of the column in `x` whose values will
#' be used to name the list elements. If `NULL`, integer row numbers of `x` are used.
#' @param id_y Character; optional name of the column in `y` whose values will
#' replace the integer indices returned in each element of the list.
#' @template new_column
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @returns 
#' For \code{ddbs_area}, \code{ddbs_length}, and \code{ddbs_perimeter}:
#' \itemize{
#'   \item \code{mode = "duckspatial"} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
#'   \item \code{mode = "sf"}: An eagerly collected vector in R memory.
#'   \item When \code{name} is provided: writes the table in the DuckDB connection and returns \code{TRUE} (invisibly).
#' }
#' 
#' For \code{ddbs_distance}: A \code{units} matrix in meters with dimensions nrow(x), nrow(y).
#'
#' For \code{ddbs_azimuth}: A numeric matrix of azimuth values (in the specified \code{unit})
#' with dimensions nrow(x) by nrow(y) when \code{mode = "sf"}, or a lazy
#' \code{tbl_duckdb_connection} with columns \code{id_x}, \code{id_y}, and \code{azimuth} otherwise.
#' Both inputs must contain only POINT geometries.
#'
#' @details
#' These functions automatically select the appropriate calculation method based on the input CRS:
#'
#' \strong{For EPSG:4326 (geographic coordinates):}
#' \itemize{
#'   \item Uses \code{ST_*_Spheroid} functions (e.g., \code{ST_Area_Spheroid}, \code{ST_Length_Spheroid})
#'   \item Leverages GeographicLib library for ellipsoidal earth model calculations
#'   \item Highly accurate but slower than planar calculations
#'   \item For \code{ddbs_distance} with POINT geometries: defaults to \code{"haversine"}
#'   \item For \code{ddbs_distance} with other geometries: defaults to \code{"spheroid"}
#' }
#'
#' \strong{For projected CRS (e.g., UTM, Web Mercator):}
#' \itemize{
#'   \item Uses planar \code{ST_*} functions (e.g., \code{ST_Area}, \code{ST_Length})
#'   \item Faster performance with accurate results in meters
#'   \item For \code{ddbs_distance}: defaults to \code{"planar"}
#' }
#'
#' \strong{Distance calculation methods} (\code{dist_type} argument):
#' \itemize{
#'   \item \code{NULL} (default): Automatically selects best method for input CRS
#'   \item \code{"planar"}: Planar distance (for projected CRS)
#'   \item \code{"geos"}: Planar distance using GEOS library (for projected CRS)
#'   \item \code{"haversine"}: Great circle distance (requires EPSG:4326 and POINT geometries)
#'   \item \code{"spheroid"}: Ellipsoidal model using GeographicLib (most accurate, slowest)
#' }
#'
#' \strong{Distance type requirements:}
#' \itemize{
#'   \item \code{"planar"} and \code{"geos"}: Require projected coordinates (not degrees)
#'   \item \code{"haversine"} and \code{"spheroid"}: Require POINT geometries and EPSG:4326
#' }
#'
#' @section Performance:
#' Speed comparison (fastest to slowest):
#' \enumerate{
#'   \item Planar calculations on projected CRS
#'   \item Haversine (spherical approximation)
#'   \item Spheroid functions (ellipsoidal model)
#' }
#'
#' @references \url{https://geographiclib.sourceforge.io/}
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#' library(dplyr)
#' 
#' # Create a DuckDB connection
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' # ===== AREA CALCULATIONS =====
#' 
#' # Load polygon data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", package = "duckspatial")
#' ) |>
#'   ddbs_transform("EPSG:3857") |> 
#'   filter(NAME_ENGL != "Antarctica")
#' 
#' # Store in DuckDB
#' ddbs_write_table(conn, countries_ddbs, "countries")
#' 
#' # Calculate area (adds a new column - area by default)
#' ddbs_area("countries", conn)
#' 
#' # Calculate area with custom column name
#' ddbs_area("countries", conn, new_column = "area_sqm")
#' 
#' # Create new table with area calculations
#' ddbs_area("countries", conn, name = "countries_with_area", new_column = "area_sqm")
#' 
#' # Calculate area from sf object directly
#' ddbs_area(countries_ddbs)
#' 
#' # Calculate area using dplyr syntax
#' countries_ddbs |> 
#'   mutate(area = ddbs_area(geom))
#' 
#' # Calculate total area 
#' countries_ddbs |> 
#'   mutate(area = ddbs_area(geom)) |> 
#'   summarise(
#'     area = sum(area),
#'     geom = ddbs_union(geom)
#'   )
#' 
#' # ===== LENGTH CALCULATIONS =====
#' 
#' # Load line data
#' rivers_ddbs <- sf::read_sf(
#'   system.file("spatial/rivers.geojson", package = "duckspatial")
#' ) |> 
#'   as_duckspatial_df()
#' 
#' # Store in DuckDB
#' ddbs_write_table(conn, rivers_ddbs, "rivers")
#' 
#' # Calculate length (add a new column - length by default)
#' ddbs_length("rivers", conn)
#' 
#' # Calculate length with custom column name
#' ddbs_length(rivers_ddbs, new_column = "length_meters")
#' 
#' # Calculate length by river name
#' rivers_ddbs |> 
#'   ddbs_union_agg("RIVER_NAME") |> 
#'   ddbs_length()
#' 
#' # Add length within dplyr
#' rivers_ddbs |> 
#'   mutate(length = ddbs_length(geometry))
#' 
#' 
#' # ===== PERIMETER CALCULATIONS =====
#' 
#' # Calculate perimeter (returns sf object with perimeter column)
#' ddbs_perimeter(countries_ddbs)
#' 
#' # Calculate perimeter within dplyr
#' countries_ddbs |> 
#'   mutate(perim = ddbs_perimeter(geom))
#' 
#' 
#' # ===== DISTANCE CALCULATIONS =====
#' 
#' # Create sample points in EPSG:4326
#' n <- 10
#' points_sf <- data.frame(
#'   id = 1:n,
#'   x = runif(n, min = -180, max = 180),
#'   y = runif(n, min = -90, max = 90)
#' ) |>
#'   ddbs_as_spatial(coords = c("x", "y"), crs = "EPSG:4326")
#' 
#' # Option 1: Using sf objects (auto-selects haversine for EPSG:4326 points)
#' dist_matrix <- ddbs_distance(x = points_sf, y = points_sf)
#' head(dist_matrix)
#' 
#' # Option 2: Explicitly specify distance type
#' dist_matrix_harv <- ddbs_distance(
#'   x = points_sf,
#'   y = points_sf,
#'   dist_type = "haversine"
#' )
#' 
#' # Option 3: Using DuckDB tables
#' ddbs_write_table(conn, points_sf, "points", overwrite = TRUE)
#' dist_matrix_sph <- ddbs_distance(
#'   conn = conn,
#'   x = "points",
#'   y = "points",
#'   dist_type = "spheroid"  # Most accurate for geographic coordinates
#' )
#' head(dist_matrix_sph)
#' 
#' # Close connection
#' ddbs_stop_conn(conn)
#' }
NULL





#' @rdname ddbs_measure_funs
#' @export
ddbs_area <- function(
  x,
  new_column = "area",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
    
  template_measure(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Area"
  )
}





#' @rdname ddbs_measure_funs
#' @export
ddbs_length <- function(
  x,
  new_column = "length",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_measure(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Length"
  )
}





#' @rdname ddbs_measure_funs
#' @export
ddbs_perimeter <- function(
  x,
  new_column = "perimeter",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
    
  template_measure(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Perimeter"
  )
}





#' @rdname ddbs_measure_funs
#' @export
ddbs_distance <- function(
  x,
  y,
  dist_type = NULL,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  id_x = NULL,
  id_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Handle errors
  assert_xy(x, "x")
  assert_xy(y, "y")
  assert_name(id_x, "id_x")
  assert_name(id_y, "id_y")
  assert_name(dist_type, "dist_type")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


  # 1. Manage connection to DB

  ## 1.1. Pre-extract attributes (CRS and geometry column name)
  ## this step should be before normalize_spatial_input()
  crs_x <- if (is.null(conn_x)) ddbs_crs(x, conn) else ddbs_crs(x, conn_x)
  crs_y <- if (is.null(conn_y)) ddbs_crs(y, conn) else ddbs_crs(y, conn_y)
  sf_col_x <- attr(x, "sf_column")
  sf_col_y <- attr(y, "sf_column")

  ## 1.2. Resolve conn_x/conn_y defaults from 'conn' for character inputs
  if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
  if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

  ## 1.3. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names
  x <- normalize_spatial_input(x, conn_x)
  y <- normalize_spatial_input(y, conn_y)

  ## 1.4. Get mode - If it's NULL, it will use the duckspatial.mode option
  mode <- get_mode(mode, name)


  # 2. Manage connection to DB

  ## 2.1. Resolve connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  y            <- resolve_conn$y
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 2.2. Get query list of table names
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)
  y_list <- get_query_list(y, target_conn)
  on.exit(y_list$cleanup(), add = TRUE)
  
  ## check if id column name exists in x or y
  assert_predicate_id(id_x, target_conn, x_list$query_name)
  assert_predicate_id(id_y, target_conn, y_list$query_name)

  ## 2.3. CRS already extracted at start of function
  if (!is.null(crs_x) && !is.null(crs_y)) {
    if (!crs_equal(crs_x, crs_y)) {
      cli::cli_abort("The Coordinates Reference System of {.arg x} and {.arg y} is different.")
    }
  } else {
    assert_crs(target_conn, x_list$query_name, y_list$query_name)
  }

  ## 2.4. Get crs units and geom type for next checks
  crs_units <- crs_x$units_gdal
  geom_type_x <- as.character(ddbs_geometry_type(x, conn = target_conn, by_feature = FALSE))
  geom_type_y <- as.character(ddbs_geometry_type(y, conn = target_conn, by_feature = FALSE))

  ## 2.4. Get the right distance type if user uses the default
  if (is.null(dist_type)) {
    if (crs_units == "degree") {
      ## Default to haversine if it's point and EPSG:4326
      if (crs_x$input == "EPSG:4326" && all(c(geom_type_x, geom_type_y) == "POINT")) {
        dist_type <- "haversine"
      } else {
        ## Default to spheroid if it's not POINT or if it's not EPSG:4326
        dist_type <- "spheroid"
      }
    } else {
      ## Otherwise, default to planar
      dist_type <- "planar"
    }
    if (!quiet) cli::cli_alert_info("Using {.arg dist_type = {.val {dist_type}}} by default.")
  }

  ## 2.5. Warnings/Errors for wrong election of dist_type
  ## Error: using an invalid distance type
  valid_types <- c("planar", "haversine", "geos", "spheroid")
  if (!dist_type %in% valid_types) {
      cli::cli_abort("{.arg dist_type} must be one of {.or {.val {valid_types}}}, not {.val {dist_type}}.")
  }

  ## Error: Using planar/geos on geographic coordinates
  if (crs_units == "degree" && dist_type %in% c("planar", "geos")) {
      cli::cli_abort(
          "When using {.arg dist_type = {.val {dist_type}}}, inputs must be in projected coordinates (e.g., UTM), not geographic (degrees)."
      )
  }

  ## Error: haversine/spheroid require POINT geometries
  if (dist_type %in% c("haversine", "spheroid") && !all(c(geom_type_x, geom_type_y) == "POINT")) {
      cli::cli_abort(
          "When using {.arg dist_type = {.val {dist_type}}}, inputs must be POINT geometries."
      )
  }

  ## Error: Using haversine/spheroid on projected coordinates
  if (crs_units == "metre" && dist_type %in% c("haversine", "spheroid")) {
      cli::cli_abort(
          "When using {.arg dist_type = {.val {dist_type}}}, inputs must be in {.val EPSG:4326} coordinates, not projected coordinates."
      )
  }

  ## Warning: Geographic CRS but not WGS84 (spheroid/haversine might be less accurate)
  if (crs_units == "degree" && dist_type %in% c("haversine", "spheroid") && crs_x$input != "EPSG:4326") {
      cli::cli_warn(
          "Inputs are in {.val {crs_x$input}}, not {.val EPSG:4326}. Distance calculations may be less accurate. Consider transforming to {.val EPSG:4326} or a projected CRS."
      )
  }


  # 3. Prepare parameters for the query

  ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
  assert_geometry_column(x_geom, x_list)
  assert_geometry_column(y_geom, y_list)

  ## 3.2. Select the right DuckD's function
  st_distance_fun <- switch(
    dist_type,
    "planar"    = "ST_Distance",
    "geos"      = "ST_Distance_GEOS",
    "haversine" = "ST_Distance_Sphere",
    "spheroid"  = "ST_Distance_Spheroid"
  )

  ## 3.3. Select the right coordinates order
  # st_distance_fun <- glue::glue("{st_distance_fun}(x.{x_geom}, y.{y_geom})")
  if (dist_type %in% c("haversine", "spheroid")) {
    # Here we flip the coordinates, but this will have to changed when spatial updates
    st_distance_fun <- glue::glue(
      "{st_distance_fun}(
        ST_Point(ST_Y(x.{x_geom}), ST_X(x.{x_geom})),
        ST_Point(ST_Y(y.{y_geom}), ST_X(y.{y_geom}))
      )"
    )
  } else {
    st_distance_fun <- glue::glue("{st_distance_fun}(x.{x_geom}, y.{y_geom})")
  }
  
  ## 3.2. Create query and get results based on mode
  if (mode == "sf") {

    ## Create the query
    tmp.query <- glue::glue("
      SELECT {st_distance_fun} as distance
      FROM {x_list$query_name} x
      CROSS JOIN {y_list$query_name} y
    ")

    ## Retrieve results
    data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)

    ## Cconvert to matrix
    ## Get number of rows
    nrowx <- get_nrow(target_conn, x_list$query_name)
    nrowy <- get_nrow(target_conn, y_list$query_name)

    ## Convert results to matrix -> to list
    ## Return sparse matrix
    dist_mat  <- matrix(
        data_tbl[["distance"]],
        nrow = nrowx,
        ncol = nrowy,
        byrow = TRUE
    )

    ## Set units and return the resulting matrix
    dist_mat <- units::set_units(dist_mat, "metre")
    return(dist_mat)

  } else {

    ## Subqueries for generating row ids
    x_id_expr <- if (is.null(id_x)) "row_number() OVER () AS id_x" else glue::glue("{id_x} AS id_x")
    y_id_expr <- if (is.null(id_y)) "row_number() OVER () AS id_y" else glue::glue("{id_y} AS id_y")

    ## Generate the query
    view_name <- ddbs_temp_table_name()
    tmp.query <- glue::glue("
        CREATE TEMP TABLE {view_name} AS
        SELECT
            x.id_x,
            y.id_y,
            {st_distance_fun} AS distance
        FROM (SELECT {x_id_expr}, * FROM {x_list$query_name}) x
        CROSS JOIN (SELECT {y_id_expr}, * FROM {y_list$query_name}) y
    ")

    ## Create a table, and return a pointer to that table
    DBI::dbExecute(target_conn, tmp.query)
    dist_tbl <- dplyr::tbl(target_conn, view_name)
    return(dist_tbl)


  }

}




#' @rdname ddbs_measure_funs
#' @param unit Character. Output unit: \code{"radians"} (default) or \code{"degrees"}.
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' # Create two sets of points in a projected CRS
#' origins <- sf::st_as_sf(
#'   data.frame(id = 1:2, x = c(0, 0), y = c(0, 0)),
#'   coords = c("x", "y"), crs = "EPSG:3857"
#' )
#' destinations <- sf::st_as_sf(
#'   data.frame(id = 1:2, x = c(0, 1), y = c(1, 0)),
#'   coords = c("x", "y"), crs = "EPSG:3857"
#' )
#'
#' # Returns a numeric matrix (nrow(origins) x nrow(destinations))
#' ddbs_azimuth(origins, destinations, mode = "sf")
#'
#' # In degrees
#' ddbs_azimuth(origins, destinations, unit = "degrees", mode = "sf")
#'
#' # Lazy tbl with all pairs (default mode)
#' ddbs_azimuth(origins, destinations)
#' }
ddbs_azimuth <- function(
  x,
  y,
  unit = "radians",
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  id_x = NULL,
  id_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_xy(y, "y")
  assert_name(unit, "unit")
  assert_name(id_x, "id_x")
  assert_name(id_y, "id_y")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")

  valid_units <- c("radians", "degrees")
  if (!unit %in% valid_units) {
    cli::cli_abort(
      "{.arg unit} must be one of {.or {.val {valid_units}}}, not {.val {unit}}."
    )
  }


  # 1. Manage connection to DB

  ## 1.1. Pre-extract attributes (CRS and geometry column name)
  ## this step should be before normalize_spatial_input()
  crs_x    <- if (is.null(conn_x)) ddbs_crs(x, conn) else ddbs_crs(x, conn_x)
  crs_y    <- if (is.null(conn_y)) ddbs_crs(y, conn) else ddbs_crs(y, conn_y)
  sf_col_x <- attr(x, "sf_column")
  sf_col_y <- attr(y, "sf_column")

  ## 1.2. Resolve conn_x/conn_y defaults from 'conn' for character inputs
  if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
  if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

  ## 1.3. Normalize inputs
  x <- normalize_spatial_input(x, conn_x)
  y <- normalize_spatial_input(y, conn_y)

  ## 1.4. Get mode
  mode <- get_mode(mode, name)


  # 2. Resolve connections and handle imports

  resolve_conn <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  y            <- resolve_conn$y
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 2.1. Get query lists
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)
  y_list <- get_query_list(y, target_conn)
  on.exit(y_list$cleanup(), add = TRUE)

  ## 2.2. Check id columns exist
  assert_predicate_id(id_x, target_conn, x_list$query_name)
  assert_predicate_id(id_y, target_conn, y_list$query_name)

  ## 2.3. Check CRS compatibility
  if (!is.null(crs_x) && !is.null(crs_y)) {
    if (!crs_equal(crs_x, crs_y)) {
      cli::cli_abort("The Coordinates Reference System of {.arg x} and {.arg y} is different.")
    }
  } else {
    assert_crs(target_conn, x_list$query_name, y_list$query_name)
  }

  ## 2.4. Check both inputs are POINT geometries
  geom_type_x <- as.character(ddbs_geometry_type(x, conn = target_conn, by_feature = FALSE))
  geom_type_y <- as.character(ddbs_geometry_type(y, conn = target_conn, by_feature = FALSE))

  if (geom_type_x != "POINT") {
    cli::cli_abort("{.arg x} must contain POINT geometries, not {.val {geom_type_x}}.")
  }
  if (geom_type_y != "POINT") {
    cli::cli_abort("{.arg y} must contain POINT geometries, not {.val {geom_type_y}}.")
  }


  # 3. Prepare parameters for the query

  ## 3.1. Get geometry column names
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
  assert_geometry_column(x_geom, x_list)
  assert_geometry_column(y_geom, y_list)

  ## 3.2. Build the ST_Azimuth expression, with optional degree conversion
  st_azimuth_expr <- if (unit == "degrees") {
    glue::glue("ST_Azimuth(x.{x_geom}, y.{y_geom}) * 180.0 / pi()")
  } else {
    glue::glue("ST_Azimuth(x.{x_geom}, y.{y_geom})")
  }


  # 4. Execute query and return based on mode

  if (mode == "sf") {

    tmp.query <- glue::glue("
      SELECT {st_azimuth_expr} AS azimuth
      FROM {x_list$query_name} x
      CROSS JOIN {y_list$query_name} y
    ")

    data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)

    nrowx <- get_nrow(target_conn, x_list$query_name)
    nrowy <- get_nrow(target_conn, y_list$query_name)

    az_mat <- matrix(
      data_tbl[["azimuth"]],
      nrow = nrowx,
      ncol = nrowy,
      byrow = TRUE
    )
    return(az_mat)

  } else {

    x_id_expr <- if (is.null(id_x)) "row_number() OVER () AS id_x" else glue::glue("{id_x} AS id_x")
    y_id_expr <- if (is.null(id_y)) "row_number() OVER () AS id_y" else glue::glue("{id_y} AS id_y")

    view_name <- ddbs_temp_table_name()
    tmp.query <- glue::glue("
      CREATE TEMP TABLE {view_name} AS
      SELECT
        x.id_x,
        y.id_y,
        {st_azimuth_expr} AS azimuth
      FROM (SELECT {x_id_expr}, * FROM {x_list$query_name}) x
      CROSS JOIN (SELECT {y_id_expr}, * FROM {y_list$query_name}) y
    ")

    DBI::dbExecute(target_conn, tmp.query)
    return(dplyr::tbl(target_conn, view_name))

  }

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## transform nc
nc_4326_sf <- ddbs_transform(nc_sf, "EPSG:4326")
nc_ddbs <- as_duckspatial_df(nc_4326_sf)

## select a sample of points
points_sample_sf <- head(points_sf, 10)
points_sample_ddbs <- as_duckspatial_df(points_sample_sf)

## write some data
duckspatial::ddbs_write_table(conn_test, countries_sf, "countries")
duckspatial::ddbs_write_table(conn_test, nc_ddbs, "nc")
duckspatial::ddbs_write_table(conn_test, nc_ddbs, "rivers")
duckspatial::ddbs_write_table(conn_test, points_sample_ddbs, "points")


# 1. ddbs_area -----------------------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns a vector by default
## - CHECK 1.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: message is shown with quiet = FALSE
## - CHECK 1.5: no message is shown with quiet = TRUE
## - CHECK 1.6: area is calculated properly
## - CHECK 1.7: materialize data, same output
## - CHECK 1.8: area calculated as vector, and added as column must be the same
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns a vector by default
## - CHECK 2.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 2.3: duckspatial is written into the database
## - CHECK 2.4: message is shown with quiet = FALSE
## - CHECK 2.5: no message is shown with quiet = TRUE
## - CHECK 2.6: area is calculated properly
## - CHECK 2.7: when creating a new duckdb table, it shows warning if they come from 
##   different connections
## - CHECK 2.8 - materialize data, same output
## - CHECK 2.9 - area calculated as vector, and added as column must be the same
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns a vector by default
## - CHECK 3.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 3.3: duckspatial is written into the database
## - CHECK 3.4: message is shown with quiet = FALSE
## - CHECK 3.5: no message is shown with quiet = TRUE
## - CHECK 3.6: area is calculated properly
## - CHECK 3.7: materialize data, same output
## - CHECK 3.8: area calculated as vector, and added as column must be the same
## - CHECK 3.9: error if conn = NULL, and x = duckdb table
## Check that errors work
## - CHECK 4.1: if name is specified, new_column cannot be NULL
## - CHECK 4.2: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.3: incorrect inputs
describe("ddbs_area()", {
  
  ### EXPECTED BEHAVIOUR - SF INPUT
  
  describe("expected behavior on sf input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_area(nc_4326_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_area(nc_4326_sf, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_area(nc_4326_sf, mode = NULL)
      output_sf <- ddbs_area(nc_4326_sf, mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_area(nc_4326_sf, conn = conn_test, name = "area_tbl", new_column = "area_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_area(nc_4326_sf, new_column = "area_calc"))
      expect_message(ddbs_area(nc_4326_sf, conn = conn_test, name = "area_tbl2", new_column = "area_calc"))
      
      expect_no_message(ddbs_area(nc_4326_sf, new_column = "area_calc", quiet = TRUE))
      expect_no_message(ddbs_area(nc_4326_sf, conn = conn_test, name = "area_tbl3", new_column = "area_calc", quiet = TRUE))
    })
    
    it("calculates area correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      area_ddbs <- ddbs_area(argentina_3857_sf, mode = "sf")
      area_sf   <- sf::st_area(argentina_3857_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })

    it("calculates area correctly on geographic CRS", {
      area_ddbs <- ddbs_area(argentina_sf, mode = "sf")
      area_sf   <- sf::st_area(argentina_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_area(nc_4326_sf, new_column = "area_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_area(nc_4326_sf, mode = "sf") |> as.numeric()
      output_column <- ddbs_area(nc_4326_sf, new_column = "area_calc", mode = NULL) |> 
        ddbs_collect()
      
      expect_identical(output_table, output_column$area_calc)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT
  
  describe("expected behavior on duckspatial_df input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_area(nc_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_area(nc_ddbs, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_area(nc_ddbs, new_column = "area_calc", mode = NULL)
      output_sf       <- ddbs_area(nc_ddbs, new_column = "area_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_area(nc_ddbs, conn = conn_test, name = "ddbs_area_tbl", new_column = "area_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_area(nc_ddbs, new_column = "area_calc"))
      expect_message(ddbs_area(nc_ddbs, conn = conn_test, name = "ddbs_area_tbl2", new_column = "area_calc"))

      expect_no_message(ddbs_area(nc_ddbs, conn = conn_test, name = "ddbs_area_tbl3", new_column = "area_calc", quiet = TRUE))      
      expect_no_message(ddbs_area(nc_ddbs, new_column = "area_calc", quiet = TRUE))
    })

    it("calculates area correctly on geographic CRS", {
      area_ddbs <- ddbs_area(argentina_sf, mode = "sf")
      area_sf   <- sf::st_area(argentina_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })
    
    it("calculates area correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      area_ddbs <- ddbs_area(argentina_3857_sf, mode = "sf")
      area_sf   <- sf::st_area(argentina_3857_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })
    
    it("warns when creating table from different connections", {
      expect_warning(ddbs_area(nc_ddbs, conn = conn_test, name = "ddbs_area_tbl4", new_column = "area_calc"))
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_area(nc_ddbs, new_column = "area_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_area(nc_ddbs, mode = "sf") |> as.numeric()
      output_column <- ddbs_area(nc_ddbs, new_column = "area_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$area_calc)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT
  
  describe("expected behavior on DuckDB table input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_area("nc", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_area("nc", conn = conn_test, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_area("nc", conn = conn_test, new_column = "area_calc", mode = NULL)
      output_sf       <- ddbs_area("nc", conn = conn_test, new_column = "area_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_area("nc", conn = conn_test, name = "conn_area_tbl", new_column = "area_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_area("nc", conn = conn_test, new_column = "area_calc"))
      expect_message(ddbs_area("nc", conn = conn_test, name = "conn_area_tbl2", new_column = "area_calc"))
      
      expect_no_message(ddbs_area("nc", conn = conn_test, new_column = "area_calc", quiet = TRUE))
      expect_no_message(ddbs_area("nc", conn = conn_test, name = "conn_area_tbl3", new_column = "area_calc", quiet = TRUE))
    })

    it("calculates area correctly on geographic CRS", {
      duckspatial::ddbs_write_table(conn_test, argentina_sf, "argentina", overwrite = TRUE)
      area_ddbs <- ddbs_area("argentina", conn = conn_test, mode = "sf")
      area_sf   <- sf::st_area(argentina_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })
    
    it("calculates area correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      duckspatial::ddbs_write_table(conn_test, argentina_3857_sf, "argentina", overwrite = TRUE)
      area_ddbs <- ddbs_area("argentina", conn = conn_test, mode = "sf")
      area_sf   <- sf::st_area(argentina_3857_sf)
      
      expect_equal(area_ddbs, area_sf, tolerance = 0.001)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_area("nc", conn = conn_test, new_column = "area_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_area("nc", conn = conn_test, mode = "sf") |> as.numeric()
      output_column <- ddbs_area("nc", conn = conn_test, new_column = "area_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$area_calc)
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_area(x = "nc", conn = NULL))
    })
    
    it("requires new_column when name is specified", {
      expect_error(ddbs_area(nc_4326_sf, name = "new_tbl"))
    })
    
    it("prevents overwriting existing tables without overwrite = TRUE", {
      expect_error(ddbs_area(nc_4326_sf, conn = conn_test, name = "countries", new_column = "area_calc"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_area(x = 999))
      expect_error(ddbs_area(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_area(nc_4326_sf, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_area(nc_4326_sf, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_area(nc_4326_sf, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_area(nc_4326_sf, quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_area(nc_4326_sf, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 2. ddbs_length ---------------------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns a vector by default
## - CHECK 1.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: message is shown with quiet = FALSE
## - CHECK 1.5: no message is shown with quiet = TRUE
## - CHECK 1.6: length is calculated properly
## - CHECK 1.7: materialize data, same output
## - CHECK 1.8: length calculated as vector, and added as column must be the same
## - CHECK 1.9: length on polygons or points is equal to 0
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns a vector by default
## - CHECK 2.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 2.3: duckspatial is written into the database
## - CHECK 2.4: message is shown with quiet = FALSE
## - CHECK 2.5: no message is shown with quiet = TRUE
## - CHECK 2.6: length is calculated properly
## - CHECK 2.7: when creating a new duckdb table, it shows warning if they come from 
##   different connections
## - CHECK 2.8 - materialize data, same output
## - CHECK 2.9 - length calculated as vector, and added as column must be the same
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns a vector by default
## - CHECK 3.2: returns the correct output (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 3.3: duckspatial is written into the database
## - CHECK 3.4: message is shown with quiet = FALSE
## - CHECK 3.5: no message is shown with quiet = TRUE
## - CHECK 3.6: length is calculated properly
## - CHECK 3.7: materialize data, same output
## - CHECK 3.8: length calculated as vector, and added as column must be the same
## - CHECK 3.9: error if conn = NULL, and x = duckdb table
## Check that errors work
## - CHECK 4.1: if name is specified, new_column cannot be NULL
## - CHECK 4.2: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.3: incorrect inputs
describe("ddbs_length()", {
  
  ### EXPECTED BEHAVIOUR - SF INPUT
  
  describe("expected behavior on sf input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_length(rivers_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_length(rivers_sf, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs <- ddbs_length(rivers_sf, new_column = "length_calc", mode = "duckspatial")
      output_sf   <- ddbs_length(rivers_sf, new_column = "length_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_length(rivers_sf, conn = conn_test, name = "length_tbl", new_column = "length_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_length(rivers_sf, new_column = "length_calc"))
      expect_message(ddbs_length(rivers_sf, conn = conn_test, name = "length_tbl2", new_column = "length_calc"))
      
      expect_no_message(ddbs_length(rivers_sf, new_column = "length_calc", quiet = TRUE))
      expect_no_message(ddbs_length(rivers_sf, conn = conn_test, name = "length_tbl3", new_column = "length_calc", quiet = TRUE))
    })

    it("calculates length correctly on geographic CRS", {
      rivers_geog <- ddbs_transform(rivers_sf, "EPSG:4326", mode = "sf")
      length_ddbs <- ddbs_length(rivers_geog, mode = "sf")
      length_sf   <- sf::st_length(rivers_geog)
      # TODO - DuckDB v1.5 has changes in the calculation, and gives different values
      # - review when ST_Spheroid_*() funs don't need to flip coords anymore
      expect_equal(length_ddbs, length_sf, tolerance = .1)
    })
    
    it("calculates length correctly on projected CRS", {
      rivers_3857_sf <- sf::st_transform(rivers_sf, "EPSG:3857")
      length_ddbs <- ddbs_length(rivers_3857_sf, mode = "sf")
      length_sf   <- sf::st_length(rivers_3857_sf)
      
      expect_equal(length_ddbs, length_sf)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_length(rivers_sf, new_column = "length_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_length(rivers_sf, mode = "sf") |> as.numeric()
      output_column <- ddbs_length(rivers_sf, new_column = "length_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$length_calc)
    })
    
    it("returns 0 for polygons and points", {
      output_polygons <- ddbs_length(countries_sf, mode = "sf") |> as.numeric()
      output_points   <- ddbs_length(points_sample_sf, mode = "sf") |> as.numeric()
      
      expect_true(all(output_polygons == 0))
      expect_true(all(output_points == 0))
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT
  
  describe("expected behavior on duckspatial_df input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_length(rivers_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_length(rivers_ddbs, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs <- ddbs_length(rivers_ddbs, new_column = "length_calc", mode = "duckspatial")
      output_sf   <- ddbs_length(rivers_ddbs, new_column = "length_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_length(rivers_ddbs, conn = conn_test, name = "ddbs_length_tbl", new_column = "length_calc") |> 
        suppressWarnings()
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_length(rivers_ddbs, new_column = "length_calc"))
      expect_message(ddbs_length(rivers_ddbs, conn = conn_test, name = "ddbs_length_tbl2", new_column = "length_calc") |> suppressWarnings())

      expect_no_message(ddbs_length(rivers_ddbs, conn = conn_test, name = "ddbs_length_tbl3", new_column = "length_calc", quiet = TRUE))      
      expect_no_message(ddbs_length(rivers_ddbs, new_column = "length_calc", quiet = TRUE))
    })

    it("calculates length correctly on geographic CRS", {
      length_ddbs <- ddbs_length(rivers_sf, mode = "sf")
      length_sf   <- sf::st_length(rivers_sf)
      
      expect_equal(length_ddbs, length_sf)
    })
    
    it("calculates length correctly on projected CRS", {
      rivers_3857_sf <- sf::st_transform(rivers_sf, "EPSG:3857")
      length_ddbs <- ddbs_length(rivers_3857_sf, mode = "sf")
      length_sf   <- sf::st_length(rivers_3857_sf)
      
      expect_equal(length_ddbs, length_sf)
    })
    
    it("warns when creating table from different connections", {
      expect_warning(ddbs_length(rivers_ddbs, conn = conn_test, name = "ddbs_length_tbl4", new_column = "length_calc"))
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_length(rivers_ddbs, new_column = "length_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_length(rivers_ddbs, mode = "sf") |> as.numeric()
      output_column <- ddbs_length(rivers_ddbs, new_column = "length_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$length_calc)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT
  
  describe("expected behavior on DuckDB table input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_length("rivers", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_length("rivers", conn = conn_test, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_length("rivers", conn = conn_test, new_column = "length_calc", mode = NULL)
      output_sf       <- ddbs_length("rivers", conn = conn_test, new_column = "length_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_length("rivers", conn = conn_test, name = "conn_length_tbl", new_column = "length_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_length("rivers", conn = conn_test, new_column = "length_calc"))
      expect_message(ddbs_length("rivers", conn = conn_test, name = "conn_length_tbl2", new_column = "length_calc"))
      
      expect_no_message(ddbs_length("rivers", conn = conn_test, new_column = "length_calc", quiet = TRUE))
      expect_no_message(ddbs_length("rivers", conn = conn_test, name = "conn_length_tbl3", new_column = "length_calc", quiet = TRUE))
    })

    it("calculates length correctly on geographic CRS", {
      duckspatial::ddbs_write_table(conn_test, rivers_sf, "rivers_4326")
      length_ddbs <- ddbs_length("rivers_4326", conn = conn_test, mode = "sf")
      length_sf   <- sf::st_length(rivers_sf)
      
      expect_equal(length_ddbs, length_sf)
    })
    
    it("calculates length correctly on projected CRS", {
      rivers_3857_sf <- sf::st_transform(rivers_sf, "EPSG:3857")
      duckspatial::ddbs_write_table(conn_test, rivers_3857_sf, "rivers_3857")
      length_ddbs <- ddbs_length("rivers_3857", conn = conn_test, mode = "sf")
      length_sf   <- sf::st_length(rivers_3857_sf)
      
      expect_equal(length_ddbs, length_sf)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_length("rivers", conn = conn_test, new_column = "length_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_length("rivers", conn = conn_test, mode = "sf") |> as.numeric()
      output_column <- ddbs_length("rivers", conn = conn_test, new_column = "length_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$length_calc)
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_length(x = "nc", conn = NULL))
    })
    
    it("requires new_column when name is specified", {
      expect_error(ddbs_length(rivers_sf, name = "new_tbl"))
    })
    
    it("prevents overwriting existing tables without overwrite = TRUE", {
      expect_error(ddbs_length(rivers_sf, conn = conn_test, name = "countries", new_column = "length_calc"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_length(x = 999))
      expect_error(ddbs_length(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_length(rivers_sf, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_length(rivers_sf, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_length(rivers_sf, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_length(rivers_sf, quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_length(rivers_sf, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 3. ddbs_distance -------------------------------------------------------

## 3.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns a matrix
## - CHECK 1.3: messages work
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_distance()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_sf_ddbs   <- ddbs_distance(points_sample_sf, points_sample_ddbs) |> collect()
      output_ddbs_sf   <- ddbs_distance(points_sample_ddbs, points_sample_sf) |> collect()
      output_sf_sf     <- ddbs_distance(points_sample_sf, points_sample_sf) |> collect()
      output_ddbs_ddbs <- ddbs_distance(points_sample_ddbs, points_sample_ddbs) |> collect()
      output_conn_sf   <- ddbs_distance("points", points_sample_sf, conn = conn_test) |> collect()
      ## This one retrieves the result in a different order, but same results
      output_sf_conn   <- ddbs_distance(points_sample_sf, "points", conn = conn_test) |> collect()
      
      expect_equal(output_sf_ddbs, output_ddbs_sf)
      expect_equal(output_ddbs_sf, output_sf_sf)
      expect_equal(output_ddbs_sf, output_ddbs_ddbs)
      expect_equal(output_ddbs_sf, output_conn_sf)
      expect_equal(output_ddbs_sf, output_sf_conn |> dplyr::arrange(id_y, id_x))
    })
    
    it("warns when mixing DuckDB table with duckspatial_df from different connections", {
      expect_warning(ddbs_distance("points", points_sample_ddbs, conn = conn_test))
      expect_warning(ddbs_distance(points_sample_ddbs, "points", conn = conn_test))
    })

    it("warns when using a geographic CRS different than WGS84", {
      points_nad83 <- ddbs_transform(points_sample_ddbs, "EPSG:4269") |> head()
      expect_warning(ddbs_distance(points_nad83, points_nad83))
    })
    
    it("returns a units matrix with mode sf", {
      output <- ddbs_distance(points_sample_sf, points_sample_ddbs, mode = "sf")
      expect_s3_class(output, "units")
      expect_equal(
        class(units::drop_units(output)), 
        c("matrix", "array")
      )
    })
    
    it("shows and suppresses messages correctly", {
      expect_message(ddbs_distance(points_sample_sf, points_sample_ddbs))
      expect_no_message(ddbs_distance(points_sample_sf, points_sample_ddbs, quiet = TRUE))
    })

    it("works with dist_type = harvesine", {
      haversine_res <- ddbs_distance(points_sample_ddbs, points_sample_ddbs, dist_type = "haversine")
      expect_s3_class(haversine_res, "tbl_duckdb_connection")
    })

    it("works with dist_type = spheroid", {
      spheroid_res <- ddbs_distance(points_sample_ddbs, points_sample_ddbs, dist_type = "spheroid")
      expect_s3_class(spheroid_res, "tbl_duckdb_connection")
    })

    it("works with dist_type = planar", {
      points_3857_ddbs <- ddbs_transform(points_sample_ddbs, "EPSG:3857")
      planar_res <- ddbs_distance(points_3857_ddbs, points_3857_ddbs, dist_type = "planar")
      expect_s3_class(planar_res, "tbl_duckdb_connection")
    })

    it("works with dist_type = geos", {
      points_3857_ddbs <- ddbs_transform(points_sample_ddbs, "EPSG:3857")
      geos_res <- ddbs_distance(points_3857_ddbs, points_3857_ddbs, dist_type = "geos")
      expect_s3_class(geos_res, "tbl_duckdb_connection")
    })

    it("works with default dist_type for geographic", {
      spheroid_res <- ddbs_distance(points_sample_ddbs, points_sample_ddbs)
      expect_s3_class(spheroid_res, "tbl_duckdb_connection")
    })

    it("works with default dist_type for projected", {
      points_3857_ddbs <- ddbs_transform(points_sample_ddbs, "EPSG:3857")
      spheroid_res <- ddbs_distance(points_3857_ddbs, points_3857_ddbs)
      expect_s3_class(spheroid_res, "tbl_duckdb_connection")
    })

  })

  
  ### EXPECTED ERRORS
  
  describe("errors", {

    it("validates dist_type argument", {
      expect_error(ddbs_distance(points_sample_sf, points_sample_ddbs, dist_type = "best_dist"))
      expect_error(ddbs_distance(points_sample_sf, points_sample_ddbs, dist_type = TRUE))
      expect_error(ddbs_distance(points_sample_sf, points_sample_ddbs, dist_type = 5))
    })

    it("error when using planar/geos with geographic coords", {
      expect_error(ddbs_distance(points_sample_ddbs, points_sample_ddbs, dist_type = "geos"))
      expect_error(ddbs_distance(points_sample_ddbs, points_sample_ddbs, dist_type = "planar"))
    })

    it("error when using geogrpahic coords in geometry different than point", {
      expect_error(ddbs_distance(points_sample_ddbs, nc_4326_sf, dist_type = "haversine"))
      expect_error(ddbs_distance(points_sample_ddbs, nc_4326_sf, dist_type = "spheroid"))
    })

    it("error when using haversine/spheroid in projected CRS", {
      points_3857_ddbs <- ddbs_transform(points_sample_ddbs, "EPSG:3857")
      expect_error(ddbs_distance(points_3857_ddbs, points_3857_ddbs, dist_type = "haversine"))
      expect_error(ddbs_distance(points_3857_ddbs, points_3857_ddbs, dist_type = "spheroid"))
    })
    
    it("requires both x and y arguments", {
      expect_error(ddbs_distance(x = points_sample_ddbs))
      expect_error(ddbs_distance(y = points_sample_ddbs))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_distance("points", "points", conn = NULL))
    })
    
    it("requires matching CRS between x and y", {
      points_3857_sf <- sf::st_transform(points_sample_sf, "EPSG:3857")
      
      expect_error(ddbs_distance(points_sample_sf, points_3857_sf))
      expect_error(ddbs_distance(points_3857_sf, points_sample_ddbs))
    })
    
    it("requires matching geometry types", {
      expect_error(ddbs_distance(argentina_sf, points_sample_ddbs))
      expect_error(ddbs_distance(points_sample_ddbs, argentina_sf))
      expect_error(ddbs_distance(points_sample_sf, sf::st_transform(rivers_sf, sf::st_crs(points_sample_sf))))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_distance(x = 999))
      expect_error(ddbs_distance(x = "999", points_sample_ddbs, conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_distance(points_sample_ddbs, points_sample_ddbs, conn = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_distance(points_sample_ddbs, points_sample_ddbs, quiet = 999))
    })
  })
})


# 4. ddbs_perimeter ------------------------------------------------------

describe("ddbs_perimeter()", {
  
  ### EXPECTED BEHAVIOUR - SF INPUT
  
  describe("expected behavior on sf input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_perimeter(nc_4326_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_perimeter(nc_4326_sf, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc", mode = NULL)
      output_sf       <- ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_perimeter(nc_4326_sf, conn = conn_test, name = "perimeter_tbl", new_column = "perimeter_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc"))
      expect_message(ddbs_perimeter(nc_4326_sf, conn = conn_test, name = "perimeter_tbl2", new_column = "perimeter_calc"))
      
      expect_no_message(ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc", quiet = TRUE))
      # expect_no_message(ddbs_perimeter(nc_4326_sf, conn = conn_test, name = "perimeter_tbl3", new_column = "perimeter_calc", quiet = TRUE))
    })
    
    it("calculates perimeter correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      perimeter_ddbs <- ddbs_perimeter(argentina_3857_sf, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_3857_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })

    it("calculates perimeter correctly on geographic CRS", {
      perimeter_ddbs <- ddbs_perimeter(argentina_sf, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_perimeter(nc_4326_sf, mode = "sf") |> as.numeric()
      output_column <- ddbs_perimeter(nc_4326_sf, new_column = "perimeter_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$perimeter_calc)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT
  
  describe("expected behavior on duckspatial_df input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_perimeter(nc_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_perimeter(nc_ddbs, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc", mode = NULL)
      output_sf       <- ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_perimeter(nc_ddbs, conn = conn_test, name = "ddbs_perimeter_tbl", new_column = "perimeter_calc") |> 
        suppressWarnings()
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc"))
      expect_message(ddbs_perimeter(nc_ddbs, conn = conn_test, name = "ddbs_perimeter_tbl2", new_column = "perimeter_calc"))

      expect_no_message(ddbs_perimeter(nc_ddbs, conn = conn_test, name = "ddbs_perimeter_tbl3", new_column = "perimeter_calc", quiet = TRUE))      
      expect_no_message(ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc", quiet = TRUE))
    })

    it("calculates perimeter correctly on geographic CRS", {
      perimeter_ddbs <- ddbs_perimeter(argentina_sf, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })
    
    it("calculates perimeter correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      perimeter_ddbs <- ddbs_perimeter(argentina_3857_sf, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_3857_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })
    
    it("warns when creating table from different connections", {
      expect_warning(ddbs_perimeter(nc_ddbs, conn = conn_test, name = "ddbs_perimeter_tbl4", new_column = "perimeter_calc"))
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_perimeter(nc_ddbs, mode = "sf") |> as.numeric()
      output_column <- ddbs_perimeter(nc_ddbs, new_column = "perimeter_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$perimeter_calc)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT
  
  describe("expected behavior on DuckDB table input", {
    
    it("returns a duckspatial_df by default", {
      output <- ddbs_perimeter("nc", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a units vector with mode sf", {
      output <- ddbs_perimeter("nc", conn = conn_test, mode = "sf")
      expect_s3_class(output, "units")
    })
    
    it("returns different mode formats (duckspatial, sf)", {
      output_ddbs     <- ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc", mode = NULL)
      output_sf       <- ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "units")
    })
    
    it("writes tables to the database", {
      output <- ddbs_perimeter("nc", conn = conn_test, name = "conn_perimeter_tbl", new_column = "perimeter_calc")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc"))
      expect_message(ddbs_perimeter("nc", conn = conn_test, name = "conn_perimeter_tbl2", new_column = "perimeter_calc"))
      
      expect_no_message(ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc", quiet = TRUE))
      expect_no_message(ddbs_perimeter("nc", conn = conn_test, name = "conn_perimeter_tbl3", new_column = "perimeter_calc", quiet = TRUE))
    })

    it("calculates perimeter correctly on geographic CRS", {
      duckspatial::ddbs_write_table(conn_test, argentina_sf, "argentina", overwrite = TRUE)
      perimeter_ddbs <- ddbs_perimeter("argentina", conn = conn_test, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })
    
    it("calculates perimeter correctly on projected CRS", {
      argentina_3857_sf <- sf::st_transform(argentina_sf, "EPSG:3857")
      duckspatial::ddbs_write_table(conn_test, argentina_3857_sf, "argentina", overwrite = TRUE)
      perimeter_ddbs <- ddbs_perimeter("argentina", conn = conn_test, mode = "sf")
      perimeter_sf   <- sf::st_perimeter(argentina_3857_sf)
      
      expect_equal(perimeter_ddbs, perimeter_sf, tolerance = 0.001)
    })
    
    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc", mode = NULL)
      
      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()
      
      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })
    
    it("produces identical results for vector and column outputs", {
      output_table <- ddbs_perimeter("nc", conn = conn_test, mode = "sf") |> as.numeric()
      output_column <- ddbs_perimeter("nc", conn = conn_test, new_column = "perimeter_calc", mode = NULL) |> st_as_sf()
      
      expect_identical(output_table, output_column$perimeter_calc)
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_perimeter(x = "nc", conn = NULL))
    })
    
    it("requires new_column when name is specified", {
      expect_error(ddbs_perimeter(nc_4326_sf, name = "new_tbl"))
    })
    
    it("prevents overwriting existing tables without overwrite = TRUE", {
      expect_error(ddbs_perimeter(nc_4326_sf, conn = conn_test, name = "countries", new_column = "perimeter_calc"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_perimeter(x = 999))
      expect_error(ddbs_perimeter(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_perimeter(nc_4326_sf, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_perimeter(nc_4326_sf, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_perimeter(nc_4326_sf, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_perimeter(nc_4326_sf, quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_perimeter(nc_4326_sf, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 5. ddbs_azimuth --------------------------------------------------------

## Create synthetic POINT data in a projected CRS for known-value tests.
## In EPSG:3857 (Web Mercator), axis order is (easting, northing).
## From origin (0,0):
##   due north → (0, 1):  azimuth = 0 rad
##   due east  → (1, 0):  azimuth = pi/2 rad
##   due south → (0,-1):  azimuth = pi rad
##   due west  → (-1,0):  azimuth = 3*pi/2 rad
origin_sf <- sf::st_as_sf(
  data.frame(id = 1L, x = 0, y = 0),
  coords = c("x", "y"), crs = "EPSG:3857"
)
dirs_sf <- sf::st_as_sf(
  data.frame(
    id = 1:4,
    x = c(0,  1, 0, -1),
    y = c(1,  0, -1,  0)
  ),
  coords = c("x", "y"), crs = "EPSG:3857"
)
origin_ddbs <- as_duckspatial_df(origin_sf)
dirs_ddbs   <- as_duckspatial_df(dirs_sf)

duckspatial::ddbs_write_table(conn_test, origin_sf, "azimuth_origin")
duckspatial::ddbs_write_table(conn_test, dirs_sf,   "azimuth_dirs")

describe("ddbs_azimuth()", {

  ### EXPECTED BEHAVIOUR - SF INPUT

  describe("expected behavior on sf input", {

    it("returns a tbl by default", {
      output <- ddbs_azimuth(origin_sf, dirs_sf)
      expect_s3_class(output, "tbl_duckdb_connection")
    })

    it("returns a numeric matrix with mode sf", {
      output <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      expect_true(is.matrix(output))
      expect_true(is.numeric(output))
    })

    it("returns a matrix with correct dimensions", {
      output <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      expect_equal(dim(output), c(nrow(origin_sf), nrow(dirs_sf)))
    })

    it("calculates azimuth correctly in radians", {
      output <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      expect_equal(output[1, 1], 0,         tolerance = 1e-6)  # due north
      expect_equal(output[1, 2], pi / 2,    tolerance = 1e-6)  # due east
      expect_equal(output[1, 3], pi,        tolerance = 1e-6)  # due south
      expect_equal(output[1, 4], 3 * pi / 2, tolerance = 1e-6) # due west
    })

    it("calculates azimuth correctly in degrees", {
      output <- ddbs_azimuth(origin_sf, dirs_sf, unit = "degrees", mode = "sf")
      expect_equal(output[1, 1],   0, tolerance = 1e-6)
      expect_equal(output[1, 2],  90, tolerance = 1e-6)
      expect_equal(output[1, 3], 180, tolerance = 1e-6)
      expect_equal(output[1, 4], 270, tolerance = 1e-6)
    })

    it("degrees output equals radians output * 180/pi", {
      rad <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      deg <- ddbs_azimuth(origin_sf, dirs_sf, unit = "degrees", mode = "sf")
      expect_equal(rad * 180 / pi, deg, tolerance = 1e-10)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT

  describe("expected behavior on duckspatial_df input", {

    it("returns a tbl by default", {
      output <- ddbs_azimuth(origin_ddbs, dirs_ddbs)
      expect_s3_class(output, "tbl_duckdb_connection")
    })

    it("returns a numeric matrix with mode sf", {
      output <- ddbs_azimuth(origin_ddbs, dirs_ddbs, mode = "sf")
      expect_true(is.matrix(output))
      expect_true(is.numeric(output))
    })

    it("warns when creating table from different connections", {
      expect_warning(ddbs_azimuth(origin_ddbs, dirs_sf, conn = conn_test))
    })

    it("produces identical results as sf input", {
      output_sf   <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      output_ddbs <- ddbs_azimuth(origin_ddbs, dirs_ddbs, mode = "sf")
      expect_equal(output_sf, output_ddbs)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT

  describe("expected behavior on DuckDB table input", {

    it("returns a tbl by default", {
      output <- ddbs_azimuth("azimuth_origin", "azimuth_dirs", conn = conn_test)
      expect_s3_class(output, "tbl_duckdb_connection")
    })

    it("returns a numeric matrix with mode sf", {
      output <- ddbs_azimuth("azimuth_origin", "azimuth_dirs", conn = conn_test, mode = "sf")
      expect_true(is.matrix(output))
      expect_true(is.numeric(output))
    })

    it("produces identical results as sf input", {
      output_sf   <- ddbs_azimuth(origin_sf, dirs_sf, mode = "sf")
      output_conn <- ddbs_azimuth("azimuth_origin", "azimuth_dirs", conn = conn_test, mode = "sf")
      expect_equal(output_sf, output_conn)
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("rejects non-POINT geometries for x", {
      expect_error(ddbs_azimuth(countries_sf, origin_sf))
    })

    it("rejects non-POINT geometries for y", {
      expect_error(ddbs_azimuth(origin_sf, countries_sf))
    })

    it("rejects mismatched CRS", {
      origin_4326_sf <- sf::st_transform(origin_sf, "EPSG:4326")
      expect_error(ddbs_azimuth(origin_sf, origin_4326_sf))
    })

    it("rejects invalid unit", {
      expect_error(ddbs_azimuth(origin_sf, dirs_sf, unit = "gradians"))
      expect_error(ddbs_azimuth(origin_sf, dirs_sf, unit = 42))
    })

    it("requires both x and y", {
      expect_error(ddbs_azimuth(x = origin_sf))
      expect_error(ddbs_azimuth(y = dirs_sf))
    })

    it("requires a connection for character table names", {
      expect_error(ddbs_azimuth("azimuth_origin", "azimuth_dirs", conn = NULL))
    })

    it("validates x and y argument types", {
      expect_error(ddbs_azimuth(x = 999, y = origin_sf))
      expect_error(ddbs_azimuth(x = "nonexistent_tbl", y = dirs_sf, conn = conn_test))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_azimuth(origin_sf, dirs_sf, quiet = 999))
    })
  })
})


## stop connection
ddbs_stop_conn(conn_test)

#' Check CRS of spatial objects or database tables
#'
#' This is an S3 generic that extracts CRS information from various spatial objects.
#'
#' @param x An object containing spatial data. Can be:
#'   - \code{duckspatial_df}: Lazy spatial data frame (CRS from attributes)
#'   - \code{sf}: sf object (CRS from sf metadata)
#'   - \code{character}: Name of table in database (requires \code{conn})
#' @param ... Additional arguments passed to methods
#'
#' @returns CRS object from \code{sf} package
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckdb)
#' library(duckspatial)
#' library(sf)
#'
#' # Method 1: duckspatial_df objects
#' nc <- ddbs_open_dataset(system.file("shape/nc.shp", package = "sf"))
#' ddbs_crs(nc)
#'
#' # Method 2: sf objects
#' nc_sf <- st_read(system.file("shape/nc.shp", package = "sf"))
#' ddbs_crs(nc_sf)
#'
#' # Method 3: table name in database
#' conn <- ddbs_create_conn(dbdir = "memory")
#' ddbs_write_table(conn, nc_sf, "nc_table")
#' ddbs_crs(conn, "nc_table")
#' ddbs_stop_conn(conn)
#' }
ddbs_crs <- function(x, ...) {
  UseMethod("ddbs_crs")
}

#' @export
#' @rdname ddbs_crs
ddbs_crs.duckspatial_df <- function(x, ...) {
  crs <- attr(x, "crs")
  if (is.null(crs)) {
    return(sf::st_crs(NA))
  }
  crs
}

#' @export
#' @rdname ddbs_crs
ddbs_crs.sf <- function(x, ...) {
  sf::st_crs(x)
}

#' @export
#' @rdname ddbs_crs
ddbs_crs.tbl_duckdb_connection <- function(x, ...) {
  # Try to auto-detect CRS from view SQL (for duckdbfs::open_dataset and similar)
  conn <- dbplyr::remote_con(x)
  
  # Try to use ST_CRS() first if we have a valid table name
  remote_table <- tryCatch({
    rem_name <- dbplyr::remote_name(x)
    if (is.null(rem_name)) NULL else as.character(rem_name)
  }, error = function(e) NULL)

  if (!is.null(remote_table)) {
    crs_data <- tryCatch({
      geom_name <- get_geom_name(conn, remote_table)
      resolve_crs(
        conn,
        remote_table,
        geom_name,
        quiet_unknown = TRUE
      )
    }, error = function(e) NULL)

    if (!is.null(crs_data) && !is.na(crs_data)) {
      return(crs_data)
    }
  }

  # Strategy 1: Try to get view SQL from duckdb_views()
  view_sql <- tryCatch({
    table_name <- dbplyr::remote_name(x)
    
    if (!is.null(table_name) && !inherits(table_name, "sql")) {
      result <- DBI::dbGetQuery(conn, glue::glue(
        "SELECT sql FROM duckdb_views() WHERE view_name = '{table_name}'"
      ))
      if (nrow(result) > 0) result$sql else NULL
    } else {
      NULL
    }
  }, error = function(e) NULL)
  
  # Strategy 2: If no view found, render the query SQL
  if (is.null(view_sql)) {
    view_sql <- tryCatch({
      as.character(dbplyr::sql_render(x, con = conn))
    }, error = function(e) NULL)
  }
  
  # Extract file path from ST_Read() calls
  if (!is.null(view_sql)) {
    # Look for ST_Read('path') or st_read("path") patterns
    path_match <- regexpr("(?:ST_Read|st_read)\\s*\\(\\s*['\"]([^'\"]+)['\"]", 
                          view_sql, 
                          perl = TRUE, 
                          ignore.case = TRUE)
    
    if (path_match[1] > 0) {
      # Extract the captured group (the path)
      start <- attr(path_match, "capture.start")[1,1]
      length <- attr(path_match, "capture.length")[1,1]
      file_path <- substr(view_sql, start, start + length - 1)
      
      # Use st_read_meta to get CRS
      crs <- tryCatch({
        get_file_crs(file_path, conn)
      }, error = function(e) NULL)
      
      if (!is.null(crs)) {
        return(crs)
      }
    }
  }
  
  # Fallback: return NA CRS
  cli::cli_warn(c(
    "Could not auto-detect CRS for {.cls tbl_duckdb_connection} object.",
    "i" = "This typically occurs when reopening a persistent DuckDB database created without recoverable CRS metadata (for example, pre-1.5 files without duckspatial comments) or when the table/file has an unknown or missing CRS.",
    "i" = "Use {.code as_duckspatial_df(x, crs = ...)} to set the CRS explicitly."
  ))
  sf::st_crs(NA)
}

#' @export
#' @rdname ddbs_crs
#' @param conn A DuckDB connection (required for character method)
ddbs_crs.character <- function(x, conn, ...) {

    # 0. Check if x is in AUTH:CODE format (e.g., "EPSG:4326")
    if (length(x) == 1 && grepl("^[A-Z]+:[0-9]+$", x)) {
        return(sf::st_crs(x))
    }
  
    # 1. Checks
    ## Check if connection is correct
    dbConnCheck(conn)
    
    name <- x
    
    ## convenient names of table and/or schema.table
    x_list <- get_query_list(x, conn)

    ## Check if table name exists in Tables OR Arrow Views
    # Use SQL check to catch temporary views which might not show up in dbListTables
    check_query <- glue::glue("
      SELECT 1 FROM information_schema.tables 
      WHERE table_name = '{x_list$query_name}' 
      UNION 
      SELECT 1 FROM duckdb_views() 
      WHERE view_name = '{x_list$query_name}'
      LIMIT 1
    ")
    
    table_exists <- tryCatch({
      nrow(DBI::dbGetQuery(conn, check_query)) > 0
    }, error = function(e) FALSE)
    
    arrow_exists <- FALSE

    if (!table_exists) {
        arrow_list <- try(duckdb::duckdb_list_arrow(conn), silent = TRUE)
        if (!inherits(arrow_list, "try-error") && x_list$table_name %in% arrow_list) {
            arrow_exists <- TRUE
        }
    }

    if (!table_exists && !arrow_exists) {
        cli::cli_abort("The provided name is not present in the database.")
    }
    
    ## Get CRS from the table
    crs_data <- tryCatch({
      geom_name <- get_geom_name(conn, x_list$query_name)
      resolve_crs(
        conn,
        x_list$query_name,
        geom_name,
        quiet_unknown = TRUE
      )
    }, error = function(e) {
      NULL
    })
  
    ## TODO - Review below for duckdb 1.5 - is this necessary with ST_CRS()?    
    if (is.null(crs_data)) {
      # Fallback: Try to auto-detect from view definition (like tbl_duckdb_connection method)
      view_sql <- tryCatch({
        result <- DBI::dbGetQuery(conn, glue::glue(
          "SELECT sql FROM duckdb_views() WHERE view_name = '{x_list$table_name}'"
        ))
        if (nrow(result) > 0) result$sql else NULL
      }, error = function(e) NULL)
      
      if (!is.null(view_sql)) {
         path_match <- regexpr("(?:ST_Read|st_read)\\s*\\(\\s*['\"]([^'\"]+)['\"]", 
                              view_sql, perl = TRUE, ignore.case = TRUE)
         if (path_match[1] > 0) {
            start <- attr(path_match, "capture.start")[1,1]
            length <- attr(path_match, "capture.length")[1,1]
            file_path <- substr(view_sql, start, start + length - 1)
            
            crs <- tryCatch({ get_file_crs(file_path, conn) }, error = function(e) NULL)
            if (!is.null(crs)) return(crs)
         }
      }
    
      cli::cli_warn(c(
        "CRS could not be detected for {.val {name}}.",
        "i" = "This can happen when reading an older DuckDB storage file that was created before duckspatial persisted CRS metadata, or when the table has no CRS information.",
        "i" = "Use {.code as_duckspatial_df(x, crs = ...)} or {.code ddbs_open_dataset(path, crs = ...)} to set the CRS explicitly."
      ))
      return(sf::st_crs(NA))
    }

    # 2. Return CRS
    return(crs_data)
}

#' @export
#' @rdname ddbs_crs
#' @param name Table name (for backward compatibility when first arg is connection)
ddbs_crs.duckdb_connection <- function(x, name, ...) {
  # Backward compatibility: ddbs_crs(conn, name)
  if (missing(name)) {
    cli::cli_abort("Must provide {.arg name} when calling {.fun ddbs_crs} with a connection.")
  }
  ddbs_crs.character(name, conn = x, ...)
}


#' @export
#' @rdname ddbs_crs
ddbs_crs.numeric <- function(x, ...) {
  # Convert numeric EPSG code to CRS
  # Assumes EPSG authority by default
  if (length(x) != 1) {
      cli::cli_abort("Numeric CRS input must be a single value (EPSG code).")
  }
  
  if (x < 1 || x != as.integer(x)) {
      cli::cli_abort("CRS code must be a positive integer.")
  }

  # Extract the CRS
  crs_x <- sf::st_crs(as.integer(x))
  
  # If the CRS doesn't exist, the previous function returns NA
  if (is.na(crs_x)) {
    cli::cli_abort("CRS code wasn't found.")
  } else {
    return(crs_x)
  }

}


#' @export
#' @rdname ddbs_crs
ddbs_crs.crs <- function(x, ...) {
  return(x)
}


#' @export
#' @rdname ddbs_crs
ddbs_crs.data.frame <- function(x, ...) {
  return(sf::st_crs(NA))
}


#' @export
#' @rdname ddbs_crs
ddbs_crs.default <- function(x, ...) {
  cli::cli_abort(c(
    "{.arg x} must be a duckspatial_df, sf object, tbl_duckdb_connection, or character table name.",
    "i" = "You provided an object of class: {.cls {class(x)}}"
  ))
}






#' Transform the coordinate reference system of geometries
#'
#' Converts geometries to a different coordinate reference system (CRS), updating 
#' their coordinates accordingly.
#'
#' @template x
#' @param y Target CRS. Can be:
#'   \itemize{
#'     \item A character string with EPSG code (e.g., "EPSG:4326")
#'     \item A \code{crs} object created with [sf::st_crs]
#'     \item An \code{sf} object (uses its CRS)
#'     \item Name of a DuckDB table (uses its CRS)
#'   }
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina")
#' 
#' ## transform to different CRS using EPSG code
#' ddbs_transform("argentina", "EPSG:3857", conn)
#' 
#' ## transform to match CRS of another object
#' argentina_3857_ddbs <- ddbs_transform(argentina_ddbs, "EPSG:3857")
#' ddbs_write_vector(conn, argentina_3857_ddbs, "argentina_3857")
#' ddbs_transform("argentina", argentina_3857_ddbs, conn)
#' 
#' ## transform to match CRS of another DuckDB table
#' ddbs_transform("argentina", "argentina_3857", conn)
#' 
#' ## transform without using a connection
#' ddbs_transform(argentina_ddbs, "EPSG:3857")
#' }
ddbs_transform <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet     = FALSE) {

    ## 0. Handle errors
    assert_xy(x, "x")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

    # 1. Manage connection to DB

    ## 1.1. Pre-extract attributes (CRS and geometry column name)
    ## this step should be before normalize_spatial_input()
    crs_x <- if (is.null(conn_x)) ddbs_crs(x, conn) else ddbs_crs(x, conn_x)
    crs_y <- if (is.null(conn_y)) ddbs_crs(y, conn) else ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")

    ## 1.2. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.3. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names
    x <- normalize_spatial_input(x, conn_x)
    try(y <- normalize_spatial_input(y, conn_y), silent = TRUE)

    ## 1.4. Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)


    # 2. Manage connection to DB

    ## 2.1. Resolve connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    y            <- resolve_conn$y
    ## register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_conn$cleanup(), add = TRUE)   
    }

    ## 2.2. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)

    ## if CRS wasn't guessed earlier
    if (is.null(crs_x)) crs_x <- ddbs_crs(x_list$query_name, target_conn)
    if (is.null(crs_y)) {
        ## try to get from `y`. if it fails, it's not sf, nor duckspatial_df
        ## therefore, it might be a CRS or character string with CRS
        try(crs_y <- ddbs_crs(y_list$query_name, target_conn), silent = TRUE)

        if (is.null(crs_y)) crs_y <- sf::st_crs(y)
    }

    ## warn if the crs is the same
    if (crs_x$input == crs_y$input) cli::cli_warn("The CRS of `x` and `y` is the same.")
    
    # 3. Prepare parameters for the query

    ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 3.2. Build the base query
    ## always_xy assumes [northing, easting]
    st_function <- glue::glue("ST_Transform({x_geom}, '{crs_x$input}', '{crs_y$input}', always_xy := true)")
    base.query <- glue::glue("
        SELECT *
        REPLACE ({build_geom_query(st_function, name, crs_y, mode)} AS {x_geom})
        FROM 
            {x_list$query_name};
    ")

    # 4. if name is not NULL (i.e. no SF returned)
    if (!is.null(name)) {

        ## convenient names of table and/or schema.table
        name_list <- get_query_name(name)

        ## handle overwrite
        overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

        ## create query (no st_as_text)
        tmp.query <- glue::glue("
            CREATE TABLE {name_list$query_name} AS
            {base.query};
        ")
        ## execute intersection query
        DBI::dbExecute(target_conn, tmp.query)
        feedback_query(quiet)
        return(invisible(TRUE))
    }

    # 5. Apply geospatial operation
    result <- ddbs_handle_query(
        query      = base.query,
        conn       = target_conn,
        mode       = mode,
        crs        = crs_y,
        x_geom     = x_geom
    )

    return(result)
}





#' Set the coordinate reference system of geometries
#'
#' Assigns or replaces the coordinate reference system (CRS) of geometries
#' without transforming their coordinates. This is useful when the CRS is
#' missing or incorrectly defined.
#'
#' @template x
#' @param y Target CRS. Can be:
#'   \itemize{
#'     \item A character string with EPSG code (e.g., "EPSG:4326")
#'     \item A \code{crs} object created with [sf::st_crs]
#'   }
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(sf)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#' 
#' ## Remove CRS
#' rivers_no_crs_ddbs <- ddbs_set_crs(rivers_ddbs, st_crs(NA))
#'
#' ## Set the CRS back
#' ddbs_set_crs(rivers_no_crs_ddbs, "EPSG:3035")
#' }
ddbs_set_crs <- function(
    x,
    y,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet     = FALSE) {

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")


    # 1. Prepare inputs

    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
  
    ## 1.5. Format CRS as "AUTH:CODE"
    y <- ddbs_crs(y)$input


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)


    ## 2.4. Build the base query (depends on the output type - sf, duckspatial_df, table)
    st_function <- glue::glue("ST_SetCRS({x_geom}, '{y}')")
  
    if (is.null(name) && mode == "sf") {
        ## If not creating a table, fallback to BLOB
        st_function <- glue::glue("ST_AsWKB({st_function})")
    }
  
    base.query <- glue::glue("
        SELECT *
        REPLACE ({st_function} AS {x_geom})
        FROM {x_list$query_name};
    ")

    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = y,
            x_geom = x_geom
        )
    }

}

testthat::skip_on_cran()

parquet_geo_crs_metadata <- function(conn, path) {
  path_sql <- as.character(DBI::dbQuoteString(conn, path))
  query <- paste0(
    "SELECT decode(value) AS geo ",
    "FROM parquet_kv_metadata(", path_sql, ") ",
    "WHERE decode(key) = 'geo' ",
    "LIMIT 1"
  )
  res <- DBI::dbGetQuery(conn, query)
  if (nrow(res) == 0) {
    return(NA_character_)
  }

  res$geo[[1]]
}

sf_read_parquet_or_null <- function(path) {
  tryCatch(
    sf::st_read(path, quiet = TRUE),
    error = function(e) NULL
  )
}

test_that("ddbs_write_dataset works for Parquet", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  tmp_file <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_file), add = TRUE)
  expect_no_error(ddbs_write_dataset(ds, tmp_file, quiet = TRUE))
  expect_true(file.exists(tmp_file))

  # Verify reading back
  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  expect_equal(dplyr::count(ds_back) |> dplyr::pull(n), 257) # Expect same count
})

test_that("ddbs_write_dataset works for GeoJSON (GDAL)", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  tmp_file <- tempfile(fileext = ".geojson")
  on.exit(unlink(tmp_file), add = TRUE)
  expect_no_error(ddbs_write_dataset(ds, tmp_file, quiet = TRUE))
  expect_true(file.exists(tmp_file))

  # Verify reading back
  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  expect_equal(dplyr::count(ds_back) |> dplyr::pull(n), 257)
})


test_that("ddbs_write_dataset overwrite behavior", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  tmp_file <- tempfile(fileext = ".parquet")
  file.create(tmp_file) # Create empty file
  on.exit(unlink(tmp_file), add = TRUE)

  # Ensure file exists to genuinely test overwrite check
  expect_true(file.exists(tmp_file))

  # Should fail by default
  expect_error(ddbs_write_dataset(ds, tmp_file), "already exists")

  # Should succeed with overwrite=TRUE
  expect_no_error(ddbs_write_dataset(ds, tmp_file, overwrite = TRUE, quiet = TRUE))
  expect_gt(file.size(tmp_file), 0)
})

test_that("ddbs_write_dataset partitioning (Parquet)", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  # Add a dummy partition column
  ds_mod <- ds |> dplyr::mutate(part_col = dplyr::if_else(CNTR_NAME == 'Argentina', "AR", "Others"))

  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)
  tmp_file <- file.path(tmp_dir, "partitioned_countries.parquet")

  expect_no_error(
    ddbs_write_dataset(ds_mod, tmp_file, partitioning = "part_col", quiet = TRUE)
  )

  # Check directory structure
  # print(list.files(tmp_dir, recursive = TRUE))
  # print(list.dirs(tmp_dir))

  # Check if directory structure exists (Hive style)
  # Directories are created INSIDE the target path
  expect_true(dir.exists(file.path(tmp_file, "part_col=AR")))
  expect_true(dir.exists(file.path(tmp_file, "part_col=Others")))
})

test_that("ddbs_write_dataset auto-detects partitioning from grouped data", {
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  conn <- ddbs_temp_conn()

  ds <- ddbs_open_dataset(path, conn = conn)

  # Group the lazy table
  ds_grouped <- ds |> dplyr::group_by(CNTR_ID)

  tmp_dir <- tempfile()
  dir.create(tmp_dir)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)
  tmp_file <- file.path(tmp_dir, "auto_partitioned.parquet")

  # Should use CNTR_ID for partitioning automatically
  expect_no_error(
    ddbs_write_dataset(ds_grouped, tmp_file, quiet = TRUE)
  )

  # Verify some partitions exist (CNTR_ID are 2-char codes)
  # Argentina is AR
  expect_true(dir.exists(file.path(tmp_file, "CNTR_ID=AR")))
})

test_that("ddbs_write_dataset validates non-spatial data for spatial format", {
  conn <- ddbs_temp_conn()
  df <- data.frame(a = 1:5, b = letters[1:5])

  tmp_file <- tempfile(fileext = ".geojson")

  expect_error(
    ddbs_write_dataset(df, tmp_file, gdal_driver = "GeoJSON", quiet = TRUE),
    "Input local data must be an 'sf' object"
  )
})

test_that("ddbs_write_dataset fails with plain local data.frame", {
  conn <- ddbs_temp_conn()

  df <- mtcars
  df$geometry <- "POLYGON((0 0, 1 1, 1 0, 0 0))" # Fake geometry column
  tmp_file <- tempfile(fileext = ".parquet")

  # Should fail because it's not an sf object
  expect_error(
    ddbs_write_dataset(df, tmp_file, quiet = TRUE),
    "Input local data must be an 'sf' object"
  )
})

test_that("ddbs_write_dataset fails with fake spatial DuckDB table (wrong type)", {
  conn <- ddbs_temp_conn()

  # Create a table with a column named 'geometry' but it's VARCHAR
  DBI::dbExecute(conn, "CREATE OR REPLACE TABLE fake_spatial AS SELECT 'POINT(0 0)'::VARCHAR AS geometry")
  data <- dplyr::tbl(conn, "fake_spatial")

  tmp_file <- tempfile(fileext = ".parquet")

  # Should fail because type is VARCHAR, not GEOMETRY
  expect_error(
    ddbs_write_dataset(data, tmp_file, quiet = TRUE),
    "does not contain a spatial column of type 'GEOMETRY'"
  )
})

test_that("ddbs_write_dataset CRS override works", {
  skip_if_not_installed("sf")
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  # Use helper with CRS override - helper passes through kwargs
  tmp_file <- ddbs_create_temp_spatial_file(ds, ext = "geojson", conn = conn, crs = "EPSG:3857")

  # Use sf to verify CRS metadata
  sf_obj <- sf::st_read(tmp_file, quiet = TRUE)
  expect_equal(sf::st_crs(sf_obj)$epsg, 3857)
})

test_that("ddbs_write_dataset works with local sf object", {
  skip_if_not_installed("sf")
  conn <- ddbs_temp_conn()

  # Create simple SF
  poly <- sf::st_sfc(sf::st_polygon(list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol=2, byrow=TRUE))))
  sf_obj <- sf::st_sf(id = 1, geometry = poly, crs = 4326)

  tmp_file <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_file), add = TRUE)

  expect_no_error(ddbs_write_dataset(sf_obj, tmp_file, conn = conn, quiet = TRUE))

  # Read back
  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  expect_equal(dplyr::count(ds_back) |> dplyr::pull(n), 1)
  expect_equal(ddbs_crs(ds_back)$epsg, 4326)

  sf_back <- sf_read_parquet_or_null(tmp_file)
  if (!is.null(sf_back)) {
    expect_equal(sf::st_crs(sf_back)$epsg, 4326)
  }
})

test_that("ddbs_write_dataset preserves local sf EPSG CRS in GeoParquet metadata", {
  skip_if_not_installed("sf")
  conn <- ddbs_temp_conn()

  coords <- matrix(
    c(0, 0, 1000000, 0, 1000000, 1000000, 0, 1000000, 0, 0),
    ncol = 2,
    byrow = TRUE
  )
  poly <- sf::st_sfc(sf::st_polygon(list(coords)), crs = 3857)
  sf_obj <- sf::st_sf(id = 1, geometry = poly)

  tmp_file <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_file), add = TRUE)

  expect_no_error(ddbs_write_dataset(sf_obj, tmp_file, conn = conn, quiet = TRUE))

  crs_meta <- parquet_geo_crs_metadata(conn, tmp_file)
  expect_match(crs_meta, '"authority"\\s*:\\s*"EPSG"')
  expect_match(crs_meta, '"code"\\s*:\\s*"?3857"?')

  sf_back <- sf_read_parquet_or_null(tmp_file)
  if (!is.null(sf_back)) {
    expect_equal(sf::st_crs(sf_back)$epsg, 3857)
    expect_gt(unname(sf::st_bbox(sf_back)[["xmax"]]), 100000)
  }

  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  expect_equal(ddbs_crs(ds_back)$epsg, 3857)
})

test_that("ddbs_write_dataset does not invent EPSG CRS metadata for local sf without CRS", {
  skip_if_not_installed("sf")
  conn <- ddbs_temp_conn()

  poly <- sf::st_sfc(sf::st_polygon(list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol=2, byrow=TRUE))))
  sf_obj <- sf::st_sf(id = 1, geometry = poly)

  tmp_file <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_file), add = TRUE)

  expect_no_error(ddbs_write_dataset(sf_obj, tmp_file, conn = conn, quiet = TRUE))

  crs_meta <- parquet_geo_crs_metadata(conn, tmp_file)
  crs_text <- if (is.na(crs_meta)) "" else crs_meta
  expect_false(grepl("EPSG", crs_text, fixed = TRUE))
})

test_that("ddbs_write_dataset preserves local sf non-EPSG CRS in GeoParquet metadata", {
  skip_if_not_installed("sf")
  conn <- ddbs_temp_conn()

  crs <- sf::st_crs("+proj=laea +lat_0=52 +lon_0=10 +x_0=0 +y_0=0 +datum=WGS84 +units=m +no_defs")
  expect_true(is.na(crs$epsg))

  poly <- sf::st_sfc(
    sf::st_polygon(list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol=2, byrow=TRUE))),
    crs = crs
  )
  sf_obj <- sf::st_sf(id = 1, geometry = poly)

  tmp_file <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_file), add = TRUE)

  expect_no_warning(ddbs_write_dataset(sf_obj, tmp_file, conn = conn, quiet = TRUE))
  expect_true(file.exists(tmp_file))

  crs_meta <- parquet_geo_crs_metadata(conn, tmp_file)
  expect_false(is.na(crs_meta))
  expect_match(crs_meta, '"crs"\\s*:\\s*\\{')
  expect_match(crs_meta, '"type"\\s*:\\s*"ProjectedCRS"')

  path_sql <- as.character(DBI::dbQuoteString(conn, tmp_file))
  desc <- DBI::dbGetQuery(conn, paste0("DESCRIBE SELECT * FROM read_parquet(", path_sql, ")"))
  geom_type <- desc$column_type[desc$column_name == attr(sf_obj, "sf_column")]
  expect_true(grepl("^GEOMETRY\\('", geom_type))
  expect_true(grepl("ProjectedCRS", geom_type, fixed = TRUE))

  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  ds_crs <- ddbs_crs(ds_back)
  expect_false(is.na(ds_crs))
  expect_true(is.na(ds_crs$epsg))

  sf_back <- sf_read_parquet_or_null(tmp_file)
  if (!is.null(sf_back)) {
    sf_crs <- sf::st_crs(sf_back)
    expect_false(is.na(sf_crs))
    expect_true(is.na(sf_crs$epsg))
  }
})

# Driver validation tests

test_that("ddbs_write_dataset validates invalid driver", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)

  tmp_file <- tempfile(fileext = ".fake")

  # Should error when specifying non-existent driver
  expect_error(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "NonExistentDriver", quiet = TRUE),
    "is not available on this system"
  )
})

# skip tests on CRAN
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

# helpers --------------------------------------------------------------

# create duckdb connection
conn_test <- ddbs_create_conn()

## store countries
ddbs_write_table(conn_test, countries_sf, "countries")

## store countries with different CRS
countries_3857_sf <- sf::st_transform(countries_sf, "EPSG:3857")
ddbs_write_table(conn_test, countries_3857_sf, "countries_3857_test")


# 1. ddbs_transform() ----------------------------------------------------

describe("ddbs_transform()", {
  
  ### EXPECTED BEHAVIOUR - AUTH:CODE CRS
  
  describe("expected behavior with AUTH:CODE CRS", {
    
    it("transforms sf to specified CRS code", {
      output <- ddbs_transform(x = countries_sf, y = "EPSG:3857")
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("transforms DuckDB table to specified CRS code", {
      output <- ddbs_transform(x = "countries", y = "EPSG:3857", conn = conn_test)
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("writes transformed DuckDB table to database", {
      output <- ddbs_transform(
        x    = "countries",
        y    = "EPSG:3857",
        conn = conn_test,
        name = "countries_3857"
      )
      
      expect_true(output)
    })
    
    it("overwrites existing table when overwrite = TRUE", {
      output <- ddbs_transform(
        x         = "countries",
        y         = "EPSG:3857",
        conn      = conn_test,
        name      = "countries_3857",
        overwrite = TRUE,
        quiet     = TRUE
      )
      
      expect_true(output)
    })
  })
  
  ### EXPECTED BEHAVIOUR - SF OBJECT CRS
  
  describe("expected behavior with sf object CRS", {
    
    it("transforms sf to CRS from another sf object", {
      countries_3857_sf <- ddbs_transform(x = countries_sf, y = "EPSG:3857", quiet = TRUE)
      output <- ddbs_transform(x = countries_sf, y = countries_3857_sf)
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("transforms DuckDB table to CRS from sf object", {
      countries_3857_sf <- ddbs_transform(x = countries_sf, y = "EPSG:3857", quiet = TRUE)
      output <- ddbs_transform(x = "countries", y = countries_3857_sf, conn = conn_test)
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("writes transformed DuckDB table to database with sf CRS", {
      countries_3857_sf <- ddbs_transform(x = countries_sf, y = "EPSG:3857", quiet = TRUE)
      output <- ddbs_transform(
        x    = "countries",
        y    = countries_3857_sf,
        conn = conn_test,
        name = "countries_3857_sf"
      )
      
      expect_true(output)
    })
    
    it("overwrites existing table when overwrite = TRUE with sf CRS", {
      countries_3857_sf <- ddbs_transform(x = countries_sf, y = "EPSG:3857", quiet = TRUE)
      output <- ddbs_transform(
        x         = "countries",
        y         = countries_3857_sf,
        conn      = conn_test,
        name      = "countries_3857_sf",
        overwrite = TRUE,
        quiet     = TRUE
      )
      
      expect_true(output)
    })
  })
  
  ### EXPECTED BEHAVIOUR - DUCKDB TABLE CRS
  
  describe("expected behavior with DuckDB table CRS", {
    
    it("transforms sf to CRS from DuckDB table", {
      output <- ddbs_transform(
        x    = countries_sf,
        y    = "countries_3857_test",
        conn = conn_test
      )
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("transforms DuckDB table to CRS from another DuckDB table", {
      countries_3857_sf <- ddbs_transform(x = countries_sf, y = "EPSG:3857", quiet = TRUE)
      output <- ddbs_transform(
        x    = "countries",
        y    = countries_3857_sf,
        conn = conn_test
      )
      
      expect_equal(sf::st_crs(output), sf::st_crs("EPSG:3857"))
    })
    
    it("writes transformed DuckDB table to database with DuckDB table CRS", {
      output <- ddbs_transform(
        x    = "countries",
        y    = "countries_3857_test",
        conn = conn_test,
        name = "countries_3857_table"
      )
      
      expect_true(output)
    })
    
    it("overwrites existing table when overwrite = TRUE with DuckDB table CRS", {
      output <- ddbs_transform(
        x         = "countries",
        y         = "countries_3857_test",
        conn      = conn_test,
        name      = "countries_3857_table",
        overwrite = TRUE,
        quiet     = TRUE
      )
      
      expect_true(output)
    })
  })
  
  ### EXPECTED BEHAVIOUR - CRS OBJECT
  
  describe("expected behavior with CRS object", {
    
    it("transforms sf to specified CRS object", {
      output <- ddbs_transform(x = rivers_sf, y = sf::st_crs(countries_sf))
      
      expect_equal(ddbs_crs(output), sf::st_crs(countries_sf))
    })
    
    it("transforms DuckDB table to specified CRS object", {
      output <- ddbs_transform(
        x    = "countries",
        y    = sf::st_crs(rivers_sf),
        conn = conn_test
      )
      
      expect_equal(ddbs_crs(output), sf::st_crs(rivers_sf))
    })
    
    it("writes transformed DuckDB table to database with CRS object", {
      output <- ddbs_transform(
        x    = "countries",
        y    = sf::st_crs(rivers_sf),
        conn = conn_test,
        name = "countries_3035_table"
      )
      
      expect_true(output)
    })
    
    it("overwrites existing table when overwrite = TRUE with CRS object", {
      output <- ddbs_transform(
        x         = "countries",
        y         = sf::st_crs(rivers_sf),
        conn      = conn_test,
        name      = "countries_3035_table",
        overwrite = TRUE,
        quiet     = TRUE
      )
      
      expect_true(output)
    })
  })
  
  ### EXPECTED WARNINGS
  
  describe("warnings", {
    
    it("warns when transforming sf to same CRS (DuckDB table)", {
      expect_warning(ddbs_transform(x = countries_sf, y = "countries", conn = conn_test))
    })
    
    it("warns when transforming DuckDB table to same CRS (sf)", {
      expect_warning(ddbs_transform(x = "countries", y = countries_sf, conn = conn_test))
    })
    
    it("warns when creating new table with same CRS", {
      expect_warning(
        ddbs_transform(
          x    = "countries",
          y    = countries_sf,
          conn = conn_test,
          name = "countries_3857_table_warn"
        )
      )
    })
    
    it("warns when overwriting table with same CRS", {
      expect_warning(
        ddbs_transform(
          x         = "countries",
          y         = countries_sf,
          conn      = conn_test,
          name      = "countries_3857_table",
          overwrite = TRUE,
          quiet     = TRUE
        )
      )
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_transform(x = "999", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_transform(x = 999))
      expect_error(ddbs_transform(x = "999", conn = conn_test))
    })
    
    it("validates y argument type", {
      expect_error(ddbs_transform(y = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_transform(conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_transform(overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_transform(quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_transform(conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



## stop connection
ddbs_stop_conn(conn_test)

# Test suite for ddbs_crs S3 generic and methods
testthat::skip_on_cran()

test_that("ddbs_crs works for different input types", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  # Setup
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn))
  
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)[1:5,]
  
  # 1. SF Object method
  expect_equal(ddbs_crs(nc_sf), sf::st_crs(nc_sf))
  
  # 2. duckspatial_df method
  ddbs_write_table(conn, nc_sf, "nc_data")
  # Use default retrieval options
  nc_ds <- as_duckspatial_df("nc_data", conn, crs = sf::st_crs(nc_sf))
    
  expect_equal(ddbs_crs(nc_ds), sf::st_crs(nc_sf))
  
  # 3. tbl_duckdb_connection (lazy) method
  # Plain lazy table without duckspatial_df wrapper
  # nc_lazy <- dplyr::tbl(conn, "nc_data")
  nc_lazy <- as_duckspatial_df("nc_data", conn)
  # ddbs_crs should fallback to looking up the table in the DB if it is a simple table ref
  # Our implementation for tbl_duckdb_connection tries to detect from VIEW SQL or ST_Read
  # BUT "nc_data" is a table with 'crs_duckspatial' column?
  # ddbs_crs.tbl_duckdb_connection implementation ONLY looks for ST_Read views
  # It falls back to NA if not found.
  
  # Wait, if I pass a tbl(conn, "table"), it's a tbl_duckdb_connection.
  # Does logic support finding CRS from 'crs_duckspatial' column for lazy tables?
  # ddbs_crs.tbl_duckdb_connection logic I wrote prioritizes auto-detection from SQL.
  # If that fails, it returns NA.
  # Ideally it should check for 'crs_duckspatial' column too?
  # Currently it prints a warning and returns NA. 
  # Let's verifying assumption - it might be NA for plain tables.
  
  # Let's test the specific feature: Auto-detection from view
  path <- system.file("shape/nc.shp", package = "sf")
  DBI::dbExecute(conn, glue::glue("CREATE VIEW nc_view AS SELECT * FROM ST_Read('{path}')"))
  # nc_view_lazy <- dplyr::tbl(conn, "nc_view")
  nc_view_lazy <- as_duckspatial_df("nc_view", conn)
  
  crs_auto <- ddbs_crs(nc_view_lazy)
  expect_false(is.na(crs_auto))
  expect_equal(crs_auto$epsg, 4267) # NC shapefile has 4267
  
  # 4. Character method
  # Uses 'crs_duckspatial' column by default
  crs_char <- ddbs_crs("nc_data", conn = conn)
  expect_true(crs_char == sf::st_crs(nc_sf))
  
  # 5. Connection method (Backward Compatibility)
  # ddbs_crs(conn, name)
  crs_compat <- ddbs_crs(conn, "nc_data")
  expect_true(crs_compat == sf::st_crs(nc_sf))
  
  # 6. Default/Error
  expect_error(ddbs_crs(NULL))
  expect_error(ddbs_crs(123))
})



# 2. ddbs_set_crs() -------------------------------------------------------

## - CHECK 1.1: works on ddbs input (string, crs object, NA)
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: works on sf input
## - CHECK 1.4: sf returns different outputs (duckspatial_df, sf)
## - CHECK 1.5: works on DuckDB table input
## - CHECK 1.6: DuckDB table returns different outputs (duckspatial_df, sf)
## - CHECK 1.7: message is shown with quiet = FALSE
## - CHECK 1.8: no message is shown with quiet = TRUE
## - CHECK 1.9: CRS is set correctly from EPSG string
## - CHECK 1.10: CRS is set correctly from sf::st_crs() object
## - CHECK 1.11: CRS is NA after removal
## - CHECK 1.12: CRS roundtrip (remove then re-assign)
## - CHECK 2.1: requires connection for character input
## - CHECK 2.2: other errors

conn_set_crs <- duckspatial::ddbs_create_conn()
ddbs_write_table(conn_set_crs, argentina_sf, "argentina_scrs")

describe("ddbs_set_crs()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("works on ddbs input", {
      output_ddbs_1 <- ddbs_set_crs(argentina_ddbs, "EPSG:4326")
      output_ddbs_2 <- ddbs_set_crs(argentina_ddbs, sf::st_crs("EPSG:4326"))
      output_ddbs_3 <- ddbs_set_crs(argentina_ddbs, sf::st_crs(NA))
      output_ddbs_4 <- ddbs_set_crs(argentina_ddbs, "EPSG:4326", quiet = TRUE)

      expect_s3_class(output_ddbs_1, "duckspatial_df")
      expect_s3_class(output_ddbs_2, "duckspatial_df")
      expect_s3_class(output_ddbs_3, "duckspatial_df")
      expect_s3_class(output_ddbs_4, "duckspatial_df")
    })

    it("returns different output formats for ddbs input", {
      output_sf <- ddbs_set_crs(argentina_ddbs, "EPSG:4326", mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on sf input", {
      output_sf_1 <- ddbs_set_crs(argentina_sf, "EPSG:4326")
      output_sf_2 <- ddbs_set_crs(argentina_sf, sf::st_crs("EPSG:4326"))
      output_sf_3 <- ddbs_set_crs(argentina_sf, sf::st_crs(NA))
      output_sf_4 <- ddbs_set_crs(argentina_sf, "EPSG:4326", quiet = TRUE)

      expect_s3_class(output_sf_1, "duckspatial_df")
      expect_s3_class(output_sf_2, "duckspatial_df")
      expect_s3_class(output_sf_3, "duckspatial_df")
      expect_s3_class(output_sf_4, "duckspatial_df")
    })

    it("returns different output formats for sf input", {
      output_sf <- ddbs_set_crs(argentina_sf, "EPSG:4326", mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("works on DuckDB table input", {
      output_conn_1 <- ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs)
      output_conn_2 <- ddbs_set_crs("argentina_scrs", sf::st_crs("EPSG:4326"), conn = conn_set_crs)
      output_conn_3 <- ddbs_set_crs("argentina_scrs", sf::st_crs(NA), conn = conn_set_crs)
      output_conn_4 <- ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs, quiet = TRUE)

      expect_s3_class(output_conn_1, "duckspatial_df")
      expect_s3_class(output_conn_2, "duckspatial_df")
      expect_s3_class(output_conn_3, "duckspatial_df")
      expect_s3_class(output_conn_4, "duckspatial_df")
    })

    it("returns different output formats for DuckDB table input", {
      output_sf <- ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs, mode = "sf")
      expect_s3_class(output_sf, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_set_crs(argentina_ddbs, "EPSG:4326"))
      expect_message(ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs, name = "argentina_scrs_out"))
      expect_message(ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs, name = "argentina_scrs_out", overwrite = TRUE))
      expect_true(ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = conn_set_crs, name = "argentina_scrs_out2"))

      expect_no_message(ddbs_set_crs(argentina_ddbs, "EPSG:4326", quiet = TRUE))
      expect_no_message(
        ddbs_set_crs(
          "argentina_scrs",
          "EPSG:4326",
          conn      = conn_set_crs,
          name      = "argentina_scrs_out",
          overwrite = TRUE,
          quiet     = TRUE
        )
      )
    })

    describe("CRS assignment", {

      it("sets CRS correctly from EPSG string", {
        result <- ddbs_set_crs(argentina_ddbs, "EPSG:4326")
        expect_equal(ddbs_crs(result)$epsg, 4326L)
      })

      it("sets CRS correctly from sf::st_crs() object", {
        result <- ddbs_set_crs(argentina_ddbs, sf::st_crs("EPSG:4326"))
        expect_equal(ddbs_crs(result)$epsg, 4326L)
      })

      it("removes CRS when set to sf::st_crs(NA)", {
        result <- ddbs_set_crs(argentina_ddbs, sf::st_crs(NA))
        expect_true(is.na(ddbs_crs(result)))
      })

      it("re-assigns CRS after removal", {
        no_crs   <- ddbs_set_crs(argentina_ddbs, sf::st_crs(NA))
        with_crs <- ddbs_set_crs(no_crs, "EPSG:4326")
        expect_equal(ddbs_crs(with_crs)$epsg, 4326L)
      })

      it("tags geometry with a different CRS without transforming coordinates", {
        result_sf   <- ddbs_set_crs(argentina_ddbs, "EPSG:3035", mode = "sf")
        orig_coords <- sf::st_coordinates(argentina_sf)[, c("X", "Y")]
        res_coords  <- sf::st_coordinates(result_sf)[, c("X", "Y")]

        expect_equal(res_coords, orig_coords, tolerance = 1e-6)
        expect_equal(sf::st_crs(result_sf)$epsg, 3035L)
      })
    })
  })

  ### EXPECTED ERRORS ---------------------------------------------------

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_set_crs("argentina_scrs", "EPSG:4326", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_set_crs(x = 999, "EPSG:4326"))
    })

    it("validates conn argument type", {
      expect_error(ddbs_set_crs(argentina_ddbs, "EPSG:4326", conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_set_crs(argentina_ddbs, "EPSG:4326", overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_set_crs(argentina_ddbs, "EPSG:4326", quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_set_crs("nonexistent_table", "EPSG:4326", conn = conn_set_crs))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_set_crs(argentina_ddbs, "EPSG:4326", conn = conn_set_crs, name = c("a", "b")))
    })
  })
})

duckspatial::ddbs_stop_conn(conn_set_crs)

testthat::skip_on_cran()

test_that("Compatibility: Arrow Views behave like Persistent Tables", {
    skip_if_not_installed("duckdb")

    # Setup
    conn <- ddbs_temp_conn()

    # Create test data with a NON-STANDARD geometry column name
    # This tests if register/read respects column naming
    data_sf <- sf::st_as_sf(
        data.frame(id = 1:5, x = 0, y = 0, val = letters[1:5]),
        coords = c("x", "y"),
        crs = 4326
    )
    # sf::st_geometry(data_sf) <- "my_custom_geom"

    # 1. Register as Arrow View
    expect_no_error(
        ddbs_register_table(conn, data_sf, "view_test", overwrite = TRUE)
    )

    # 2. Test ddbs_crs on View
    expect_no_error(crs_out <- ddbs_crs(conn, "view_test"))
    expect_equal(crs_out, sf::st_crs(4326))

    # 3. Test ddbs_read_table on View
    # Should handle WKB conversion automatically and preserve "my_custom_geom"
    read_view <- ddbs_read_table(conn, "view_test")

    expect_s3_class(read_view, "sf")
    # expect_equal(attr(read_view, "sf_column"), "my_custom_geom")
    expect_equal(nrow(read_view), 5)

    # 4. Verify Data Integrity vs Persistent Table
    ddbs_write_table(conn, data_sf, "table_test", overwrite = TRUE)
    read_table <- ddbs_read_table(conn, "table_test")

    # Compare View result vs Table result
    # (Ignore attribute order if necessary, but data should match)
    expect_equal(sf::st_drop_geometry(read_view), sf::st_drop_geometry(read_table))
    expect_equal(sf::st_geometry(read_view), sf::st_geometry(read_table))
})


test_that("Round trip: write -> read for various geometry types", {
    skip_if_not_installed("duckdb")
    conn <- ddbs_temp_conn()

    # Test data
    line <- sf::st_as_sfc("LINESTRING(0 0, 1 1)") |> sf::st_sf(id = 1, geom = _, crs = 4326)
    polygon <- sf::st_as_sfc("POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))") |> sf::st_sf(id = 1, geom = _, crs = 4326)
    multipoint <- sf::st_as_sfc("MULTIPOINT(0 0, 1 1)") |> sf::st_sf(id = 1, geom = _, crs = 4326)

    datasets <- list(line = line, polygon = polygon, multipoint = multipoint)

    for (name in names(datasets)) {
        data <- datasets[[name]]
        table_name <- paste0("rt_write_", name)

        # Write -> Read
        ddbs_write_table(conn, data, table_name, overwrite = TRUE)
        result <- ddbs_read_table(conn, table_name)

        # Verification
        expect_s3_class(result, "sf")
        expect_equal(nrow(result), 1)
        expect_equal(sf::st_crs(result), sf::st_crs(data))
        # use all.equal for comparing geometries to handle potential precision issues
        expect_true(all.equal(sf::st_geometry(result), sf::st_geometry(data), check.attributes = FALSE))
    }
})

test_that("Round trip: register -> read for various geometry types", {
    skip_if_not_installed("duckdb")
    conn <- ddbs_temp_conn()

    # Test data
    line <- sf::st_as_sfc("LINESTRING(0 0, 1 1)") |> sf::st_sf(id = 1, geom = _, crs = 4326)
    polygon <- sf::st_as_sfc("POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))") |> sf::st_sf(id = 1, geom = _, crs = 4326)
    multipoint <- sf::st_as_sfc("MULTIPOINT(0 0, 1 1)") |> sf::st_sf(id = 1, geom = _, crs = 4326)

    datasets <- list(line = line, polygon = polygon, multipoint = multipoint)

    for (name in names(datasets)) {
        data <- datasets[[name]]
        view_name <- paste0("rt_register_", name)

        # Register -> Read
        ddbs_register_table(conn, data, view_name, overwrite = TRUE)
        result <- ddbs_read_table(conn, view_name)

        # Verification
        expect_s3_class(result, "sf")
        expect_equal(nrow(result), 1)
        expect_equal(sf::st_crs(result), sf::st_crs(data))
        expect_true(all.equal(sf::st_geometry(result), sf::st_geometry(data), check.attributes = FALSE))
    }
})

test_that("Compatibility: Writing from file path and reading back", {
    skip_if_not_installed("duckdb")
    conn <- ddbs_temp_conn()

    file_path <- system.file("spatial/countries.geojson", package = "duckspatial")
    table_name <- "countries_from_file_compat"

    # Write from file, then read back
    ddbs_write_table(conn, file_path, table_name, overwrite = TRUE)
    read_data <- ddbs_read_table(conn, table_name)

    # Read the original file directly with sf for comparison
    original_sf <- sf::st_read(file_path, quiet = TRUE)

    # Basic checks
    expect_s3_class(read_data, "sf")
    expect_equal(nrow(read_data), nrow(original_sf))

    # CRS check (read_data will have CRS)
    expect_equal(sf::st_crs(read_data)$srid, sf::st_crs(original_sf)$srid)

    expect_true("NAME_ENGL" %in% names(read_data))
    expect_true(all.equal(
        sf::st_geometry(read_data),
        sf::st_geometry(original_sf),
        check.attributes = FALSE
    ))
})

# skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
skip_on_cran()
skip_if_not_installed("duckdb")


# helpers --------------------------------------------------------------

# create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

# helper function
tester <- function(data = points_sf,
                   name = "test_view",
                   conn = conn_test,
                   overwrite = FALSE,
                   quiet = FALSE) {
    ddbs_register_table(
        conn,
        data,
        name,
        overwrite,
        quiet
    )
}


# expected behavior --------------------------------------------------------------

test_that("can register sf object as arrow view", {

    # register sf object as view
    result <- tester(
        data = points_sf,
        name = "points_view"
    )

    expect_true(result)

    # check that view exists in arrow views (under hidden raw prefix)
    arrow_views <- duckdb::duckdb_list_arrow(conn_test)
    expect_true("__raw_points_view" %in% arrow_views)

})

test_that("can read registered view back with ddbs_read_table", {

    # register sf object as view
    ddbs_register_table(conn_test, points_sf, "points_view2", overwrite = TRUE)

    # read back
    result <- ddbs_read_table(conn_test, "points_view2")
  

    # check that result is sf object
    expect_true(inherits(result, "sf"))

    # check that number of rows matches
    expect_equal(nrow(result), nrow(points_sf))

    # check that CRS matches
    expect_equal(sf::st_crs(result), sf::st_crs(points_sf))

})


test_that("overwrite=TRUE replaces existing view", {

    # register first view
    ddbs_register_table(conn_test, points_sf, "overwrite_test", overwrite = FALSE)

    # try to register again without overwrite - should error
    expect_error(
        ddbs_register_table(conn_test, countries_sf, "overwrite_test", overwrite = FALSE)
    )

    # register again with overwrite - should succeed
    result <- ddbs_register_table(conn_test, countries_sf, "overwrite_test", overwrite = TRUE)
    expect_true(result)

    # check that the view now contains countries data
    count_result <- DBI::dbGetQuery(conn_test, "SELECT COUNT(*) as n FROM overwrite_test")
    expect_equal(count_result$n, nrow(countries_sf))

})

test_that("can register sf object from file path", {

    # get path to test file
    file_path <- system.file("spatial/countries.geojson", package = "duckspatial")

    # register from file path
    result <- ddbs_register_table(conn_test, file_path, "countries_from_file", overwrite = TRUE)

    expect_true(result)

    # check that view exists
    arrow_views <- duckdb::duckdb_list_arrow(conn_test)
    expect_true("__raw_countries_from_file" %in% arrow_views)

})

# test_that("registered view contains geometry column", {

#     # register sf object
#     ddbs_register_table(conn_test, points_sf, "geom_test", overwrite = TRUE)

#     # check columns
#     columns <- DBI::dbListFields(conn_test, "geom_test")

#     expect_true("geometry" %in% columns)

# })

test_that("registered view contains CRS column or has CRS (in duckdb 1.5+)", {
    # register sf object
    ddbs_register_table(conn_test, points_sf, "crs_test", overwrite = TRUE)

    # check that the requested view is queryable
    count_result <- DBI::dbGetQuery(conn_test, "SELECT COUNT(*) as n FROM crs_test")
    expect_equal(count_result$n, nrow(points_sf))

    # check that the exposed view has a CRS-bearing geometry type
    desc <- DBI::dbGetQuery(conn_test, "DESCRIBE crs_test")
    geom_type <- desc$column_type[desc$column_name == attr(points_sf, "sf_column")]
    expect_equal(geom_type, "GEOMETRY('EPSG:4326')")

    duckdb_crs <- DBI::dbGetQuery(
        conn_test,
        "SELECT ST_CRS(geometry) AS crs FROM crs_test LIMIT 1"
    )
    expect_equal(duckdb_crs$crs, "EPSG:4326")

    crs_test <- ddbs_crs(conn_test, "crs_test")

    expect_s3_class(crs_test, "crs")
})

test_that("registered sf object without CRS exposes generic geometry", {
    points_no_crs <- points_sf
    sf::st_crs(points_no_crs) <- NA

    ddbs_register_table(conn_test, points_no_crs, "no_crs_test", overwrite = TRUE)

    count_result <- DBI::dbGetQuery(conn_test, "SELECT COUNT(*) as n FROM no_crs_test")
    expect_equal(count_result$n, nrow(points_no_crs))

    desc <- DBI::dbGetQuery(conn_test, "DESCRIBE no_crs_test")
    geom_type <- desc$column_type[desc$column_name == attr(points_no_crs, "sf_column")]
    expect_equal(geom_type, "GEOMETRY")
})

test_that("registered sf object with non-EPSG CRS keeps queryable DuckDB CRS", {
    custom_crs <- "+proj=longlat +a=6378137 +b=6378137 +no_defs"
    custom_sf <- sf::st_as_sf(
        data.frame(id = 1, x = 0, y = 0),
        coords = c("x", "y"),
        crs = custom_crs
    )

    expect_true(is.na(sf::st_crs(custom_sf)$epsg))

    ddbs_register_table(conn_test, custom_sf, "custom_crs_test", overwrite = TRUE)

    duckdb_crs <- DBI::dbGetQuery(
        conn_test,
        "SELECT ST_CRS(geometry) AS crs FROM custom_crs_test LIMIT 1"
    )
    expect_false(is.na(duckdb_crs$crs[[1]]))
    expect_s3_class(sf::st_crs(duckdb_crs$crs[[1]]), "crs")
})

# expected errors --------------------------------------------------------------

test_that("error when view name exists and overwrite=FALSE", {

    # register first view
    ddbs_register_table(conn_test, points_sf, "duplicate_test", overwrite = TRUE)

    # try to register again - should error
    expect_error(
        ddbs_register_table(conn_test, points_sf, "duplicate_test", overwrite = FALSE),
        "overwrite = TRUE"
    )

})

# New tests for duckspatial_df and existing crs_duckspatial column ----

test_that("can register duckspatial_df directly", {
    # Create a duckspatial_df by reading from existing view
    ddbs_register_table(conn_test, points_sf, "points_for_lazy", overwrite = TRUE)
    df_lazy <- ddbs_read_table(conn_test, "points_for_lazy") |>
        as_duckspatial_df()

    # Register duckspatial_df as new view
    result <- ddbs_register_table(conn_test, df_lazy, "lazy_view_direct", overwrite = TRUE)
    expect_true(result)

    # Verify view exists
    arrow_views <- duckdb::duckdb_list_arrow(conn_test)
    expect_true("__raw_lazy_view_direct" %in% arrow_views)

    # Verify data is queryable
    count_result <- DBI::dbGetQuery(conn_test, "SELECT COUNT(*) as n FROM lazy_view_direct")
    expect_equal(count_result$n, nrow(points_sf))
})


test_that("error for unsupported data types", {
    expect_error(
        ddbs_register_table(conn_test, list(a = 1), "bad_input"),
        "must be an"
    )

    expect_error(
        ddbs_register_table(conn_test, 123, "bad_input"),
        "must be an"
    )
})

#' Generate random points within bounding boxes of geometries
#'
#' Creates random points within the bounding box of each geometry, which may 
#' fall outside the geometry itself.
#'
#' @template x
#' @param n Number of random points to generate within each geometry
#' @param seed A number for the random number generator
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#'
#' ## generate 100 random points within each geometry
#' ddbs_generate_points("argentina", n = 100, conn)
#'
#' ## generate points without using a connection
#' ddbs_generate_points(argentina_ddbs, n = 100)
#' }
ddbs_generate_points <- function(
  x,
  n,
  seed = NULL,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_numeric(n, "n")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


 # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x <- ddbs_crs(x, conn)
  mode  <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query
  
  ## 2.1. Get the bounding box of the input geometries
  bbox <- ddbs_bbox(x_list$query_name, by_feature = FALSE, conn = target_conn, quiet = TRUE)

  ## 2.2. Create the table and store it as a view
  view_name_tbl <- ddbs_temp_view_name()
  generate_points_query <- if (is.null(seed)) {
    glue::glue("ST_GeneratePoints({{min_x: {bbox$xmin}, min_y: {bbox$ymin}, max_x: {bbox$xmax}, max_y: {bbox$ymax}}}::BOX_2D, {n}) as geometry")
  } else {
    glue::glue("ST_GeneratePoints({{min_x: {bbox$xmin}, min_y: {bbox$ymin}, max_x: {bbox$xmax}, max_y: {bbox$ymax}}}::BOX_2D, {n}, {seed}) as geometry")
  }
  tmp.query   <- glue::glue("
    CREATE TEMP VIEW '{view_name_tbl}' AS 
    SELECT
      ST_X(point) AS x,
      ST_Y(point) AS y,
    FROM 
       {generate_points_query};
  ")
  DBI::dbExecute(target_conn, tmp.query)

  ## 2.3. Build base query  
  st_function <- "ST_Point(x, y)"
  base.query <- glue::glue("
    SELECT {build_geom_query(st_function, name, crs_x, mode)} as geometry
    FROM {view_name_tbl};
  ")  


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query      = base.query,
      conn       = target_conn,
      mode       = mode,
      crs        = crs_x,
      x_geom     = "geometry"
    )
  }

}

#' @param id_x Character; optional name of the column in `x` whose values will
#' be used to name the list elements. If `NULL`, integer row numbers of `x` are used.
#' @param id_y Character; optional name of the column in `y` whose values will
#' replace the integer indices returned in each element of the list.
#' @param sparse A logical value. If `TRUE`, it returns a sparse table/list. If `FALSE`,
#' it returns a wide table/matrix.

.onLoad <- function(libname, pkgname) {
  op <- options()
  op_duckspatial <- list(
    duckspatial.output_type = "duckspatial_df",
    duckspatial.max_rows = 1e6,
    duckspatial.duckdb_storage_version = "v1.5.0"
  )
  toset <- !(names(op_duckspatial) %in% names(op))
  if (any(toset)) options(op_duckspatial[toset])
  
  # Make internal duckdb functions available if needed
  # This serves as a reminder that we use asNamespace("duckdb") in the code
}

.onAttach <- function(libname, pkgname) {
  packageStartupMessage(
    paste("duckspatial", utils::packageVersion("duckspatial"), "attached"),
    "\n* Compatible with DuckDB >= v1.5.4.2"
  )

  # Notify about default output change
  packageStartupMessage(
      "\nDefault output has changed on v1.0.0:",
      "\n  duckspatial now returns lazy `duckspatial_df` (dbplyr) objects",
      "\n  instead of `sf` objects.",
      "\n\nTo restore the previous behaviour:",
      "\n  ddbs_options(mode = 'sf')"
    )
}

.onUnload <- function(libpath) {
  conn <- getOption("duckspatial_conn", NULL)
  if (!is.null(conn) && DBI::dbIsValid(conn)) {
    tryCatch(DBI::dbDisconnect(conn), error = function(e) NULL)
  }
  options(duckspatial_conn = NULL)
}

#' Creates a buffer around geometries
#'
#' Computes a polygon that represents all locations within a specified distance from the
#' original geometry
#'
#' @template x
#' @param distance a numeric value specifying the buffer distance. Units correspond to
#' the coordinate system of the geometry (e.g. degrees or meters)
#' @param num_triangles an integer representing how many triangles will be produced to 
#' approximate a quarter circle. The larger the number, the smoother the resulting geometry. 
#' Default is 8.
#' @param cap_style a character string specifying the cap style. Must be one of 
#' "CAP_ROUND" (default), "CAP_FLAT", or "CAP_SQUARE". Case-insensitive.
#' @param join_style a character string specifying the join style. Must be one of 
#' "JOIN_ROUND" (default), "JOIN_MITRE", or "JOIN_BEVEL". Case-insensitive.
#' @param mitre_limit a numeric value specifying the mitre limit ratio. Only applies when 
#' \code{join_style} is "JOIN_MITRE". It is the ratio of the distance from the corner to 
#' the mitre point to the corner radius. Default is 1.0.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina")
#'
#' ## basic buffer
#' ddbs_buffer(conn = conn, "argentina", distance = 1)
#'
#' ## buffer with custom parameters
#' ddbs_buffer(conn = conn, "argentina", distance = 1, 
#'             num_triangles = 16, cap_style = "CAP_SQUARE")
#'
#' ## buffer without using a connection
#' ddbs_buffer(argentina_ddbs, distance = 1)
#' }
ddbs_buffer <- function(
  x,
  distance,
  num_triangles = 8L,
  cap_style = "CAP_ROUND",
  join_style = "JOIN_ROUND",
  mitre_limit = 1.0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  
  # 0. Handle function-specific errors
  assert_numeric(distance, "distance")
  assert_integer_scalar(num_triangles, "num_triangles")
  assert_character_scalar(cap_style, "cap_style")
  assert_character_scalar(join_style, "join_style")
  assert_numeric(mitre_limit, "mitre_limit")

  ## Validate cap_style
  valid_cap_styles <- c("CAP_ROUND", "CAP_FLAT", "CAP_SQUARE")
  if (!toupper(cap_style) %in% valid_cap_styles) {
      cli::cli_abort("{.arg cap_style} must be one of: {.val {paste0(valid_cap_styles, collapse = ', ')}}.")
  }
  
  ## Validate join_style
  valid_join_styles <- c("JOIN_ROUND", "JOIN_MITRE", "JOIN_BEVEL")
  if (!toupper(join_style) %in% valid_join_styles) {
      cli::cli_abort("{.arg join_style} must be one of: {.val {paste0(valid_join_styles, collapse = ', ')}}.")
  }
  
  ## Check num_triangles is positive
  if (num_triangles < 1) cli::cli_abort("{.arg num_triangles} must be a positive integer")
  
  ## Check mitre_limit is striclty positive
  if (mitre_limit <= 0) cli::cli_abort("{.arg mitre_limit} must be a positive number")
  

  # 1. Build ST_Buffer parameters string
  buffer_args <- glue::glue("{distance}, {as.integer(num_triangles)}, '{toupper(cap_style)}', '{toupper(join_style)}', {mitre_limit}")
  

  # 2. Pass to template
  template_unary_ops(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Buffer",
    other_args = buffer_args
  )

}





#' Calculates the centroid of geometries
#'
#' Returns the geometric center (centroid) of a geometry as a point, 
#' representing its average position.
#'
#' @template x
#' @param method Character string specifying the method to calculate the 
#' centroid. Must be one of "centroid" (default) or "surface". "centroid" 
#' calculates the default centroid, which may fall outside the geometry 
#' for certain shapes (e.g., donuts). "surface" calculates a point guaranteed 
#' to fall within the geometry.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina")
#'
#' ## centroid
#' ddbs_centroid("argentina", conn)
#'
#' ## centroid without using a connection
#' ddbs_centroid(argentina_ddbs)
#' }
ddbs_centroid <- function(
    x,
    method = "centroid",
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet     = FALSE) {
    
    st_fun <- switch(
        method,
        centroid = "ST_Centroid",
        surface  = "ST_PointOnSurface",
        cli::cli_abort("Invalid method. Must be one of 'centroid', or 'surface'.")
    )
    
    template_unary_ops(
        x          = x,
        conn       = conn,
        name       = name,
        mode       = mode,
        overwrite  = overwrite,
        quiet      = quiet,
        fun        = st_fun,
        other_args = NULL
    )

}





#' Make invalid geometries valid
#'
#' Attempts to correct invalid geometries so they conform to the rules of well-formed 
#' geometries (e.g., fixing self-intersections or improper rings) and returns the 
#' corrected geometries.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, countries_ddbs, "countries")
#'
#' ## make valid
#' ddbs_make_valid("countries", conn)
#'
#' ## make valid without using a connection
#' ddbs_make_valid(countries_ddbs)
#' }
ddbs_make_valid <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_MakeValid",
        other_args = NULL
    )
    
}






#' Simplify geometries
#'
#' Reduces the complexity of geometries by removing unnecessary vertices while preserving 
#' the overall shape.
#'
#' @template x
#' @param tolerance Tolerance distance for simplification. Larger values result in more 
#' simplified geometries.
#' @param preserve_topology If FALSE, uses the Douglas-Peucker algorithm, which reduces
#' the vertices by removing points that are within a given distance. If TRUE, uses a
#' topology-preserving variant of Douglas-Peucker that guarantees the output geometry
#' remains valid (slower).
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, countries_ddbs, "countries")
#'
#' ## simplify with tolerance of 0.01
#' ddbs_simplify("countries", tolerance = 0.01, conn = conn)
#'
#' ## simplify without using a connection
#' ddbs_simplify(countries_ddbs, tolerance = 0.01)
#' }
ddbs_simplify <- function(
    x,
    tolerance = 0,
    preserve_topology = FALSE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Handle function-specific errors
    assert_positive_numeric(tolerance, "tolerance")
    assert_logic(preserve_topology, "preserve_topology")
  
    
    # 1. Build ST_Simplify parameters string
  
    ## 1.1. Choose the right function
    st_simplify_fun <- 
        if (isTRUE(preserve_topology)) "ST_SimplifyPreserveTopology" else "ST_Simplify"
  
    ## 1.2. Build tolerance parameter
    simplify_args <- tolerance
  
    # 2. Pass to template  
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = st_simplify_fun,
        other_args = simplify_args
    )

}






#' Reduce the coordinate precision of geometries
#'
#' Snaps the coordinates of geometries to a regular grid, reducing their
#' precision. Coordinates are rounded to the nearest multiple of \code{precision}
#' (e.g. \code{precision = 0.1} keeps one decimal place, \code{precision = 1}
#' snaps to integers). The CRS is preserved.
#'
#' @template x
#' @param precision Numeric grid size to snap coordinates to. Smaller values keep
#' more precision (e.g. \code{0.001}); larger values snap more aggressively
#' (e.g. \code{1} snaps to integers).
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## snap coordinates to a 0.01 grid
#' ddbs_reduce_precision(countries_ddbs, precision = 0.01)
#' }
ddbs_reduce_precision <- function(
    x,
    precision,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Handle function-specific errors
    assert_positive_numeric(precision, "precision")

    # 1. Pass to template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_ReducePrecision",
        other_args = precision
    )

}






#' Extract the exterior ring of polygons
#'
#' Returns the outer boundary (exterior ring) of polygon geometries. For multi-polygons, 
#' returns the exterior ring of each individual polygon.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, countries_ddbs, "countries")
#'
#' ## extract exterior ring
#' ddbs_exterior_ring(conn = conn, "countries")
#'
#' ## extract exterior ring without using a connection
#' ddbs_exterior_ring(countries_ddbs)
#' }
ddbs_exterior_ring <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    # 0. Handle function-specific error
    assert_geom_type(x = x, conn = conn, geom = "POLYGON", multi = FALSE)
    
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_ExteriorRing",
        other_args = NULL
    )

}









#' Compute the concave hull of geometries
#'
#' Returns the concave hull that tightly encloses the geometry, capturing its overall 
#' shape more closely than a convex hull.
#'
#' @template x
#' @param ratio Numeric. The ratio parameter dictates the level of concavity; `1`
#'        returns the convex hull, while `0` indicates to return the most concave
#'        hull possible. Defaults to `0.5`.
#' @param allow_holes Boolean. If `TRUE` (the default), it allows the output to
#'        contain holes.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(sf)
#'
#' # create points data
#' n <- 5
#' points_ddbs <- data.frame(
#'   id = 1,
#'   x = runif(n, min = -180, max = 180),
#'   y = runif(n, min = -90, max = 90)
#' ) |>
#'   ddbs_as_spatial(coords = c("x", "y")) |>
#'   ddbs_combine()
#'
#' # option 1: passing ddbs or sf objects
#' output1 <- duckspatial::ddbs_concave_hull(points_ddbs, mode = "sf")
#'
#' plot(output1)
#'
#'
#' # option 2: passing the name of a table in a duckdb db
#'
#' # creates a duckdb
#' conn <- duckspatial::ddbs_create_conn()
#'
#' # write sf to duckdb
#' ddbs_write_vector(conn, points_ddbs, "points_tbl")
#'
#' # spatial join
#' output2 <- duckspatial::ddbs_concave_hull(
#'  conn = conn,
#'  x = "points_tbl",
#'  mode = "sf"
#' )
#'
#' plot(output2)
#'
#' }
ddbs_concave_hull <- function(
    x,
    ratio = 0.5,
    allow_holes = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    # 0. Handle function-specific errors
    assert_numeric_interval(ratio, 0, 1, "ratio")
    assert_logic(allow_holes, "allow_holes")

    # 1. Build ST_ConcaveHull parameters string
    concavehull_args <- glue::glue("{ratio}, {toupper(allow_holes)}")    

    # 2. Pass to template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_ConcaveHull",
        other_args = concavehull_args
    )
}





#' Compute the convex hull of geometries
#'
#' Returns the convex hull that encloses the geometry, forming the smallest convex 
#' polygon that contains all points of the geometry.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(sf)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' # read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' # option 1: passing sf objects
#' output1 <- duckspatial::ddbs_convex_hull(x = argentina_ddbs, mode = "sf")
#'
#' plot(output1["CNTR_NAME"])#' # store in duckdb
#'
#' # option 2: passing the name of a table in a duckdb db
#'
#' # creates a duckdb
#' conn <- duckspatial::ddbs_create_conn()
#'
#' # write sf to duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina_tbl")
#'
#' # spatial join
#' output2 <- duckspatial::ddbs_convex_hull(
#'  conn = conn,
#'  x = "argentina_tbl",
#'  mode = "sf"
#' )
#'
#' plot(output2["CNTR_NAME"])
#' }
ddbs_convex_hull <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_ConvexHull",
        other_args = NULL
    )

}





#' Reverse the vertex order of geometries
#'
#' Returns each geometry with the order of its vertices reversed (e.g. a
#' LINESTRING running A -> B -> C becomes C -> B -> A). The geometry type and
#' CRS are preserved.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## reverse vertex order
#' ddbs_reverse(argentina_ddbs)
#' }
ddbs_reverse <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Reverse",
        other_args = NULL
    )

}





#' Normalize geometries to a canonical form
#'
#' Returns each geometry in its normalized (canonical) form, ordering and
#' orienting its components in a deterministic way. This is useful when comparing
#' geometries that are topologically equal but differ in vertex order or
#' component ordering. The geometry type and CRS are preserved.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## normalize geometries
#' ddbs_normalize(argentina_ddbs)
#' }
ddbs_normalize <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Normalize",
        other_args = NULL
    )

}





#' Get the geometry type of features
#'
#' Returns the type of each geometry (e.g., POINT, LINESTRING, POLYGON) in the 
#' input features.
#'
#' @template x
#' @template by_feature
#' @template conn_null
#'
#' @returns A factor with geometry type(s)
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' # option 1: passing sf objects
#' # Get geometry type for each feature
#' ddbs_geometry_type(countries_ddbs)
#' 
#' # Get overall geometry type
#' ddbs_geometry_type(countries_ddbs, by_feature = FALSE)
#' }
ddbs_geometry_type <- function(
  x,
  by_feature = TRUE,
  conn = NULL) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_character(conn, x)
  assert_logic(by_feature, "by_feature")
  

  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = TRUE)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)


  ## 2.3. Create the query
  if (isTRUE(by_feature)) {
    tmp.query <- glue::glue("
          SELECT ST_GeometryType({x_geom}) as geometry
          FROM {x_list$query_name};
      ")
  } else {
    tmp.query <- glue::glue("
          SELECT DISTINCT ST_GeometryType({x_geom}) as geometry
          FROM {x_list$query_name};
      ")
  }
  
  ## 2.4 Return a factor vector
  data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)
  return(data_tbl$geometry)
    
}





#' Computes a Voronoi diagram from point geometries
#'
#' Returns a Voronoi diagram (Thiessen polygons) from a collection of points.
#' Each polygon represents the region closer to one point than to any other
#' point in the set. This function only works with MULTIPOINT geometries.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## create some points, and combine them to MULTIPOINT
#' set.seed(42)
#' n <- 1000
#' points_ddbs <- data.frame(
#'     id = 1:n,
#'     x = runif(n, min = -20, max = 20),
#'     y = runif(n, min = -20, max = 02)
#' ) |>
#'   ddbs_as_spatial(coords = c("x", "y"), crs = 4326) |> 
#'   ddbs_combine()
#' 
#' ## create voronoi diagrama
#' ddbs_voronoi(points_ddbs)
#' }
ddbs_voronoi <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {


    # 0. Handle function-specific errors
    assert_geom_type(x, conn, geom = "MULTIPOINT")

    # 1. Run the template
    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_VoronoiDiagram",
        other_args = NULL
    )

}






#' Flips the X and Y coordinates of geometries
#'
#' Returns a geometry with the X and Y coordinates swapped. This is useful
#' for correcting geometries where longitude and latitude are in the wrong
#' order, or for converting between coordinate systems with different axis
#' orders.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## store in duckdb
#' ddbs_write_vector(conn, argentina_ddbs, "argentina")
#'
#' ## flip coordinates
#' ddbs_flip_coordinates("argentina", conn)
#'
#' ## flip coordinates without using a connection
#' ddbs_flip_coordinates(argentina_ddbs)
#' }
ddbs_flip_coordinates <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_FlipCoordinates",
        other_args = NULL
    )

}





#' Drop geometry column from a duckspatial_df object
#'
#' Removes the geometry column from a \code{duckspatial_df} object, returning a 
#' lazy tibble without spatial information.
#'
#' @template x
#'
#' @return A lazy tibble backed by dbplyr
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## drop geometry column
#' countries_tbl <- ddbs_drop_geometry(countries_ddbs)
#' }
ddbs_drop_geometry <- function(x) {

  ## Get geometry column name
  geometry_col <- attr(x, "sf_column")

  ## Drop duckspatial_df class
  class(x) <- setdiff(class(x), c("duckspatial_df"))

  ## Unselect geometry column
  dplyr::select(x, -dplyr::all_of(geometry_col))
}





#' Convert geometries to multi-type
#'
#' Converts single geometries to their multi-type equivalent (e.g.,
#' \code{POLYGON} to \code{MULTIPOLYGON}). Geometries that are already multi-type are
#' returned unchanged.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## convert to multi-type
#' ddbs_multi(countries_ddbs)
#' }
ddbs_multi <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Multi",
        other_args = NULL
    )

}





#' Computes the maximum inscribed circle of a geometry
#'
#' Returns the largest circle that fits inside the input geometry. The result
#' is derived from a struct containing the circle's center point, the nearest
#' point on the geometry boundary to that center, and the circle's radius.
#'
#' @template x
#' @param geom a character string specifying which component of the inscribed
#' circle to return. Must be one of \code{"center"} (default) or
#' \code{"nearest"}. \code{"center"} returns the center point of the maximum
#' inscribed circle; \code{"nearest"} returns the closest point on the
#' geometry boundary to that center.
#' @param tolerance a numeric value specifying the tolerance used when
#' computing the inscribed circle. If \code{NULL} (default), tolerance is
#' computed automatically as \code{max(width, height) / 1000}.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## return the center point of the maximum inscribed circle
#' ddbs_maximum_inscribed_circle(argentina_ddbs)
#'
#' ## return the nearest boundary point instead
#' ddbs_maximum_inscribed_circle(argentina_ddbs, geom = "nearest")
#'
#' ## use a custom tolerance
#' ddbs_maximum_inscribed_circle(argentina_ddbs, tolerance = 0.01)
#'
#' ## without a connection
#' ddbs_maximum_inscribed_circle(argentina_ddbs)
#' }
ddbs_maximum_inscribed_circle <- function(
  x,
  geom = "center",
  tolerance = NULL,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the base query (depends on the output type - sf, duckspatial_df, table)
  ### Add tolerance parameter if provided
  if (is.null(tolerance)) {
    st_function <- glue::glue("ST_MaximumInscribedCircle({x_geom})")
  } else {
    st_function <- glue::glue("ST_MaximumInscribedCircle({x_geom}, {tolerance})")
  }
  ### This function returns a data frame column. We select the radius, and the one of 
  ### the two geometry columns (nearest or center)
  if (geom == "nearest") {
    nearest_function <- glue::glue("{st_function}.nearest")
    geom_function <- glue::glue("
      {build_geom_query(nearest_function, name, crs_x, mode)} AS {x_geom}
    ")
  } else if (geom == "center") {
    center_function <- glue::glue("{st_function}.center")
    geom_function <- glue::glue("
      {build_geom_query(center_function, name, crs_x, mode)}  AS {x_geom}
    ")
  }
  ### Finally, build the query
  base.query <- glue::glue("
    SELECT * EXCLUDE ({x_geom}),
    {st_function}.radius  AS geom_radius,
    {geom_function}
    FROM {x_list$query_name};
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }

}





#' Remove repeated points from a geometry
#'
#' Removes duplicate consecutive vertices from geometries, optionally within
#' a tolerance distance.
#'
#' @template x
#' @param tolerance A numeric value specifying the minimum distance between
#'   consecutive vertices. Vertices closer than this threshold are considered
#'   repeated and removed. Default is \code{0}, which removes only exact
#'   duplicates.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(sf)
#'
#' ## Create a polygon with repeated points
#' poly <- st_polygon(list(matrix(
#'   c(0, 0,
#'     1, 0,
#'     1, 0,  # repeated point
#'     1, 1,
#'     0, 1,
#'     0, 0),
#'   ncol = 2, byrow = TRUE
#' )))
#' 
#' poly_sf <- st_as_sf(st_sfc(poly))
#'
#' ## remove exact duplicate consecutive vertices
#' ddbs_remove_repeated_points(poly_sf)
#'
#' ## remove vertices within a tolerance of 1 unit
#' ddbs_remove_repeated_points(poly_sf, tolerance = 1)
#' }
ddbs_remove_repeated_points <- function(
  x,
  tolerance = 0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  # 0. Handle function-specific errors
  assert_positive_numeric(tolerance, "tolerance")

  # 2. Pass to template  
  template_unary_ops(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_RemoveRepeatedPoints",
    other_args = tolerance
  )

}





#' Create lines from point geometries
#'
#' Aggregates point geometries into a single LINESTRING by connecting them in
#' their original order. Optionally, lines can be created per group using the
#' \code{by} argument.
#'
#' @template x
#' @param by A character vector of column names to group by before aggregating
#' points into lines. If \code{NULL} (default), all points are combined into a
#' single line.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#' 
#' @details
#' Connects input POINT geometries into a LINESTRING in row order (row 1 → row 2 → …).
#' To control the connection order, sort the data beforehand with [dplyr::arrange()].
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' points_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/points.gpkg", package = "duckspatial")
#' )
#' 
#' ## create a single line from all points
#' ddbs_make_line(points_ddbs)
#' 
#' ## create lines grouped by a column
#' ddbs_make_line(points_ddbs, by = "type")
#' 
#' ## return as sf object
#' ddbs_make_line(points_ddbs, by = "type", mode = "sf")
#' 
#' ## screate lines groupping by 2 columns
#' ddbs_make_line(points_ddbs, by = c("type", "class"))
#' }
ddbs_make_line <- function(
  x,
  by   = NULL,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  
  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  assert_geom_type(x = x, conn = conn, geom = "POINT", multi = FALSE)


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the base query (depends on the output type - sf, duckspatial_df, table)
  st_function <- glue::glue("ST_MakeLine(LIST({x_geom}))")

  ## Add groups if specified
  if (!is.null(by)) {
    grps <- paste0(by, collapse = ", ")
    base.query <- glue::glue("
      SELECT 
        {grps},
        {build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}
      FROM {x_list$query_name}
      GROUP BY {grps};
    ")
  } else {
    base.query <- glue::glue("
      SELECT {build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}
      FROM {x_list$query_name};
    ")
  }
  


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }

}

#' Collect all vertices of a geometry into a MULTIPOINT
#'
#' Returns a MULTIPOINT containing all vertices of the input geometry. This
#' works for any geometry type: points, linestrings, polygons, and their
#' multi- and collection variants.
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## collect vertices
#' ddbs_vertices(argentina_ddbs)
#'
#' ## collect vertices and return as sf
#' ddbs_vertices(argentina_ddbs, mode = "sf")
#' }
ddbs_vertices <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    template_unary_ops(
        x = x,
        conn = conn,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet,
        fun = "ST_Points",
        other_args = NULL
    )

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")


# 1. Geometry validation functions ---------------------------------------

testthat::describe("Geometry validation functions", {

  ## DDBS_IS_EMPTY
  testthat::it("ddbs_is_empty() macro works", {
    normal_result <- ddbs_is_empty(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(is_empty = ddbs_is_empty(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$is_empty, macro_result$is_empty)
  })

  ## DDBS_IS_VALID
  testthat::it("ddbs_is_valid() macro works", {
    normal_result <- ddbs_is_valid(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(is_valid = ddbs_is_valid(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$is_valid, macro_result$is_valid)
  })

  ## DDBS_IS_SIMPLE
  testthat::it("ddbs_is_simple() macro works", {
    normal_result <- ddbs_is_simple(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(is_simple = ddbs_is_simple(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$is_simple, macro_result$is_simple)
  })

  ## DDBS_IS_CLOSED
  testthat::it("ddbs_is_closed() macro works on lines", {
    normal_result <- ddbs_is_closed(rivers_ddbs) |> ddbs_collect()
    macro_result  <- rivers_ddbs |>
      dplyr::mutate(is_closed = ddbs_is_closed(geom)) |>
      ddbs_collect()
    expect_equal(normal_result$is_closed, macro_result$is_closed)
  })

  ## DDBS_IS_RING
  testthat::it("ddbs_is_ring() macro works on lines", {
    normal_result <- ddbs_is_ring(rivers_ddbs) |> ddbs_collect()
    macro_result  <- rivers_ddbs |>
      dplyr::mutate(is_ring = ddbs_is_ring(geom)) |>
      ddbs_collect()
    expect_equal(normal_result$is_ring, macro_result$is_ring)
  })

  ## DDBS_GEOMETRY_TYPE
  testthat::it("ddbs_geometry_type() macro works", {
    normal_result <- ddbs_geometry_type(countries_ddbs)
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry_type = ddbs_geometry_type(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result, macro_result$geometry_type)
  })

})


# 2. Measure functions ---------------------------------------------------

testthat::describe("Measure functions", {

  ## DDBS_AREA
  testthat::it("ddbs_area() macro works on polygons", {
    normal_result <- ddbs_area(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(area = ddbs_area(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$area, macro_result$area)
  })

  ## DDBS_LENGTH
  testthat::it("ddbs_length() macro works on lines", {
    normal_result <- ddbs_length(rivers_ddbs) |> ddbs_collect()
    macro_result  <- rivers_ddbs |>
      dplyr::mutate(length = ddbs_length(geom)) |>
      ddbs_collect()
    expect_equal(normal_result$length, macro_result$length)
  })

  ## DDBS_PERIMETER
  testthat::it("ddbs_perimeter() macro works on polygons", {
    normal_result <- ddbs_perimeter(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(perimeter = ddbs_perimeter(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$perimeter, macro_result$perimeter)
  })

})


# 3. Coordinate operations -----------------------------------------------

testthat::describe("Coordinate operations", {

  ## DDBS_X
  testthat::it("ddbs_x() macro works on points", {
    normal_result <- ddbs_x(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(x = ddbs_x(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$X, macro_result$x)
  })

  ## DDBS_Y
  testthat::it("ddbs_y() macro works on points", {
    normal_result <- ddbs_y(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(y = ddbs_y(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$Y, macro_result$y)
  })

  ## DDBS_Z
  testthat::it("ddbs_z() macro works on points", {
    normal_result <- ddbs_z(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(z = ddbs_z(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$Z, macro_result$z)
  })

  ## DDBS_M
  testthat::it("ddbs_m() macro works on points", {
    normal_result <- ddbs_m(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(m = ddbs_m(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$M, macro_result$m)
  })

})


# 4. Dimension operations ------------------------------------------------

testthat::describe("Dimension operations", {

  ## DDBS_HAS_Z
  testthat::it("ddbs_has_z() macro works", {
    normal_result <- ddbs_has_z(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(has_z = ddbs_has_z(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$has_z, macro_result$has_z)
  })

  ## DDBS_HAS_M
  testthat::it("ddbs_has_m() macro works", {
    normal_result <- ddbs_has_m(points_ddbs) |> ddbs_collect()
    macro_result  <- points_ddbs |>
      dplyr::mutate(has_m = ddbs_has_m(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$has_m, macro_result$has_m)
  })

})


# 5. Format conversion ---------------------------------------------------

testthat::describe("Format conversion functions", {

  ## DDBS_AS_TEXT
  testthat::it("ddbs_as_text() macro works", {
    normal_result <- ddbs_as_text(countries_ddbs)
    macro_result  <- countries_ddbs |>
      dplyr::mutate(wkt = ddbs_as_text(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result, macro_result$wkt)
  })

  ## DDBS_AS_WKB
  testthat::it("ddbs_as_wkb() macro works", {
    normal_result <- ddbs_as_wkb(countries_ddbs)
    macro_result  <- countries_ddbs |>
      dplyr::mutate(wkb = ddbs_as_wkb(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result, macro_result$wkb)
  })

  ## DDBS_AS_HEXWKB
  testthat::it("ddbs_as_hexwkb() macro works", {
    normal_result <- ddbs_as_hexwkb(countries_ddbs)
    macro_result  <- countries_ddbs |>
      dplyr::mutate(hexwkb = ddbs_as_hexwkb(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result, macro_result$hexwkb)
  })

})


# 6. Extent functions ----------------------------------------------------

testthat::describe("Extent functions", {

  ## DDBS_BBOX
  testthat::it("ddbs_bbox() macro works", {
    normal_result <- ddbs_bbox(countries_ddbs, by_feature = TRUE) |> collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(bbox = ddbs_bbox(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$xmin, macro_result$bbox$xmin)
    expect_equal(normal_result$xmax, macro_result$bbox$xmax)
    expect_equal(normal_result$ymin, macro_result$bbox$ymin)
    expect_equal(normal_result$ymax, macro_result$bbox$ymax)
  })

})

# 7. Union ---------------------------------------------------------------

testthat::describe("Extent functions", {

  ## Calculate length
  rivers_length_ddbs <- ddbs_length(rivers_ddbs)

  ## DDBS_BBOX
  testthat::it("ddbs_union() macro works", {

    normal_result <- rivers_length_ddbs |> 
      ddbs_union_agg("RIVER_NAME") |> 
      ddbs_length() |> 
      ddbs_collect(as = "tibble") |> 
      dplyr::arrange(RIVER_NAME)

    macro_result  <- rivers_length_ddbs |>
      dplyr::summarise(
        geom = ddbs_union(geom),
        length = sum(length),
        .by = RIVER_NAME
      )  |> 
      ddbs_collect(as = "tibble") |> 
      dplyr::arrange(RIVER_NAME)

    expect_equal(normal_result$length, macro_result$length)
  })

  ## DDBS_INTERSECTION_AGG
  testthat::it("ddbs_intersection_agg() macro works", {
    sq <- function(xmin, ymin, side) sf::st_polygon(list(matrix(
      c(xmin, ymin, xmin + side, ymin, xmin + side, ymin + side, xmin, ymin + side, xmin, ymin),
      ncol = 2, byrow = TRUE
    )))
    inter_ddbs <- as_duckspatial_df(sf::st_sf(
      grp = c("a", "a", "a"),
      geometry = sf::st_sfc(sq(0, 0, 3), sq(1, 1, 3), sq(2, 2, 3), crs = 4326)
    ))

    normal_result <- ddbs_intersection_agg(inter_ddbs, by = "grp") |>
      ddbs_collect()
    macro_result  <- inter_ddbs |>
      dplyr::summarise(geometry = ddbs_intersection_agg(geometry), .by = grp) |>
      ddbs_collect()

    expect_equal(as.numeric(sf::st_area(normal_result)), as.numeric(sf::st_area(macro_result)))
  })

})


# 8. Geometry processing functions ---------------------------------------

# 7. Geometry processing functions ---------------------------------------

testthat::describe("Geometry processing functions", {

  countries_3857_ddbs <- ddbs_transform(countries_ddbs, "EPSG:3857")

  ## DDBS_BOUNDARY
  testthat::it("ddbs_boundary() macro works", {
    normal_result <- ddbs_boundary(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_boundary(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_BUFFER
  testthat::it("ddbs_buffer() macro works with default parameters", {
    normal_result <- ddbs_buffer(countries_3857_ddbs, distance = 0.1) |> 
      ddbs_collect()
    macro_result  <- countries_3857_ddbs |>
      dplyr::mutate(geometry = ddbs_buffer(geometry, 0.1)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  testthat::it("ddbs_buffer() macro works with custom parameters", {
    normal_result <- ddbs_buffer(
      countries_3857_ddbs,
      distance      = 0.1,
      num_triangles = 16,
      cap_style     = "CAP_FLAT",
      join_style    = "JOIN_MITRE",
      mitre_limit   = 2
    ) |> ddbs_collect()
    macro_result  <- countries_3857_ddbs |>
      dplyr::mutate(geometry = ddbs_buffer(geometry, 0.1, 16, "CAP_FLAT", "JOIN_MITRE", 2)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_CENTROID
  testthat::it("ddbs_centroid() macro works", {
    normal_result <- ddbs_centroid(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_centroid(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_CONCAVE_HULL
  testthat::it("ddbs_concave_hull() macro works with default parameters", {
    normal_result <- ddbs_concave_hull(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_concave_hull(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  testthat::it("ddbs_concave_hull() macro works with custom parameters", {
    normal_result <- ddbs_concave_hull(
      countries_ddbs,
      ratio       = 0.3,
      allow_holes = FALSE
    ) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_concave_hull(geometry, 0.3, FALSE)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_CONVEX_HULL
  testthat::it("ddbs_convex_hull() macro works", {
    normal_result <- ddbs_convex_hull(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_convex_hull(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_EXTERIOR_RING
  testthat::it("ddbs_exterior_ring() macro works on polygons", {
    normal_result <- ddbs_exterior_ring(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_exterior_ring(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_VORONOI
  testthat::it("ddbs_voronoi() macro works on points", {
    points_multi_ddbs <- ddbs_multi(points_ddbs)
    normal_result <- ddbs_voronoi(points_multi_ddbs) |> ddbs_collect()
    macro_result  <- points_multi_ddbs |>
      dplyr::mutate(geometry = ddbs_voronoi(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_BUILD_AREA
  testthat::it("ddbs_build_area() macro works on lines", {
    normal_result <- ddbs_build_area(rivers_ddbs) |> ddbs_collect()
    macro_result  <- rivers_ddbs |>
      dplyr::mutate(geom = ddbs_build_area(geom)) |>
      ddbs_collect()
    expect_equal(normal_result$geom, macro_result$geom)
  })

  ## DDBS_REVERSE
  testthat::it("ddbs_reverse() macro works", {
    normal_result <- ddbs_reverse(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_reverse(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_NORMALIZE
  testthat::it("ddbs_normalize() macro works", {
    normal_result <- ddbs_normalize(countries_ddbs) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_normalize(geometry)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

  ## DDBS_LINE_NODE
  testthat::it("ddbs_line_node() macro works on lines", {
    normal_result <- ddbs_line_node(rivers_ddbs) |> ddbs_collect()
    macro_result  <- rivers_ddbs |>
      dplyr::mutate(geom = ddbs_line_node(geom)) |>
      ddbs_collect()
    expect_equal(normal_result$geom, macro_result$geom)
  })

  ## DDBS_REDUCE_PRECISION
  testthat::it("ddbs_reduce_precision() macro works", {
    normal_result <- ddbs_reduce_precision(countries_ddbs, precision = 0.01) |> ddbs_collect()
    macro_result  <- countries_ddbs |>
      dplyr::mutate(geometry = ddbs_reduce_precision(geometry, 0.01)) |>
      ddbs_collect()
    expect_equal(normal_result$geometry, macro_result$geometry)
  })

})

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write data
duckspatial::ddbs_write_table(conn_test, points_sf, "points")
duckspatial::ddbs_write_table(conn_test, argentina_sf, "argentina")
duckspatial::ddbs_write_table(conn_test, countries_sf, "countries")


# 1. ddbs_buffer() -------------------------------------------------------

describe("ddbs_buffer()", {
  
  ### EXPECTED BEHAVIOUR

  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_buffer(points_ddbs, 50)
      output_sf   <- ddbs_buffer(points_sf, 50)
      output_conn <- ddbs_buffer("points", 50, conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_buffer(points_ddbs, 10, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_buffer(points_ddbs, 10))
      expect_message(ddbs_buffer("points", 50, conn = conn_test, name = "buffer"))
      expect_message(ddbs_buffer("points", 50, conn = conn_test, name = "buffer", overwrite = TRUE))
      expect_true(ddbs_buffer("points", 50, conn = conn_test, name = "buffer2"))

      expect_no_message(ddbs_buffer(points_ddbs, 50, quiet = TRUE))
      expect_no_message(ddbs_buffer("points", 50, conn = conn_test, name = "buffer", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_buffer(points_ddbs, 50)
      output_tbl <- ddbs_read_table(conn_test, "buffer")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
    
    it("matches sf::st_buffer results", {
      point_planar <- sf::st_transform(points_sf[1, ], "EPSG:3857")
      sf_output   <- sf::st_buffer(point_planar, 100, nQuadSegs = 8)
      ddbs_output <- ddbs_buffer(point_planar, 100) |> 
        sf::st_as_sf()

      expect_equal(sf_output$geometry, ddbs_output$geometry)
    })
    
    describe("num_triangles parameter", {
      
      it("works with default value", {
        output <- ddbs_buffer(points_ddbs, 50)
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("works with custom values", {
        output_16 <- ddbs_buffer(points_ddbs, 50, num_triangles = 16)
        output_4 <- ddbs_buffer(points_ddbs, 50, num_triangles = 4)
        
        expect_s3_class(output_16, "duckspatial_df")
        expect_s3_class(output_4, "duckspatial_df")
      })
      
      it("matches sf nQuadSegs parameter", {
        point_planar <- sf::st_transform(points_sf[1, ], "EPSG:3857")
        sf_output <- sf::st_buffer(point_planar, 100, nQuadSegs = 16)
        ddbs_output <- ddbs_buffer(point_planar, 100, num_triangles = 16) |> 
          sf::st_as_sf()
        
        expect_equal(sf_output$geometry, ddbs_output$geometry)
      })
    })
    
    describe("cap_style parameter", {
      
      it("accepts CAP_ROUND (default)", {
        output <- ddbs_buffer(points_ddbs, 50, cap_style = "CAP_ROUND")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("accepts CAP_FLAT on line geometries", {
        output <- ddbs_buffer(rivers_ddbs, 50, cap_style = "CAP_FLAT")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("accepts CAP_SQUARE", {
        output <- ddbs_buffer(points_ddbs, 50, cap_style = "CAP_SQUARE")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("is case-insensitive", {
        output <- ddbs_buffer(points_ddbs, 50, cap_style = "cap_round")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("matches sf endCapStyle parameter", {
        point_planar <- sf::st_transform(points_sf[1, ], "EPSG:3857")
        sf_output <- sf::st_buffer(point_planar, 100, nQuadSegs = 8, endCapStyle = "SQUARE")
        ddbs_output <- ddbs_buffer(point_planar, 100, num_triangles = 8, cap_style = "CAP_SQUARE") |> 
          sf::st_as_sf()
        
        expect_equal(sf_output$geometry, ddbs_output$geometry)
      })
    })
    
    describe("join_style parameter", {
      
      it("accepts JOIN_ROUND (default)", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "JOIN_ROUND")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("accepts JOIN_MITRE", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "JOIN_MITRE")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("accepts JOIN_BEVEL", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "JOIN_BEVEL")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("is case-insensitive", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "join_round")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("matches sf joinStyle parameter", {
        point_planar <- sf::st_transform(points_sf[1, ], "EPSG:3857")
        sf_output <- sf::st_buffer(point_planar, 100, nQuadSegs = 8, joinStyle = "MITRE")
        ddbs_output <- ddbs_buffer(point_planar, 100, num_triangles = 8, join_style = "JOIN_MITRE") |> 
          sf::st_as_sf()
        
        expect_equal(sf_output$geometry, ddbs_output$geometry)
      })
    })
    
    describe("mitre_limit parameter", {
      
      it("works with default value", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "JOIN_MITRE")
        expect_s3_class(output, "duckspatial_df")
      })
      
      it("works with custom values", {
        output <- ddbs_buffer(points_ddbs, 50, join_style = "JOIN_MITRE", mitre_limit = 5.0)
        expect_s3_class(output, "duckspatial_df")
      })
    })
    
    it("works with all parameters combined", {
      output <- ddbs_buffer(
        points_ddbs, 
        distance = 50, 
        num_triangles = 16,
        cap_style = "CAP_SQUARE",
        join_style = "JOIN_MITRE",
        mitre_limit = 2.5
      )
      
      expect_s3_class(output, "duckspatial_df")
    })
    
    it("works when creating tables with custom parameters", {
      expect_true(
        ddbs_buffer(
          "points", 
          50, 
          conn = conn_test, 
          name = "buffer_custom",
          num_triangles = 12,
          cap_style = "CAP_SQUARE"
        )
      )
      
      output_custom_tbl <- ddbs_read_table(conn_test, "buffer_custom")
      expect_s3_class(output_custom_tbl, "sf")
    })
  })

  ### EXPECTED ERRORS
  
  describe("errors", {
    
    describe("basic argument validation", {
      
      it("requires distance argument", {
        expect_error(ddbs_buffer(points_ddbs))
      })
      
      it("requires distance to be numeric", {
        expect_error(ddbs_buffer(points_ddbs, distance = "12"))
      })
      
      it("requires connection when using table names", {
        expect_error(ddbs_buffer("points", conn = NULL))
      })
      
      it("validates x argument type", {
        expect_error(ddbs_buffer(x = 999))
      })
      
      it("validates conn argument type", {
        expect_error(ddbs_buffer(points_ddbs, conn = 999))
      })
      
      it("validates overwrite argument type", {
        expect_error(ddbs_buffer(points_ddbs, overwrite = 999))
      })
      
      it("validates quiet argument type", {
        expect_error(ddbs_buffer(points_ddbs, quiet = 999))
      })
      
      it("validates table name exists", {
        expect_error(ddbs_buffer(x = "999", conn = conn_test))
      })
      
      it("requires name to be single character string", {
        expect_error(ddbs_buffer(points_ddbs, conn = conn_test, name = c('banana', 'banana')))
      })
    })
    
    describe("num_triangles validation", {
      
      it("rejects non-numeric values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, num_triangles = "8"),
          "must be a single integer value"
        )
      })
      
      it("rejects vector inputs", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, num_triangles = c(8, 16)),
          "must be a single integer value"
        )
      })
      
      it("rejects decimal values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, num_triangles = 8.5),
          "must be a single integer value"
        )
      })
      
      it("rejects zero", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, num_triangles = 0),
          "`num_triangles` must be a positive integer"
        )
      })
      
      it("rejects negative values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, num_triangles = -5),
          "`num_triangles` must be a positive integer"
        )
      })
    })
    
    describe("cap_style validation", {
      
      it("rejects non-character values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, cap_style = 123),
          "must be a single character string"
        )
      })
      
      it("rejects vector inputs", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, cap_style = c("CAP_ROUND", "CAP_FLAT")),
          "must be a single character string"
        )
      })
      
      it("rejects invalid cap style names", {
        expect_error(ddbs_buffer(points_ddbs, 50, cap_style = "INVALID_STYLE"))
        expect_error(ddbs_buffer(points_ddbs, 50, cap_style = "ROUND"))
      })
    })
    
    describe("join_style validation", {
      
      it("rejects non-character values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, join_style = 123),
          "must be a single character string"
        )
      })
      
      it("rejects vector inputs", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, join_style = c("JOIN_ROUND", "JOIN_MITRE")),
          "must be a single character string"
        )
      })
      
      it("rejects invalid join style names", {
        expect_error(ddbs_buffer(points_ddbs, 50, join_style = "INVALID_JOIN"))
        expect_error(ddbs_buffer(points_ddbs, 50, join_style = "ROUND"))
      })
    })
    
    describe("mitre_limit validation", {
      
      it("rejects non-numeric values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, mitre_limit = "1.0"),
          "must be a single numeric value"
        )
      })
      
      it("rejects vector inputs", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, mitre_limit = c(1.0, 2.0)),
          "must be a single numeric value"
        )
      })
      
      it("rejects zero", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, mitre_limit = 0),
          "`mitre_limit` must be a positive number"
        )
      })
      
      it("rejects negative values", {
        expect_error(
          ddbs_buffer(points_ddbs, 50, mitre_limit = -1.5),
          "`mitre_limit` must be a positive number"
        )
      })
    })
  })
})


# 2. ddbs_centroid() -----------------------------------------------------


## 2.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
#  - CHECK 1.4: writting table works
## - CHECK 1.5: expected errors
describe("ddbs_centroid()", {

  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_centroid(argentina_ddbs)
      output_sf   <- ddbs_centroid(argentina_sf)
      output_conn <- ddbs_centroid("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_centroid(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_centroid(argentina_ddbs))
      expect_message(ddbs_centroid("argentina", conn = conn_test, name = "centroid"))
      expect_message(ddbs_centroid("argentina", conn = conn_test, name = "centroid", overwrite = TRUE))
      expect_true(ddbs_centroid("argentina", conn = conn_test, name = "centroid2"))

      expect_no_message(ddbs_centroid(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_centroid("argentina", conn = conn_test, name = "centroid", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_centroid(argentina_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "centroid")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
  })
  
  ### EXPECTED ERRORS

  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_centroid("argentina", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_centroid(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_centroid(argentina_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_centroid(argentina_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_centroid(argentina_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_centroid(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_centroid(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})




# 3. ddbs_make_valid() -----------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 2.1: errors

describe("ddbs_make_valid()", {
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_make_valid(argentina_ddbs)
      output_sf   <- ddbs_make_valid(argentina_sf)
      output_conn <- ddbs_make_valid("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_make_valid(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_make_valid(argentina_ddbs))
      expect_message(ddbs_make_valid("argentina", conn = conn_test, name = "make_valid"))
      expect_message(ddbs_make_valid("argentina", conn = conn_test, name = "make_valid", overwrite = TRUE))
      expect_true(ddbs_make_valid("argentina", conn = conn_test, name = "make_valid2"))

      expect_no_message(ddbs_make_valid(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_make_valid("argentina", conn = conn_test, name = "make_valid", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_make_valid(argentina_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "make_valid")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
  })
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_make_valid("argentina", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_make_valid(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_make_valid(argentina_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_make_valid(argentina_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_make_valid(argentina_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_make_valid(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_make_valid(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})





# 4. ddbs_simplify() -----------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 2.1: errors
describe("ddbs_simplify()", {
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_simplify(argentina_ddbs, tolerance = 0.01)
      output_sf   <- ddbs_simplify(argentina_sf, tolerance = 0.01)
      output_conn <- ddbs_simplify("argentina", tolerance = 0.01, conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_simplify(argentina_ddbs, tolerance = 0.01, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_simplify(argentina_ddbs, tolerance = 0.01))
      expect_message(ddbs_simplify("argentina", tolerance = 0.01, conn = conn_test, name = "simplify"))
      expect_message(ddbs_simplify("argentina", tolerance = 0.01, conn = conn_test, name = "simplify", overwrite = TRUE))
      expect_true(ddbs_simplify("argentina", tolerance = 0.01, conn = conn_test, name = "simplify2"))

      expect_no_message(ddbs_simplify(argentina_ddbs, tolerance = 0.01, quiet = TRUE))
      expect_no_message(ddbs_simplify("argentina", tolerance = 0.01, conn = conn_test, name = "simplify", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_simplify(argentina_ddbs, tolerance = 0.01)
      output_tbl <- ddbs_read_table(conn_test, "simplify")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("works with preserve_topology = TRUE", {
      output_ddbs <- ddbs_simplify(argentina_ddbs, preserve_topology = TRUE)
      expect_s3_class(output_ddbs, "duckspatial_df")
    })

  })
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_simplify("argentina", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_simplify(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_simplify(argentina_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_simplify(argentina_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_simplify(argentina_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_simplify(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_simplify(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 5. ddbs_exterior_ring() -----------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 1.5: geometry type
## - CHECK 2.1: errors
describe("ddbs_exterior_ring()", {
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_exterior_ring(argentina_ddbs)
      output_sf   <- ddbs_exterior_ring(argentina_sf)
      output_conn <- ddbs_exterior_ring("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_exterior_ring(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_exterior_ring(argentina_ddbs))
      expect_message(ddbs_exterior_ring("argentina", conn = conn_test, name = "exterior_ring"))
      expect_message(ddbs_exterior_ring("argentina", conn = conn_test, name = "exterior_ring", overwrite = TRUE))
      expect_true(ddbs_exterior_ring("argentina", conn = conn_test, name = "exterior_ring2"))

      expect_no_message(ddbs_exterior_ring(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_exterior_ring("argentina", conn = conn_test, name = "exterior_ring", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_exterior_ring(argentina_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "exterior_ring")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
    
    it("returns LINESTRING geometry type", {
      output_ddbs <- ddbs_exterior_ring(argentina_ddbs)
      geom_type <- ddbs_collect(output_ddbs) |> sf::st_geometry_type() |> as.character()
      
      expect_equal(geom_type, "LINESTRING")
    })
  })
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_exterior_ring("argentina", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_exterior_ring(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_exterior_ring(argentina_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_exterior_ring(argentina_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_exterior_ring(argentina_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_exterior_ring(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_exterior_ring(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})




# 6. ddbs_convex_hull() -----------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 2.1: errors
describe("ddbs_convex_hull()", {
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_convex_hull(argentina_ddbs)
      output_sf   <- ddbs_convex_hull(argentina_sf)
      output_conn <- ddbs_convex_hull("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_convex_hull(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_convex_hull(argentina_ddbs))
      expect_message(ddbs_convex_hull("argentina", conn = conn_test, name = "convex_hull"))
      expect_message(ddbs_convex_hull("argentina", conn = conn_test, name = "convex_hull", overwrite = TRUE))
      expect_true(ddbs_convex_hull("argentina", conn = conn_test, name = "convex_hull2"))

      expect_no_message(ddbs_convex_hull(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_convex_hull("argentina", conn = conn_test, name = "convex_hull", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_convex_hull(argentina_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "convex_hull")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
  })
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_convex_hull("argentina", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_convex_hull(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_convex_hull(argentina_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_convex_hull(argentina_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_convex_hull(argentina_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_convex_hull(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_convex_hull(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 7. ddbs_concave_hull() ---------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 1.5: ratio work
## - CHECK 1.6: allow_holes work
## - CHECK 1.7: same result as sf
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_concave_hull()", {
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_concave_hull(argentina_ddbs)
      output_sf   <- ddbs_concave_hull(argentina_sf)
      output_conn <- ddbs_concave_hull("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_concave_hull(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_concave_hull(argentina_ddbs))
      expect_message(ddbs_concave_hull("argentina", conn = conn_test, name = "concave_hull"))
      expect_message(ddbs_concave_hull("argentina", conn = conn_test, name = "concave_hull", overwrite = TRUE))
      expect_true(ddbs_concave_hull("argentina", conn = conn_test, name = "concave_hull2"))

      expect_no_message(ddbs_concave_hull(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_concave_hull("argentina", conn = conn_test, name = "concave_hull", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output_ddbs <- ddbs_concave_hull(argentina_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "concave_hull")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
    
    describe("ratio parameter", {
      
      it("produces different results with different ratios", {
        output_ratio_1 <- ddbs_concave_hull(argentina_ddbs, ratio = 1, mode = "sf")
        output_ratio_2 <- ddbs_concave_hull(argentina_ddbs, ratio = 0.2, mode = "sf")

        expect_false(identical(output_ratio_1, output_ratio_2))
      })
    })
    
    describe("allow_holes parameter", {
      
      it("produces different results with different allow_holes values", {
        output_holes_1 <- ddbs_concave_hull(argentina_ddbs, allow_holes = TRUE, mode = "sf")
        output_holes_2 <- ddbs_concave_hull(argentina_ddbs, allow_holes = FALSE, mode = "sf")

        expect_false(identical(output_holes_1, output_holes_2))
      })
    })
    
    it("matches sf::st_concave_hull results", {
      geos_version <- package_version(sf::sf_extSoftVersion()["GEOS"])
      skip_if(geos_version < "3.11.0")

      sf_output   <- sf::st_concave_hull(argentina_sf, ratio = 0.5, allow_holes = FALSE)
      ddbs_output <- ddbs_concave_hull(argentina_sf, ratio = 0.5, allow_holes = FALSE, mode = "sf")
      
      expect_equal(sf_output$geometry, ddbs_output$geometry)
    })
  })
  
  describe("errors", {
    
    describe("ratio parameter validation", {
      
      it("rejects values less than 0", {
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = -1))
      })
      
      it("rejects values greater than 1", {
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = 2))
      })
      
      it("rejects vector inputs", {
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = c(0.1, 0.5)))
      })
      
      it("rejects non-numeric values", {
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = "0.5"))
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = TRUE))
      })
      
      it("rejects NULL", {
        expect_error(ddbs_concave_hull(argentina_ddbs, ratio = NULL))
      })
    })
    
    describe("allow_holes parameter validation", {
      
      it("rejects non-logical values", {
        expect_error(ddbs_concave_hull(argentina_ddbs, allow_holes = 3))
        expect_error(ddbs_concave_hull(argentina_ddbs, allow_holes = "TRUE"))
      })
      
      it("rejects NULL", {
        expect_error(ddbs_concave_hull(argentina_ddbs, allow_holes = NULL))
      })
    })
    
    describe("basic argument validation", {
      
      it("requires connection when using table names", {
        expect_error(ddbs_concave_hull("argentina", conn = NULL))
      })
      
      it("validates x argument type", {
        expect_error(ddbs_concave_hull(x = 999))
      })
      
      it("validates conn argument type", {
        expect_error(ddbs_concave_hull(argentina_ddbs, conn = 999))
      })
      
      it("validates overwrite argument type", {
        expect_error(ddbs_concave_hull(argentina_ddbs, overwrite = 999))
      })
      
      it("validates quiet argument type", {
        expect_error(ddbs_concave_hull(argentina_ddbs, quiet = 999))
      })
      
      it("validates table name exists", {
        expect_error(ddbs_concave_hull(x = "999", conn = conn_test))
      })
      
      it("requires name to be single character string", {
        expect_error(ddbs_concave_hull(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
      })
    })
  })
})





# 8. ddbs_geometry_type() ---------------------------------------------------

## - CHECK 1.1: works on all formats with by_feature = TRUE
## - CHECK 1.2: works on all formats with by_feature = FALSE
## - CHECK 1.3: messages work
## - CHECK 1.4: returns the right geometry type
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_geometry_type()", {
  
  describe("expected behavior", {
    
    it("works on all formats with by_feature = TRUE", {
      output_ddbs <- ddbs_geometry_type(points_ddbs, by_feature = TRUE)
      output_sf   <- ddbs_geometry_type(points_sf, by_feature = TRUE)
      output_conn <- ddbs_geometry_type("points", conn = conn_test, by_feature = TRUE)

      expect_s3_class(output_ddbs, "factor")
      expect_equal(output_ddbs, output_sf)
      expect_equal(output_ddbs, output_conn)
    })
    
    it("works on all formats with by_feature = FALSE", {
      output_ddbs <- ddbs_geometry_type(points_ddbs, by_feature = FALSE)
      output_sf   <- ddbs_geometry_type(points_sf, by_feature = FALSE)
      output_conn <- ddbs_geometry_type("points", conn = conn_test, by_feature = FALSE)

      expect_s3_class(output_ddbs, "factor")
      expect_equal(output_ddbs, output_sf)
      expect_equal(output_ddbs, output_conn)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_geometry_type(argentina_ddbs))
    })
    
    describe("geometry type detection", {
      
      it("correctly identifies POINT geometries", {
        output <- ddbs_geometry_type(points_ddbs, by_feature = FALSE) |> as.character()
        expect_equal(output, "POINT")
      })
      
      it("correctly identifies POLYGON geometries", {
        output <- ddbs_geometry_type(argentina_ddbs, by_feature = FALSE) |> as.character()
        expect_equal(output, "POLYGON")
      })
      
      it("correctly identifies LINESTRING geometries", {
        output <- ddbs_geometry_type(rivers_ddbs, by_feature = FALSE) |> as.character()
        expect_equal(output, "LINESTRING")
      })
    })
  })
  
  describe("errors", {
    
    describe("by_feature parameter validation", {
      
      it("rejects non-logical values", {
        expect_error(ddbs_geometry_type(argentina_ddbs, by_feature = 3))
        expect_error(ddbs_geometry_type(argentina_ddbs, by_feature = "TRUE"))
      })
      
      it("rejects NULL", {
        expect_error(ddbs_geometry_type(argentina_ddbs, by_feature = NULL))
      })
    })
    
    describe("basic argument validation", {
      
      it("requires connection when using table names", {
        expect_error(ddbs_geometry_type("argentina", conn = NULL))
      })
      
      it("validates x argument type", {
        expect_error(ddbs_geometry_type(x = 999))
      })
      
      it("validates conn argument type", {
        expect_error(ddbs_geometry_type(argentina_ddbs, conn = 999))
      })
      
      it("validates table name exists", {
        expect_error(ddbs_geometry_type(x = "999", conn = conn_test))
      })
    })
  })
})



# 9. ddbs_flip_coordinates() -----------------------------------------------

describe("ddbs_flip_coordinates()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_flip_coordinates(argentina_ddbs)
      output_sf   <- ddbs_flip_coordinates(argentina_sf)
      output_conn <- ddbs_flip_coordinates("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_flip_coordinates(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_flip_coordinates(argentina_ddbs))
      expect_message(ddbs_flip_coordinates("argentina", conn = conn_test, name = "flip_coords"))
      expect_message(ddbs_flip_coordinates("argentina", conn = conn_test, name = "flip_coords", overwrite = TRUE))
      expect_true(ddbs_flip_coordinates("argentina", conn = conn_test, name = "flip_coords2"))

      expect_no_message(ddbs_flip_coordinates(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_flip_coordinates("argentina", conn = conn_test, name = "flip_coords", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_flip_coordinates(argentina_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "flip_coords")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("double flip returns original geometry", {
      flipped_twice <- ddbs_flip_coordinates(argentina_ddbs) |>
        ddbs_flip_coordinates()

      expect_equal(
        ddbs_collect(flipped_twice)$geometry,
        ddbs_collect(argentina_ddbs)$geometry
      )
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_flip_coordinates("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_flip_coordinates(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_flip_coordinates(argentina_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_flip_coordinates(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_flip_coordinates(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_flip_coordinates(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_flip_coordinates(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})




# 10. ddbs_drop_geometry() --------------------------------------------------

describe("ddbs_drop_geometry()", {

  describe("expected behavior", {

    it("removes the geometry column", {
      output    <- ddbs_drop_geometry(argentina_ddbs)
      col_names <- dplyr::collect(output) |> names()
      expect_false("geometry" %in% col_names)
    })

    it("result is not a duckspatial_df", {
      output <- ddbs_drop_geometry(argentina_ddbs)
      expect_false(inherits(output, "duckspatial_df"))
    })

    it("preserves non-geometry columns", {
      output        <- ddbs_drop_geometry(argentina_ddbs)
      expected_cols <- argentina_sf |> sf::st_drop_geometry() |> names()
      output_cols   <- dplyr::collect(output) |> names()
      expect_true(all(expected_cols %in% output_cols))
    })
  })
})




# 11. ddbs_make_line() -------------------------------------------------------

describe("ddbs_make_line()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_make_line(points_ddbs)
      output_sf   <- ddbs_make_line(points_sf)
      output_conn <- ddbs_make_line("points", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns LINESTRING geometry", {
      output    <- ddbs_make_line(points_ddbs, mode = "sf")
      geom_type <- sf::st_geometry_type(output) |> as.character()
      expect_equal(geom_type, "LINESTRING")
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_make_line(points_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_make_line(points_ddbs))
      expect_message(ddbs_make_line("points", conn = conn_test, name = "make_line"))
      expect_message(ddbs_make_line("points", conn = conn_test, name = "make_line", overwrite = TRUE))
      expect_true(ddbs_make_line("points", conn = conn_test, name = "make_line2"))

      expect_no_message(ddbs_make_line(points_ddbs, quiet = TRUE))
      expect_no_message(ddbs_make_line("points", conn = conn_test, name = "make_line", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_make_line(points_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "make_line")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    describe("by parameter", {

      it("creates one line per group", {
        pts_grp <- dplyr::mutate(points_sf[1:10, ], grp = rep(c("A", "B"), each = 5))
        output  <- ddbs_make_line(pts_grp, by = "grp", mode = "sf")

        expect_equal(nrow(output), 2L)
      })

      it("preserves group column in output", {
        pts_grp <- dplyr::mutate(points_sf[1:10, ], grp = rep(c("A", "B"), each = 5))
        output  <- ddbs_make_line(pts_grp, by = "grp", mode = "sf")

        expect_true("grp" %in% names(output))
      })
    })
  })

  describe("errors", {

    it("requires POINT geometry", {
      expect_error(ddbs_make_line(argentina_ddbs))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_make_line("points", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_make_line(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_make_line(points_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_make_line(points_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_make_line(points_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_make_line(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_make_line(points_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})




# 12. ddbs_maximum_inscribed_circle() ----------------------------------------

describe("ddbs_maximum_inscribed_circle()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_maximum_inscribed_circle(argentina_ddbs)
      output_sf   <- ddbs_maximum_inscribed_circle(argentina_sf)
      output_conn <- ddbs_maximum_inscribed_circle("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns POINT geometry for center", {
      output    <- ddbs_maximum_inscribed_circle(argentina_ddbs, mode = "sf")
      geom_type <- sf::st_geometry_type(output) |> as.character()
      expect_true(all(geom_type == "POINT"))
    })

    it("includes geom_radius column", {
      output <- ddbs_maximum_inscribed_circle(argentina_ddbs, mode = "sf")
      expect_true("geom_radius" %in% names(output))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_maximum_inscribed_circle(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_maximum_inscribed_circle(argentina_ddbs))
      expect_message(ddbs_maximum_inscribed_circle("argentina", conn = conn_test, name = "max_inscribed"))
      expect_message(ddbs_maximum_inscribed_circle("argentina", conn = conn_test, name = "max_inscribed", overwrite = TRUE))
      expect_true(ddbs_maximum_inscribed_circle("argentina", conn = conn_test, name = "max_inscribed2"))

      expect_no_message(ddbs_maximum_inscribed_circle(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_maximum_inscribed_circle("argentina", conn = conn_test, name = "max_inscribed", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_maximum_inscribed_circle(argentina_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "max_inscribed")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    describe("geom parameter", {

      it("accepts 'center' (default)", {
        output <- ddbs_maximum_inscribed_circle(argentina_ddbs, geom = "center")
        expect_s3_class(output, "duckspatial_df")
      })

      it("accepts 'nearest'", {
        output <- ddbs_maximum_inscribed_circle(argentina_ddbs, geom = "nearest")
        expect_s3_class(output, "duckspatial_df")
      })
    })

    describe("tolerance parameter", {

      it("works with custom tolerance", {
        output <- ddbs_maximum_inscribed_circle(argentina_ddbs, tolerance = 0.01)
        expect_s3_class(output, "duckspatial_df")
      })
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_maximum_inscribed_circle("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_maximum_inscribed_circle(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_maximum_inscribed_circle(argentina_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_maximum_inscribed_circle(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_maximum_inscribed_circle(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_maximum_inscribed_circle(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_maximum_inscribed_circle(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})




# 13. ddbs_remove_repeated_points() ------------------------------------------

describe("ddbs_remove_repeated_points()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_remove_repeated_points(argentina_ddbs)
      output_sf   <- ddbs_remove_repeated_points(argentina_sf)
      output_conn <- ddbs_remove_repeated_points("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_remove_repeated_points(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_remove_repeated_points(argentina_ddbs))
      expect_message(ddbs_remove_repeated_points("argentina", conn = conn_test, name = "remove_repeated"))
      expect_message(ddbs_remove_repeated_points("argentina", conn = conn_test, name = "remove_repeated", overwrite = TRUE))
      expect_true(ddbs_remove_repeated_points("argentina", conn = conn_test, name = "remove_repeated2"))

      expect_no_message(ddbs_remove_repeated_points(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_remove_repeated_points("argentina", conn = conn_test, name = "remove_repeated", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_remove_repeated_points(argentina_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "remove_repeated")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("removes repeated consecutive vertices", {
      poly_repeated <- sf::st_as_sf(sf::st_sfc(
        sf::st_polygon(list(matrix(
          c(0, 0,  1, 0,  1, 0,  1, 1,  0, 1,  0, 0),
          ncol = 2, byrow = TRUE
        ))),
        crs = 4326
      ))

      output       <- ddbs_remove_repeated_points(poly_repeated, mode = "sf")
      n_pts_before <- nrow(sf::st_coordinates(poly_repeated))
      n_pts_after  <- nrow(sf::st_coordinates(output))

      expect_lt(n_pts_after, n_pts_before)
    })

    describe("tolerance parameter", {

      it("works with default value (0)", {
        output <- ddbs_remove_repeated_points(argentina_ddbs, tolerance = 0)
        expect_s3_class(output, "duckspatial_df")
      })

      it("works with custom tolerance", {
        output <- ddbs_remove_repeated_points(argentina_ddbs, tolerance = 0.01)
        expect_s3_class(output, "duckspatial_df")
      })
    })
  })

  describe("errors", {

    describe("tolerance parameter validation", {

      it("rejects negative values", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, tolerance = -1))
      })

      it("rejects non-numeric values", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, tolerance = "0.5"))
      })
    })

    describe("basic argument validation", {

      it("requires connection when using table names", {
        expect_error(ddbs_remove_repeated_points("argentina", conn = NULL))
      })

      it("validates x argument type", {
        expect_error(ddbs_remove_repeated_points(x = 999))
      })

      it("validates conn argument type", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, conn = 999))
      })

      it("validates overwrite argument type", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, overwrite = 999))
      })

      it("validates quiet argument type", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, quiet = 999))
      })

      it("validates table name exists", {
        expect_error(ddbs_remove_repeated_points(x = "999", conn = conn_test))
      })

      it("requires name to be single character string", {
        expect_error(ddbs_remove_repeated_points(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
      })
    })
  })
})




# N. ddbs_vertices() -------------------------------------------------------

describe("ddbs_vertices()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_vertices(argentina_ddbs)
      output_sf   <- ddbs_vertices(argentina_sf)
      output_conn <- ddbs_vertices("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns MULTIPOINT geometries", {
      output <- ddbs_vertices(argentina_ddbs, mode = "sf")
      expect_true(all(sf::st_geometry_type(output) == "MULTIPOINT"))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_vertices(argentina_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_vertices(argentina_ddbs))
      expect_message(ddbs_vertices("argentina", conn = conn_test, name = "vertices"))
      expect_message(ddbs_vertices("argentina", conn = conn_test, name = "vertices", overwrite = TRUE))
      expect_true(ddbs_vertices("argentina", conn = conn_test, name = "vertices2"))

      expect_no_message(ddbs_vertices(argentina_ddbs, quiet = TRUE))
      expect_no_message(ddbs_vertices("argentina", conn = conn_test, name = "vertices", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_vertices(argentina_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "vertices")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_vertices("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_vertices(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_vertices(argentina_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_vertices(argentina_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_vertices(argentina_ddbs, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_vertices(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_vertices(argentina_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 15. ddbs_reverse() ---------------------------------------------------------

## A linestring with a known vertex order to check the reversal
line_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(
    sf::st_linestring(matrix(c(0,0, 1,1, 2,2), ncol = 2, byrow = TRUE)),
    crs = 4326
  )
)
line_ddbs <- as_duckspatial_df(line_sf)

describe("ddbs_reverse()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_reverse(argentina_ddbs)
      output_sf   <- ddbs_reverse(argentina_sf)
      output_conn <- ddbs_reverse("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns an sf object with mode = 'sf'", {
      expect_s3_class(ddbs_reverse(argentina_ddbs, mode = "sf"), "sf")
    })

    it("reverses the vertex order", {
      output <- ddbs_reverse(line_ddbs, mode = "sf")
      coords <- sf::st_coordinates(output)[, c("X", "Y")]
      expect_equal(unname(coords), matrix(c(2,2, 1,1, 0,0), ncol = 2, byrow = TRUE))
    })

    it("preserves the CRS", {
      output <- ddbs_reverse(line_ddbs, mode = "sf")
      expect_equal(sf::st_crs(output), sf::st_crs(4326))
    })

    it("is its own inverse (reverse twice = identity)", {
      twice <- ddbs_reverse(ddbs_reverse(line_ddbs), mode = "sf")
      expect_equal(sf::st_coordinates(twice), sf::st_coordinates(line_sf))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_reverse(argentina_ddbs))
      expect_message(ddbs_reverse("argentina", conn = conn_test, name = "reversed"))
      expect_no_message(ddbs_reverse("argentina", conn = conn_test, name = "reversed", overwrite = TRUE, quiet = TRUE))
    })

    it("writes a table when name is provided", {
      expect_true(ddbs_reverse("argentina", conn = conn_test, name = "reversed2"))
      expect_true(DBI::dbExistsTable(conn_test, "reversed2"))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_reverse("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_reverse(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_reverse(argentina_ddbs, conn = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_reverse(x = "999", conn = conn_test))
    })
  })
})


# 16. ddbs_normalize() -------------------------------------------------------

describe("ddbs_normalize()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_normalize(argentina_ddbs)
      output_sf   <- ddbs_normalize(argentina_sf)
      output_conn <- ddbs_normalize("argentina", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns an sf object with mode = 'sf'", {
      expect_s3_class(ddbs_normalize(argentina_ddbs, mode = "sf"), "sf")
    })

    it("preserves geometry type and CRS", {
      output <- ddbs_normalize(argentina_ddbs, mode = "sf")
      expect_equal(sf::st_crs(output), sf::st_crs(argentina_sf))
      expect_setequal(
        as.character(sf::st_geometry_type(output)),
        as.character(sf::st_geometry_type(argentina_sf))
      )
    })

    it("is idempotent (normalize twice = normalize once)", {
      once  <- ddbs_normalize(argentina_ddbs, mode = "sf")
      twice <- ddbs_normalize(ddbs_normalize(argentina_ddbs), mode = "sf")
      expect_equal(sf::st_coordinates(twice), sf::st_coordinates(once))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_normalize(argentina_ddbs))
      expect_message(ddbs_normalize("argentina", conn = conn_test, name = "normalized"))
      expect_no_message(ddbs_normalize("argentina", conn = conn_test, name = "normalized", overwrite = TRUE, quiet = TRUE))
    })

    it("writes a table when name is provided", {
      expect_true(ddbs_normalize("argentina", conn = conn_test, name = "normalized2"))
      expect_true(DBI::dbExistsTable(conn_test, "normalized2"))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_normalize("argentina", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_normalize(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_normalize(argentina_ddbs, conn = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_normalize(x = "999", conn = conn_test))
    })
  })
})


# 17. ddbs_reduce_precision() ------------------------------------------------

## A point with imprecise coordinates to snap to a grid
prec_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(sf::st_point(c(1.23456, 2.98765)), crs = 4326)
)
prec_ddbs <- as_duckspatial_df(prec_sf)
ddbs_write_table(conn_test, prec_sf, "prec")

describe("ddbs_reduce_precision()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_reduce_precision(prec_ddbs, precision = 0.1)
      output_sf   <- ddbs_reduce_precision(prec_sf, precision = 0.1)
      output_conn <- ddbs_reduce_precision("prec", precision = 0.1, conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns an sf object with mode = 'sf'", {
      expect_s3_class(ddbs_reduce_precision(prec_ddbs, precision = 0.1, mode = "sf"), "sf")
    })

    it("snaps coordinates to the grid", {
      out01 <- ddbs_reduce_precision(prec_ddbs, precision = 0.1, mode = "sf")
      expect_equal(unname(sf::st_coordinates(out01)[1, c("X", "Y")]), c(1.2, 3.0))

      out1 <- ddbs_reduce_precision(prec_ddbs, precision = 1, mode = "sf")
      expect_equal(unname(sf::st_coordinates(out1)[1, c("X", "Y")]), c(1, 3))
    })

    it("preserves the CRS", {
      expect_equal(sf::st_crs(ddbs_reduce_precision(prec_ddbs, precision = 0.1, mode = "sf")), sf::st_crs(4326))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_reduce_precision(prec_ddbs, precision = 0.1))
      expect_message(ddbs_reduce_precision("prec", precision = 0.1, conn = conn_test, name = "prec_out"))
      expect_no_message(
        ddbs_reduce_precision("prec", precision = 0.1, conn = conn_test, name = "prec_out", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("writes a table when name is provided", {
      expect_true(ddbs_reduce_precision("prec", precision = 0.1, conn = conn_test, name = "prec_out2"))
      expect_true(DBI::dbExistsTable(conn_test, "prec_out2"))
    })
  })

  describe("errors", {

    it("requires the precision argument", {
      expect_error(ddbs_reduce_precision(prec_ddbs))
    })

    it("validates precision is numeric", {
      expect_error(ddbs_reduce_precision(prec_ddbs, precision = "a"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_reduce_precision("prec", precision = 0.1, conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_reduce_precision(999, precision = 0.1))
    })

    it("validates table name exists", {
      expect_error(ddbs_reduce_precision("999", precision = 0.1, conn = conn_test))
    })
  })
})


## stop connection
ddbs_stop_conn(conn_test)

#' Open spatial dataset lazily via DuckDB
#'
#' Reads spatial data directly from disk using DuckDB's spatial extension or 
#' native Parquet reader, returning a \code{duckspatial_df} object for lazy processing.
#'
#' @param path Path to spatial file. Supports Parquet (`.parquet`, with optional GeoParquet metadata),
#'   GeoJSON, GeoPackage, Shapefile, FlatGeoBuf, OSM PBF, DuckDB database files
#'   (`.duckdb`, `.db`, and `.ddb`), and other GDAL-supported formats.
#' @param crs Coordinate reference system. Can be an EPSG code (e.g., 4326),
#'   a CRS string, or an \code{sf} crs object. If \code{NULL} (default),
#'   attempts to auto-detect from the file, including native DuckDB CRS metadata
#'   and duckspatial-managed column-comment metadata for compatibility
#'   `.duckdb` files.
#' @param layer Layer name or index to read (ST_Read). For DuckDB database
#'   files, this is required and specifies the table name to read. Default is
#'   NULL (first layer for ST_Read).
#' @param geom_col Name of the geometry column. Default is \code{NULL}, which attempts 
#'   auto-detection.
#' @param conn DuckDB connection to use. If NULL, uses the default connection.
#' 
#' @param parquet_binary_as_string Logical. (Parquet) If TRUE, load binary columns as strings.
#' @param parquet_file_row_number Logical. (Parquet) If TRUE, include a `file_row_number` column.
#' @param parquet_filename Logical. (Parquet) If TRUE, include a `filename` column.
#' @param parquet_hive_partitioning Logical. (Parquet) If TRUE, interpret path as Hive partitioned.
#' @param parquet_union_by_name Logical. (Parquet) If TRUE, unify columns by name.
#' @param parquet_encryption_config List/Struct. (Parquet) Encryption configuration (advanced).
#' 
#' @param read_shp_mode Mode for reading Shapefiles. "ST_ReadSHP" (default, fast native reader) or "GDAL" (ST_Read).
#' @param read_osm_mode Mode for reading OSM PBF files. "GDAL" (default, ST_Read) or "ST_ReadOSM" (fast native reader, no geometry).
#' @param shp_encoding Encoding for Shapefiles when using "ST_ReadSHP" (e.g., "UTF-8", "ISO-8859-1").
#' 
#' @param gdal_spatial_filter Optional WKB geometry (as raw vector or hex string) to filter spatially (ST_Read only).
#' @param gdal_spatial_filter_box Optional bounding box (as numeric vector \code{c(minx, miny, maxx, maxy)}) 
#'   (ST_Read only).
#' @param gdal_keep_wkb Logical. If TRUE, return WKB blobs instead of GEOMETRY type (ST_Read only).
#' @param gdal_max_batch_size Integer. Maximum batch size for reading (ST_Read only).
#' @param gdal_sequential_layer_scan Logical. If TRUE, scan layers sequentially (ST_Read only).
#' @param gdal_sibling_files Character vector. List of sibling files (ST_Read only).
#' @param gdal_allowed_drivers Character vector. List of allowed GDAL drivers (ST_Read only).
#' @param gdal_open_options Character vector. Driver-specific open options (ST_Read only).
#'
#' @returns A \code{duckspatial_df} object.
#' 
#' @references 
#' This function is inspired by the dataset opening logic in the 
#' \code{duckdbfs} package (\url{https://github.com/cboettig/duckdbfs}).
#' 
#' @export
ddbs_open_dataset <- function(path, 
                                   # Common
                                   crs = NULL, 
                                   layer = NULL,
                                   geom_col = NULL,
                                   conn = NULL,
                                   
                                   # Parquet Options
                                   parquet_binary_as_string = NULL,
                                   parquet_file_row_number = NULL,
                                   parquet_filename = NULL,
                                   parquet_hive_partitioning = NULL,
                                   parquet_union_by_name = NULL,
                                   parquet_encryption_config = NULL,
                                   
                                   # Dedicated Reader Modes
                                   read_shp_mode = c("ST_ReadSHP", "GDAL"),
                                   read_osm_mode = c("GDAL", "ST_ReadOSM"),
                                   shp_encoding = NULL,
                                   
                                   # GDAL / ST_Read Options
                                   gdal_spatial_filter = NULL,
                                   gdal_spatial_filter_box = NULL,
                                   gdal_keep_wkb = NULL,
                                   gdal_max_batch_size = NULL,
                                   gdal_sequential_layer_scan = NULL,
                                   gdal_sibling_files = NULL,
                                   gdal_allowed_drivers = NULL,
                                   gdal_open_options = NULL) {
  
  ddbs_assert_duckdb_crs_support()

  # Capture the call for error reporting
  fn_call <- rlang::current_call()
  crs_override <- crs
  
  # Get or create connection
  if (is.null(conn)) {
    conn <- ddbs_default_conn()
  }
  
  # Ensure spatial extension
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)
  
  # Determine format
  fmt <- get_file_format(path)
  
  # Resolve modes
  read_shp_mode <- match.arg(read_shp_mode)
  read_osm_mode <- match.arg(read_osm_mode)
  
  # Check for dedicated readers dispatch
  is_dedicated_shp <- (fmt == "shp" && read_shp_mode == "ST_ReadSHP")
  is_dedicated_osm <- (fmt == "osm.pbf" || grepl("\\.osm\\.pbf$", path)) && read_osm_mode == "ST_ReadOSM"
  
  # Support opening DuckDB files natively for explicit DuckDB file extensions.
  # Magic bytes are still the source of truth for whether a candidate is valid.
  duckdb_ext <- has_duckdb_file_extension(path)
  duckdb_info <- duckdb_file_info(path)
  if (duckdb_ext) {
    if (!duckdb_info$exists) {
      cli::cli_abort("File {.file {path}} does not exist.")
    }

    if (!isTRUE(duckdb_info$valid)) {
      if (isTRUE(duckdb_info$empty) || tolower(tools::file_ext(path)) %in% c("duckdb", "ddb")) {
        cli::cli_abort(c(
          "File {.file {path}} is not a valid DuckDB database.",
          "x" = "The file is missing the DuckDB magic bytes."
        ))
      }
      # Non-DuckDB .db files are common in the wild. Let GDAL/ST_Read attempt
      # them instead of treating the extension alone as authoritative.
    } else {
      if (is.null(layer)) {
        cli::cli_abort(c(
          "You must specify the {.arg layer} (table name) when opening a DuckDB database file.",
          "i" = "Example: {.code ddbs_open_dataset('{path}', layer = 'my_table')}"
        ))
      }

      local_conn <- withCallingHandlers(
        ddbs_create_conn(path),
        duckspatial_storage_mismatch = function(w) {
          invokeRestart("muffleWarning")
        }
      )

      result <- tryCatch({
        name_list <- get_query_name(layer)
        if (!table_exists(local_conn, name_list$table_name, name_list$schema_name)) {
          cli::cli_abort("Table {.val {layer}} is not present in DuckDB database {.file {path}}.")
        }
        as_duckspatial_df(layer, conn = local_conn, crs = crs_override, geom_col = geom_col)
      }, error = function(e) {
        ddbs_stop_conn(local_conn)
        stop(e)
      })

      return(result)
    }
  }
    
  # Helper for temporary table construction
  # As for duckdb v1.5 we cannot work with views as the geometry type is not recognized
  # and we need to alter it (column modification is not allowed in views)
  create_temp_table <- function(name, from) {
      glue::glue("
            CREATE OR REPLACE TEMPORARY VIEW {name} AS 
            SELECT * FROM {from};
        ")
  }
  
  view_name <- ddbs_temp_view_name()
  
  # -- QUERY CONSTRUCTION --
  
  # Helper to format args for SQL
  fmt_arg <- function(x, quote = TRUE) {
     if (is.null(x)) return("NULL")
     if (is.character(x) && quote) return(paste0("'", x, "'"))
     if (is.logical(x)) return(ifelse(x, "TRUE", "FALSE"))
     if (is.numeric(x)) return(as.character(x))
     return("NULL") 
  }
  
  fmt_list_arg <- function(x) {
     if (is.null(x)) return("NULL")
     if (length(x) == 0) return("[]")
     items <- paste0("'", x, "'", collapse = ", ")
     paste0("[", items, "]")
  }
  
  # Strategy for "unknown" format (auto-detection):
  # 1. Attempt Native Parquet first (efficient, but extension-less parquet fails ST_Read)
  # 2. Fallback to ST_Read (generic GDAL)
  
  force_parquet <- (fmt == "parquet")
  is_likely_parquet <- force_parquet
  
  if (fmt == "unknown") {
      # Probing: check if it's parquet
      # We check by trying to read the first row with read_parquet
      # We use tryCatch to detect failure (bad magic bytes)
      is_parquet_check <- tryCatch({
          DBI::dbGetQuery(conn, glue::glue("SELECT 1 FROM read_parquet('{path}') LIMIT 1"))
          TRUE
      }, error = function(e) FALSE)
      
      if (is_parquet_check) {
          is_likely_parquet <- TRUE
      }
  }

  if (is_likely_parquet) {
      # WARN on mismatched arguments if user provided GDAL specifics
      if (!is.null(layer) || !is.null(gdal_spatial_filter) || !is.null(gdal_spatial_filter_box) ||
          !is.null(gdal_max_batch_size) || !is.null(gdal_sequential_layer_scan) || 
          !is.null(gdal_sibling_files) || !is.null(gdal_allowed_drivers) || !is.null(gdal_open_options)) {
          cli::cli_warn("Arguments specific to ST_Read (gdal_*) and 'layer' are ignored for Parquet files.")
      }
      
      # Build read_parquet args
      p_args <- c()
      if (!is.null(parquet_binary_as_string)) p_args <- c(p_args, glue::glue("binary_as_string := {fmt_arg(parquet_binary_as_string)}"))
      if (!is.null(parquet_file_row_number)) p_args <- c(p_args, glue::glue("file_row_number := {fmt_arg(parquet_file_row_number)}"))
      if (!is.null(parquet_filename)) p_args <- c(p_args, glue::glue("filename := {fmt_arg(parquet_filename)}"))
      if (!is.null(parquet_hive_partitioning)) p_args <- c(p_args, glue::glue("hive_partitioning := {fmt_arg(parquet_hive_partitioning)}"))
      if (!is.null(parquet_union_by_name)) p_args <- c(p_args, glue::glue("union_by_name := {fmt_arg(parquet_union_by_name)}"))
      
      p_args_str <- ""
      if (length(p_args) > 0) {
          p_args_str <- paste0(", ", paste(p_args, collapse = ", "))
      }
      
      scan_query <- glue::glue("read_parquet('{path}'{p_args_str})")
      
      # Resolve geometry column
      try_cols <- tryCatch({
          DBI::dbGetQuery(conn, glue::glue("DESCRIBE SELECT * FROM {scan_query}"))
      }, error = function(e) NULL)

      if (is.null(geom_col)) {
         if (!is.null(try_cols)) {
             geom_col <- ddbs_describe_geometry_col(try_cols)
         } else {
             geom_col <- NULL
         }
      }

      # Intercept GeoArrow structs (native Arrow encoding) which DuckDB cannot parse
      if (!is.null(geom_col) && !is.null(try_cols)) {
          col_type <- try_cols$column_type[try_cols$column_name == geom_col]
          if (length(col_type) > 0 && grepl("STRUCT", toupper(col_type[1]))) {
              cli::cli_abort(c(
                  "The geometry column {.val {geom_col}} uses a native Arrow/GeoArrow struct encoding that DuckDB's spatial extension cannot parse here.",
                  "x" = "This file uses a GeoParquet/GeoArrow geometry encoding that {.pkg duckspatial} cannot open through DuckDB yet.",
                  "i" = "To work around this, rewrite the geometry column to WKB GeoParquet, for example using:",
                  " " = "  {.code duckspatial::ddbs_write_dataset(data, path)}",
                  " " = "  # OR if using geoarrow:",
                  " " = "  {.code data${geom_col} <- geoarrow::as_geoarrow_vctr(data${geom_col}, schema = geoarrow::geoarrow_wkb())}",
                  " " = "  {.code arrow::write_parquet(data, path)}"
              ))
          }
      }
      
      view_query <- create_temp_table(
        name = view_name,
        from = scan_query
      )
      
  } else if (is_dedicated_shp) {      
      geom_col <- "geom" # Standard for ST_ReadSHP
      # Dedicated ST_ReadSHP path
      if (!is.null(shp_encoding)) {
        view_query <- create_temp_table(
            name = view_name,
            from = glue::glue("ST_ReadSHP('{path}', encoding := '{shp_encoding}')")
        )
      } else {
        view_query <- create_temp_table(
            name = view_name,
            from = glue::glue("ST_ReadSHP('{path}')")
        )
      }

  } else if (is_dedicated_osm) {
       # Dedicated ST_ReadOSM path
        geom_col <- NA_character_ # Signal no geometry
        view_query <- create_temp_table(
            name = view_name,
            from = glue::glue("ST_ReadOSM('{path}')")
        )

  } else {
      # Standard ST_Read (GDAL)
      # WARN on mismatched arguments
      if (!is.null(parquet_binary_as_string) || !is.null(parquet_file_row_number) || 
          !is.null(parquet_filename) || !is.null(parquet_hive_partitioning) || 
          !is.null(parquet_union_by_name) || !is.null(parquet_encryption_config)) {
          cli::cli_warn("Arguments specific to Parquet (parquet_*) are ignored for this format.")
      }
      
      if (!is.null(shp_encoding)) {
          cli::cli_warn("Argument `shp_encoding` is ignored when `read_shp_mode` is not 'ST_ReadSHP'.")
      }

      # ST_Read ARGS
      sp_filter_sql <- "NULL" 
      if (!is.null(gdal_spatial_filter)) {
          if (is.raw(gdal_spatial_filter)) {
              hex <- paste0(as.character(gdal_spatial_filter), collapse = "")
              sp_filter_sql <- paste0("x'", hex, "'")
          } else if (is.character(gdal_spatial_filter)) {
               sp_filter_sql <- paste0("x'", gdal_spatial_filter, "'")
          }
      }
      
      sp_box_sql <- "NULL"
      if (!is.null(gdal_spatial_filter_box)) {
          if (is.numeric(gdal_spatial_filter_box) && length(gdal_spatial_filter_box) == 4) {
              sp_box_sql <- glue::glue("ROW({gdal_spatial_filter_box[1]}, {gdal_spatial_filter_box[2]}, {gdal_spatial_filter_box[3]}, {gdal_spatial_filter_box[4]})::BOX_2D")
          }
      }
      
      args_list <- c()
      if (!is.null(gdal_keep_wkb)) args_list <- c(args_list, glue::glue("keep_wkb := {fmt_arg(gdal_keep_wkb)}"))
      if (!is.null(gdal_max_batch_size)) args_list <- c(args_list, glue::glue("max_batch_size := {fmt_arg(gdal_max_batch_size)}"))
      if (!is.null(gdal_sequential_layer_scan)) args_list <- c(args_list, glue::glue("sequential_layer_scan := {fmt_arg(gdal_sequential_layer_scan)}"))
      if (!is.null(layer)) args_list <- c(args_list, glue::glue("layer := {fmt_arg(layer)}"))
      if (!is.null(gdal_sibling_files)) args_list <- c(args_list, glue::glue("sibling_files := {fmt_list_arg(gdal_sibling_files)}"))
      
      if (sp_filter_sql != "NULL") args_list <- c(args_list, glue::glue("spatial_filter := {sp_filter_sql}"))
      if (sp_box_sql != "NULL") args_list <- c(args_list, glue::glue("spatial_filter_box := {sp_box_sql}"))
      
      if (!is.null(gdal_allowed_drivers)) args_list <- c(args_list, glue::glue("allowed_drivers := {fmt_list_arg(gdal_allowed_drivers)}"))
      if (!is.null(gdal_open_options)) args_list <- c(args_list, glue::glue("open_options := {fmt_list_arg(gdal_open_options)}"))
      
      args_str <- ""
      if (length(args_list) > 0) {
          args_str <- paste0(", ", paste(args_list, collapse = ", "))
      }
      
      query_str <- glue::glue("ST_Read('{path}'{args_str})")
      
      if (is.null(geom_col)) {
         try_cols <- tryCatch({
             DBI::dbGetQuery(conn, glue::glue("DESCRIBE SELECT * FROM {query_str}"))
         }, error = function(e) {
             # Catch ST_Read errors gracefully and provide user-friendly message
             msg <- e$message
             
             # Check if it's a format recognition error
             if (grepl("not recognized as a supported file format", msg, ignore.case = TRUE) || 
                 grepl("No extension found", msg, ignore.case = TRUE) ||
                 grepl("Could not open", msg, ignore.case = TRUE)) {
                 
                 # Extract the GDAL error message if present
                 gdal_match <- regexpr("GDAL Error \\([0-9]+\\): .*", msg)
                 gdal_msg <- if (gdal_match > 0) {
                     regmatches(msg, gdal_match)
                 } else {
                     "File format not recognized"
                 }
                 
                 # Clean, user-friendly error with technical details preserved
                 cli::cli_abort(c(
                     "Unable to open file {.path {basename(path)}}",
                     "x" = "The file format could not be detected or is not supported.",
                     "i" = "Supported formats include: Parquet, GeoJSON, GeoPackage, Shapefile, FlatGeoBuf, and other GDAL formats.",
                     "i" = "GDAL error: {gdal_msg}"
                 ), call = fn_call)
             }
             
             # For other errors, just return NULL to continue
             NULL
         })
         
         if (!is.null(try_cols)) {
             geom_col <- ddbs_describe_geometry_col(try_cols)
         }
      }
      
      view_query <- create_temp_table(
        name = view_name,
        from = query_str
      )

  }

  # -- EXECUTE VIEW CREATION --
  tryCatch({
    # We use dbExecute here to ensure the view is created successfully.
    # This is where most file-opening and format errors will surface.
    DBI::dbExecute(conn, view_query)
  }, error = function(e) {
    msg <- e$message
    
    # 1. Format recognition error (DuckDB or GDAL)
    if (grepl("not recognized as a supported file format|No extension found", msg, ignore.case = TRUE)) {
        # Extract the GDAL error message if present
        gdal_match <- regexpr("GDAL Error \\([0-9]+\\): .*", msg)
        gdal_msg <- if (gdal_match > 0) regmatches(msg, gdal_match) else "File format not recognized"
        
        cli::cli_abort(c(
            "Unable to open file {.path {basename(path)}}",
            "x" = "The file format could not be detected or is not supported.",
            "i" = "Supported formats include: Parquet, GeoJSON, GeoPackage, Shapefile, FlatGeoBuf, and other GDAL formats.",
            "i" = "GDAL error: {gdal_msg}"
        ), call = fn_call)
    }
    
    # 2. File not found error (IO Error)
    if (grepl("Cannot open file|No such file|IO Error.*No such file", msg, ignore.case = TRUE)) {
        cli::cli_abort(c(
            "Unable to open file {.path {basename(path)}}",
            "x" = "The file does not exist or cannot be accessed.",
            "i" = "Please check if the file path is correct and the file is not moved or deleted."
        ), call = fn_call)
    }
    
    # 3. Generic error fallback
    cli::cli_abort(c(
        "Failed to open dataset at {.path {path}}",
        "x" = msg
    ), call = fn_call)
  })
    
    # Get lazy table reference
    duck_tbl <- dplyr::tbl(conn, view_name)
    
    view_cols <- tryCatch(
      DBI::dbGetQuery(conn, glue::glue("DESCRIBE {view_name}")),
      error = function(e) NULL
    )
    
    # Resolve geometry column: uses user-supplied name if valid, 
    # otherwise falls back to auto-detection heuristic.
    geom_col <- ddbs_describe_geometry_col(view_cols, geom_col)

    # Return already if there's no geometry
    if (is.null(geom_col)) return(duck_tbl)

    crs <- ddbs_open_dataset_crs(
      crs = crs_override,
      conn = conn,
      view_name = view_name,
      geom_col = geom_col,
      path = path,
      fmt = if (is_likely_parquet) "parquet" else fmt
    )
    
    # Return duckspatial if there's geometry col
    result <- new_duckspatial_df(
      duck_tbl, 
      crs = crs, 
      geom_col = geom_col, 
      source_table = view_name,
      source_conn = conn
    )    

    return(result)
}

#' Detect a geometry column from DuckDB DESCRIBE output
#'
#' @keywords internal
#' @noRd
ddbs_describe_geometry_col <- function(desc, geom_col = NULL) {
  if (is.null(desc) || nrow(desc) == 0) {
    return(NULL)
  }
  
  col_type <- if ("column_type" %in% names(desc)) desc$column_type else desc$data_type
  
  is_geometry_type <- grepl("^GEOMETRY(\\(|$)", col_type, ignore.case = TRUE)
  is_wkb_type <- grepl("WKB_BLOB", col_type, ignore.case = TRUE)
  is_blob_type <- grepl("^BLOB$", col_type, ignore.case = TRUE)
  is_struct_type <- grepl("^STRUCT", col_type, ignore.case = TRUE)
  is_spatial_type <- is_geometry_type | is_wkb_type | is_blob_type | is_struct_type

  # 1. If user supplied a geom_col, validate it
  if (!is.null(geom_col) && !is.na(geom_col)) {
    match_idx <- which(desc$column_name == geom_col)
    if (length(match_idx) > 0 && is_spatial_type[match_idx[1]]) {
      return(geom_col)
    }
  }

  # 2. Auto-detection heuristics
  
  # 2.1. First priority: Native GEOMETRY types
  geom_cols <- desc$column_name[is_geometry_type]
  if (length(geom_cols) > 0) {
    return(geom_cols[1])
  }

  # 2.2. Second priority: Spatial column names with compatible types (WKB/BLOB/STRUCT)
  known <- c("geom", "geometry", "wkb_geometry")
  found <- desc$column_name[desc$column_name %in% known & is_spatial_type]
  if (length(found) > 0) {
    return(found[1])
  }

  # 2.3. Third priority: Known spatial formats fallback (WKB_BLOB)
  # Do not treat arbitrary BLOB columns as geometry unless they were 
  # explicitly selected by name above or exposed as native GEOMETRY.
  wkb_cols <- desc$column_name[is_wkb_type]
  if (length(wkb_cols) > 0) {
    return(wkb_cols[1])
  }

  NULL
}

#' Resolve CRS for an opened DuckDB-backed dataset
#'
#' @keywords internal
#' @noRd
ddbs_open_dataset_crs <- function(crs, conn, view_name, geom_col, path, fmt) {
  if (!is.null(crs)) {
    return(if (inherits(crs, "crs")) crs else sf::st_crs(crs))
  }

  crs_from_duckdb <- tryCatch({
    q_geom <- DBI::dbQuoteIdentifier(conn, geom_col)
    res <- DBI::dbGetQuery(conn, glue::glue(
      "SELECT ST_CRS({q_geom}) AS crs ",
      "FROM {view_name} ",
      "WHERE {q_geom} IS NOT NULL ",
      "LIMIT 1"
    ))

    if (nrow(res) == 0 || is.na(res$crs[1]) || identical(res$crs[1], "")) {
      NULL
    } else {
      sf::st_crs(res$crs[1])
    }
  }, error = function(e) NULL)

  if (!is.null(crs_from_duckdb) && !is.na(crs_from_duckdb)) {
    return(crs_from_duckdb)
  }

  if (!identical(fmt, "parquet")) {
    crs_from_gdal_meta <- withCallingHandlers(
      tryCatch(get_file_crs(path, conn), error = function(e) NULL),
      warning = function(w) {
        if (grepl("Cannot open|No such file|not recognized|missing value where", w$message, ignore.case = TRUE)) {
          invokeRestart("muffleWarning")
        }
      }
    )

    if (!is.null(crs_from_gdal_meta) && !is.na(crs_from_gdal_meta)) {
      return(crs_from_gdal_meta)
    }
  }

  if (identical(fmt, "parquet")) {
    crs_from_geoparquet <- suppressWarnings(
      tryCatch(get_parquet_crs(path, conn), error = function(e) NULL)
    )

    if (!is.null(crs_from_geoparquet) && !is.na(crs_from_geoparquet)) {
      return(crs_from_geoparquet)
    }
  }

  warning("CRS could not be auto-detected; returning NA CRS.", call. = FALSE)
  sf::st_crs(NA)
}

# skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")


# helpers --------------------------------------------------------------

# create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

# helper function
tester <- function(x = points_sf,
                   y = argentina_sf,
                   predicate = "intersects",
                   conn = NULL,
                   name = NULL,
                   distance = NULL,
                   mode = "sf",
                   overwrite = FALSE,
                   quiet = FALSE) {
    ddbs_filter(
        x = x,
        y = y,
        predicate = predicate,
        conn = conn,
        name = name,
        distance = distance,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet
    )
}


# expected behavior --------------------------------------------------------------


testthat::test_that("expected behavior", {

    # option 1: passing sf objects
    output1 <- tester(
        x = points_sf,
        y = argentina_sf,
        predicate = "intersects"
    )

    testthat::expect_true(is(output1 , 'sf'))

    # option 2: passing the names of tables in a duckdb db, returing sf
    # write sf to duckdb
    ddbs_write_table(conn_test, points_sf, "points", overwrite = TRUE)
    ddbs_write_table(conn_test, argentina_sf, "argentina", overwrite = TRUE)

    # spatial filter
    output2 <- tester(
        conn = conn_test,
        x = "points",
        y = "argentina",
        predicate = "intersects"
    )

    testthat::expect_true(is(output2 , 'sf'))

    # option 3: passing the names of tables in a duckdb db, creating new table in db
    output3 <- tester(
        conn = conn_test,
        x = "points",
        y = "argentina",
        predicate = "intersects",
        name = "filter_result",
        overwrite = TRUE
    )

    testthat::expect_true(output3)

    ddbs_read_table(conn = conn_test, name = "filter_result")


    # show and suppress messages
    testthat::expect_no_message( tester() )


})


testthat::test_that("error if table already exists", {

    # write table for the 1st time
    testthat::expect_true(tester(x = "points",
                                    y = "argentina",
                                    conn = conn_test,
                                    name = 'banana_filter',
                                    overwrite = FALSE)
                             )

    # expected error if overwrite = FALSE
    testthat::expect_error(tester(x = "points",
                                    y = "argentina",
                                    conn = conn_test,
                                    name = 'banana_filter',
                                    overwrite = FALSE))

    # overwrite table
    testthat::expect_true(tester(x = "points",
                                    y = "argentina",
                                    conn = conn_test,
                                    name = 'banana_filter',
                                    overwrite = TRUE))


})

# expected errors --------------------------------------------------------------

testthat::test_that("errors with incorrect input", {

    testthat::expect_error(tester(x = 999))
    testthat::expect_error(tester(y = 999))
    testthat::expect_error(tester(predicate = 999))
    testthat::expect_error(tester(conn = 999))
    testthat::expect_error(tester(overwrite = 999))
    testthat::expect_error(tester(quiet = 999))

    testthat::expect_error(tester(x = "999", conn = conn_test))
    testthat::expect_error(tester(y = "999", conn = conn_test))

    testthat::expect_error(tester(conn = conn_test, name = c('banana', 'banana')))


    })



# duckspatial_df inputs --------------------------------------------------------

testthat::test_that("ddbs_filter works with duckspatial_df inputs", {
  argentina_path <- system.file("spatial/argentina.geojson", package = "duckspatial")
  
  # Create a distinct connection for this test to avoid interference
  conn <- ddbs_temp_conn()
  
  # Load as duckspatial_df
  argentina_ds <- ddbs_open_dataset(argentina_path, conn = conn)
  
  # Create points as sf
  points_small_sf <- sf::st_as_sf(
    data.frame(id = 1:5, x = c(-60, -60, 0, 0, 0), y = c(-34, -34, 0, 0, 0)), 
    coords = c("x", "y"), 
    crs = 4326
  )
  
  # Register points to the same connection as a duckspatial_df
  ddbs_write_table(conn, points_small_sf, "test_points_filter")
  points_ds <- ddbs_read_table(conn, "test_points_filter")
  
  # 1. duckspatial_df x duckspatial_df
  result1 <- ddbs_filter(points_ds, argentina_ds, predicate = "intersects")
  expect_s3_class(result1, "duckspatial_df")
  expect_true(nrow(dplyr::collect(result1)) >= 0)
  
  # 2. duckspatial_df x sf
  result2 <- ddbs_filter(points_ds, argentina_sf, predicate = "intersects")
  expect_s3_class(result2, "duckspatial_df")
  
  # 3. sf x duckspatial_df
  withr::with_options(list(duckspatial.mode = "sf"), {
      r_sf <- ddbs_filter(points_small_sf, argentina_ds, predicate = "intersects")
      expect_s3_class(r_sf, "sf")
  })
})

# predicates -------------------------------------------------------------------

testthat::test_that("ddbs_filter works with different predicates", {
  # Polygon: square (0,0) to (10,10)
  p1 <- sf::st_polygon(list(matrix(c(0,0, 10,0, 10,10, 0,10, 0,0), ncol=2, byrow=TRUE)))
  poly_sf <- sf::st_sf(id = 1, geometry = sf::st_sfc(p1), crs=4326)
  
  # Points: inside (5,5), edge (0,0), outside (20,20)
  pts <- matrix(c(5,5, 0,0, 20,20), ncol=2, byrow=TRUE)
  pts_sf <- sf::st_sf(id = 1:3, geometry = sf::st_sfc(lapply(1:3, function(i) sf::st_point(pts[i,]))), crs=4326)
  
  res_within <- ddbs_filter(pts_sf, poly_sf, predicate = "within") |> dplyr::collect()
  expect_true(1 %in% res_within$id)
  expect_false(3 %in% res_within$id)
  
  res_disjoint <- ddbs_filter(pts_sf, poly_sf, predicate = "disjoint") |> dplyr::collect()
  expect_true(3 %in% res_disjoint$id)
  expect_false(1 %in% res_disjoint$id)
  
  # ST_DWithin
  res_dwithin <- ddbs_filter(pts_sf, pts_sf[1, ], predicate = "dwithin", distance = 15000000) |> dplyr::collect()
  expect_true(3 %in% res_dwithin$id)

})

# output parameters ------------------------------------------------------------

testthat::test_that("ddbs_filter respects mode parameter", {
  result_sf <- ddbs_filter(points_sf, argentina_sf, mode = "sf")
  expect_s3_class(result_sf, "sf")
  
  result_ds <- ddbs_filter(points_sf, argentina_sf, mode = "duckspatial")
  expect_s3_class(result_ds, "duckspatial_df")
})

# error handling ---------------------------------------------------------------

testthat::test_that("ddbs_filter throws error on CRS mismatch", {
  points_3857 <- sf::st_transform(points_sf[1:10,], 3857)
  
  expect_error(
    ddbs_filter(points_3857, argentina_sf),
    "Coordinates Reference System"
  )
})

testthat::test_that("dwithin fails in non-point geometries", {
  expect_error(ddbs_filter(pts_sf, poly_sf, predicate = "dwithin", distance = 100))
})

# Comprehensive validation tests for ddbs_write_dataset

testthat::skip_on_cran()

test_that("warns when extension and driver mismatch", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  tmp_file <- tempfile(fileext = ".geojson")
  on.exit(unlink(tmp_file), add = TRUE)
  
  # .geojson with ESRI Shapefile driver should warn
  expect_warning(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "ESRI Shapefile", quiet = TRUE),
    "Extension/driver mismatch"
  )
  
  # Should still create the file
  expect_true(file.exists(tmp_file))
})

test_that("warns with specific extension and driver details in mismatch", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  tmp_file <- tempfile(fileext = ".shp")
  on.exit(unlink(tmp_file), add = TRUE)
  
  # .shp maps to "ESRI Shapefile", but we specify "GeoJSON"
  expect_warning(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "GeoJSON", quiet = TRUE),
    "File extension.*\\.shp.*ESRI Shapefile"
  )
  
  # Verify mentions GeoJSON and .geojson
  tmp_file2 <- tempfile(fileext = ".shp")
  on.exit(unlink(tmp_file2), add = TRUE)
  
  expect_warning(
    ddbs_write_dataset(ds, tmp_file2, gdal_driver = "GeoJSON", quiet = TRUE),
    "You specified driver.*GeoJSON.*\\.geojson"
  )
})

test_that("errors on unknown extension without gdal_driver", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  expect_error(
    ddbs_write_dataset(ds, "output.xyz"),
    "Cannot determine GDAL driver"
  )
  
  # Error should suggest using gdal_driver
  expect_error(
    ddbs_write_dataset(ds, "output.xyz"),
    "gdal_driver"
  )
  
  # Error should suggest ddbs_drivers()
  expect_error(
    ddbs_write_dataset(ds, "output.xyz"),
    "ddbs_drivers"
  )
})

test_that("works with unknown extension when gdal_driver provided", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  tmp_file <- tempfile(fileext = ".xyz")
  on.exit(unlink(tmp_file), add = TRUE)
  
  expect_no_error(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "GeoJSON", quiet = TRUE)
  )
  expect_true(file.exists(tmp_file))
  
  # Verify it's actually GeoJSON by reading it back
  ds_back <- ddbs_open_dataset(tmp_file, conn = conn)
  expect_equal(dplyr::count(ds_back) |> dplyr::pull(n), 257)
})

test_that("errors with helpful message for invalid driver", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  tmp_file <- tempfile(fileext = ".geojson")
  
  expect_error(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "NonExistentDriver", quiet = TRUE),
    "is not available on this system"
  )
  
  # Should suggest available drivers
  expect_error(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "NonExistentDriver", quiet = TRUE),
    "Available writable drivers"
  )
  
  # Should mention ddbs_drivers()
  expect_error(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "NonExistentDriver", quiet = TRUE),
    "ddbs_drivers"
  )
})

test_that("auto-detects from standard extensions", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  # Test various extensions
  ## fgb: it's not deleted on exit, raises a CRAN note of detritus
  extensions <- c("geojson", "shp")
  
  for (ext in extensions) {
    tmp_file <- tempfile(fileext = paste0(".", ext))
    on.exit(unlink(tmp_file), add = TRUE)
    
    # Should work without specifying gdal_driver
    expect_no_error(
      ddbs_write_dataset(ds, tmp_file, quiet = TRUE)
    )
    expect_true(file.exists(tmp_file))
  }
})

test_that("handles native formats without GDAL validation", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  # Parquet
  tmp_parquet <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_parquet), add = TRUE)
  expect_no_error(ddbs_write_dataset(ds, tmp_parquet, quiet = TRUE))
  expect_true(file.exists(tmp_parquet))
  
  # CSV
  tmp_csv <- tempfile(fileext = ".csv")
  on.exit(unlink(tmp_csv), add = TRUE)
  expect_no_error(ddbs_write_dataset(ds, tmp_csv, quiet = TRUE))
  expect_true(file.exists(tmp_csv))
})

test_that("no warning when extension and driver match", {
  conn <- ddbs_temp_conn()
  ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"), conn = conn)
  
  tmp_file <- tempfile(fileext = ".geojson")
  on.exit(unlink(tmp_file), add = TRUE)
  
  # .geojson with GeoJSON driver should not warn
  expect_no_warning(
    ddbs_write_dataset(ds, tmp_file, gdal_driver = "GeoJSON", quiet = TRUE)
  )
  expect_true(file.exists(tmp_file))
})

# Tests for ddbs_open_dataset following the ssf_read testing pattern
# Uses internal package datasets (countries, argentina, rivers) from setup.R

# =============================================================================
# Format-specific tests
# =============================================================================
testthat::skip_on_cran()

expect_duckdb_st_crs <- function(ds, conn) {
  view_name <- attr(ds, "source_table")
  geom_col <- attr(ds, "sf_column")
  q_geom <- as.character(DBI::dbQuoteIdentifier(conn, geom_col))

  res <- DBI::dbGetQuery(
    conn,
    glue::glue(
      "SELECT ST_CRS({q_geom}) AS crs ",
      "FROM {view_name} ",
      "WHERE {q_geom} IS NOT NULL ",
      "LIMIT 1"
    )
  )

  expect_equal(nrow(res), 1)
  expect_false(is.na(res$crs[[1]]))
  expect_false(identical(res$crs[[1]], ""))
  res$crs[[1]]
}

test_that("ddbs_open_dataset works with GeoJSON", {
  conn <- ddbs_temp_conn()

  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(countries_path, conn = conn)

  expect_s3_class(ds, "duckspatial_df")

  # Verify row count via SQL (countries.geojson has 257 countries)
  view_name <- attr(ds, "source_table")
  res <- DBI::dbGetQuery(conn, sprintf("SELECT count(*) FROM %s", view_name))
  expect_equal(as.numeric(res[[1]]), 257)

  # Verify CRS detection delegates to DuckDB when available
  duckdb_crs <- expect_duckdb_st_crs(ds, conn)
  expect_equal(attr(ds, "crs"), sf::st_crs(duckdb_crs))

  # Verify geometry is valid
  res_geom <- DBI::dbGetQuery(
    conn,
    sprintf("SELECT ST_AsText(geom) FROM %s LIMIT 1", view_name)
  )
  expect_true(nrow(res_geom) == 1)
  expect_true(grepl("POLYGON", res_geom[[1]]))
})

test_that("ddbs_open_dataset works with GeoPackage", {
  conn <- ddbs_temp_conn()

  # Create temp GeoPackage
  tmp_gpkg <- ddbs_create_temp_spatial_file(
    countries_sf,
    ext = "gpkg",
    conn = conn
  )

  ds <- ddbs_open_dataset(tmp_gpkg, conn = conn)

  expect_s3_class(ds, "duckspatial_df")

  # Verify row count
  view_name <- attr(ds, "source_table")
  res <- DBI::dbGetQuery(conn, sprintf("SELECT count(*) FROM %s", view_name))
  expect_equal(as.numeric(res[[1]]), nrow(countries_sf))

  # Verify CRS detection delegates to DuckDB when available
  duckdb_crs <- expect_duckdb_st_crs(ds, conn)
  expect_equal(attr(ds, "crs"), sf::st_crs(duckdb_crs))
})

test_that("ddbs_open_dataset works with Parquet (GeoArrow)", {
  skip_if_not_installed("arrow")

  conn <- ddbs_temp_conn()

  # Create temp Parquet from internal data using helper
  tmp_parquet <- ddbs_create_temp_spatial_file(countries_sf, ext = "parquet", conn = conn)

  ds <- ddbs_open_dataset(tmp_parquet, conn = conn)

  expect_s3_class(ds, "duckspatial_df")

  # Verify row count
  view_name <- attr(ds, "source_table")
  res <- DBI::dbGetQuery(conn, sprintf("SELECT count(*) FROM %s", view_name))
  expect_equal(as.numeric(res[[1]]), nrow(countries_sf))

  # Check geometry column detection
  expect_equal(attr(ds, "sf_column"), "geometry")

  # Verify CRS detection delegates to DuckDB when available
  duckdb_crs <- expect_duckdb_st_crs(ds, conn)
  expect_equal(attr(ds, "crs"), sf::st_crs(duckdb_crs))
})

test_that("ddbs_open_dataset works with Shapefile", {
  conn <- ddbs_temp_conn()

  # Create temp shapefile from internal data using helper
  tmp_shp <- ddbs_create_temp_spatial_file(rivers_sf, ext = "shp", conn = conn)
  expect_true(file.exists(tmp_shp))

  ds <- ddbs_open_dataset(tmp_shp, conn = conn)

  expect_s3_class(ds, "duckspatial_df")

  # Verify row count (rivers_sf has 100 features)
  view_name <- attr(ds, "source_table")
  res <- DBI::dbGetQuery(conn, sprintf("SELECT count(*) FROM %s", view_name))
  expect_equal(as.numeric(res[[1]]), nrow(rivers_sf))

  # Verify CRS detection (rivers is EPSG:3035)
  detected_crs <- attr(ds, "crs")
  expect_equal(sf::st_crs(detected_crs)$epsg, 3035)

  # Verify geometry is valid (rivers are linestrings)
  res_geom <- DBI::dbGetQuery(
    conn,
    sprintf("SELECT ST_AsText(geom) FROM %s LIMIT 1", view_name)
  )
  expect_true(nrow(res_geom) == 1)
  expect_true(grepl("LINESTRING", res_geom[[1]]))
})

test_that("ddbs_open_dataset detects CRS across generated and bundled formats", {
  skip_if_not_installed("arrow")

  conn <- ddbs_temp_conn()

  generated_geojson <- ddbs_create_temp_spatial_file(countries_sf, ext = "geojson", conn = conn)
  generated_geoparquet <- ddbs_create_temp_spatial_file(countries_sf, ext = "parquet", conn = conn)
  bundled_gpkg <- system.file("spatial/points.gpkg", package = "duckspatial")
  bundled_shp <- system.file("shape/nc.shp", package = "sf")

  generated_geojson_ds <- ddbs_open_dataset(generated_geojson, conn = conn)
  expect_equal(attr(generated_geojson_ds, "crs"), sf::st_crs(expect_duckdb_st_crs(generated_geojson_ds, conn)))

  generated_geoparquet_ds <- ddbs_open_dataset(generated_geoparquet, conn = conn)
  expect_equal(attr(generated_geoparquet_ds, "crs"), sf::st_crs(expect_duckdb_st_crs(generated_geoparquet_ds, conn)))
  expect_false(is.na(attr(generated_geoparquet_ds, "crs")))

  bundled_gpkg_ds <- ddbs_open_dataset(bundled_gpkg, conn = conn)
  expect_equal(attr(bundled_gpkg_ds, "crs"), sf::st_crs(expect_duckdb_st_crs(bundled_gpkg_ds, conn)))

  bundled_shp_ds <- ddbs_open_dataset(bundled_shp, conn = conn)
  expect_equal(sf::st_crs(attr(bundled_shp_ds, "crs"))$epsg, 4267)
})

# =============================================================================
# Dedicated reader dispatch tests
# =============================================================================

test_that("ddbs_open_dataset dispatches to ST_ReadSHP vs GDAL correctly", {
  conn <- ddbs_temp_conn()

  # Create temp shapefile using helper
  tmp_shp <- ddbs_create_temp_spatial_file(argentina_sf, ext = "shp", conn = conn)
  expect_true(file.exists(tmp_shp))

  # Default mode: ST_ReadSHP
  ds_shp <- ddbs_open_dataset(tmp_shp, conn = conn)
  expect_s3_class(ds_shp, "duckspatial_df")

  view_sql_shp <- DBI::dbGetQuery(
    conn,
    glue::glue("SELECT sql FROM duckdb_views() WHERE view_name = '{attr(ds_shp, 'source_table')}'")
  )$sql
  expect_true(grepl("st_readshp", view_sql_shp, ignore.case = TRUE))

  # Explicit GDAL mode
  ds_gdal <- ddbs_open_dataset(tmp_shp, conn = conn, read_shp_mode = "GDAL")
  expect_s3_class(ds_gdal, "duckspatial_df")

  view_sql_gdal <- DBI::dbGetQuery(
    conn,
    glue::glue("SELECT sql FROM duckdb_views() WHERE view_name = '{attr(ds_gdal, 'source_table')}'")
  )$sql
  expect_true(grepl("st_read", view_sql_gdal, ignore.case = TRUE))
  expect_false(grepl("st_readshp", view_sql_gdal, ignore.case = TRUE))

  # Data integrity: counts should match
  expect_equal(
    ds_shp |> dplyr::count() |> dplyr::collect() |> dplyr::pull(n),
    ds_gdal |> dplyr::count() |> dplyr::collect() |> dplyr::pull(n)
  )
})

test_that("ddbs_open_dataset handles shp_encoding argument", {
  conn <- ddbs_temp_conn()

  # Create temp shapefile using helper
  tmp_shp <- ddbs_create_temp_spatial_file(argentina_sf, ext = "shp", conn = conn)
  expect_true(file.exists(tmp_shp))

  ds_enc <- ddbs_open_dataset(tmp_shp, conn = conn, shp_encoding = "UTF-8")

  expect_s3_class(ds_enc, "duckspatial_df")

  # view_sql_enc <- DBI::dbGetQuery(
  #   conn,
  #   glue::glue("SELECT sql FROM duckdb_tables() WHERE table_name = '{attr(ds_enc, 'source_table')}'")
  # )$sql
  # expect_true(grepl("encoding.*UTF-8", view_sql_enc, ignore.case = TRUE))
})

test_that("ddbs_open_dataset OSM mode dispatch", {

  conn <- ddbs_temp_conn()

  # GDAL mode (default) - uses dummy file that doesn't exist.
  # Since our logic now correctly dispatches to ST_Read (GDAL), and we have strict error handling,
  # this should error "Unable to open file" because ST_Read validates existence.
  expect_error(
    ddbs_open_dataset("dummy.osm.pbf", conn = conn, read_osm_mode = "GDAL"),
    # "Could not open GDAL"
    "File format not recognized"
  )
  
  # ST_ReadOSM mode
  # This path is lazy/permissive and might not error on open, allowing us to inspect SQL.
  # ST_ReadOSM mode
  # Note: ST_ReadOSM does not produce a spatial column DuckDB understands,
  # so it returns a regular table, not a duckspatial_df.
  ds_osm_read <- ddbs_open_dataset("dummy.osm.pbf", conn = conn, read_osm_mode = "ST_ReadOSM")
  expect_false(inherits(ds_osm_read, "duckspatial_df"))

  # The view name is stored in the remote_name for dbplyr objects
  view_name_osm <- as.character(dbplyr::remote_name(ds_osm_read))

  view_sql_osm_read <- DBI::dbGetQuery(
    conn,
    glue::glue("SELECT sql FROM duckdb_views() WHERE view_name = '{view_name_osm}'")
  )$sql

  expect_true(grepl("st_readosm", view_sql_osm_read, ignore.case = TRUE))
})

# =============================================================================
# Argument validation and warnings
# =============================================================================

test_that("ddbs_open_dataset warns when ST_Read args are passed to Parquet", {
  skip_if_not_installed("arrow")

  conn <- ddbs_temp_conn()

  tmp_parquet <- tempfile(fileext = ".parquet")
  arrow::write_parquet(data.frame(x = 1, y = 2), tmp_parquet)
  on.exit(unlink(tmp_parquet), add = TRUE)

  expect_warning(
    ddbs_open_dataset(tmp_parquet, conn = conn, layer = "foo"),
    "Arguments specific to ST_Read .* are ignored"
  )
})

test_that("ddbs_open_dataset detects format correctly with GDAL options", {
  conn <- ddbs_temp_conn()

  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")

  ds <- ddbs_open_dataset(
    countries_path,
    conn = conn,
    gdal_open_options = c("FLATTEN_NESTED_ATTRIBUTES=YES")
  )

  expect_s3_class(ds, "duckspatial_df")
  res <- collect(ds)
  expect_true(nrow(res) > 0)
})


# =============================================================================
# CRS and geometry handling
# =============================================================================

test_that("ddbs_open_dataset handles explicit CRS override", {
  conn <- ddbs_temp_conn()

  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")

  # countries.geojson is EPSG:4326, override to 3857
  ds <- ddbs_open_dataset(countries_path, crs = 3857, conn = conn)

  expect_s3_class(ds, "duckspatial_df")

  # Verify CRS is the overridden value (3857), not the original (4326)
  detected_crs <- attr(ds, "crs")
  expect_equal(sf::st_crs(detected_crs)$epsg, 3857)
})

test_that("ddbs_open_dataset with custom geom_col", {
  conn <- ddbs_temp_conn()

  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")

  ds <- ddbs_open_dataset(countries_path, geom_col = "geom", conn = conn)

  expect_s3_class(ds, "duckspatial_df")
  expect_equal(attr(ds, "sf_column"), "geom")
})

# =============================================================================
# Error handling
# =============================================================================

test_that("ddbs_open_dataset handles missing file gracefully", {
  conn <- ddbs_temp_conn()

  expect_error(
    ddbs_open_dataset("/path/to/nonexistent/file.geojson", conn = conn),
    regexp = "Unable to open file",
    ignore.case = TRUE
  )
})

test_that("ddbs_open_dataset fails gracefully on non-compliant GeoArrow structs", {
  skip_if_not_installed("arrow")
  skip_if_not_installed("geoarrow")
  
  # Create a file with Arrow native encoding (NOT WKB) using direct write_parquet
  # This triggers the GeoArrow native struct encoding
  tmp_bad <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp_bad))
  arrow::write_parquet(countries_sf, tmp_bad)
  
  expect_error(ddbs_open_dataset(tmp_bad), "uses a native Arrow/GeoArrow struct encoding")
})

# =============================================================================
# DuckDB Native Format Support
# =============================================================================

test_that("ddbs_open_dataset validates DuckDB files properly", {
  expect_error(
    ddbs_open_dataset("/nonexistent_path.duckdb"),
    "does not exist"
  )

  tmp_empty_duck <- tempfile(fileext = ".duckdb")
  on.exit(unlink(tmp_empty_duck), add = TRUE)
  file.create(tmp_empty_duck)
  expect_error(
    ddbs_open_dataset(tmp_empty_duck),
    "not a valid DuckDB database"
  )

  tmp_bad_duck <- tempfile(fileext = ".duckdb")
  on.exit(unlink(tmp_bad_duck), add = TRUE)
  writeLines("this is just text", tmp_bad_duck)
  expect_error(
    ddbs_open_dataset(tmp_bad_duck),
    "not a valid DuckDB database"
  )

  # Use a unique file for the valid DuckDB test to avoid any interference
  # We use the established ddbs_temp_conn helper for clean connection management
  tmp_good_duck <- tempfile(fileext = ".duckdb")
  on.exit(unlink(tmp_good_duck), add = TRUE)

  local({
    conn <- ddbs_temp_conn(file = tmp_good_duck, cleanup = FALSE)
    ddbs_write_table(conn, countries_sf, "countries", quiet = TRUE)
  })

  expect_error(
    ddbs_open_dataset(tmp_good_duck),
    "layer"
  )

  expect_error(
    ddbs_open_dataset(tmp_good_duck, layer = "missing_table"),
    "not present"
  )

  ds <- ddbs_open_dataset(tmp_good_duck, layer = "countries", crs = 4326)
  expect_s3_class(ds, "duckspatial_df")
  expect_equal(as.character(dbplyr::remote_name(ds)), "countries")
  expect_equal(nrow(dplyr::collect(ds)), nrow(countries_sf))
})

test_that("ddbs_open_dataset opens supported DuckDB file extensions", {
  # Helper to create a fresh DB file for each extension to avoid lock issues
  create_duckdb_test_file <- function(ext) {
    db_path <- tempfile(fileext = ext)
    local({
      conn <- ddbs_temp_conn(file = db_path, cleanup = FALSE)
      ddbs_write_table(conn, countries_sf, "countries", quiet = TRUE)
    })
    db_path
  }

  expect_duckdb_open <- function(db_path) {
    on.exit(unlink(db_path), add = TRUE)
    ds <- ddbs_open_dataset(db_path, layer = "countries", crs = 4326)

    expect_s3_class(ds, "duckspatial_df")
    expect_equal(as.character(dbplyr::remote_name(ds)), "countries")
    expect_equal(nrow(dplyr::collect(ds)), nrow(countries_sf))
  }

  for (ext in c(".duckdb", ".db", ".ddb")) {
    expect_duckdb_open(create_duckdb_test_file(ext))
  }
})

test_that("ddbs_open_dataset does not open DuckDB files with unsupported extensions natively", {
  tmp_txt <- tempfile(fileext = ".txt")
  on.exit(unlink(tmp_txt), add = TRUE)

  local({
    conn <- ddbs_temp_conn(file = tmp_txt, cleanup = FALSE)
    DBI::dbExecute(conn, "CREATE TABLE countries AS SELECT 1 AS value")
  })

  expect_error(
    ddbs_open_dataset(tmp_txt, layer = "countries"),
    regexp = "Unable to open file",
    ignore.case = TRUE
  )
})

test_that("ddbs_open_dataset lets non-DuckDB .db files fall through to ST_Read", {
  tmp_db <- tempfile(fileext = ".db")
  on.exit(unlink(tmp_db), add = TRUE)
  writeLines("not a duckdb database", tmp_db)

  expect_error(
    ddbs_open_dataset(tmp_db),
    regexp = "Unable to open file",
    ignore.case = TRUE
  )
})

testthat::skip_on_cran()

test_that("ddbs_create_temp_spatial_file creates file and cleans up automatically", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)
  
  # Track the temp directory
  temp_path <- NULL
  temp_file_path <- NULL
  
  # Use a local scope to trigger cleanup
  local({
    temp_file_path <<- ddbs_create_temp_spatial_file(ds, ext = "geojson", conn = conn)
    temp_path <<- dirname(temp_file_path)
    
    # File should exist within the function scope
    expect_true(file.exists(temp_file_path))
    expect_true(dir.exists(temp_path))
    
    # Verify it's actually geojson by reading it back
    ds_back <- ddbs_open_dataset(temp_file_path, conn = conn)
    expect_equal(dplyr::count(ds_back) |> dplyr::pull(n), 257)
  })
  
  # After exiting scope, directory should be cleaned up
  expect_false(dir.exists(temp_path))
})

test_that("ddbs_create_temp_spatial_file works with shapefile format", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)
  
  temp_path <- NULL
  
  local({
    # Shapefiles create multiple files, so directory cleanup is crucial
    temp_file <- ddbs_create_temp_spatial_file(ds, ext = "shp", conn = conn)
    temp_path <<- dirname(temp_file)
    
    expect_true(file.exists(temp_file))
    
    # Shapefiles create .shp, .shx, .dbf, etc.
    # Check that multiple files exist in the directory
    all_files <- list.files(temp_path, recursive = TRUE)
    expect_gt(length(all_files), 1) # Should have multiple associated files
  })
  
  # Cleanup should remove all shapefile components
  expect_false(dir.exists(temp_path))
})

test_that("ddbs_create_temp_spatial_file passes through additional arguments", {
  conn <- ddbs_temp_conn()
  path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds <- ddbs_open_dataset(path, conn = conn)
  
  local({
    # Test passing crs argument through
    temp_file <- ddbs_create_temp_spatial_file(
      ds, 
      ext = "geojson", 
      conn = conn,
      crs = "EPSG:3857"
    )
    
    expect_true(file.exists(temp_file))
    
    # Verify CRS was set
    sf_obj <- sf::st_read(temp_file, quiet = TRUE)
    expect_equal(sf::st_crs(sf_obj)$epsg, 3857)
  })
})

#' Create temporary spatial file for testing
#'
#' This helper creates a temporary spatial file in an isolated directory
#' that is automatically cleaned up when the calling function exits.
#' Useful for tests involving file-based spatial formats (e.g., shapefiles)
#' that create multiple associated files.
#'
#' @param data A `duckspatial_df`, `tbl_lazy`, or `sf` object to write
#' @param ext File extension (e.g., "geojson", "shp", "parquet"). Defaults to "geojson".
#' @param gdal_driver Optional explicit GDAL driver (overrides extension detection)
#' @param conn DuckDB connection. If NULL, attempts to infer from data.
#' @param envir Environment for cleanup context. Defaults to parent.frame()
#'   to ensure cleanup happens in the test context.
#' @param ... Additional arguments passed to ddbs_write_dataset()
#'
#' @return Character string path to the created file
#'
#' @examples
#' \dontrun{
#' test_that("example test", {
#'   conn <- ddbs_temp_conn()
#'   ds <- ddbs_open_dataset(system.file("spatial/countries.geojson", package = "duckspatial"))
#'   
#'   # Create temp file - automatically cleaned up after test
#'   temp_file <- ddbs_create_temp_spatial_file(ds, ext = "shp", conn = conn)
#'   expect_true(file.exists(temp_file))
#' })
#' }
#' @noRd
ddbs_create_temp_spatial_file <- function(
  data,
  ext = "geojson",
  gdal_driver = NULL,
  conn = NULL,
  envir = parent.frame(),
  ...
) {
  # Create isolated temporary directory with automatic cleanup
  temp_dir <- withr::local_tempdir(.local_envir = envir)
  
  # Generate file path
  file_path <- file.path(temp_dir, paste0("test_data.", ext))
  
  # Write the dataset using new API (auto-detects from extension)
  ddbs_write_dataset(
    data = data,
    path = file_path,
    gdal_driver = gdal_driver,
    conn = conn,
    quiet = TRUE,
    ...
  )
  
  return(file_path)
}

#' Areal-Weighted Interpolation using DuckDB
#'
#' @description
#' Transfers attribute data from a source spatial layer to a target spatial layer based
#' on the area of overlap between their geometries. This function executes all spatial
#' calculations within DuckDB, enabling efficient processing of large datasets without
#' loading all geometries into R memory.
#'
#' @details
#' Areal-weighted interpolation is used when the source and target geometries are incongruent (they do not align). It relies on the assumption of **uniform distribution**: values in the source polygons are assumed to be spread evenly across the polygon's area.
#'
#' **Coordinate Systems:**
#' Area calculations are highly sensitive to the Coordinate Reference System (CRS).
#' While the function can run on geographic coordinates (lon/lat), it is strongly recommended
#' to use a **projected CRS** (e.g., EPSG:3857, UTM, or Albers) to ensure accurate area measurements.
#' Use the \code{join_crs} argument to project data on-the-fly during the interpolation.
#'
#' **Extensive vs. Intensive Variables:**
#' \itemize{
#'   \item **Extensive** variables are counts or absolute amounts (e.g., total population,
#'   number of voters). When a source polygon is split, the value is divided proportionally
#'   to the area.
#'   \item **Intensive** variables are ratios, rates, or densities (e.g., population density,
#'   cancer rates). When a source polygon is split, the value remains constant for each piece.
#' }
#'
#' **Mass Preservation (The \code{weight} argument):**
#' For extensive variables, the choice of weight determines the denominator used in calculations:
#' \itemize{
#'   \item \code{"sum"} (default): The denominator is the sum of all overlapping areas
#'   for that source feature. This preserves the "mass" of the variable *relative to the target's coverage*.
#'   If the target polygons do not completely cover a source polygon, some data is technically "lost"
#'   because it falls outside the target area. This matches \code{areal::aw_interpolate(weight="sum")}.
#'   \item \code{"total"}: The denominator is the full geometric area of the source feature.
#'   This assumes the source value is distributed over the entire source polygon. If the target
#'   covers only 50% of the source, only 50% of the value is transferred. This is strictly
#'   mass-preserving relative to the source. This matches \code{sf::st_interpolate_aw(extensive=TRUE)}.
#' }
#' *Note:* Intensive variables are always calculated using the \code{"sum"} logic (averaging
#' based on intersection areas) regardless of this parameter.
#'
#' @param target An \code{sf} object or the name of a persistent table in the DuckDB connection
#'   representing the destination geometries.
#' @param source An \code{sf} object or the name of a persistent table in the DuckDB connection
#'   containing the data to be interpolated.
#' @param tid Character. The name of the column in \code{target} that uniquely identifies features.
#' @param sid Character. The name of the column in \code{source} that uniquely identifies features.
#' @param extensive Character vector. Names of columns in \code{source} to be treated as
#'   spatially extensive (e.g., population counts).
#' @param intensive Character vector. Names of columns in \code{source} to be treated as
#'   spatially intensive (e.g., population density).
#' @param weight Character. Determines the denominator calculation for extensive variables.
#'   Either \code{"sum"} (default) or \code{"total"}. See **Mass Preservation** in Details.
#' @template mode
#' @param keep_NA Logical. If \code{TRUE} (default), returns all features from the target,
#'   even those that do not overlap with the source (values will be NA). If \code{FALSE},
#'   performs an inner join, dropping non-overlapping target features.
#' @param na.rm Logical. If \code{TRUE}, source features with \code{NA} values in the
#'   interpolated variables are completely removed from the calculation (area calculations
#'   will behave as if that polygon did not exist). Defaults to \code{FALSE}.
#' @param join_crs Numeric or Character (optional). EPSG code or WKT for the CRS to use
#'   for area calculations. If provided, both \code{target} and \code{source} are transformed
#'   to this CRS within the database before interpolation.
#' @template conn_null
#' @template name
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @examples
#' \donttest{
#' library(sf)
#'
#' # 1. Prepare Data
#' # Load NC counties (Source) and project to Albers (EPSG:5070)
#' nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
#' nc <- st_transform(nc, 5070)
#' nc$sid <- seq_len(nrow(nc)) # Create Source ID
#'
#' # Create a target grid
#' g <- st_make_grid(nc, n = c(10, 5))
#' g_sf <- st_as_sf(g)
#' g_sf$tid <- seq_len(nrow(g_sf)) # Create Target ID
#'
#' # 2. Extensive Interpolation (Counts)
#' # Use weight = "total" for strict mass preservation (e.g., total births)
#' res_ext <- ddbs_interpolate_aw(
#'   target = g_sf, source = nc,
#'   tid = "tid", sid = "sid",
#'   extensive = "BIR74",
#'   weight = "total",
#'   mode = "sf"
#' )
#'
#' # Check mass preservation
#' sum(res_ext$BIR74, na.rm = TRUE) / sum(nc$BIR74) # Should be ~1
#'
#' # 3. Intensive Interpolation (Density/Rates)
#' # Calculates area-weighted average (e.g., assumption of uniform density)
#' res_int <- ddbs_interpolate_aw(
#'   target = g_sf, source = nc,
#'   tid = "tid", sid = "sid",
#'   intensive = "BIR74",
#'   mode = "sf"
#' )
#'
#' # 4. Quick Visualization
#' par(mfrow = c(1, 2))
#' plot(res_ext["BIR74"], main = "Extensive (Total Count)", border = NA)
#' plot(res_int["BIR74"], main = "Intensive (Weighted Avg)", border = NA)
#' }
#'
#' @seealso
#' \code{\link[areal:aw_interpolate]{areal::aw_interpolate()}} — reference implementation.
#'
#' @references
#' Prener, C. and Revord, C. (2019). \emph{areal: An R package for areal weighted interpolation}.
#' \emph{Journal of Open Source Software}, 4(37), 1221.
#' Available at: \doi{10.21105/joss.01221}
#'
#' @export
ddbs_interpolate_aw <- function(
    target,
    source,
    tid,
    sid,
    extensive = NULL,
    intensive = NULL,
    weight = "sum",
    mode = NULL,
    keep_NA = TRUE,
    na.rm = FALSE,
    join_crs = NULL,
    conn = NULL,
    name = NULL,
    overwrite = FALSE,
    quiet = FALSE
) {

  # 0. Handle inputs and errors
  assert_xy(target, "target")
  assert_xy(source, "source")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  assert_logic(keep_NA, "keep_NA")
  assert_logic(na.rm, "na.rm")
  assert_conn_character(conn, target, source)

  # 0. Validation Logic
  if (missing(tid)) cli::cli_abort("{.arg tid} must be provided.")
  if (missing(sid)) cli::cli_abort("{.arg sid} must be provided.")
  if (is.null(extensive) && is.null(intensive)) {
    cli::cli_abort("At least one of {.arg extensive} or {.arg intensive} must be provided.")
  }
  if (!weight %in% c("sum", "total")) {
    cli::cli_abort("{.arg weight} must be either 'sum' or 'total'.")
  }

  # Strict validation: Intensive variables cannot use "total" weight
  # because summing densities across total source areas is mathematically invalid.
  if (!is.null(intensive) && weight == "total") {
    cli::cli_abort("Spatially intensive variables must use {.code weight = 'sum'}.")
  }

  # 2. Normalize inputs

  # Pre-extract CRS and sf_column (before normalize_spatial_input converts types)
  t_geom <- attr(target, "sf_column")
  s_geom <- attr(source, "sf_column")
  t_crs  <- ddbs_crs(target, conn)
  s_crs  <- ddbs_crs(source, conn)

  # Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, validate character table names
  target <- normalize_spatial_input(target, conn)
  source <- normalize_spatial_input(source, conn)

  ## Get mode - If it's NULL, it will use the duckspatial.mode option
  mode <- get_mode(mode, name)

    
  # 3. Manage connection to DB

  ## 3.1. Resolve connections and handle imports
  resolve_res <- resolve_spatial_connections(target, source, conn, quiet = quiet)

  target_conn <- resolve_res$conn
  target <- resolve_res$x
  source <- resolve_res$y
  
  # 3.2 Get query list
  t_list <- get_query_list(target, target_conn)
  on.exit(t_list$cleanup(), add = TRUE)
  s_list <- get_query_list(source, target_conn)
  on.exit(s_list$cleanup(), add = TRUE)
  

  # 4. Prepare parameters for query

  ## 4.1. predicate already validated early (sel_pred above)
  ## get names of geometry columns (use saved sf_col_x/y from before transformation)
  t_geom <- t_geom %||% get_geom_name(target_conn, t_list$query_name)
  s_geom <- s_geom %||% get_geom_name(target_conn, s_list$query_name)
  assert_geometry_column(t_geom, t_list)
  assert_geometry_column(s_geom, s_list)

  ## Default to raw geometry columns (overwritten below if transformation is requested)
  t_geom_expr <- t_geom
  s_geom_expr <- s_geom

  ## 4.2. Manage CRS
  if (!is.null(join_crs)) {

    # If we need to reproject (join_crs provided), both inputs MUST have a known CRS.
    if (is.na(t_crs)) cli::cli_abort("Target CRS unknown. Cannot transform to {.arg join_crs}.")
    if (is.na(s_crs)) cli::cli_abort("Source CRS unknown. Cannot transform to {.arg join_crs}.")


    join_crs_sql <- crs_to_sql(join_crs)

    # Convert those objects to SQL literals
    t_crs_sql <- crs_to_sql(t_crs)
    s_crs_sql <- crs_to_sql(s_crs)

    if (t_crs_sql == "NULL") cli::cli_abort("Target CRS value is NULL. Cannot transform to {.arg join_crs}.")
    if (s_crs_sql == "NULL") cli::cli_abort("Source CRS value is NULL. Cannot transform to {.arg join_crs}.")

    t_geom_expr <- glue::glue("ST_Transform({t_geom}, {t_crs_sql}, {join_crs_sql})")
    s_geom_expr <- glue::glue("ST_Transform({s_geom}, {s_crs_sql}, {join_crs_sql})")
  } else {
    # If NO join_crs provided, inputs MUST match.
    if (!is.null(t_crs) && !is.null(s_crs)) {
      if (!crs_equal(t_crs, s_crs)) {
        cli::cli_abort("The Coordinates Reference System of {.arg target} and {.arg source} is different.")
      }
  } else {
      assert_crs(target_conn, t_list$query_name, s_list$query_name)
  }
    # If neither has CRS (t_has_crs=FALSE, s_has_crs=FALSE),
    # we assume they are both planar/NA and proceed without error.
  }

  ## 4.3. Get Attribute Columns (target columns to keep)
  t_rest <- get_geom_name(target_conn, t_list$query_name, rest = TRUE, collapse = FALSE)

  ## 4.4. Column Conflict Prevention
  # Check if interpolated vars already exist in target (excluding tid)
  interp_vars <- c(extensive, intensive)
  conflicts <- intersect(interp_vars, t_rest)
  conflicts <- setdiff(conflicts, tid) # ignore tid overlap as we join on it

  if (length(conflicts) > 0) {
    cli::cli_warn(c(
      "Column name conflict detected.",
      "i" = "The variables {.val {conflicts}} exist in both the target table and the interpolation list.",
      "!" = "The output will contain the interpolated values, overwriting the original target columns."
    ))
    # We remove conflicts from t_rest so the SQL doesn't select the original target cols
    t_rest <- setdiff(t_rest, conflicts)
  }

  ## 4.5. Validate IDs and Variables exist
  assert_col_exists(target_conn, t_list$query_name, tid, "target")
  assert_col_exists(target_conn, s_list$query_name, sid, "source")

  if (!is.null(extensive)) assert_col_exists(target_conn, s_list$query_name, extensive, "source")
  if (!is.null(intensive)) assert_col_exists(target_conn, s_list$query_name, intensive, "source")


  # 5. Build CTE Query
  s_alias <- "s_geom_proj"
  t_alias <- "t_geom_proj"

  # Apply na.rm: Filter source table if requested
  s_source_sql <- s_list$query_name
  if (isTRUE(na.rm)) {
    vars_to_check <- c(extensive, intensive)
    where_clause <- paste(paste0(vars_to_check, " IS NOT NULL"), collapse = " AND ")
    s_source_sql <- glue::glue("(SELECT * FROM {s_list$query_name} WHERE {where_clause})")
  }

  # 5.1 Intersection/Overlap CTE
  overlap_cte <- glue::glue("
    overlap_calc AS (
      SELECT
        s.{sid} AS sid,
        t.{tid} AS tid,
        COALESCE(ST_Area(ST_Intersection({s_alias}, {t_alias})), 0) AS overlap_area
      FROM
        (SELECT *, {s_geom_expr} AS {s_alias} FROM {s_source_sql}) s
      INNER JOIN
        (SELECT *, {t_geom_expr} AS {t_alias} FROM {t_list$query_name}) t
      ON ST_Intersects({s_alias}, {t_alias})
    )
  ")

  # 5.2 Denominators
  denom_ctes <- character()

  if (!is.null(extensive)) {
    if (weight == "sum") {
      # Matches areal::aw_interpolate(weight="sum")
      denom_ctes <- c(denom_ctes, glue::glue("
        denom_extensive AS (
          SELECT sid, SUM(overlap_area) as total_area_sid
          FROM overlap_calc
          GROUP BY sid
        )
      "))
    } else {
      # Matches sf::st_interpolate_aw(extensive=TRUE)
      denom_ctes <- c(denom_ctes, glue::glue("
        denom_extensive AS (
          SELECT {sid} as sid, ST_Area({s_geom_expr}) as total_area_sid
          FROM {s_source_sql}
        )
      "))
    }
  }

  if (!is.null(intensive)) {
    # Matches sf::st_interpolate_aw(extensive=FALSE) logic
    denom_ctes <- c(denom_ctes, glue::glue("
      denom_intensive AS (
        SELECT tid, SUM(overlap_area) as total_area_tid
        FROM overlap_calc
        GROUP BY tid
      )
    "))
  }

  # 5.3 Aggregation Logic
  select_exprs <- character()
  if (!is.null(extensive)) {
    for (v in extensive) {
      # NULLIF protects against division by zero (empty geometry or zero area)
      select_exprs <- c(select_exprs, glue::glue(
        "SUM( (src.{v} * o.overlap_area) / NULLIF(dens.total_area_sid, 0) ) AS {v}"
      ))
    }
  }
  if (!is.null(intensive)) {
    for (v in intensive) {
      select_exprs <- c(select_exprs, glue::glue(
        "SUM( (src.{v} * o.overlap_area) / NULLIF(deni.total_area_tid, 0) ) AS {v}"
      ))
    }
  }

  agg_fields <- paste(select_exprs, collapse = ", ")

  # Join back to filtered source (if na.rm=TRUE) or original
  # Since we might have filtered source in overlap_calc but need attributes here
  src_join_sql <- s_source_sql # Reuse the filtered subquery logic

  joins_sql <- glue::glue("
    FROM overlap_calc o
    JOIN {src_join_sql} src ON o.sid = src.{sid}
  ")

  if (!is.null(extensive)) {
    joins_sql <- paste(joins_sql, "LEFT JOIN denom_extensive dens ON o.sid = dens.sid")
  }
  if (!is.null(intensive)) {
    joins_sql <- paste(joins_sql, "LEFT JOIN denom_intensive deni ON o.tid = deni.tid")
  }

  agg_cte <- glue::glue("
    aggregated_values AS (
      SELECT
        o.tid,
        {agg_fields}
      {joins_sql}
      GROUP BY o.tid
    )
  ")

  # 6. Final Execution
  # Explicitly select target columns to keep attributes
  t_cols_select <- if(length(t_rest) > 0) paste0("tgt.", t_rest, collapse = ", ") else ""
  if (t_cols_select != "") t_cols_select <- paste0(t_cols_select, ", ")

  # Apply keep_NA logic
  # if keep_NA=TRUE, use LEFT JOIN (preserve all targets)
  # if keep_NA=FALSE, use INNER JOIN (keep only targets with interpolated data)
  final_join_type <- if (isTRUE(keep_NA)) "LEFT JOIN" else "INNER JOIN"
  full_ctes <- paste(c(overlap_cte, denom_ctes, agg_cte), collapse = ",\n")

  # 6.1 Handle Output Type: SF vs Tibble (no geometry)
  st_function <- glue::glue("tgt.{t_geom}")
  final_select <- glue::glue("
    SELECT
      {t_cols_select}
      {build_geom_query(st_function, name, t_crs, mode)} as {t_geom},
      av.* EXCLUDE (tid)
    FROM {t_list$query_name} tgt
    {final_join_type} aggregated_values av ON tgt.{tid} = av.tid
  ")
  table_select <- glue::glue("
    SELECT
      {t_cols_select}
      {build_geom_query(st_function, name, t_crs, mode)} as {t_geom},
      av.* EXCLUDE (tid)
    FROM {t_list$query_name} tgt
    {final_join_type} aggregated_values av ON tgt.{tid} = av.tid
  ")

  # 6.2 Execute

  if (!is.null(name)) {
    name_list <- get_query_name(name)
    overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

    # CREATE TABLE must precede WITH for standard CTE usage in DuckDB statements like this
    full_sql <- glue::glue("
      CREATE TABLE {name_list$query_name} AS
      WITH {full_ctes}
      {table_select}
    ")

    DBI::dbExecute(target_conn, full_sql)
    feedback_query(quiet)
    return(invisible(TRUE))
  }

  full_sql <- glue::glue("
    WITH {full_ctes}
    {final_select}
  ")

  result <- ddbs_handle_query(
        query  = full_sql,
        conn   = target_conn,
        mode   = mode,
        crs    = t_crs,
        x_geom = t_geom
    )

  return(result)

}

#' Create a duckspatial lazy spatial data frame
#'
#' Extends tbl_duckdb_connection with spatial metadata (CRS, geometry column).
#'
#' @param x Input: tbl_duckdb_connection, tbl_lazy, or similar dbplyr object
#' @param crs CRS object or string
#' @param geom_col Name of geometry column (default: "geom")
#' @param source_table Name of the source table if applicable
#' @param source_conn Name of the source connection if applicable
#' @param create_view Logical. If TRUE, creates a temporary view for the input query. 
#' Otherwise it generates a temporary table.
#' @return A duckspatial_df object
#' @keywords internal
new_duckspatial_df <- function(
  x, 
  crs = NULL, 
  geom_col = NULL, 
  source_table = NULL,
  source_conn = NULL,
  create_view = FALSE
) {
  # Avoid double wrapping
  if (is_duckspatial_df(x)) return(x)
  
  # This will manage dplyr methods
  # Maybe move to duckspatial_df.tbl_duckdb_connection in the future
  if (inherits(x, "tbl_sql") && is.null(source_table) && !is.null(source_conn)) {

    # Here we won't have a source table, so we will need to create it
    if (create_view) {
      which <- "VIEW"
      source_table <- ddbs_temp_view_name()
    } else {
      which <- "TABLE"
      source_table <- ddbs_temp_table_name()
    }

    # Use sql_render to extract the query
    inner_query <- dbplyr::sql_render(x)

    # Create the table that will be returned as source_table
    # This executes the dplyr verb
    DBI::dbExecute(
      source_conn,
      glue::glue("
        CREATE OR REPLACE TEMP {which} {source_table} AS
        ({inner_query});"
      )
    )
    
    # Handle as a lazy duckdb table in the next step
    x <- dplyr::tbl(source_conn, source_table)
  }
  
  if (!inherits(x, "tbl_sql")) {
    cli::cli_abort("{.arg x} must be a {.cls tbl_sql} (lazy DuckDB table). Use {.fn as_duckspatial_df} for other objects.")
  }

  # If geometry column is not provided, use geom by default
  geom_col <- geom_col %||% "geom"
  
  # Prepend our class
  structure(
    x,
    class = c("duckspatial_df", class(x)),
    sf_column = geom_col, # Keeping attribute name as sf_column for compatibility
    crs = if (inherits(crs, "crs")) crs else sf::st_crs(crs),
    source_table = source_table,
    source_conn = source_conn
  )
}


#' Check if object is a duckspatial_df
#' @param x Object to test
#' @return Logical
#' @export
is_duckspatial_df <- function(x) {
  inherits(x, "duckspatial_df")
}

#' Convert objects to duckspatial_df
#'
#' @description
#' `as_duckspatial_df()` creates a lazy spatial data frame (`duckspatial_df`) from 
#' various inputs. When `x` is a table name (character) or an existing DuckDB 
#' table (`tbl_duckdb_connection`), the function creates a zero-copy representation 
#' of the data directly from the database without loading it into memory. This is 
#' the canonical way to "register" or wrap existing persistent spatial tables.
#' 
#' **CRS Persistence:** `duckspatial` reads native DuckDB 1.5.0+ CRS metadata 
#' and, for compatibility with files written by older versions of `duckspatial`, 
#' CRS metadata stored in column comments. DuckDB files saved in pre-1.5.0 
#' format without `duckspatial`-managed comments will not have CRS information 
#' and will default to `NA` with a warning.
#'
#' @param x Object to convert (sf, tbl_lazy, data.frame, or table name)
#' @param conn DuckDB connection (required for character table names)
#' @param crs CRS object or string. Auto-detected from `sf` objects and 
#'   persistent DuckDB tables.
#' @param geom_col Geometry column name (default: "geom")
#' @param ... Additional arguments passed to methods:
#'   \describe{
#'     \item{\code{coords}}{Character vector of length 2 for point ingestion}
#'     \item{\code{wkt}}{Character name of WKT column for ingestion}
#'     \item{\code{remove}}{Logical. If TRUE (default), coordinate/WKT columns are removed}
#'     \item{\code{na.fail}}{Logical. If TRUE (default), errors on missing values}
#'   }
#' @return A duckspatial_df object
#' @export
as_duckspatial_df <- function(x, conn = NULL, crs = NULL, geom_col = NULL, ...) {
  ddbs_assert_duckdb_crs_support()
  UseMethod("as_duckspatial_df")
}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.duckspatial_df <- function(x, conn = NULL, crs = NULL, 
                                              geom_col = NULL, ...) {
  if (is.null(crs) && is.null(geom_col)) return(x)
  
  # Update metadata if requested
  if (!is.null(crs)) attr(x, "crs") <- sf::st_crs(crs)
  if (!is.null(geom_col)) attr(x, "sf_column") <- geom_col
  
  x
}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.sf <- function(x, conn = NULL, crs = NULL, geom_col = NULL, ...) {
  # Get CRS and geom column from sf object
  if (is.null(crs)) crs <- sf::st_crs(x)
  if (is.null(geom_col)) geom_col <- attr(x, "sf_column")
  
  # Get or create connection
  if (is.null(conn)) {
    conn <- ddbs_default_conn()
  }
  
  # Register sf as temp view
  view_name <- ddbs_temp_view_name()
  ddbs_write_table(
    conn = conn,
    data = x,
    name = view_name,
    quiet = TRUE,
    temp_view = TRUE
  )
  
  # Create lazy table
  lazy_tbl <- dplyr::tbl(conn, view_name)
  result <- new_duckspatial_df(
    lazy_tbl, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = view_name,
    source_conn = conn
  )
  
  return(result)

}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.tbl_duckdb_connection <- function(
  x, 
  conn = NULL, 
  crs = NULL,
  geom_col = NULL, ...
) {
  dots <- list(...)
  if (!is.null(dots$coords) || !is.null(dots$wkt)) {
    return(handle_heterogeneous_ingestion(x, conn, crs, geom_col, ...))
  }

  # Auto-detect CRS if not provided (DuckDB-specific)
  conn <- conn %||% dbplyr::remote_con(x)
  if (is.null(crs) && is.null(dots$coords) && is.null(dots$wkt)) {
    crs <- ddbs_crs(x, conn)
  }
  
  # Auto-detect geometry column
  if (is.null(geom_col)) {
    cols <- colnames(x)
    if ("geometry" %in% cols) {
      geom_col <- "geometry"
    } else if ("geom" %in% cols) {
      geom_col <- "geom"
    } else {
      geom_col <- "geom"
    }
  }
  
  # Extract source table for efficient get_query_list path
  source_table <- tryCatch({
    rem_name <- dbplyr::remote_name(x)
    if (is.null(rem_name)) NULL else as.character(rem_name)
  },
    error = function(e) NULL
  )
  
  # Auto-detect geometry type for collect optimization
  geom_type <- tryCatch({
    desc <- DBI::dbGetQuery(conn, glue::glue("DESCRIBE {dbplyr::sql_render(x)}"))
    desc$column_type[desc$column_name == geom_col]
  }, error = function(e) NULL)
  
  res <- new_duckspatial_df(
    x, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = source_table,
    source_conn = conn
  )
  if (!is.null(geom_type)) attr(res, "geom_type") <- geom_type
  res
}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.tbl_lazy <- function(
  x, 
  conn = NULL, 
  crs = NULL, 
  geom_col = NULL, ...
) {
  dots <- list(...)
  if (!is.null(dots$coords) || !is.null(dots$wkt)) {
    return(handle_heterogeneous_ingestion(x, conn, crs, geom_col, ...))
  }

  # Auto-detect CRS if not provided
  conn <- conn %||% dbplyr::remote_con(x)
  if (is.null(crs) && is.null(dots$coords) && is.null(dots$wkt)) {
    crs <- ddbs_crs(x, conn)
  }
  
  # Auto-detect geometry column
  if (is.null(geom_col)) {
    geom_col <- get_geom_name(conn, x)
  }
  
  # Extract source table for efficient get_query_list path
  source_table <- tryCatch(
    as.character(dbplyr::remote_name(x)),
    error = function(e) NULL
  )
  
  new_duckspatial_df(
    x, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = source_table,
    source_conn = conn
  )
}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.character <- function(
  x, 
  conn = NULL, 
  crs = NULL, 
  geom_col = NULL, ...
) {
  if (is.null(conn)) {
    conn <- ddbs_default_conn(create = FALSE)
    if (is.null(conn)) {
      cli::cli_abort("{.arg conn} must be provided when using table names as input.")
    }
  }

  dots <- list(...)
  if (!is.null(dots$coords) || !is.null(dots$wkt)) {
    return(handle_heterogeneous_ingestion(x, conn, crs, geom_col, ...))
  }
  
  lazy_tbl <- dplyr::tbl(conn, x)
  
  # Auto-detect geometry column
  if (is.null(geom_col)) {
    geom_col <- get_geom_name(conn, x)
  }

  # Auto-detect CRS if not provided
  if (is.null(crs) && is.null(dots$coords) && is.null(dots$wkt)) {
    crs <- ddbs_crs(x, conn)
  }
  
  new_duckspatial_df(
    lazy_tbl, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = x,
    source_conn = conn
  )

}

#' @rdname as_duckspatial_df
#' @export
as_duckspatial_df.data.frame <- function(
  x, 
  conn = NULL, 
  crs = NULL, 
  geom_col = NULL, ...
) {
   # Detect if we have an sfc column that matches geom_col or any sfc column
   is_sfc <- vapply(x, inherits, logical(1), "sfc")
   
   if (any(is_sfc)) {
     # If user provided geom_col, check if it's one of the sfc columns
     if (!is.null(geom_col) && geom_col %in% names(x)) {
        if (!inherits(x[[geom_col]], "sfc")) {
           cli::cli_abort("Column {.val {geom_col}} is not an {.cls sfc} column.")
        }
        # Delegate to sf handler
        return(as_duckspatial_df(sf::st_as_sf(x, sf_column_name = geom_col), conn = conn, crs = crs, ...))
     } else if (inherits(x, "sf")) {
        # Already sf, delegate
        return(as_duckspatial_df.sf(x, conn = conn, crs = crs, geom_col = geom_col, ...))
     } else {
        # Pick the first sfc column as geometry
        first_sfc <- names(x)[which(is_sfc)[1]]
        cli::cli_inform("Detected {.cls sfc} column {.val {first_sfc}}, converting to {.cls sf} first.")
        return(as_duckspatial_df(sf::st_as_sf(x, sf_column_name = first_sfc), conn = conn, crs = crs, ...))
     }
   }

   # Support coords/wkt ingestion for data.frames
   dots <- list(...)
   if (!is.null(dots$coords) || !is.null(dots$wkt)) {
     return(handle_heterogeneous_ingestion(x, conn, crs, geom_col, ...))
   }
   
   # Non-spatial data frame path (only if we can't find sfc)
   # NOTE: Usually we want to error here if it's not spatial, 
   # but maybe someone wants to upload a table and then set_geometry later?
   # For now, let's enforce that it must have at least one sfc or we error.
   if (!any(is_sfc)) {
     # Auto-detect geometry column for raw upload
     if (is.null(geom_col)) {
       if ("geometry" %in% names(x)) {
         geom_col <- "geometry"
       } else if ("geom" %in% names(x)) {
         geom_col <- "geom"
       }
     }
     
     if (!is.null(geom_col) && geom_col %in% names(x)) {
        cli::cli_warn("Column {.val {geom_col}} exists but is not {.cls sfc}. Attempting raw upload.")
     } else {
        cli::cli_abort(c(
          "No {.cls sfc} geometry column found in data frame.",
          "i" = "Use {.fn st_as_sf} first, or provide {.arg coords} or {.arg wkt} to ingest non-spatial data."
        ))
     }
   }

   # Upload to DuckDB
   if (is.null(conn)) conn <- ddbs_default_conn()
   
   view_name <- ddbs_temp_table_name()
   DBI::dbWriteTable(conn, view_name, x)
   
   lazy_tbl <- dplyr::tbl(conn, view_name)
   new_duckspatial_df(
    lazy_tbl, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = view_name,
    source_conn = conn
  )
}

#' Handle ingestion from coords or WKT
#' @keywords internal
handle_heterogeneous_ingestion <- function(x, conn, crs, geom_col, ...) {
  dots <- list(...)
  coords <- dots$coords
  wkt <- dots$wkt
  remove <- dots$remove %||% TRUE
  na.fail <- dots$na.fail %||% TRUE
  
  # 1. Normalize/Resolve
  # We avoid normalize_spatial_input here because it calls as_duckspatial_df,
  # which might call us back. We just need a connection and a query name.
  target_conn <- conn %||% get_conn_from_input(x) %||% ddbs_default_conn()
  
  if (inherits(x, "sf")) {
     # Already handled by as_duckspatial_df.sf before calling us
  } else if (is.data.frame(x)) {
     view_name <- ddbs_temp_table_name()
     DBI::dbWriteTable(target_conn, view_name, x)
     x <- view_name
  }
  
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = target_conn, quiet = TRUE)
  target_conn <- resolve_conn$conn
  on.exit(resolve_conn$cleanup(), add = TRUE)
  
  x_list <- get_query_list(resolve_conn$x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)
  
  # We use geom_col as intermediate name, ddbs_handle_query will use it
  target_geom <- geom_col %||% "geom"
  
  if (!is.null(coords)) {
    if (!is.character(coords) || length(coords) != 2) {
      cli::cli_abort("{.arg coords} must be a character vector of length 2.")
    }
    crs <- crs %||% "EPSG:4326"
    coords_quoted <- paste0('"', coords, '"')
    
    tryCatch({
      DBI::dbExecute(target_conn, glue::glue("SELECT {paste0(coords_quoted, collapse = ', ')} FROM {x_list$query_name} WHERE 1=0"))
    }, error = function(e) {
      cli::cli_abort("Coordinate columns {.val {coords}} not found in input.")
    })
    
    if (na.fail) {
      null_count <- DBI::dbGetQuery(
        target_conn, 
        glue::glue("SELECT COUNT(*) as n FROM {x_list$query_name} WHERE {paste0(coords_quoted, ' IS NULL', collapse = ' OR ')}")
      )$n
      if (null_count > 0) {
        cli::cli_abort("Missing values found in coordinate columns {.val {coords}}. Set {.code na.fail = FALSE} to ignore.")
      }
    }
    
    coords_str <- paste0(coords_quoted, collapse = ", ")
    st_function <- glue::glue("ST_Point({coords_str})")
    select_clause <- if (remove) glue::glue("SELECT * EXCLUDE ({coords_str}),") else "SELECT *,"
    
    query <- glue::glue("
      {select_clause}
      {build_geom_query(st_function, name = NULL, crs = crs, mode = 'duckspatial')} as {target_geom}
      FROM {x_list$query_name}
    ")
    
    return(ddbs_handle_query(
      query = query,
      conn = target_conn,
      mode = "duckspatial",
      crs = crs,
      x_geom = target_geom
    ))
  }
  
  if (!is.null(wkt)) {
    # WKT internal logic
    assert_character_scalar(wkt, "wkt")
    
    # 2. Validation
    wkt_quoted <- paste0('"', wkt, '"')
    tryCatch({
      DBI::dbExecute(target_conn, glue::glue("SELECT {wkt_quoted} FROM {x_list$query_name} WHERE 1=0"))
    }, error = function(e) {
      cli::cli_abort("WKT column {.val {wkt}} not found in input.")
    })
    
    if (na.fail) {
      null_count <- DBI::dbGetQuery(
        target_conn, 
        glue::glue("SELECT COUNT(*) as n FROM {x_list$query_name} WHERE {wkt_quoted} IS NULL")
      )$n
      if (null_count > 0) {
        cli::cli_abort("Missing values found in WKT column {.val {wkt}}. Set {.code na.fail = FALSE} to ignore.")
      }
    }
    
    # 3. Build Query
    st_function <- glue::glue("ST_GeomFromText({wkt_quoted})")
    select_clause <- if (remove) {
      glue::glue("SELECT * EXCLUDE ({wkt_quoted}),")
    } else {
      "SELECT *,"
    }
    
    query <- glue::glue("
      {select_clause}
      {build_geom_query(st_function, name = NULL, crs = crs, mode = 'duckspatial')} as {target_geom}
      FROM {x_list$query_name}
    ")
    
    return(ddbs_handle_query(
      query = query,
      conn = target_conn,
      mode = "duckspatial",
      crs = crs,
      x_geom = target_geom
    ))
  }
}

# skip tests on CRAN
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
skip_if_not_installed("sf")

# 0. Test ddbs_temp_conn helper ----
test_that("ddbs_temp_conn creates valid auto-closing connection", {
  # Test that connection is valid
  conn <- ddbs_temp_conn()
  expect_true(DBI::dbIsValid(conn))
  
  # Test that connection works
  result <- DBI::dbGetQuery(conn, "SELECT 1 AS test")
  expect_equal(result$test, 1)
})

test_that("ddbs_temp_conn auto-closes on function exit", {
  # Create a helper function that uses ddbs_temp_conn
  get_conn_status <- function() {
    conn <- ddbs_temp_conn()
    # Return the connection object so we can check it after func exits
    conn
  }
  
  # Call the function - connection should be closed after it returns
  returned_conn <- get_conn_status()
  
  # Connection should be invalid (closed) after function returned
  expect_false(DBI::dbIsValid(returned_conn))
})

# create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

# 1. Basic functionality with sf objects ----
test_that("ddbs_write_table writes an sf object to a new table", {
  # Write the points_sf data to a new table
  table_name <- "test_points_write"
  expect_true(ddbs_write_table(conn_test, points_sf, table_name, overwrite = TRUE))

  # Verify the table exists
  all_tables <- DBI::dbListTables(conn_test)
  expect_true(table_name %in% all_tables)

  # Verify the content
  result <- ddbs_read_table(conn_test, table_name)
  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(points_sf))
  expect_equal(sf::st_crs(result), sf::st_crs(points_sf))
})

test_that("ddbs_write_table respects the overwrite argument", {
  table_name <- "test_overwrite"
  # Write initial data
  ddbs_write_table(conn_test, points_sf, table_name, overwrite = TRUE)

  # Expect an error when trying to overwrite without permission
  expect_error(
    ddbs_write_table(conn_test, points_sf, table_name, overwrite = FALSE),
    "already present"
  )

  # Overwrite with new data (countries_sf)
  expect_true(ddbs_write_table(conn_test, countries_sf, table_name, overwrite = TRUE))

  # Verify the new data is in the table
  result <- ddbs_read_table(conn_test, table_name)
  expect_equal(nrow(result), nrow(countries_sf))
})

test_that("ddbs_write_table handles different geometry types", {
  # Create some test data with different geometry types
  line <- sf::st_as_sfc("LINESTRING(0 0, 1 1)") |>
    sf::st_sf(id = 1, geom = _)
  polygon <- sf::st_as_sfc("POLYGON((0 0, 1 0, 1 1, 0 1, 0 0))") |>
    sf::st_sf(id = 1, geom = _)

  # Write LINESTRING
  table_line <- "test_line"
  ddbs_write_table(conn_test, line, table_line, overwrite = TRUE)
  result_line <- ddbs_read_table(conn_test, table_line)
  expect_s3_class(result_line, "sf")
  expect_equal(sf::st_geometry_type(result_line) |> as.character(), "LINESTRING")

  # Write POLYGON
  table_polygon <- "test_polygon"
  ddbs_write_table(conn_test, polygon, table_polygon, overwrite = TRUE)
  result_polygon <- ddbs_read_table(conn_test, table_polygon)
  expect_s3_class(result_polygon, "sf")
  expect_equal(sf::st_geometry_type(result_polygon) |> as.character(), "POLYGON")
})

test_that("ddbs_write_table stores CRS information correctly", {
  table_name <- "test_crs_storage"
  ddbs_write_table(conn_test, points_sf, table_name, overwrite = TRUE)

  # Query the crs
  crs_info <- DBI::dbGetQuery(
    conn_test, 
    glue::glue("SELECT ST_CRS(geometry) FROM {table_name} LIMIT 1")
  ) |> as.character()
  expect_equal(crs_info, "EPSG:4326")
})

# 2. Writing from file paths ----
test_that("ddbs_write_table can write from a .geojson file path", {
  file_path <- system.file("spatial/countries.geojson", package = "duckspatial")
  table_name <- "countries_from_file_write"

  expect_true(ddbs_write_table(conn_test, file_path, table_name, overwrite = TRUE))

  # Verify table exists and has content
  all_tables <- DBI::dbListTables(conn_test)
  expect_true(table_name %in% all_tables)
  count <- DBI::dbGetQuery(conn_test, glue::glue("SELECT COUNT(*) FROM {table_name}"))
  expect_gt(count[[1]], 0)

  # Verify CRS is stored
  crs_info <- DBI::dbGetQuery(
    conn_test, 
    glue::glue("SELECT ST_CRS(geom) FROM {table_name} LIMIT 1")
  ) |> as.character()
  expect_true(!is.na(crs_info) && nchar(crs_info) > 0)
})

# 3. Edge cases and errors ----
test_that("ddbs_write_table handles sf objects with no CRS", {
  points_no_crs <- points_sf
  sf::st_crs(points_no_crs) <- NA
  table_name <- "test_no_crs"

  ddbs_write_table(conn_test, points_no_crs, table_name, overwrite = TRUE)

  # Check that the crs_duckspatial column does NOT exist
  # Because we skip creating it when CRS is missing
  crs <- DBI::dbGetQuery(
    conn_test,
    "SELECT ST_CRS(geometry) as crs FROM test_no_crs LIMIT 1;"
  )$crs
  expect_true(is.na(crs))
})

test_that("ddbs_write_table respects temp_view = TRUE", {
  table_name <- "test_temp_view_write"
  
  # Write with temp_view = TRUE
  expect_true(ddbs_write_table(conn_test, points_sf, table_name, temp_view = TRUE, overwrite = TRUE))
  
  # The requested name should be queryable through the typed temp view
  all_tables <- DBI::dbListTables(conn_test)
  expect_true(table_name %in% all_tables)

  # The raw Arrow view is hidden behind the typed SQL view
  arrow_views <- duckdb::duckdb_list_arrow(conn_test)
  expect_true(paste0("__raw_", table_name) %in% arrow_views)
  
  # Should be readable
  result <- ddbs_read_table(conn_test, table_name)
  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(points_sf))
})

# Disconnect
duckdb::dbDisconnect(conn_test)

# 4. New functionality: duckspatial_df and existing CRS columns ----
test_that("ddbs_write_table can write a duckspatial_df object", {
  conn_new <- ddbs_temp_conn()
  
  # Create a duckspatial_df
  ddbs_write_table(conn_new, points_sf, "points_source", overwrite = TRUE)
  df_lazy <- ddbs_read_table(conn_new, "points_source") |>
    as_duckspatial_df()
  
  table_name <- "points_from_lazy"
  
  # This should now work (previously failed with "Expected string vector of length 1")
  expect_true(ddbs_write_table(conn_new, df_lazy, table_name, overwrite = TRUE))
  
  # Verify result
  result <- ddbs_read_table(conn_new, table_name)
  expect_equal(nrow(result), nrow(points_sf))
  expect_equal(sf::st_crs(result), sf::st_crs(points_sf))
})


test_that("ddbs_write_table throws error for unsupported input", {
  conn_new <- ddbs_temp_conn()
  
  expect_error(
    ddbs_write_table(conn_new, list(a = 1), "bad_input"),
    "must be an .*sf.* object"
  )
})

test_that("ddbs_write_table with temp_view=TRUE works for duckspatial_df", {
  conn_new <- ddbs_temp_conn()
  
  # Create a duckspatial_df first
  ddbs_write_table(conn_new, points_sf, "points_src", overwrite = TRUE)
  df_lazy <- ddbs_read_table(conn_new, "points_src") |>
    as_duckspatial_df()
  
  view_name <- "points_lazy_view"
  
  # This should work with temp_view = TRUE
  expect_true(ddbs_write_table(conn_new, df_lazy, view_name, temp_view = TRUE, overwrite = TRUE))
  
  # Verify the typed view was created and backed by a hidden raw Arrow view
  all_tables <- DBI::dbListTables(conn_new)
  expect_true(view_name %in% all_tables)

  arrow_views <- duckdb::duckdb_list_arrow(conn_new)
  expect_true(paste0("__raw_", view_name) %in% arrow_views)
  
  # Should be queryable
  result <- ddbs_read_table(conn_new, view_name)
  expect_equal(nrow(result), nrow(points_sf))
})

# =============================================================================
# Tests for duckspatial_df core functionality
# Tests: new_duckspatial_df, as_duckspatial_df.*, is_duckspatial_df
# Note: nc_sf is loaded from setup.R
# =============================================================================
testthat::skip_on_cran()

test_that("new_duckspatial_df creates valid duckspatial_df objects", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  lazy_tbl <- dplyr::tbl(conn, "nc_test")
  
  result <- duckspatial:::new_duckspatial_df(
    lazy_tbl, 
    crs = sf::st_crs(nc_sf), 
    geom_col = "geometry",
    source_table = "nc_test"
  )
  
  expect_s3_class(result, "duckspatial_df")
  expect_s3_class(result, "tbl_lazy")
  expect_equal(attr(result, "sf_column"), "geometry")
  expect_true(duckspatial:::crs_equal(attr(result, "crs"), sf::st_crs(nc_sf)))
  expect_equal(attr(result, "source_table"), "nc_test")
})

test_that("new_duckspatial_df avoids double wrapping", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  lazy_tbl <- dplyr::tbl(conn, "nc_test")
  result1 <- duckspatial:::new_duckspatial_df(lazy_tbl, crs = sf::st_crs(nc_sf))
  result2 <- duckspatial:::new_duckspatial_df(result1, crs = sf::st_crs(nc_sf))
  
  expect_identical(result1, result2)
})

test_that("as_duckspatial_df.duckspatial_df can update metadata", {
  conn <- duckspatial:::ddbs_temp_conn()
  ds <- as_duckspatial_df(nc_sf, conn = conn)
  
  ds_new <- as_duckspatial_df(ds, crs = "EPSG:3857")
  expect_equal(sf::st_crs(ds_new), sf::st_crs("EPSG:3857"))
  
  ds_new2 <- as_duckspatial_df(ds, geom_col = "new_geom")
  expect_equal(attr(ds_new2, "sf_column"), "new_geom")
  
  expect_identical(ds, as_duckspatial_df(ds))
})

test_that("as_duckspatial_df.sf works correctly", {
  conn <- duckspatial:::ddbs_temp_conn()
  
  result <- as_duckspatial_df(nc_sf, conn = conn)
  
  expect_s3_class(result, "duckspatial_df")
  expect_s3_class(result, "tbl_lazy")
  expect_true(duckspatial:::crs_equal(attr(result, "crs"), sf::st_crs(nc_sf)))
  expect_equal(attr(result, "sf_column"), attr(nc_sf, "sf_column"))
})

test_that("as_duckspatial_df.tbl_duckdb_connection works correctly", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  lazy_tbl <- dplyr::tbl(conn, "nc_test")
  # Auto-detection should work now!
  result <- as_duckspatial_df(lazy_tbl)
  
  expect_s3_class(result, "duckspatial_df")
  expect_true(duckspatial:::crs_equal(attr(result, "crs"), sf::st_crs(nc_sf)))
})

test_that("as_duckspatial_df.tbl_lazy works correctly", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  lazy_tbl <- dplyr::tbl(conn, "nc_test") |> dplyr::filter(AREA > 0)
  result <- as_duckspatial_df(lazy_tbl, crs = sf::st_crs(nc_sf))  
  
  expect_s3_class(result, "duckspatial_df")
  expect_true(duckspatial:::crs_equal(attr(result, "crs"), sf::st_crs(nc_sf)))
})

test_that("as_duckspatial_df.character works correctly", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  result <- as_duckspatial_df("nc_test", conn = conn, crs = sf::st_crs(nc_sf))
  
  expect_s3_class(result, "duckspatial_df")
  # expect_equal(attr(result, "source_table"), "nc_test")
})

test_that("as_duckspatial_df.character requires connection", {
  expect_error(
    as_duckspatial_df("some_table"),
    "conn|table|Table.*does not exist"
  )
})

test_that("as_duckspatial_df.data.frame works with sfc columns", {
  conn <- duckspatial:::ddbs_temp_conn()
  
  df <- data.frame(id = 1, val = "a")
  df$geom <- sf::st_sfc(sf::st_point(c(0, 0)), crs = 4326)
  
  result <- as_duckspatial_df(df, conn = conn)
  
  expect_s3_class(result, "duckspatial_df")
  expect_equal(sf::st_crs(result), sf::st_crs(4326))
  expect_equal(attr(result, "sf_column"), "geom")
  
  result2 <- as_duckspatial_df(df, conn = conn, geom_col = "geom")
  expect_equal(attr(result2, "sf_column"), "geom")
  
  df_no_geom <- data.frame(id = 1)
  expect_error(as_duckspatial_df(df_no_geom, conn = conn), "sfc")
  
  df_wrong_geom_multi <- data.frame(id = 1, not_geom = 2)
  df_wrong_geom_multi$real_geom <- sf::st_sfc(sf::st_point(c(0, 0)))
  expect_error(
    as_duckspatial_df(df_wrong_geom_multi, conn = conn, geom_col = "not_geom"), 
    "sfc"
  )
})

test_that("as_duckspatial_df handles coords ingestion across inputs", {
  conn <- duckspatial:::ddbs_temp_conn()
  df <- data.frame(id = 1:3, lon = 1:3, lat = 4:6)
  
  # 1. Data frame
  ds_df <- as_duckspatial_df(df, coords = c("lon", "lat"), conn = conn, crs = 4326)
  expect_s3_class(ds_df, "duckspatial_df")
  expect_equal(sf::st_crs(ds_df), sf::st_crs(4326))
  expect_false("lon" %in% colnames(ds_df))
  
  # 2. Character (table name)
  DBI::dbWriteTable(conn, "points_tbl", df, overwrite = TRUE)
  ds_char <- as_duckspatial_df("points_tbl", coords = c("lon", "lat"), conn = conn, crs = 3857)
  expect_s3_class(ds_char, "duckspatial_df")
  expect_equal(sf::st_crs(ds_char), sf::st_crs(3857))
  
  # 3. Lazy table
  lazy_tbl <- dplyr::tbl(conn, "points_tbl")
  ds_lazy <- as_duckspatial_df(lazy_tbl, coords = c("lon", "lat"), crs = 4326)
  expect_s3_class(ds_lazy, "duckspatial_df")
  
  # na.fail = TRUE (default)
  df_na <- df
  df_na$lon[1] <- NA
  expect_error(as_duckspatial_df(df_na, coords = c("lon", "lat"), conn = conn, crs = 4326))
  
  # remove = FALSE
  ds_keep <- as_duckspatial_df(df, coords = c("lon", "lat"), remove = FALSE, conn = conn, crs = 4326)
  expect_true("lon" %in% colnames(ds_keep))
})

test_that("as_duckspatial_df handles WKT ingestion across inputs", {
  conn <- duckspatial:::ddbs_temp_conn()
  df <- data.frame(id = 1:3, wkt_col = c("POINT(0 0)", "POINT(1 1)", "POINT(2 2)"))
  
  # 1. Data frame
  ds_df <- as_duckspatial_df(df, wkt = "wkt_col", conn = conn, crs = 4326)
  expect_s3_class(ds_df, "duckspatial_df")
  expect_equal(sf::st_crs(ds_df), sf::st_crs(4326))
  expect_false("wkt_col" %in% colnames(ds_df))
  
  # 2. Character (table name)
  DBI::dbWriteTable(conn, "wkt_tbl", df, overwrite = TRUE)
  ds_char <- as_duckspatial_df("wkt_tbl", wkt = "wkt_col", conn = conn, crs = 3857)
  expect_s3_class(ds_char, "duckspatial_df")
  
  # 3. Lazy table
  lazy_tbl <- dplyr::tbl(conn, "wkt_tbl")
  ds_lazy <- as_duckspatial_df(lazy_tbl, wkt = "wkt_col", crs = 4326)
  expect_s3_class(ds_lazy, "duckspatial_df")
  
  # na.fail = TRUE (default)
  df_na <- df
  df_na$wkt_col[1] <- NA
  expect_error(as_duckspatial_df(df_na, wkt = "wkt_col", conn = conn, crs = 4326))
  
  # remove = FALSE
  ds_keep <- as_duckspatial_df(df, wkt = "wkt_col", remove = FALSE, conn = conn, crs = 4326)
  expect_true("wkt_col" %in% colnames(ds_keep))
})

test_that("as_duckspatial_df ingestion respects geom_col", {
  conn <- duckspatial:::ddbs_temp_conn()
  df <- data.frame(id = 1, x = 0, y = 0)
  
  ds <- as_duckspatial_df(df, coords = c("x", "y"), geom_col = "my_geom", conn = conn, crs = 4326)
  expect_equal(attr(ds, "sf_column"), "my_geom")
  
  # Collect and check column names
  res_sf <- ddbs_collect(ds)
  expect_true("my_geom" %in% colnames(res_sf))
})

test_that("is_duckspatial_df works correctly", {
  conn <- duckspatial:::ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # lazy_tbl <- dplyr::tbl(conn, "nc_test")
  # result <- as_duckspatial_df(lazy_tbl, crs = sf::st_crs(nc_sf))
  result <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  expect_true(is_duckspatial_df(result))
  # expect_false(is_duckspatial_df(lazy_tbl))
  expect_false(is_duckspatial_df(nc_sf))
  expect_false(is_duckspatial_df(NULL))
})

# =============================================================================
# Tests for duckspatial_df sf methods
# Tests: st_crs, st_bbox, st_geometry, st_as_sf, print, ddbs_geom_col
# Note: nc_sf is loaded from setup.R
# =============================================================================
testthat::skip_on_cran()

test_that("st_crs.duckspatial_df returns correct CRS", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  expect_equal(sf::st_crs(nc_lazy), sf::st_crs(nc_sf))
})

test_that("st_bbox.duckspatial_df works correctly", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  bbox <- sf::st_bbox(nc_lazy)
  expect_s3_class(bbox, "bbox")
  
  bbox_orig <- sf::st_bbox(nc_sf)
  expect_equal(as.numeric(bbox), as.numeric(bbox_orig))
})

test_that("st_geometry.duckspatial_df works correctly", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  geom <- sf::st_geometry(nc_lazy)
  
  expect_s3_class(geom, "sfc")
  expect_equal(length(geom), nrow(nc_sf))
  expect_equal(sf::st_crs(geom), sf::st_crs(nc_sf))
})

test_that("st_as_sf.duckspatial_df works correctly", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  result_sf <- sf::st_as_sf(nc_lazy)
  
  expect_s3_class(result_sf, "sf")
  expect_s3_class(result_sf, "data.frame")
  expect_equal(nrow(result_sf), nrow(nc_sf))
})

test_that("print.duckspatial_df shows informative output", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  output <- capture.output(print(nc_lazy))
  
  expect_true(any(grepl("duckspatial lazy spatial table", output)))
  expect_true(any(grepl("CRS:", output)))
  expect_true(any(grepl("Geometry column:", output)))
})

test_that("ddbs_geom_col returns correct geometry column name", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf))
  
  expect_equal(ddbs_geom_col(nc_lazy), "geometry")
  expect_equal(ddbs_geom_col(nc_sf), attr(nc_sf, "sf_column"))
})

## Coordinates:
## - ddbs_x, ddbs_y, ddbs_z, ddbs_m (point coordinates)
## - ddbs_xmax, ddbs_xmin, ddbs_ymax, ddbs_ymin,
##   ddbs_zmax, ddbs_zmin, ddbs_mmax, ddbs_mmin (coordinate bounds)


#' Extract coordinates from geometries
#'
#' Extracts the X, Y, M, or Z coordinates from `POINT` geometries
#'
#' @template x
#' @param new_column Name of the new column to store the extracted coordinate.
#'   Defaults to `"X"`, `"Y"`, `"M"`, or `"Z"`.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' 
#' @details
#' - `ddbs_x()`: Extracts the X coordinate (longitude).
#' - `ddbs_y()`: Extracts the Y coordinate (latitude).
#' - `ddbs_m()`: Extracts the M coordinate (measure).
#' - `ddbs_z()`: Extracts the Z coordinate (elevation). 
#' 
#' @name ddbs_xy
#' @rdname ddbs_xy
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## extract coordinates
#' ddbs_x(argentina_ddbs)
#' ddbs_y(argentina_ddbs)
#' }
NULL




#' @rdname ddbs_xy
#' @export
ddbs_x <- function(
  x,
  new_column = "X",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_X"
  )
  
}



#' @rdname ddbs_xy
#' @export
ddbs_y <- function(
  x,
  new_column = "Y",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Y"
  )
  
}



#' @rdname ddbs_xy
#' @export
ddbs_m <- function(
  x,
  new_column = "M",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_M"
  )
  
}


#' @rdname ddbs_xy
#' @export
ddbs_z <- function(
  x,
  new_column = "Z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Z"
  )

}




#' Coordinate bounds of geometries
#'
#' Returns the minimum or maximum value of a specific coordinate axis across
#' all points of a geometry. When \code{by_feature = TRUE} (default), a value
#' is computed per row. When \code{by_feature = FALSE}, a single global
#' value is returned for the entire dataset.
#'
#' @template x
#' @param new_column Name of the new column. Defaults to the lowercase function
#'   name (e.g. \code{"xmax"}, \code{"xmin"}, \code{"ymax"}, \ldots).
#' @template by_feature
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' - `ddbs_xmax()` / `ddbs_xmin()`: maximum / minimum X coordinate.
#' - `ddbs_ymax()` / `ddbs_ymin()`: maximum / minimum Y coordinate.
#' - `ddbs_zmax()` / `ddbs_zmin()`: maximum / minimum Z coordinate.
#' - `ddbs_mmax()` / `ddbs_mmin()`: maximum / minimum M (measure) coordinate.
#'
#' When \code{by_feature = FALSE}, the result is always a single \code{numeric}
#' scalar representing the global extreme across the entire dataset.
#'
#' @returns
#' \itemize{
#'   \item \code{by_feature = TRUE} and \code{mode = "duckspatial"} (default):
#'     A \code{duckspatial_df}.
#'   \item \code{by_feature = TRUE} and \code{mode = "sf"}: A numeric vector.
#'   \item \code{by_feature = FALSE}: A single \code{numeric} scalar.
#'   \item When \code{name} is provided: writes the table in DuckDB and
#'     returns \code{TRUE} (invisibly).
#' }
#'
#' @name ddbs_coord_bounds
#' @rdname ddbs_coord_bounds
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", package = "duckspatial")
#' )
#'
#' ## per-feature X extent (default)
#' ddbs_xmax(argentina_ddbs)
#' ddbs_xmin(argentina_ddbs)
#'
#' ## global bounding values
#' ddbs_xmax(argentina_ddbs, by_feature = FALSE)
#' ddbs_ymin(argentina_ddbs, by_feature = FALSE)
#' }
NULL


#' @rdname ddbs_coord_bounds
#' @export
ddbs_xmax <- function(
    x,
    new_column = "xmax",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_XMax"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_xmin <- function(
    x,
    new_column = "xmin",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_XMin"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_ymax <- function(
    x,
    new_column = "ymax",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_YMax"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_ymin <- function(
    x,
    new_column = "ymin",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_YMin"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_zmax <- function(
    x,
    new_column = "zmax",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_ZMax"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_zmin <- function(
    x,
    new_column = "zmin",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_ZMin"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_mmax <- function(
    x,
    new_column = "mmax",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_MMax"
  )
}


#' @rdname ddbs_coord_bounds
#' @export
ddbs_mmin <- function(
    x,
    new_column = "mmin",
    by_feature = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_MMin"
  )
}

# 0. Set up ------------------------------------------------------------------

skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

conn_test <- duckspatial::ddbs_create_conn()
ddbs_write_table(conn_test, argentina_sf, "argentina")
ddbs_write_table(conn_test, points_sf,   "points")


# 1. ddbs_xmax() -------------------------------------------------------------

## - CHECK 1.1: works on all input types with by_feature = TRUE
## - CHECK 1.2: returns numeric vector with mode = "sf"
## - CHECK 1.3: returns single scalar with by_feature = FALSE
## - CHECK 1.4: messages shown/suppressed correctly
## - CHECK 1.5: writes table to database
## - CHECK 1.6: by_feature = FALSE equals max of by_feature = TRUE vector
## - CHECK 1.7: xmax >= xmin (sanity)
## - CHECK 2.1: standard errors
describe("ddbs_xmax()", {

  describe("expected behavior", {

    it("works on all input types with by_feature = TRUE", {
      out_ddbs <- ddbs_xmax(argentina_ddbs)
      out_sf   <- ddbs_xmax(argentina_sf)
      out_conn <- ddbs_xmax("argentina", conn = conn_test)

      expect_s3_class(out_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(out_ddbs), ddbs_collect(out_sf))
      expect_equal(ddbs_collect(out_ddbs), ddbs_collect(out_conn))
    })

    it("returns a numeric vector with mode = 'sf'", {
      out <- ddbs_xmax(argentina_ddbs, mode = "sf")
      expect_true(is.numeric(out))
    })

    it("returns a single numeric scalar with by_feature = FALSE", {
      out_ddbs <- ddbs_xmax(argentina_ddbs, by_feature = FALSE)
      out_sf   <- ddbs_xmax(argentina_sf,   by_feature = FALSE)
      out_conn <- ddbs_xmax("argentina", conn = conn_test, by_feature = FALSE)

      expect_length(out_ddbs, 1)
      expect_true(is.numeric(out_ddbs))
      expect_equal(out_ddbs, out_sf)
      expect_equal(out_ddbs, out_conn)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_xmax(argentina_ddbs))
      expect_message(ddbs_xmax("argentina", conn = conn_test, name = "xmax_tbl"))
      expect_no_message(ddbs_xmax(argentina_ddbs, quiet = TRUE))
      expect_no_message(
        ddbs_xmax("argentina", conn = conn_test, name = "xmax_tbl_q", quiet = TRUE)
      )
    })

    it("writes table to database", {
      out <- ddbs_xmax("argentina", conn = conn_test, name = "xmax_tbl2", new_column = "x_hi")
      expect_true(out)
    })

    it("by_feature = FALSE equals max of by_feature = TRUE vector", {
      global_max   <- ddbs_xmax(argentina_ddbs, by_feature = FALSE)
      per_feat_max <- ddbs_xmax(argentina_ddbs, mode = "sf")
      expect_equal(global_max, max(per_feat_max))
    })

    it("xmax >= xmin for the same dataset", {
      expect_gte(
        ddbs_xmax(argentina_ddbs, by_feature = FALSE),
        ddbs_xmin(argentina_ddbs, by_feature = FALSE)
      )
    })
  })

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_xmax("argentina", conn = conn_test, name = "dup_xmax")
      expect_error(ddbs_xmax("argentina", conn = conn_test, name = "dup_xmax"))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_xmax("argentina", conn = NULL))
    })

    it("errors on invalid x type", {
      expect_error(ddbs_xmax(999))
      expect_error(ddbs_xmax(TRUE))
    })
  })
})


# 2. ddbs_xmin() -------------------------------------------------------------

## - CHECK 1.1: works on all input types
## - CHECK 1.2: returns scalar with by_feature = FALSE
## - CHECK 1.3: by_feature = FALSE equals min of by_feature = TRUE
describe("ddbs_xmin()", {

  describe("expected behavior", {

    it("works on all input types with by_feature = TRUE", {
      out_ddbs <- ddbs_xmin(argentina_ddbs)
      out_sf   <- ddbs_xmin(argentina_sf)
      out_conn <- ddbs_xmin("argentina", conn = conn_test)

      expect_s3_class(out_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(out_ddbs), ddbs_collect(out_sf))
      expect_equal(ddbs_collect(out_ddbs), ddbs_collect(out_conn))
    })

    it("returns a single numeric scalar with by_feature = FALSE", {
      out <- ddbs_xmin(argentina_ddbs, by_feature = FALSE)
      expect_length(out, 1)
      expect_true(is.numeric(out))
    })

    it("by_feature = FALSE equals min of by_feature = TRUE vector", {
      global_min   <- ddbs_xmin(argentina_ddbs, by_feature = FALSE)
      per_feat_min <- ddbs_xmin(argentina_ddbs, mode = "sf")
      expect_equal(global_min, min(per_feat_min))
    })
  })
})


# 3. ddbs_ymax() / ddbs_ymin() -----------------------------------------------

## - CHECK 1.1: return duckspatial_df by default
## - CHECK 1.2: return scalar with by_feature = FALSE, ymax >= ymin
## - CHECK 1.3: by_feature = FALSE equals max/min of by_feature = TRUE
describe("ddbs_ymax() and ddbs_ymin()", {

  it("return duckspatial_df by default", {
    expect_s3_class(ddbs_ymax(argentina_ddbs), "duckspatial_df")
    expect_s3_class(ddbs_ymin(argentina_ddbs), "duckspatial_df")
  })

  it("return a single numeric scalar with by_feature = FALSE, ymax >= ymin", {
    out_max <- ddbs_ymax(argentina_ddbs, by_feature = FALSE)
    out_min <- ddbs_ymin(argentina_ddbs, by_feature = FALSE)
    expect_length(out_max, 1)
    expect_length(out_min, 1)
    expect_true(is.numeric(out_max))
    expect_gte(out_max, out_min)
  })

  it("by_feature = FALSE equals max/min of by_feature = TRUE vector", {
    expect_equal(
      ddbs_ymax(argentina_ddbs, by_feature = FALSE),
      max(ddbs_ymax(argentina_ddbs, mode = "sf"))
    )
    expect_equal(
      ddbs_ymin(argentina_ddbs, by_feature = FALSE),
      min(ddbs_ymin(argentina_ddbs, mode = "sf"))
    )
  })
})


# 4. ddbs_zmax() / ddbs_zmin() / ddbs_mmax() / ddbs_mmin() ------------------

## - CHECK 1.1: all return duckspatial_df by default
## - CHECK 1.2: all return a single numeric scalar with by_feature = FALSE
describe("ddbs_zmax(), ddbs_zmin(), ddbs_mmax(), ddbs_mmin()", {

  it("all return duckspatial_df by default", {
    expect_s3_class(ddbs_zmax(argentina_ddbs), "duckspatial_df")
    expect_s3_class(ddbs_zmin(argentina_ddbs), "duckspatial_df")
    expect_s3_class(ddbs_mmax(argentina_ddbs), "duckspatial_df")
    expect_s3_class(ddbs_mmin(argentina_ddbs), "duckspatial_df")
  })

  it("all return a single numeric scalar with by_feature = FALSE", {
    expect_length(ddbs_zmax(argentina_ddbs, by_feature = FALSE), 1)
    expect_length(ddbs_zmin(argentina_ddbs, by_feature = FALSE), 1)
    expect_length(ddbs_mmax(argentina_ddbs, by_feature = FALSE), 1)
    expect_length(ddbs_mmin(argentina_ddbs, by_feature = FALSE), 1)
    expect_true(is.numeric(ddbs_zmax(argentina_ddbs, by_feature = FALSE)))
    expect_true(is.numeric(ddbs_mmin(argentina_ddbs, by_feature = FALSE)))
  })
})

#' Union and combine geometries
#'
#' @description
#' Perform union and combine operations on spatial geometries in DuckDB.
#'
#' * `ddbs_union()` - Union all geometries into one, or perform pairwise union between two datasets
#' * `ddbs_union_agg()` - Union geometries grouped by one or more columns
#' * `ddbs_combine()` - Combine geometries into a MULTI-geometry without dissolving boundaries
#'
#' @template x
#' @param y Input spatial data. Can be:
#'   \itemize{
#'    \item \code{NULL} (default): performs only the union of `x`
#'     \item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
#'     \item An \code{sf} object
#'     \item A \code{tbl_lazy} from dbplyr
#'     \item A character string naming a table/view in \code{conn}
#'   }
#' @param by_feature Logical. When `y` is provided:
#'   * `FALSE` (default) - Union all geometries from both `x` and `y` into a single geometry
#'   * `TRUE` - Perform row-by-row union between matching features from `x` and `y` (requires same number of rows)
#' @param by Character vector specifying one or more column names to
#' group by when computing unions. Geometries will be unioned within each group.
#' Default is \code{NULL}
#' @param mem Logical. If \code{TRUE}, uses \code{ST_MemUnion_Agg()} instead of
#' \code{ST_Union_Agg()} — slower but more memory efficient. Default is
#' \code{FALSE}. Only applies to \code{ddbs_union_agg()}.
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' ## ddbs_union(x, y, by_feature)
#' Performs geometric union operations that dissolve internal boundaries:
#' * When `y = NULL`: Unions all geometries in `x` into a single geometry
#' * When `y != NULL` and `by_feature = FALSE`: Unions all geometries from both `x` and `y` into a single geometry
#' * When `y != NULL` and `by_feature = TRUE`: Performs row-wise union, pairing the first geometry from `x` with the first from `y`, second with second, etc.
#'
#' ## ddbs_union_agg(x, by)
#' Groups geometries by one or more columns, then unions geometries within each group.
#' Useful for dissolving boundaries between features that share common attributes.
#' Set \code{mem = TRUE} to use \code{ST_MemUnion_Agg()} when memory is a constraint.
#'
#' ## ddbs_combine(x)
#' Combines all geometries into a single MULTI-geometry (e.g., MULTIPOLYGON, MULTILINESTRING)
#' without dissolving shared boundaries. This is faster than union but preserves all
#' original geometry boundaries.
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(dplyr)
#' library(duckspatial)
#' 
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |> 
#'   filter(ISO3_CODE != "ATA")
#' 
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson", 
#'   package = "duckspatial")
#' ) |> 
#'  ddbs_transform("EPSG:4326")
#' 
#' ## combine countries into a single MULTI-geometry
#' ## (without solving boundaries)
#' combined_countries_ddbs <- ddbs_combine(countries_ddbs)
#' 
#' ## combine countries into a single MULTI-geometry
#' ## (solving boundaries)
#' union_countries_ddbs <- ddbs_union(countries_ddbs)
#' 
#' ## union of geometries of two objects, into 1 geometry
#' union_countries_rivers_ddbs <- ddbs_union(countries_ddbs, rivers_ddbs)
#' }
#'
#' @name ddbs_union_funs
#' @rdname ddbs_union_funs
NULL



#' @rdname ddbs_union_funs
#' @export
ddbs_union <- function(
  x,
  y           = NULL,
  by_feature  = FALSE,
  conn        = NULL,
  conn_x      = NULL,
  conn_y      = NULL,
  name        = NULL,
  mode        = NULL,
  overwrite   = FALSE,
  quiet       = FALSE) {

  ## 0. Validate inputs
  assert_xy(x, "x")
  assert_logic(by_feature, "by_feature")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  if (isTRUE(by_feature) && is.null(y)) {
    cli::cli_warn("When {.arg y} is NULL, {.arg by_feature = TRUE} is ignored.")
  }

  ## Pre-extract `x` attributes (CRS and geometry column name)
  crs_x    <- if (is.null(conn_x)) ddbs_crs(x, conn) else ddbs_crs(x, conn_x)
  sf_col_x <- attr(x, "sf_column")


  # ------------------------------------------------------------------
  # 1. Pairwise union: ST_Union(x, y)
  # ------------------------------------------------------------------
  if (!is.null(y)) {

    ## Validate y
    assert_xy(y, "y")
    assert_conn_character(conn, y)

    ## Pre-extract `y` attributes
    crs_y    <- if (is.null(conn_y)) ddbs_crs(y, conn) else ddbs_crs(y, conn_y)
    sf_col_y <- attr(y, "sf_column")

    ## Resolve conn_x/conn_y defaults from conn for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## Normalize inputs
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## Resolve connections
    resolve_conn <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    y            <- resolve_conn$y
    ## register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_conn$cleanup(), add = TRUE)   
    }

    ## Get query names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)

    ## CRS check
    if (!is.null(crs_x) && !is.null(crs_y)) {
      if (!crs_equal(crs_x, crs_y)) {
        cli::cli_abort("The Coordinates Reference System of {.arg x} and {.arg y} is different.")
      }
    } else {
      assert_crs(target_conn, x_list$query_name, y_list$query_name)
    }

    ## Geometry column names
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## Get mode - If it's NULL, it will use the duckspatial.mode option
    mode <- get_mode(mode, name)

    ## Named table: write and return
    if (!is.null(name)) {
      name_list <- get_query_name(name)
      overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

      tmp.query <- build_union_query(
        by_feature = by_feature,
        mode       = "duckspatial",
        name       = name,
        crs        = crs_x,
        name_query = name_list$query_name,
        x_geom     = x_geom,
        y_geom     = y_geom,
        x_query    = x_list$query_name,
        y_query    = y_list$query_name
      )

      DBI::dbExecute(target_conn, tmp.query)
      feedback_query(quiet)
      return(invisible(TRUE))
    }

    ## Create the base query
    base.query <- build_union_query(
      by_feature = by_feature,
      name       = name,
      crs        = crs_x,
      mode       = mode,
      name_query = NULL,
      x_geom     = x_geom,
      y_geom     = y_geom,
      x_query    = x_list$query_name,
      y_query    = y_list$query_name
    )

    result <- ddbs_handle_query(
        query  = base.query,
        conn   = target_conn,
        mode   = mode,
        crs    = crs_x,
        x_geom = x_geom
    )

    return(result)

  }


  # ------------------------------------------------------------------
  # 2. Aggregate union: ST_Union(x)
  # ------------------------------------------------------------------

  ## Normalize input
  x <- normalize_spatial_input(x, conn)

  ## Resolve connection
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## Get query name
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  ## Geometry column name
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## Resolve mode
  mode <- get_mode(mode, name)

  ## Named table: write and return
  if (!is.null(name)) {
    name_list <- get_query_name(name)
    overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

    tmp.query <- build_union_query(
      by_feature = FALSE,
      name       = name,
      crs        = crs_x,
      mode       = "duckspatial",
      name_query = name_list$query_name,
      x_geom     = x_geom,
      x_query    = x_list$query_name
    )

    DBI::dbExecute(target_conn, tmp.query)
    feedback_query(quiet)
    return(invisible(TRUE))
  }

  ## Create the base query

  base.query <- build_union_query(
    by_feature = FALSE,
    name       = name,
    crs        = crs_x,
    mode       = mode,
    name_query = NULL,
    x_geom     = x_geom,
    x_query    = x_list$query_name
  )

  result <- ddbs_handle_query(
      query      = base.query,
      conn       = target_conn,
      mode       = mode,
      crs        = crs_x,
      x_geom     = x_geom
  )

  return(result)
  
}




#' @rdname ddbs_union_funs
#' @export
ddbs_combine <- function(
    x,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
  

    # 1. Prepare inputs
  
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)
  
    ## 2.2. Build the base query (depends on the output type - sf, duckspatial_df, table)
    st_function <- glue::glue("ST_Collect(LIST({x_geom}))")
    base.query <- glue::glue("
      SELECT
        {build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}
      FROM
        {x_list$query_name};
    ")

    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
      create_duckdb_table(
        conn      = target_conn,
        name      = name,
        query     = base.query,
        overwrite = overwrite,
        quiet     = quiet
      )
    } else {
      ddbs_handle_query(
        query  = base.query,
        conn   = target_conn,
        mode   = mode,
        crs    = crs_x,
        x_geom = x_geom
      )
    }
  
}



#' @rdname ddbs_union_funs
#' @export
ddbs_union_agg <- function(
  x,
  by,
  mem = FALSE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {


  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(mem, "mem")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Get names of the rest of the groupping columns
  by_cols <- paste0(by, collapse = ", ")

  ## 2.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
  agg_fn      <- if (isTRUE(mem)) "ST_MemUnion_Agg" else "ST_Union_Agg"
  st_function <- glue::glue("{agg_fn}({x_geom})")
  base.query <- glue::glue("
    SELECT 
      {by_cols},
      {build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}
    FROM 
      {x_list$query_name}
    GROUP BY 
      {by_cols};
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }
}





#' Aggregate the intersection of geometries
#'
#' Computes the geometric intersection of a set of geometries — the area common
#' to all of them — using DuckDB's `ST_Intersection_Agg()` aggregate. This is the
#' intersection counterpart to [ddbs_union_agg()].
#'
#' @template x
#' @param by Character vector of one or more column names to group by. The
#'   intersection is computed within each group. When \code{NULL} (default), all
#'   geometries are intersected into a single geometry.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## create a connection and three overlapping polygons
#' conn <- ddbs_create_conn()
#' polys <- sf::st_as_sf(
#'   data.frame(grp = c("a", "a", "a")),
#'   geometry = sf::st_sfc(
#'     sf::st_polygon(list(matrix(c(0,0, 3,0, 3,3, 0,3, 0,0), ncol = 2, byrow = TRUE))),
#'     sf::st_polygon(list(matrix(c(1,1, 4,1, 4,4, 1,4, 1,1), ncol = 2, byrow = TRUE))),
#'     sf::st_polygon(list(matrix(c(2,2, 5,2, 5,5, 2,5, 2,2), ncol = 2, byrow = TRUE))),
#'     sf::st_polygon(list(matrix(c(3,3, 6,3, 6,6, 3,6, 3,3), ncol = 2, byrow = TRUE)))
#'   ),
#'   crs = 4326
#' )
#'
#' ## intersect all geometries into their common area
#' polys_inters <- ddbs_intersection_agg(polys, mode = "sf")
#' 
#' ## intersect within groups
#' polys_inters_grp <- ddbs_intersection_agg(polys, by = "grp", mode = "sf")
#' 
#' ## plot them
#' plot(polys["grp"], key.pos = NULL, reset = FALSE)
#' plot(polys_inters_grp["grp"], pal = c("red", "blue"), add = TRUE)
#' plot(sf::st_geometry(polys_inters), col = "black", pch = 19, cex = 2, add = TRUE)
#' }
ddbs_intersection_agg <- function(
  x,
  by = NULL,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {


  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  if (!is.null(by) && !is.character(by)) {
    cli::cli_abort("{.arg by} must be a character vector of column names or {.code NULL}.")
  }


  # 1. Prepare inputs

  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df,
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the base query. With `by` we group; without it we aggregate the
  ## whole dataset into a single geometry.
  st_function <- glue::glue("ST_Intersection_Agg({x_geom})")
  geom_expr   <- build_geom_query(st_function, name, crs_x, mode)

  if (is.null(by)) {
    base.query <- glue::glue("
      SELECT
        {geom_expr} AS {x_geom}
      FROM
        {x_list$query_name};
    ")
  } else {
    by_cols <- paste0(by, collapse = ", ")
    base.query <- glue::glue("
      SELECT
        {by_cols},
        {geom_expr} AS {x_geom}
      FROM
        {x_list$query_name}
      GROUP BY
        {by_cols};
    ")
  }


  # 3. Table creation if name is provided, or
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }
}





#' Dumps geometries into their component parts
#'
#' Decomposes multi-part or complex geometries into individual simple geometry
#' components, returning one row per component geometry
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## aggregate rivers by name
#' rivers_agg_ddbs <- ddbs_union_agg(rivers_ddbs, by = "RIVER_NAME")
#'
#' ## dump into individual geometries
#' ddbs_dump(rivers_agg_ddbs)
#' }
ddbs_dump <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")

  # Warn if geometry type is not multi
  geom_type <- ddbs_geometry_type(x, by_feature = FALSE, conn)
  if (all(geom_type %in% c("POINT", "LINESTRING", "POLYGON")) & !quiet) {
    cli::cli_warn(c("The geometry type of {.arg x} is {.val {geom_type}}",
      "*" = "{.fun ddbs_dump()} is typically used for multi-part geometries." ,
      "*" = "With simple geometries it doesn't have any effect.")
  )}


  # 1. Prepare inputs

  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)


  ## 2.4. Build the base query (depends on the output type - sf, duckspatial_df, table)
  st_function <- "dump.geom"
  base.query <- glue::glue("
    WITH dumped AS (
    SELECT * EXCLUDE {x_geom},
          UNNEST(ST_Dump({x_geom})) AS dump
    FROM {x_list$query_name}
    )
    SELECT * EXCLUDE dump,
          /* dump.path AS path, */
          {build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}
    FROM dumped;
  ")


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }
  
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## add a grouping column to cuntries
set.seed(123)
countries_group_sf <- countries_sf |> 
  dplyr::mutate(n = sample(1:2, nrow(countries_sf), replace = TRUE)) |> 
  dplyr::mutate(n_2 = sample(c("A", "B"), nrow(countries_sf), replace = TRUE)) 

countries_group_ddbs <- as_duckspatial_df(countries_group_sf)

## write data
duckspatial::ddbs_write_table(conn_test, points_sf, "points")
duckspatial::ddbs_write_table(conn_test, countries_group_sf, "countries")

## two-part MULTIPOLYGON for ddbs_dump tests
two_poly_sf <- sf::st_sf(
  id       = 1L,
  geometry = sf::st_sfc(sf::st_multipolygon(list(
    list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol = 2, byrow = TRUE)),
    list(matrix(c(2,2, 3,2, 3,3, 2,3, 2,2), ncol = 2, byrow = TRUE))
  ))),
  crs = 4326
)
two_poly_ddbs <- as_duckspatial_df(two_poly_sf)
duckspatial::ddbs_write_table(conn_test, two_poly_sf, "two_poly")


# 1. ddbs_union_agg() ----------------------------------------------------

## 1.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 1.5: there must be the same number of rows as the number of groups (2)
## - CHECK 1.6: grouping with more than 1 column
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_union_agg()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_union_agg(countries_group_ddbs, by = "n")
      output_sf   <- ddbs_union_agg(countries_group_sf, by = "n")
      output_conn <- ddbs_union_agg("countries", by = "n", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")

      ## Sometimes they are arranged differently, but the results are
      ## the same when sorted
      expect_equal(
        ddbs_collect(output_ddbs) |> dplyr::arrange(n), 
        ddbs_collect(output_sf) |> dplyr::arrange(n)
      )
      expect_equal(
        ddbs_collect(output_ddbs) |> dplyr::arrange(n), 
        ddbs_collect(output_conn) |> dplyr::arrange(n)
      )
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_union_agg(countries_group_ddbs, "n", mode = NULL)
      output_sf <- ddbs_union_agg(countries_group_ddbs, "n", mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_union_agg(countries_group_ddbs, "n"))
      expect_message(ddbs_union_agg("countries", "n", conn = conn_test, name = "union_agg"))
      expect_message(ddbs_union_agg("countries", "n", conn = conn_test, name = "union_agg", overwrite = TRUE))
      
      expect_no_message(ddbs_union_agg(countries_group_ddbs, "n", quiet = TRUE))
      expect_no_message(ddbs_union_agg("countries", "n", conn = conn_test, name = "union_agg", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output <- ddbs_union_agg("countries", "n", conn = conn_test, name = "union_agg2")
      expect_true(output)
    })
    
    it("written table matches computed output", {
      output_conn <- ddbs_union_agg("countries", by = "n", conn = conn_test)
      output_tbl  <- ddbs_read_table(conn_test, "union_agg")
      
      expect_equal(
        ddbs_collect(output_conn)$geometry,
        output_tbl$geometry
      )
    })
    
    it("returns same number of rows as number of groups", {
      output <- ddbs_union_agg(countries_group_ddbs, by = "n")
      n_rows <- ddbs_collect(output) |> nrow()
      
      expect_equal(n_rows, length(unique(countries_group_sf$n)))
    })
    
    it("supports grouping by multiple columns", {
      output <- ddbs_union_agg(countries_group_ddbs, by = c("n", "n_2"))

      expect_s3_class(output, "duckspatial_df")
    })

    it("mem = TRUE works", {
      output_default <- ddbs_union_agg(countries_group_ddbs, by = "n")
      output_mem     <- ddbs_union_agg(countries_group_ddbs, by = "n", mem = TRUE)

      expect_s3_class(output_mem, "duckspatial_df")
      expect_equal(
        ddbs_collect(output_default) |> dplyr::arrange(n) |> dplyr::pull(n),
        ddbs_collect(output_mem)     |> dplyr::arrange(n) |> dplyr::pull(n)
      )
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_union_agg("countries", conn = NULL))
    })
    
    it("requires by argument", {
      expect_error(ddbs_union_agg(countries_group_ddbs, by = NULL))
    })
    
    it("validates by argument type", {
      expect_error(ddbs_union_agg(countries_group_ddbs, by = 3))
    })
    
    it("validates by argument column existence", {
      expect_error(ddbs_union_agg(countries_group_ddbs, by = "banana"))
      expect_error(ddbs_union_agg(countries_group_ddbs, by = c("n", "banana")))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_union_agg(x = 999))
      expect_error(ddbs_union_agg(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_union_agg(countries_group_ddbs, conn = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_union_agg(countries_group_ddbs, quiet = 999))
    })

    it("validates mem argument type", {
      expect_error(ddbs_union_agg(countries_group_ddbs, mem = 999))
    })
  })
})


# 2. ddbs_union() --------------------------------------------------------

## 2.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats (3 function ways)
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 1.5: check number of rows of the result
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_union()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats with only x argument", {
      output_ddbs <- ddbs_union(countries_ddbs)
      output_sf   <- ddbs_union(countries_sf)
      output_conn <- ddbs_union("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("works on all formats with x and y, by_feature = FALSE", {
      output_ddbs <- ddbs_union(countries_ddbs, countries_sf)
      output_sf   <- ddbs_union(countries_sf, countries_ddbs)
      output_conn <- ddbs_union("countries", countries_sf, conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("works on all formats with x and y, by_feature = TRUE", {
      output_ddbs <- ddbs_union(countries_ddbs, countries_sf, by_feature = TRUE)
      output_sf   <- ddbs_union(countries_sf, countries_ddbs, by_feature = TRUE)
      output_conn <- ddbs_union(countries_sf, "countries", conn = conn_test, by_feature = TRUE)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_union(countries_ddbs, mode = NULL)
      output_sf <- ddbs_union(countries_ddbs, mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_union(countries_ddbs))
      expect_message(ddbs_union("countries", conn = conn_test, name = "union_test"))
      expect_message(ddbs_union("countries", conn = conn_test, name = "union_test", overwrite = TRUE))
      
      expect_no_message(ddbs_union(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_union("countries", conn = conn_test, name = "union_test", overwrite = TRUE, quiet = TRUE))
    })
    
    it("warns when using by_feature = TRUE with only x argument", {
      expect_warning(ddbs_union(countries_ddbs, by_feature = TRUE))
    })
    
    it("writes tables to the database", {
      output <- ddbs_union("countries", conn = conn_test, name = "union_test2")
      expect_true(output)
    })
    
    it("written table matches computed output", {
      output_ddbs <- ddbs_union(countries_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "union_test")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
    
    it("returns 1 row when using only x argument", {
      output <- ddbs_union(countries_ddbs)
      n_rows <- ddbs_collect(output) |> nrow()
      
      expect_equal(n_rows, 1)
    })
    
    it("returns 1 row when using x and y with by_feature = FALSE", {
      output <- ddbs_union(countries_ddbs, countries_sf)
      n_rows <- ddbs_collect(output) |> nrow()
      
      expect_equal(n_rows, 1)
    })
    
    it("returns minimum number of rows between x and y when by_feature = TRUE", {
      output <- ddbs_union(countries_ddbs, countries_sf, by_feature = TRUE)
      n_rows <- ddbs_collect(output) |> nrow()
      
      expect_equal(n_rows, nrow(countries_sf))
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_union("countries", conn = NULL))
    })
    
    it("requires x argument when y is provided", {
      expect_error(ddbs_union(y = countries_ddbs))
    })
    
    it("validates by_feature argument type", {
      expect_error(ddbs_union(countries_ddbs, by_feature = 3))
      expect_error(ddbs_union(countries_ddbs, by_feature = NULL))
      expect_error(ddbs_union(countries_ddbs, by_feature = "banana"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_union(x = 999))
      expect_error(ddbs_union(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_union(countries_ddbs, conn = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_union(countries_ddbs, quiet = 999))
    })
  })
})



# 3. ddbs_combine() -----------------------------------------------------

## 2.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, sf)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting table works
## - CHECK 1:5: always returns 1 row
## - CHECK 2.1: errors
describe("ddbs_combine()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_combine(countries_ddbs)
      output_sf   <- ddbs_combine(countries_sf)
      output_conn <- ddbs_combine("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_combine(countries_ddbs, mode = NULL)
      output_sf <- ddbs_combine(countries_ddbs, mode = "sf")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "sf")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_combine(countries_ddbs))
      expect_message(ddbs_combine("countries", conn = conn_test, name = "combine"))
      expect_message(ddbs_combine("countries", conn = conn_test, name = "combine", overwrite = TRUE))
      
      expect_no_message(ddbs_combine(countries_ddbs, quiet = TRUE))
      expect_no_message(ddbs_combine("countries", conn = conn_test, name = "combine", overwrite = TRUE, quiet = TRUE))
    })
    
    it("writes tables to the database", {
      output <- ddbs_combine("countries", conn = conn_test, name = "combine2")
      expect_true(output)
    })
    
    it("written table matches computed output", {
      output_ddbs <- ddbs_combine(countries_ddbs)
      output_tbl  <- ddbs_read_table(conn_test, "combine")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })
    
    it("always returns 1 row", {
      output <- ddbs_combine(countries_ddbs)
      n_rows <- ddbs_collect(output) |> nrow()
      
      expect_equal(n_rows, 1)
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_combine("countries", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_combine(x = 999))
      expect_error(ddbs_combine(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_combine(countries_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_combine(countries_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_combine(countries_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_combine(countries_ddbs, quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_combine(countries_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


# 4. ddbs_dump() ---------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns different output formats
## - CHECK 1.3: messages work
## - CHECK 1.4: writing a table works
## - CHECK 1.5: decomposes multi-part geometry into individual component rows
## - CHECK 1.6: returns more rows than input for multi-part geometries
## - CHECK 1.7: warns for simple geometries
## - CHECK 2.1: general errors

describe("ddbs_dump()", {

  describe("expected behavior", {

    it("works on all formats and returns matching row counts", {
      output_ddbs <- ddbs_dump(two_poly_ddbs)
      output_sf   <- ddbs_dump(two_poly_sf)
      output_conn <- ddbs_dump("two_poly", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(nrow(ddbs_collect(output_ddbs)), nrow(ddbs_collect(output_sf)))
      expect_equal(nrow(ddbs_collect(output_ddbs)), nrow(ddbs_collect(output_conn)))
    })

    it("returns different output formats", {
      output_sf_fmt <- ddbs_dump(two_poly_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_dump(two_poly_ddbs))
      expect_message(ddbs_dump("two_poly", conn = conn_test, name = "dump"))
      expect_message(ddbs_dump("two_poly", conn = conn_test, name = "dump", overwrite = TRUE))
      expect_true(ddbs_dump("two_poly", conn = conn_test, name = "dump2"))

      expect_no_message(ddbs_dump(two_poly_ddbs, quiet = TRUE))
      expect_no_message(ddbs_dump("two_poly", conn = conn_test, name = "dump", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "dump")
      expect_equal(
        ddbs_collect(ddbs_dump(two_poly_ddbs))$geometry,
        output_tbl$geometry
      )
    })

    it("decomposes a 2-part MULTIPOLYGON into exactly 2 rows", {
      n_output <- nrow(ddbs_collect(ddbs_dump(two_poly_ddbs)))
      expect_equal(n_output, 2L)
    })

    it("returns more rows than the input multi-part geometry", {
      n_input  <- nrow(ddbs_collect(two_poly_ddbs))
      n_output <- nrow(ddbs_collect(ddbs_dump(two_poly_ddbs)))
      expect_gt(n_output, n_input)
    })

    it("warns when input geometry is not multi-part", {
      expect_warning(ddbs_dump(argentina_ddbs))
    })

  })

  describe("errors", {

    it("requires a valid connection when using table name", {
      expect_error(ddbs_dump("two_poly", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_dump(x = 999))
      expect_error(ddbs_dump(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_dump(countries_ddbs, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_dump(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_dump(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_dump(countries_ddbs, conn = conn_test, name = c("a", "b")))
    })

    it("errors on unexpected arguments", {
      expect_error(ddbs_dump(countries_ddbs, new_column = 999))
    })

  })

})


## stop connection
ddbs_stop_conn(conn_test)

#' Convert geometries to standard interchange formats
#'
#' @description
#' Convert spatial geometries to common interchange formats using DuckDB spatial
#' serialization functions.
#'
#' * `ddbs_as_text()` – Convert geometries to Well-Known Text (WKT)
#' * `ddbs_as_wkb()` – Convert geometries to Well-Known Binary (WKB)
#' * `ddbs_as_hexwkb()` – Convert geometries to hexadecimal Well-Known Binary (HEXWKB)
#' * `ddbs_as_geojson()` – Convert to GeoJSON (a `FeatureCollection` or one `Feature` per row)
#'
#' @template x
#' @template conn_null
#' @param feature_collection Logical, only used by `ddbs_as_geojson()`. If `TRUE`
#'   (default), all rows are returned as a single GeoJSON `FeatureCollection`
#'   string. If `FALSE`, a character vector with one `Feature` string per row is
#'   returned.
#'
#' @details
#' These functions are thin wrappers around DuckDB spatial serialization
#' functions (`ST_AsText`, `ST_AsWKB`, `ST_AsHEXWKB`, and `ST_AsGeoJSON`).
#'
#' They are useful for exporting geometries into widely supported formats for
#' interoperability with external spatial tools, databases, and web services.
#'
#' Unlike the other serializers, which only encode the geometry, `ddbs_as_geojson()`
#' produces complete GeoJSON `Feature`s: the geometry is placed in the `geometry`
#' member and all remaining (non-geometry) columns are included as feature
#' `properties`. By default the features are wrapped in a single `FeatureCollection`,
#' the canonical GeoJSON artifact for writing `.geojson` files and feeding web
#' services and mapping libraries.
#'
#' @return
#' Depending on the function:
#' \itemize{
#'   \item \code{ddbs_as_text()} returns a character vector of WKT geometries
#'   \item \code{ddbs_as_wkb()} returns a list of raw vectors (binary WKB)
#'   \item \code{ddbs_as_hexwkb()} returns a character vector of HEXWKB strings
#'   \item \code{ddbs_as_geojson()} returns a single GeoJSON \code{FeatureCollection}
#'     string (class \code{"geojson"}) when \code{feature_collection = TRUE}, or a
#'     character vector of \code{Feature} strings (one per row) when \code{FALSE};
#'     non-geometry columns are included as feature properties
#' }
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", package = "duckspatial")
#' )
#'
#' ddbs_as_text(argentina_ddbs)
#' ddbs_as_wkb(argentina_ddbs)
#' ddbs_as_hexwkb(argentina_ddbs)
#'
#' ## a single FeatureCollection (default)
#' ddbs_as_geojson(argentina_ddbs)
#'
#' ## one Feature string per row
#' ddbs_as_geojson(argentina_ddbs, feature_collection = FALSE)
#' }
#'
#' @name ddbs_as_format
#' @rdname ddbs_as_format
NULL





#' @rdname ddbs_as_format
#' @export
ddbs_as_text <- function(
  x,
  conn = NULL) {

  template_geometry_conversion(
    x = x,
    conn = conn,
    fun = "ST_AsText"
  )

}





#' @rdname ddbs_as_format
#' @export
ddbs_as_wkb <- function(
  x,
  conn = NULL) {

  template_geometry_conversion(
    x = x,
    conn = conn,
    fun = "ST_AsWKB"
  )

}





#' @rdname ddbs_as_format
#' @export
ddbs_as_hexwkb <- function(
  x,
  conn = NULL) {

  template_geometry_conversion(
    x = x,
    conn = conn,
    fun = "ST_AsHEXWKB"
  )

}





#' @rdname ddbs_as_format
#' @export
ddbs_as_geojson <- function(
  x,
  conn = NULL,
  feature_collection = TRUE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_character(conn, x)
  assert_logic(feature_collection, "feature_collection")

  # 1. Prepare inputs
  x <- normalize_spatial_input(x, conn)
  sf_col_x <- attr(x, "sf_column")

  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = TRUE)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  on.exit(resolve_conn$cleanup(), add = TRUE)

  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  # 2. Resolve the geometry column and the remaining (property) columns
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  desc       <- DBI::dbGetQuery(target_conn, glue::glue("DESCRIBE SELECT * FROM {x_list$query_name};"))
  prop_cols  <- desc$column_name[desc$column_name != x_geom]

  ## Per-row properties as compact JSON text. to_json() handles type formatting
  ## and string escaping; '{}' when there are no non-geometry columns.
  ## to_json() comes from the DuckDB `json` extension, which is only needed when
  ## there are non-geometry columns to encode as feature properties.
  if (length(prop_cols)) {
    if (isFALSE(check_loaded_extension(target_conn, "json"))) {
      cli::cli_abort(c(
        "{.fn ddbs_as_geojson} needs the DuckDB {.pkg json} extension to encode feature properties.",
        "i" = "Install it with {.code ddbs_install(conn, extension = \"json\")}, or drop the non-geometry columns first."
      ))
    }
    prop_fields <- paste(sprintf('"%s" := "%s"', prop_cols, prop_cols), collapse = ", ")
    props_sql   <- sprintf("to_json(struct_pack(%s))::VARCHAR", prop_fields)
  } else {
    props_sql <- "'{}'"
  }

  ## Build each Feature by plain string concatenation. This avoids round-tripping
  ## the geometry through the JSON type (json_object / json_group_array), which is
  ## markedly slower and far more memory-hungry on large datasets. Properties
  ## precede the geometry to match geojsonsf::sf_geojson().
  feature_sql <- sprintf(
    "'{\"type\":\"Feature\",\"properties\":' || %s || ',\"geometry\":' || coalesce(ST_AsGeoJSON(\"%s\"), 'null') || '}'",
    props_sql, x_geom
  )

  # 3. Build and retrieve the result
  if (feature_collection) {
    ## Single FeatureCollection. coalesce() keeps `features` as [] (not null) when empty.
    tmp.query <- sprintf(
      "SELECT '{\"type\":\"FeatureCollection\",\"features\":[' || coalesce(string_agg(%s, ','), '') || ']}' AS geojson FROM %s;",
      feature_sql, x_list$query_name
    )
    out <- DBI::dbGetQuery(target_conn, tmp.query)$geojson
    ## tag like geojsonsf so downstream tools treat it as raw JSON, not a string
    class(out) <- c("geojson", "json")
    out
  } else {
    tmp.query <- sprintf("SELECT %s AS geojson FROM %s;", feature_sql, x_list$query_name)
    DBI::dbGetQuery(target_conn, tmp.query)$geojson
  }

}




#' Create geometries from standard interchange formats
#'
#' @description
#' Parse serialized geometries from common interchange formats into a spatial
#' object, using DuckDB spatial deserialization functions. These are the
#' inverses of the [ddbs_as_text()] / [ddbs_as_wkb()] / [ddbs_as_hexwkb()] /
#' [ddbs_as_geojson()] serializers.
#'
#' * `ddbs_geom_from_text()` – Parse Well-Known Text (WKT)
#' * `ddbs_geom_from_wkb()` – Parse Well-Known Binary (WKB)
#' * `ddbs_geom_from_hexwkb()` – Parse hexadecimal Well-Known Binary (HEXWKB)
#' * `ddbs_geom_from_hexewkb()` – Parse hexadecimal Extended Well-Known Binary (HEXEWKB)
#' * `ddbs_geom_from_geojson()` – Parse GeoJSON
#'
#' @param x For `ddbs_geom_from_wkb()`, a list of raw vectors (binary WKB). For
#'   all other functions, a character vector of serialized geometries.
#' @param crs Character or numeric CRS specification (e.g. `"EPSG:4326"` or
#'   `4326`) to assign to the resulting geometries. Defaults to `NULL` (no CRS
#'   assigned).
#' @param ... Named vectors of additional attribute columns to include in the
#'   output. Each must have the same length as `x`.
#' @param geom_col Name of the geometry column in the output. Defaults to
#'   `"geometry"`.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' These functions are thin wrappers around the DuckDB spatial deserialization
#' functions (`ST_GeomFromText`, `ST_GeomFromWKB`, `ST_GeomFromHEXWKB`,
#' `ST_GeomFromHEXEWKB`, and `ST_GeomFromGeoJSON`).
#'
#' They are useful for importing geometries produced by external spatial tools,
#' databases, and web services back into a `duckspatial` workflow.
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' ## from WKT
#' ddbs_geom_from_text(
#'   c("POINT (-58.38 -34.60)", "POINT (-64.18 -31.42)"),
#'   crs = 4326
#' )
#'
#' ## from GeoJSON, with an extra attribute column
#' ddbs_geom_from_geojson(
#'   '{"type":"Point","coordinates":[-58.38,-34.60]}',
#'   city = "Buenos Aires",
#'   crs  = 4326
#' )
#' }
#'
#' @name ddbs_geom_from
#' @rdname ddbs_geom_from
NULL




#' @rdname ddbs_geom_from
#' @export
ddbs_geom_from_text <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_geom_from(
    x = x,
    crs = crs,
    ...,
    geom_col = geom_col,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_GeomFromText",
    input = "character"
  )

}




#' @rdname ddbs_geom_from
#' @export
ddbs_geom_from_wkb <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_geom_from(
    x = x,
    crs = crs,
    ...,
    geom_col = geom_col,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_GeomFromWKB",
    input = "blob"
  )

}




#' @rdname ddbs_geom_from
#' @export
ddbs_geom_from_hexwkb <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_geom_from(
    x = x,
    crs = crs,
    ...,
    geom_col = geom_col,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_GeomFromHEXWKB",
    input = "character"
  )

}




#' @rdname ddbs_geom_from
#' @export
ddbs_geom_from_hexewkb <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_geom_from(
    x = x,
    crs = crs,
    ...,
    geom_col = geom_col,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_GeomFromHEXEWKB",
    input = "character"
  )

}




#' @rdname ddbs_geom_from
#' @export
ddbs_geom_from_geojson <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_geom_from(
    x = x,
    crs = crs,
    ...,
    geom_col = geom_col,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_GeomFromGeoJSON",
    input = "character"
  )

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## insert data
ddbs_write_table(conn_test, countries_sf, "countries")


# 1. ddbs_as_text() ------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: message works
## - CHECK 1.3: compare to SF (we can't because DuckDB retrieves more decimals)
## - CHECK 2.1: errors
describe("ddbs_as_text()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------
  describe("expected behavior", {

    it("works on all input formats", {
      output_sf   <- ddbs_as_text(countries_sf)
      output_ddbs <- ddbs_as_text(countries_ddbs)
      output_conn <- ddbs_as_text("countries", conn_test)

      expect_equal(output_sf, output_ddbs)
      expect_equal(output_sf, output_conn)
    })

    it("doesn't display a message", {
      expect_no_message(ddbs_as_text(countries_sf))
    })

  })

  ### ERRORS ------------------------------------------------------------
  describe("errors work", {

    it("throws errors for invalid inputs", {
      expect_error(ddbs_as_text(x = 999))
      expect_error(ddbs_as_text(countries_ddbs, conn = 999))
      expect_error(ddbs_as_text(x = "999", conn = conn_test))
    })

  })

})



# 2. ddbs_as_wkb ---------------------------------------------------------

## 2.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all formats
## - CHECK 1.2: message works
## - CHECK 1.3: compare to SF (the class is different, so we compare the first 
## and last elements)
## - CHECK 2.1: errors
describe("ddbs_as_wkb()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------
  describe("expected behavior", {

    it("works on all input formats and produces consistent WKB", {
      output_sf   <- ddbs_as_wkb(countries_sf)
      output_ddbs <- ddbs_as_wkb(countries_ddbs)
      output_conn <- ddbs_as_wkb("countries", conn_test)

      expect_equal(output_sf, output_ddbs)
      expect_equal(output_sf, output_conn)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_as_wkb(countries_sf))
    })

    it("matches SF WKB output at first and last elements", {
      sf_output <- sf::st_as_binary(countries_sf$geometry)
      output_sf   <- ddbs_as_wkb(countries_sf)

      expect_equal(output_sf[[1]], sf_output[[1]])
      expect_equal(length(output_sf), length(sf_output))
      expect_equal(output_sf[[length(output_sf)]], sf_output[[length(sf_output)]])
    })

  })

  ### ERRORS ------------------------------------------------------------
  describe("errors work", {

    it("throws errors for invalid inputs", {
      expect_error(ddbs_as_wkb(x = 999))
      expect_error(ddbs_as_wkb(countries_ddbs, conn = 999))
      expect_error(ddbs_as_wkb(x = "999", conn = conn_test))
    })

  })

})


# 3. ddbs_as_hexwkb() ----------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: message works
## - CHECK 1.3: compare to SF (the class is different, so we compare the first 
## and last elements)
## - CHECK 2.1: errors
describe("ddbs_as_hexwkb()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------
  describe("expected behavior", {

    it("works on all input formats and produces consistent HEX WKB", {
      output_sf   <- ddbs_as_hexwkb(countries_sf)
      output_ddbs <- ddbs_as_hexwkb(countries_ddbs)
      output_conn <- ddbs_as_hexwkb("countries", conn_test)

      expect_equal(output_sf, output_ddbs)
      expect_equal(output_sf, output_conn)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_as_hexwkb(countries_sf))
    })

    it("matches SF HEX WKB output at first and last elements", {
      sf_output <- sf::st_as_binary(countries_sf$geometry, hex = TRUE)
      output_sf   <- ddbs_as_hexwkb(countries_sf)
      output_sf_lower <- lapply(output_sf, tolower)

      expect_equal(output_sf_lower[[1]], sf_output[[1]])
      expect_equal(length(output_sf_lower), length(sf_output))
      expect_equal(output_sf_lower[[length(output_sf_lower)]], sf_output[[length(sf_output)]])
    })

  })

  ### ERRORS ------------------------------------------------------------
  describe("errors work", {

    it("throws errors for invalid inputs", {
      expect_error(ddbs_as_hexwkb(x = 999))
      expect_error(ddbs_as_hexwkb(countries_ddbs, conn = 999))
      expect_error(ddbs_as_hexwkb(x = "999", conn = conn_test))
    })

  })

})



# 4. ddbs_as_geojson() ---------------------------------------------------

## - CHECK 1.1: works on all input formats and they agree
## - CHECK 1.2: default returns a single FeatureCollection (class geojson/json)
## - CHECK 1.3: feature_collection = FALSE returns one Feature per row
## - CHECK 1.4: non-geometry columns are included as properties
## - CHECK 1.5: output is valid, parseable GeoJSON
## - CHECK 1.6: structurally matches geojsonsf::sf_geojson()
## - CHECK 1.7: geometry-only input yields empty properties
## - CHECK 1.8: empty input yields an empty FeatureCollection
## - CHECK 2.1: errors
describe("ddbs_as_geojson()", {

  ## encoding feature properties needs the DuckDB json extension (to_json);
  ## skip where it is not available (e.g. offline CI such as r-hub)
  skip_if(isFALSE(check_installed_extension(conn_test, "json")), "json extension not available")

  ### EXPECTED BEHAVIOR -------------------------------------------------
  describe("expected behavior", {

    it("works on all input formats and they agree", {
      output_sf   <- ddbs_as_geojson(countries_sf)
      output_ddbs <- ddbs_as_geojson(countries_ddbs)
      output_conn <- ddbs_as_geojson("countries", conn_test)

      expect_type(output_sf, "character")
      expect_equal(output_sf, output_ddbs)
      expect_equal(output_sf, output_conn)
    })

    it("returns a single FeatureCollection by default", {
      output <- ddbs_as_geojson(countries_sf)
      expect_length(output, 1L)
      expect_s3_class(output, "geojson")
      parsed <- jsonlite::fromJSON(output, simplifyVector = FALSE)
      expect_equal(parsed$type, "FeatureCollection")
      expect_length(parsed$features, nrow(countries_sf))
    })

    it("returns one Feature per row when feature_collection = FALSE", {
      output <- ddbs_as_geojson(countries_sf, feature_collection = FALSE)
      expect_type(output, "character")
      expect_length(output, nrow(countries_sf))
      expect_true(all(grepl('"type":"Feature"', output, fixed = TRUE)))
    })

    it("includes all non-geometry columns as properties", {
      output <- ddbs_as_geojson(countries_sf, feature_collection = FALSE)
      parsed <- jsonlite::fromJSON(output[1])

      expect_named(parsed, c("type", "properties", "geometry"))
      prop_names <- setdiff(names(countries_sf), attr(countries_sf, "sf_column"))
      expect_setequal(names(parsed$properties), prop_names)
    })

    it("produces valid, parseable GeoJSON geometry", {
      parsed <- jsonlite::fromJSON(ddbs_as_geojson(countries_sf), simplifyVector = FALSE)
      expect_equal(parsed$features[[1]]$type, "Feature")
      expect_false(is.null(parsed$features[[1]]$geometry$type))
    })

    it("matches the structure of geojsonsf::sf_geojson()", {
      skip_if_not_installed("geojsonsf")
      ours <- jsonlite::fromJSON(ddbs_as_geojson(countries_sf), simplifyVector = FALSE)
      ref  <- jsonlite::fromJSON(geojsonsf::sf_geojson(countries_sf), simplifyVector = FALSE)
      expect_equal(ours$type, ref$type)
      expect_equal(length(ours$features), length(ref$features))
      expect_equal(ours$features[[1]]$properties, ref$features[[1]]$properties)
    })

    it("yields empty properties when there are no other columns", {
      geom_only <- countries_sf[, attr(countries_sf, "sf_column")]
      output    <- ddbs_as_geojson(geom_only, feature_collection = FALSE)
      expect_true(all(grepl('"properties":{}', output, fixed = TRUE)))
    })

    it("returns an empty FeatureCollection for empty input", {
      DBI::dbExecute(conn_test, "CREATE OR REPLACE TABLE empty_geo AS SELECT * FROM countries WHERE 1=0;")
      output <- ddbs_as_geojson("empty_geo", conn_test)
      parsed <- jsonlite::fromJSON(output, simplifyVector = FALSE)
      expect_equal(parsed$type, "FeatureCollection")
      expect_length(parsed$features, 0L)
    })

    it("doesn't display a message", {
      expect_no_message(ddbs_as_geojson(countries_sf))
    })
  })

  ### ERRORS ------------------------------------------------------------
  describe("errors work", {

    it("throws errors for invalid inputs", {
      expect_error(ddbs_as_geojson(x = 999))
      expect_error(ddbs_as_geojson(countries_ddbs, conn = 999))
      expect_error(ddbs_as_geojson(x = "999", conn = conn_test))
      expect_error(ddbs_as_geojson(countries_ddbs, feature_collection = "yes"))
    })
  })
})



# 5. ddbs_geom_from_*() --------------------------------------------------

## Build serialized representations of two known points to round-trip back
## - CHECK 1.x: each parser returns the expected output class/format
## - CHECK 2.x: CRS is assigned, extra columns kept, named tables written
## - CHECK 3.x: round-trips reproduce the original coordinates
## - CHECK 4.x: errors on invalid inputs
geom_from_ser <- DBI::dbGetQuery(conn_test, "
  SELECT ST_AsText(geom)    AS wkt,
         ST_AsHEXWKB(geom)  AS hex,
         ST_AsGeoJSON(geom) AS gj,
         ST_AsWKB(geom)     AS wkb
  FROM (VALUES
    (ST_Point(10, 20)::GEOMETRY),
    (ST_Point(-58.38, -34.60)::GEOMETRY)
  ) t(geom);
")

describe("ddbs_geom_from_text()", {

  describe("expected behavior", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns an sf object with mode = 'sf'", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326, mode = "sf")
      expect_s3_class(output, "sf")
      expect_equal(nrow(output), 2L)
    })

    it("assigns the CRS", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326, mode = "sf")
      expect_equal(sf::st_crs(output), sf::st_crs(4326))
    })

    it("round-trips coordinates correctly", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326, mode = "sf")
      expect_equal(sf::st_coordinates(output)[1, ], c(X = 10, Y = 20))
    })

    it("keeps extra columns passed via ...", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, id = 1:2, crs = 4326, mode = "sf")
      expect_true("id" %in% names(output))
      expect_equal(output$id, 1:2)
    })

    it("respects geom_col", {
      output <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326, geom_col = "the_geom", mode = "sf")
      expect_equal(attr(output, "sf_column"), "the_geom")
    })

    it("writes a table when name is provided", {
      ok <- ddbs_geom_from_text(geom_from_ser$wkt, crs = 4326, conn = conn_test, name = "gft_tbl")
      expect_true(ok)
      expect_true(DBI::dbExistsTable(conn_test, "gft_tbl"))
    })
  })

  describe("errors work", {

    it("errors on non-character input", {
      expect_error(ddbs_geom_from_text(123))
    })

    it("errors when name is given without conn", {
      expect_error(ddbs_geom_from_text(geom_from_ser$wkt, name = "no_conn"))
    })

    it("errors on unnamed ... arguments", {
      expect_error(ddbs_geom_from_text(geom_from_ser$wkt, 1:2))
    })

    it("errors when extra columns have the wrong length", {
      expect_error(ddbs_geom_from_text(geom_from_ser$wkt, id = 1:3))
    })
  })
})

describe("ddbs_geom_from_hexwkb()", {

  it("returns a duckspatial_df by default", {
    expect_s3_class(ddbs_geom_from_hexwkb(geom_from_ser$hex, crs = 4326), "duckspatial_df")
  })

  it("round-trips coordinates correctly", {
    output <- ddbs_geom_from_hexwkb(geom_from_ser$hex, crs = 4326, mode = "sf")
    expect_equal(sf::st_coordinates(output)[1, ], c(X = 10, Y = 20))
  })

  it("errors on non-character input", {
    expect_error(ddbs_geom_from_hexwkb(123))
  })
})

describe("ddbs_geom_from_hexewkb()", {

  it("returns a duckspatial_df by default", {
    expect_s3_class(ddbs_geom_from_hexewkb(geom_from_ser$hex, crs = 4326), "duckspatial_df")
  })

  it("round-trips coordinates correctly", {
    output <- ddbs_geom_from_hexewkb(geom_from_ser$hex, crs = 4326, mode = "sf")
    expect_equal(sf::st_coordinates(output)[1, ], c(X = 10, Y = 20))
  })
})

describe("ddbs_geom_from_geojson()", {

  it("returns a duckspatial_df by default", {
    expect_s3_class(ddbs_geom_from_geojson(geom_from_ser$gj, crs = 4326), "duckspatial_df")
  })

  it("round-trips coordinates correctly", {
    output <- ddbs_geom_from_geojson(geom_from_ser$gj, crs = 4326, mode = "sf")
    expect_equal(sf::st_coordinates(output)[1, ], c(X = 10, Y = 20))
  })

  it("errors on non-character input", {
    expect_error(ddbs_geom_from_geojson(123))
  })
})

describe("ddbs_geom_from_wkb()", {

  it("returns a duckspatial_df by default", {
    expect_s3_class(ddbs_geom_from_wkb(geom_from_ser$wkb, crs = 4326), "duckspatial_df")
  })

  it("round-trips coordinates correctly", {
    output <- ddbs_geom_from_wkb(geom_from_ser$wkb, crs = 4326, mode = "sf")
    expect_equal(sf::st_coordinates(output)[1, ], c(X = 10, Y = 20))
  })

  it("errors when input is not a list of raw vectors", {
    expect_error(ddbs_geom_from_wkb("not raw"))
  })
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

## dbConnCheck

#' Check if a supported DuckDB connection
#'
#' @template conn
#'
#' @keywords internal
#' @returns TRUE (invisibly) for successful import
dbConnCheck <- function(conn) {  # nocov start
    if (inherits(conn, "duckdb_connection")) {
        return(invisible(TRUE))

    } else if (is.null(conn)) { return(invisible(FALSE))

    } else {
        cli::cli_abort("'conn' must be connection object: <duckdb_connection> from `duckdb`")
    }
}  # nocov end

#' Normalize spatial input for processing
#'
#' Standardizes all input types before spatial operations:
#' - sf objects: passed through unchanged
#' - duckspatial_df: passed through unchanged
#' - tbl_duckdb_connection: coerced to duckspatial_df (with CRS/source_table attributes)
#' - character: verified to exist in connection
#'
#' This normalization enables downstream functions to work with a consistent set of types
#' (sf, duckspatial_df, or character) rather than handling all possible input variations.
#'
#' @param x Spatial input (sf, duckspatial_df, tbl_duckdb_connection, or character)
#' @param conn DuckDB connection (required for character table names)
#' @return Normalized input ready for get_query_list()
#' @keywords internal
#' @noRd
normalize_spatial_input <- function(x, conn = NULL, geom_col = NULL) {
  # 1. sf: pass through
  if (inherits(x, "sf")) return(x)
  
  # 2. duckspatial_df: already normalized
  if (inherits(x, "duckspatial_df")) return(x)
  
  # 3. tbl_duckdb_connection: coerce to duckspatial_df
  if (inherits(x, "tbl_duckdb_connection")) {
    return(as_duckspatial_df(x, geom_col = geom_col))
  }
  
  # 4. character: verify table/view exists
  if (is.character(x)) {
    if (is.null(conn)) {
      cli::cli_abort("{.arg conn} required when using character table names.")
    }
    if (!table_exists(conn, x) & !arrow_view_exists(conn, x)) {
      cli::cli_abort("Table or view {.val {x}} does not exist in connection.")
    }
    return(x)
  }
  
  # Unsupported type - let downstream handle/error
  x
}


table_exists <- function(conn, name, schema = "main") {
  DBI::dbGetQuery(
    conn,
    sprintf(
      "SELECT EXISTS (
         SELECT 1 FROM information_schema.tables
         WHERE table_schema='%s' AND table_name='%s'
       )",
      schema, name
    )
  )[1,1]
}


arrow_view_exists <- function(conn, name) {
  name %in% duckdb::duckdb_list_arrow(conn)
}

#' Get DuckDB connection from an object
#'
#' Extracts the connection from duckspatial_df, tbl_lazy, or validates a direct connection.
#'
#' @param x A duckspatial_df, tbl_lazy, duckdb_connection, or NULL
#' @return A duckdb_connection or NULL
#' @keywords internal
get_conn_from_input <- function(x) {  # nocov start
  if (is.null(x)) return(NULL)
  
  if (inherits(x, "duckdb_connection")) return(x)
  
  if (inherits(x, c("duckspatial_df", "tbl_lazy"))) {
    return(dbplyr::remote_con(x))
  }
  
  NULL
}  # nocov end

#' Compare two CRS objects for equality
#'
#' Properly compares CRS objects, handling different representations of the same CRS.
#'
#' @param crs1 First CRS object
#' @param crs2 Second CRS object
#' @return Logical indicating if CRS are equal
#' @keywords internal
crs_equal <- function(crs1, crs2) {  # nocov start
  if (is.null(crs1) || is.null(crs2)) return(FALSE)
  isTRUE(sf::st_crs(crs1) == sf::st_crs(crs2))
}  # nocov end

#' Import a view/table from one connection to another
#'
#' Enables cross-connection operations by importing views using one of three strategies.
#'
#' @param target_conn Target DuckDB connection
#' @param source_conn Source DuckDB connection  
#' @param source_object duckspatial_df, tbl_lazy, or tbl_duckdb_connection from source_conn
#' @param target_name Name for view in target connection (auto-generated if NULL)
#' @template quiet
#' @return List with imported view name and cleanup function
#' @keywords internal
import_view_to_connection <- function(target_conn, source_conn, source_object, target_name = NULL, quiet = FALSE) {
  
  if (is.null(target_name)) {
    target_name <- paste0("imported_", gsub("-", "_", uuid::UUIDgenerate()))
  }
  
  # Track cleanup actions
  cleanup_actions <- list()
  
  # STRATEGY 1: SQL recreation (same DB, direct view reference)
  if (inherits(source_object, c("duckspatial_df", "tbl_duckdb_connection", "tbl_lazy"))) {
    source_table <- dbplyr::remote_name(source_object)
    
    if (!is.null(source_table) && !inherits(source_table, "sql")) {
      source_table_clean <- gsub('^"|"$', "", as.character(source_table))
      
      view_sql <- tryCatch({
        q_sql <- glue::glue(
          "SELECT sql FROM duckdb_views() WHERE view_name = {DBI::dbQuoteString(source_conn, source_table_clean)}"
        )
        result <- DBI::dbGetQuery(source_conn, q_sql)
        if (nrow(result) > 0) result$sql else NULL
      }, error = function(e) NULL)
      
      if (!is.null(view_sql) && length(view_sql) > 0) {
        source_pat_clean <- paste0("VIEW\\s+\"", source_table_clean, "\"")
        source_pat <- paste0("VIEW\\s+", source_table)
        
        if (grepl(source_pat_clean, view_sql, ignore.case = TRUE)) {
          new_sql <- sub(source_pat_clean, paste0("VIEW ", target_name), view_sql, ignore.case = TRUE)
        } else {
          new_sql <- sub(source_pat, paste0("VIEW ", target_name), view_sql, ignore.case = TRUE)
        }
        new_sql <- sub("CREATE VIEW", "CREATE OR REPLACE TEMPORARY VIEW", new_sql, ignore.case = TRUE)
        
        tryCatch({
          DBI::dbExecute(target_conn, new_sql)
          if (!quiet) cli::cli_inform("Imported view using SQL recreation (zero overhead)")
          return(list(name = target_name, method = "sql_recreation", cleanup = function() NULL))
        }, error = function(e) {
          if (isTRUE(getOption("duckspatial.debug", FALSE))) {
            cli::cli_alert_info("Strategy 1 (SQL recreation) failed: {conditionMessage(e)}")
          }
        })
      }
    }
  }
  
  # STRATEGY 2: SQL render (same DB, lazy query)
  if (inherits(source_object, c("tbl_lazy", "tbl_duckdb_connection"))) {
    query_sql <- tryCatch({
      as.character(dbplyr::sql_render(source_object, con = source_conn))
    }, error = function(e) NULL)
    
    if (!is.null(query_sql)) {
      view_query <- glue::glue("CREATE OR REPLACE TEMPORARY VIEW {target_name} AS {query_sql}")
      
      tryCatch({
        DBI::dbExecute(target_conn, view_query)
        if (!quiet) cli::cli_inform("Imported view using SQL query (zero overhead)")
        return(list(name = target_name, method = "sql_render", cleanup = function() NULL))
      }, error = function(e) {
        if (isTRUE(getOption("duckspatial.debug", FALSE))) {
          cli::cli_alert_info("Strategy 2 (SQL render) failed: {conditionMessage(e)}")
        }
      })
    }
  }
  
  # STRATEGY 3: ATTACH file-based source DB (READ_ONLY)
  # Try multiple ways to get dbdir since dbGetInfo may return NULL
  source_dbdir <- tryCatch({
    info <- DBI::dbGetInfo(source_conn)
    dbdir <- info$dbdir
    if (is.null(dbdir) || length(dbdir) == 0) {
       dbdir <- info$dbname
    }
    
    if (is.null(dbdir) || length(dbdir) == 0) {
      # Fallback: try to query from DuckDB itself
      res <- DBI::dbGetQuery(source_conn, "SELECT current_database()")
      NULL # Still can't get file path from this
    }
    dbdir
  }, error = function(e) NULL)
  
  if (!is.null(source_dbdir) && source_dbdir != ":memory:" && file.exists(source_dbdir)) {
    source_table <- tryCatch(dbplyr::remote_name(source_object), error = function(e) NULL)
    
    if (!is.null(source_table) && !inherits(source_table, "sql")) {
      attach_alias <- paste0("src_", gsub("-", "_", uuid::UUIDgenerate()))
      source_table_clean <- gsub('^"|"$', "", as.character(source_table))
      
      tryCatch({
        DBI::dbExecute(target_conn, glue::glue("ATTACH '{source_dbdir}' AS {attach_alias} (READ_ONLY)"))
        
        view_query <- glue::glue("
          CREATE OR REPLACE TEMPORARY VIEW {target_name} AS
          SELECT * FROM {attach_alias}.{source_table_clean}
        ")
        DBI::dbExecute(target_conn, view_query)
        
        if (!quiet) cli::cli_inform("Imported view using ATTACH (zero-copy, READ_ONLY)")
        return(list(
          name = target_name, 
          method = "attach",
          cleanup = function() {
            tryCatch(DBI::dbExecute(target_conn, glue::glue("DETACH {attach_alias}")), error = function(e) NULL)
          }
        ))
      }, error = function(e) {
        if (isTRUE(getOption("duckspatial.debug", FALSE))) {
          cli::cli_alert_info("Strategy 3 (ATTACH) failed: {conditionMessage(e)}")
        }
        tryCatch(DBI::dbExecute(target_conn, glue::glue("DETACH IF EXISTS {attach_alias}")), error = function(e) NULL)
      })
    }
  }
  
  # STRATEGY 4: Nanoarrow streaming (cross-DB)
  # Materialize into target to avoid Arrow lifecycle issues
  if (inherits(source_object, c("tbl_lazy", "tbl_duckdb_connection", "duckspatial_df"))) {
    query_sql <- tryCatch({
      as.character(dbplyr::sql_render(source_object, con = source_conn))
    }, error = function(e) NULL)
    
    if (!is.null(query_sql)) {
      tryCatch({
        res <- DBI::dbSendQuery(source_conn, query_sql, arrow = TRUE)
        
        reader <- duckdb::duckdb_fetch_arrow(res)
        stream <- nanoarrow::as_nanoarrow_array_stream(reader)
        
        # Register temporarily, then materialize into a view
        temp_arrow_name <- paste0("temp_arrow_", gsub("-", "_", uuid::UUIDgenerate()))
        duckdb::duckdb_register_arrow(target_conn, temp_arrow_name, stream)
        
        # Create permanent view from Arrow data (materializes)
        DBI::dbExecute(target_conn, glue::glue(
          "CREATE OR REPLACE TEMPORARY VIEW {target_name} AS SELECT * FROM {temp_arrow_name}"
        ))
        
        # Cleanup Arrow registration
        DBI::dbClearResult(res)
        tryCatch(duckdb::duckdb_unregister_arrow(target_conn, temp_arrow_name), error = function(e) NULL)
        
        if (!quiet) cli::cli_inform("Imported via nanoarrow streaming (zero R materialization)")
        return(list(name = target_name, method = "nanoarrow", cleanup = function() NULL))
      }, error = function(e) {
        if (isTRUE(getOption("duckspatial.debug", FALSE))) {
          cli::cli_alert_info("Strategy 4 (nanoarrow) failed: {conditionMessage(e)}")
        }
      })
    }
  }
  
  # STRATEGY 5: Collect + register (last resort)
  df <- dplyr::collect(source_object)
  
  if (inherits(df, "sf")) {
    duckspatial::ddbs_write_table(target_conn, df, target_name, temp_view = TRUE, quiet = quiet)
    cli::cli_warn("Imported via collection (materialized to R, then uploaded)")
    return(list(name = target_name, method = "collect_and_write", data = df, cleanup = function() NULL))
  } else if (is.data.frame(df)) {
    # For non-spatial data, use duckdb_register (zero-copy from R)
    duckdb::duckdb_register(target_conn, target_name, df)
    cli::cli_warn("Imported via duckdb_register (collected to R, zero-copy to target)")
    return(list(name = target_name, method = "duckdb_register", data = df, cleanup = function() NULL))
  } else {
    cli::cli_abort("Import failed: Cannot import object of class {.cls {class(df)}}.")
  }
}



#' Get column names in a DuckDB database
#'
#' @template conn
#' @param x name of the table
#' @param rest whether to return geometry column name, of the rest of the columns
#'
#' @keywords internal
#' @returns name of the geometry column of a table
get_geom_name <- function(conn, x, rest = FALSE, collapse = FALSE, table_id = NULL) {  # nocov start

    # check if the table exists (via DESCRIBE which works for temp views too)
    info_tbl <- try(DBI::dbGetQuery(conn, glue::glue("DESCRIBE {x};")), silent = TRUE)
    
    if (inherits(info_tbl, "try-error")) {
        cli::cli_abort("The table <{x}> does not exist.")
    }
    other_cols <- if (rest) {
        info_tbl[!grepl("GEOMETRY", info_tbl$column_type, ignore.case = TRUE), "column_name"]
    } else {
        info_tbl[grepl("GEOMETRY", info_tbl$column_type, ignore.case = TRUE), "column_name"]
    }

    # collapse columns with quoted names
    if (isTRUE(collapse)) {
      if (length(other_cols) > 0) {
        prefix <- if (is.null(table_id)) "" else paste0(table_id, ".")
        other_cols <- paste0(prefix, '"', other_cols, '"', collapse = ', ')
        other_cols <- paste0(other_cols, ", ") # trailing comma for SELECT {rest} something
      } else {
        other_cols <- ""
      }
    }

    return(other_cols)
}  # nocov end


#' Get names for the query
#'
#' @param name table name
#'
#' @keywords internal
#' @returns list with fixed names
get_query_name <- function(name) {  # nocov start
    if (length(name) == 2) {
        table_name <- name[2]
        schema_name <- name[1]
        query_name <- paste0(name, collapse = ".")
    } else {
        table_name   <- name
        schema_name <- "main"
        query_name <- name
    }
    list(
        table_name = table_name,
        schema_name = schema_name,
        query_name = query_name
    )
} # nocov end



#' Get names for the query
#'
#' @param x sf, duckspatial_df, tbl_lazy, or character
#' @template conn_null
#'
#' @keywords internal
#' @noRd
#' @returns list with fixed names
# IMPORTANT: This function returns a cleanup function instead of using on.exit() internally.
# 
# Why? R's on.exit() runs when the function containing it exits, NOT when the caller exits.
# If we used on.exit() here, the temporary views would be dropped as soon as get_query_list()
# returns, BEFORE the caller can execute their SQL query that references the view.
#
# The caller MUST register the cleanup function with on.exit() in their own scope:
#   result <- get_query_list(x, conn)
#   on.exit(result$cleanup(), add = TRUE)
#   # ... use result$query_name ...
#   # ... use result$query_name ...
get_query_list <- function(x, conn) {

  if (inherits(x, "sf")) {
    temp_view_name <- ddbs_temp_view_name()
    duckspatial::ddbs_write_table(conn = conn, data = x, name = temp_view_name,
                                    quiet = TRUE, temp_view = TRUE)
    x_list <- get_query_name(temp_view_name)
    x_list$cleanup <- function() {
      tryCatch(DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {temp_view_name};")), error = function(e) NULL)
      tryCatch(duckdb::duckdb_unregister_arrow(conn, temp_view_name), error = function(e) NULL)
    }
    x_list$owned <- FALSE   # created here, caller should not clean up
    return(x_list)

  } else if (inherits(x, "duckspatial_df")) {
    source_table <- attr(x, "source_table")
    if (!is.null(source_table)) {
      remote_name_result <- tryCatch(dbplyr::remote_name(x), error = function(e) NULL)
      if (!is.null(remote_name_result) &&
          !inherits(remote_name_result, "sql") &&
          as.character(remote_name_result) == source_table) {
        result <- get_query_name(source_table)
        result$cleanup <- function() NULL
        result$owned <- TRUE
        return(result)
      }
    }
    ## Test duckdb 1.5
    x_list <- get_query_name(source_table)
    if (!is.null(x_list$table_name)) {
      x_list$cleanup <- function() NULL
      x_list$owned <- TRUE
      return(x_list)
    }
    ## Modified by dplyr verbs: render to a new temp view
    temp_view_name <- ddbs_temp_view_name()
    query_sql <- dbplyr::sql_render(x, con = conn)
    DBI::dbExecute(conn, glue::glue(
      "CREATE OR REPLACE TEMPORARY VIEW {temp_view_name} AS {query_sql}"
    ))
    x_list <- get_query_name(temp_view_name)
    x_list$cleanup <- function() {
      tryCatch(DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {temp_view_name};")), error = function(e) NULL)
    }
    x_list$owned <- TRUE

    return(x_list)

  } else if (inherits(x, "tbl_lazy")) {
    temp_view_name <- ddbs_temp_view_name()
    query_sql <- dbplyr::sql_render(x)
    DBI::dbExecute(conn, glue::glue(
      "CREATE OR REPLACE TEMPORARY VIEW {temp_view_name} AS {query_sql}"
    ))
    x_list <- get_query_name(temp_view_name)
    x_list$cleanup <- function() {
      tryCatch(DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {temp_view_name};")), error = function(e) NULL)
    }
    x_list$owned <- TRUE
    return(x_list)

  } else if (inherits(x, "data.frame")) {
    temp_view_name <- ddbs_temp_view_name()
    duckdb::duckdb_register(conn, temp_view_name, x)
    x_list <- get_query_name(temp_view_name)
    x_list$cleanup <- function() {
      tryCatch(DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {temp_view_name};")), error = function(e) NULL)
      tryCatch(duckdb::duckdb_unregister_arrow(conn, temp_view_name), error = function(e) NULL)
    }
    x_list$owned <- TRUE
    return(x_list)

  } else {
    ## Character table name: pre-existing, never clean up
    x_list <- get_query_name(x)
    x_list$cleanup <- function() NULL
    x_list$owned <- TRUE
    return(x_list)
  }
}




#' Gets predicate name
#'
#' Gets a full predicate name from the shorter version
#'
#' @template predicate
#'
#' @keywords internal
#' @returns character
get_st_predicate <- function(predicate) { # nocov start
    switch(predicate,
      "intersects"            = "ST_Intersects",
      "intersects_extent"     = "ST_Intersects_Extent",
      "covers"                = "ST_Covers",
      "touches"               = "ST_Touches",
      "contains"              = "ST_Contains",
      "contains_properly"     = "ST_ContainsProperly",
      "within"                = "ST_Within",
      "within_properly"       = "ST_WithinProperly",
      "disjoint"              = "ST_Disjoint",
      "equals"                = "ST_Equals",
      "overlaps"              = "ST_Overlaps",
      "crosses"               = "ST_Crosses",
      "covered_by"            = "ST_CoveredBy",
      "dwithin"               = "ST_DWithin",
      cli::cli_abort(c(
          "Invalid spatial predicate: {.val {predicate}}",
          "i" = "Valid options: {.val {c('intersects', 'intersects_extent', 'covers', 'touches', 'contains', 'contains_properly', 'within', 'within_properly', 'dwithin', 'disjoint', 'equals', 'overlaps', 'crosses', 'covered_by')}}"
        ))
      )
} # nocov end


#' Converts from data frame to sf using WKB conversion
#'
#' Converts a table that has been read from DuckDB into an sf object.
#'
#' @param data a tibble or data frame
#' @template crs
#' @param x_geom name of geometry column
#'
#' @keywords internal
#' @returns sf
convert_to_sf_wkb <- function(data, crs, x_geom) { # nocov start

  # 1. Resolve CRS
  target_crs <- crs

  # Add warning if still no CRS found
  if (is.null(target_crs)) {
    cli::cli_alert_warning("No CRS found for the imported table.")
  }

  # 2. Check Geometry Type and Convert
  geom_data <- data[[x_geom]]

  if (inherits(geom_data, "blob") || is.list(geom_data)) {
    # --- FAST PATH: Binary Data ---

    # Attempt to use wk directly.
    # We use tryCatch because:
    # 1. It handles lists where the first element is NULL (which is.raw() misses)
    # 2. It safely falls back if the list contains non-WKB data
    
    wk_success <- tryCatch({
      # Strip attributes (like 'blob') to ensure it's a clean list for wk
      attributes(geom_data) <- NULL
      
      # OPTIMIZATION: Zero-copy wrap and convert
      wkb_obj <- wk::new_wk_wkb(geom_data)
      data[[x_geom]] <- sf::st_as_sfc(wkb_obj)
      TRUE
    }, error = function(e) {
      FALSE
    })

    if (!wk_success) {
      # --- FALLBACK PATH ---
      # Used if wk failed (e.g., data is native arrow structure or complex list)
      tryCatch({
        ga_vctr <- geoarrow::as_geoarrow_vctr(geom_data)
        data[[x_geom]] <- sf::st_as_sfc(ga_vctr)
      }, error = function(e) {
        # Final fallback: standard sf blob reading
        data[[x_geom]] <- sf::st_as_sfc(structure(geom_data, class = "WKB"))
      })
    }

  } else if (is.character(geom_data)) {
    # --- SLOW PATH: WKT Strings ---
    # Used if the query explicitly used ST_AsText() or older DuckDB versions
    data[[x_geom]] <- sf::st_as_sfc(geom_data)
  }

  # 3. Construct SF Object
  # Use st_as_sf with the pre-converted geometry column
  # We explicitly set the geometry column name to handle cases where x_geom isn't "geometry"
  sf_obj <- sf::st_as_sf(data, sf_column_name = x_geom)

  # 4. Assign CRS if found
  if (!is.null(target_crs)) {
    sf::st_crs(sf_obj) <- sf::st_crs(target_crs)
  }

  return(sf_obj)
} # nocov end






#' Feedback for overwrite argument
#'
#' @param x table name
#' @template conn
#' @template quiet
#' @template overwrite
#'
#' @keywords internal
#' @returns cli message
overwrite_table <- function(x, conn, quiet, overwrite) { # nocov start
  if (overwrite) {
    DBI::dbExecute(conn, glue::glue("DROP TABLE IF EXISTS {x};"))
    if (isFALSE(quiet)) cli::cli_alert_info("Table <{x}> dropped")
  }
} # nocov end





#' Feedback for query success
#'
#' @template quiet
#'
#' @keywords internal
#' @returns cli message
feedback_query <- function(quiet) { # nocov start
  if (isFALSE(quiet)) cli::cli_alert_success("Query successful")
} # nocov end



#' Get the number of rows in a table
#'
#' @template conn
#' @param table name of the table
#'
#' @keywords internal
#' @returns number of rows in the table
get_nrow <- function(conn, table) { # nocov start
  DBI::dbGetQuery(conn, glue::glue("SELECT COUNT(*) as n FROM {table}"))$n
} # nocov end





reframe_predicate_data <- function(
  conn, 
  data, 
  x_list, 
  y_list, 
  id_x, 
  id_y, 
  sparse) { # nocov start

  ## get number of rows
  nrowx <- get_nrow(conn, x_list$query_name)
  nrowy <- get_nrow(conn, y_list$query_name)

  ## convert results to matrix -> to list
  ## return matrix if sparse = FALSE
  pred_mat  <- matrix(data$predicate, nrow = nrowx, ncol = nrowy, byrow = TRUE)
  if (isFALSE(sparse)) return(pred_mat)

  pred_list <- apply(pred_mat, 1, function(row) which(row), simplify = FALSE)

  ## return if no matches have been found
  if (length(pred_list) == 0) return(NULL)

  ## rename list if id is provided
  if (!is.null(id_x)) {
    idx_names <- DBI::dbGetQuery(conn, glue::glue("SELECT {id_x} as id FROM {x_list$query_name}"))$id
    names(pred_list) <- idx_names
  }

  ## rename list if id is provided
  if (!is.null(id_y)) {
    idy_names <- DBI::dbGetQuery(conn, glue::glue("SELECT {id_y} as id FROM {y_list$query_name}"))$id
    pred_list <- lapply(pred_list, function(ind) {
      if (length(ind) == 0) return(ind)
      idy_names[ind]
    })
  }

  return(pred_list)

} # nocov end

#' Convert CRS input to DuckDB SQL literal
#'
#' Helper to format numeric EPSG codes, WKT strings, or `sf::st_crs` objects
#' into a SQL literal string compatible with `ST_Transform`.
#'
#' @param x numeric (EPSG), character (WKT/Proj), or `sf` crs object
#'
#' @keywords internal
#' @noRd
#' @returns character string (e.g. "'EPSG:4326'") or "NULL"
crs_to_sql <- function(x) {  # nocov start
  if (is.null(x)) return("NULL")
  if (inherits(x, "crs") && is.na(x)) return("NULL")
  if (is.atomic(x) && all(is.na(x))) return("NULL")

  if (inherits(x, "crs")) {
    if (!is.na(x$epsg)) return(paste0("'EPSG:", x$epsg, "'"))
    if (!is.null(x$wkt)) {
      # Escape single quotes for SQL
      val_clean <- gsub("'", "''", x$wkt)
      return(paste0("'", val_clean, "'"))
    }
    return("NULL")
  }

  if (is.numeric(x)) {
    return(paste0("'EPSG:", as.integer(x), "'"))
  }

  if (is.character(x)) {
    val_clean <- gsub("'", "''", x)
    return(paste0("'", val_clean, "'"))
  }

  return("NULL")
} # nocov end






#' Handle output type for duckspatial functions
#'
#' Returns a `sf` or a `duckspatial_df` from a query, depending
#' on the `mode` parameter or global options.
#'
#' @param query A query
#' @param conn DuckDB connection
#' @template mode
#' @template crs
#' @param x_geom Name of the geometry column
#'
#' @keywords internal
#' @noRd
#' @returns Object of the specified mode type
ddbs_handle_query <- function(
  query, 
  conn, 
  mode = NULL, 
  crs = NULL,
  x_geom = "geometry",
  fun_group = 1,
  units = NULL
) { # nocov start

  # First, handle simple data frames
  crs_is_na <- is.null(crs) || (inherits(crs, "crs") && is.na(crs)) || (is.atomic(crs) && all(is.na(crs)))
  if (crs_is_na && length(x_geom) == 0) {

    ## Create the table
    view_name <- ddbs_temp_table_name()
    DBI::dbExecute(
      conn, 
      glue::glue("CREATE TEMP TABLE {view_name} AS {query};")
    )

    ## Return a lazy table
    return(dplyr::tbl(conn, view_name))

  }
  
  # Resolve mode type: parameter > global option > default
  if (is.null(mode)) {
    mode <- getOption("duckspatial.mode", "duckspatial")
  }
  
  # Validate mode type
  valid_modes <- c("duckspatial", "sf")
  if (!mode %in% valid_modes) {
    cli::cli_abort(
      "{.arg mode} must be one of {.val {valid_modes}}, not {.val {mode}}."
    )
  }
  
  # Handle based on mode type
  if (mode == "sf") {

    ## Get the query as a data frame
    data_tbl <- DBI::dbGetQuery(conn, query)

    ## Manage sf output depending on the function group
    ## - Group 1: functions that return a normal {sf} object (most of the funs)
    ## - Group 2: functions that return a vector (units or not units)
    if (fun_group == 1) {

      ## Convert to sf object
      data_sf <- convert_to_sf_wkb(
        data       = data_tbl,
        crs        = crs,
        x_geom     = x_geom
      )
      return(data_sf)

    } else if (fun_group == 2) {
      
      ## Return units/non-units vector
      if (is.null(units)) {
        return(data_tbl[, 1])
      } else {
        return(units::as_units(data_tbl[, 1], units))
      }
      
    }
    
  } else {
    # mode == "duckspatial"
    # Create a view name and the query
    view_name <- ddbs_temp_table_name()
    query <- glue::glue("
      CREATE TEMP TABLE {view_name} AS
      {query};
    ")
    # on.exit(DBI::dbExecute(conn, glue::glue("DROP TABLE IF EXISTS {view_name}")))

    # Create the view
    DBI::dbExecute(conn, query)

    # Open lazily as duckspatial_df
    lazy_tbl <- duckdb::tbl_function(conn, view_name)

    result <- new_duckspatial_df(
      lazy_tbl, 
      crs = crs, 
      geom_col = x_geom, 
      source_table = view_name,
      source_conn = conn
    )
    
    return(result)
  }
  # nocov end
}



#' Get CRS from a spatial file
#'
#' @param path Path to the file
#' @param conn DuckDB connection
#' @return crs object or NULL
#' @keywords internal
#' @noRd
get_file_crs <- function(path, conn) {
    tryCatch({
      meta_query <- glue::glue("
        SELECT 
          layers[1].geometry_fields[1].crs.auth_name as auth_name,
          layers[1].geometry_fields[1].crs.auth_code as auth_code
        FROM st_read_meta('{path}')
      ")
      meta <- DBI::dbGetQuery(conn, meta_query)
      
      if (!is.na(meta$auth_code) && !is.na(meta$auth_name)) {
        crs_string <- paste0(meta$auth_name, ":", meta$auth_code)
        sf::st_crs(crs_string)
      } else {
        NULL
      }
    }, error = function(e) {
      cli::cli_warn("Could not auto-detect CRS from file: {e$message}")
      NULL
    })
}



#' Get or create default DuckDB connection with spatial extension installed and loaded
#'
#'
#' @param create Logical. If TRUE and no connection exists, create one.
#'   Default is TRUE.
#' @param ... Additional parameters to pass to `ddbs_create_conn()`
#'
#' @returns A `duckdb_connection` or NULL if no connection exists and
#'   create = FALSE
#'
#' @keywords internal
ddbs_default_conn <- function(create = TRUE, ...) {
  conn <- getOption("duckspatial_conn", NULL)

  # Check if existing connection is still valid

  if (!is.null(conn)) {
    if (!DBI::dbIsValid(conn)) {
      options(duckspatial_conn = NULL)
      conn <- NULL
    }
  }

  # Create new connection if needed
  if (is.null(conn) && create) {
    conn <- ddbs_create_conn(dbdir = "memory", ...)
    options(duckspatial_conn = conn)
  }

  # Activate macros
  if (!is.null(conn)) {
    create_duckspatial_macros(conn)
  }

  conn
}

#' Generate unique temporary view name
#'
#' Creates a unique name for temporary views to avoid collisions.
#'
#' @returns Character string with unique view name
#' @keywords internal
ddbs_temp_view_name <- function() { # nocov start
  paste0("temp_view_", gsub("-", "_", uuid::UUIDgenerate()))
} # nocov end

#' Generate unique temporary table name
#'
#' Creates a unique name for temporary tables to avoid collisions.
#'
#' @returns Character string with unique table name
#' @keywords internal
ddbs_temp_table_name <- function() { # nocov start
  paste0("temp_table_", gsub("-", "_", uuid::UUIDgenerate()))
} # nocov end

ddbs_checkpoint_if_possible <- function(conn) {
  if (!DBI::dbIsValid(conn)) {
    return(invisible(FALSE))
  }

  ok <- tryCatch({
    DBI::dbExecute(conn, "FORCE CHECKPOINT")
    TRUE
  }, error = function(e) {
    FALSE
  })

  invisible(ok)
}


#' Create an ephemeral DuckDB connection
#'
#' Creates a DuckDB connection that is automatically closed when the calling
#' function exits (either normally or due to an error). For file-based 
#' connections, the database file can also be automatically deleted on cleanup.
#'
#' @param file If TRUE, creates a file-based temporary database instead of 
#'   in-memory. If a character string, uses that as the database file path.
#'   Default is FALSE (in-memory).
#' @param read_only If TRUE and file is provided, opens the connection as 
#'   read-only. Has no effect on in-memory connections. Default is FALSE.
#' @param cleanup If TRUE (default), the connection will be closed (with 
#'   shutdown = TRUE for file-based) and for file-based connections, the 
#'   database file will be deleted.
#' @param envir The environment in which to schedule cleanup. Default is the
#'   parent frame (the caller's environment).
#' @template threads
#' @template memory_limit_gb
#' @param duckdb_storage_version Storage compatibility for newly created persistent
#'   native DuckDB files (\code{.duckdb}, \code{.db}, \code{.ddb}).
#'   \itemize{
#'     \item \code{"v1.5.0"} (\strong{Native Spatial Storage}, Default): Preserves
#'           CRS metadata in native DuckDB \code{GEOMETRY} columns. Requires
#'           DuckDB >= 1.5.0 to open the file.
#'     \item \code{"v1.0.0"} (\strong{Legacy Compatibility}): Creates
#'           files readable by older DuckDB versions (>= 1.0.0). Persists CRS
#'           metadata in duckspatial-managed column comments (a convention not
#'           recognized by other spatial software).
#'   }
#'
#' @returns A `duckdb_connection` that will be automatically closed on exit.
#'   For file-based connections, also returns the file path as an attribute 
#'   `db_file`.
#' @noRd
#' @keywords internal
ddbs_temp_conn <- function(file = FALSE, read_only = FALSE, cleanup = TRUE, 
                            envir = parent.frame(), threads = NULL, memory_limit_gb = NULL,
                            duckdb_storage_version = duckspatial_storage_default()) {

  assert_threads(threads)
  assert_memory_limit_gb(memory_limit_gb)
  duckdb_storage_version <- match_duckdb_storage_version(duckdb_storage_version)

  if (isTRUE(file) || is.character(file)) {
    # File-based connection
    if (is.character(file)) {
      db_file <- file
    } else {
      db_file <- tempfile(fileext = ".duckdb")
    }
    
    # IMPORTANT: DuckDB cannot open non-existent files in read-only mode
    # If creating a new tempfile with read_only=TRUE, we must create it first
    if (isTRUE(read_only) && !file.exists(db_file)) {
      # Create the database file first in writable mode
      conn_init <- ddbs_open_persistent(
        db_file,
        duckdb_storage_version = duckdb_storage_version,
        read_only = FALSE
      )
      ddbs_checkpoint_if_possible(conn_init)
      drv_init <- conn_init@driver
      DBI::dbDisconnect(conn_init)
      if (inherits(drv_init, "duckdb_driver")) {
        duckdb::duckdb_shutdown(drv_init)
      }
    }
    
    conn <- ddbs_open_persistent(
      db_file,
      duckdb_storage_version = duckdb_storage_version,
      read_only = read_only
    )
    
    # Checks and installs the Spatial extension
    # NOTE: These operations work fine on read-only connections:
    # - INSTALL writes to global extension dir (~/.duckdb/extensions), not the database
    # - LOAD just loads extension into session memory
    # - SET operations are session-level settings
    ddbs_install(conn, upgrade = FALSE, quiet = TRUE)
    ddbs_load(conn, quiet = TRUE, create_macros = !read_only)

    # Configure resources
    ddbs_set_resources(conn, threads = threads, memory_limit_gb = memory_limit_gb)

    # Cleanup: disconnect and optionally delete file
    withr::defer({
      if (DBI::dbIsValid(conn)) {
        if (!isTRUE(read_only)) {
          ddbs_checkpoint_if_possible(conn)
        }
        drv <- conn@driver
        tryCatch(suppressWarnings(DBI::dbDisconnect(conn)), error = function(e) NULL)
        if (inherits(drv, "duckdb_driver")) {
          tryCatch(duckdb::duckdb_shutdown(drv), error = function(e) NULL)
        }
      }
      if (isTRUE(cleanup) && file.exists(db_file)) unlink(db_file)
    }, envir = envir)
  } else {
    # In-memory connection
    conn <- ddbs_create_conn(
      dbdir = "memory",
      threads = threads,
      memory_limit_gb = memory_limit_gb,
      duckdb_storage_version = duckdb_storage_version
    )
    withr::defer({
      if (DBI::dbIsValid(conn)) {
        drv <- conn@driver
        tryCatch(suppressWarnings(DBI::dbDisconnect(conn)), error = function(e) NULL)
        if (inherits(drv, "duckdb_driver")) {
          tryCatch(duckdb::duckdb_shutdown(drv), error = function(e) NULL)
        }
      }
    }, envir = envir)
  }
  
  conn
}

#' Resolve connections and handle cross-connection imports
#'
#' @param x Input x (sf, duckspatial_df, tbl, character, etc.)
#' @param y Input y
#' @param conn Explicit target connection (optional)
#' @param conn_x Connection for x (optional, resolved if NULL)
#' @param conn_y Connection for y (optional, resolved if NULL)
#' @template quiet
#' 
#' @return List containing:
#'   - conn: The target connection to use
#'   - x: Updated x (may be new view name if imported)
#'   - y: Updated y (may be new view name if imported)
#'   - cleanup: A function to call on exit to drop temporary views
#' @keywords internal
#' @noRd
resolve_spatial_connections <- function(
  x, 
  y, 
  conn = NULL, 
  conn_x = NULL, 
  conn_y = NULL, 
  quiet = FALSE) {
    
    cleanup_funs <- list()
    add_cleanup <- function(fn) {
        cleanup_funs <<- c(cleanup_funs, list(fn))
    }
    
    # 1. Resolve source connections
    # If not provided, try to extract from objects
    # Note: Character inputs will return NULL from get_conn_from_input
    source_conn_x <- conn_x %||% get_conn_from_input(x)
    source_conn_y <- conn_y %||% get_conn_from_input(y)
    source_conn_ds <- attr(x, "source_conn")
    
    # 2. Determine target connection
    # Priority: explicit conn > conn_x > conn_y > default
    target_conn <- if (!is.null(conn)) {
        conn
    } else if (!is.null(source_conn_x)) {
        source_conn_x
    } else if (!is.null(source_conn_y)) {
        source_conn_y
    } else if (!is.null(source_conn_ds)) {
      source_conn_ds
    } else {
      ddbs_default_conn()
    }
  
    
    # 2.1 Validate target connection
    if (!DBI::dbIsValid(target_conn)) {
        cli::cli_abort("Target connection is not valid. Please provide a valid DuckDB connection.")
    }
    
    # 2.2 Warn if conn_x and conn_y differ but no explicit conn was provided
    if (is.null(conn) && 
        !is.null(source_conn_x) && 
        !is.null(source_conn_y) && 
        !identical(source_conn_x, source_conn_y)) {
        cli::cli_warn(c(
            "{.arg x} and {.arg y} come from different DuckDB connections.",
            "i" = "Using {.arg x}'s connection as the target. Provide {.arg conn} to override."
        ))
    }
    
    # 3. Handle imports if source connections differ from target
    
    # Check x
    # We only import x if it HAS a source connection that is different from target
    if (!is.null(source_conn_x) && !identical(target_conn, source_conn_x)) {
         # Need to import x
         cli::cli_warn(c(
            "{.arg x} and the target connection are different.",
            "i" = "Importing {.arg x} to the target connection.",
            "i" = "This may require materializing data."
         ))
         
         x_to_import <- x
         if (is.character(x)) {
             x_to_import <- tryCatch({
                 tbl_obj <- dplyr::tbl(source_conn_x, x)
                 suppressWarnings(as_duckspatial_df(tbl_obj))
             }, error = function(e) {
                 tryCatch(dplyr::tbl(source_conn_x, x), error = function(ex) x)
             })
         }
         
         res <- import_view_to_connection(target_conn, source_conn_x, x_to_import, quiet = quiet)
         x <- res$name
         
         add_cleanup(function() {
             tryCatch(DBI::dbExecute(target_conn, glue::glue("DROP VIEW IF EXISTS {res$name}")), error = function(e) NULL)
             if (is.function(res$cleanup)) tryCatch(res$cleanup(), error = function(e) NULL)
         })
    }
    
    # Check y
    # We only import y if it HAS a source connection that is different from target
    if (!is.null(source_conn_y) && !identical(target_conn, source_conn_y)) {
         # Need to import y
         cli::cli_warn(c(
            "{.arg y} and the target connection are different.",
            "i" = "Importing {.arg y} to the target connection.",
            "i" = "This may require materializing data."
         ))
         
         y_to_import <- y
         if (is.character(y)) {
             y_to_import <- tryCatch({
                 tbl_obj <- dplyr::tbl(source_conn_y, y)
                 suppressWarnings(as_duckspatial_df(tbl_obj))
             }, error = function(e) {
                 tryCatch(dplyr::tbl(source_conn_y, y), error = function(ex) y)
             })
         }
         
         res <- import_view_to_connection(target_conn, source_conn_y, y_to_import, quiet = quiet)
         y <- res$name
         
         add_cleanup(function() {
             tryCatch(DBI::dbExecute(target_conn, glue::glue("DROP VIEW IF EXISTS {res$name}")), error = function(e) NULL)
             if (is.function(res$cleanup)) tryCatch(res$cleanup(), error = function(e) NULL)
         })
    }
    
    list(
        conn = target_conn,
        x = x,
        y = y,
        cleanup = function() {
            for (fn in cleanup_funs) fn()
        }
    )
}



#' Builds query in geometries
#' 
#' @param fun Duckdb spatial function
#' @template mode
#'
#' @keywords internal
#' @noRd
build_geom_query <- function(fun, name, crs, mode) { # nocov start
  ## Get mode
  if (is.null(name) && mode == "sf") {
    ## If not creating a table, fallback to BLOB
    glue::glue("ST_AsWKB({fun})")
  } else {
    ## When creating a table in a connection, we preserve the CRS
    ## in the geometry column
    geom_field <- duckdb_geometry_type(conn = NULL, crs)
    glue::glue("{fun}::{geom_field}")
  }
} # nocov end


#' Gets the current mode
#' 
#' @param mode duckspatial, sf or NULL
#'
#' @keywords internal
#' @noRd
get_mode <- function(mode, name) { # nocov start

  ## If name is not NULL, the geospatial operations will create a new table,
  ## so we are interested in the duckspatial version (i.e. not add ST_AsWKB())
  ## This is the same as ignoring the option
  if (!is.null(name)) {
    return("duckspatial")
  } 

  ## Get the current option
  if (is.null(mode)) {
    mode <- getOption("duckspatial.mode", "duckspatial")
  }

  return(mode)

} # nocov end



#' Builds verbose query of ddbs_union
#'
#' @keywords internal
#' @noRd
build_union_sql <- function(
  by_feature, 
  x_geom, 
  y_geom = NULL,
  x_query, 
  y_query = NULL) { # nocov start
  if (!is.null(y_query)) {
    if (by_feature) {
      list(
        geom_call = glue::glue("ST_Union(v1.{x_geom}, v2.{y_geom})"),
        geom_alias = x_geom,
        from      = glue::glue(
          "(SELECT ROW_NUMBER() OVER () as rn, * FROM {x_query}) v1
           JOIN (SELECT ROW_NUMBER() OVER () as rn, * FROM {y_query}) v2
           ON v1.rn = v2.rn"
        )
      )
    } else {
      list(
        geom_call  = glue::glue("ST_Union_Agg(geom)"),
        geom_alias = x_geom,
        from       = glue::glue(
          "(SELECT {x_geom} as geom FROM {x_query}
            UNION ALL
            SELECT {y_geom} as geom FROM {y_query}) v1"
        )
      )
    }
  } else {
    list(
      geom_call  = glue::glue("ST_Union_Agg({x_geom})"),
      geom_alias = x_geom,
      from       = x_query
    )
  }
} # nocov end





#' Builds verbose query of ddbs_union
#'
#' @keywords internal
#' @noRd
build_union_query <- function(
  by_feature, 
  name,
  crs,
  mode, 
  name_query,
  x_geom, 
  y_geom = NULL, 
  x_query, 
  y_query = NULL) { # nocov start

  parts     <- build_union_sql(by_feature, x_geom, y_geom, x_query, y_query)
  geom_expr <- glue::glue("{build_geom_query(parts$geom_call, name, crs, mode)} as {parts$geom_alias}")

  if (!is.null(y_query)) {
    row_id  <- if (by_feature) "ROW_NUMBER() OVER () as row_id," else "1 as row_id,"
  } else {
    row_id  <- ""
  }

  prefix <- if (!is.null(name_query)) {
    glue::glue("CREATE TABLE {name_query} AS")
  } else {
    ""
  }

  glue::glue("
    {prefix}
    SELECT {row_id} {geom_expr}
    FROM {parts$from};
  ")
} # nocov end


#' Gets the CRS of a table's geometry column
#' 
#' @param conn DuckDB connection
#' @param geom_name Name of the geometry column
#' @param table_name Name of the table
#' 
#' @return CRS object or NULL if not found
#'
#' @keywords internal
#' @noRd
get_table_crs <- function(conn, geom_name, table_name) { # nocov start

  DBI::dbGetQuery(
    conn,
    glue::glue("
        SELECT 
            ST_CRS({geom_name}) AS crs 
        FROM 
            {table_name}
        LIMIT 1;")
    )$crs

} # nocov end


#' Formats the geometry type for DuckDB queries based on the CRS of the data
#' 
#' @param x An sf object or CRS object to extract the CRS from
#' 
#' @return A character string representing the geometry type for DuckDB
#'
#' @keywords internal
#' @noRd
get_geometry_type_duckdb <- function(x, conn = NULL) { # nocov start
  duckdb_geometry_type(conn, x)
} # nocov end




#' Helper to create a table during a geospatial operation
#' 
#' @returns TRUE invisibly after creating the table
#'
#' @keywords internal
#' @noRd
create_duckdb_table <- function(
  conn,
  name,
  query,
  overwrite,
  quiet
) { # nocov start
  ## Convenient names of table and/or schema.table
  name_list <- get_query_name(name)

  ## Overwrite handling
  overwrite_table(name_list$query_name, conn, quiet, overwrite)

  ## Create and execute the query
  tmp.query <- glue::glue("
      CREATE TABLE {name_list$query_name} AS
      {query}
  ")
  DBI::dbExecute(conn, tmp.query)
  feedback_query(quiet)
  return(invisible(TRUE))
} # nocov end




#' Helper to create a the predicate clause for a spatial join/filter
#' 
#' @returns A character string with the predicate clause
#'
#' @keywords internal
#' @noRd
generate_predicate_clause <- function(
  predicate,
  conn,
  x_list,
  y_list,
  x_geom,
  y_geom,
  distance,
  crs_x
) { # nocov start

  ## For ST_Dwithin we have an extra distance argument, and the duckdb functions
  ## differ based on the CRS units (ST_DWithin vs ST_DWithin_Spheroid)
  if (predicate == "ST_DWithin") {

      ## if distance is not specified, it will fall back to ST_Within
      if (is.null(distance)) {
          cli::cli_warn("{.val distance} wasn't specified. Using ST_Within.")
          distance <- 0
      }

      ## check the CRS units to use the right function
      crs_units <- crs_x$units_gdal
      if (crs_units != "metre") {

          ## When using Spheroid version, only point geometry is allowed
          geom_type_x <- ddbs_geometry_type(x_list$query_name, FALSE, conn)
          geom_type_y <- ddbs_geometry_type(y_list$query_name, FALSE, conn)
          if (!geom_type_x %in% c("POINT", "MULTIPOINT") || !geom_type_y %in% c("POINT", "MULTIPOINT")) {
              cli::cli_abort(c(
                  "ST_DWithin with non-meter units only supports POINT or MULTIPOINT geometries.",
                  "i" = "Detected types: {.val {geom_type_x}} and {.val {geom_type_y}}."
              ))
          }

          # st_predicate <- glue::glue("ST_DWithin_Spheroid(v1.{x_geom}, v2.{y_geom}, {distance})")
          st_predicate <- glue::glue("
              ST_DWithin_Spheroid(
                  ST_Point(ST_Y(v1.{x_geom}), ST_X(v1.{x_geom})), 
                  ST_Point(ST_Y(v2.{y_geom}), ST_X(v2.{y_geom})), 
                  {distance})
              ")
          if (crs_x$input != "EPSG:4326") {
              cli::cli_warn(c(
                "Inputs are in {.val {crs_x$input}}, not {.val EPSG:4326}.",
                "i" = "Distance calculations may be less accurate.",
                "i" = "Consider transforming to {.val EPSG:4326} or a projected CRS."
              ))
          }
      } else {
          st_predicate <- glue::glue("ST_DWithin(v1.{x_geom}, v2.{y_geom}, {distance})")
      }

  } else {
      ## In every other case, it's a simple binary predicate with no extra arguments
      st_predicate <- glue::glue("{predicate}(v1.{x_geom}, v2.{y_geom})")
  }

  return(st_predicate)
} # nocov end



#' Helper to validate that the CRS of two tables match before a spatial operation
#'
#' @keywords internal
#' @noRd
validate_xy_crs <- function(
  crs_x,
  crs_y,
  conn,
  x_list,
  y_list
) { # nocov start
  if (!is.null(crs_x) && !is.null(crs_y)) {
       if (!crs_equal(crs_x, crs_y)) {
         cli::cli_abort("The Coordinates Reference System of {.arg x} and {.arg y} is different.")
       }
    } else {
       assert_crs(conn, x_list$query_name, y_list$query_name)
    }
} # nocov end




#' Helper to create macros for duckspatial functions
#'
#' @keywords internal
#' @noRd
create_duckspatial_macros <- function(conn) { # nocov start

  macros <- list(
    
    # --- geometry validation
    "CREATE OR REPLACE MACRO ddbs_is_simple(geom) AS ST_IsSimple(geom);",
    "CREATE OR REPLACE MACRO ddbs_is_valid(geom) AS ST_IsValid(geom);",
    "CREATE OR REPLACE MACRO ddbs_is_closed(geom) AS ST_IsClosed(geom);",
    "CREATE OR REPLACE MACRO ddbs_is_ring(geom) AS ST_IsRing(geom);",
    "CREATE OR REPLACE MACRO ddbs_is_empty(geom) AS ST_IsEmpty(geom);",
    "CREATE OR REPLACE MACRO ddbs_geometry_type(geom) AS ST_GeometryType(geom);",

    # --- measure functions
    "CREATE OR REPLACE MACRO ddbs_area(geom) AS (
      CASE 
        WHEN ST_CRS(geom) = 'EPSG:4326' THEN ST_Area_Spheroid(ST_FlipCoordinates(geom))
        ELSE ST_Area(geom)
      END
    );",

    "CREATE OR REPLACE MACRO ddbs_length(geom) AS (
      CASE 
        WHEN ST_CRS(geom) = 'EPSG:4326' THEN ST_Length_Spheroid(ST_FlipCoordinates(geom))
        ELSE ST_Length(geom)
      END
    );",

    "CREATE OR REPLACE MACRO ddbs_perimeter(geom) AS (
      CASE 
        WHEN ST_CRS(geom) = 'EPSG:4326' THEN ST_Perimeter_Spheroid(ST_FlipCoordinates(geom))
        ELSE ST_Perimeter(geom)
      END
    );",

    # --- aggregation functions
    "CREATE OR REPLACE MACRO ddbs_union_agg(geom) AS ST_Union_Agg(geom);",
    "CREATE OR REPLACE MACRO ddbs_union(geom) AS ST_Union_Agg(geom);",
    "CREATE OR REPLACE MACRO ddbs_intersection_agg(geom) AS ST_Intersection_Agg(geom);",

    # --- coordinate operations
    "CREATE OR REPLACE MACRO ddbs_x(geom) AS ST_X(geom);",
    "CREATE OR REPLACE MACRO ddbs_y(geom) AS ST_Y(geom);",
    "CREATE OR REPLACE MACRO ddbs_m(geom) AS ST_M(geom);",
    "CREATE OR REPLACE MACRO ddbs_z(geom) AS ST_Z(geom);",

    # --- dimension operations
    "CREATE OR REPLACE MACRO ddbs_has_z(geom) AS ST_HasZ(geom);",
    "CREATE OR REPLACE MACRO ddbs_has_m(geom) AS ST_HasM(geom);",

    # --- format conversion
    "CREATE OR REPLACE MACRO ddbs_as_text(geom) AS ST_AsText(geom);",
    "CREATE OR REPLACE MACRO ddbs_as_wkb(geom) AS ST_AsWKB(geom);",
    "CREATE OR REPLACE MACRO ddbs_as_hexwkb(geom) AS ST_AsHexWKB(geom);",
    "CREATE OR REPLACE MACRO ddbs_as_geojson(geom) AS ST_AsGeoJSON(geom);",

    # --- extent functions
    "CREATE OR REPLACE MACRO ddbs_bbox(geom) AS (
      SELECT {
        xmin: ST_XMin(ST_Extent(geom)),
        ymin: ST_YMin(ST_Extent(geom)),
        xmax: ST_XMax(ST_Extent(geom)),
        ymax: ST_YMax(ST_Extent(geom))
      }
    );",

    # --- geometry processing general
    "CREATE OR REPLACE MACRO ddbs_boundary(geom) AS ST_Boundary(geom);",
    "CREATE OR REPLACE MACRO 
      ddbs_buffer(geometry, distance, num_triangles := 8, cap_style := 'CAP_ROUND', join_style := 'JOIN_ROUND', mitre_limit := 1) AS 
      ST_Buffer(geometry, distance::DOUBLE, num_triangles::INTEGER, cap_style::VARCHAR, join_style::VARCHAR, mitre_limit::DOUBLE);
    ",
    "CREATE OR REPLACE MACRO ddbs_centroid(geom) AS ST_Centroid(geom);",
    "CREATE OR REPLACE MACRO 
      ddbs_concave_hull(geom, ratio := 0.5, allow_holes := TRUE) AS 
      ST_ConcaveHull(geom, ratio, allow_holes);
    ",
    "CREATE OR REPLACE MACRO ddbs_convex_hull(geom) AS ST_ConvexHull(geom);",
    "CREATE OR REPLACE MACRO ddbs_exterior_ring(geom) AS ST_ExteriorRing(geom);",
    "CREATE OR REPLACE MACRO ddbs_voronoi(geom) AS ST_VoronoiDiagram(geom);",
    "CREATE OR REPLACE MACRO ddbs_build_area(geom) AS ST_BuildArea(geom);",
    "CREATE OR REPLACE MACRO ddbs_line_node(geom) AS ST_Node(geom);",
    "CREATE OR REPLACE MACRO ddbs_reduce_precision(geom, grid_size) AS ST_ReducePrecision(geom, grid_size::DOUBLE);",
    "CREATE OR REPLACE MACRO ddbs_reverse(geom) AS ST_Reverse(geom);",
    "CREATE OR REPLACE MACRO ddbs_normalize(geom) AS ST_Normalize(geom);"


  )

  invisible(lapply(macros, DBI::dbExecute, conn = conn))

} # nocov end




#' Checks if a extension is installed
#'
#' @keywords internal
#' @noRd
check_installed_extension <- function(conn = NULL, extension) {

  ## Resolve connection
  conn <- conn %||% ddbs_default_conn()
  dbConnCheck(conn)
  assert_character_scalar(extension, "extension")

  ## Get connection information
  ext_info <- DBI::dbGetQuery(
    conn,
    "SELECT * FROM duckdb_extensions() WHERE extension_name = ?",
    params = list(tolower(extension))
  )

  ## Error if extension is not installed
  if (isFALSE(ext_info$installed) || nrow(ext_info) == 0) {
    return(FALSE)
  } else {
    return(TRUE)
  }

}



#' Ensure a DuckDB extension is loaded
#'
#' Checks that an extension is installed (via `check_installed_extension`) and,
#' if it is not already loaded, loads it. Returns `TRUE` when the extension is
#' available and loaded, `FALSE` when it is not installed (and therefore cannot
#' be loaded).
#'
#' @keywords internal
#' @noRd
check_loaded_extension <- function(conn = NULL, extension) {

  ## Resolve connection
  conn <- conn %||% ddbs_default_conn()
  dbConnCheck(conn)
  assert_character_scalar(extension, "extension")

  ## Must be installed before it can be loaded
  if (isFALSE(check_installed_extension(conn, extension))) {
    return(FALSE)
  }

  ## Load it if it is not already loaded
  ext_info <- DBI::dbGetQuery(
    conn,
    "SELECT * FROM duckdb_extensions() WHERE extension_name = ?",
    params = list(tolower(extension))
  )

  if (isFALSE(ext_info$loaded)) {
    suppressMessages(DBI::dbExecute(conn, glue::glue("LOAD {extension};")))
  }

  TRUE

}

# Tests for duckdbfs interoperability
# All tests involving duckdbfs::open_dataset integration with duckspatial

# =============================================================================
# normalize_spatial_input with duckdbfs
# =============================================================================
testthat::skip_on_cran()

## Create connection
conn <- ddbs_default_conn()

## Try to install h3 before duckdbfs
h3_ext <- tryCatch(ddbs_install(conn, extension = "h3"), error = function(e) NULL)

## Skip tests if h3 installation failed
if (is.null(h3_ext)) testthat::skip()

test_that("normalize_spatial_input works for duckdbfs::open_dataset inputs", {
  skip_if_not_installed("duckdbfs")

  # Create a temporary parquet file

  conn <- ddbs_temp_conn()
  sf_obj <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(0, 0))), a = 1)
  ddbs_write_table(conn, sf_obj, "test_table")

  tmp_file <- tempfile(fileext = ".parquet")
  DBI::dbExecute(conn, glue::glue("COPY test_table TO '{tmp_file}' (FORMAT PARQUET)"))

  # Open with duckdbfs
  ds <- duckdbfs::open_dataset(tmp_file)

  # Should be converted to duckspatial_df
  result <- normalize_spatial_input(ds)
  expect_s3_class(result, "duckspatial_df")
  expect_s3_class(result, "tbl_duckdb_connection")

  # Clean up
  unlink(tmp_file)
})

# =============================================================================
# ddbs_join with duckdbfs inputs
# =============================================================================

test_that("ddbs_join works with raw duckdbfs tbl_duckdb_connection", {
  skip_if_not_installed("duckdbfs")

  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")

  # Both from duckdbfs - different connections but same CRS
  # Using open_dataset directly (no head) preserves source table for CRS detection
  conn1 <- ddbs_temp_conn()
  conn2 <- ddbs_temp_conn()

  countries1 <- duckdbfs::open_dataset(
    countries_path, format = "sf", conn = conn1
  )
  countries2 <- duckdbfs::open_dataset(
    countries_path, format = "sf", conn = conn2
  )

  # Cross-connection join should work with warning about different connections
  testthat::expect_warning(
    result <- ddbs_join(countries1, countries2, join = "intersects"),
    "connection are different"
  )

  testthat::expect_true(
    inherits(result, "duckspatial_df") || inherits(result, "sf")
  )
})

testthat::skip_on_cran()

test_that("normalize_spatial_input works for sf objects", {
  sf_obj <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(0, 0))), a = 1)
  expect_identical(duckspatial:::normalize_spatial_input(sf_obj), sf_obj)
})

test_that("normalize_spatial_input works for duckspatial_df objects", {
  conn <- duckspatial:::ddbs_temp_conn()
  
  sf_obj <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(0, 0))), a = 1)
  duckspatial::ddbs_write_table(conn, sf_obj, "test_table")
  
  ds_df <- duckspatial::ddbs_read_table(conn, "test_table")
  expect_identical(duckspatial:::normalize_spatial_input(ds_df), ds_df)
})

test_that("normalize_spatial_input works for tbl_duckdb_connection objects", {
  conn <- duckspatial:::ddbs_temp_conn()
  
  sf_obj <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(0, 0))), a = 1)
  duckspatial::ddbs_write_table(conn, sf_obj, "test_table")
  
  # tbl_obj <- dplyr::tbl(conn, "test_table")
  tbl_obj <- as_duckspatial_df("test_table", conn)
  result <- duckspatial:::normalize_spatial_input(tbl_obj)
  
  expect_s3_class(result, "duckspatial_df")
  expect_identical(attr(result, "source_table"), "test_table")
})

test_that("normalize_spatial_input works for character inputs", {
  conn <- duckspatial:::ddbs_temp_conn()
  
  DBI::dbExecute(conn, "CREATE TABLE test_table (a INTEGER)")
  
  # Valid existing table
  expect_identical(duckspatial:::normalize_spatial_input("test_table", conn), "test_table")
  
  # Missing connection
  expect_error(duckspatial:::normalize_spatial_input("test_table"), "required when using character table names")
  
  # Non-existent table
  expect_error(duckspatial:::normalize_spatial_input("non_existent", conn), "does not exist in connection")
})

test_that("normalize_spatial_input passes through unsupported types", {
  expect_identical(duckspatial:::normalize_spatial_input(123), 123)
  expect_identical(duckspatial:::normalize_spatial_input(NULL), NULL)
})

test_that("normalize_spatial_input works for duckspatial::ddbs_open_dataset inputs", {
  # Create a distinct temporary parquet file
  conn <- duckspatial:::ddbs_temp_conn()
  sf_obj <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(1, 1))), b = 2) # Different data
  duckspatial::ddbs_write_table(conn, sf_obj, "test_table_2")
  
  tmp_file_ds <- tempfile(fileext = ".parquet")
  DBI::dbExecute(conn, glue::glue("COPY test_table_2 TO '{tmp_file_ds}' (FORMAT PARQUET)"))
  
  # Open with duckspatial
  ds_ddbs <- duckspatial::ddbs_open_dataset(tmp_file_ds)
  
  # Should be passed through (as it's already duckspatial_df)
  result <- duckspatial:::normalize_spatial_input(ds_ddbs)
  expect_s3_class(result, "duckspatial_df")
  
  # Clean up
  unlink(tmp_file_ds)
})

testthat::skip_on_cran()

test_that("duckspatial_df uses efficient SQL render fallback when source_table is missing", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_temp_conn()
  
  # Setup: Create a table and a duckspatial_df from a query
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # Create a query-based duckspatial_df (no direct source table)
  # This typically results in source_table=NULL in as_duckspatial_df
  # providing we do it in a way that generates a query.
  
  lazy_tbl <- dplyr::tbl(conn, "nc_test") |> dplyr::filter(AREA > 0)
  
  # # Ensure it is a duckspatial_df
  ds <- as_duckspatial_df(lazy_tbl, crs = sf::st_crs(nc_sf))
  
  # Manually ensure source_table is NULL to force the fallback path logic
  attr(ds, "source_table") <- NULL
  
  # Call get_query_list (internal function)
  # We need to access it via triple colon or having it exported (it's not exported)
  # Since we are in the package tests, we can access internal functions if using devtools::test() 
  # but distinct test files run in separate environments. 
  # Usually testthat runs with access to internal functions.
  
  x_list <- get_query_list(ds, conn)
  
  # Verify we got a temp view name
  temp_view_name <- x_list$query_name
  expect_true(grepl("temp_view_", temp_view_name))
  
  # CLEANUP check: Register cleanup to occur on exit of this test block
  on.exit(x_list$cleanup(), add = TRUE)
  
  # REGRESSION CHECK:
  # The efficient method creates a VIEW using SQL.
  # The inefficient method used duckdb_register_arrow (via ddbs_write_table(temp_view=TRUE)),
  # which creates a replacement scan view (typically no SQL definition in duckdb_views or different).
  
  # Let's inspect the view definition
  view_info <- DBI::dbGetQuery(conn, glue::glue(
    "SELECT sql FROM duckdb_views() WHERE view_name = '{temp_view_name}'"
  ))
  
  expect_equal(nrow(view_info), 1)
  
  view_sql <- view_info$sql
  
  # The new efficient method should show the CREATE VIEW AS SELECT ... statement
  expect_false(is.null(view_sql))
  expect_false(is.na(view_sql))
  # It should contain the original query logic (e.g. "AREA" > 0)
  # TODO - REVIEW LATER
  # expect_true(grepl("AREA", view_sql) || grepl("area", view_sql, ignore.case = TRUE))
  expect_true(grepl("SELECT", view_sql, ignore.case = TRUE))
  
  # If it were the old method (Arrow registration), the SQL for the view is usually internal
  # or at least wouldn't reflect the SQL query "AREA > 0", because it would be serving
  # materialized data (points/polygons) from the Arrow stream.
})

testthat::skip_on_cran()

test_that("resolve_spatial_connections handles single connection correctly", {
  conn <- ddbs_temp_conn()
  
  # Case 1: All on same connection, explicit conn
  res <- resolve_spatial_connections("t1", "t2", conn = conn)
  expect_identical(res$conn, conn)
  expect_equal(res$x, "t1")
  expect_equal(res$y, "t2")
  
  # Case 2: Implicit connection from objects
  # We need get_conn_from_input to work. It uses dbplyr::remote_con(x) for tbl_lazy/duckspatial_df
  # For the test, we can use real tables to avoid mocking internals that might fail
  ddbs_write_table(conn, sf::st_sf(geom=sf::st_sfc(sf::st_point(1:2)), crs=4326), "t1")
  # t1_df <- dplyr::tbl(conn, "t1")
  # t2_df <- dplyr::tbl(conn, "t1") # reuse
  t1_df <- as_duckspatial_df("t1", conn)
  t2_df <- as_duckspatial_df("t1", conn) # reuse
  
  res2 <- resolve_spatial_connections(t1_df, t2_df)
  expect_identical(res2$conn, conn)
})

test_that("resolve_spatial_connections handles cross-connection correctly", {
  conn1 <- ddbs_temp_conn()
  conn2 <- ddbs_temp_conn()
  
  # Setup data
  ddbs_write_table(conn1, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(0,0))), crs=4326), "t1")
  ddbs_write_table(conn2, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(1,1))), crs=4326), "t2")
  
  # t2 (foreign) should be imported to conn1 (target)
  # Case: conn explicit
  expect_warning(
    expect_warning(
      res <- resolve_spatial_connections("t1", "t2", conn = conn1, conn_x = conn1, conn_y = conn2),
      "target connection are different"
    ),
    "Imported via collection"
  )
  
  expect_identical(res$conn, conn1)
  expect_equal(res$x, "t1")
  expect_false(res$y == "t2") # imported view name
  
  # Verify import worked
  # imported_y <- DBI::dbReadTable(conn1, res$y)
  imported_y <- ddbs_read_table(conn1, res$y)
  expect_equal(nrow(imported_y), 1)
  
  # Note: cleanup behavior depends on import strategy used
  # For in-memory DBs, collect+register is used which creates a table not a view
})

test_that("resolve_spatial_connections imports x when explicit conn differs", {
  conn_target <- ddbs_temp_conn()
  conn_source <- ddbs_temp_conn()
  
  ddbs_write_table(conn_source, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(0,0))), crs=4326), "t1")
  
  expect_warning(
    expect_warning(
      res <- resolve_spatial_connections("t1", "t2", conn = conn_target, conn_x = conn_source),
      "Importing `x` to the target connection"
    ),
    "Imported via collection"
  )
  
  expect_identical(res$conn, conn_target)
  # x should be imported (new name), y should be "t2" (assuming strict mode off or character pass-through)
  expect_false(res$x == "t1") 
  expect_equal(res$y, "t2")
})

test_that("resolve_spatial_connections defaults to conn_x if available", {
  conn1 <- ddbs_temp_conn()
  
  ddbs_write_table(conn1, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(0,0))), crs=4326), "t1")
  # t1_df <- dplyr::tbl(conn1, "t1")
  t1_df <- as_duckspatial_df("t1", conn1)
  
  res <- resolve_spatial_connections(t1_df, "t2")
  expect_identical(res$conn, conn1)
})

test_that("resolve_spatial_connections warns when conn_x != conn_y without explicit conn", {
  conn1 <- ddbs_temp_conn()
  conn2 <- ddbs_temp_conn()
  
  ddbs_write_table(conn1, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(0,0))), crs=4326), "t1")
  ddbs_write_table(conn2, sf::st_sf(geom=sf::st_sfc(sf::st_point(c(1,1))), crs=4326), "t2")
  
  # t1_df <- dplyr::tbl(conn1, "t1")
  # t2_df <- dplyr::tbl(conn2, "t2")
  t1_df <- as_duckspatial_df("t1", conn1)
  t2_df <- as_duckspatial_df("t2", conn2)
  
  # Should warn about different connections when no explicit conn provided
  expect_warning(
    expect_warning(
      expect_warning(
        res <- resolve_spatial_connections(t1_df, t2_df),
        "different DuckDB connections"
      ),
      "target connection are different"
    ),
    "Imported via collection" #"Imported via duckdb_register"
  )
  
  # Should use x's connection as target
  expect_identical(res$conn, conn1)
})

test_that("resolve_spatial_connections cleanup is always a function", {
  conn <- ddbs_temp_conn()
  
  res <- resolve_spatial_connections("t1", "t2", conn = conn)
  
  # cleanup should always be a function

  expect_true(is.function(res$cleanup))
  
  # Should be callable without error
  expect_no_error(res$cleanup())
})

testthat::skip_on_cran()

# Tests for import_view_to_connection -------------------------------------

test_that("Strategy 1: SQL recreation works for same DB (direct view)", {
  skip_if_not_installed("duckdb")
  
  # File-based DB required for testing same-file connections
  conn1 <- ddbs_temp_conn(file = TRUE)
  db_file <- attr(conn1, "db_file")
  
  DBI::dbExecute(conn1, "CREATE TABLE data (id INTEGER, val DOUBLE)")
  DBI::dbExecute(conn1, "INSERT INTO data VALUES (1, 1.1), (2, 2.2)")
  DBI::dbExecute(conn1, "CREATE VIEW my_view AS SELECT * FROM data")
  
  tbl_ref <- dplyr::tbl(conn1, "my_view")
  
  # Second connection to same file
  conn2 <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file)
  withr::defer(DBI::dbDisconnect(conn2), envir = parent.frame())
  
  expect_message(
    res <- import_view_to_connection(conn2, conn1, tbl_ref, "imported_view"),
    "SQL recreation"
  )
  
  expect_equal(res$method, "sql_recreation")
  df_res <- DBI::dbGetQuery(conn2, "SELECT * FROM imported_view ORDER BY id")
  expect_equal(nrow(df_res), 2)
})

test_that("Strategy 2: SQL render works for same DB (lazy query)", {
  skip_if_not_installed("duckdb")
  
  # File-based DB required for testing same-file connections
  conn1 <- ddbs_temp_conn(file = TRUE)
  db_file <- attr(conn1, "db_file")
  
  DBI::dbExecute(conn1, "CREATE TABLE data (id INTEGER, val DOUBLE)")
  DBI::dbExecute(conn1, "INSERT INTO data VALUES (1, 10), (2, 20)")
  
  tbl_lazy <- dplyr::tbl(conn1, "data") |> 
    dplyr::filter(id == 1) |>
    dplyr::mutate(val = val * 2)
  
  # Second connection to same file
  conn2 <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file, read_only = TRUE)
  withr::defer(DBI::dbDisconnect(conn2, shutdown = TRUE), envir = parent.frame())
  
  expect_message(
    res <- import_view_to_connection(conn2, conn1, tbl_lazy, "imported_lazy"),
    "SQL query"
  )
  
  expect_equal(res$method, "sql_render")
  df <- DBI::dbGetQuery(conn2, "SELECT * FROM imported_lazy")
  expect_equal(nrow(df), 1)
})

test_that("Strategy 3: ATTACH works for file-based cross-connection", {

  skip_if_not_installed("duckdb")
  
  # Create source file, write data, then close
  source_file <- tempfile(fileext = ".duckdb")
  source_conn <- DBI::dbConnect(duckdb::duckdb(), dbdir = source_file)
  DBI::dbExecute(source_conn, "CREATE TABLE test_data (id INTEGER, name TEXT)")
  DBI::dbExecute(source_conn, "INSERT INTO test_data VALUES (1, 'a'), (2, 'b')")
  DBI::dbDisconnect(source_conn)
  
  # Re-connect as read-only using helper
  source_conn <- ddbs_temp_conn(file = source_file, read_only = TRUE)
  tbl_ref <- dplyr::tbl(source_conn, "test_data")
  
  # Target: memory DB
  target_conn <- ddbs_temp_conn()
  
  # Since dbGetInfo may return NULL for read-only, ATTACH might not trigger
  res <- import_view_to_connection(target_conn, source_conn, tbl_ref, "attached_view")
  
  # Accept any successful method
  expect_true(res$method %in% c("attach", "nanoarrow", "duckdb_register"))
})

test_that("Strategy 4: Nanoarrow streaming works for cross-memory-DB", {

  skip_if_not_installed("duckdb")
  skip_if_not_installed("nanoarrow")
  
  conn1 <- ddbs_temp_conn()
  conn2 <- ddbs_temp_conn()
  
  DBI::dbExecute(conn1, "CREATE TABLE data (id INTEGER, val TEXT)")
  DBI::dbExecute(conn1, "INSERT INTO data VALUES (1, 'a'), (2, 'b')")
  
  tbl_ref <- dplyr::tbl(conn1, "data")
  
  expect_warning(
    res <- import_view_to_connection(conn2, conn1, tbl_ref, "arrow_imported"),
    "Imported via"
  )
  
  expect_true(res$method %in% c("nanoarrow", "duckdb_register"))
})

test_that("Strategy 5: Collect+sf works for duckspatial_df", {

  skip_if_not_installed("duckdb")
  skip_if_not_installed("sf")
  
  conn1 <- ddbs_temp_conn()
  conn2 <- ddbs_temp_conn()
  
  pts <- sf::st_sfc(sf::st_point(c(0, 0)), sf::st_point(c(1, 1)), crs = 4326)
  sf_data <- sf::st_sf(id = 1:2, geometry = pts)
  
  dsdf <- as_duckspatial_df(sf_data, conn = conn1)
  
  expect_warning(
    res <- import_view_to_connection(conn2, conn1, dsdf, "sf_imported"),
    "Imported via collection"
  )
  
  expect_true(res$method %in% c("nanoarrow", "collect_and_write"))
})

test_that("Quoted identifiers work correctly", {
  skip_if_not_installed("duckdb")
  
  conn1 <- ddbs_temp_conn(file = TRUE)
  db_file <- attr(conn1, "db_file")
  
  DBI::dbExecute(conn1, "CREATE TABLE \"quoted table\" (id INTEGER)")
  DBI::dbExecute(conn1, "CREATE VIEW \"quoted view\" AS SELECT * FROM \"quoted table\"")
  
  tbl_ref <- dplyr::tbl(conn1, "quoted view")
  
  conn2 <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file)
  withr::defer(DBI::dbDisconnect(conn2, shutdown = TRUE), envir = parent.frame())
  
  expect_message(
    res <- import_view_to_connection(conn2, conn1, tbl_ref, "imported_quoted"),
    "SQL recreation"
  )
  
  expect_equal(res$method, "sql_recreation")
})

testthat::skip_on_cran()

test_that("get_query_list handles sf objects", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_temp_conn()
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  
  # Action
  res <- get_query_list(nc_sf, conn)
  
  # Verify
  expect_true(grepl("temp_view_", res$query_name))
  expect_true(DBI::dbExistsTable(conn, res$query_name))
  
  # Verify cleanup
  res$cleanup()
  expect_disjoint(
    res$query_name,
    ddbs_list_tables(conn)$table_name
  )
  expect_false(DBI::dbExistsTable(conn, res$query_name))
})

test_that("get_query_list handles duckspatial_df with source_table", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_temp_conn()
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  ddbs_write_table(conn, nc_sf, "nc_source", quiet = TRUE)
  
  ds <- as_duckspatial_df("nc_source", conn = conn, crs = sf::st_crs(nc_sf))
  
  # Action
  res <- get_query_list(ds, conn)
  
  # Verify - should return source name directly
  # expect_equal(res$query_name, "nc_source")
  expect_in(
    "nc_source",
    ddbs_list_tables(conn)$table_name
  )
  
  # Cleanup should be no-op (source table remains)
  res$cleanup()
  # expect_true(DBI::dbExistsTable(conn, "nc_source"))
  expect_in(
    "nc_source",
    ddbs_list_tables(conn)$table_name
  )
})

test_that("get_query_list handles duckspatial_df WITHOUT source_table (efficient fallback)", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_temp_conn()
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # Create query-based duckspatial_df
  ds <- as_duckspatial_df("nc_test", conn) |> dplyr::filter(AREA > 0)
  
  # Force source_table to NULL to trigger fallback
  attr(ds, "source_table") <- NULL
  
  # Action
  res <- get_query_list(ds, conn)
  
  # Verify
  expect_true(grepl("temp_view_", res$query_name))
  
  # Check it's a view with SQL definition (proving sql_render was used)
  view_def <- DBI::dbGetQuery(conn, glue::glue("SELECT sql FROM duckdb_views() WHERE view_name = '{res$query_name}'"))
  expect_true(nrow(view_def) == 1)
  expect_true(grepl("SELECT", view_def$sql, ignore.case = TRUE))
  
  # Verify cleanup
  res$cleanup()
  expect_false(DBI::dbExistsTable(conn, res$query_name))
})

test_that("get_query_list handles tbl_lazy objects", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_temp_conn()
  DBI::dbWriteTable(conn, "mtcars_test", mtcars)
  
  tbl_lazy <- dplyr::tbl(conn, "mtcars_test") |> dplyr::filter(mpg > 20)
  
  # Action
  res <- get_query_list(tbl_lazy, conn)
  
  # Verify
  expect_true(grepl("temp_view_", res$query_name))
  
  # Check view
  view_def <- DBI::dbGetQuery(conn, glue::glue("SELECT sql FROM duckdb_views() WHERE view_name = '{res$query_name}'"))
  expect_true(nrow(view_def) == 1)
  expect_true(grepl("mpg > 20", view_def$sql) || grepl('"mpg" > 20.0', view_def$sql))
  
  # Verify cleanup
  res$cleanup()
  expect_false(DBI::dbExistsTable(conn, res$query_name))
})

test_that("get_query_list handles character inputs", {
  conn <- ddbs_temp_conn()
  DBI::dbWriteTable(conn, "plain_table", data.frame(id = 1))
  
  # Action
  res <- get_query_list("plain_table", conn)
  
  # Verify
  expect_equal(res$query_name, "plain_table")
  expect_equal(res$table_name, "plain_table")
  
  # Cleanup is no-op
  res$cleanup()
  expect_true(DBI::dbExistsTable(conn, "plain_table"))
})

testthat::skip_on_cran()

test_that("Strategy 1: SQL Recreation (View-to-View)", {
  skip_if_not_installed("sf")
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })

  # Setup: Create a view in conn1 explicitly via SQL
  # Strategy 1 ONLY works for SQL views (registered in duckdb_views),
  # NOT for arrow/replacement scans created by register()
  DBI::dbExecute(conn1, "CREATE VIEW view1 AS SELECT 1 as id, 2 as val")
  
  # Check if strategy 1 is used
  res <- duckspatial:::import_view_to_connection(conn2, conn1, dplyr::tbl(conn1, "view1"))
  
  expect_equal(res$method, "sql_recreation")
  
  # Verify data in conn2
  n <- DBI::dbGetQuery(conn2, glue::glue("SELECT count(*) as n FROM {res$name}"))$n
  expect_equal(n, 1)
})

test_that("Strategy 2: SQL Render (Query-to-View)", {
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })
  
  # Ensure clean slate
  # Strategy 2 relies on the query referencing tables that exist in the target
  
  # 1. Create a physical table in BOTH connections
  # We use dbWriteTable to ensure it's a real table, not a view
  df <- data.frame(id = 1:5)
  DBI::dbWriteTable(conn1, "t1", df)
  DBI::dbWriteTable(conn2, "t1", df)

  # 2. Create lazy query in conn1
  q <- dplyr::tbl(conn1, "t1") |> dplyr::filter(id > 2)
  
  # 3. Import to conn2
  # Should use SQL render because "t1" exists in conn2 (same name)
  res <- duckspatial:::import_view_to_connection(conn2, conn1, q)
  
  expect_equal(res$method, "sql_render")
  
  # Verify
  n <- DBI::dbGetQuery(conn2, glue::glue("SELECT count(*)::int as n FROM {res$name}"))$n
  expect_equal(n, 3)
})

test_that("Strategy 3: ATTACH (File-based DB)", {
  conn_target <- ddbs_create_conn()
  
  # Create a file-based source DB
  db_file <- tempfile(fileext = ".duckdb")
  # Open read-write to populate
  conn_init <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file, read_only = FALSE)
  
  # Create PERMANENT table
  df <- data.frame(id = 1:10)
  DBI::dbWriteTable(conn_init, "persistent_table", df)
  
  # Close init connection to release write lock
  DBI::dbDisconnect(conn_init, shutdown = TRUE)
  
  # Now open source connection as READ_WRITE (simulating user having it open)
  # NOTE: ATTACH requires that we can access the file.
  # If we open it here as READ_ONLY, DBI::dbGetInfo(conn)$dbdir should return the path.
  # Let's verify what happens if we open it normally.
  conn_source <- DBI::dbConnect(duckdb::duckdb(), dbdir = db_file, read_only = TRUE)
  
  on.exit({
    ddbs_stop_conn(conn_target)
    DBI::dbDisconnect(conn_source, shutdown = TRUE)
    if (file.exists(db_file)) unlink(db_file)
  }, add = FALSE)

  # Verify connection info explicitly
  info <- DBI::dbGetInfo(conn_source)
  # Debug: Check what's happening
  if (is.na(info$dbname) || info$dbname == ":memory:") {
      warning("Connection is memory based! path: ", info$dbname)
  }
  
  # Import
  # Strategy 3 should work because conn_source points to a file that exists
  res <- duckspatial:::import_view_to_connection(conn_target, conn_source, dplyr::tbl(conn_source, "persistent_table"))
  
  expect_equal(res$method, "attach")
  
  # Verify
  n <- DBI::dbGetQuery(conn_target, glue::glue("SELECT count(*) as n FROM {res$name}"))$n
  expect_equal(n, 10)
  
  # Cleanup
  res$cleanup()
})

test_that("Strategy 4: Nanoarrow Streaming (Force fallback via mocks or specific conditions)", {
  skip_if_not_installed("nanoarrow")
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })
  
  # Strategy 1, 2, 3 must fail for 4 to trigger
  # 1 fails: not a view
  # 2 fails: table doesn't exist in target
  # 3 fails: in-memory DB
  
  df <- data.frame(a = 1:10)
  duckdb::duckdb_register(conn1, "local_table", df)
  
  q <- dplyr::tbl(conn1, "local_table") |> dplyr::filter(a < 5)
  
  # This should skip 1 (not view), skip 2 (local_table not in conn2), skip 3 (memory)
  # And hit Strategy 4 (Nanoarrow)
  
  # NOTE: Depending on nanoarrow/duckdb versions, this might fail and hit Strategy 5
  # We check if we got EITHER 4 or 5, but aim for 4 if environment supports it
  
  # Suppress the "Imported via..." info messages and warnings
  suppressWarnings({
     res <- duckspatial:::import_view_to_connection(conn2, conn1, q)
  })
 
  # If nanoarrow works, method is "nanoarrow". If not, it falls back to "duckdb_register" (Strategy 5 for DF)
  expect_true(res$method %in% c("nanoarrow", "duckdb_register", "collect_and_write"))
})

test_that("Strategy 5: Collect Fallback works", {
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })
  
  # Create a query that fails Nanoarrow (e.g., complex type not supported or mock error)
  # Easier: Just verify that if we pass a raw dataframe (not a remote source), it goes to strategy 5 logic
  # But import_view_to_connection takes a remote source.
  
  # Verification: Strategy 5 is valid if it produces correct results
  df <- data.frame(id = 1:3)
  duckdb::duckdb_register(conn1, "t5", df)
  
  suppressWarnings({
    res <- duckspatial:::import_view_to_connection(conn2, conn1, dplyr::tbl(conn1, "t5"))
  })
  n <- DBI::dbGetQuery(conn2, glue::glue("SELECT count(*) as n FROM {res$name}"))$n
  expect_equal(n, 3)
})

#' @param conn_x A `DBIConnection` object to a DuckDB database for the input `x`. 
#'        If `NULL` (default), it is resolved from `conn` or extracted from `x`.
#' @param conn_y A `DBIConnection` object to a DuckDB database for the input `y`. 
#'        If `NULL` (default), it is resolved from `conn` or extracted from `y`.
#' 

#' Template for unary functions without extra arguments (e.g. ST_Centroid)
#'
#' @template x
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#' @param fun The duckdb function to use
#' @param other_params string with other function-specific parameters
#' 
#' @template returns_mode
#' @keywords internal
#' @noRd
template_unary_ops <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE,
  fun,
  other_args = NULL,
  additional_clauses = NULL,
  drop_crs = FALSE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the function arguments
  ## if other_args is NULL, use only the geometry column name
  ## if is not NULL, append the rest of the function arguments
  if (is.null(other_args)) {
    args <- x_geom
  } else {
    args <- sprintf(
      "%s, %s",
      x_geom,
      other_args
    )
  }

  ## Additional clauses for some functions
  additional_clauses <- if (is.function(additional_clauses)) {
    additional_clauses(x_geom)
  } else {
    additional_clauses %||% ""
  }

  ## 2.3. Other function-specific handling
  ## - ST_Buffer, check the units and warn if they aren't in meters
  if (tolower(fun) == "st_buffer") {
    crs_units <- crs_x$units_gdal
    if (crs_units != "metre") cli::cli_warn("The input CRS is in {crs_units}s. This function calculates the buffer in those units.")
  }

  ## 2.4. Build the base query (depends on the output type - sf, duckspatial_df, table)
  ## When the operation moves geometries out of their CRS (e.g. ST_AsMVTGeom,
  ## which maps to tile pixel coordinates), the output must not carry the source
  ## CRS metadata.
  out_crs <- if (drop_crs) sf::st_crs(NA) else crs_x
  st_function <- glue::glue("{fun}({args})")
  base.query <- glue::glue("
    SELECT *
    REPLACE ({build_geom_query(st_function, name, out_crs, mode)} AS {x_geom})
    FROM {x_list$query_name}
    {additional_clauses};
  ")


  # 3. Table creation if name is provided, or
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = out_crs,
      x_geom = x_geom
    )
  }

}






#' Template for geom converstion funs (e.g. ST_AsWKB)
#'
#' @template x
#' @template conn_null
#' @param fun The duckdb function to use
#' 
#' @returns Character vector or list
#' @keywords internal
#' @noRd
template_geometry_conversion <- function(
  x,
  conn = NULL,
  fun
) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_character(conn, x)


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  sf_col_x <- attr(x, "sf_column")

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = TRUE)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)


  # 3. Create the query and retrieve the results

  ## 3.1. create query
  tmp.query <- glue::glue("
      SELECT {fun}({x_geom}) as geometry
      FROM {x_list$query_name};
  ")

  ## 3.2. retrieve results from the query
  data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)
  data_vec <- data_tbl$geometry

  return(data_vec)

}





#' Template for parsing serialized geometries into a spatial object
#'
#' Inverse of \code{template_geometry_conversion()}: takes an R vector of
#' serialized geometries (WKT, HEXWKB, GeoJSON or a list of raw WKB vectors)
#' and builds a geometry column via a DuckDB \code{ST_GeomFrom*} function.
#'
#' @param x Character vector (or list of raw vectors for WKB) of serialized
#'   geometries.
#' @param crs CRS to assign to the resulting geometries.
#' @param ... Named vectors of additional attribute columns.
#' @param geom_col Name of the output geometry column.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#' @param fun The DuckDB \code{ST_GeomFrom*} function to use.
#' @param input Either \code{"character"} (text input) or \code{"blob"} (raw WKB).
#'
#' @keywords internal
#' @noRd
template_geom_from <- function(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE,
  fun,
  input = c("character", "blob")) {

  input <- match.arg(input)

  # 0. Validate inputs
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  assert_character_scalar(geom_col, "geom_col")

  if (input == "character") {
    if (!is.character(x)) {
      cli::cli_abort("{.arg x} must be a character vector of serialized geometries.")
    }
  } else {
    if (!is.list(x) || !all(vapply(x, is.raw, logical(1)))) {
      cli::cli_abort("{.arg x} must be a list of raw vectors (WKB).")
    }
  }

  n <- length(x)

  dots <- list(...)
  if (length(dots) > 0) {
    if (is.null(names(dots)) || any(names(dots) == "")) {
      cli::cli_abort("All arguments passed via {.arg ...} must be named.")
    }
    conflict <- intersect(names(dots), geom_col)
    if (length(conflict) > 0) {
      cli::cli_abort(
        "Column name{?s} {.val {conflict}} in {.arg ...} conflict with the geometry column name."
      )
    }
    bad_len <- names(dots)[vapply(dots, length, integer(1)) != n]
    if (length(bad_len) > 0) {
      cli::cli_abort(
        "Extra column{?s} {.val {bad_len}} must have the same length as {.arg x} ({n})."
      )
    }
  }

  if (!is.null(name) && is.null(conn)) {
    cli::cli_abort(
      "If {.arg name} is provided, {.arg conn} must be a valid DuckDB connection."
    )
  }

  # 1. Resolve connection and mode
  target_conn <- conn %||% ddbs_default_conn()
  mode        <- get_mode(mode, name)

  # 2. Build a data frame with the serialized column (+ extras) and write a temp
  #    table. Use an index column to guarantee the right number of rows when the
  #    serialized column is a list of raw vectors.
  scol <- ".ddbs_serialized"
  df   <- data.frame(.ddbs_idx = seq_len(n))
  df[[scol]] <- x
  for (nm in names(dots)) df[[nm]] <- dots[[nm]]
  df[[".ddbs_idx"]] <- NULL

  tmp_name <- ddbs_temp_table_name()
  DBI::dbWriteTable(target_conn, tmp_name, df, temporary = TRUE)
  on.exit(DBI::dbRemoveTable(target_conn, tmp_name), add = TRUE)

  # 3. Build the ST_GeomFrom* expression (cast to GEOMETRY so build_geom_query
  #    can further annotate it with the CRS)
  st_fun    <- glue::glue('({fun}("{scol}")::GEOMETRY)')
  geom_expr <- build_geom_query(st_fun, name, crs, mode)

  # 4. Build the SELECT clause, dropping the serialized column and keeping extras
  select_clause <- if (length(dots) > 0) {
    glue::glue('SELECT * EXCLUDE ("{scol}"),')
  } else {
    "SELECT"
  }
  base_query <- glue::glue(
    '{select_clause} {geom_expr} AS "{geom_col}" FROM "{tmp_name}";'
  )

  # 5. Handle output
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base_query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base_query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs,
      x_geom = geom_col
    )
  }

}





#' Template for measure functions that return a vector (e.g. ST_Area)
#'
#' @template x
#' @template conn_null
#' @template name
#' @template new_column
#' @template mode
#' @template overwrite
#' @template quiet
#' @param fun The duckdb function to use
#' 
#' @returns When `new_column = NULL` it returns a `units` vector in \eqn{m^2}. When `new_column` is not NULL, the
#' output depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
#'   \itemize{
#'     \item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
#'     \item \code{sf}: An eagerly collected \code{sf} object in R memory.
#' }
#' When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
#' 
#' @keywords internal
#' @noRd
template_measure <- function(
  x,
  conn = NULL,
  name = NULL,
  new_column = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE,
  fun = c("ST_Area", "ST_Length", "ST_Perimeter")) {
  
  # Match and validate fun
  fun <- match.arg(fun)

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_character_scalar(new_column, "new_column")
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  
  if (!is.null(name) && is.null(new_column)) {
      cli::cli_abort("Please, specify the {.arg new_column} name.")
  }

  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x     <- ddbs_crs(x, conn)
  sf_col_x  <- attr(x, "sf_column")
  mode      <- get_mode(mode, name)
  crs_units <- crs_x$units_gdal

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 3.2. Warn if the units aren't meters or EPSG:4326
  ## for EPSG:4326, we can use ST_*_Spheroid to get the measurement in meters
  ## so that will be an exception
  if (crs_units != "metre" && !crs_x$input %in% c("EPSG:4326", "WGS 84")) {
      cli::cli_warn(
        "Input is in {.val {crs_x$input}}, not {.val EPSG:4326}. {fun} calculations may be less accurate. Consider transforming to {.val EPSG:4326} or a projected CRS."
      )
  }

  ## 3.3. Build the appropriate ST function based on fun and CRS
  ## Use spheroid version for geographic coordinates
  if (crs_units == "metre") {
    st_function <- glue::glue("{fun}({x_geom})")
  } else {
    # st_function <- glue::glue("{fun}_Spheroid({x_geom})") # when the issue #109 is solved
    st_function <- glue::glue("{fun}_Spheroid(ST_FlipCoordinates({x_geom}))")
  }
  
  ## 3.4. Determine units for output
  output_units <- switch(
    fun,
    "ST_Area"      = "m^2",
    "ST_Length"    = "metre",
    "ST_Perimeter" = "metre"
  )

  ## 3.5. Build the base query. For sf we will return an units vector
  if (mode == "sf") {
    base.query <- glue::glue("
      SELECT {st_function} AS {new_column}
      FROM {x_list$query_name};
    ")
  } else {
    base.query <- glue::glue("
      SELECT 
        * EXCLUDE {x_geom},
        {st_function} AS {new_column},
        {build_geom_query(x_geom, name, crs_x, mode)} AS {x_geom}
      FROM 
        {x_list$query_name};
    ")
  }


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query      = base.query,
      conn       = target_conn,
      mode       = mode,
      crs        = crs_x,
      x_geom     = x_geom,
      fun_group  = 2,
      units      = output_units
    )
  }

}






#' Template for unary functions without extra arguments (e.g. ST_IsValid)
#'
#' @template x
#' @template by_feature
#' @template conn_null
#' @template name
#' @template new_column
#' @template mode
#' @template overwrite
#' @template quiet
#' @param fun The duckdb function to use
#' 
#' @returns
#' \itemize{
#'   \item \code{mode = "duckspatial"} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
#'   \item \code{mode = "sf"}: An eagerly collected vector in R memory.
#'   \item When \code{name} is provided: writes the table in the DuckDB connection and returns \code{TRUE} (invisibly).
#' }
#' 
#' @keywords internal
#' @noRd
template_new_column <- function(
  x,
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  new_column = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE,
  fun) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_logic(by_feature, "by_feature")
  assert_name(name)
  assert_character_scalar(new_column, "new_column")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)

  if (!is.null(name) && is.null(new_column)) cli::cli_abort("Please, specify the {.arg new_column} name.")


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Compute if by_feature = FALSE (returns always a single value)
  ## - For functions ST_Has*() - if 1 is TRUE, return TRUE
  ## - For functions ST_is_*() - if 1 is FALSE, return FALSE
  ## - For coordinate bound functions - aggregate in SQL, return global max/min
  if (isFALSE(by_feature)) {
    fun_lower <- tolower(fun)

    ## For coordinate max functions: return the global maximum across all features
    if (fun_lower %in% c("st_xmax", "st_ymax", "st_zmax", "st_mmax")) {
      tmp.query <- glue::glue("
        SELECT MAX({fun}({x_geom})) as {new_column}
        FROM {x_list$query_name};
      ")
      data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)
      return(max(data_tbl[1, ]))
    }

    ## For coordinate min functions: return the global minimum across all features
    if (fun_lower %in% c("st_xmin", "st_ymin", "st_zmin", "st_mmin")) {
      tmp.query <- glue::glue("
        SELECT MIN({fun}({x_geom})) as {new_column}
        FROM {x_list$query_name};
      ")
      data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)
      return(min(data_tbl[1, ]))
    }

    ## Original behavior for logical functions
    tmp.query <- glue::glue("
      SELECT {fun}({x_geom}) as {new_column}
      FROM {x_list$query_name};
    ")
    data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)

    ## Return TRUE if any is TRUE
    if (fun_lower %in% c("st_hasz", "st_hasm")) {
      return(any(data_tbl[1, ]))
    } else {
      return(all(data_tbl[1, ]))
    }
  }

  ## 2.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
  st_function <- glue::glue("{x_geom}")

  if (mode == "sf") {
    base.query <- glue::glue("
      SELECT {fun}({x_geom}) as {new_column},
      FROM {x_list$query_name};
    ")
  } else {
    base.query <- glue::glue("
      SELECT 
        * EXCLUDE {x_geom},
        {fun}({x_geom}) AS {new_column},
        {build_geom_query(x_geom, name, crs_x, mode)} AS {x_geom}
      FROM 
        {x_list$query_name};
    ")
  }


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom,
      fun_group  = 2,
      units      = NULL
    )
  }

}

#' Geometry binary operations
#'
#' Perform geometric set operations between two sets of geometries.
#'
#' @details
#' These functions perform different geometric set operations:
#' \describe{
#'   \item{`ddbs_intersection`}{Returns the geometric intersection of two sets
#'   of geometries, producing the area, line, or point shared by both.}
#'   \item{`ddbs_crop`}{Returns the geometric intersection of two sets of
#'   geometries, using the bounding box of `y`, rather than its original geometry}
#'   \item{`ddbs_difference`}{Returns the portion of the first geometry that
#'   does not overlap with the second geometry.}
#'   \item{`ddbs_sym_difference`}{Returns the portions of both geometries
#'   that do not overlap with each other. Equivalent to
#'   `(A - B) UNION (B - A)`.}
#'   \item{`ddbs_shortest_line`}{Returns a LINESTRING connecting the closest
#'   points between each pair of geometries from \code{x} and \code{y}. Performs
#'   a cross join (all combinations), so the output contains columns from both
#'   \code{x} and \code{y} (excluding \code{y}'s geometry column). For each pair
#'   the line has the same length as \code{ST_Distance}.}
#' }
#'
#' @template x
#' @template y
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#' library(sf)
#'
#' # Create two overlapping polygons for testing
#' poly1 <- st_polygon(list(matrix(c(
#'   0, 0,
#'   4, 0,
#'   4, 4,
#'   0, 4,
#'   0, 0
#' ), ncol = 2, byrow = TRUE)))
#'
#' poly2 <- st_polygon(list(matrix(c(
#'   2, 2,
#'   6, 2,
#'   6, 6,
#'   2, 6,
#'   2, 2
#' ), ncol = 2, byrow = TRUE)))
#'
#' x <- st_sf(id = 1, geometry = st_sfc(poly1), crs = 4326)
#' y <- st_sf(id = 2, geometry = st_sfc(poly2), crs = 4326)
#'
#' # Visualize the input polygons
#' plot(st_geometry(x), col = "lightblue", main = "Input Polygons")
#' plot(st_geometry(y), col = "lightcoral", add = TRUE, alpha = 0.5)
#'
#' # Intersection: only the overlapping area (2,2 to 4,4)
#' result_intersect <- ddbs_intersection(x, y)
#' plot(st_geometry(result_intersect), col = "purple", 
#'      main = "Intersection")
#'
#' # Difference: part of x not in y (L-shaped area)
#' result_diff <- ddbs_difference(x, y)
#' plot(st_geometry(result_diff), col = "lightblue", 
#'      main = "Difference (x - y)")
#'
#' # Symmetric Difference: parts of both that don't overlap
#' result_symdiff <- ddbs_sym_difference(x, y)
#' plot(st_geometry(result_symdiff), col = "orange", 
#'      main = "Symmetric Difference")
#'
#' # Using with database connection
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ddbs_write_vector(conn, x, "poly_x")
#' ddbs_write_vector(conn, y, "poly_y")
#'
#' # Perform operations with connection
#' ddbs_intersection("poly_x", "poly_y", conn)
#' ddbs_difference("poly_x", "poly_y", conn)
#' ddbs_sym_difference("poly_x", "poly_y", conn)
#'
#' # Save results to database table
#' ddbs_difference("poly_x", "poly_y", conn, name = "diff_result")
#' }
#'
#' @name ddbs_binary_funs
#' @rdname ddbs_binary_funs
#' @aliases ddbs_intersection ddbs_difference ddbs_sym_difference ddbs_crop ddbs_shortest_line
NULL





#' @rdname ddbs_binary_funs
#' @export
ddbs_intersection <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_xy(y, "y")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")


    # 1. Prepare inputs

    ## 1.1. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## 1.3. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn_x)
    crs_y    <- ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    # NOTE: Inline connection resolution logic was replaced by resolve_spatial_connections()
    # helper (defined in db_utils_not_exported.R) to maintain consistency with ddbs_join
    # and other two-input spatial functions. See tests/testthat/test-resolve_connections.R
    # for regression tests covering cross-connection scenarios.
    target_conn <- resolve_res$conn
    x           <- resolve_res$x
    y           <- resolve_res$y
    
    ## 1.4. register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_res$cleanup(), add = TRUE)   
    }
    
    ## 1.5. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)
    
    ## 1.6. Validate the CRS of x and y
    validate_xy_crs(
        crs_x = crs_x,
        crs_y = crs_y,
        conn = target_conn,
        x_list = x_list,
        y_list = y_list
    )


    # 2. Prepare the query

    ## 2.1. Get names of geometry columns (use saved sf_col_x/y from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## 2.2. Build the base query
    st_function <- glue::glue("ST_Intersection(v1.{x_geom}, v2.{y_geom})")
    base.query <- glue::glue("
        SELECT 
            v1.* REPLACE({build_geom_query(st_function, name, crs_x, mode)} AS {x_geom})
        FROM 
            {x_list$query_name} v1,
            {y_list$query_name} v2
        WHERE 
            ST_Intersects(v2.{y_geom}, v1.{x_geom});
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs_x,
            x_geom = x_geom
        )
    }
}






#' @rdname ddbs_binary_funs
#' @export
ddbs_difference <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_xy(y, "y")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")


    # 1. Prepare inputs

    ## 1.1. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## 1.3. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn_x)
    crs_y    <- ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    # NOTE: Inline connection resolution logic was replaced by resolve_spatial_connections()
    # helper (defined in db_utils_not_exported.R) to maintain consistency with ddbs_join
    # and other two-input spatial functions. See tests/testthat/test-resolve_connections.R
    # for regression tests covering cross-connection scenarios.
    target_conn <- resolve_res$conn
    x           <- resolve_res$x
    y           <- resolve_res$y
    
    ## 1.4. register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_res$cleanup(), add = TRUE)   
    }
    
    ## 1.5. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)
    
    ## 1.6. Validate the CRS of x and y
    validate_xy_crs(
        crs_x = crs_x,
        crs_y = crs_y,
        conn = target_conn,
        x_list = x_list,
        y_list = y_list
    )


    # 2. Prepare the query

    ## 2.1. Get names of geometry columns (use saved sf_col_x/y from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## 2.2. Build base query
    st_function <- glue::glue("{x_geom}")
    base.query <- glue::glue("
        WITH diff_geom AS (
            SELECT 
                v1.* REPLACE (
                    ST_Difference(
                        ST_MakeValid(v1.{x_geom}),
                        ST_MakeValid(v2.{y_geom})
                    ) AS {x_geom}
                )
            FROM 
                {x_list$query_name} v1, 
                {y_list$query_name} v2
            WHERE NOT ST_IsEmpty(
                ST_Difference(
                    ST_MakeValid(v1.{x_geom}),
                    ST_MakeValid(v2.{y_geom})
                )
            )
        )
        SELECT 
            * REPLACE ({build_geom_query(st_function, name, crs_x, mode)} AS {x_geom})
        FROM diff_geom;
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs_x,
            x_geom = x_geom
        )
    }

}





#' @rdname ddbs_binary_funs
#' @export
ddbs_sym_difference <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {
    

    # 0. Handle errors
    assert_xy(x, "x")
    assert_xy(y, "y")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

    # 1. Prepare inputs

    ## 1.1. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.2. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## 1.3. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn_x)
    crs_y    <- ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    # NOTE: Inline connection resolution logic was replaced by resolve_spatial_connections()
    # helper (defined in db_utils_not_exported.R) to maintain consistency with ddbs_join
    # and other two-input spatial functions. See tests/testthat/test-resolve_connections.R
    # for regression tests covering cross-connection scenarios.
    target_conn <- resolve_res$conn
    x           <- resolve_res$x
    y           <- resolve_res$y
    
    ## 1.4. register cleanup of the connection
    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_res$cleanup(), add = TRUE)   
    }
    
    ## 1.5. Get query list of table names
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)
    
    ## 1.6. Validate the CRS of x and y
    validate_xy_crs(
        crs_x = crs_x,
        crs_y = crs_y,
        conn = target_conn,
        x_list = x_list,
        y_list = y_list
    )


    # 2. Prepare the query

    ## 2.1. Get names of geometry columns (use saved sf_col_x/y from before transformation)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## 2.2. Build the base query
    st_function <- glue::glue("{x_geom}")
    base.query <- glue::glue("
        WITH symdiff_geom AS (
            SELECT 
                v1.* REPLACE (
                    ST_Union(
                        ST_Difference(
                            ST_MakeValid(v1.{x_geom}),
                            ST_MakeValid(v2.{y_geom})
                        ),
                        ST_Difference(
                            ST_MakeValid(v2.{y_geom}),
                            ST_MakeValid(v1.{x_geom})
                        )
                    ) AS {x_geom}
                ),
                v2.* EXCLUDE ({y_geom})
            FROM 
                {x_list$query_name} v1, 
                {y_list$query_name} v2
            WHERE NOT ST_IsEmpty(
                ST_Union(
                    ST_Difference(
                        ST_MakeValid(v1.{x_geom}),
                        ST_MakeValid(v2.{y_geom})
                    ),
                    ST_Difference(
                        ST_MakeValid(v2.{y_geom}),
                        ST_MakeValid(v1.{x_geom})
                    )
                )
            )
        )
        SELECT 
            * REPLACE ({build_geom_query(st_function, name, crs_x, mode)} AS {x_geom})
        FROM symdiff_geom;
    ")


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs_x,
            x_geom = x_geom
        )
    }
}



#' @rdname ddbs_binary_funs
#' @export
ddbs_crop <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    ## When we crop we use the intersection of the envelope of y
    y <- ddbs_envelope(y)


    ddbs_intersection(
        x = x,
        y = y,
        conn = conn,
        conn_x = conn_x,
        conn_y = conn_y,
        name = name,
        mode = mode,
        overwrite = overwrite,
        quiet = quiet
    )

}




#' @rdname ddbs_binary_funs
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#' library(sf)
#'
#' # Two separate point clouds in a projected CRS
#' origins <- st_as_sf(
#'   data.frame(id = 1:3, x = c(0, 5, 10), y = c(0, 0, 0)),
#'   coords = c("x", "y"), crs = "EPSG:3857"
#' )
#' targets <- st_as_sf(
#'   data.frame(id = 1:3, x = c(0, 5, 10), y = c(3, 4, 5)),
#'   coords = c("x", "y"), crs = "EPSG:3857"
#' )
#'
#' # Returns LINESTRING geometries connecting the closest points
#' ddbs_shortest_line(origins, targets)
#' ddbs_shortest_line(origins, targets, mode = "sf")
#'
#' # Works with any geometry type
#' ddbs_shortest_line(origins, argentina_ddbs)
#' }
ddbs_shortest_line <- function(
    x,
    y,
    conn = NULL,
    conn_x = NULL,
    conn_y = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {


    # 0. Validate inputs
    assert_xy(x, "x")
    assert_xy(y, "y")
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")


    # 1. Prepare inputs

    ## 1.1. Resolve conn_x/conn_y defaults from 'conn' for character inputs
    if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
    if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

    ## 1.2. Normalize inputs
    x <- normalize_spatial_input(x, conn_x)
    y <- normalize_spatial_input(y, conn_y)

    ## 1.3. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn_x)
    crs_y    <- ddbs_crs(y, conn_y)
    sf_col_x <- attr(x, "sf_column")
    sf_col_y <- attr(y, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.4. Resolve spatial connections and handle imports
    resolve_res <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
    target_conn <- resolve_res$conn
    x           <- resolve_res$x
    y           <- resolve_res$y

    if (any(is.null(conn_x), is.null(conn_y))) {
        on.exit(resolve_res$cleanup(), add = TRUE)
    }

    ## 1.5. Get query lists
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)
    y_list <- get_query_list(y, target_conn)
    on.exit(y_list$cleanup(), add = TRUE)

    ## 1.6. Validate CRS
    validate_xy_crs(
        crs_x  = crs_x,
        crs_y  = crs_y,
        conn   = target_conn,
        x_list = x_list,
        y_list = y_list
    )


    # 2. Prepare the query

    ## 2.1. Get geometry column names
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
    assert_geometry_column(x_geom, x_list)
    assert_geometry_column(y_geom, y_list)

    ## 2.2. Build the base query (cross join, no filter)
    ## y's non-geometry columns are included so the user knows which y each row was paired with.
    st_function <- glue::glue("ST_ShortestLine(v1.{x_geom}, v2.{y_geom})")
    base.query <- glue::glue("
        SELECT
            v1.* REPLACE({build_geom_query(st_function, name, crs_x, mode)} AS {x_geom}),
            v2.* EXCLUDE ({y_geom})
        FROM
            {x_list$query_name} v1,
            {y_list$query_name} v2
    ")


    # 3. Return result
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs_x,
            x_geom = x_geom
        )
    }
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create two overlapping polygons for testing
poly1 <- sf::st_polygon(list(matrix(c(
  0, 0,
  4, 0,
  4, 4,
  0, 4,
  0, 0
), ncol = 2, byrow = TRUE)))

poly2 <- sf::st_polygon(list(matrix(c(
  2, 2,
  6, 2,
  6, 6,
  2, 6,
  2, 2
), ncol = 2, byrow = TRUE)))

poly1_sf <- sf::st_sf(id = 1, geometry = sf::st_sfc(poly1), crs = 4326)
poly2_sf <- sf::st_sf(id = 2, geometry = sf::st_sfc(poly2), crs = 4326)

poly1_ddbs <- as_duckspatial_df(poly1_sf)
poly2_ddbs <- as_duckspatial_df(poly2_sf)

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()
conn_test_2 <- duckspatial::ddbs_create_conn()

## write data in the database
ddbs_write_table(conn_test, poly1_sf, "poly1")
ddbs_write_table(conn_test, poly2_sf, "poly2")
ddbs_write_table(conn_test_2, poly2_sf, "poly2")


# 1. ddbs_intersection() -------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting a table works
## - CHECK 1.5: conn_x and conn_y work
## - CHECK 1.6: compare to sf
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
# ddbs_intersection() -----------------------------------------------------

describe("ddbs_intersection()", {

  describe("expected behavior", {

    it("works on all formats and matches results", {
      output_1 <- ddbs_intersection(poly1_sf, poly2_sf)
      output_2 <- ddbs_intersection(poly1_ddbs, poly2_sf)
      output_3 <- ddbs_intersection(poly1_sf, poly2_ddbs)
      output_4 <- ddbs_intersection(poly1_ddbs, poly2_ddbs)

      expect_warning(ddbs_intersection("poly1", poly2_ddbs, conn = conn_test))
      output_6 <- ddbs_intersection("poly1", poly2_sf, conn = conn_test)
      output_7 <- ddbs_intersection(poly1_sf, "poly2", conn = conn_test)
      expect_warning(ddbs_intersection(poly1_ddbs, "poly2", conn = conn_test))
      output_9 <- ddbs_intersection("poly1", "poly2", conn = conn_test)

      expect_s3_class(output_1, "duckspatial_df")
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_2))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_3))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_4))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_6))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_7))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_9))
    })

    it("returns different outputs depending on 'mode' argument", {
      output_sf_fmt <- ddbs_intersection(poly1_sf, poly2_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("handles messages correctly", {
      expect_no_message(ddbs_intersection(poly1_sf, poly2_sf))
      expect_message(ddbs_intersection(poly1_sf, poly2_sf, conn = conn_test, name = "intersection"))
      expect_message(ddbs_intersection(poly1_sf, poly2_sf, conn = conn_test, name = "intersection", overwrite = TRUE))
      expect_true(ddbs_intersection(poly1_sf, poly2_sf, conn = conn_test, name = "intersection2"))

      expect_no_message(ddbs_intersection(poly1_sf, poly2_sf, quiet = TRUE))
      expect_no_message(ddbs_intersection("poly1", "poly2", conn = conn_test, name = "intersection", overwrite = TRUE, quiet = TRUE))
    })

    it("writes a table correctly", {
      output_1 <- ddbs_intersection(poly1_sf, poly2_sf)
      output_tbl <- ddbs_read_table(conn_test, "intersection")
      expect_equal(ddbs_collect(output_1)$geometry, output_tbl$geometry)
    })

    it("works with separate connections for x and y", {
      output_1 <- ddbs_intersection(poly1_sf, poly2_sf)
      output_10 <- ddbs_intersection("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2)
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_10))

      expect_message(ddbs_intersection("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2, name = "test"))
      
      expect_in("test", ddbs_list_tables(conn_test)$table_name)
      expect_disjoint("test", ddbs_list_tables(conn_test_2)$table_name)
    })

    it("matches sf::st_intersection results", {
      sf_output   <- sf::st_intersection(poly1_sf, poly2_sf)
      ddbs_output <- ddbs_intersection(poly1_sf, poly2_sf) |> sf::st_as_sf()
      
      ## Note that ddbs_intersection produces more accurate results,
      ## therefore, just check the bounding box
      bbox_ddbs <- round(ddbs_bbox(ddbs_output, mode = "sf"))
      bbox_sf   <- round(ddbs_bbox(sf_output, mode = "sf"))

      expect_equal(bbox_ddbs, bbox_sf)
    })

  })

  describe("errors", {

    it("errors on missing or invalid arguments", {
      expect_error(ddbs_intersection(poly2_ddbs))
      expect_error(ddbs_intersection(y = poly2_ddbs))
      expect_error(ddbs_intersection("poly2", conn = NULL))
      expect_error(ddbs_intersection("poly1", "poly2", conn_x = conn_test))
      expect_error(ddbs_intersection("poly1", "poly2", conn_y = conn_test))
    })

    it("errors on other invalid inputs", {
      expect_error(ddbs_intersection(x = 999))
      expect_error(ddbs_intersection(poly2_ddbs, poly1_sf, conn = 999))
      expect_error(ddbs_intersection(poly2_ddbs, poly1_sf, overwrite = 999))
      expect_error(ddbs_intersection(poly2_ddbs, poly1_sf, quiet = 999))
      expect_error(ddbs_intersection(x = "999", poly1_sf, conn = conn_test))
      expect_error(ddbs_intersection(poly2_ddbs, poly1_sf, conn = conn_test, name = c('banana', 'banana')))
    })

  })

})



# 2. ddbs_difference() -------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting a table works
## - CHECK 1.5: conn_x and conn_y work
## - CHECK 1.6: compare to sf
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_difference()", {

  describe("expected behavior", {

    it("works on all formats and matches results", {
      output_1 <- ddbs_difference(poly1_sf, poly2_sf)
      output_2 <- ddbs_difference(poly1_ddbs, poly2_sf)
      output_3 <- ddbs_difference(poly1_sf, poly2_ddbs)
      output_4 <- ddbs_difference(poly1_ddbs, poly2_ddbs)

      expect_warning(ddbs_difference("poly1", poly2_ddbs, conn = conn_test))
      output_6 <- ddbs_difference("poly1", poly2_sf, conn = conn_test)
      output_7 <- ddbs_difference(poly1_sf, "poly2", conn = conn_test)
      expect_warning(ddbs_difference(poly1_ddbs, "poly2", conn = conn_test))
      output_9 <- ddbs_difference("poly1", "poly2", conn = conn_test)

      expect_s3_class(output_1, "duckspatial_df")
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_2))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_3))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_4))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_6))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_7))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_9))
    })

    it("returns different outputs depending on 'mode' argument", {
      output_sf_fmt <- ddbs_difference(poly1_sf, poly2_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("handles messages correctly", {
      expect_no_message(ddbs_difference(poly1_sf, poly2_sf))
      expect_message(ddbs_difference(poly1_sf, poly2_sf, conn = conn_test, name = "difference"))
      expect_message(ddbs_difference(poly1_sf, poly2_sf, conn = conn_test, name = "difference", overwrite = TRUE))
      expect_true(ddbs_difference(poly1_sf, poly2_sf, conn = conn_test, name = "difference2"))

      expect_no_message(ddbs_difference(poly1_sf, poly2_sf, quiet = TRUE))
      expect_no_message(ddbs_difference("poly1", "poly2", conn = conn_test, name = "difference", overwrite = TRUE, quiet = TRUE))
    })

    it("writes a table correctly", {
      output_1 <- ddbs_difference(poly1_sf, poly2_sf)
      output_tbl <- ddbs_read_table(conn_test, "difference")
      expect_equal(ddbs_collect(output_1)$geometry, output_tbl$geometry)
    })

    it("works with separate connections for x and y", {
      output_1 <- ddbs_difference(poly1_sf, poly2_sf)
      output_10 <- ddbs_difference("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2)
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_10))

      expect_message(ddbs_difference("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2, name = "diff3"))

      expect_in("diff3", ddbs_list_tables(conn_test)$table_name)
      expect_disjoint("diff3", ddbs_list_tables(conn_test_2)$table_name)
      # expect_true(DBI::dbExistsTable(conn_test, "diff3"))
      # expect_false(DBI::dbExistsTable(conn_test_2, "diff3"))
    })

    it("matches sf::st_difference results", {
      sf_output   <- sf::st_difference(poly1_sf, poly2_sf)
      ddbs_output <- ddbs_difference(poly1_sf, poly2_sf) |> sf::st_as_sf()
      
      ## Note that ddbs_difference produces more accurate results,
      ## therefore, just check the bounding box
      bbox_ddbs <- round(ddbs_bbox(ddbs_output, mode = "sf"))
      bbox_sf   <- round(ddbs_bbox(sf_output, mode = "sf"))

      expect_equal(bbox_ddbs, bbox_sf)
    })

  })

  describe("errors", {

    it("errors on missing or invalid arguments", {
      expect_error(ddbs_difference(poly2_ddbs))
      expect_error(ddbs_difference(y = poly2_ddbs))
      expect_error(ddbs_difference("poly2", conn = NULL))
      expect_error(ddbs_difference("poly1", "poly2", conn_x = conn_test))
      expect_error(ddbs_difference("poly1", "poly2", conn_y = conn_test))
    })

    it("errors on other invalid inputs", {
      expect_error(ddbs_difference(x = 999))
      expect_error(ddbs_difference(poly2_ddbs, poly1_sf, conn = 999))
      expect_error(ddbs_difference(poly2_ddbs, poly1_sf, overwrite = 999))
      expect_error(ddbs_difference(poly2_ddbs, poly1_sf, quiet = 999))
      expect_error(ddbs_difference(x = "999", poly1_sf, conn = conn_test))
      expect_error(ddbs_difference(poly2_ddbs, poly1_sf, conn = conn_test, name = c('banana', 'banana')))
    })

  })

})


# 3. ddbs_sym_difference() -----------------------------------------------

describe("ddbs_sym_difference()", {

  describe("expected behavior", {

    it("works on all formats and matches results", {
      output_1 <- ddbs_sym_difference(poly1_sf, poly2_sf)
      output_2 <- ddbs_sym_difference(poly1_ddbs, poly2_sf)
      output_3 <- ddbs_sym_difference(poly1_sf, poly2_ddbs)
      output_4 <- ddbs_sym_difference(poly1_ddbs, poly2_ddbs)

      expect_warning(ddbs_sym_difference("poly1", poly2_ddbs, conn = conn_test))
      output_6 <- ddbs_sym_difference("poly1", poly2_sf, conn = conn_test)
      output_7 <- ddbs_sym_difference(poly1_sf, "poly2", conn = conn_test)
      expect_warning(ddbs_sym_difference(poly1_ddbs, "poly2", conn = conn_test))
      output_9 <- ddbs_sym_difference("poly1", "poly2", conn = conn_test)

      expect_s3_class(output_1, "duckspatial_df")
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_2))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_3))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_4))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_6))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_7))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_9))
    })

    it("returns different outputs depending on 'mode' argument", {
      output_sf_fmt <- ddbs_sym_difference(poly1_sf, poly2_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("handles messages correctly", {
      expect_no_message(ddbs_sym_difference(poly1_sf, poly2_sf))
      expect_message(ddbs_sym_difference(poly1_sf, poly2_sf, conn = conn_test, name = "symdifference"))
      expect_message(ddbs_sym_difference(poly1_sf, poly2_sf, conn = conn_test, name = "symdifference", overwrite = TRUE))
      expect_true(ddbs_sym_difference(poly1_sf, poly2_sf, conn = conn_test, name = "symdifference2"))

      expect_no_message(ddbs_sym_difference(poly1_sf, poly2_sf, quiet = TRUE))
      expect_no_message(ddbs_sym_difference("poly1", "poly2", conn = conn_test, name = "symdifference", overwrite = TRUE, quiet = TRUE))
    })

    it("writes a table correctly", {
      output_1 <- ddbs_sym_difference(poly1_sf, poly2_sf)
      output_tbl <- ddbs_read_table(conn_test, "symdifference")
      expect_equal(ddbs_collect(output_1)$geometry, output_tbl$geometry)
    })

    it("works with separate connections for x and y", {
      output_1 <- ddbs_sym_difference(poly1_sf, poly2_sf)
      output_10 <- ddbs_sym_difference("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2)
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_10))

      expect_message(ddbs_sym_difference("poly1", "poly2", conn_x = conn_test, conn_y = conn_test_2, name = "symdiff3"))

      expect_in("symdiff3", ddbs_list_tables(conn_test)$table_name)
      expect_disjoint("symdiff3", ddbs_list_tables(conn_test_2)$table_name)
      # expect_true(DBI::dbExistsTable(conn_test, "symdiff3"))
      # expect_false(DBI::dbExistsTable(conn_test_2, "symdiff3"))
    })

    it("matches sf::st_sym_difference results", {
      sf_output   <- sf::st_sym_difference(poly1_sf, poly2_sf)
      ddbs_output <- ddbs_sym_difference(poly1_sf, poly2_sf) |> sf::st_as_sf()
      
      ## Note that ddbs_sym_difference produces more accurate results,
      ## therefore, just check the bounding box
      bbox_ddbs <- round(ddbs_bbox(ddbs_output, mode = "sf"))
      bbox_sf   <- round(ddbs_bbox(sf_output, mode = "sf"))

      expect_equal(bbox_ddbs, bbox_sf)
    })

    it("produces symmetric results", {
      # Symmetric difference should be commutative: symdiff(A,B) == symdiff(B,A)
      output_xy <- ddbs_sym_difference(poly1_sf, poly2_sf)
      output_yx <- ddbs_sym_difference(poly2_sf, poly1_sf)
      
      expect_equal(
        ddbs_bbox(output_xy, mode = "sf"),
        ddbs_bbox(output_yx, mode = "sf")
      )
    })

  })

  describe("errors", {

    it("errors on missing or invalid arguments", {
      expect_error(ddbs_sym_difference(poly2_ddbs))
      expect_error(ddbs_sym_difference(y = poly2_ddbs))
      expect_error(ddbs_sym_difference("poly2", conn = NULL))
      expect_error(ddbs_sym_difference("poly1", "poly2", conn_x = conn_test))
      expect_error(ddbs_sym_difference("poly1", "poly2", conn_y = conn_test))
    })

    it("errors on other invalid inputs", {
      expect_error(ddbs_sym_difference(x = 999))
      expect_error(ddbs_sym_difference(poly2_ddbs, poly1_sf, conn = 999))
      expect_error(ddbs_sym_difference(poly2_ddbs, poly1_sf, overwrite = 999))
      expect_error(ddbs_sym_difference(poly2_ddbs, poly1_sf, quiet = 999))
      expect_error(ddbs_sym_difference(x = "999", poly1_sf, conn = conn_test))
      expect_error(ddbs_sym_difference(poly2_ddbs, poly1_sf, conn = conn_test, name = c('banana', 'banana')))
    })

  })

})





# 4. ddbs_crop() ---------------------------------------------------------

## - CHECK 1.1: works on all formats
## - CHECK 1.2: returns different output modes
## - CHECK 1.3: messages work
## - CHECK 1.4: writing a table works
## - CHECK 1.5: warns when mixing connections
## - CHECK 1.6: compare to sf
## - CHECK 2.1: missing / invalid arguments
## - CHECK 2.2: other errors

describe("ddbs_crop()", {

  describe("expected behavior", {

    it("works on all formats and matches results", {
      output_1 <- ddbs_crop(poly1_sf, poly2_sf)
      output_2 <- ddbs_crop(poly1_ddbs, poly2_sf)
      output_3 <- ddbs_crop(poly1_sf, poly2_ddbs)
      output_4 <- ddbs_crop(poly1_ddbs, poly2_ddbs)

      expect_s3_class(output_1, "duckspatial_df")
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_2))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_3))
      expect_equal(ddbs_collect(output_1), ddbs_collect(output_4))
    })

    it("returns different outputs depending on 'mode' argument", {
      output_sf_fmt <- ddbs_crop(poly1_sf, poly2_ddbs, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("handles messages correctly", {
      expect_no_message(ddbs_crop(poly1_sf, poly2_sf))
      expect_message(ddbs_crop(poly1_sf, poly2_sf, conn = conn_test, name = "crop"))
      expect_message(ddbs_crop(poly1_sf, poly2_sf, conn = conn_test, name = "crop", overwrite = TRUE))
      expect_true(ddbs_crop(poly1_sf, poly2_sf, conn = conn_test, name = "crop2"))

      expect_no_message(ddbs_crop(poly1_sf, poly2_sf, quiet = TRUE))
    })

    it("writes a table correctly", {
      output_1   <- ddbs_crop(poly1_sf, poly2_sf)
      output_tbl <- ddbs_read_table(conn_test, "crop")
      expect_equal(ddbs_collect(output_1)$geometry, output_tbl$geometry)
    })

    it("warns when mixing a table name with a duckspatial_df from another connection", {
      expect_warning(ddbs_crop("poly1", poly2_ddbs, conn = conn_test))
    })

    it("matches sf::st_crop results", {
      sf_output   <- sf::st_crop(poly1_sf, poly2_sf)
      ddbs_output <- ddbs_crop(poly1_sf, poly2_sf) |> sf::st_as_sf()

      bbox_ddbs <- round(ddbs_bbox(ddbs_output, mode = "sf"))
      bbox_sf   <- round(ddbs_bbox(sf_output, mode = "sf"))

      expect_equal(bbox_ddbs, bbox_sf)
    })

  })

  describe("errors", {

    it("errors on missing or invalid arguments", {
      expect_error(ddbs_crop(poly2_ddbs))
      expect_error(ddbs_crop(y = poly2_ddbs))
      expect_error(ddbs_crop("poly2", poly1_sf, conn = NULL))
    })

    it("errors on other invalid inputs", {
      expect_error(ddbs_crop(999, poly1_sf))
      expect_error(ddbs_crop(poly2_ddbs, poly1_sf, conn = 999))
      expect_error(ddbs_crop(poly2_ddbs, poly1_sf, overwrite = 999))
      expect_error(ddbs_crop(poly2_ddbs, poly1_sf, quiet = 999))
    })

  })

})



# ddbs_shortest_line() ---------------------------------------------------

## Fixtures: two separate POINT datasets in a projected CRS.
## (0,0) → (3,4): known Euclidean distance = 5 m in EPSG:3857.
pt_a_sf   <- sf::st_as_sf(data.frame(id = 1L, x = 0, y = 0), coords = c("x", "y"), crs = "EPSG:3857")
pt_b_sf   <- sf::st_as_sf(data.frame(id = 2L, x = 3, y = 4), coords = c("x", "y"), crs = "EPSG:3857")
pt_a_ddbs <- as_duckspatial_df(pt_a_sf)
pt_b_ddbs <- as_duckspatial_df(pt_b_sf)
ddbs_write_table(conn_test, pt_a_sf, "pt_a")
ddbs_write_table(conn_test, pt_b_sf, "pt_b")

describe("ddbs_shortest_line()", {

  describe("expected behavior on sf input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_shortest_line(pt_a_sf, pt_b_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns an sf object with mode sf", {
      output <- ddbs_shortest_line(pt_a_sf, pt_b_sf, mode = "sf")
      expect_s3_class(output, "sf")
    })

    it("output geometry is LINESTRING", {
      output <- ddbs_shortest_line(pt_a_sf, pt_b_sf, mode = "sf")
      expect_equal(as.character(sf::st_geometry_type(output)), "LINESTRING")
    })

    it("shortest line length equals the planar distance between the points", {
      line   <- ddbs_shortest_line(pt_a_sf, pt_b_sf, mode = "sf")
      expect_equal(as.numeric(sf::st_length(line)), 5, tolerance = 1e-6)
    })

    it("writes to a table when name is provided", {
      output <- ddbs_shortest_line(pt_a_sf, pt_b_sf, conn = conn_test, name = "shortest_tbl")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_message(
        ddbs_shortest_line(pt_a_sf, pt_b_sf, conn = conn_test, name = "shortest_tbl2")
      )
      expect_no_message(
        ddbs_shortest_line(pt_a_sf, pt_b_sf, conn = conn_test, name = "shortest_tbl3", quiet = TRUE)
      )
    })
  })

  describe("expected behavior on duckspatial_df input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_shortest_line(pt_a_ddbs, pt_b_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns an sf object with mode sf", {
      output <- ddbs_shortest_line(pt_a_ddbs, pt_b_ddbs, mode = "sf")
      expect_s3_class(output, "sf")
    })

    it("warns when mixing connections", {
      expect_warning(ddbs_shortest_line(pt_a_ddbs, pt_b_sf, conn = conn_test))
    })

    it("matches sf input result", {
      out_sf   <- ddbs_shortest_line(pt_a_sf,   pt_b_sf,   mode = "sf")
      out_ddbs <- ddbs_shortest_line(pt_a_ddbs, pt_b_ddbs, mode = "sf")
      expect_equal(sf::st_geometry(out_sf), sf::st_geometry(out_ddbs))
    })
  })

  describe("expected behavior on DuckDB table input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_shortest_line("pt_a", "pt_b", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns an sf object with mode sf", {
      output <- ddbs_shortest_line("pt_a", "pt_b", conn = conn_test, mode = "sf")
      expect_s3_class(output, "sf")
    })

    it("matches sf input result", {
      out_sf   <- ddbs_shortest_line(pt_a_sf, pt_b_sf, mode = "sf")
      out_conn <- ddbs_shortest_line("pt_a",  "pt_b",  conn = conn_test, mode = "sf")
      expect_equal(sf::st_geometry(out_sf), sf::st_geometry(out_conn))
    })
  })

  describe("errors", {

    it("requires both x and y", {
      expect_error(ddbs_shortest_line(pt_a_sf))
      expect_error(ddbs_shortest_line(y = pt_b_sf))
    })

    it("requires a connection for character table names", {
      expect_error(ddbs_shortest_line("pt_a", "pt_b", conn = NULL))
    })

    it("rejects mismatched CRS", {
      pt_b_4326_sf <- sf::st_transform(pt_b_sf, "EPSG:4326")
      expect_error(ddbs_shortest_line(pt_a_sf, pt_b_4326_sf))
    })

    it("rejects invalid x and y types", {
      expect_error(ddbs_shortest_line(999, pt_b_sf))
      expect_error(ddbs_shortest_line(pt_a_sf, 999))
    })

    it("validates conn, overwrite, quiet argument types", {
      expect_error(ddbs_shortest_line(pt_a_sf, pt_b_sf, conn = 999))
      expect_error(ddbs_shortest_line(pt_a_sf, pt_b_sf, overwrite = 999))
      expect_error(ddbs_shortest_line(pt_a_sf, pt_b_sf, quiet = 999))
    })
  })
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' @param by_feature Logical. If \code{TRUE}, the geometric operation is applied
#'        separately to each geometry. If \code{FALSE}, the geometric
#'        operation is applied to the data as a whole.

#' Convert a duckspatial_df to a nanoarrow_array_stream
#'
#' @param x A \code{duckspatial_df} object
#' @param ... Additional arguments passed to downstream methods
#' @param schema Optional target schema for the entire stream.
#' @param native If TRUE, transforms WKB to a "Native" GeoArrow layout (e.g., 
#'   Point, Polygon) using optimized Arrow-to-Arrow kernels. This layout is 
#'   optimized for high-performance rendering in tools like Deck.GL.
#'
#' @return A \code{nanoarrow_array_stream}
#' @exportS3Method nanoarrow::as_nanoarrow_array_stream duckspatial_df
as_nanoarrow_array_stream.duckspatial_df <- function(x, ..., 
                                                     schema = NULL, 
                                                     native = FALSE) {
  
  geom_col <- attr(x, "sf_column") %||% "geom"
  conn <- dbplyr::remote_con(x)
  
  if (is.null(conn)) {
    cli::cli_abort("Cannot stream {.cls duckspatial_df}: connection is missing.")
  }

  # Strip class to treat as standard lazy table for further manipulation.
  x_lazy <- x
  class(x_lazy) <- setdiff(class(x_lazy), "duckspatial_df")

  # 1. Force DuckDB to output standard WKB binary
  x_wkb <- dplyr::mutate(
    x_lazy,
    !!rlang::sym(geom_col) := dbplyr::sql(glue::glue("ST_AsWKB({DBI::dbQuoteIdentifier(conn, geom_col)})"))
  )
  query_sql <- dbplyr::sql_render(x_wkb)

  # 2. Execute and get Arrow data
  res <- DBI::dbSendQuery(conn, query_sql, arrow = TRUE)
  arrow_obj <- duckdb::duckdb_fetch_arrow(res)
  
  # 3. Native Path: Transform WKB to Native GeoArrow entirely in Arrow memory
  if (isTRUE(native)) {
    tab <- arrow::as_arrow_table(arrow_obj)
    
    # Infer native schema from WKB column
    wkb_col <- nanoarrow::as_nanoarrow_array(arrow::as_arrow_array(tab[[geom_col]]))
    
    # Try to infer native schema
    target_geom_schema <- tryCatch({
      geoarrow::infer_geoarrow_schema(wkb_col, coord_type = "SEPARATE")
    }, error = function(e) {
      geoarrow::geoarrow_wkb(crs = ddbs_crs(x))
    })
    
    # Cast WKB to Native layout using geoarrow kernels
    tab_list <- as.list(tab)
    # geoarrow::as_geoarrow_array needs a nanoarrow_array or wk object
    tab_list[[geom_col]] <- geoarrow::as_geoarrow_array(wkb_col, schema = target_geom_schema)
    
    new_tab <- arrow::as_arrow_table(arrow::record_batch(!!!tab_list))
    return(nanoarrow::as_nanoarrow_array_stream(new_tab, schema = schema))
  }

  # 4. Lazy Path: Zero-copy metadata injection into the stream
  old_schema <- nanoarrow::as_nanoarrow_schema(arrow_obj$schema)
  
  if (!is.null(schema)) {
    new_schema <- nanoarrow::as_nanoarrow_schema(schema)
  } else {
    dots <- list(...)
    wkb_args <- dots[names(dots) %in% c("edges")]
    
    ga_type <- do.call(geoarrow::geoarrow_wkb, c(list(crs = ddbs_crs(x)), wkb_args))
    ga_schema <- nanoarrow::as_nanoarrow_schema(ga_type)
    
    geom_found <- FALSE
    new_children <- lapply(old_schema$children, function(child) {
      if (child$name == geom_col) {
        geom_found <<- TRUE
        child_new <- ga_schema
        child_new$name <- child$name
        child_new$flags <- child$flags
        child_new
      } else {
        child
      }
    })
    
    if (!geom_found) {
      cli::cli_abort("Geometry column {.field {geom_col}} not found in query results.")
    }
    
    new_schema <- nanoarrow::nanoarrow_schema_modify(old_schema, list(children = new_children))
  }
  
  nanoarrow::as_nanoarrow_array_stream(arrow_obj, schema = new_schema)
}

#' @exportS3Method arrow::as_arrow_table duckspatial_df
as_arrow_table.duckspatial_df <- function(x, ...) {
  # Delegate to our optimized stream, then collect to a table
  stream <- nanoarrow::as_nanoarrow_array_stream(x, ...)
  arrow::as_arrow_table(stream)
}

#' @exportS3Method geoarrow::infer_geoarrow_schema duckspatial_df
infer_geoarrow_schema.duckspatial_df <- function(x, ...) {
  dots <- list(...)
  wkb_args <- dots[names(dots) %in% c("edges")]
  nanoarrow::as_nanoarrow_schema(do.call(geoarrow::geoarrow_wkb, c(list(crs = ddbs_crs(x)), wkb_args)))
}

#' @exportS3Method arrow::as_record_batch_reader duckspatial_df
as_record_batch_reader.duckspatial_df <- function(x, ...) {
  # Delegate to nanoarrow which handles the DuckDB/Arrow interface and metadata injection
  stream <- nanoarrow::as_nanoarrow_array_stream(x, ...)
  # Use the arrow package's importer to ensure C-stream metadata is respected
  arrow::RecordBatchReader$import_from_c(stream)
}

# Set up --------------------------------------------------------------

testthat::skip_if_not_installed("nanoarrow")
testthat::skip_if_not_installed("geoarrow")
testthat::skip_if_not_installed("arrow")

# Use nc_ddbs from setup.R
# North Carolina data has 100 features and 14 columns (geom is #14)

# 1. as_nanoarrow_array_stream.duckspatial_df ---------------------------

describe("as_nanoarrow_array_stream.duckspatial_df()", {

  it("produces a valid nanoarrow_array_stream with geoarrow.wkb extension", {
    # Default path (native = FALSE)
    stream <- nanoarrow::as_nanoarrow_array_stream(nc_ddbs)
    expect_s3_class(stream, "nanoarrow_array_stream")
    
    schema <- stream$get_schema()
    geom_col <- attr(nc_ddbs, "sf_column")
    
    # Find geometry child in the schema
    child_names <- vapply(schema$children, function(x) x$name, character(1))
    geom_idx <- which(child_names == geom_col)
    
    expect_length(geom_idx, 1)
    geom_schema <- schema$children[[geom_idx]]
    
    # Check for geoarrow.wkb metadata
    expect_equal(geom_schema$metadata[["ARROW:extension:name"]], "geoarrow.wkb")
    
    # Check CRS in metadata
    ext_meta <- geom_schema$metadata[["ARROW:extension:metadata"]]
    expect_true(nchar(ext_meta) > 0)
    expect_true(grepl("crs", ext_meta))
    
    # Clean up
    stream$release()
  })

  it("works with native = TRUE (materialized path)", {
    # Skip if sf can't be loaded (unlikely here but safe)
    testthat::skip_if_not_installed("sf")
    
    stream <- nanoarrow::as_nanoarrow_array_stream(nc_ddbs, native = TRUE)
    expect_s3_class(stream, "nanoarrow_array_stream")
    
    schema <- stream$get_schema()
    child_names <- vapply(schema$children, function(x) x$name, character(1))
    geom_idx <- which(child_names == attr(nc_ddbs, "sf_column"))
    
    geom_schema <- schema$children[[geom_idx]]
    
    # native = TRUE should produce native geoarrow types (e.g. geoarrow.polygon)
    # depending on the geometry type of the input
    ext_name <- geom_schema$metadata[["ARROW:extension:name"]]
    expect_true(grepl("^geoarrow\\.", ext_name))
    
    stream$release()
  })

  it("works with geometry_schema (Native layout)", {
    # Request a native point schema
    target_schema <- geoarrow::geoarrow_point()
    # Use points_ddbs instead of nc_ddbs (polygons) to avoid conversion error
    stream <- nanoarrow::as_nanoarrow_array_stream(points_ddbs, geometry_schema = target_schema)
    
    expect_s3_class(stream, "nanoarrow_array_stream")
    schema <- stream$get_schema()
    geom_col <- attr(points_ddbs, "sf_column")
    child_names <- vapply(schema$children, function(x) x$name, character(1))
    geom_idx <- which(child_names == geom_col)
    geom_schema <- schema$children[[geom_idx]]
    
    expect_equal(geom_schema$metadata[["ARROW:extension:name"]], "geoarrow.wkb")
    
    stream$release()
  })

  it("errors if connection is missing", {
    # Create a dummy duckspatial_df without a connection
    x_bad <- structure(
        list(),
        class = c("duckspatial_df", "tbl_duckdb_connection", "tbl_dbi", "tbl_sql", "tbl_lazy", "tbl"),
        sf_column = "geom",
        crs = sf::st_crs(4326)
    )
    # Ensure remote_con(x_bad) is NULL
    
    expect_error(
      nanoarrow::as_nanoarrow_array_stream(x_bad),
      "connection is missing"
    )
  })

  it("errors if geometry column is missing from the query results", {
    # Point the sf_column attribute to something that doesn't exist
    nc_broken <- nc_ddbs
    attr(nc_broken, "sf_column") <- "non_existent_geom"
    
    # This might error in mutate (DuckDB Binder Error) or in our geom_found check
    expect_error(
      nanoarrow::as_nanoarrow_array_stream(nc_broken)
    )
  })

})

# 2. infer_geoarrow_schema.duckspatial_df -------------------------------

describe("infer_geoarrow_schema.duckspatial_df()", {

  it("returns a geoarrow.wkb schema", {
    schema <- geoarrow::infer_geoarrow_schema(nc_ddbs)
    expect_s3_class(schema, "nanoarrow_schema")
    expect_equal(schema$metadata[["ARROW:extension:name"]], "geoarrow.wkb")
  })

})

# 3. as_record_batch_reader.duckspatial_df ------------------------------

describe("as_record_batch_reader.duckspatial_df()", {

  it("returns an arrow RecordBatchReader", {
    reader <- arrow::as_record_batch_reader(nc_ddbs)
    expect_s3_class(reader, "RecordBatchReader")
    
    # Verify the reader's schema has the extension
    geom_col <- attr(nc_ddbs, "sf_column")
    field <- reader$schema$GetFieldByName(geom_col)
    
    # In the arrow R package, extension metadata is mapped to the type's extension_name()
    expect_true(inherits(field$type, "ExtensionType"))
    expect_equal(field$type$extension_name(), "geoarrow.wkb")
  })

})

#' Geometry validation functions
#'
#' Functions to check various geometric properties and validity conditions of spatial
#' geometries using DuckDB's spatial extension.
#' 
#' @template x
#' @template new_column
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' These functions provide different types of geometric validation. Note that by default,
#' the functions add a new column as a logical vector. This behaviour allows to filter the
#' data within DuckDB without the need or materializating a vector in R (see details).
#'
#' - `ddbs_is_valid()` checks if a geometry is valid according to the OGC Simple Features
#'   specification. Invalid geometries may have issues like self-intersections in polygons,
#'   duplicate points, or incorrect ring orientations.
#'
#' - `ddbs_is_simple()` determines whether geometries are simple, meaning they are free of
#'   self-intersections. For example, a linestring that crosses itself is not simple.
#'
#' - `ddbs_is_ring()` checks if a linestring geometry is closed (first and last points are
#'   identical) and simple (no self-intersections), forming a valid ring.
#'
#' - `ddbs_is_empty()` tests whether a geometry is empty, containing no points. Empty
#'   geometries are valid but represent the absence of spatial information.
#'
#' - `ddbs_is_closed()` determines if a linestring geometry is closed, meaning the first
#'   and last coordinates are identical. Unlike `ddbs_is_ring()`, this does not check for
#'   simplicity.
#' 
#' @returns
#' \itemize{
#'   \item \code{mode = "duckspatial"} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
#'   \item \code{mode = "sf"}: An eagerly collected vector in R memory.
#'   \item When \code{name} is provided: writes the table in the DuckDB connection and returns \code{TRUE} (invisibly).
#' }
#' 
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#' library(dplyr)
#'
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## geometry validation
#' ddbs_is_valid(countries_ddbs)
#' ddbs_is_simple(countries_ddbs)
#' ddbs_is_ring(rivers_ddbs)
#' ddbs_is_empty(countries_ddbs)
#' ddbs_is_closed(countries_ddbs)
#' 
#' ## filter invalid countries
#' ddbs_is_valid(countries_ddbs) |> filter(!is_valid)
#' }
#'
#' @name ddbs_geom_validation_funs
#' @rdname ddbs_geom_validation_funs
NULL




#' @rdname ddbs_geom_validation_funs
#' @export
ddbs_is_simple <- function(
  x,
  new_column = "is_simple",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_IsSimple"
  )
  
}





#' @rdname ddbs_geom_validation_funs
#' @export
ddbs_is_valid <- function(
  x,
  new_column = "is_valid",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_IsValid"
  )
  
}




#' @rdname ddbs_geom_validation_funs
#' @export
ddbs_is_closed <- function(
  x,
  new_column = "is_closed",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_IsClosed"
  )
  
}




#' @rdname ddbs_geom_validation_funs
#' @export
ddbs_is_empty <- function(
  x,
  new_column = "is_empty",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_IsEmpty"
  )
  
}





#' @rdname ddbs_geom_validation_funs
#' @export
ddbs_is_ring <- function(
  x,
  new_column = "is_ring",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_IsRing"
  )
  
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## read countries
countries_2_sf <- sf::read_sf(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |> 
    sf::st_transform("EPSG:4326")

## add an empty polygon to countries
countries_2_sf[nrow(countries_2_sf) + 1, ] <- countries_2_sf[nrow(countries_2_sf) + 1, ] 
countries_2_ddbs <- countries_2_sf

## write data
duckspatial::ddbs_write_table(conn_test, countries_2_sf, "countries")


# 1. ddbs_is_valid() -----------------------------------------------------

## - CHECK 1.1: vector works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: compare with SF
## - CHECK 2.1: errors
describe("ddbs_is_valid()", {
  
  describe("expected behavior", {
    
    it("returns logical vector on all formats with sf", {
      output_ddbs_vec <- ddbs_is_valid(countries_2_ddbs, mode = "sf")
      output_sf_vec   <- ddbs_is_valid(countries_2_sf, mode = "sf")
      output_conn_vec <- ddbs_is_valid("countries", conn = conn_test, mode = "sf")

      expect_type(output_ddbs_vec, "logical")
      expect_equal(output_ddbs_vec, output_sf_vec)
      expect_equal(output_ddbs_vec, output_conn_vec)
    })
    
    it("returns duckpatial_df on all formats by default", {
      output_ddbs <- ddbs_is_valid(countries_2_ddbs)
      output_sf   <- ddbs_is_valid(countries_2_sf)
      output_conn <- ddbs_is_valid("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats", {
      output_sf_fmt <- ddbs_is_valid(countries_2_ddbs, mode = "sf")
      expect_type(output_sf_fmt, "logical")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_is_valid(countries_2_ddbs))
      expect_message(ddbs_is_valid("countries", conn = conn_test, name = "is_valid_tbl"))
      expect_message(ddbs_is_valid("countries", conn = conn_test, name = "is_valid_tbl", overwrite = TRUE))
      expect_true(ddbs_is_valid("countries", conn = conn_test, name = "is_valid_tbl2"))
      expect_no_message(ddbs_is_valid(countries_2_ddbs, quiet = TRUE))
      expect_no_message(ddbs_is_valid("countries", conn = conn_test, name = "is_valid_tbl", overwrite = TRUE, quiet = TRUE))
    })
    
    it("saves to database table correctly", {
      output_ddbs <- ddbs_is_valid(countries_2_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "is_valid_tbl")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("matches sf::st_is_valid results", {
      ## Review why Antarctica is not valid in Duckdb, but it's valid in sf
      output_ddbs_vec <- ddbs_is_valid(countries_2_ddbs, mode = "sf")
      sf_output <- sf::st_is_valid(countries_2_sf)
      
      expect_equal(output_ddbs_vec[-4], sf_output[-4])
    })
    
  })
  
  describe("errors", {
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_valid(countries_2_sf, new_column = 5))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_is_valid("countries", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_is_valid(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_is_valid(countries_2_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_valid(countries_2_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_is_valid(countries_2_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_is_valid(countries_2_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_is_valid(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_is_valid(countries_2_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 2. ddbs_is_simple() -----------------------------------------------------

## - CHECK 1.1: vector works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: compare with SF
## - CHECK 2.1: errors
describe("ddbs_is_simple()", {
  
  describe("expected behavior", {
    
    it("returns logical vector on all formats with sf", {
      output_ddbs_vec <- ddbs_is_simple(countries_2_ddbs, mode = "sf")
      output_sf_vec   <- ddbs_is_simple(countries_2_sf, mode = "sf")
      output_conn_vec <- ddbs_is_simple("countries", conn = conn_test, mode = "sf")

      expect_type(output_ddbs_vec, "logical")
      expect_equal(output_ddbs_vec, output_sf_vec)
      expect_equal(output_ddbs_vec, output_conn_vec)
    })
    
    it("returns duckpatial_df on all formats by default", {
      output_ddbs <- ddbs_is_simple(countries_2_ddbs)
      output_sf   <- ddbs_is_simple(countries_2_sf)
      output_conn <- ddbs_is_simple("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats", {
      output_sf_fmt <- ddbs_is_simple(countries_2_ddbs, mode = "sf")
      expect_type(output_sf_fmt, "logical")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_is_simple(countries_2_ddbs))
      expect_message(ddbs_is_simple("countries", conn = conn_test, name = "is_simple_tbl"))
      expect_message(ddbs_is_simple("countries", conn = conn_test, name = "is_simple_tbl", overwrite = TRUE))
      expect_true(ddbs_is_simple("countries", conn = conn_test, name = "is_simple_tbl2"))
      expect_no_message(ddbs_is_simple(countries_2_ddbs, quiet = TRUE))
      expect_no_message(ddbs_is_simple("countries", conn = conn_test, name = "is_simple_tbl", overwrite = TRUE, quiet = TRUE))
    })
    
    it("saves to database table correctly", {
      output_ddbs <- ddbs_is_simple(countries_2_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "is_simple_tbl")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("matches sf::st_is_valid results", {
      output_ddbs_vec <- ddbs_is_simple(countries_2_ddbs, mode = "sf")
      sf_output <- sf::st_is_simple(countries_2_sf)
      
      expect_equal(output_ddbs_vec, sf_output)
    })
    
  })
  
  describe("errors", {
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_simple(countries_2_sf, new_column = 5))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_is_simple("countries", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_is_simple(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_is_simple(countries_2_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_simple(countries_2_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_is_simple(countries_2_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_is_simple(countries_2_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_is_simple(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_is_simple(countries_2_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 3. ddbs_is_empty() -----------------------------------------------------

## - CHECK 1.1: vector works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: compare with SF
## - CHECK 2.1: errors
describe("ddbs_is_empty()", {
  
  describe("expected behavior", {
    
    it("returns logical vector on all formats with sf", {
      output_ddbs_vec <- ddbs_is_empty(countries_2_ddbs, mode = "sf")
      output_sf_vec   <- ddbs_is_empty(countries_2_sf, mode = "sf")
      output_conn_vec <- ddbs_is_empty("countries", conn = conn_test, mode = "sf")

      expect_type(output_ddbs_vec, "logical")
      expect_equal(output_ddbs_vec, output_sf_vec)
      expect_equal(output_ddbs_vec, output_conn_vec)
    })
    
    it("returns duckpatial_df on all formats by default", {
      output_ddbs <- ddbs_is_empty(countries_2_ddbs)
      output_sf   <- ddbs_is_empty(countries_2_sf)
      output_conn <- ddbs_is_empty("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats", {
      output_sf_fmt <- ddbs_is_empty(countries_2_ddbs, mode = "sf")
      expect_type(output_sf_fmt, "logical")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_is_empty(countries_2_ddbs))
      expect_message(ddbs_is_empty("countries", conn = conn_test, name = "is_empty_tbl"))
      expect_message(ddbs_is_empty("countries", conn = conn_test, name = "is_empty_tbl", overwrite = TRUE))
      expect_true(ddbs_is_empty("countries", conn = conn_test, name = "is_empty_tbl2"))
      expect_no_message(ddbs_is_empty(countries_2_ddbs, quiet = TRUE))
      expect_no_message(ddbs_is_empty("countries", conn = conn_test, name = "is_empty_tbl", overwrite = TRUE, quiet = TRUE))
    })
    
    it("saves to database table correctly", {
      output_ddbs <- ddbs_is_empty(countries_2_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "is_empty_tbl")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("matches sf::st_is_valid results", {
      output_ddbs_vec <- ddbs_is_empty(countries_2_ddbs, mode = "sf")
      sf_output <- sf::st_is_empty(countries_2_sf)
      
      expect_equal(output_ddbs_vec, sf_output)
    })
    
  })
  
  describe("errors", {
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_empty(countries_2_sf, new_column = 5))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_is_empty("countries", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_is_empty(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_is_empty(countries_2_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_empty(countries_2_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_is_empty(countries_2_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_is_empty(countries_2_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_is_empty(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_is_empty(countries_2_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 4. ddbs_is_ring() -----------------------------------------------------

## - CHECK 1.1: vector works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: compare with SF
## - CHECK 2.1: errors
describe("ddbs_is_ring()", {
  
  describe("expected behavior", {
    
    it("returns logical vector on all formats with sf", {
      output_ddbs_vec <- ddbs_is_ring(countries_2_ddbs, mode = "sf")
      output_sf_vec   <- ddbs_is_ring(countries_2_sf, mode = "sf")
      output_conn_vec <- ddbs_is_ring("countries", conn = conn_test, mode = "sf")

      expect_type(output_ddbs_vec, "logical")
      expect_equal(output_ddbs_vec, output_sf_vec)
      expect_equal(output_ddbs_vec, output_conn_vec)
    })
    
    it("returns duckpatial_df on all formats by default", {
      output_ddbs <- ddbs_is_ring(countries_2_ddbs)
      output_sf   <- ddbs_is_ring(countries_2_sf)
      output_conn <- ddbs_is_ring("countries", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats", {
      output_sf_fmt <- ddbs_is_ring(countries_2_ddbs, mode = "sf")
      expect_type(output_sf_fmt, "logical")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_is_ring(countries_2_ddbs))
      expect_message(ddbs_is_ring("countries", conn = conn_test, name = "is_ring_tbl"))
      expect_message(ddbs_is_ring("countries", conn = conn_test, name = "is_ring_tbl", overwrite = TRUE))
      expect_true(ddbs_is_ring("countries", conn = conn_test, name = "is_ring_tbl2"))
      expect_no_message(ddbs_is_ring(countries_2_ddbs, quiet = TRUE))
      expect_no_message(ddbs_is_ring("countries", conn = conn_test, name = "is_ring_tbl", overwrite = TRUE, quiet = TRUE))
    })
    
    it("saves to database table correctly", {
      output_ddbs <- ddbs_is_ring(countries_2_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "is_ring_tbl")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("returns TRUE in a closed linestring", {
      coords <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))
      closed_ring_sf <- sf::st_sf(
        geometry = sf::st_sfc(sf::st_linestring(coords))
      )

      expect_true(ddbs_is_ring(closed_ring_sf, mode = "sf"))

    })
    
  })
  
  describe("errors", {
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_ring(countries_2_sf, new_column = 5))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_is_ring("countries", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_is_ring(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_is_ring(countries_2_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_ring(countries_2_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_is_ring(countries_2_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_is_ring(countries_2_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_is_ring(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_is_ring(countries_2_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 5. ddbs_is_closed() -----------------------------------------------------

## - CHECK 1.1: vector works on all formats
## - CHECK 1.2: messages work
## - CHECK 1.3: compare with SF
## - CHECK 2.1: errors
describe("ddbs_is_closed()", {

  
  # Closed ring - first and last points are identical
  coords_closed <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))

  # Not closed - first and last points are different
  coords_open <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1))

  # Combine in one SF object
  lines_sf <- sf::st_sf(
    id = c(1, 2),
    type = c("closed", "open"),
    geometry = sf::st_sfc(
      sf::st_linestring(coords_closed),
      sf::st_linestring(coords_open)
    ),
    crs = 4326
  )

  lines_ddbs <- as_duckspatial_df(lines_sf)
  ddbs_write_table(conn_test, lines_ddbs, "lines")

  ## EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("returns logical vector on all formats with sf", {
      output_ddbs_vec <- ddbs_is_closed(lines_ddbs, mode = "sf")
      output_sf_vec   <- ddbs_is_closed(lines_sf, mode = "sf")
      output_conn_vec <- ddbs_is_closed("lines", conn = conn_test, mode = "sf")

      expect_type(output_ddbs_vec, "logical")
      expect_equal(output_ddbs_vec, output_sf_vec)
      expect_equal(output_ddbs_vec, output_conn_vec)
    })
    
    it("returns duckpatial_df on all formats by default", {
      output_ddbs <- ddbs_is_closed(lines_ddbs)
      output_sf   <- ddbs_is_closed(lines_sf)
      output_conn <- ddbs_is_closed("lines", conn = conn_test)
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats", {
      output_sf_fmt <- ddbs_is_closed(lines_ddbs, mode = "sf")
      expect_type(output_sf_fmt, "logical")
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_is_closed(lines_ddbs))
      expect_message(ddbs_is_closed("lines", conn = conn_test, name = "is_closed_tbl"))
      expect_message(ddbs_is_closed("lines", conn = conn_test, name = "is_closed_tbl", overwrite = TRUE))
      expect_true(ddbs_is_closed("lines", conn = conn_test, name = "is_closed_tbl2"))
      expect_no_message(ddbs_is_closed(lines_ddbs, quiet = TRUE))
      expect_no_message(ddbs_is_closed("lines", conn = conn_test, name = "is_closed_tbl", overwrite = TRUE, quiet = TRUE))
    })
    
    it("saves to database table correctly", {
      output_ddbs <- ddbs_is_closed(lines_ddbs)
      output_tbl <- ddbs_read_table(conn_test, "is_closed_tbl")
      
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("returns TRUE in a closed linestring", {
      coords <- rbind(c(0, 0), c(1, 0), c(1, 1), c(0, 1), c(0, 0))
      closed_ring_sf <- sf::st_sf(
        geometry = sf::st_sfc(sf::st_linestring(coords))
      )

      expect_true(ddbs_is_closed(closed_ring_sf, mode = "sf"))

    })
    
  })
  
  describe("errors", {
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_closed(lines_sf, new_column = 5))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_is_closed("lines", conn = NULL))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_is_closed(x = 999))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_is_closed(lines_ddbs, conn = 999))
    })
    
    it("validates new_column argument type", {
      expect_error(ddbs_is_closed(lines_ddbs, new_column = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_is_closed(lines_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_is_closed(lines_ddbs, quiet = 999))
    })
    
    it("validates table name exists", {
      expect_error(ddbs_is_closed(x = "999", conn = conn_test))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_is_closed(lines_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})


## close connection
ddbs_stop_conn(conn_test)

# Mapbox Vector Tiles (MVT) -----------------------------------------------------
#
# Everything related to MVT lives in this file:
#   - ddbs_as_mvt_geom()  : low-level ST_AsMVTGeom wrapper (geometry -> tile space)
#   - ddbs_write_mbtiles(): generate a vector tile pyramid into an MBTiles file
#   - internal helpers     : mvt_normalize_bounds(), mvt_tile_grid(),
#                            mvt_property_casts(), mvt_write_mbtiles_sqlite(),
#                            mvt_vector_layers_json()



#' Transform geometries into Mapbox Vector Tile coordinate space
#'
#' Prepares geometries for encoding as Mapbox Vector Tiles (MVT) by clipping
#' them to a tile's bounding box and transforming the coordinates into the
#' tile's integer pixel space. This is a low-level building block; to produce a
#' ready-to-serve tile pyramid use \code{\link{ddbs_write_mbtiles}}.
#'
#' @template x
#' @param bounds The tile bounding box, in the same CRS as \code{x}. Either a
#'   numeric vector \code{c(xmin, ymin, xmax, ymax)} or an object understood by
#'   \code{sf::st_bbox()} (e.g. an \code{sf}, \code{sfc}, or \code{bbox}
#'   object).
#' @param extent Integer. The width and height of the tile coordinate space
#'   (the geometry is mapped into the square \code{[0, extent]}). Defaults to
#'   \code{4096}.
#' @param buffer Integer. The number of pixels by which the tile bounds are
#'   expanded before clipping, to avoid rendering artefacts at tile edges.
#'   Defaults to \code{256}.
#' @param clip_geom Logical. If \code{TRUE} (default), geometries are clipped to
#'   the (buffered) tile bounds; geometries falling entirely outside become
#'   \code{NULL}. If \code{FALSE}, geometries are transformed but not clipped.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' This is a wrapper around DuckDB's \code{ST_AsMVTGeom}. The output geometries
#' use the MVT tile coordinate system: the origin is the top-left corner and the
#' y-axis points downwards, so the result is no longer in the input CRS (the CRS
#' is dropped). Geometries that fall completely outside the tile bounds return
#' \code{NULL} when \code{clip_geom = TRUE}.
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", package = "duckspatial")
#' )
#'
#' ## use the layer's own bounding box as the tile bounds
#' ddbs_as_mvt_geom(argentina_ddbs, bounds = sf::st_bbox(sf::st_as_sf(argentina_ddbs)))
#'
#' ## explicit numeric bounds and a smaller tile extent
#' ddbs_as_mvt_geom(
#'   argentina_ddbs,
#'   bounds = c(-74, -56, -53, -21),
#'   extent = 256
#' )
#' }
ddbs_as_mvt_geom <- function(
  x,
  bounds,
  extent = 4096L,
  buffer = 256L,
  clip_geom = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Handle function-specific errors
  assert_integer_scalar(extent, "extent")
  assert_integer_scalar(buffer, "buffer")
  assert_logic(clip_geom, "clip_geom")

  if (extent <= 0) cli::cli_abort("{.arg extent} must be a positive integer.")
  if (buffer < 0)  cli::cli_abort("{.arg buffer} must be a non-negative integer.")

  # 1. Normalize bounds to c(xmin, ymin, xmax, ymax)
  bbox <- mvt_normalize_bounds(bounds)

  # 2. Build the ST_AsMVTGeom argument string (bounds as a BOX_2D struct literal)
  bounds_sql <- glue::glue(
    "{{'min_x': {bbox[['xmin']]}, 'min_y': {bbox[['ymin']]}, ",
    "'max_x': {bbox[['xmax']]}, 'max_y': {bbox[['ymax']]}}}::BOX_2D"
  )
  mvt_args <- glue::glue(
    "{bounds_sql}, {as.integer(extent)}, {as.integer(buffer)}, {tolower(as.character(clip_geom))}"
  )

  # 3. Pass to template (drop CRS: output is in tile pixel coordinates)
  template_unary_ops(
    x = x,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_AsMVTGeom",
    other_args = mvt_args,
    drop_crs = TRUE
  )

}



#' Write a Mapbox Vector Tile pyramid to an MBTiles file
#'
#' Generates a pyramid of Mapbox Vector Tiles (MVT) from a spatial dataset and
#' stores them in an \href{https://github.com/mapbox/mbtiles-spec}{MBTiles} file
#' (a SQLite database). For each tile in the requested zoom range the geometries
#' are clipped and re-projected into tile space (\code{ST_AsMVTGeom}) and encoded
#' into a binary vector tile (\code{ST_AsMVT}).
#'
#' @template x
#' @param dsn Path to the output \code{.mbtiles} file.
#' @param layer_name Name of the vector tile layer. This is the value you pass
#'   as \code{source_layer} when adding the layer in a web map. Defaults to
#'   \code{"layer"}.
#' @param zoom_levels Integer vector of zoom levels to generate (e.g.
#'   \code{0:8}). Defaults to \code{0:6}.
#' @param extent Integer tile extent (resolution). Defaults to \code{4096}.
#' @param buffer Integer tile buffer in pixels. Defaults to \code{256}.
#' @template conn_null
#' @param overwrite Logical. Overwrite \code{dsn} if it already exists. Defaults
#'   to \code{FALSE}.
#' @template quiet
#'
#' @details
#' The input is re-projected to web mercator (\code{EPSG:3857}), so \code{x} must
#' have a CRS. All non-geometry columns are carried into the tiles as feature
#' properties; columns whose type is not supported by the vector tile format are
#' cast to \code{DOUBLE} (numeric types) or \code{VARCHAR} (everything else), and
#' \code{BLOB} columns are dropped.
#'
#' Writing the SQLite container relies on DuckDB's \code{sqlite} extension, which
#' is installed and loaded automatically (an internet connection may be required
#' the first time).
#'
#' MBTiles can be served with a vector tile server, or converted to
#' \href{https://github.com/protomaps/PMTiles}{PMTiles} (e.g. with the
#' \code{pmtiles convert} CLI) for static hosting. The resulting tiles can then
#' be displayed with the \pkg{mapgl} package, for example:
#'
#' \preformatted{
#' library(mapgl)
#' maplibre() |>
#'   add_pmtiles_source(id = "src", url = "https://example.com/argentina.pmtiles") |>
#'   add_fill_layer(id = "fill", source = "src", source_layer = "argentina")
#' }
#'
#' @return The path to \code{dsn}, invisibly.
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", package = "duckspatial")
#' )
#'
#' ddbs_write_mbtiles(
#'   argentina_ddbs,
#'   dsn         = tempfile(fileext = ".mbtiles"),
#'   layer_name  = "argentina",
#'   zoom_levels = 0:6
#' )
#' }
ddbs_write_mbtiles <- function(
  x,
  dsn,
  layer_name = "layer",
  zoom_levels = 0:6,
  extent = 4096L,
  buffer = 256L,
  conn = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_character(conn, x)
  assert_character_scalar(dsn, "dsn")
  assert_character_scalar(layer_name, "layer_name")
  assert_integer_scalar(extent, "extent")
  assert_integer_scalar(buffer, "buffer")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")

  if (!is.numeric(zoom_levels) || length(zoom_levels) < 1 ||
      any(zoom_levels != as.integer(zoom_levels)) ||
      any(zoom_levels < 0) || any(zoom_levels > 22)) {
    cli::cli_abort("{.arg zoom_levels} must be a vector of integers between 0 and 22.")
  }
  zoom_levels <- sort(unique(as.integer(zoom_levels)))

  if (extent <= 0) cli::cli_abort("{.arg extent} must be a positive integer.")
  if (buffer < 0)  cli::cli_abort("{.arg buffer} must be a non-negative integer.")

  if (file.exists(dsn)) {
    if (!overwrite) {
      cli::cli_abort("{.arg dsn} already exists. Use {.code overwrite = TRUE} to replace it.")
    }
    unlink(dsn)
  }

  # 1. Resolve input to a query in a DuckDB connection
  x <- normalize_spatial_input(x, conn)
  crs_x  <- sf::st_crs(ddbs_crs(x, conn))
  if (is.na(crs_x)) {
    cli::cli_abort("{.arg x} must have a CRS so it can be re-projected to web mercator (EPSG:3857).")
  }

  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  on.exit(resolve_conn$cleanup(), add = TRUE)

  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)

  x_geom <- attr(x, "sf_column") %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  # 2. Source CRS string for ST_Transform
  src_crs <- if (!is.na(crs_x$epsg)) paste0("EPSG:", crs_x$epsg) else crs_x$wkt
  src_q   <- ddbs_quote_sql_string(target_conn, src_crs)

  # 3. Introspect attribute columns and build supported-type property casts
  desc      <- DBI::dbGetQuery(target_conn, glue::glue("DESCRIBE SELECT * FROM {x_list$query_name};"))
  attr_cols <- desc[desc$column_name != x_geom, c("column_name", "column_type"), drop = FALSE]
  prop_casts    <- mvt_property_casts(attr_cols$column_name, attr_cols$column_type)
  struct_fields <- if (length(prop_casts)) {
    paste0(paste(prop_casts, collapse = ",\n               "), ",\n               ")
  } else {
    ""
  }

  # 4. Re-project to web mercator in a temp table
  tmp3857 <- ddbs_temp_table_name()
  DBI::dbExecute(target_conn, glue::glue(
    'CREATE TEMP TABLE {tmp3857} AS
     SELECT * EXCLUDE ("{x_geom}"),
            ST_Transform("{x_geom}", {src_q}, \'EPSG:3857\', always_xy := true) AS mvt_src_geom
     FROM {x_list$query_name};'
  ))
  on.exit(DBI::dbExecute(target_conn, glue::glue("DROP TABLE IF EXISTS {tmp3857};")), add = TRUE)

  # 5. Data bbox in web mercator (cheap: per-row min/max, no union)
  bb <- DBI::dbGetQuery(target_conn, glue::glue(
    "SELECT min(ST_XMin(mvt_src_geom)) xmin, min(ST_YMin(mvt_src_geom)) ymin,
            max(ST_XMax(mvt_src_geom)) xmax, max(ST_YMax(mvt_src_geom)) ymax
     FROM {tmp3857};"
  ))

  # 6. Tile grid covering the data for each requested zoom level
  tiles <- mvt_tile_grid(zoom_levels, bb)
  if (nrow(tiles) == 0) {
    cli::cli_abort("No tiles cover the data for the requested {.arg zoom_levels}.")
  }
  tmp_tiles <- ddbs_temp_table_name()
  DBI::dbWriteTable(target_conn, tmp_tiles, tiles, temporary = TRUE)
  on.exit(DBI::dbExecute(target_conn, glue::glue("DROP TABLE IF EXISTS {tmp_tiles};")), add = TRUE)

  # 7. Generate the vector tiles (one binary blob per z/x/y)
  tmp_out <- ddbs_temp_table_name()
  layer_q <- ddbs_quote_sql_string(target_conn, layer_name)
  DBI::dbExecute(target_conn, glue::glue(
    "CREATE TEMP TABLE {tmp_out} AS
     WITH prepped AS (
       SELECT t.z, t.x, t.y,
              struct_pack(
               {struct_fields}geom := ST_AsMVTGeom(
                  d.mvt_src_geom, ST_Extent(ST_TileEnvelope(t.z, t.x, t.y)),
                  {as.integer(extent)}, {as.integer(buffer)}, true)
              ) AS feature
       FROM {tmp_tiles} t
       JOIN {tmp3857} d ON ST_Intersects(d.mvt_src_geom, ST_TileEnvelope(t.z, t.x, t.y))
     )
     SELECT z, x, y, ST_AsMVT(feature, {layer_q}) AS tile
     FROM prepped
     WHERE feature.geom IS NOT NULL
     GROUP BY z, x, y;"
  ))
  on.exit(DBI::dbExecute(target_conn, glue::glue("DROP TABLE IF EXISTS {tmp_out};")), add = TRUE)

  n_tiles <- DBI::dbGetQuery(target_conn, glue::glue("SELECT count(*) n FROM {tmp_out};"))$n

  # 8. Data bbox in lon/lat (EPSG:4326) for the MBTiles metadata
  ll <- DBI::dbGetQuery(target_conn, glue::glue(
    "SELECT min(ST_XMin(g)) xmin, min(ST_YMin(g)) ymin,
            max(ST_XMax(g)) xmax, max(ST_YMax(g)) ymax
     FROM (SELECT ST_Transform(mvt_src_geom, 'EPSG:3857', 'EPSG:4326', always_xy := true) g
           FROM {tmp3857}) t;"
  ))

  # 9. Write the MBTiles SQLite container
  mvt_write_mbtiles_sqlite(target_conn, tmp_out, dsn, layer_name, attr_cols, zoom_levels, ll)

  if (!quiet) {
    cli::cli_alert_success("Wrote {n_tiles} tile{?s} to {.file {dsn}}.")
  }

  invisible(dsn)
}



# Internal helpers -------------------------------------------------------------

#' Normalize a `bounds` argument to c(xmin, ymin, xmax, ymax)
#'
#' @param bounds A numeric vector of length 4 or an object understood by
#'   \code{sf::st_bbox()}.
#'
#' @keywords internal
#' @noRd
mvt_normalize_bounds <- function(bounds) {

  if (inherits(bounds, c("bbox", "sf", "sfc"))) {
    bb <- sf::st_bbox(bounds)
    bbox <- c(xmin = bb[["xmin"]], ymin = bb[["ymin"]],
              xmax = bb[["xmax"]], ymax = bb[["ymax"]])
  } else if (is.numeric(bounds) && length(bounds) == 4) {
    nms <- names(bounds)
    if (!is.null(nms) && all(c("xmin", "ymin", "xmax", "ymax") %in% nms)) {
      bbox <- c(xmin = bounds[["xmin"]], ymin = bounds[["ymin"]],
                xmax = bounds[["xmax"]], ymax = bounds[["ymax"]])
    } else {
      bbox <- c(xmin = bounds[1], ymin = bounds[2],
                xmax = bounds[3], ymax = bounds[4])
    }
  } else {
    cli::cli_abort(
      "{.arg bounds} must be a numeric vector {.code c(xmin, ymin, xmax, ymax)} or an {.cls sf}/{.cls sfc}/{.cls bbox} object."
    )
  }

  if (anyNA(bbox)) {
    cli::cli_abort("{.arg bounds} must not contain missing values.")
  }
  if (bbox[["xmax"]] <= bbox[["xmin"]] || bbox[["ymax"]] <= bbox[["ymin"]]) {
    cli::cli_abort("{.arg bounds} must have {.code xmax > xmin} and {.code ymax > ymin}.")
  }

  bbox
}



#' Compute the XYZ tile grid covering a web-mercator bbox over a zoom range
#'
#' @param zoom_levels Integer vector of zoom levels.
#' @param bb A one-row data frame with `xmin`, `ymin`, `xmax`, `ymax` in
#'   EPSG:3857.
#'
#' @return A data frame with integer columns `z`, `x`, `y` (XYZ / Google scheme).
#' @keywords internal
#' @noRd
mvt_tile_grid <- function(zoom_levels, bb) {

  world <- 20037508.342789244  # half the web-mercator extent

  grids <- lapply(zoom_levels, function(z) {
    n  <- 2^z
    ts <- (2 * world) / n
    xmin_t <- max(0,     floor((bb$xmin + world) / ts))
    xmax_t <- min(n - 1, floor((bb$xmax + world) / ts))
    ## y is flipped: tile row 0 is at the top (north)
    ymin_t <- max(0,     floor((world - bb$ymax) / ts))
    ymax_t <- min(n - 1, floor((world - bb$ymin) / ts))
    if (xmax_t < xmin_t || ymax_t < ymin_t) return(NULL)
    expand.grid(
      z = as.integer(z),
      x = as.integer(seq(xmin_t, xmax_t)),
      y = as.integer(seq(ymin_t, ymax_t))
    )
  })

  out <- do.call(rbind, grids)
  out %||% data.frame(z = integer(), x = integer(), y = integer())
}



#' Build struct_pack property assignments, casting to MVT-supported types
#'
#' `ST_AsMVT` only supports VARCHAR, FLOAT, DOUBLE, INTEGER, BIGINT and BOOLEAN
#' property columns. Numeric types are cast to DOUBLE, other unsupported types to
#' VARCHAR, and BLOB columns are dropped.
#'
#' @param cols,types Character vectors of column names and DuckDB type names.
#'
#' @return A character vector of `"name" := d."name"[::CAST]` expressions.
#' @keywords internal
#' @noRd
mvt_property_casts <- function(cols, types) {

  supported <- c("VARCHAR", "FLOAT", "DOUBLE", "INTEGER", "BIGINT", "BOOLEAN")
  numeric_t <- c("TINYINT", "SMALLINT", "UTINYINT", "USMALLINT", "UINTEGER",
                 "UBIGINT", "HUGEINT", "UHUGEINT", "DECIMAL", "REAL", "NUMERIC")

  out <- character(0)
  for (i in seq_along(cols)) {
    nm      <- cols[i]
    base_ty <- toupper(sub("\\(.*", "", types[i]))  # strip DECIMAL(p,s) etc.

    if (base_ty == "BLOB") next

    expr <- if (base_ty %in% supported) {
      sprintf('"%s" := d."%s"', nm, nm)
    } else if (base_ty %in% numeric_t) {
      sprintf('"%s" := d."%s"::DOUBLE', nm, nm)
    } else {
      sprintf('"%s" := d."%s"::VARCHAR', nm, nm)
    }
    out <- c(out, expr)
  }

  out
}



#' Build the MBTiles `json` metadata value describing the vector layer
#'
#' @keywords internal
#' @noRd
mvt_vector_layers_json <- function(layer_name, attr_cols, minz, maxz) {

  numeric_t <- c("TINYINT", "SMALLINT", "INTEGER", "BIGINT", "UTINYINT",
                 "USMALLINT", "UINTEGER", "UBIGINT", "HUGEINT", "UHUGEINT",
                 "FLOAT", "DOUBLE", "DECIMAL", "REAL", "NUMERIC")

  fields <- list()
  for (i in seq_len(nrow(attr_cols))) {
    base_ty <- toupper(sub("\\(.*", "", attr_cols$column_type[i]))
    if (base_ty == "BLOB") next
    fields[[attr_cols$column_name[i]]] <-
      if (base_ty == "BOOLEAN") "Boolean"
      else if (base_ty %in% numeric_t) "Number"
      else "String"
  }

  obj <- list(vector_layers = list(list(
    id      = layer_name,
    fields  = fields,
    minzoom = minz,
    maxzoom = maxz
  )))

  ## force `fields` to render as a JSON object even when empty
  json <- jsonlite::toJSON(obj, auto_unbox = TRUE)
  as.character(json)
}



#' Write generated tiles + metadata into an MBTiles SQLite file via DuckDB
#'
#' @keywords internal
#' @noRd
mvt_write_mbtiles_sqlite <- function(conn, tiles_table, dsn, layer_name,
                                     attr_cols, zoom_levels, ll) {

  ## Ensure the sqlite extension is available
  tryCatch({
    DBI::dbExecute(conn, "INSTALL sqlite;")
    DBI::dbExecute(conn, "LOAD sqlite;")
  }, error = function(e) {
    cli::cli_abort(c(
      "Could not load the DuckDB {.pkg sqlite} extension required to write MBTiles.",
      "x" = conditionMessage(e)
    ))
  })

  dsn_q <- ddbs_quote_sql_string(conn, dsn)
  DBI::dbExecute(conn, glue::glue("ATTACH {dsn_q} AS mbtiles_out (TYPE sqlite);"))
  on.exit(try(DBI::dbExecute(conn, "DETACH mbtiles_out;"), silent = TRUE), add = TRUE)

  DBI::dbExecute(conn, "CREATE TABLE mbtiles_out.metadata (name TEXT, value TEXT);")
  DBI::dbExecute(conn,
    "CREATE TABLE mbtiles_out.tiles (zoom_level INTEGER, tile_column INTEGER, tile_row INTEGER, tile_data BLOB);")

  ## MBTiles uses the TMS scheme (y flipped): tile_row = 2^z - 1 - y
  DBI::dbExecute(conn, glue::glue(
    "INSERT INTO mbtiles_out.tiles
     SELECT z, x, (CAST(pow(2, z) AS INTEGER) - 1 - y), tile FROM {tiles_table};"
  ))

  ## Metadata rows
  minz <- min(zoom_levels)
  maxz <- max(zoom_levels)
  meta <- data.frame(
    name = c("name", "format", "minzoom", "maxzoom", "bounds", "center", "type", "json"),
    value = c(
      layer_name,
      "pbf",
      as.character(minz),
      as.character(maxz),
      sprintf("%f,%f,%f,%f", ll$xmin, ll$ymin, ll$xmax, ll$ymax),
      sprintf("%f,%f,%d", (ll$xmin + ll$xmax) / 2, (ll$ymin + ll$ymax) / 2, minz),
      "overlay",
      mvt_vector_layers_json(layer_name, attr_cols, minz, maxz)
    ),
    stringsAsFactors = FALSE
  )

  values_sql <- paste(
    sprintf("(%s, %s)",
            ddbs_quote_sql_string(conn, meta$name),
            ddbs_quote_sql_string(conn, meta$value)),
    collapse = ",\n"
  )
  DBI::dbExecute(conn, glue::glue(
    "INSERT INTO mbtiles_out.metadata (name, value) VALUES {values_sql};"
  ))

  invisible(TRUE)
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## A 10x10 polygon (in EPSG:4326) that exactly fills the c(0,0,10,10) bounds
mvt_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(
    sf::st_polygon(list(matrix(c(0,0, 10,0, 10,10, 0,10, 0,0), ncol = 2, byrow = TRUE))),
    crs = 4326
  )
)
mvt_ddbs <- as_duckspatial_df(mvt_sf)
ddbs_write_table(conn_test, mvt_sf, "mvt_poly")


# 1. ddbs_as_mvt_geom() --------------------------------------------------

describe("ddbs_as_mvt_geom()", {

  describe("expected behavior", {

    it("returns a duckspatial_df by default", {
      expect_s3_class(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10)), "duckspatial_df")
    })

    it("returns an sf object with mode = 'sf'", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), mode = "sf")
      expect_s3_class(output, "sf")
    })

    it("works on all input formats", {
      out_sf   <- ddbs_as_mvt_geom(mvt_sf, bounds = c(0, 0, 10, 10), mode = "sf")
      out_ddbs <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), mode = "sf")
      out_conn <- ddbs_as_mvt_geom("mvt_poly", bounds = c(0, 0, 10, 10), conn = conn_test, mode = "sf")

      expect_equal(sf::st_as_text(sf::st_geometry(out_sf)), sf::st_as_text(sf::st_geometry(out_ddbs)))
      expect_equal(sf::st_as_text(sf::st_geometry(out_sf)), sf::st_as_text(sf::st_geometry(out_conn)))
    })

    it("maps the geometry into [0, extent] tile space", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), extent = 4096, mode = "sf")
      coords <- sf::st_coordinates(output)[, c("X", "Y")]
      expect_equal(range(coords), c(0, 4096))
    })

    it("respects the extent argument", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), extent = 256, mode = "sf")
      expect_equal(max(sf::st_coordinates(output)[, c("X", "Y")]), 256)
    })

    it("drops the CRS (output is in tile pixel coordinates)", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10))
      expect_true(is.na(ddbs_crs(output)))
    })

    it("accepts an sf bbox as bounds", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = sf::st_bbox(mvt_sf), mode = "sf")
      expect_s3_class(output, "sf")
    })

    it("keeps non-geometry columns", {
      output <- ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), mode = "sf")
      expect_true("id" %in% names(output))
    })

    it("writes a table when name is provided", {
      ok <- ddbs_as_mvt_geom("mvt_poly", bounds = c(0, 0, 10, 10), conn = conn_test, name = "mvt_out")
      expect_true(ok)
      expect_true(DBI::dbExistsTable(conn_test, "mvt_out"))
    })
  })

  describe("errors", {

    it("errors on bounds of the wrong length", {
      expect_error(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10)))
    })

    it("errors when xmax <= xmin or ymax <= ymin", {
      expect_error(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(10, 10, 0, 0)))
    })

    it("errors on a non-positive extent", {
      expect_error(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), extent = 0))
    })

    it("errors on a negative buffer", {
      expect_error(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), buffer = -1))
    })

    it("errors on a non-logical clip_geom", {
      expect_error(ddbs_as_mvt_geom(mvt_ddbs, bounds = c(0, 0, 10, 10), clip_geom = "yes"))
    })

    it("errors on invalid x type", {
      expect_error(ddbs_as_mvt_geom(999, bounds = c(0, 0, 10, 10)))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_as_mvt_geom("mvt_poly", bounds = c(0, 0, 10, 10)))
    })
  })
})


# 2. ddbs_write_mbtiles() ------------------------------------------------

## sample data with a string, an integer and a double attribute column
tile_sf <- sf::st_read(
  system.file("spatial/argentina.geojson", package = "duckspatial"), quiet = TRUE
)
tile_sf$name <- "Argentina"
tile_sf$pop  <- 45000000L
tile_sf$dens <- 16.5
tile_ddbs <- as_duckspatial_df(tile_sf[, c("name", "pop", "dens")])

## helper to read an mbtiles file back through a fresh duckdb sqlite attach
read_mbtiles <- function(conn, dsn) {
  DBI::dbExecute(conn, "INSTALL sqlite; LOAD sqlite;")
  DBI::dbExecute(conn, sprintf("ATTACH '%s' AS mbr (TYPE sqlite, READ_ONLY);", dsn))
  on.exit(try(DBI::dbExecute(conn, "DETACH mbr;"), silent = TRUE), add = TRUE)
  list(
    tiles    = DBI::dbGetQuery(conn, "SELECT zoom_level, tile_column, tile_row, octet_length(tile_data) nbytes FROM mbr.tiles;"),
    metadata = DBI::dbGetQuery(conn, "SELECT name, value FROM mbr.metadata;")
  )
}

## the sqlite extension may be unavailable offline; skip the section if so
sqlite_ok <- tryCatch({
  DBI::dbExecute(conn_test, "INSTALL sqlite; LOAD sqlite;")
  TRUE
}, error = function(e) FALSE)

describe("ddbs_write_mbtiles()", {

  testthat::skip_if_not(sqlite_ok, "DuckDB sqlite extension not available")

  describe("expected behavior", {

    it("writes an mbtiles file and returns its path invisibly", {
      dsn <- tempfile(fileext = ".mbtiles")
      res <- withVisible(ddbs_write_mbtiles(tile_ddbs, dsn = dsn, layer_name = "argentina", zoom_levels = 0:4, quiet = TRUE))
      expect_false(res$visible)
      expect_equal(res$value, dsn)
      expect_true(file.exists(dsn))
    })

    it("produces a tile pyramid across the requested zoom levels", {
      dsn <- tempfile(fileext = ".mbtiles")
      ddbs_write_mbtiles(tile_ddbs, dsn = dsn, layer_name = "argentina", zoom_levels = 0:4, quiet = TRUE)
      mb <- read_mbtiles(conn_test, dsn)
      expect_true(nrow(mb$tiles) > 0)
      expect_setequal(sort(unique(mb$tiles$zoom_level)), 0:4)
      expect_true(all(mb$tiles$nbytes > 0))
    })

    it("writes the expected metadata (format, zoom range, layer fields)", {
      dsn <- tempfile(fileext = ".mbtiles")
      ddbs_write_mbtiles(tile_ddbs, dsn = dsn, layer_name = "argentina", zoom_levels = 0:3, quiet = TRUE)
      meta <- read_mbtiles(conn_test, dsn)$metadata
      kv <- stats::setNames(meta$value, meta$name)
      expect_equal(kv[["format"]], "pbf")
      expect_equal(kv[["minzoom"]], "0")
      expect_equal(kv[["maxzoom"]], "3")
      expect_match(kv[["json"]], "\"argentina\"")
      expect_match(kv[["json"]], "\"pop\":\"Number\"")
      expect_match(kv[["json"]], "\"name\":\"String\"")
    })

    it("works with a character table name input", {
      ddbs_write_table(conn_test, tile_sf[, c("name", "pop", "dens")], "tile_src")
      dsn <- tempfile(fileext = ".mbtiles")
      expect_invisible(ddbs_write_mbtiles("tile_src", dsn = dsn, conn = conn_test, zoom_levels = 0:2, quiet = TRUE))
      expect_true(file.exists(dsn))
    })
  })

  describe("errors", {

    it("errors when dsn exists and overwrite = FALSE", {
      dsn <- tempfile(fileext = ".mbtiles")
      ddbs_write_mbtiles(tile_ddbs, dsn = dsn, zoom_levels = 0:1, quiet = TRUE)
      expect_error(ddbs_write_mbtiles(tile_ddbs, dsn = dsn, zoom_levels = 0:1))
    })

    it("errors on invalid zoom_levels", {
      expect_error(ddbs_write_mbtiles(tile_ddbs, dsn = tempfile(fileext = ".mbtiles"), zoom_levels = c(-1, 2)))
    })

    it("errors when the input has no CRS", {
      no_crs <- sf::st_set_crs(tile_sf[, "name"], NA)
      expect_error(ddbs_write_mbtiles(as_duckspatial_df(no_crs), dsn = tempfile(fileext = ".mbtiles")))
    })

    it("errors on invalid x type", {
      expect_error(ddbs_write_mbtiles(999, dsn = tempfile(fileext = ".mbtiles")))
    })
  })
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' sf methods for duckspatial_df
#'
#' These methods provide sf compatibility for duckspatial_df objects,
#' allowing them to work with sf functions like st_crs(), st_geometry(), etc.
#'
#' @name duckspatial_df_sf
#' @keywords internal
NULL

#' @rdname duckspatial_df_sf
#' @export
#' @importFrom sf st_crs
st_crs.duckspatial_df <- function(x, ...) {
  res <- attr(x, "crs") %||% sf::st_crs(NA)
  
  # If we have dots (like parameters=TRUE), we must re-dispatch
  # to sf::st_crs to trigger the full parameter extraction.
  # sf only extracts these parameters (srid, axes, etc.) if the input
  # is an sf or sfc object. 
  if (length(list(...)) > 0 && !is.na(res)) {
    # Wrap in dummy sfc to force sf to extract parameters
    dummy_sfc <- sf::st_sfc(sf::st_point(), crs = res)
    return(sf::st_crs(dummy_sfc, ...))
  }
  
  res
}

#' @rdname duckspatial_df_sf
#' @export
#' @importFrom sf st_geometry
st_geometry.duckspatial_df <- function(obj, ...) {
  # For duckspatial_df, st_geometry is a materialization point.
  # We delegate to st_as_sf to ensure we use the robust collect() path
  # which handles ST_AsWKB conversion correctly.
  sf::st_geometry(sf::st_as_sf(obj, ...))
}

#' Convert WKB data to sfc
#' @keywords internal
convert_wkb_to_sfc <- function(geom_data, crs) {
  if (inherits(geom_data, "sfc")) return(geom_data)
  
  # Use wk for fast conversion if possible
  tryCatch({
    attributes(geom_data) <- NULL
    wkb_obj <- wk::new_wk_wkb(geom_data)
    sf::st_as_sfc(wkb_obj, crs = crs)
  }, error = function(e) {
    # Fallback to slow path
    sf::st_as_sfc(structure(geom_data, class = "WKB"), crs = crs)
  })
}

#' @rdname duckspatial_df_sf
#' @export
#' @importFrom sf st_bbox
st_bbox.duckspatial_df <- function(obj, ...) {
  geom_col <- attr(obj, "sf_column") %||% "geom"
  crs_obj <- st_crs(obj)
  
  # Try to use DuckDB's ST_Extent for efficiency
  # We use the lazy table directly
  tryCatch({
    conn <- dbplyr::remote_con(obj)
    query_sql <- dbplyr::sql_render(obj)
    
    # We construct the extent query
    # ST_Extent returns the MBR of all geometries in the table/query
    extent_query <- glue::glue(
      "SELECT 
        ST_XMin(ext) as xmin, 
        ST_YMin(ext) as ymin, 
        ST_XMax(ext) as xmax, 
        ST_YMax(ext) as ymax 
       FROM (SELECT ST_Extent(ST_Collect(LIST({DBI::dbQuoteIdentifier(conn, geom_col)}))) as ext FROM ({query_sql}))"
    )
    
    res <- DBI::dbGetQuery(conn, extent_query)
    
    if (nrow(res) > 0 && !is.na(res$xmin[1])) {
      return(sf::st_bbox(c(
        xmin = res$xmin[1], 
        ymin = res$ymin[1], 
        xmax = res$xmax[1], 
        ymax = res$ymax[1]
      ), crs = crs_obj))
    }
    
    # Empty result or all NAs
    sf::st_bbox(c(xmin = NA_real_, ymin = NA_real_, 
                  xmax = NA_real_, ymax = NA_real_), 
                crs = crs_obj)
  }, error = function(e) {
    # If anything fails, return NA bbox
    sf::st_bbox(c(xmin = NA_real_, ymin = NA_real_, 
                  xmax = NA_real_, ymax = NA_real_), 
                crs = crs_obj)
  })
}

#' Collect and materialize a duckspatial_df
#'
#' Materializes a lazy \code{duckspatial_df} object by executing the underlying
#' DuckDB query. Supports multiple output formats.
#'
#' @param x A \code{duckspatial_df} object
#' @param ... Additional arguments passed to \code{collect}
#' @param as Output format. One of:
#'   \describe{
#'     \item{\code{"sf"}}{(Default) Returns an \code{sf} object with \code{sfc} geometry}
#'     \item{\code{"tibble"}}{Returns a tibble with geometry column dropped (fastest)}
#'     \item{\code{"raw"}}{Returns a tibble with geometry as raw WKB bytes}
#'     \item{\code{"geoarrow"}}{Returns a tibble with geometry as \code{geoarrow_vctr}}
#'   }
#'
#' @returns Data in the specified format
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' # Load lazy spatial data
#' nc <- ddbs_open_dataset(system.file("shape/nc.shp", package = "sf"))
#'
#' # Perform lazy operations
#' result <- nc |> dplyr::filter(AREA > 0.1)
#'
#' # Collect to sf (default)
#' result_sf <- ddbs_collect(result)
#' plot(result_sf["AREA"])
#'
#' # Collect as tibble without geometry (fast)
#' result_tbl <- ddbs_collect(result, as = "tibble")
#'
#' # Collect with raw WKB bytes
#' result_raw <- ddbs_collect(result, as = "raw")
#'
#' # Collect as geoarrow for Arrow workflows
#' result_ga <- ddbs_collect(result, as = "geoarrow")
#' }
ddbs_collect <- function(x, ..., as = c("sf", "tibble", "raw", "geoarrow")) {
  if (!inherits(x, "duckspatial_df")) {
    cli::cli_abort("{.arg x} must be a {.cls duckspatial_df} object.")
  }
  as <- match.arg(as)
  dplyr::collect(x, ..., as = as)
}

#' Force computation of a lazy duckspatial_df
#'
#' Executes the accumulated query and stores the result in a DuckDB temporary
#' table. The result remains lazy (a \code{duckspatial_df}) but points to the
#' materialized data, avoiding repeated computation of complex query plans.
#'
#' This is useful when you want to:
#' \itemize{
#'   \item Cache intermediate results for reuse across multiple subsequent operations
#'   \item Simplify complex query plans before heavy operations like spatial joins
#'   \item Force execution at a specific point without pulling data into R memory
#' }
#'
#' @param x A \code{duckspatial_df} object
#' @param ... Additional arguments passed to \code{dplyr::compute}
#' @param name Optional name for the result table. If NULL, a unique temporary
#'   name is generated.
#' @param temporary If TRUE (default), creates a temporary table that is
#'   automatically cleaned up when the connection closes.
#'
#' @returns A new \code{duckspatial_df} pointing to the materialized table
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#' library(dplyr)
#'
#' # Load lazy spatial data
#' countries <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", package = "duckspatial")
#' )
#'
#' # Complex pipeline - ddbs_compute() caches intermediate result
#' cached <- countries |>
#'   filter(CNTR_ID %in% c("DE", "FR", "IT")) |>
#'   ddbs_compute()  # Execute and store in temp table
#'
#' # Check query plan - should reference temp table
#' show_query(cached)
#'
#' # Further operations continue from cached result
#' result <- cached |>
#'   ddbs_filter(other_layer, predicate = "intersects") |>
#'   st_as_sf()
#' }
ddbs_compute <- function(x, ..., name = NULL, temporary = TRUE) {
  if (!inherits(x, "duckspatial_df")) {
    cli::cli_abort("{.arg x} must be a {.cls duckspatial_df} object.")
  }
  dplyr::compute(x, name = name, temporary = temporary, ...)
}

#' @rdname duckspatial_df_sf
#' @export
#' @importFrom sf st_as_sf
st_as_sf.duckspatial_df <- function(x, ...) {
  # st_as_sf always returns sf, ignore any as= argument
  dplyr::collect(x, ..., as = "sf")
}

#' @rdname duckspatial_df_sf
#' @export
print.duckspatial_df <- function(x, ..., n = 10) {

  ## get metadata for the header
  geom_col <- attr(x, "sf_column") %||% "geom"
  crs <- ddbs_crs(x)
  # bbox <- st_bbox(x)
  bbox <- ddbs_bbox(x)
  geomtype <- ddbs_geometry_type(x, by_feature = FALSE)
  
  ## header with visual separator
  cat(cli::col_white("# A duckspatial lazy spatial table\n"))
  
  ## metadata with icons/symbols
  cat(cli::col_white("#"), cli::col_blue("\u25cf CRS:"), cli::col_white(ddbs_format_crs(crs)), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Geometry column:"), cli::col_white(geom_col), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Geometry type:"), cli::col_white(geomtype), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Bounding box:"), 
      cli::col_white(sprintf("xmin: %.5g ymin: %.5g xmax: %.5g ymax: %.5g", 
                              bbox["xmin"], bbox["ymin"], bbox["xmax"], bbox["ymax"])), "\n")
  
  ## info box
  cat(cli::col_white("# Data backed by DuckDB (dbplyr lazy evaluation)\n"))
  cat(cli::col_white("# Use"), 
      cli::style_bold(cli::col_green("ddbs_collect()")), 
      cli::col_white("or"), 
      cli::style_bold(cli::col_green("st_as_sf()")), 
      cli::col_white("to materialize to sf\n"))
  cat(cli::col_white("#\n"))
  
  ## print preview
  tryCatch({
    remote_name <- attr(x, "source_table")
    remote_conn <- attr(x, "source_conn") %||% dbplyr::remote_con(x)
    head_data <- dplyr::tbl(remote_conn, remote_name)
    print(head_data, n = n)
  }, error = function(e) {
    cat(cli::col_yellow("\u26a0 Preview unavailable\n"))
  })
  
  invisible(x)
}

#' Format a CRS object compactly for printing
#' @param crs An sf crs object
#' @return A character string
#' @keywords internal
ddbs_format_crs <- function(crs) {
  if (is.na(crs)) return("NA")
  
  # 1. If EPSG code is available, use it (shortest, most recognizable)
  if (!is.null(crs$epsg) && !is.na(crs$epsg)) {
    return(paste0("EPSG:", crs$epsg))
  }
  
  # 2. Try to extract ID["Authority", "Code"] from WKT if EPSG is missing
  # This often covers well-known CRSs that aren't fully resolved in the object
  wkt <- crs$wkt
  if (!is.null(wkt)) {
    # Match ID at the end of the WKT block, e.g. ID["OGC","CRS84"]]
    id_match <- regmatches(wkt, regexec('ID\\["([A-Za-z0-9_]+)",\\s*"?([A-Za-z0-9_]+)"?\\]\\]$', wkt))
    if (length(id_match[[1]]) == 3) {
      return(paste0(id_match[[1]][2], ":", id_match[[1]][3]))
    }
  }

  # 3. Check for OGC:CRS84 specifically if it wasn't caught
  if (!is.null(crs$input) && crs$input == "OGC:CRS84") return("OGC:CRS84")

  # 4. If input is short and not JSON, use it
  input <- crs$input
  if (!is.null(input) && nchar(input) > 0 && nchar(input) < 50 && !grepl("^\\{", input)) {
    return(input)
  }
  
  # 5. Fallback to Name if available
  if (!is.null(crs$Name) && nchar(crs$Name) > 0) return(crs$Name)
  
  # 6. Absolute fallback: truncated input or just 'custom'
  if (is.null(input)) return("unknown")
  if (nchar(input) > 50) return(paste0(substr(input, 1, 47), "..."))
  input
}

#' Get the geometry column name
#' @param x A duckspatial_df object
#' @return Character string with geometry column name
#' @export
ddbs_geom_col <- function(x) {
  if (!is_duckspatial_df(x)) {
    if (inherits(x, "sf")) {
      return(attr(x, "sf_column"))
    }
    cli::cli_abort("{.arg x} must be a duckspatial_df or sf object.")
  }
  attr(x, "sf_column") %||% "geom"
}

# =============================================================================
# Tests for duckspatial_df dplyr methods
# Tests: dplyr_reconstruct, collect, compute, left_join, inner_join, head
# Note: nc_sf is loaded from setup.R
# =============================================================================

# =============================================================================
# dplyr verb class preservation
# =============================================================================

testthat::skip_on_cran()

test_that("dplyr verbs preserve duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  # Test filter
  filtered <- nc_lazy |> dplyr::filter(AREA > 0.1)
  expect_s3_class(filtered, "duckspatial_df")
  expect_equal(attr(filtered, "crs"), attr(nc_lazy, "crs"))
  
  # Test mutate
  mutated <- nc_lazy |> dplyr::mutate(area_sq = AREA * AREA)
  expect_s3_class(mutated, "duckspatial_df")
  expect_equal(attr(mutated, "crs"), attr(nc_lazy, "crs"))
  
  # Test select
  selected <- nc_lazy |> dplyr::select(NAME, AREA, geometry)
  expect_s3_class(selected, "duckspatial_df")
  expect_equal(attr(selected, "crs"), attr(nc_lazy, "crs"))
  
  # Test arrange
  arranged <- nc_lazy |> dplyr::arrange(AREA)
  expect_s3_class(arranged, "duckspatial_df")
  expect_equal(attr(arranged, "crs"), attr(nc_lazy, "crs"))
})

# TODO - Implement geometry aggregation in summarize and then re-enable this test
# test_that("group_by preserves duckspatial_df class", {
#   conn <- ddbs_temp_conn()
#   ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
#   # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
#   #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
#   nc_lazy <- as_duckspatial_df("nc_test", conn)
  
#   grouped <- nc_lazy |> dplyr::group_by(SID74)
  
#   expect_s3_class(grouped, "duckspatial_df")
#   expect_equal(attr(grouped, "crs"), attr(nc_lazy, "crs"))
#   expect_equal(attr(grouped, "sf_column"), attr(nc_lazy, "sf_column"))
# })

# TODO - Implement geometry aggregation in summarize and then re-enable this test
# test_that("summarize preserves duckspatial_df class", {
#   conn <- ddbs_temp_conn()
#   ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
#   # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
#   #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
#   nc_lazy <- as_duckspatial_df("nc_test", conn)
  
#   summarized <- nc_lazy |> 
#     dplyr::group_by(SID74) |> 
#     dplyr::summarize(total_area = sum(AREA, na.rm = TRUE), .groups = "drop")
  
#   expect_s3_class(summarized, "duckspatial_df")
#   expect_equal(attr(summarized, "crs"), attr(nc_lazy, "crs"))
# })

test_that("distinct preserves duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  distinct_result <- nc_lazy |> dplyr::distinct(SID74, .keep_all = TRUE)
  
  expect_s3_class(distinct_result, "duckspatial_df")
  expect_equal(attr(distinct_result, "crs"), attr(nc_lazy, "crs"))
})

test_that("rename preserves duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  renamed <- nc_lazy |> dplyr::rename(county_name = NAME)
  
  expect_s3_class(renamed, "duckspatial_df")
  expect_equal(attr(renamed, "crs"), attr(nc_lazy, "crs"))
})

test_that("slice_min preserves duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  sliced <- nc_lazy |> dplyr::slice_min(AREA, n = 5)
  
  expect_s3_class(sliced, "duckspatial_df")
  expect_equal(attr(sliced, "crs"), attr(nc_lazy, "crs"))
})

test_that("head preserves duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  headed <- nc_lazy |> head(10)
  
  expect_s3_class(headed, "duckspatial_df")
  expect_equal(attr(headed, "crs"), attr(nc_lazy, "crs"))
  expect_equal(attr(headed, "sf_column"), attr(nc_lazy, "sf_column"))
})

test_that("chained dplyr operations preserve duckspatial_df class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn)
  
  result <- nc_lazy |>
    dplyr::filter(AREA > 0.1) |>
    dplyr::mutate(area_double = AREA * 2) |>
    dplyr::select(NAME, AREA, area_double, geometry) |>
    dplyr::arrange(dplyr::desc(AREA)) |>
    head(10)
  
  expect_s3_class(result, "duckspatial_df")
  expect_equal(attr(result, "crs"), attr(nc_lazy, "crs"))
  expect_equal(attr(result, "sf_column"), attr(nc_lazy, "sf_column"))
  
  collected <- dplyr::collect(result, as = "tibble")
  expect_s3_class(collected, "tbl_df")
  expect_equal(nrow(collected), 10)
})

# =============================================================================
# collect() tests
# =============================================================================

test_that("ddbs_collect works with duckspatial_df", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(geom_col = "geometry", crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn, geom_col = "geometry")
  
  result <- ddbs_collect(nc_lazy)
  
  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(nc_sf))
})

# =============================================================================
# compute() tests
# =============================================================================

test_that("compute.duckspatial_df forces execution and preserves class", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf), geom_col = "geometry")
  nc_lazy <- as_duckspatial_df("nc_test", conn, crs = sf::st_crs(nc_sf), geom_col = "geometry")
  
  computed <- dplyr::compute(nc_lazy)
  
  expect_s3_class(computed, "duckspatial_df")
  expect_s3_class(computed, "tbl_lazy")
  expect_equal(attr(computed, "crs"), attr(nc_lazy, "crs"))
  expect_equal(attr(computed, "sf_column"), attr(nc_lazy, "sf_column"))
})

test_that("compute.duckspatial_df simplifies query plan", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  ## TODO - does not pass in v1.5.1
  testthat::skip()
  nc_lazy <- 
    as_duckspatial_df("nc_test", conn) |> 
    # dplyr::tbl(conn, "nc_test") |>
    # as_duckspatial_df(crs = sf::st_crs(nc_sf), geom_col = "geometry") |>
    dplyr::filter(AREA > 0.1) |>
    dplyr::mutate(area_sq = AREA * AREA)
  
  query_before <- as.character(dbplyr::sql_render(nc_lazy))
  expect_true(grepl("AREA", query_before))
  
  computed <- dplyr::compute(nc_lazy)
  query_after <- as.character(dbplyr::sql_render(computed))
  
  expect_true(grepl("dbplyr_", query_after))
  expect_false(grepl("nc_test", query_after))
})

test_that("ddbs_compute wrapper works correctly", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf), geom_col = "geometry")
  nc_lazy <- as_duckspatial_df("nc_test", conn) 
  
  computed <- ddbs_compute(nc_lazy)
  
  expect_s3_class(computed, "duckspatial_df")
  expect_equal(attr(computed, "crs"), attr(nc_lazy, "crs"))
  
  expect_error(ddbs_compute(data.frame(x = 1)), "duckspatial_df")
})

# =============================================================================
# join tests
# =============================================================================

test_that("left_join.duckspatial_df preserves spatial attributes", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn) 
  
  
  extra_data <- data.frame(NAME = nc_sf$NAME[1:5], extra_col = 1:5)
  DBI::dbWriteTable(conn, "extra_data", extra_data)
  extra_lazy <- dplyr::tbl(conn, "extra_data")
  
  result <- dplyr::left_join(nc_lazy, extra_lazy, by = "NAME")
  
  expect_s3_class(result, "duckspatial_df")
  expect_equal(attr(result, "crs"), attr(nc_lazy, "crs"))
  expect_equal(attr(result, "sf_column"), attr(nc_lazy, "sf_column"))
})

test_that("inner_join.duckspatial_df preserves spatial attributes", {
  conn <- ddbs_temp_conn()
  ddbs_write_table(conn, nc_sf, "nc_test", quiet = TRUE)
  
  # nc_lazy <- dplyr::tbl(conn, "nc_test") |>
  #   as_duckspatial_df(crs = sf::st_crs(nc_sf))
  nc_lazy <- as_duckspatial_df("nc_test", conn) 
  
  extra_data <- data.frame(NAME = nc_sf$NAME[1:5], extra_col = 1:5)
  DBI::dbWriteTable(conn, "extra_data", extra_data)
  extra_lazy <- dplyr::tbl(conn, "extra_data")
  
  result <- dplyr::inner_join(nc_lazy, extra_lazy, by = "NAME")
  
  expect_s3_class(result, "duckspatial_df")
  expect_equal(attr(result, "crs"), attr(nc_lazy, "crs"))
  expect_equal(attr(result, "sf_column"), attr(nc_lazy, "sf_column"))
})

# =============================================================================
# Regression tests
# =============================================================================

test_that("dplyr::filter is preserved when chaining with ddbs_filter", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Test 1: filter to subset, then spatial filter
  # Brazil, Uruguay, Chile, France - only first 3 touch Argentina
  subset <- countries |>
    dplyr::filter(CNTR_ID %in% c("BR", "UY", "CL", "FR"))
  
  result <- subset |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  # Should be 3 (BR, UY, CL touch Argentina), not 4 (FR doesn't touch) 
  # and not 5 (all Argentina neighbors, which was the bug)
  expect_equal(nrow(result), 3)
  expect_setequal(result$CNTR_ID, c("BR", "UY", "CL"))
})

test_that("dplyr::select is preserved when chaining with ddbs_filter", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Select only certain columns before spatial filter
  result <- countries |>
    dplyr::select(CNTR_ID, NAME_ENGL, geom) |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  # Should only have selected columns
  expect_true("CNTR_ID" %in% names(result))
  expect_true("NAME_ENGL" %in% names(result))
  # Other original columns should NOT be present
  expect_false("ISO3_CODE" %in% names(result))
})

test_that("chained filter + select + ddbs_filter works", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  result <- countries |>
    dplyr::filter(CNTR_ID %in% c("BR", "UY", "CL", "FR", "DE")) |>
    dplyr::select(CNTR_ID, geom) |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  # Should be 3 (BR, UY, CL) and only selected columns (maybe crs_duckspatial added)
  expect_equal(nrow(result), 3)
  expect_setequal(result$CNTR_ID, c("BR", "UY", "CL"))
  expect_true("CNTR_ID" %in% names(result))
  expect_false("ISO3_CODE" %in% names(result))  # This was NOT selected
})

test_that("unmodified duckspatial_df still uses optimization", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Without any dplyr operations, source_table optimization should still work
  result <- countries |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  # All countries touching Argentina: BR, UY, PY, BO, CL
  expect_equal(nrow(result), 5)
})

test_that("mutate is preserved in ddbs_filter", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Create a dummy column and filter on it
  # If mutate is ignored, "dummy_col" won't exist or filter won't work
  result <- countries |>
    dplyr::mutate(dummy_col = 1) |>
    dplyr::filter(dummy_col == 1) |>
    # Add a filter that relies on mutate result
    dplyr::mutate(is_ar = grepl("^AR", ISO3_CODE)) |>
    dplyr::filter(is_ar) |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  # Should work just like the direct filter case (0 rows)
  expect_equal(nrow(result), 0)
})

test_that("rename is preserved in ddbs_filter", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Rename ID column, then filter using new name
  # If rename ignored, new name won't exist
  result <- countries |>
    dplyr::rename(new_id = CNTR_ID) |>
    dplyr::filter(new_id %in% c("BR", "UY", "CL")) |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
  
  expect_equal(nrow(result), 3)
  expect_true("new_id" %in% names(result))
  expect_false("CNTR_ID" %in% names(result))
})

test_that("slice/head is preserved in ddbs_filter", {
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  argentina <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", package = "duckspatial")
  )
  
  # Take top 1 country (Afghanistan usually), confirm it doesn't touch Argentina
  # If slice ignored, we get all neighbors
  result <- countries |>
    dplyr::arrange(NAME_ENGL) |>
    head(1) |>
    ddbs_filter(argentina, predicate = "touches") |>
    dplyr::collect()
    
  expect_equal(nrow(result), 0)
})

test_that("summarize before spatial op fails correctly (missing geom)", {
  # Summarize drops geometry for non-spatial summaries
  countries <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  )
  
  # Summarize to drop geometry, result is just a tibble (lazy)
  summarized <- countries |>
    dplyr::group_by(CNTR_ID) |>
    dplyr::summarize(n = dplyr::n())
    
  # Should fail because geometry column is missing in the summarized view
  # If it ignored summarize and used source_table, it would mistakenly succeed
  expect_error(
    ddbs_filter(summarized, countries),
    # "Values list .* does not have a column named .*geom"
  )
})

assert_logic <- function(arg, ref = "quiet") { # nocov start

    if (!is.logical(arg)) {
        cli::cli_abort(
            "{.arg {ref}} must be either TRUE or FALSE.",
            .frame = parent.frame()
            )
        }
} # nocov end


assert_xy <- function(xy, ref = "x") { # nocov start

    valid_types <- inherits(xy, "sf") || 
                   inherits(xy, "duckspatial_df") ||
                   inherits(xy, "tbl_sql") ||
                   inherits(xy, "tbl_lazy") ||
                   is.character(xy)
    
    if (!valid_types) {
        cli::cli_abort(
            "{.arg {ref}} must be an sf object, duckspatial_df, tbl_lazy, or a table name string.",
            .frame = parent.frame()
        )
    }
} # nocov end

assert_name <- function(name = parent.frame()$name, ref = "name") { # nocov start

    if (!is.null(name) && (!is.character(name) || length(name) != 1)) {
        cli::cli_abort(
            "{.arg {ref}} must be a single character string or NULL.",
            .frame = parent.frame()
        )
    }
} # nocov end

assert_character_scalar <- function(arg, ref) { # nocov start

    if (!is.character(arg) || length(arg) != 1) {
        cli::cli_abort(
            "{.arg {ref}} must be a single character string.",
            .frame = parent.frame()
        )
    }
} # nocov end

assert_numeric <- function(arg, ref) { # nocov start

    if (!is.numeric(arg) || length(arg) != 1) {
        cli::cli_abort(
            "{.arg {ref}} must be a single numeric value.",
            .frame = parent.frame()
        )
    }
} # nocov end

assert_strict_positive_numeric <- function(arg, ref) { # nocov start

    if (!is.numeric(arg) || length(arg) != 1 || arg <= 0) {
        cli::cli_abort(
            "{.arg {ref}} must be a single positive numeric value (> 0).",
            .frame = parent.frame()
        )
    }
} # nocov end


assert_positive_numeric <- function(arg, ref) { # nocov start

    if (!is.numeric(arg) || length(arg) != 1 || arg < 0) {
        cli::cli_abort(
            "{.arg {ref}} must be a single positive numeric value (>= 0).",
            .frame = parent.frame()
        )
    }
} # nocov end


assert_integer_scalar <- function(arg, ref) { # nocov start

    if (!is.numeric(arg) || length(arg) != 1 || arg != as.integer(arg)) {
        cli::cli_abort(
            "{.arg {ref}} must be a single integer value.",
            .frame = parent.frame()
        )
    }
} # nocov end

assert_numeric_interval <- function(arg, minn, maxx, ref) { # nocov start

    if (!is.numeric(arg) || min(arg) < minn || max(arg) > maxx) {
        cli::cli_abort(
            "{.arg {ref}} must be a single numeric value between {minn} and {maxx}.",
            .frame = parent.frame()
        )
    }
} # nocov end


# whether the function takes sf of string as xy input
assert_connflict <- function(conn, xy, ref = "x") { # nocov start

    is_duckdn_conn <- dbConnCheck(conn)


    if (inherits(xy, "sf") && is_duckdn_conn) {
        cli::cli_abort(
            "If you pass a {.arg {ref}}, it must be a string when {.arg conn} is provided.",
            .frame = parent.frame()
        )
    }
} # nocov end


assert_geometry_column <- function(geom, name_list) { # nocov start
    if (length(geom) == 0) cli::cli_abort("Geometry column wasn't found in table <{name_list$query_name}>.")
} # nocov end



## assert crs_column (needed for ddbs_filter and ddbs_join)
## cols can be a character vector or a collapsed string
assert_crs_column <- function(crs_column, cols) { # nocov start
    if (!is.null(crs_column)) {
        # Handle both vector and string inputs
        found <- crs_column %in% cols | any(grepl(crs_column, cols, fixed = TRUE))
        if (!found)
            cli::cli_abort("CRS column <{crs_column}> do not found in the table. If the data do not have CRS column, set the argument `crs_column = NULL`")
    }
} # nocov end


## assert id argument in predicate functions
assert_predicate_id <- function(id, conn, lst) { # nocov start
    if (!is.null(id)) {
        x_rest <- get_geom_name(conn, lst, rest = TRUE, collapse = FALSE)
        if (!id %in% x_rest) cli::cli_abort("<id> must be NULL or a column name of <x>")
    }
} # nocov end


## assert if the CRS of `x` and `y` is the same
assert_crs <- function(conn, x, y) { # nocov start

  ## get CRS using character method
  crs_x <- duckspatial::ddbs_crs(x, conn = conn)
  crs_y <- duckspatial::ddbs_crs(y, conn = conn)

  ## abort if CRS is different (use crs_equal for proper comparison)
  if (!crs_equal(crs_x, crs_y)) {
    cli::cli_abort("The Coordinates Reference System of `x` and `y` is different.")
  }

} # nocov end




## assert that if `x` or `y` are character, connection cannot be NULL
assert_conn_character <- function(conn, ...) { # nocov start

  ## get inputs into a list
  inputs <- list(...)

  ## check if they are character
  inputs_check <- lapply(inputs, is.character) |> as.logical()

  if (any(inputs_check) & is.null(conn)) {
    cli::cli_abort("If a table name is provided in `x` or `y`, the connection cannot be NULL.")
  }

} # nocov end

#' Assert that columns exist in a table
#'
#' @param conn DuckDB connection
#' @param table Table name
#' @param cols Vector of column names to check
#' @param ref Reference name for error message
#' @noRd
#' @returns invisible(TRUE)
assert_col_exists <- function(conn, table, cols, ref) { # nocov start

    avail <- DBI::dbGetQuery(conn, glue::glue("DESCRIBE {table}"))$column_name
    missing <- setdiff(cols, avail)
    if (length(missing) > 0) {
        # Rephrased to put the vector first, helping cli resolve pluralization quantity
        cli::cli_abort(
            "{.val {missing}} column{?s} not found in {.arg {ref}} table."
        )
    }
    invisible(TRUE)
} # nocov end

assert_threads <- function(threads, ref = "threads") { # nocov start
  if (!is.null(threads)) {
    valid <- is.numeric(threads) && length(threads) == 1 && 
             threads > 0 && as.integer(threads) == threads
    
    if (!valid) {
      cli::cli_abort("{.arg {ref}} must be a positive integer or NULL.")
    }
  }
} # nocov end

assert_memory_limit_gb <- function(memory_limit, ref = "memory_limit_gb") { # nocov start
  if (!is.null(memory_limit)) {
    valid <- is.numeric(memory_limit) && length(memory_limit) == 1 && memory_limit > 0
    
    if (!valid) {
      cli::cli_abort("{.arg {ref}} must be a positive number (GB) or NULL.")
    }
  }
} # nocov end


assert_geom_type <- function(x, conn, geom, multi = FALSE) { # nocov start
  
  ## add multi if required
  if (isTRUE(multi)) {
    geom <- c(geom, paste0("MULTI", geom))
  }
  
  ## get the unique geometry types
  geom_type <- ddbs_geometry_type(x, conn, by_feature = FALSE)
  
  ## check geometry type
  invalid_types <- setdiff(geom_type, geom)
  
  if (length(invalid_types) > 0) {
    cli::cli_abort(c(
      "Invalid geometry type{?s} found: {.val {invalid_types}}",
      "i" = "Allowed type{?s}: {.val {geom}}"
    ))
  }
  
} # nocov end


assert_conn_x_name <- function(conn, x, name) { # nocov start

  if (!is.null(conn) && is.null(name) && !is.character(x)) {
    cli::cli_abort("If {.arg conn} is not NULL, {.arg x} or {.arg name} must be a table name.")
  }

  if (!is.null(name) && is.null(conn)) {
    cli::cli_abort("If {.arg name} is not NULL, {.arg conn} must be a valid connection.")
  }

} # nocov end

#' Evaluate spatial predicates between geometries
#'
#' Determines which geometries in one dataset satisfy a specified spatial 
#' relationship with geometries in another dataset, such as intersection, 
#' containment, or touching.
#'
#' @template x
#' @template y
#' @template predicate
#' @template conn_null
#' @template conn_x_conn_y
#' @template name
#' @template predicate_args
#' @param distance a numeric value specifying the distance for ST_DWithin. Units correspond to
#' the coordinate system of the geometry (e.g. degrees or meters)
#' @template mode
#' @template overwrite
#' @template quiet
#' @param ... Passed to [ddbs_predicate]
#'
#' @details
#'
#' This function provides a unified interface to all spatial predicate operations
#' in DuckDB's spatial extension. It performs pairwise comparisons between all
#' geometries in `x` and `y` using the specified predicate.
#'
#' ## Available Predicates
#'
#' - **intersects**: Geometries share at least one point
#' - **covers**: Geometry `x` completely covers geometry `y`
#' - **touches**: Geometries share a boundary but interiors do not intersect
#' - **disjoint**: Geometries have no points in common
#' - **within**: Geometry `x` is completely inside geometry `y`
#' - **dwithin**: Geometry `x` is completely within a distance of geometry `y`
#' - **contains**: Geometry `x` completely contains geometry `y`
#' - **overlaps**: Geometries share some but not all points
#' - **crosses**: Geometries have some interior points in common
#' - **equals**: Geometries are spatially equal
#' - **covered_by**: Geometry `x` is completely covered by geometry `y`
#' - **intersects_extent**: Bounding boxes of geometries intersect (faster but less precise)
#' - **contains_properly**: Geometry `x` contains geometry `y` without boundary contact
#' - **within_properly**: Geometry `x` is within geometry `y` without boundary contact
#'
#' If `x` or `y` are not DuckDB tables, they are automatically copied into a
#' temporary in-memory DuckDB database (unless a connection is supplied via `conn`).
#'
#' `id_x` or `id_y` may be used to replace the default integer indices with the
#' values of an identifier column in `x` or `y`, respectively.
#'
#' @returns Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
#' \itemize{
#'   \item \code{duckspatial} (default): A \code{tbl_duckdb_connection} (lazy data frame) backed by dbplyr/DuckDB.
#'   \item \code{sf}: An eagerly collected list.
#' }
#' When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
#' 
#'
#' @examples
#' \dontrun{
#' ## Load packages
#' library(duckspatial)
#' library(dplyr)
#' 
#' ## create in-memory DuckDB database
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ## read countries data, and rivers
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |>
#'   filter(CNTR_ID %in% c("PT", "ES", "FR", "IT"))
#' 
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson", 
#'   package = "duckspatial")
#' ) |>
#'   ddbs_transform(ddbs_crs(countries_ddbs))
#' 
#' ## Store in DuckDB
#' ddbs_write_vector(conn, countries_ddbs, "countries")
#' ddbs_write_vector(conn, rivers_ddbs, "rivers")
#' 
#' ## Example 1: Check which rivers intersect each country
#' ddbs_predicate(countries_ddbs, rivers_ddbs, predicate = "intersects")
#' ddbs_intersects(countries_ddbs, rivers_ddbs)
#' 
#' ## Example 2: Find neighboring countries
#' ddbs_predicate(
#'   countries_ddbs, 
#'   countries_ddbs, 
#'   predicate = "touches",
#'   id_x = "NAME_ENGL", 
#'   id_y = "NAME_ENGL"
#' )
#' 
#' ddbs_touches(
#'   countries_ddbs, 
#'   countries_ddbs, 
#'   id_x = "NAME_ENGL", 
#'   id_y = "NAME_ENGL"
#' )
#' 
#' ## Example 3: Find rivers that don't intersect countries
#' ddbs_predicate(
#'   countries_ddbs, 
#'   rivers_ddbs, 
#'   predicate = "disjoint",
#'   id_x = "NAME_ENGL", 
#'   id_y = "RIVER_NAME"
#' )
#' 
#' ## Example 4: Use table names inside duckdb
#' ddbs_predicate("countries", "rivers", predicate = "within", conn, id_x = "NAME_ENGL")
#' ddbs_within("countries", "rivers", conn,  id_x = "NAME_ENGL")
#' }
#' @name ddbs_predicate
#' @rdname ddbs_predicate
NULL



#' @rdname ddbs_predicate
#' @export
ddbs_predicate <- function(
  x,
  y,
  predicate = "intersects",
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  id_x = NULL,
  id_y = NULL,
  sparse = TRUE,
  distance = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = TRUE) {

  
  ## 0. Handle errors
  assert_xy(x, "x")
  assert_xy(y, "y")
  assert_name(id_x, "id_x")
  assert_name(id_y, "id_y")
  assert_logic(sparse, "sparse")
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")

  ## Validate predicate early (it aborts on invalid)
  st_predicate <- get_st_predicate(predicate)


  # 1. Manage connection to DB

  ## 1.1. Pre-extract attributes (CRS and geometry column name)
  ## this step should be before normalize_spatial_input()
  crs_x <- if (is.null(conn_x)) ddbs_crs(x, conn) else ddbs_crs(x, conn_x)
  crs_y <- if (is.null(conn_y)) ddbs_crs(y, conn) else ddbs_crs(y, conn_y)
  sf_col_x <- attr(x, "sf_column")
  sf_col_y <- attr(y, "sf_column")

  ## 1.2. Resolve conn_x/conn_y defaults from 'conn' for character inputs
  if (is.null(conn_x) && !is.null(conn) && is.character(x)) conn_x <- conn
  if (is.null(conn_y) && !is.null(conn) && is.character(y)) conn_y <- conn

  ## 1.3. Normalize inputs: coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names
  x <- normalize_spatial_input(x, conn_x)
  y <- normalize_spatial_input(y, conn_y)

  ## 1.4. Get mode - If it's NULL, it will use the duckspatial.mode option
  mode <- get_mode(mode, name)


  # 2. Manage connection to DB

  ## 2.1. Resolve connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y, conn, conn_x, conn_y, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  y            <- resolve_conn$y
  ## register cleanup of the connection
  if (any(is.null(conn_x), is.null(conn_y))) {
      on.exit(resolve_conn$cleanup(), add = TRUE)   
  }

  ## 2.2. Get query list of table names
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)
  y_list <- get_query_list(y, target_conn)
  on.exit(y_list$cleanup(), add = TRUE)

  ## check if id column name exists in x or y
  assert_predicate_id(id_x, target_conn, x_list$query_name)
  assert_predicate_id(id_y, target_conn, y_list$query_name)

  ## CRS already extracted at start of function
  if (!is.null(crs_x) && !is.null(crs_y)) {
      if (!crs_equal(crs_x, crs_y)) {
        cli::cli_abort("The Coordinates Reference System of {.arg x} and {.arg y} is different.")
      }
  } else {
      assert_crs(target_conn, x_list$query_name, y_list$query_name)
  }


  # 3. Prepare parameters for the query

  ## 3.1. Get names of geometry columns (use saved sf_col_x from before transformation)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  y_geom <- sf_col_y %||% get_geom_name(target_conn, y_list$query_name)
  assert_geometry_column(x_geom, x_list)
  assert_geometry_column(y_geom, y_list)
  
  ## 3.2. Build predicate expression
  if (st_predicate == "ST_DWithin") {

    ## Warn if the distance arg wasn't specified, and give it a value of 0
    if (is.null(distance)) {
      cli::cli_warn("{.val distance} wasn't specified. Using ST_Within.")
      distance <- 0
    }
    
    ## check the CRS units to use the right function
    crs_units <- crs_x$units_gdal
    if (crs_units != "metre") {
      # predicate_expr <- glue::glue("ST_DWithin_Spheroid(x.{x_geom}, y.{y_geom}, {distance})")
      # predicate_expr <- glue::glue("ST_DWithin_Spheroid(ST_FlipCoordinates(x.{x_geom}), ST_FlipCoordinates(y.{y_geom}), {distance})")
        predicate_expr <- glue::glue(
        "ST_DWithin_Spheroid(
          ST_Point(ST_Y(x.{x_geom}), ST_X(x.{x_geom})),
          ST_Point(ST_Y(y.{y_geom}), ST_X(y.{y_geom})),
          {distance}
        )"
      )
      if (crs_x$input != "EPSG:4326") {
        cli::cli_warn(
          "Inputs are in {.val {crs_x$input}}, not {.val EPSG:4326}. Distance calculations may be less accurate. Consider transforming to {.val EPSG:4326} or a projected CRS."
        )
      }
    } else {
      predicate_expr <- glue::glue("ST_DWithin(x.{x_geom}, y.{y_geom}, {distance})")
    }
    
    
  } else {
    predicate_expr <- glue::glue("{st_predicate}(x.{x_geom}, y.{y_geom})")
  }


  # 4. Build query and return based on mode
  ## - mode sf: it will return a list-like object
  ## - mode duckspatial: it will return a lazy-tbl object
  if (mode == "sf") {
    
    ## materialize full predicate matrix and reframe as sf/sparse when required
    tmp.query <- glue::glue("
      SELECT {predicate_expr} AS predicate
      FROM {x_list$query_name} x
      CROSS JOIN {y_list$query_name} y
    ")
    
    data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)
    
    result <- reframe_predicate_data(
      conn   = target_conn,
      data   = data_tbl,
      x_list = x_list,
      y_list = y_list,
      id_x   = id_x,
      id_y   = id_y,
      sparse = sparse
    )
    
  } else if (mode == "duckspatial") {
    
    ## Resolve identifiers
    x_id_expr <- if (is.null(id_x)) "row_number() OVER () AS id_x" else glue::glue("{id_x} AS id_x")
    y_id_expr <- if (is.null(id_y)) "row_number() OVER () AS id_y" else glue::glue("{id_y} AS id_y")

    ## Name for the table to be created
    view_name <- ddbs_temp_table_name()

     if (sparse) {
    
      ## long format - only TRUE pairs
      tmp.query <- glue::glue("
        CREATE TEMP TABLE {view_name} AS
        SELECT 
          x.id_x,
          y.id_y
        FROM (SELECT {x_id_expr}, * FROM {x_list$query_name}) x
        CROSS JOIN (SELECT {y_id_expr}, * FROM {y_list$query_name}) y
        WHERE {predicate_expr}
      ")
      
    } else {
      
      ## Wide format - all pairs with TRUE/FALSE
      ## need to fetch y_ids eagerly to build pivot columns
      y_ids <- DBI::dbGetQuery(
        target_conn,
        glue::glue("SELECT {y_id_expr} FROM {y_list$query_name}")
      )$id_y
      
      pivot_list <- paste(
        glue::glue("SUM(CASE WHEN id_y = '{y_ids}' AND predicate THEN 1 ELSE 0 END)::BOOLEAN AS \"{y_ids}\""),
        collapse = ",\n"
      )
      
      ## Generate the query
      tmp.query <- glue::glue("
        CREATE TEMP TABLE {view_name} AS
        WITH long AS (
          SELECT 
            x.id_x,
            y.id_y,
            {predicate_expr} AS predicate
          FROM (SELECT {x_id_expr}, * FROM {x_list$query_name}) x
          CROSS JOIN (SELECT {y_id_expr}, * FROM {y_list$query_name}) y
        )
        SELECT 
          id_x,
          {pivot_list}
        FROM long
        GROUP BY id_x
        ORDER BY id_x
      ")
      
     }
    
    ## Create a table, and return a pointer to that table
    DBI::dbExecute(target_conn, tmp.query)
    result <- dplyr::tbl(target_conn, view_name)
    
  }

  return(result)

}




#' @rdname ddbs_predicate
#' @export
ddbs_intersects <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "intersects", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_covers <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "covers", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_touches <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "touches", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_is_within_distance <- function(x, y, distance = NULL, ...) {
  ddbs_predicate(x = x, y = y, predicate = "dwithin", distance = distance, ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_disjoint <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "disjoint", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_within <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "within", ...)
}

#' @rdname ddbs_predicate
ddbs_contains <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "contains", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_overlaps <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "overlaps", ...)
}

#' @rdname ddbs_predicate
ddbs_crosses <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "crosses", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_equals <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "equals", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_covered_by <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "covered_by", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_intersects_extent <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "intersects_extent", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_contains_properly <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "contains_properly", ...)
}

#' @rdname ddbs_predicate
#' @export
ddbs_within_properly <- function(x, y, ...) {
  ddbs_predicate(x = x, y = y, predicate = "within_properly", ...)
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()
conn_test_2 <- duckspatial::ddbs_create_conn()

## write data in the database
ddbs_write_table(conn_test, points_sf, "points")
ddbs_write_table(conn_test, argentina_ddbs, "argentina")
ddbs_write_table(conn_test_2, argentina_ddbs, "argentina")


# 1. ddbs_predicate ------------------------------------------------------

## 1.1. Expected behaviour ----------

## expected behaviour
## - CHECK 1.1: combination of sf/ddbs/duckdb table work
## - CHECK 1.2: all predicates work
## - CHECK 1.3: conn_x and conn_y work
## - CHECK 1.4: sparse returns a matrix
## - CHECK 1.5: returns same as SF
## - CHECK 1.6: id_x and id_y work
## - CHECK 2.1: Combination of inputs / missing arguments
## - CHECK 2.2: other errors
describe("ddbs_predicate()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all format combinations (sf, duckspatial_df, DuckDB table)", {
      output_sf_sf       <- ddbs_predicate(points_sf, argentina_sf)
      output_ddbs_sf     <- ddbs_predicate(points_ddbs, argentina_sf) |> collect()
      output_sf_ddbs     <- ddbs_predicate(points_sf, argentina_ddbs) |> collect()
      output_ddbs_ddbs   <- ddbs_predicate(points_ddbs, argentina_ddbs) |> collect()
      output_conn_sf     <- ddbs_predicate("points", argentina_sf, conn = conn_test) |> collect()
      output_sf_conn     <- ddbs_predicate(points_sf, "argentina", conn = conn_test) |> collect()
      output_conn_conn   <- ddbs_predicate("points", "argentina", conn = conn_test) |> collect()
      
      expect_s3_class(output_sf_sf, "tbl_duckdb_connection")
      expect_equal(collect(output_sf_sf), output_ddbs_sf)
      expect_equal(output_ddbs_sf, output_sf_ddbs)
      expect_equal(output_ddbs_sf, output_ddbs_ddbs)
      expect_equal(output_ddbs_sf, output_conn_sf)
      expect_equal(output_ddbs_sf, output_sf_conn)
      expect_equal(output_ddbs_sf, output_conn_conn)
    })
    
    it("warns when mixing DuckDB table with duckspatial_df from different connections", {
      expect_warning(ddbs_predicate("points", argentina_ddbs, conn = conn_test))
      expect_warning(ddbs_predicate(points_ddbs, "argentina", conn = conn_test))
    })
    
    it("works with intersects predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "intersects") |> collect()
      output_function  <- ddbs_intersects(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with covers predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "covers") |> collect()
      output_function  <- ddbs_covers(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with touches predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "touches") |> collect()
      output_function  <- ddbs_touches(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with disjoint predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "disjoint") |> collect()
      output_function  <- ddbs_disjoint(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with within predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "within") |> collect()
      output_function  <- ddbs_within(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with contains predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "contains") |> collect()
      output_function  <- ddbs_contains(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with overlaps predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "overlaps") |> collect()
      output_function  <- ddbs_overlaps(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with covered_by predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "covered_by") |> collect()
      output_function  <- ddbs_covered_by(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with intersects_extent predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "intersects_extent") |> collect()
      output_function  <- ddbs_intersects_extent(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with contains_properly predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "contains_properly") |> collect()
      output_function  <- ddbs_contains_properly(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with within_properly predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "within_properly") |> collect()
      output_function  <- ddbs_within_properly(countries_sf, argentina_sf) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("works with dwithin predicate", {
      point_sf <- ddbs_collect(points_ddbs)[1, ]
      output_predicate <- ddbs_predicate(point_sf, points_ddbs, predicate = "dwithin", distance = 100) |> collect()
      output_function  <- ddbs_is_within_distance(point_sf, points_ddbs, distance = 100) |> collect()
      
      expect_equal(output_predicate, output_function)
    })
    
    it("supports conn_x and conn_y for different connections", {
      expect_warning(ddbs_predicate("points", "argentina", conn_x = conn_test, conn_y = conn_test_2))
      
      output_different_conn <- suppressWarnings(ddbs_predicate("points", "argentina", conn_x = conn_test, conn_y = conn_test_2)) |> collect()
      output_same_result    <- ddbs_predicate(points_sf, argentina_sf) |> collect()
      
      expect_equal(output_different_conn, output_same_result)
    })

    it("returns a wide table when sparse = FALSE and mode = 'duckspatial'", {
      output_sparse <- ddbs_predicate(points_ddbs, argentina_ddbs, sparse = FALSE)

      expect_equal(
        nrow(collect(output_sparse)),
        nrow(ddbs_collect(points_ddbs))
      )

      expect_equal(
        ncol(collect(output_sparse)) - 1, #remove x_id
        nrow(ddbs_collect(argentina_ddbs))
      )

      expect_equal(
        names(collect(output_sparse)),
        c("id_x", "1")
      )
      
    })

    it("returns a long table when sparse = FALSE and mode = 'duckspatial'", {
      output_sparse <- ddbs_predicate(points_ddbs, argentina_ddbs)

      expect_equal(ncol(collect(output_sparse)), 2)

      expect_equal(
        names(collect(output_sparse)),
        c("id_x", "id_y")
      )
      
    })
    
    it("returns matrix when sparse = FALSE and mode = 'sf'", {
      output_sparse <- ddbs_predicate(points_ddbs, argentina_ddbs, sparse = FALSE, mode = "sf")
      expect_true(inherits(output_sparse, "matrix"))
    })
    
    it("returns same results as sf when sparse = FALSE for covers", {
      output_ddbs <- ddbs_covers(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_covers(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for touches", {
      output_ddbs <- ddbs_touches(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_touches(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for disjoint", {
      output_ddbs <- ddbs_disjoint(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_disjoint(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for within", {
      output_ddbs <- ddbs_within(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_within(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for contains", {
      output_ddbs <- ddbs_contains(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_contains(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for overlaps", {
      output_ddbs <- ddbs_overlaps(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_overlaps(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for covered_by", {
      output_ddbs <- ddbs_covered_by(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_covered_by(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for intersects_extent", {
      output_ddbs <- ddbs_intersects_extent(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_intersects(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("returns same results as sf when sparse = FALSE for contains_properly", {
      output_ddbs <- ddbs_contains_properly(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_contains_properly(countries_sf, argentina_sf, sparse = FALSE)
      
      expect_equal(output_ddbs, output_sf)
    })
    
    it("supports id_x parameter to name output list elements", {
      output <- ddbs_predicate(countries_sf, argentina_sf, "touches", id_x = "CNTR_ID", mode = "sf")
      
      expect_equal(names(output), countries_sf$CNTR_ID)
      expect_equal(output[[2]], 1)
    })
    
    it("supports id_y parameter to use custom IDs in results", {
      output <- ddbs_predicate(countries_sf, argentina_sf, "touches", id_y = "CNTR_ID", mode = "sf")
      
      expect_null(names(output))
      expect_equal(output[[2]], "AR")
    })
    
    it("supports both id_x and id_y parameters together", {
      output <- ddbs_predicate(countries_sf, argentina_sf, "touches", id_x = "CNTR_ID", id_y = "CNTR_ID", mode = "sf")

      expect_equal(names(output), countries_sf$CNTR_ID)
      expect_equal(output[[2]], "AR")
    })

    it("works with equals predicate", {
      output_predicate <- ddbs_predicate(countries_sf, argentina_sf, predicate = "equals") |> collect()
      output_function  <- ddbs_equals(countries_sf, argentina_sf) |> collect()

      expect_equal(output_predicate, output_function)
    })

    it("returns same results as sf when sparse = FALSE for equals", {
      output_ddbs <- ddbs_equals(countries_sf, argentina_sf, sparse = FALSE, mode = "sf")
      output_sf   <- sf::st_equals(countries_sf, argentina_sf, sparse = FALSE)

      expect_equal(output_ddbs, output_sf)
    })

    it("correctly identifies identical geometries", {
      result_ddbs <- ddbs_equals(argentina_sf, argentina_sf, sparse = FALSE, mode = "sf")
      result_sf   <- sf::st_equals(argentina_sf, argentina_sf, sparse = FALSE)

      expect_equal(result_ddbs, result_sf)
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {

    it("requires x and y to be points", {
      expect_error(
        ddbs_predicate(points_sf, argentina_ddbs, predicate = "dwithin", distance = 100)
      )
      expect_error(
        ddbs_predicate(points_sf, ddbs_transform(rivers_ddbs, 4326), predicate = "dwithin", distance = 100)
      )
    })
    
    it("requires both x and y arguments", {
      expect_error(ddbs_predicate(argentina_ddbs))
      expect_error(ddbs_predicate(y = argentina_ddbs))
    })
    
    it("requires connection when using table names", {
      expect_error(ddbs_predicate("argentina", conn = NULL))
    })
    
    it("validates predicate argument", {
      expect_error(ddbs_predicate(argentina_ddbs, points_sf, predicate = "intersect_this"))
    })
    
    it("validates sparse argument type", {
      expect_error(ddbs_predicate(argentina_ddbs, points_sf, sparse = "TRUE"))
    })
    
    it("validates distance argument type for dwithin", {
      expect_error(ddbs_is_within_distance(argentina_ddbs, distance = "many kilometers"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_predicate(x = 999))
      expect_error(ddbs_predicate(x = "999", points_sf, conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_predicate(argentina_ddbs, points_sf, conn = 999))
    })

    it("validates conn_x argument type", {
      expect_error(ddbs_predicate(argentina_ddbs, "points", conn_y = 999))
    })

    it("validates conn_y argument type", {
      expect_error(ddbs_predicate("argentina", points_sf, conn_x = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_predicate(argentina_ddbs, points_sf, overwrite = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_predicate(argentina_ddbs, points_sf, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})

## stop connection
ddbs_stop_conn(conn_test)

#' Reads a vectorial table from DuckDB into R
#'
#' Retrieves the data from a DuckDB table, view, or Arrow view with a geometry
#' column, and converts it to an R \code{sf} object. This function works with
#' both persistent tables created by \code{ddbs_write_table} and temporary
#' Arrow views created by \code{ddbs_register_table}.
#'
#' @template conn
#' @template name
#' @param clauses character, additional SQL code to modify the query from the
#' table (e.g. "WHERE ...", "ORDER BY...")
#' @template quiet
#'
#' @returns an sf object
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(sf)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## create random points
#' random_points <- data.frame(
#'   id = 1:5,
#'   x = runif(5, min = -180, max = 180),
#'   y = runif(5, min = -90, max = 90)
#' )
#'
#' ## convert to sf
#' sf_points <- st_as_sf(random_points, coords = c("x", "y"), crs = 4326)
#'
#' ## Example 1: Write and read persistent table
#' ddbs_write_vector(conn, sf_points, "points")
#' ddbs_read_table(conn, "points")
#'
#' ## Example 2: Register and read Arrow view (faster, temporary)
#' ddbs_register_vector(conn, sf_points, "points_view")
#' ddbs_read_table(conn, "points_view")
#'
#' ## disconnect from db
#' ddbs_stop_conn(conn)
#' }
ddbs_read_table <- function(
    conn,
    name,
    clauses = NULL,
    quiet = FALSE) {
    
    # 0. Handle errors
    dbConnCheck(conn)
    assert_name(name)
    assert_name(clauses, "clauses")
    assert_logic(quiet, "quiet")
    

    # 1. Checks
    ## convenient names of table and/or schema.table
    name_list <- get_query_name(name)

    ## Check if table/view name exists in regular tables or Arrow views
    table_exists <- name_list$table_name %in% DBI::dbListTables(conn)
    object_type <- NULL

    if (table_exists) {
        # Determine if it's a table or view
        tables_df <- ddbs_list_tables(conn)
        db_tables <- paste0(tables_df$table_schema, ".", tables_df$table_name) |>
            sub(pattern = "^main\\.", replacement = "")
        match_idx <- which(db_tables == name_list$query_name)[1]
        if (!is.na(match_idx)) {
            table_type <- tables_df$table_type[match_idx]
            object_type <- if (!is.na(table_type) && identical(table_type, "VIEW")) {
                "view"
            } else {
                "table"
            }
        } else {
            object_type <- "table"
        }
    } else {
        # Check if it exists as an Arrow view
        arrow_views <- try(
            duckdb::duckdb_list_arrow(conn),
            silent = TRUE
        )
        arrow_exists <- if (inherits(arrow_views, "try-error")) {
            FALSE
        } else {
            name_list$query_name %in% arrow_views
        }

        if (!arrow_exists) {
            cli::cli_abort("The provided name is not present in the database as a table, view, or Arrow view.")
        } else {
            object_type <- "Arrow view"
        }
    }

    ## get column names and prepare SQL
    if (object_type == "Arrow view") {
        # For Arrow views, PRAGMA table_info doesn't work, so we need to get columns differently
        all_cols <- DBI::dbListFields(conn, name_list$query_name)

        # Dynamically identify geometry column (heuristic: look for standard names)
        candidates <- c("geometry", "geom", "shape", "wkb_geometry")
        geom_name <- intersect(all_cols, candidates)[1]

        # Fallback if standard name not found: find column that's not crs_duckspatial
        # The geometry column is added before crs_duckspatial, so it should be the
        # last column before crs_duckspatial (or last column if excluding crs_duckspatial)
        if (is.na(geom_name)) {
            if (length(all_cols) > 0) {
                # Take the LAST non-CRS column (geometry is added last during registration)
                geom_name <- all_cols[length(all_cols)]
            }
        }

        if (is.na(geom_name) || !geom_name %in% all_cols) {
            cli::cli_abort("Geometry column wasn't found in Arrow view <{name_list$query_name}>.")
        }

        no_geom_cols <- setdiff(all_cols, geom_name) 
        no_geom_cols <-  if (length(no_geom_cols) > 0) paste0('"', no_geom_cols, '",', collapse = ' ') else ""

        # For Arrow views: Try ST_AsWKB directly first (geoarrow may already be recognized as GEOMETRY)
        # If that fails, ST_GeomFromWKB will be needed, but geoarrow registration makes it GEOMETRY type
        select_geom_sql <- glue::glue("ST_AsWKB({geom_name}) AS {geom_name}")
    } else {
        # For regular tables and views, use get_geom_name
        geom_name    <- get_geom_name(conn, name_list$query_name)
        no_geom_cols <- get_geom_name(conn, name_list$query_name, rest = TRUE, collapse = TRUE)
        if (length(geom_name) == 0) cli::cli_abort("Geometry column wasn't found in table <{name_list$query_name}>.")

        # For regular tables: already GEOMETRY type
        select_geom_sql <- glue::glue("ST_AsWKB({geom_name}) AS {geom_name}")
    }

    # 2. Retrieve data
    ## Retrieve data as data frame
    tmp.query <- glue::glue(
            "SELECT
            {no_geom_cols}
            {select_geom_sql}
            FROM {name_list$query_name}"
    )
    tmp.query <- paste(tmp.query, clauses)
    data_tbl <- DBI::dbGetQuery(conn, tmp.query)
  
    ## Get the CRS
    crs <- get_table_crs(
        conn = conn,
        geom_name = geom_name,
        table_name = name_list$query_name
    )

    ## 5. convert to SF
    data_sf <- convert_to_sf_wkb(
        data       = data_tbl,
        crs        = crs,
        x_geom     = geom_name
    )

    ## return result
    if (isFALSE(quiet)) {
        cli::cli_alert_success("{object_type} {name} successfully imported.")
    }
    return(data_sf)

}





#' Load spatial vector data from DuckDB into R
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `ddbs_read_vector()` was renamed to \code{\link{ddbs_read_table}}.
#'
#' @inheritParams ddbs_read_table
#' @returns an sf object
#' @export
#' @keywords internal
ddbs_read_vector <- function(
    conn,
    name,
    clauses = NULL,
    quiet = FALSE) {
    
    lifecycle::deprecate_soft(
        when    = "1.0.0",
        what    = "ddbs_read_vector()",
        with    = "ddbs_read_table()"
    )
    
    ddbs_read_table(
        conn       = conn,
        name       = name,
        clauses    = clauses,
        quiet      = quiet
    )
}




#' Read metadata from a spatial file
#'
#' Retrieves file-level metadata from a spatial vector file (e.g. GeoPackage,
#' Shapefile, GeoJSON) using DuckDB's \code{ST_Read_Meta()} function. Returns
#' information about the file's driver and its layers as a tibble.
#'
#' @param path character, path to the spatial file to inspect.
#' @template conn
#'
#' @returns A \code{tibble} with one row per file and the
#'   following columns:
#'   \describe{
#'     \item{file_name}{Path to the file.}
#'     \item{driver_short_name}{Short name of the GDAL driver (e.g.
#'       \code{"GPKG"}).}
#'     \item{driver_long_name}{Full name of the GDAL driver (e.g.
#'       \code{"GeoPackage"}).}
#'     \item{layers}{A list-column of data frames, one per file, each
#'       describing the layers contained in the file. Unnest with
#'       \code{\link[tidyr]{unnest}} to access individual layer attributes
#'       such as name, geometry type, and feature count.}
#'   }
#' @export
#'
#' @examples
#' \dontrun{
#' ## Read metadata from a GeoPackage
#' meta <- ddbs_read_meta(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#' 
#' ## View file-level metadata
#' meta
#'
#' ## Inspect layer details
#' tidyr::unnest(meta, layers)
#' }
ddbs_read_meta <- function(path, conn = NULL) {

  if (is.null(conn)) {
    target_conn <- ddbs_default_conn()
  } else {
    target_conn <- conn
  }

  DBI::dbGetQuery(target_conn, glue::glue(
    "SELECT * FROM ST_Read_Meta('{path}')"
  )) |> 
    tibble::as_tibble()

}

# skip tests on CRAN
skip_on_cran()
skip_if_not_installed("duckdb")
skip_if_not_installed("sf")

# Setup: connection and data
conn_test <- duckspatial::ddbs_create_conn()

# Write some data to test against
ddbs_write_table(conn_test, points_sf, "test_points_read", overwrite = TRUE)
ddbs_write_table(conn_test, countries_sf, "test_countries_read", overwrite = TRUE)
ddbs_register_table(conn_test, points_sf, "test_points_view_read", overwrite = TRUE)

# 1. Basic reading from tables and views ----
test_that("ddbs_read_table reads a table into an sf object", {
  # Read from a table created with ddbs_write_table
  result <- ddbs_read_table(conn_test, "test_points_read")

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(points_sf))
  expect_equal(sf::st_crs(result), sf::st_crs(points_sf))
  expect_true(all(c("id", "geometry") %in% names(result)))
})

test_that("ddbs_read_table reads an Arrow view into an sf object", {
  # Read from a view created with ddbs_register_table
  result <- ddbs_read_table(conn_test, "test_points_view_read")

  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(points_sf))
  expect_equal(sf::st_crs(result), sf::st_crs(points_sf))
})

test_that("ddbs_read_table reads a standard SQL view into an sf object", {
  # Create a standard SQL view from an existing table
  # First ensure the base table exists
  ddbs_write_table(conn_test, points_sf, "base_table_for_view", overwrite = TRUE)
  
  # Create SQL view
  DBI::dbExecute(conn_test, "CREATE OR REPLACE VIEW test_sql_view AS SELECT * FROM base_table_for_view")
  
  # Read from the SQL view
  result <- ddbs_read_table(conn_test, "test_sql_view")
  
  expect_s3_class(result, "sf")
  expect_equal(nrow(result), nrow(points_sf))
  expect_equal(sf::st_crs(result), sf::st_crs(points_sf))
  
  # Cleanup
  DBI::dbExecute(conn_test, "DROP VIEW IF EXISTS test_sql_view")
})

# 2. Parameters and clauses ----
test_that("ddbs_read_table's 'clauses' argument works", {
  # Use a WHERE clause to filter the data
  result <- ddbs_read_table(conn_test, "test_points_read", clauses = "WHERE id <= 3")

  expect_equal(nrow(result), 3)
  expect_true(all(result$id <= 3))
})

test_that("ddbs_read_table works with different geometry types", {
  # Test with polygons
  result <- ddbs_read_table(conn_test, "test_countries_read")
  expect_s3_class(result, "sf")
  expect_true(any(grepl("POLYGON", sf::st_geometry_type(result))))
})

# 3. CRS handling ----
test_that("ddbs_read_table infers CRS from crs_duckspatial column", {
  # The 'test_points_read' table has crs_duckspatial = '4326'
  result <- ddbs_read_table(conn_test, "test_points_read")
  expect_equal(sf::st_crs(result), sf::st_crs(4326))
})

test_that("ddbs_read_table handles tables with no CRS info", {
  # Create a table with no CRS
  points_no_crs <- points_sf
  sf::st_crs(points_no_crs) <- NA
  ddbs_write_table(conn = conn_test, data = points_no_crs, name = "test_no_crs_read", overwrite = TRUE)

  # Read without specifying CRS (should result in NA CRS)
  result <- ddbs_read_table(conn_test, "test_no_crs_read")
  expect_true(is.na(sf::st_crs(result)))
})

# 4. Errors ----
test_that("ddbs_read_table throws an error for non-existent tables/views", {
  expect_error(
    ddbs_read_table(conn_test, "this_table_does_not_exist"),
    "not present in the database"
  )
})

# Disconnect
duckdb::dbDisconnect(conn_test, shutdown = TRUE)

#' dplyr methods for duckspatial_df
#'
#' These methods use dplyr's extension mechanism (dplyr_reconstruct) to
#' properly preserve spatial metadata through operations.
#'
#' @name duckspatial_df_dplyr
#' @keywords internal
NULL

# =============================================================================
# Core Extension: dplyr_reconstruct
# =============================================================================

#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr dplyr_reconstruct
dplyr_reconstruct.duckspatial_df <- function(data, template) {
  # Get the base classes from data
  base_classes <- class(data)
  
  # Reconstruct proper class hierarchy based on template
  # Template has: duckspatial_df, tbl_duckdb_connection, tbl_dbi, tbl_sql, tbl_lazy, tbl
  template_classes <- class(template)
  
  # Use template classes but only if data is likely a lazy table
  # (dplyr verbs on tbl_lazy usually return tbl_lazy)
  if (inherits(data, "tbl_sql")) {
    # Set class to match template structure
    # Ensure duckspatial_df is at the top
    if (!"duckspatial_df" %in% template_classes) {
        # Should not happen if template is duckspatial_df, but safe guard
        template_classes <- c("duckspatial_df", template_classes)
    }
    class(data) <- template_classes
  }
  
  # Restore spatial attributes from template
  # NOTE: We deliberately do NOT copy source_table here.

  # The source_table attribute is an optimization hint for get_query_list()
  # that allows direct table reference when the lazy query is unmodified.
  # However, dplyr_reconstruct is called AFTER dplyr verbs modify the query,
  # so copying source_table would cause get_query_list() to use the original
  # table instead of the modified lazy query (losing filters, selects, etc).
  #
  # By not setting source_table, get_query_list() will use sql_render() to
  # create a temporary view from the full lazy query, preserving all operations.
  attr(data, "sf_column") <- attr(template, "sf_column")
  attr(data, "crs") <- attr(template, "crs")
  
  data
}

# =============================================================================
# collect - Special handling for geometry conversion
# =============================================================================

#' Collect a duckspatial_df with flexible output formats
#'
#' @param x A duckspatial_df object
#' @param ... Additional arguments passed to dplyr::collect
#' @param as Output format: "sf" (default), "tibble" (no geometry), 
#'   "raw" (WKB bytes), or "geoarrow" (geoarrow_vctr)
#' @return Collected data in the specified format
#' @rdname ddbs_collect
#' @export
#' @importFrom dplyr collect
#' @importFrom rlang :=
collect.duckspatial_df <- function(x, ..., as = NULL) {
  # Resolve output type: parameter > global option > default "sf"
  # Note: if global option is "duckspatial_df" (lazy), collect defaults to "sf" (eager)
  if (is.null(as)) {
    as <- getOption("duckspatial.output_type", "sf")
    if (as == "duckspatial_df") as <- "sf"
  }
  
  as <- match.arg(as, c("sf", "tibble", "raw", "geoarrow"))
  geom_col <- attr(x, "sf_column") %||% "geom"
  crs_obj <- attr(x, "crs")
  
  # Strip class to treat as standard lazy table for further manipulation
  x_lazy <- x
  class(x_lazy) <- setdiff(class(x_lazy), "duckspatial_df")
  
  # --- Handle tibble case: just drop geometry and collect ---
  if (as == "tibble") {
    # If the user requested tibble, they likely don't want the geometry column
    # or don't care about its format. 
    # But if we just collect, it might fail if geom is effectively 'GEOMETRY' type?
    # DuckDB returns native blobs for GEOMETRY type. R handles them as list(raw).
    # So strictly speaking, simple collect works.
    
    # We drop geometry to be safe/consistent with expectation of "non-spatial tibble"
    if (geom_col %in% colnames(x_lazy)) {
       x_lazy <- dplyr::select(x_lazy, -dplyr::all_of(geom_col))
    }
    return(dplyr::collect(x_lazy, ...))
  }
  
  # --- For sf/raw/geoarrow: we need Valid WKB data ---
  # Native DuckDB geometry blobs are NOT standard WKB.
  # We MUST inject ST_AsWKB() execution into the query.
  
  # Check if geometry column exists in the output
  if (geom_col %in% colnames(x_lazy)) {
      conn <- dbplyr::remote_con(x_lazy)
      query_sql <- dbplyr::sql_render(x_lazy)
      
      # Check column type in the lazy table
      # Use cached type from attributes if available to avoid extra DESCRIBE round-trip
      cached_type <- attr(x, "geom_type")
      
      # Inject ST_AsWKB() conversion
      # We use dbplyr::sql to pass the raw SQL function
      # We assume the column name is safe or quoted by dbplyr if we used sym?
      # But inside sql() we must quote manually if needed. 
      # dbplyr::ident handles quoting.
      
      # Safer: use dplyr::mutate with sql snippet
      # We need to construct the SQL "ST_AsWKB("colname")" safely.
      # rlang::sym(geom_col) allows dbplyr to quote the column name in the generated SQL.
      # But we need to wrap it in function.
      
      # Method: use explicit SQL string construction
      # x_lazy |> mutate(geom = sql("ST_AsWKB(geom)"))
      # But quoting: ST_AsWKB("geom")
      
      # Let's use dbplyr's translation if available, or raw sql.
      # ST_AsWKB is standard.
      
      # Check column type in the lazy table
      # Use cached type from attributes if available to avoid extra DESCRIBE round-trip
      # Use cached type from attributes if available to avoid extra DESCRIBE round-trip
      cached_type <- attr(x, "geom_type")
      
      # Variables to hold resolved state
      target_col_sql <- geom_col # Default to attribute name
      should_convert <- FALSE
      
      if (!is.null(cached_type)) {
          should_convert <- grepl("GEOMETRY|BLOB", cached_type, ignore.case = TRUE)
      } else {
          tryCatch({
              # Check type of geom_col
              # DESCRIBE (query) is standard DuckDB
              desc <- DBI::dbGetQuery(conn, glue::glue("DESCRIBE {query_sql}"))
              
              # Match geom_col (case insensitive search)
              # use numeric index to handle NAs safely
              match_idx <- which(tolower(desc$column_name) == tolower(geom_col))
              
              if (length(match_idx) > 0) {
                  # Found column in DB stats
                  idx <- match_idx[1]
                  col_info <- desc[idx, ]
                  
                  # resolving correct casing from DB for quoting
                  target_col_sql <- col_info$column_name
                  
                  ctype <- if ("column_type" %in% names(col_info)) col_info$column_type else col_info$data_type
                  # We only wrap ST_AsWKB if it is GEOMETRY or BLOB/WKB_BLOB
                  should_convert <- grepl("GEOMETRY|BLOB", ctype, ignore.case = TRUE)
              } else {
                  # Column not found in DESCRIBE? 
                  # It might be present but DESCRIBE logic missed it or view shenanigans.
                  # Fallback: Assume it exists and needs conversion (Safe path)
                  should_convert <- TRUE
              }
          }, error = function(e) {
              # Fallback: safer to wrap than to get raw DuckDB internal blobs
              should_convert <<- TRUE 
          })
      }

      if (should_convert) {
           x_lazy <- dplyr::mutate(
               x_lazy, 
                !!rlang::sym(geom_col) := dbplyr::sql(glue::glue("ST_AsWKB({DBI::dbQuoteIdentifier(conn, target_col_sql)})"))
           )
      }

  }
  
  # Collect with dbplyr
  collected <- dplyr::collect(x_lazy, ...)
  
  # --- Convert based on output format ---
  if (!geom_col %in% names(collected)) {
    # No geometry column found (maybe user selected it out), return as tibble
    return(tibble::as_tibble(collected))
  }
  
  if (as == "raw") {
    # Return tibble with raw WKB bytes (no conversion)
    return(tibble::as_tibble(collected))
  }
  
  if (as == "geoarrow") {
    # Convert WKB to geoarrow_vctr
    geom_data <- collected[[geom_col]]
    if (!inherits(geom_data, "geoarrow_vctr")) {
      # Strip blob attributes if present (DuckDB blobs sometimes have extra attrs)
      attributes(geom_data) <- NULL
      col_converted <- tryCatch({
         geoarrow::as_geoarrow_vctr(
            wk::new_wk_wkb(geom_data),
            schema = geoarrow::geoarrow_wkb()
         )
      }, error = function(e) {
         cli::cli_warn("Failed to convert to geoarrow: {conditionMessage(e)}")
         geom_data
      })
      collected[[geom_col]] <- col_converted
    }
    return(tibble::as_tibble(collected))
  }
  
  # as == "sf" (default)
  convert_to_sf_wkb(
    data = collected,
    crs = crs_obj,
    x_geom = geom_col
  )
}




# =============================================================================
# compute - Force execution while staying lazy
# =============================================================================

#' Force computation of a duckspatial_df
#'
#' Executes the accumulated query and stores the result in a DuckDB temporary
#' table. The result remains lazy (a `duckspatial_df`) but points to the
#' materialized data, avoiding repeated computation of complex query plans.
#'
#' This is useful when you want to:
#' - Cache intermediate results for reuse across multiple subsequent operations
#' - Simplify complex query plans before heavy operations like spatial joins
#' - Force execution at a specific point without pulling data into R memory
#'
#' @param x A `duckspatial_df` object
#' @param name Optional name for the result table. If NULL, a unique temporary
#'   name is generated.
#' @param temporary If TRUE (default), creates a temporary table that is
#'   automatically cleaned up when the connection closes.
#' @param ... Additional arguments passed to [dplyr::compute()]
#' @return A new `duckspatial_df` pointing to the materialized table, with
#'   spatial metadata (CRS, geometry column) preserved.
#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr compute
#' @examples
#' \dontrun{
#' library(dplyr)
#'
#' # Complex pipeline - compute() caches intermediate result
#' result <- countries |>
#'   filter(POP_EST > 50000000) |>
#'   ddbs_filter(argentina, predicate = "touches") |>
#'   compute() |>  # Execute and cache here
#'   select(NAME_ENGL, POP_EST) |>
#'   ddbs_join(rivers, join = "intersects")
#'
#' # Check query plan - should reference the cached table
#' show_query(result)
#' }
compute.duckspatial_df <- function(x, name = NULL, temporary = TRUE, ...) {
  # Extract spatial metadata before compute
  crs <- attr(x, "crs")
  geom_col <- attr(x, "sf_column")
  
  # Strip our class to delegate to dbplyr's compute.tbl_sql
  class(x) <- setdiff(class(x), "duckspatial_df")
  
  # Execute via dbplyr
  result <- dplyr::compute(x, name = name, temporary = temporary, ...)
  
  # Get the new table name
  new_source <- tryCatch(
    as.character(dbplyr::remote_name(result)),
    error = function(e) NULL
  )

  # Get the new connection
  new_conn <- tryCatch(
    dbplyr::remote_con(result),
    error = function(e) NULL
  )
  
  # Re-wrap as duckspatial_df with preserved metadata
  new_duckspatial_df(
    result, 
    crs = crs, 
    geom_col = geom_col, 
    source_table = new_source,
    source_conn = new_conn
  )
}





# =============================================================================
# Passthrough verbs — geometry unchanged, use dplyr_reconstruct pattern
# =============================================================================


#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr select
select.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  
  # Strip geometry, select requested cols, re-add geometry as WKB
  res <- dplyr::select(.data, ..., dplyr::all_of(atts$sf_column))
  
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr filter
filter.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr arrange
arrange.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr rename
rename.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  
  # Check if geometry column is being renamed
  dots <- rlang::enquos(...)
  new_geom_col <- atts$sf_column
  for (nm in names(dots)) {
    if (rlang::as_name(dots[[nm]]) == atts$sf_column) {
      new_geom_col <- nm
      break
    }
  }
  
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = new_geom_col,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr slice
slice.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom utils head
head.duckspatial_df <- function(x, n = 6L, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}


#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr glimpse
glimpse.duckspatial_df <- function(x, width = NULL, ...) {
  # Preserve spatial metadata
  crs <- attr(x, "crs")
  geom_col <- attr(x, "sf_column")
  bbox <- st_bbox(x)
  geomtype <- ddbs_geometry_type(x, by_feature = FALSE) |> 
    as.character()

  # Strip class to delegate to dplyr's glimpse.tbl_lazy
  class(x) <- setdiff(class(x), "duckspatial_df")

  # Header
  cat(cli::col_white("# A duckspatial lazy spatial table\n"))
  cat(cli::col_white("#"), cli::col_blue("\u25cf CRS:"), cli::col_white(ddbs_format_crs(crs)), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Geometry column:"), cli::col_white(geom_col), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Geometry type:"), cli::col_white(geomtype), "\n")
  cat(cli::col_white("#"), cli::col_blue("\u25cf Bounding box:"), 
      cli::col_white(sprintf("xmin: %.5g ymin: %.5g xmax: %.5g ymax: %.5g", 
                              bbox["xmin"], bbox["ymin"], bbox["xmax"], bbox["ymax"])), "\n")
  cat(cli::col_white("# Data backed by DuckDB (dbplyr lazy evaluation)\n"))
  cat(cli::col_white("# Use ddbs_collect() or st_as_sf() to materialize to sf\n"))
  cat(cli::col_white("#\n"))

  # Execute via dplyr
  NextMethod()

  invisible(x)
}





# =============================================================================
# Transforming verbs — may change geometry, use new_duckspatial_df pattern
# =============================================================================

#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr mutate
mutate.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr count
count.duckspatial_df <- function(x, ..., wt = NULL, sort = FALSE, name = NULL) {
  # count() aggregates — intentionally drops spatial class (no geometry)
  class(x) <- setdiff(class(x), "duckspatial_df")
  NextMethod()
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr distinct
distinct.duckspatial_df <- function(.data, ..., .keep_all = FALSE) {
  crs <- attr(.data, "crs")
  geom_col <- attr(.data, "sf_column")
  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()

  ## if geometry column is in the table, return duckspatial_df
  ## otherwise, return a lazy tbl
  if (geom_col %in% colnames(res)) {
    class(res) <- c("duckspatial_df", class(res))
    attr(res, "sf_column") <- geom_col
    attr(res, "crs") <- crs
    res
  } else {
    res
  }
}





# =============================================================================
# Join methods - preserve spatial class from left side
# =============================================================================
# Note: dbplyr join methods typically return tbl_lazy.
# Our dplyr_reconstruct should handle class restoration if dbplyr calls it.
# But often generic joins dispatch to dbplyr methods directly.
# Let's enable generic join methods just in case to verify attributes.

# #' @rdname duckspatial_df_dplyr
# #' @export
# #' @importFrom dplyr left_join
# left_join.duckspatial_df <- function(x, y, by = NULL, copy = FALSE, 
#                                       suffix = c(".x", ".y"), ...,
#                                       keep = NULL, na_matches = c("na", "never"),
#                                       relationship = NULL) {
#   # Strip class to avoid infinite recursion if NextMethod doesn't strip it?
#   # Actually NextMethod works fine.
#   out <- NextMethod()
  
#   # If out lost the class (common with dbplyr), restore it
#   if (!inherits(out, "duckspatial_df")) {
#       # Re-wrap
#       class(out) <- c("duckspatial_df", class(out))
#       attr(out, "sf_column") <- attr(x, "sf_column")
#       attr(out, "crs") <- attr(x, "crs")
#   }
  
#   out
# }

# #' @rdname duckspatial_df_dplyr
# #' @export
# #' @importFrom dplyr inner_join
# inner_join.duckspatial_df <- function(x, y, by = NULL, copy = FALSE, 
#                                        suffix = c(".x", ".y"), ...,
#                                        keep = NULL, na_matches = c("na", "never"),
#                                        relationship = NULL) {
#   out <- NextMethod()
#   if (!inherits(out, "duckspatial_df")) {
#       class(out) <- c("duckspatial_df", class(out))
#       attr(out, "sf_column") <- attr(x, "sf_column")
#       attr(out, "crs") <- attr(x, "crs")
#   }
#   out
# }

#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr left_join
left_join.duckspatial_df <- function(x, y, by = NULL, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  if (inherits(y, "duckspatial_df")) {
    class(y) <- setdiff(class(y), "duckspatial_df")
  }
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr inner_join
inner_join.duckspatial_df <- function(x, y, by = NULL, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  if (inherits(y, "duckspatial_df")) {
    class(y) <- setdiff(class(y), "duckspatial_df")
  }
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr right_join
right_join.duckspatial_df <- function(x, y, by = NULL, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  if (inherits(y, "duckspatial_df")) {
    class(y) <- setdiff(class(y), "duckspatial_df")
  }
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}



#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr full_join
full_join.duckspatial_df <- function(x, y, by = NULL, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  if (inherits(y, "duckspatial_df")) {
    class(y) <- setdiff(class(y), "duckspatial_df")
  }
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}





# =============================================================================
# Grouping and aggregating methods
# =============================================================================

.strip_spatial_attrs <- function(x) {
  attr(x, "sf_column")    <- NULL
  attr(x, "crs")          <- NULL
  attr(x, "source_table") <- NULL
  attr(x, "source_conn")  <- NULL
  x
}


#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr group_by
group_by.duckspatial_df <- function(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data)) {
  NextMethod()
}

#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr ungroup
ungroup.duckspatial_df <- function(x, ...) {
  atts <- attributes(x)
  class(x) <- setdiff(class(x), "duckspatial_df")
  res <- NextMethod()
  new_duckspatial_df(
    x            = res,
    crs          = atts$crs,
    geom_col     = atts$sf_column,
    source_table = NULL,
    source_conn  = atts$source_conn,
    create_view  = TRUE
  )
}

#' @rdname duckspatial_df_dplyr
#' @export
#' @importFrom dplyr summarise
summarise.duckspatial_df <- function(.data, ...) {
  atts <- attributes(.data)
  geom_col <- atts$sf_column

  # Check if geometry column is preserved in the summary by name only
  dots <- rlang::enquos(...)
  geom_preserved <- geom_col %in% names(dots)

  class(.data) <- setdiff(class(.data), "duckspatial_df")
  res <- NextMethod()

  if (geom_preserved) {
    new_duckspatial_df(
      x            = res,
      crs          = atts$crs,
      geom_col     = geom_col,
      source_table = NULL,
      source_conn  = atts$source_conn,
      create_view  = TRUE
    )
  } else {
    cli::cli_warn(c(
      "Geometry column {.field {geom_col}} was dropped by {.fn summarise}.",
      "i" = "Use a spatial aggregate like {.fn {geom_col} = ddbs_union({geom_col})} to preserve geometry.",
      "i" = "Result is no longer a {.cls duckspatial_df}."
    ))
    .strip_spatial_attrs(res)
  }
}

#' Create point geometries from coordinate vectors
#'
#' Constructs POINT geometries from numeric coordinate vectors, optionally
#' including Z (elevation) and M (measure) dimensions and extra attribute
#' columns.
#'
#' @param x Numeric vector of X (longitude) coordinates.
#' @param y Numeric vector of Y (latitude) coordinates.
#' @param z Optional numeric vector of Z (elevation) coordinates.
#' @param m Optional numeric vector of M (measure) coordinates. Requires
#'   \code{z}.
#' @param ... Named vectors of additional attribute columns to include in the
#'   output. Each must have the same length as \code{x}.
#' @param crs Character or numeric CRS specification (e.g. \code{"EPSG:4326"}
#'   or \code{4326}). Defaults to \code{NULL} (no CRS assigned).
#' @param geom_col Name of the geometry column in the output. Defaults to
#'   \code{"geometry"}.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' ## 2D points
#' ddbs_point(
#'   x = c(-58.38, -64.18, -60.64),
#'   y = c(-34.60, -31.42, -32.95),
#'   crs = 4326
#' )
#'
#' ## 3D points with extra columns
#' ddbs_point(
#'   x   = c(0, 1, 2),
#'   y   = c(0, 1, 2),
#'   z   = c(10, 20, 30),
#'   id  = 1:3,
#'   crs = 4326
#' )
#' }
ddbs_point <- function(
    x,
    y,
    z = NULL,
    m = NULL,
    ...,
    crs = NULL,
    geom_col = "geometry",
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Validate inputs
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")

    if (!is.numeric(x)) cli::cli_abort("{.arg x} must be a numeric vector.")
    if (!is.numeric(y)) cli::cli_abort("{.arg y} must be a numeric vector.")

    n <- length(x)
    if (length(y) != n) {
        cli::cli_abort("{.arg x} and {.arg y} must have the same length.")
    }

    if (!is.null(m) && is.null(z)) {
        cli::cli_abort("{.arg m} requires {.arg z} to also be provided.")
    }
    if (!is.null(z)) {
        if (!is.numeric(z)) cli::cli_abort("{.arg z} must be a numeric vector.")
        if (length(z) != n) {
            cli::cli_abort("{.arg z} must have the same length as {.arg x} ({n}).")
        }
    }
    if (!is.null(m)) {
        if (!is.numeric(m)) cli::cli_abort("{.arg m} must be a numeric vector.")
        if (length(m) != n) {
            cli::cli_abort("{.arg m} must have the same length as {.arg x} ({n}).")
        }
    }

    dots <- list(...)
    if (length(dots) > 0) {
        if (is.null(names(dots)) || any(names(dots) == "")) {
            cli::cli_abort("All arguments passed via {.arg ...} must be named.")
        }
        reserved <- c("x", "y", "z", "m", geom_col)
        conflict <- intersect(names(dots), reserved)
        if (length(conflict) > 0) {
            cli::cli_abort(
                "Column name{?s} {.val {conflict}} in {.arg ...} conflict with reserved coordinate or geometry column names."
            )
        }
        bad_len <- names(dots)[vapply(dots, length, integer(1)) != n]
        if (length(bad_len) > 0) {
            cli::cli_abort(
                "Extra column{?s} {.val {bad_len}} must have the same length as {.arg x} ({n})."
            )
        }
    }

    if (!is.null(name) && is.null(conn)) {
        cli::cli_abort(
            "If {.arg name} is provided, {.arg conn} must be a valid DuckDB connection."
        )
    }

    # 1. Resolve connection and mode
    target_conn <- conn %||% ddbs_default_conn()
    mode        <- get_mode(mode, name)

    # 2. Build data frame and write to a temp table
    df <- data.frame(x = x, y = y, check.names = FALSE, stringsAsFactors = FALSE)
    if (!is.null(z)) df[["z"]] <- z
    if (!is.null(m)) df[["m"]] <- m
    for (nm in names(dots)) df[[nm]] <- dots[[nm]]

    tmp_name <- ddbs_temp_table_name()
    DBI::dbWriteTable(target_conn, tmp_name, df, temporary = TRUE)
    on.exit(DBI::dbRemoveTable(target_conn, tmp_name), add = TRUE)

    # 3. Build ST_MakePoint expression (cast to GEOMETRY so build_geom_query
    #    can further annotate with CRS without hitting POINT_2D cast limits)
    st_fun <- if (!is.null(m)) {
        '(ST_MakePoint("x", "y", "z", "m")::GEOMETRY)'
    } else if (!is.null(z)) {
        '(ST_MakePoint("x", "y", "z")::GEOMETRY)'
    } else {
        '(ST_MakePoint("x", "y")::GEOMETRY)'
    }

    # 4. Build SELECT clause for extra columns
    extra_select <- if (length(dots) > 0) {
        paste0(paste0('"', names(dots), '"', collapse = ", "), ", ")
    } else {
        ""
    }

    geom_expr  <- build_geom_query(st_fun, name, crs, mode)
    base_query <- glue::glue(
        'SELECT {extra_select}{geom_expr} AS "{geom_col}" FROM "{tmp_name}";'
    )

    # 5. Handle output
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base_query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base_query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs,
            x_geom = geom_col
        )
    }
}


#' Generate point geometries from coordinates
#'
#' Converts a data frame with coordinate columns into spatial point geometries.
#'
#' @template x
#' @param coords Character vector of length 2 specifying the names of the
#'        longitude and latitude columns (or X and Y coordinates). Defaults to
#'        \code{c("lon", "lat")}.
#' @param crs Character or numeric. The Coordinate Reference System (CRS) of the
#'        input coordinates. Can be specified as an EPSG code (e.g., \code{"EPSG:4326"}
#'        or \code{4326}) or a WKT string. Defaults to \code{"EPSG:4326"} (WGS84
#'        longitude/latitude).
#' @param remove Logical. If \code{TRUE} (default), the coordinate columns
#'        specified in \code{coords} are removed from the output.
#' @param na.fail Logical. If \code{TRUE} (default), the function errors if
#'        any missing values (NAs) are found in the coordinate columns.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#' @param ... Additional arguments. Currently supports \code{geom_col} to
#'        specify the name of the geometry column in the output.
#'
#' @template returns_mode
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' ## create sample data with coordinates
#' cities_df <- data.frame(
#'   city = c("Buenos Aires", "Córdoba", "Rosario"),
#'   lon = c(-58.3816, -64.1811, -60.6393),
#'   lat = c(-34.6037, -31.4201, -32.9468),
#'   population = c(3075000, 1391000, 1193605)
#' )
#'
#' # option 1: convert data frame to sf object
#' cities_ddbs <- ddbs_as_points(cities_df)
#'
#' # specify custom coordinate column names and keep them in output
#' cities_df2 <- data.frame(
#'   city = c("Mendoza", "Tucumán"),
#'   longitude = c(-68.8272, -65.2226),
#'   latitude = c(-32.8895, -26.8241)
#' )
#'
#' ddbs_as_points(cities_df2, coords = c("longitude", "latitude"), remove = FALSE)
#'
#'
#' ## option 2: convert table in duckdb to spatial table
#'
#' # create a duckdb connection and write data
#' conn <- duckspatial::ddbs_create_conn()
#' DBI::dbWriteTable(conn, "cities_tbl", cities_df, overwrite = TRUE)
#'
#' # convert to spatial table in database
#' ddbs_as_points(
#'     x = "cities_tbl",
#'     conn = conn,
#'     name = "cities_spatial",
#'     overwrite = TRUE
#' )
#'
#' # read the spatial table
#' ddbs_read_table(conn, "cities_spatial")
#' }
ddbs_as_points <- function(
    x,
    coords = c("lon", "lat"),
    crs = "EPSG:4326",
    remove = TRUE,
    na.fail = TRUE,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE,
    ...) {
        
    dots <- list(...)
    
    # 0. Validate inputs
    assert_name(name)
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
    assert_logic(remove, "remove")
    assert_logic(na.fail, "na.fail")

    if (length(coords) != 2) {
        cli::cli_abort("{.arg coords} must be a character vector of length 2.")
    }
  

    # Respect geom_col
    target_geom <- dots$geom_col
  
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn, geom_col = target_geom)

    ## 1.2. Pre-extract attributes
    mode <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Validate coordinates existence and NAs
    # Use double quotes for column names to handle spaces/special characters
    coords_quoted <- paste0('"', coords, '"')
    
    # Check if columns exist
    tryCatch({
      DBI::dbExecute(target_conn, glue::glue("SELECT {paste0(coords_quoted, collapse = ', ')} FROM {x_list$query_name} WHERE 1=0"))
    }, error = function(e) {
      cli::cli_abort("Coordinate columns {.val {coords}} not found in input.")
    })

    # na.fail check
    if (na.fail) {
      null_check_sql <- glue::glue(
        "SELECT COUNT(*) as null_count FROM {x_list$query_name} 
         WHERE {paste0(coords_quoted, ' IS NULL', collapse = ' OR ')}"
      )
      null_count <- DBI::dbGetQuery(target_conn, null_check_sql)$null_count
      if (null_count > 0) {
        cli::cli_abort("Missing values found in coordinate columns {.val {coords}}. Set {.code na.fail = FALSE} to ignore.")
      }
    }

    ## 2.2. Coords as character string for ST_Point
    coords_str <- paste0(coords_quoted, collapse = ", ")
  
    ## 2.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
    st_function <- glue::glue("ST_Point({coords_str})")
    
    # Handle 'remove' using DuckDB EXCLUDE
    select_clause <- if (remove) {
      glue::glue("SELECT * EXCLUDE ({coords_str}),")
    } else {
      "SELECT *,"
    }

    # Respect geom_col for query building
    query_geom <- target_geom %||% "geom"

    base.query <- glue::glue("
      {select_clause}
      {build_geom_query(st_function, name, crs, mode)} as {query_geom}
      FROM {x_list$query_name};
    ")    


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
            conn      = target_conn,
            name      = name,
            query     = base.query,
            overwrite = overwrite,
            quiet     = quiet
        )
    } else {
        ddbs_handle_query(
            query  = base.query,
            conn   = target_conn,
            mode   = mode,
            crs    = crs,
            x_geom = query_geom
        )
    }
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## create data
cities_tbl <- data.frame(
  city = c("Buenos Aires", "Córdoba", "Rosario"),
  lon = c(-58.3816, -64.1811, -60.6393),
  lat = c(-34.6037, -31.4201, -32.9468),
  population = c(3075000, 1391000, 1193605)
)

## write data
DBI::dbWriteTable(conn_test, "cities", cities_tbl)


# 1. ddbs_as_points() -------------------------------------------------------

## - CHECK 1.1: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.2: messages work
## - CHECK 1.3: writting a table works
## - CHECK 1.4: different column names, and CRS
## - CHECK 2.1: specific errors
## - CHECK 2.2: general errors
describe("ddbs_as_points()", {

  ### EXPECTED BEHAVIOR -------------------------------------------------

  describe("expected behavior", {

    it("returns different output formats", {
      output_ddbs_fmt <- ddbs_as_points(cities_tbl)
      output_sf_fmt <- ddbs_as_points(cities_tbl, mode = "sf")

      expect_s3_class(output_ddbs_fmt, "duckspatial_df")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_as_points(cities_tbl))
      expect_message(ddbs_as_points("cities", conn = conn_test, name = "as_spatial"))
      expect_message(ddbs_as_points("cities", conn = conn_test, name = "as_spatial", overwrite = TRUE))
      expect_true(ddbs_as_points("cities", conn = conn_test, name = "as_spatial2"))

      expect_no_message(ddbs_as_points(cities_tbl, quiet = TRUE))
      expect_no_message(ddbs_as_points("cities", conn = conn_test, name = "as_spatial", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_tbl <- ddbs_read_table(conn_test, "as_spatial")
      expect_equal(
        ddbs_collect(ddbs_as_points(cities_tbl))$geometry,
        output_tbl$geometry
      )
    })

    it("handles different column names and CRS correctly", {
      output_tbl <- ddbs_read_table(conn_test, "as_spatial")
      cities_3857 <- output_tbl |> 
        sf::st_transform("EPSG:3847") %>% 
        dplyr::mutate(
          xx = sf::st_coordinates(.)[, 1],
          yy = sf::st_coordinates(.)[, 2]
        ) |> 
        sf::st_drop_geometry()

      output_3857 <- ddbs_as_points(
        cities_3857,
        coords = c("xx", "yy"),
        crs = "EPSG:3857"
      )

      expect_equal(sf::st_crs(output_3857), sf::st_crs("EPSG:3857"))
    })

    it("supports remove and na.fail arguments", {
      # remove = TRUE (default)
      out_remove <- ddbs_as_points(cities_tbl, coords = c("lon", "lat"), remove = TRUE)
      expect_false(any(c("lon", "lat") %in% colnames(out_remove)))
      
      # remove = FALSE
      out_keep <- ddbs_as_points(cities_tbl, coords = c("lon", "lat"), remove = FALSE)
      expect_true(all(c("lon", "lat") %in% colnames(out_keep)))
      
      # na.fail = TRUE (default)
      cities_na <- cities_tbl
      cities_na$lon[1] <- NA
      expect_error(ddbs_as_points(cities_na), "Missing values found")
      
      # na.fail = FALSE
      out_na <- ddbs_as_points(cities_na, na.fail = FALSE)
      expect_s3_class(out_na, "duckspatial_df")
    })
  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("validates coords argument", {
      expect_error(ddbs_as_points(cities_tbl, coords = c("longitude", "latitude")))
      expect_error(ddbs_as_points(cities_tbl, coords = c("longitude", "latitude", "z")))
    })

    it("validates CRS argument", {
      expect_error(ddbs_as_points(cities_tbl, crs = "NICE_CRS"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_as_points("cities", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_as_points(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_as_points(cities_tbl, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_as_points(cities_tbl, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_as_points(cities_tbl, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_as_points(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_as_points(cities_tbl, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})



# 2. ddbs_point() -----------------------------------------------------------

## - CHECK 1.1: 2D, 3D, and 4D output classes
## - CHECK 1.2: returns sf when mode = "sf"
## - CHECK 1.3: CRS is assigned correctly
## - CHECK 1.4: extra columns appear in output
## - CHECK 1.5: geom_col renames the geometry column
## - CHECK 1.6: messages shown/suppressed correctly
## - CHECK 1.7: writes table to database
## - CHECK 2.1: standard errors (x/y type, length mismatch, m without z,
##              named ..., conflicting names, length mismatch in ...,
##              name without conn)
describe("ddbs_point()", {

  x_vec <- c(-58.38, -64.18, -60.64)
  y_vec <- c(-34.60, -31.42, -32.95)
  z_vec <- c(100, 200, 300)
  m_vec <- c(1.0, 2.0, 3.0)

  describe("expected behavior", {

    it("returns duckspatial_df by default for 2D, 3D, and 4D", {
      expect_s3_class(ddbs_point(x_vec, y_vec),                                        "duckspatial_df")
      expect_s3_class(ddbs_point(x_vec, y_vec, z = z_vec),                             "duckspatial_df")
      expect_s3_class(ddbs_point(x_vec, y_vec, z = z_vec, m = m_vec),                  "duckspatial_df")
    })

    it("returns sf when mode = 'sf'", {
      out <- ddbs_point(x_vec, y_vec, crs = 4326, mode = "sf")
      expect_s3_class(out, "sf")
    })

    it("assigns CRS correctly", {
      out <- ddbs_point(x_vec, y_vec, crs = 4326, mode = "sf")
      expect_equal(sf::st_crs(out), sf::st_crs(4326))
    })

    it("includes extra columns in output", {
      out <- ddbs_point(x_vec, y_vec, id = 1:3, label = c("a", "b", "c"), mode = "sf")
      expect_true("id"    %in% colnames(out))
      expect_true("label" %in% colnames(out))
    })

    it("renames geometry column when geom_col is set", {
      out <- ddbs_point(x_vec, y_vec, geom_col = "geom", mode = "sf")
      expect_true("geom" %in% colnames(out))
      expect_false("geometry" %in% colnames(out))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_point(x_vec, y_vec))
      expect_message(
        ddbs_point(x_vec, y_vec, conn = conn_test, name = "pts_tbl")
      )
      expect_no_message(
        ddbs_point(x_vec, y_vec, conn = conn_test, name = "pts_tbl_q",
                   quiet = TRUE)
      )
    })

    it("writes table to database and returns TRUE invisibly", {
      out <- ddbs_point(x_vec, y_vec, crs = 4326, conn = conn_test,
                        name = "pts_tbl2")
      expect_true(out)
      expect_true(DBI::dbExistsTable(conn_test, "pts_tbl2"))
    })

    it("coordinates round-trip correctly", {
      out <- ddbs_point(x_vec, y_vec, crs = 4326, mode = "sf")
      coords <- sf::st_coordinates(out)
      expect_equal(coords[, "X"], x_vec, tolerance = 1e-6)
      expect_equal(coords[, "Y"], y_vec, tolerance = 1e-6)
    })
  })

  describe("errors", {

    it("errors if x is not numeric", {
      expect_error(ddbs_point("a", y_vec))
    })

    it("errors if y is not numeric", {
      expect_error(ddbs_point(x_vec, "b"))
    })

    it("errors if x and y have different lengths", {
      expect_error(ddbs_point(x_vec, y_vec[-1]))
    })

    it("errors if m is provided without z", {
      expect_error(ddbs_point(x_vec, y_vec, m = m_vec))
    })

    it("errors if z is not numeric", {
      expect_error(ddbs_point(x_vec, y_vec, z = c("a", "b", "c")))
    })

    it("errors if z has wrong length", {
      expect_error(ddbs_point(x_vec, y_vec, z = z_vec[-1]))
    })

    it("errors if ... contains unnamed arguments", {
      expect_error(ddbs_point(x_vec, y_vec, z_vec, m_vec, 1:3))
    })

    it("errors if ... names conflict with reserved column names", {
      expect_error(ddbs_point(x_vec, y_vec, geometry = 1:3))
    })

    it("errors if ... columns have wrong length", {
      expect_error(ddbs_point(x_vec, y_vec, id = 1:2))
    })

    it("errors if name is provided without conn", {
      expect_error(ddbs_point(x_vec, y_vec, name = "pts_err"))
    })

    it("errors if overwrite is not logical", {
      expect_error(ddbs_point(x_vec, y_vec, overwrite = 1))
    })

    it("errors if quiet is not logical", {
      expect_error(ddbs_point(x_vec, y_vec, quiet = 1))
    })
  })
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

#' @param mode Character. Controls the return type. Options:
#'   \itemize{
#'     \item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
#'     \item \code{"sf"}: Eagerly collected sf object (uses memory)
#'   }
#'   Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
#'   per-function via this argument. Per-function overrides global setting.

#' Write an SF Object to a DuckDB Database
#'
#' This function writes a Simple Features (SF) object into a DuckDB database as a new table.
#' The table is created in the specified schema of the DuckDB database.
#' 
#' **CRS Persistence:** \code{duckspatial} ensures CRS metadata is retained across 
#' sessions using two strategies: \strong{Native Spatial Storage} (for DuckDB 1.5.0+ 
#' databases) and \strong{Legacy Compatibility} (using column comments for older 
#' database versions). For file-based spatial data interchange, 
#' [ddbs_write_dataset()] to GeoParquet (\code{.parquet}) is recommended.
#'
#' @template conn
#' @param data A \code{sf} object to write to the DuckDB database, or the path to
#' a local file that can be read with `ST_READ`
#' @template name
#' @template overwrite
#' @param temp_view If `TRUE`, registers the `sf` object as a temporary Arrow-backed database 'view' 
#' using [ddbs_register_table] instead of creating a persistent table. This is much faster but the view 
#' will not persist. Defaults to `FALSE`.
#' @template quiet
#'
#' @returns TRUE (invisibly) for successful import
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(sf)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## create random points
#' random_points <- data.frame(
#'   id = 1:5,
#'   x = runif(5, min = -180, max = 180),  # Random longitude values
#'   y = runif(5, min = -90, max = 90)     # Random latitude values
#' )
#'
#' ## convert to sf
#' sf_points <- st_as_sf(random_points, coords = c("x", "y"), crs = 4326)
#'
#' ## insert data into the database
#' ddbs_write_table(conn, sf_points, "points")
#'
#' ## read data back into R
#' ddbs_read_table(conn, "points")
#'
#' ## disconnect from db
#' ddbs_stop_conn(conn)
#' }
ddbs_write_table <- function(
    conn,
    data,
    name,
    overwrite = FALSE,
    temp_view = FALSE,
    quiet = FALSE
) {
    # 1. Checks
    ## Check if connection is correct
    dbConnCheck(conn)
    input_crs <- ddbs_input_crs_for_write(data, conn)

    ## Handle temp_view
    if (temp_view) {
        return(ddbs_register_table(conn, data, name, overwrite, quiet))
    }

    # Handle duckspatial_df/tbl_lazy - try cross-connection import first
    if (inherits(data, "duckspatial_df") || inherits(data, "tbl_lazy")) {
        source_conn <- get_conn_from_input(data)
        
        # Check if same connection (no import needed, just materialize)
        same_conn <- !is.null(source_conn) && identical(source_conn, conn)
        
        if (!same_conn && !is.null(source_conn)) {
            # Cross-connection: try efficient import strategies first
            # Use a temp name for the import view, we'll convert to table after
            temp_import_name <- paste0("temp_import_", gsub("-", "_", uuid::UUIDgenerate()))
            
            import_result <- tryCatch({
                import_view_to_connection(conn, source_conn, data, temp_import_name)
            }, error = function(e) NULL)
            
            if (!is.null(import_result)) {
                # Successful import - convert temp view to permanent table
                ## convenient names of table and/or schema.table
                name_list <- get_query_name(name)
                ## get schema.table available in the database
                tables_df <- ddbs_list_tables(conn)
                db_tables <- paste0(tables_df$table_schema, ".", tables_df$table_name) |>
                    sub(pattern = "^main\\.", replacement = "")
                ## Check if table name already exists
                if (name_list$query_name %in% db_tables & !overwrite)
                    cli::cli_abort("The provided name is already present in the database. Please, use `overwrite = TRUE` or choose a different name.")
                
                # Handle overwrite
                overwrite_table(name_list$query_name, conn, quiet, overwrite)
                
                # Create permanent table from the imported view
                DBI::dbExecute(conn, glue::glue(
                    "CREATE TABLE {name_list$query_name} AS SELECT * FROM {import_result$name}"
                ))
                
                # Cleanup temp view
                tryCatch(
                    DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {import_result$name}")),
                    error = function(e) NULL
                )
                import_result$cleanup()
                
                if (isFALSE(quiet)) {
                    cli::cli_alert_success("Table {name_list$query_name} imported via {import_result$method}")
                }
                geom_col <- tryCatch(get_geom_name(conn, name_list$query_name), error = function(e) NULL)
                ddbs_write_legacy_crs_comment_if_needed(
                    conn,
                    name_list$query_name,
                    geom_col = geom_col,
                    crs = input_crs
                )
                return(invisible(TRUE))
            }
        }
        
        # Fallback: collect to sf and re-upload (slow but always works)
        if (inherits(data, "duckspatial_df")) {
             data <- ddbs_collect(data, as = "sf")
        } else {
             data <- dplyr::collect(data) |> sf::st_as_sf()
        }
    }

    ## convenient names of table and/or schema.table
    name_list <- get_query_name(name)
    ## get schema.table available in the database
    tables_df <- ddbs_list_tables(conn)
    db_tables <- paste0(tables_df$table_schema, ".", tables_df$table_name) |>
        sub(pattern = "^main\\.", replacement = "")
    ## Check if table name already exists
    if (name_list$query_name %in% db_tables & !overwrite)
        cli::cli_abort("The provided name is already present in the database. Please, use `overwrite = TRUE` or choose a different name.")

    # 2. Handle overwrite
    overwrite_table(name_list$query_name, conn, quiet, overwrite)

    ## 3. insert data
    if (inherits(data, "sf")) {

        # 3. Handle unsupported geometries (TOO SLOW)
        # unsupported_types <- c("GEOMETRYCOLLECTION")
        # geom_types <- unique(sf::st_geometry_type(data))
        # if (any(geom_types %in% unsupported_types)) {
        #     cli::cli_abort("Unsupported geometry types found: {paste(geom_types[geom_types %in% unsupported_types], collapse = ', ')}")
        # }

        # 4. Prepare data for writing - import as data frame with geom as binary
        ## Get geometry column name
        geom_name <- setdiff(names(data), names(sf::st_drop_geometry(data)))
        ## Extract geometry as binary and append to data frame
        wkb_data <- sf::st_as_binary(sf::st_geometry(data), EWKB = TRUE)
        data_df <- as.data.frame(data)
        data_df[[geom_name]] <- wkb_data  # Ensure raw data is preserved

        ## Get the CRS, and define the geometry type for duckdb
        geom_field <- get_geometry_type_duckdb(data, conn)

        ## Warn if no CRS was found in the input data
        if (geom_field == "GEOMETRY") {
            cli::cli_alert_warning("No CRS found in the input data. The table will be created without CRS information.")
        }

        ## Write data into DuckDB
        # duckdb::duckdb_register(conn, "temp_view", data_df, experimental = TRUE) # check later
        duckdb::dbWriteTable(
            conn, 
            DBI::Id(schema = name_list$schema_name, table = name_list$table_name), 
            data_df, 
            field.types = c(geom_name = "BLOB")
        )
 
        ## Convert to spatial with CRS
        DBI::dbExecute(conn, glue::glue("
            ALTER TABLE {name_list$query_name}
            ALTER COLUMN {geom_name} SET DATA TYPE {geom_field} USING ST_GeomFromWKB({geom_name});
        "))
        ddbs_write_legacy_crs_comment_if_needed(
            conn,
            name_list$query_name,
            geom_col = geom_name,
            crs = input_crs
        )
        # duckdb::duckdb_unregister(conn, "temp_view") |> on.exit()
        

    } else if (!is.character(data) || length(data) != 1) {
        cli::cli_abort("{.arg data} must be an {.cls sf} object, a {.cls duckspatial_df}, or a file path string.")
    } else {
        ## check file extension
        file_ext <- tools::file_ext(data)
        if (file_ext == "parquet") {
            cli::cli_abort("Direct parquet file writting is not currently supported.")
            ## Current state v1.5.0 - we can write parquet files, but the geometry is stored as STRUCT, and
            ## we cannot convert it to GEOMETRY type in duckdb
            
            ## insert data
            # DBI::dbExecute(
            #     conn,
            #     glue::glue("CREATE TABLE {name_list$query_name} AS SELECT * FROM read_parquet('{data}')")
            # )
            # ## Get the CRS
            # get_parquet_crs(data, conn)
            # ## specify geometry column
            # ## - try to get geom column name
            # metadata_df <- DBI::dbGetQuery(conn, glue::glue("DESCRIBE {name_list$query_name}"))
            # geom_name <- metadata_df$column_name[grepl("STRUCT", metadata_df$column_type)]
            # DBI::dbExecute(conn, glue::glue("
            #     ALTER TABLE {name_list$query_name}
            #     ALTER COLUMN {geom_name} SET DATA TYPE GEOMETRY USING ST_GeomFromWKB({geom_name});
            # "))
        
        } else {
            ## insert files (in duckdb v1.5, ST_Read() already manages CRS)
            DBI::dbExecute(
                conn,
                glue::glue("CREATE TABLE {name_list$query_name} AS SELECT * FROM ST_Read('{data}')")
            )
            ddbs_write_legacy_crs_comment_if_needed(
                conn,
                name_list$query_name,
                crs = input_crs
            )
        }
    }


    # 6. User feedback
    if (isFALSE(quiet)) {
        cli::cli_alert_success("Table {name_list$query_name} successfully imported")
        }

    return(invisible(TRUE))

}




#' Write an SF Object to a DuckDB Database
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `ddbs_write_vector()` was renamed to \code{\link{ddbs_write_table}}.
#'
#' @inheritParams ddbs_write_table
#' @returns TRUE (invisibly) for successful import
#' @export
#' @keywords internal
ddbs_write_vector <- function(
    conn,
    data,
    name,
    overwrite = FALSE,
    temp_view = FALSE,
    quiet = FALSE) {
    
    lifecycle::deprecate_soft(
        when    = "1.0.0",
        what    = "ddbs_write_vector()",
        with    = "ddbs_write_table()"
    )
    
    ddbs_write_table(
        conn = conn,
        data = data,
        name = name,
        overwrite = overwrite,
        temp_view = temp_view,
        quiet = quiet
    )
}

#' @returns Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
#' \itemize{
#'   \item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
#'   \item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
#' \code{sf} equivalent (e.g. `sf` or `units` vector).
#' }
#' When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).

#' Get or set global duckspatial options
#'
#' @param output_type Character string. Controls the default return type for [ddbs_collect].
#'   Must be one of:
#'   \itemize{
#'     \item \code{"sf"} (default): Eagerly collected \code{sf} object (in-memory).
#'     \item \code{"tibble"}: Eagerly collected \code{tibble} without geometry.
#'     \item \code{"raw"}: Eagerly collected \code{tibble} with geometry as raw WKB bytes.
#'     \item \code{"geoarrow"}: Eagerly collected \code{tibble} with geometry as \code{geoarrow_vctr}.
#'   }
#'   If \code{NULL} (the default), the existing option is not changed.
#' @param mode Character. Controls the return type. Options:
#'   \itemize{
#'     \item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
#'     \item \code{"sf"}: Eagerly collected sf object (uses memory)
#'   }
#'   If \code{NULL} (the default), the existing option is not changed.
#' @param duckdb_storage_version Character. The default DuckDB storage compatibility version
#'   for newly created database files. See \url{https://duckdb.org/docs/internals/storage}
#'   for details.
#'
#' @return Invisibly returns a list containing the currently set options.
#' @export
#'
#' @examples
#' \dontrun{
#' # Set default mode to geoarrow
#' ddbs_options(mode = "geoarrow")
#' 
#' # Set default output to tibble
#' ddbs_options(output_type = "tibble")
#'
#' # Check current settings
#' ddbs_options()
#' }
ddbs_options <- function(output_type = NULL, mode = NULL, duckdb_storage_version = NULL) {
  
  # 1. SETTER logic
  if (!is.null(output_type)) {
    if (!output_type %in% c("sf", "tibble", "raw", "geoarrow")) {
      cli::cli_abort(paste0(
        "Invalid output_type: {.val {output_type}}. ",
        "Must be one of {.val sf}, {.val tibble}, {.val raw}, or {.val geoarrow}."
      ))
    }
    options(duckspatial.output_type = output_type)
  }

  if (!is.null(mode)) {
    if (!mode %in% c("duckspatial", "sf")) {
      cli::cli_abort(paste0(
        "Invalid mode: {.val {mode}}. ",
        "Must be one of {.val duckspatial}, or {.val sf}."
      ))
    }
    options(duckspatial.mode = mode)
  }

  if (!is.null(duckdb_storage_version)) {
    # Use internal helper to validate
    duckdb_storage_version <- match_duckdb_storage_version(duckdb_storage_version)
    options(duckspatial.duckdb_storage_version = duckdb_storage_version)
  }
  
  # 2. GETTER logic (Always retrieve current state)
  op <- list(
    duckspatial.output_type = getOption("duckspatial.output_type"),
    duckspatial.mode = getOption("duckspatial.mode"),
    duckspatial.duckdb_storage_version = getOption("duckspatial.duckdb_storage_version")
  )
  
  # If we set a value, return invisibly. If just checking, return visibly.
  if (all(is.null(output_type), is.null(mode), is.null(duckdb_storage_version))) {
    op
  } else {
    invisible(op)
  }
}


#' Report duckspatial configuration status
#'
#' Displays useful information about the current configuration, including global options
#' and the status of the default DuckDB connection.
#'
#' @return Invisibly returns a list with the current status configuration.
#' @export
#'
#' @examples
#' ddbs_sitrep()
ddbs_sitrep <- function() {
  
  cli::cli_h1("duckspatial Status Report")
  
  # 1. Global Options
  cli::cli_h2("Global Options")
  
  out_type <- getOption("duckspatial.output_type", "duckspatial_df (default)")
  mode     <- getOption("duckspatial.mode", "duckspatial (default)")
  storage_ver <- getOption("duckspatial.duckdb_storage_version", "v1.5.0")
  cli::cli_ul()
  cli::cli_li("Output Type: {.val {out_type}}")
  cli::cli_li("Mode: {.val {mode}}")
  cli::cli_li("DuckDB Storage Version: {.val {storage_ver}}")
  cli::cli_end()
  
  # 2. Connection Status
  cli::cli_h2("Default Connection")
  
  conn <- getOption("duckspatial_conn", NULL)
  
  if (is.null(conn)) {
    cli::cli_alert_warning("No default connection active (will be created on demand).")
    conn_status <- "NULL"
  } else {
    is_valid <- DBI::dbIsValid(conn)
    if (is_valid) {
      cli::cli_alert_success("Active DuckDB connection found.")
      
      # Check if memory or file
      # Generic way to check might be difficult without specific duckdb info, 
      # but dbGetInfo might reveal it or simple "Active" is enough.
      # duckdb connections don't easily reveal their path via public API usually?
      # We can try inspection if needed, but "Active" is good enough.
      conn_status <- "Active"
    } else {
      cli::cli_alert_danger("Connection object exists but is invalid (disconnected).")
      conn_status <- "Invalid"
    }
  }
  
  invisible(list(
    output_type = out_type,
    connection_status = conn_status
  ))
}

testthat::skip_on_cran()

test_that("ddbs_options sets and retrieves options", {
  # Save original options to restore after test
  op_orig <- options()
  on.exit(options(op_orig))
  
  # Check default (empty call)
  defaults <- ddbs_options()
  expect_type(defaults, "list")
  expect_true("duckspatial.output_type" %in% names(defaults))

  # Set new option
  res <- ddbs_options(output_type = "sf")
  expect_equal(getOption("duckspatial.output_type"), "sf")
  expect_equal(res$duckspatial.output_type, "sf")
  
  res <- ddbs_options(output_type = "tibble")
  expect_equal(getOption("duckspatial.output_type"), "tibble")

  # Set new extended options
  ddbs_options(output_type = "raw")
  expect_equal(getOption("duckspatial.output_type"), "raw")
  
  ddbs_options(output_type = "geoarrow")
  expect_equal(getOption("duckspatial.output_type"), "geoarrow")
  
  # Invalid option should error
  expect_error(ddbs_options(output_type = "invalid"), "Invalid output_type")
  
  # Null should do nothing
  before <- getOption("duckspatial.output_type")
  ddbs_options(output_type = NULL)
  expect_equal(getOption("duckspatial.output_type"), before)
})

test_that("ddbs_sitrep runs without error", {
  # Just check it doesn't crash and returns invisible list
  expect_no_error({
    res <- ddbs_sitrep()
  })
  expect_type(res, "list")
  expect_true("output_type" %in% names(res))
  expect_true("connection_status" %in% names(res))
})

#' Register an SF Object as an Arrow Table in DuckDB
#'
#' This function registers a Simple Features (SF) object as a temporary Arrow-backed
#' view in a DuckDB database. This is a zero-copy operation and is significantly
#' faster than `ddbs_write_table` for workflows that do not require data to be
#' permanently materialized in the database.
#'
#' @inheritParams ddbs_write_table
#' @returns TRUE (invisibly) on successful registration.
#' @export
#' @examples
#' \dontrun{
#' library(duckdb)
#' library(duckspatial)
#' library(sf)
#'
#' conn <- ddbs_create_conn("memory")
#'
#' nc <- st_read(system.file("shape/nc.shp", package="sf"), quiet = TRUE)
#'
#' ddbs_register_table(conn, nc, "nc_arrow_view")
#'
#' dbGetQuery(conn, "SELECT COUNT(*) FROM nc_arrow_view;")
#'
#' ddbs_stop_conn(conn)
#'}
ddbs_register_table <- function(
    conn,
    data,
    name,
    overwrite = FALSE,
    quiet = FALSE
) {
    # 1. Checks
    dbConnCheck(conn)
    name_list <- get_query_name(name)
    view_name <- name_list$query_name

    # Handle duckspatial_df/tbl_lazy by collecting to sf first
    data_sf <- if (inherits(data, "duckspatial_df")) {
        ddbs_collect(data, as = "sf")
    } else if (inherits(data, "tbl_lazy")) {
        dplyr::collect(data) |> sf::st_as_sf()
    } else if (inherits(data, "sf")) {
        data
    } else if (is.character(data) && length(data) == 1) {
        sf::st_read(data, quiet = TRUE)
    } else {
        cli::cli_abort(
            "{.arg data} must be an {.cls sf} object, {.cls duckspatial_df}, or a readable file path."
        )
    }

    tables_df <- ddbs_list_tables(conn)
    db_tables <- paste0(tables_df$table_schema, ".", tables_df$table_name) |>
        sub(pattern = "^main\\.", replacement = "")
    name_exists <- view_name %in% db_tables
    
    # Prefix only the table component if it's schema-qualified to avoid 
    # broken names like __raw_schema.table. Also, duckdb_register_arrow 
    # doesn't support schema-qualified names, so we use a single component.
    raw_view_name <- paste0("__raw_", gsub("\\.", "_", view_name))

    arrow_views <- try(
        duckdb::duckdb_list_arrow(conn),
        silent = TRUE
    )
    if (inherits(arrow_views, "try-error")) {
        arrow_views <- character(0)
    }
    
    arrow_exists <- view_name %in% arrow_views
    raw_exists <- raw_view_name %in% arrow_views

    if ((name_exists || arrow_exists || raw_exists) && !overwrite) {
        cli::cli_abort(
            "The provided view (or table) name is already present in the database. Please, use `overwrite = TRUE` or choose a different name."
        )
    }

    if (overwrite && (name_exists || arrow_exists || raw_exists)) {
        if (name_exists) {
            match_idx <- which(db_tables == view_name)[1]
            table_type <- tables_df$table_type[match_idx]
            drop_stmt <- if (
                !is.na(table_type) && identical(table_type, "VIEW")
            ) {
                glue::glue("DROP VIEW IF EXISTS {view_name};")
            } else {
                glue::glue("DROP TABLE IF EXISTS {view_name};")
            }
            DBI::dbExecute(conn, drop_stmt)
            if (isFALSE(quiet)) {
                cli::cli_alert_info("Existing object {view_name} dropped")
            }
        }
        
        # Also clean up any potential hidden raw arrow view
        if (raw_exists) {
            try(duckdb::duckdb_unregister_arrow(conn, raw_view_name), silent = TRUE)
        }
        
        if (arrow_exists) {
            try(
                duckdb::duckdb_unregister_arrow(conn, view_name),
                silent = TRUE
            )
        }
    }

    # Try to register geoarrow extensions when available
    try(
        DBI::dbExecute(conn, "CALL register_geoarrow_extensions();"),
        silent = TRUE
    )

    # 2. Register table
  
    ## First, split data and geometry
    df <- sf::st_drop_geometry(data_sf)
    wkb <- wk::as_wkb(sf::st_geometry(data_sf))

    ## Get original geometry column name and CRS
    geom_name <- attr(data_sf, "sf_column")
    crs_obj <- sf::st_crs(data_sf)
    
    # DuckDB accepts WKT in GEOMETRY('<crs>'), but GeoArrow metadata consumed by
    # DuckDB cannot serialize WKT here. Keep the two CRS representations separate.
    if (!is.na(crs_obj)) {
        if (!is.na(crs_obj$epsg)) {
            crs_input <- paste0("EPSG:", crs_obj$epsg)
        } else {
            crs_input <- crs_obj$wkt
        }
        geoarrow_crs <- if (!is.na(crs_obj$epsg)) crs_input else crs_obj$proj4string
        if (length(geoarrow_crs) == 0 || is.na(geoarrow_crs) || identical(geoarrow_crs, "")) {
            geoarrow_crs <- NULL
        }
    } else {
        crs_input <- NULL
        geoarrow_crs <- NULL
    }

    ## Estimate chunk size from a sample. This is needed to avoid OOM errors
    ## when creating the Arrow table for very large datasets. The chunk size is
    ## determined based on the size of the first 1000 rows
    n <- length(wkb)
    sample_n <- min(1000L, n)
    sample_bytes <- sum(vapply(unclass(wkb[seq_len(sample_n)]), length, integer(1)))
    avg_bytes_per_feature <- sample_bytes / sample_n

    ## Target ~500MB per chunk (safely under ~2GB limit)
    target_bytes <- 500L * 1024L^2
    chunk_size   <- max(1000L, floor(target_bytes / avg_bytes_per_feature))
    idx          <- split(seq_len(n), ceiling(seq_len(n) / chunk_size))

    ## Create:
    ## - Arrow table: when the dataset is small (<500MB)
    ## - Arrow Batch: when the dataset is large (>500MB)
    if (length(idx) == 1L) {
        ## Create a single Arrow table
        arrow_table <- {
            df[[geom_name]] <- geoarrow::as_geoarrow_vctr(
                wkb,
                schema = geoarrow::geoarrow_wkb(crs = geoarrow_crs)
            )
            arrow::Table$create(df)
        }
    } else {
        ## Create an Arrow RecordBatchReader for chunked processing
        batches <- lapply(idx, function(i) {
            chunk <- df[i, , drop = FALSE]
            chunk[[geom_name]] <- geoarrow::as_geoarrow_vctr(
                wkb[i],
                schema = geoarrow::geoarrow_wkb(crs = geoarrow_crs)
            )
            arrow::record_batch(chunk)
        })
        
        ## Create the table as a batch
        schema <- batches[[1]]$schema
        arrow_table <- arrow::RecordBatchReader$create(
            batches = batches, 
            schema = schema
        )
    }

    ## Register the raw Arrow table under a hidden name
    tryCatch({
        duckdb::duckdb_register_arrow(conn, raw_view_name, arrow_table)
        
        ## Wrap it in a user-facing typed view
        q_geom <- DBI::dbQuoteIdentifier(conn, geom_name)
        
        # If view_name is qualified, DuckDB doesn't allow TEMP VIEW in other schemas.
        # We use a regular VIEW if qualified, or a TEMP VIEW if not.
        view_type <- if (name_list$schema_name != "main") "VIEW" else "TEMP VIEW"

        if (!is.null(crs_input)) {
            # Escape single quotes in WKT for SQL safety
            safe_crs <- gsub("'", "''", crs_input)
            DBI::dbExecute(conn, glue::glue(
                "CREATE OR REPLACE {view_type} {view_name} AS ",
                "SELECT * EXCLUDE {q_geom}, ",
                "({q_geom}::GEOMETRY('{safe_crs}')) AS {q_geom} ",
                "FROM {raw_view_name}"
            ))
        } else {
            # No CRS, just create a direct view casting to generic GEOMETRY
            DBI::dbExecute(conn, glue::glue(
                "CREATE OR REPLACE {view_type} {view_name} AS ",
                "SELECT * EXCLUDE {q_geom}, ",
                "({q_geom}::GEOMETRY) AS {q_geom} ",
                "FROM {raw_view_name}"
            ))
        }
    }, error = function(e) {
        # Cleanup on failure to avoid stale registrations
        try(duckdb::duckdb_unregister_arrow(conn, raw_view_name), silent = TRUE)
        stop(e)
    })

    ## User feedback
    if (isFALSE(quiet)) {
        cli::cli_alert_success("Temporary view {view_name} registered")
    }

    invisible(TRUE)
}

# ddbs_register_table <- function(
#     conn,
#     data,
#     name,
#     overwrite = FALSE,
#     quiet = FALSE
# ) {
#     # 1. Checks
#     dbConnCheck(conn)
#     name_list <- get_query_name(name)
#     view_name <- name_list$query_name

#     # Handle duckspatial_df/tbl_lazy by collecting to sf first
#     data_sf <- if (inherits(data, "duckspatial_df")) {
#         ddbs_collect(data, as = "sf")
#     } else if (inherits(data, "tbl_lazy")) {
#         dplyr::collect(data) |> sf::st_as_sf()
#     } else if (inherits(data, "sf")) {
#         data
#     } else if (is.character(data) && length(data) == 1) {
#         sf::st_read(data, quiet = TRUE)
#     } else {
#         cli::cli_abort(
#             "{.arg data} must be an {.cls sf} object, {.cls duckspatial_df}, or a readable file path."
#         )
#     }

#     tables_df <- ddbs_list_tables(conn)
#     db_tables <- paste0(tables_df$table_schema, ".", tables_df$table_name) |>
#         sub(pattern = "^main\\.", replacement = "")
#     name_exists <- view_name %in% db_tables
#     arrow_views <- try(
#         duckdb::duckdb_list_arrow(conn),
#         silent = TRUE
#     )
#     arrow_exists <- if (inherits(arrow_views, "try-error")) {
#         FALSE
#     } else {
#         view_name %in% arrow_views
#     }

#     if ((name_exists || arrow_exists) && !overwrite) {
#         cli::cli_abort(
#             "The provided view (or table) name is already present in the database. Please, use `overwrite = TRUE` or choose a different name."
#         )
#     }

#     if (overwrite && (name_exists || arrow_exists)) {
#         if (name_exists) {
#             match_idx <- which(db_tables == view_name)[1]
#             table_type <- tables_df$table_type[match_idx]
#             drop_stmt <- if (
#                 !is.na(table_type) && identical(table_type, "VIEW")
#             ) {
#                 glue::glue("DROP VIEW IF EXISTS {view_name};")
#             } else {
#                 glue::glue("DROP TABLE IF EXISTS {view_name};")
#             }
#             DBI::dbExecute(conn, drop_stmt)
#             if (isFALSE(quiet)) {
#                 cli::cli_alert_info("Existing object {view_name} dropped")
#             }
#         }
#         if (arrow_exists) {
#             try(
#                 duckdb::duckdb_unregister_arrow(conn, view_name),
#                 silent = TRUE
#             )
#         }
#     }

#     # Try to register geoarrow extensions when available
#     try(
#         DBI::dbExecute(conn, "CALL register_geoarrow_extensions();"),
#         silent = TRUE
#     )

#     # 2. Register table
#     df <- sf::st_drop_geometry(data_sf)
#     wkb <- wk::as_wkb(sf::st_geometry(data_sf))

#     # Get original geometry column name
#     geom_col_name <- attr(data_sf, "sf_column")

#     # Use geoarrow to create a geoarrow vector from WKB
#     # Assign to original geometry column name instead of hardcoded "geometry"
#     df[[geom_col_name]] <- geoarrow::as_geoarrow_vctr(
#         wkb,
#         schema = geoarrow::geoarrow_wkb()
#     )

#     # Get the CRS (however, right now it doesnt recognize the CRS and defaults to GEOMETRY 
#     # without SRID, which is not ideal)
#     # data_crs <- get_geometry_type_duckdb(data_sf)
#     data_crs <- ddbs_crs(data_sf)

#     # arrow_table <- tryCatch({
#     #    arrow::Table$create(df)
#     # }, error = function(e) {
#     #    # Fallback to standard WKB (binary) if geoarrow fails
#     #    # (e.g. "NotImplemented: MakeBuilder: cannot construct builder for type geoarrow.wkb")
#     #    df[[geom_col_name]] <- wkb
#     #    arrow::Table$create(df)
#     # })
#     arrow_table <- tryCatch({
#         # Attach CRS to the geometry column via geoarrow type
#         df[[geom_col_name]] <- as_geoarrow_vctr(
#             df[[geom_col_name]],
#             type = geoarrow_wkb(crs = data_crs$input)
#         )
#         arrow::Table$create(df)
#         }, error = function(e) {
#         # Fallback: use raw WKB and embed CRS metadata manually in the field
#         df[[geom_col_name]] <- wkb
#         arrow::Table$create(df)
#     })

#     duckdb::duckdb_register_arrow(conn, view_name, arrow_table)

#     if (isFALSE(quiet)) {
#         cli::cli_alert_success("Temporary view {view_name} registered")
#     }

#     invisible(TRUE)
# }




#' Register an SF Object as an Arrow Table in DuckDB
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#'
#' `ddbs_register_vector()` was renamed to \code{\link{ddbs_register_table}}.
#'
#' @inheritParams ddbs_register_table
#' @returns TRUE (invisibly) on successful registration.
#' @export
#' @keywords internal
ddbs_register_vector <- function(
    conn,
    data,
    name,
    overwrite = FALSE,
    quiet = FALSE) {
    
    lifecycle::deprecate_soft(
        when    = "1.0.0",
        what    = "ddbs_register_vector()",
        with    = "ddbs_register_table()"
    )
    
    ddbs_register_table(
        conn = conn,
        data = data,
        name = name,
        overwrite = overwrite,
        quiet = quiet
    )
}

#' Check geometry dimensions
#'
#' Functions to check whether geometries have Z (elevation) or M (measure) dimensions
#' 
#' @template x
#' @template by_feature
#' @template new_column
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' These functions check for additional coordinate dimensions beyond X and Y:
#'
#' - `ddbs_has_z()` checks if a geometry has Z coordinates (elevation/altitude values).
#'   Geometries with Z dimension are often referred to as 3D geometries and have
#'   coordinates in the form (X, Y, Z).
#'
#' - `ddbs_has_m()` checks if a geometry has M coordinates (measure values). The M
#'   dimension typically represents a measurement along the geometry, such as distance
#'   or time, and results in coordinates of the form (X, Y, M) or (X, Y, Z, M).
#'
#' @template returns_mode
#' 
#' @examples
#' \dontrun{
#' ## load packages
#' library(dplyr)
#' library(duckspatial)
#' 
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |> 
#'   filter(ISO3_CODE != "ATA")
#' 
#' ## check if it has Z or M
#' ddbs_has_m(countries_ddbs)
#' ddbs_has_z(countries_ddbs)
#' }
#'
#' @name ddbs_has_dim
#' @rdname ddbs_has_dim
NULL






#' @rdname ddbs_has_dim
#' @export
ddbs_has_z <- function(
  x,
  by_feature = TRUE,
  new_column = "has_z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(  
    x = x,
    by_feature = by_feature,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_HasZ"
  )

}




#' @rdname ddbs_has_dim
#' @export
ddbs_has_m <- function(
  x,
  by_feature = TRUE,
  new_column = "has_m",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  
  template_new_column(  
    x = x,
    new_column = new_column,
    by_feature = by_feature,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_HasM"
  )

}







#' Force geometry dimensions
#'
#' Functions to force geometries to have specific coordinate dimensions (X, Y, Z, M)
#' 
#' @template x
#' @param var A numeric variable in `x` to be converted in the dimension specified
#' in the argument `dim`
#' @param var_z A numeric variable in `x` to be convered in `Z` dimension
#' @param var_m A numeric variable in `x` to be convered in `M` dimension
#' @param dim The dimension to add: either `"z"` (default) for
#' elevation or `"m"` for measure values.
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' These functions modify the dimensionality of geometries:
#'
#' - `ddbs_force_2d()` removes Z and M coordinates from geometries, returning only
#'   X and Y coordinates. This is useful for simplifying 3D or measured geometries
#'   to 2D.
#'
#' - `ddbs_force_3d()` forces geometries to have three dimensions. When `dim = "z"`
#'   (default), adds or retains Z coordinates (X, Y, Z). When `dim = "m"`, adds or
#'   retains M coordinates (X, Y, M). Missing values are typically set to 0. If the
#'   input geometry has a third dimension already, it will be replaced by the new one.
#'   If the input geometry has 4 dimensions, it will drop the dimension that wasn't 
#'   specified.
#'
#' - `ddbs_force_4d()` forces geometries to have all four dimensions (X, Y, Z, M).
#'   Missing Z or M values are typically set to 0.
#'
#' @template returns_mode
#' 
#' @examples
#' \dontrun{
#' ## load packages
#' library(dplyr)
#' library(duckspatial)
#' 
#' ## load data and add 2 numeric vars
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' ) |> 
#'   dplyr::filter(ISO3_CODE != "ATA") |> 
#'   ddbs_area(new_column = "area") |> 
#'   ddbs_perimeter(new_column = "perim") 
#' 
#' ## add a Z dimension
#' countries_z_ddbs <- ddbs_force_3d(countries_ddbs, "area")
#' ddbs_has_z(countries_z_ddbs)
#' 
#' ## add a M dimension as 3D (removes current Z)
#' countries_m_ddbs <- ddbs_force_3d(countries_z_ddbs, "area", "m")
#' ddbs_has_z(countries_m_ddbs)
#' ddbs_has_m(countries_m_ddbs)
#' 
#' ## add both Z and M
#' countries_zm_ddbs <- ddbs_force_4d(countries_ddbs, "area", "perim")
#' ddbs_has_z(countries_zm_ddbs)
#' ddbs_has_m(countries_zm_ddbs)
#' 
#' ## drop both ZM
#' countries_drop_ddbs <- ddbs_force_2d(countries_zm_ddbs)
#' ddbs_has_z(countries_drop_ddbs)
#' ddbs_has_m(countries_drop_ddbs)
#' }
#'
#' @name ddbs_force_dim
#' @rdname ddbs_force_dim
NULL





#' @rdname ddbs_force_dim
#' @export
ddbs_force_2d <- function(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  
  template_unary_ops(
      x = x,
      conn = conn,
      name = name,
      mode = mode,
      overwrite = overwrite,
      quiet = quiet,
      fun = "ST_Force2D",
      other_args = NULL
  )
  
}





#' @rdname ddbs_force_dim
#' @export
ddbs_force_3d <- function(
  x,
  var,
  dim = "z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Handle function-specific errors
  if (!tolower(dim) %in% c("m", "z")) cli::cli_abort("{.arg dim} must be {.val Z} or {.val M}.")
  
  # 1. Function to use
  st_fun <- glue::glue("ST_Force3D{dim}")

  
  template_unary_ops(
      x = x,
      conn = conn,
      name = name,
      mode = mode,
      overwrite = overwrite,
      quiet = quiet,
      fun = st_fun,
      other_args = var
  )
  
}





#' @rdname ddbs_force_dim
#' @export
ddbs_force_4d <- function(
  x,
  var_z,
  var_m,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {
  

  # 1. Build ST_Force4D parameters string
  force_args <- glue::glue("{var_z}, {var_m}")

  
  # 2. Pass to template
  template_unary_ops(
      x = x,
      conn = conn,
      name = name,
      mode = mode,
      overwrite = overwrite,
      quiet = quiet,
      fun = "ST_Force4D",
      other_args = force_args
  )
  
}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## nc_sf uses NAD27 CRS, which DuckDB returns as "EPSG:4267" — a string mismatch
## that breaks expect_equal. Reproject to EPSG:4326 so the CRS round-trips cleanly.
## nc still has the AREA and PERIMETER columns needed for var/var_z/var_m.
nc_sf_local    <- sf::st_transform(nc_sf, "EPSG:4326")
nc_ddbs_local  <- duckspatial::as_duckspatial_df(nc_sf_local)

## write data
duckspatial::ddbs_write_table(conn_test, nc_sf_local, "nc")


# 1. ddbs_force_2d() -----------------------------------------------------

describe("ddbs_force_2d()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_force_2d(nc_ddbs_local)
      output_sf   <- ddbs_force_2d(nc_sf_local)
      output_conn <- ddbs_force_2d("nc", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_force_2d(nc_ddbs_local, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_force_2d(nc_ddbs_local))
      expect_message(ddbs_force_2d("nc", conn = conn_test, name = "force_2d"))
      expect_message(ddbs_force_2d("nc", conn = conn_test, name = "force_2d", overwrite = TRUE))
      expect_true(ddbs_force_2d("nc", conn = conn_test, name = "force_2d2"))

      expect_no_message(ddbs_force_2d(nc_ddbs_local, quiet = TRUE))
      expect_no_message(ddbs_force_2d("nc", conn = conn_test, name = "force_2d", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_force_2d(nc_ddbs_local)
      output_tbl  <- ddbs_read_table(conn_test, "force_2d")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("strips Z dimension from 3D geometry", {
      nc_3d      <- ddbs_force_3d(nc_ddbs_local, "AREA")
      result     <- ddbs_force_2d(nc_3d, mode = "sf")
      first_geom <- sf::st_geometry(result)[[1]]
      expect_false("XYZ"  %in% class(first_geom))
      expect_false("XYZM" %in% class(first_geom))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_force_2d("nc", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_force_2d(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_force_2d(nc_ddbs_local, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_force_2d(nc_ddbs_local, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_force_2d(nc_ddbs_local, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_force_2d(x = "999", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_force_2d(nc_ddbs_local, conn = conn_test, name = c("banana", "banana")))
    })
  })
})




# 2. ddbs_force_3d() -----------------------------------------------------

describe("ddbs_force_3d()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_force_3d(nc_ddbs_local, "AREA")
      output_sf   <- ddbs_force_3d(nc_sf_local, "AREA")
      output_conn <- ddbs_force_3d("nc", "AREA", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_force_3d(nc_ddbs_local, "AREA", mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_force_3d(nc_ddbs_local, "AREA"))
      expect_message(ddbs_force_3d("nc", "AREA", conn = conn_test, name = "force_3d"))
      expect_message(ddbs_force_3d("nc", "AREA", conn = conn_test, name = "force_3d", overwrite = TRUE))
      expect_true(ddbs_force_3d("nc", "AREA", conn = conn_test, name = "force_3d2"))

      expect_no_message(ddbs_force_3d(nc_ddbs_local, "AREA", quiet = TRUE))
      expect_no_message(ddbs_force_3d("nc", "AREA", conn = conn_test, name = "force_3d", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_force_3d(nc_ddbs_local, "AREA")
      output_tbl  <- ddbs_read_table(conn_test, "force_3d")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    describe("dim parameter", {

      it("adds Z coordinate when dim = 'z' (default)", {
        result     <- ddbs_force_3d(nc_ddbs_local, "AREA", dim = "z", mode = "sf")
        first_geom <- sf::st_geometry(result)[[1]]
        expect_true("XYZ" %in% class(first_geom) || "XYZM" %in% class(first_geom))
      })

      it("adds M coordinate when dim = 'm'", {
        result     <- ddbs_force_3d(nc_ddbs_local, "AREA", dim = "m", mode = "sf")
        first_geom <- sf::st_geometry(result)[[1]]
        expect_true("XYM" %in% class(first_geom) || "XYZM" %in% class(first_geom))
      })

      it("is case-insensitive", {
        output_upper <- ddbs_force_3d(nc_ddbs_local, "AREA", dim = "Z")
        output_lower <- ddbs_force_3d(nc_ddbs_local, "AREA", dim = "z")
        expect_equal(ddbs_collect(output_upper), ddbs_collect(output_lower))
      })
    })
  })

  describe("errors", {

    it("rejects invalid dim values", {
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", dim = "w"))
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", dim = "xyz"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_force_3d("nc", "AREA", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_force_3d(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_force_3d(x = "999", "AREA", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_force_3d(nc_ddbs_local, "AREA", conn = conn_test, name = c("banana", "banana")))
    })
  })
})




# 3. ddbs_force_4d() -----------------------------------------------------

describe("ddbs_force_4d()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER")
      output_sf   <- ddbs_force_4d(nc_sf_local, "AREA", "PERIMETER")
      output_conn <- ddbs_force_4d("nc", "AREA", "PERIMETER", conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER"))
      expect_message(ddbs_force_4d("nc", "AREA", "PERIMETER", conn = conn_test, name = "force_4d"))
      expect_message(ddbs_force_4d("nc", "AREA", "PERIMETER", conn = conn_test, name = "force_4d", overwrite = TRUE))
      expect_true(ddbs_force_4d("nc", "AREA", "PERIMETER", conn = conn_test, name = "force_4d2"))

      expect_no_message(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", quiet = TRUE))
      expect_no_message(ddbs_force_4d("nc", "AREA", "PERIMETER", conn = conn_test, name = "force_4d", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER")
      output_tbl  <- ddbs_read_table(conn_test, "force_4d")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("adds both Z and M coordinates", {
      result     <- ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", mode = "sf")
      first_geom <- sf::st_geometry(result)[[1]]
      expect_true("XYZM" %in% class(first_geom))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_force_4d("nc", "AREA", "PERIMETER", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_force_4d(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_force_4d(x = "999", "AREA", "PERIMETER", conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_force_4d(nc_ddbs_local, "AREA", "PERIMETER", conn = conn_test, name = c("banana", "banana")))
    })
  })
})


## stop connection
ddbs_stop_conn(conn_test)

# skip tests on CRAN because they take too much time
# skip_if(Sys.getenv("TEST_ONE") != "")
# testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## Try to limit threads
Sys.setenv("OMP_THREAD_LIMIT" = 2)

# read polygons data from duckspatial package
countries_sf <- sf::st_read(system.file("spatial/countries.geojson", package = "duckspatial")) |> 
    sf::st_transform("EPSG:4326")
countries_sf <- subset(countries_sf, CNTR_ID %in% c("AR", "BR", "BO", "PE", "PY", "UY", "CL"))
argentina_sf <- sf::st_read(system.file("spatial/argentina.geojson", package = "duckspatial")) |> 
    sf::st_transform("EPSG:4326")
argentina_ddbs <- duckspatial::as_duckspatial_df(argentina_sf)
countries_ddbs <- duckspatial::as_duckspatial_df(countries_sf)

# read lines data
rivers_sf <- sf::st_read(system.file("spatial/rivers.geojson", package = "duckspatial")) |> 
    sf::st_transform("EPSG:3035")
rivers_ddbs <- duckspatial::ddbs_open_dataset(system.file("spatial/rivers.geojson", package = "duckspatial"))

## create points data
set.seed(42)
n <- 1000
points_sf <- data.frame(
    id = 1:n,
    x = runif(n, min = -180, max = 180),
    y = runif(n, min = -90, max = 90)
) |>
    sf::st_as_sf(coords = c("x", "y"), crs = 4326)

points_ddbs <- duckspatial::as_duckspatial_df(points_sf)

# North Carolina data from sf package - used by duckspatial_df tests
nc_sf   <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc_ddbs <- duckspatial::ddbs_open_dataset(system.file("shape/nc.shp", package = "sf"))

#' Check and create schema
#'
#' @template conn
#' @param name A character string with the name of the schema to be created
#' @template quiet
#'
#' @returns TRUE (invisibly) for successful schema creation
#' @export
#'
#' @examples
#' ## load packages
#' \dontrun{
#' library(duckspatial)
#' library(duckdb)
#'
#' ## connect to in memory database
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## create a new schema
#' ddbs_create_schema(conn, "new_schema")
#'
#' ## check schemas
#' dbGetQuery(conn, "SELECT * FROM information_schema.schemata;")
#'
#' ## disconnect from db
#' ddbs_stop_conn(conn)
#' }
ddbs_create_schema <- function(conn, name, quiet = FALSE) {

    # 1. Checks
    ## Check if connection is correct
    dbConnCheck(conn)
    assert_name(name)
    assert_logic(quiet, "quiet")
    ## Check if schema already exists
    namechar  <- DBI::dbQuoteString(conn,name)
    tmp.query <- paste0("SELECT EXISTS(SELECT 1 FROM pg_namespace WHERE nspname = ",
                        namechar, ");")
    schema    <- DBI::dbGetQuery(conn, tmp.query)[1, 1]
    ## If it exists return TRUE, otherwise, create the schema
    if (schema) {
        cli::cli_abort("Schema <{name}> already exists.")
    } else {
        DBI::dbExecute(
            conn,
            glue::glue("CREATE SCHEMA {name};")
        )

        if (isFALSE(quiet)) {
            cli::cli_alert_success("Schema {name} created")
        }
    }
    return(invisible(TRUE))

}





#' Check tables and schemas inside a database
#'
#' @template conn
#'
#' @returns `data.frame`
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' 
#' ## create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#' 
#' ## read some data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' argentina_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/argentina.geojson", 
#'   package = "duckspatial")
#' )
#' 
#' ## insert into the database
#' ddbs_write_table(conn, argentina_ddbs, "argentina")
#' ddbs_write_table(conn, countries_ddbs, "countries")
#' 
#' ## list tables in the database
#' ddbs_list_tables(conn)
#' }
ddbs_list_tables <- function(conn) {
  DBI::dbGetQuery(conn, "
      SELECT table_schema, table_name, table_type
      FROM information_schema.tables
    ")
}





#' Check first rows of the data 
#' 
#' Prints a transposed table of the first rows of a DuckDB table, similarly
#' as the S3 [dplyr::glimpse] method.
#'
#' @template conn
#' @template name
#' @template quiet
#'
#' @returns Invisibly `duckspatial_df` object
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## read data
#' argentina_sf <- ddbs_open_dataset(system.file("spatial/argentina.geojson", package = "duckspatial"))
#'
#' ## store in duckdb
#' ddbs_write_table(conn, argentina_sf, "argentina")
#'
#' ## glimpse the inserted table
#' ddbs_glimpse(conn, "argentina")
#' }
ddbs_glimpse <- function(
  conn,
  name,
  quiet = FALSE) {

  
  ## 1. Handle errors
  dbConnCheck(conn)
  assert_name(name)
  assert_logic(quiet, "quiet")


  # 2. Prepare parameters for the query

  ## 2.1. Convenient names of table and/or schema.table
  name_list <- get_query_name(name)

  ## 2.2. Get column names
  x_geom    <- get_geom_name(conn, name_list$query_name)
  no_geom_cols <- get_geom_name(conn, name_list$query_name, rest = TRUE, collapse = TRUE)

  ## 2.3. Get the CRS
  crs <- ddbs_crs(name, conn)

  # 4. Get data

  ## 4.1. Build the query and retrieve the results
  data_tbl <- DBI::dbGetQuery(conn, glue::glue("
    SELECT
    {no_geom_cols}
    ST_AsWKB({x_geom}) AS {x_geom}
    FROM {name}
    LIMIT 10;
  "))

  ## 4.2. Convert to sf
  data_sf <- convert_to_sf_wkb(
      data   = data_tbl,
      crs    = crs,
      x_geom = x_geom
  )
  
  ## 4.4. Convert sf to duckspatial_df
  result <- as_duckspatial_df(
    x        = data_sf,
    conn     = conn,
    crs      = crs,
    geom_col = x_geom
  )
  
  ## 4.5. Return glimpse.duckspatial.df() and the result
  glimpse(result)
  return(invisible(result))

}



#' Create a DuckDB connection with spatial extension
#'
#' It creates a DuckDB connection, and then it installs and loads the
#' spatial extension
#'
#' @param dbdir String. Either `"tempdir"`, `"memory"`, or a DuckDB database
#' file path with `.duckdb`, `.db`, or `.ddb` extension. Defaults to `"memory"`.
#' @template threads
#' @template memory_limit_gb
#' @param upgrade if TRUE, it upgrades the DuckDB extension to the latest version
#' @param ... Additional parameters to be passed to \code{\link[DBI]{dbConnect}}
#' @param duckdb_storage_version Storage compatibility for newly created persistent
#'   native DuckDB files (\code{.duckdb}, \code{.db}, \code{.ddb}). See
#'   \url{https://duckdb.org/docs/internals/storage} for more information on
#'   DuckDB storage versions and compatibility.
#'   \itemize{
#'     \item \code{"v1.5.0"} (\strong{Native Spatial Storage}, Default): Preserves
#'           CRS metadata in native DuckDB \code{GEOMETRY} columns. Requires
#'           DuckDB >= 1.5.0 to open the file.
#'     \item \code{"v1.0.0"} (\strong{Legacy Compatibility}): Creates
#'           files readable by older DuckDB versions (>= 1.0.0). Persists CRS
#'           metadata in duckspatial-managed column comments (a convention not
#'           recognized by other spatial software).
#'     \item \code{"latest"}: Use the highest storage version supported by your
#'           installed DuckDB engine.
#'   }
#'   Other major version strings like \code{"v1.4.0"}, \code{"v1.3.0"}, etc., are also supported.
#'
#' @returns A `duckdb_connection`
#' @export
#'
#' @examples
#' \dontrun{
#' # load packages
#' library(duckspatial)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' # create a duckdb database in disk  (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "tempdir")
#'
#' # create a connection with 1 thread and 2GB memory limit
#' conn <- ddbs_create_conn(threads = 1, memory_limit_gb = 2)
#' ddbs_stop_conn(conn)
#' }
ddbs_create_conn <- function(
  dbdir = "memory", 
  threads = NULL, 
  memory_limit_gb = NULL,
  upgrade = FALSE,
  ...,
  duckdb_storage_version = duckspatial_storage_default()) {

    duckdb_storage_version <- match_duckdb_storage_version(duckdb_storage_version)

    if (!dbdir %in% c("tempdir", "memory") && !has_duckdb_file_extension(dbdir)) {
      cli::cli_abort(
        "{.arg dbdir} should be {.val tempdir}, {.val memory}, or a file path with {.file .duckdb}, {.file .db}, or {.file .ddb} extension."
      )
    }

    assert_threads(threads)
    assert_memory_limit_gb(memory_limit_gb)

    # this creates a local database which allows DuckDB to
    # perform **larger-than-memory** workloads
    if(dbdir == 'tempdir'){
      
      db_path <- tempfile(pattern = 'duckspatial', fileext = '.duckdb')
      conn <- ddbs_open_persistent(
        db_path,
        duckdb_storage_version = duckdb_storage_version,
        read_only = FALSE,
        ...
      )
    } else if (dbdir == 'memory') {
      conn <- duckdb::dbConnect(
        duckdb::duckdb(
          dbdir = ":memory:"
          #, bigint = "integer64" ## in case the data includes big int
        ),
        geometry = "wk",
        ...
      )
    } else {
      conn <- ddbs_open_persistent(
        dbdir,
        duckdb_storage_version = duckdb_storage_version,
        read_only = FALSE,
        ...
      )
    }

    # Checks and installs the Spatial extension
    ddbs_install(conn, upgrade = upgrade, quiet = TRUE)
    ddbs_load(conn, quiet = TRUE)

    # Configure resources if requested
    ddbs_set_resources(conn, threads = threads, memory_limit_gb = memory_limit_gb)

    return(conn)
}





#' Get list of GDAL drivers and file formats
#'
#' @template conn_default
#'
#' @returns `data.frame`
#' @export
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## database setup
#' conn <- ddbs_create_conn()
#'
#' ## check drivers
#' ddbs_drivers(conn)
#' }
ddbs_drivers <- function(conn = NULL) {
  if (is.null(conn)) {
    conn <- ddbs_default_conn()
    if (is.null(conn)) {
       conn <- ddbs_create_conn(dbdir = "memory")
       on.exit(ddbs_stop_conn(conn), add = TRUE)
    }
  }
  DBI::dbGetQuery(conn, "
      SELECT * FROM ST_Drivers()
    ")
}

#' Close a DuckDB connection
#'
#' @template conn
#'
#' @returns TRUE (invisibly) for successful disconnection
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#'
#' ## create an in-memory duckdb database
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## close the connection
#' ddbs_stop_conn(conn)
#' }
ddbs_stop_conn <- function(conn) {
    # Check if connection is correct
    dbConnCheck(conn)

    # Disconnect from database and shutdown driver
    # Explicit driver shutdown is required on Windows to release file locks
    ddbs_checkpoint_if_possible(conn)
    drv <- conn@driver
    DBI::dbDisconnect(conn)
    if (inherits(drv, "duckdb_driver")) {
        duckdb::duckdb_shutdown(drv)
    }

    return(invisible(TRUE))
}

# testthat::skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")
testthat::skip_if_not_installed("sf")
testthat::skip_if_not_installed("areal")

library(sf)
library(areal)

# Data Setup
# race: Source data (polygons with data)
# wards: Target data (polygons to interpolate to)
race <- areal::ar_stl_race
wards <- areal::ar_stl_wards

## Duckdb v1.5 doesnt support the ESRI CRS
race <- sf::st_transform(race, "EPSG:3548")
wards <- sf::st_transform(wards, "EPSG:3548")

## Return sf by default
ddbs_options(mode = "sf")

# -------------------------------------------------------------------------
# Core Logic Tests (Accuracy & Semantics)
# -------------------------------------------------------------------------

test_that("ddbs_interpolate_aw matches areal::aw_interpolate (weight='sum')", {
  # Logic: weight='sum' implies mass preservation relative to target coverage.
  # Denominator = Sum of overlapping areas.
  # This is the default behavior of areal::aw_interpolate.

  # 1. Run areal
  res_areal <- areal::aw_interpolate(
    wards,
    tid = WARD,
    source = race,
    sid = GEOID,
    weight = "sum",
    output = "tibble",
    extensive = "TOTAL_E"
  )

  # 2. Run duckspatial
  res_duck <- ddbs_interpolate_aw(
    target = wards,
    source = race,
    tid = "WARD",
    sid = "GEOID",
    extensive = "TOTAL_E",
    weight = "sum",
    keep_NA = TRUE, # Matches areal default
    conn = NULL
  )

  # Compare
  res_duck_df <- sf::st_drop_geometry(res_duck)
  cmp <- merge(
    res_areal[, c("WARD", "TOTAL_E")],
    res_duck_df[, c("WARD", "TOTAL_E")],
    by = "WARD",
    suffixes = c("_areal", "_duck")
  )

  # Check 1: Per-target agreement
  expect_equal(cmp$TOTAL_E_areal, cmp$TOTAL_E_duck, tolerance = 1e-6)

  # Check 2: Mass preservation
  expect_equal(
    sum(cmp$TOTAL_E_areal, na.rm = TRUE),
    sum(cmp$TOTAL_E_duck, na.rm = TRUE),
    tolerance = 1e-6
  )
})

test_that("ddbs_interpolate_aw matches sf::st_interpolate_aw (Extensive / weight='total')", {
  # Logic: weight='total' implies strict mass preservation of source.
  # Denominator = Total area of source polygon.
  # This matches sf::st_interpolate_aw(extensive=TRUE).

  # 1. Run sf (defaults: keep_NA=FALSE)
  res_sf <- suppressWarnings(sf::st_interpolate_aw(
    x = race["TOTAL_E"],
    to = wards,
    extensive = TRUE,
    keep_NA = FALSE
  ))

  # 2. Run duckspatial
  res_duck <- ddbs_interpolate_aw(
    target = wards,
    source = race,
    tid = "WARD",
    sid = "GEOID",
    extensive = "TOTAL_E",
    weight = "total",
    keep_NA = FALSE, # Explicitly match sf default
    conn = NULL
  )

  # Match rows by ID (sf doesn't keep ID by default, so we match on WARD via rowname mapping or assume order)
  # Here we rely on the fact that both use 'wards' as base.
  # To be robust, we attach WARD ID back to sf result based on geometry or index.
  res_sf$WARD <- wards$WARD[match(row.names(res_sf), row.names(wards))]

  cmp <- merge(
    sf::st_drop_geometry(res_sf),
    sf::st_drop_geometry(res_duck),
    by = "WARD"
  )

  # Values check
  expect_equal(cmp$TOTAL_E.x, cmp$TOTAL_E.y, tolerance = 1e-6)

  # Row count check (keep_NA=FALSE should drop non-overlapping targets)
  expect_equal(nrow(res_sf), nrow(res_duck))
})

test_that("ddbs_interpolate_aw matches sf::st_interpolate_aw (Intensive)", {
  # Logic: Intensive variables (densities).
  # Denominator = Sum of overlap areas per target.
  # Logic implies weight='sum' on the target aggregation side.

  # Create a dummy intensive variable
  race$density <- race$TOTAL_E / sf::st_area(race)

  # 1. Run sf
  res_sf <- suppressWarnings(sf::st_interpolate_aw(
    x = race["density"],
    to = wards,
    extensive = FALSE
  ))

  # 2. Run duckspatial
  res_duck <- ddbs_interpolate_aw(
    target = wards,
    source = race,
    tid = "WARD",
    sid = "GEOID",
    intensive = "density",
    weight = "sum",
    conn = NULL
  )

  vals_sf <- as.numeric(res_sf$density)
  vals_duck <- as.numeric(res_duck$density)

  # Compare only non-NA values
  idx <- !is.na(vals_sf) & !is.na(vals_duck)
  expect_equal(vals_sf[idx], vals_duck[idx], tolerance = 1e-6)
})

test_that("ddbs_interpolate_aw handles Mixed Interpolation (Extensive + Intensive)", {
  # areal supports mixed interpolation in one call.
  # weight argument applies to extensive variables (sum/total).
  # intensive variables always use sum-overlap logic.
  
  # Prepare Source Data with both types
  race_mixed <- race
  race_mixed$pop_density <- race_mixed$TOTAL_E / as.numeric(sf::st_area(race_mixed))

  # 1. Run areal
  res_areal <- areal::aw_interpolate(
    wards, tid = WARD, source = race_mixed, sid = GEOID,
    weight = "sum", output = "tibble", 
    extensive = "TOTAL_E", intensive = "pop_density"
  )
  
  # 2. Run duckspatial
  res_duck <- ddbs_interpolate_aw(
    target = wards, source = race_mixed, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", intensive = "pop_density",
    weight = "sum", keep_NA = TRUE, conn = NULL
  )
  
  res_duck_df <- sf::st_drop_geometry(res_duck)
  cmp <- merge(
    res_areal[, c("WARD", "TOTAL_E", "pop_density")],
    res_duck_df[, c("WARD", "TOTAL_E", "pop_density")],
    by = "WARD", suffixes = c("_areal", "_duck")
  )
  
  # Check Extensive
  expect_equal(cmp$TOTAL_E_areal, cmp$TOTAL_E_duck, tolerance = 1e-6)
  
  # Check Intensive
  expect_equal(cmp$pop_density_areal, cmp$pop_density_duck, tolerance = 1e-6)
})

# -------------------------------------------------------------------------
# Feature & Argument Tests (keep_NA, na.rm, output types)
# -------------------------------------------------------------------------

test_that("ddbs_interpolate_aw respects keep_NA=FALSE", {
  # Create a target that definitely does NOT overlap with the source
  # Source: St. Louis (race)
  # Target: St. Louis (wards) + 1 dummy polygon far away
  dummy_poly <- sf::st_sfc(sf::st_polygon(list(rbind(c(0,0), c(1,0), c(1,1), c(0,1), c(0,0)))))
  sf::st_crs(dummy_poly) <- sf::st_crs(wards)
  dummy_row <- wards[1, ]
  dummy_row$geometry <- dummy_poly
  dummy_row$WARD <- "DUMMY"
  wards_expanded <- rbind(wards, dummy_row)

  # 1. keep_NA = FALSE (Inner Join behavior)
  res_drop <- ddbs_interpolate_aw(
    target = wards_expanded, source = race, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", keep_NA = FALSE
  )

  # The dummy row should be gone
  expect_false("DUMMY" %in% res_drop$WARD)
  expect_equal(nrow(res_drop), nrow(wards)) # Should match original overlapping count

  # 2. keep_NA = TRUE (Left Join behavior)
  res_keep <- ddbs_interpolate_aw(
    target = wards_expanded, source = race, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", keep_NA = TRUE
  )

  # The dummy row should be present, with NA/0 for data
  expect_true("DUMMY" %in% res_keep$WARD)
  expect_equal(nrow(res_keep), nrow(wards_expanded))
})

## TODO - THIS TEST DOES NOT PASS - WHY??
test_that("ddbs_interpolate_aw respects na.rm=TRUE", {
  testthat::skip()
  # Inject NAs into source data
  race_na <- race
  race_na$TOTAL_E[1:5] <- NA # First 5 rows are NA

  # 1. Run with na.rm=TRUE
  # Source rows with NA should be completely ignored during interpolation
  res_na_rm <- ddbs_interpolate_aw(
    target = wards, source = race_na, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", weight = "sum",
    na.rm = TRUE
  )

  # 2. Run with na.rm=FALSE
  # NAs propagate (results for overlapping targets will be NA or skewed depending on SQL sum behavior with nulls)
  # In SQL, SUM(col) usually ignores NULLs, but if ALL are NULL, result is NULL.
  # However, na.rm=TRUE filters the geometry entirely from the overlap calculation,
  # whereas na.rm=FALSE keeps the geometry but with NULL value.
  # For weighted interpolation, keeping geometry with NULL value affects denominators!
  res_keep_na <- ddbs_interpolate_aw(
    target = wards, source = race_na, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", weight = "sum",
    na.rm = FALSE
  )

  # Comparison:
  # Check if results differ. Specifically, if we remove source polygons,
  # the denominators (total area) might change for 'sum' weight if overlapping.
  # Just checking they are not identical validates the switch works.
  expect_false(identical(res_na_rm$TOTAL_E, res_keep_na$TOTAL_E))

  # Expect fewer NAs (or different values) when we actively remove bad source rows vs letting them sit in the join.
})

test_that("ddbs_interpolate_aw handles projection via join_crs", {
  conn <- ddbs_create_conn()
  ddbs_write_table(conn, wards, "wards_tbl", overwrite = TRUE)
  ddbs_write_table(conn, race, "race_tbl", overwrite = TRUE)

  # Run with explicit reprojection to Mercator (3857)
  res_proj <- ddbs_interpolate_aw(
    target = "wards_tbl",
    source = "race_tbl",
    tid = "WARD",
    sid = "GEOID",
    extensive = "TOTAL_E",
    weight = "sum",
    join_crs = 3857,
    conn = conn
  )

  expect_s3_class(res_proj, "sf")
  expect_true("TOTAL_E" %in% names(res_proj))

  # Total sum check
  total_pop <- sum(race$TOTAL_E, na.rm = TRUE)
  res_pop <- sum(res_proj$TOTAL_E, na.rm = TRUE)
  expect_equal(res_pop, total_pop, tolerance = 0.05)

  ddbs_stop_conn(conn)
})

test_that("ddbs_interpolate_aw handles output to table", {
  conn <- ddbs_create_conn()

  ddbs_interpolate_aw(
    target = wards,
    source = race,
    tid = "WARD",
    sid = "GEOID",
    extensive = "TOTAL_E",
    name = "result_table",
    conn = conn
  )

  expect_true("result_table" %in% DBI::dbListTables(conn))

  # Check content
  res <- ddbs_read_table(conn, "result_table")
  expect_true("TOTAL_E" %in% names(res))
  expect_equal(nrow(res), nrow(wards))

  ddbs_stop_conn(conn)
})

test_that("ddbs_interpolate_aw throws errors for missing arguments", {
  expect_error(
    ddbs_interpolate_aw(wards, race, tid = "WARD"),
    "sid"
  )
  expect_error(
    ddbs_interpolate_aw(wards, race, sid = "GEOID"),
    "tid"
  )
  expect_error(
    ddbs_interpolate_aw(wards, race, tid = "WARD", sid = "GEOID"),
    "extensive"
  )
})

# -------------------------------------------------------------------------
# Edge Cases (CRS Mismatch, Disjoint)
# -------------------------------------------------------------------------

test_that("ddbs_interpolate_aw errors on CRS mismatch if join_crs is NULL", {
  # Create version with different CRS (Mercator)
  wards_3857 <- sf::st_transform(wards, 3857)
  
  # 1. Should error because wards is 3857 and race is NAD83, and no join_crs provided
  # assert_crs is called internally which triggers "different" message
  expect_error(
    ddbs_interpolate_aw(
      target = wards_3857, source = race, tid = "WARD", sid = "GEOID",
      extensive = "TOTAL_E"
    ),
    "different"
  )
  
  # 2. Should succeed if we provide join_crs (forcing reprojection of both to common CRS)
  expect_no_error(
    ddbs_interpolate_aw(
      target = wards_3857, source = race, tid = "WARD", sid = "GEOID",
      extensive = "TOTAL_E", join_crs = 3857
    )
  )
})

test_that("ddbs_interpolate_aw handles disjoint data correctly", {
  # Create a source that is miles away from target
  source_far <- race
  
  # shift coordinates by 1,000,000 meters (1000km) to ensure disjointness
  # (100m was not enough for projected data)
  shifted_geom <- sf::st_geometry(source_far) + c(1000000, 1000000)
  shifted_geom <- sf::st_set_crs(shifted_geom, sf::st_crs(source_far))
  sf::st_geometry(source_far) <- shifted_geom
  
  # 1. keep_NA = TRUE -> All target rows returned, but values should be NA (no overlap)
  res <- ddbs_interpolate_aw(
    target = wards, source = source_far, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", keep_NA = TRUE
  )
  
  expect_equal(nrow(res), nrow(wards))
  # All values should be NA because there is no overlap
  expect_true(all(is.na(res$TOTAL_E)))
  
  # 2. keep_NA = FALSE -> Empty result (targets with no overlap are dropped)
  res_drop <- ddbs_interpolate_aw(
    target = wards, source = source_far, tid = "WARD", sid = "GEOID",
    extensive = "TOTAL_E", keep_NA = FALSE
  )
  expect_equal(nrow(res_drop), 0)
})


# -------------------------------------------------------------------------
# New Feature Tests (Validation, Conflicts)
# -------------------------------------------------------------------------


test_that("ddbs_interpolate_aw enforces strict logic validation", {
  # 1. Error if weight='total' is used with intensive variables
  # (Mathematically invalid to sum densities over total areas)
  expect_error(
    ddbs_interpolate_aw(
      target = wards, source = race, tid = "WARD", sid = "GEOID",
      intensive = "density", # Assume density exists or checks pass before logic
      weight = "total"
    ),
    "intensive variables must use" # Expecting the error message from cli_abort
  )
  
  # 2. Error if output argument is invalid
  expect_error(
    ddbs_interpolate_aw(
      target = wards, source = race, tid = "WARD", sid = "GEOID",
      extensive = "TOTAL_E", 
      mode = "geojson"
    )
  )
})

test_that("ddbs_interpolate_aw warns on Column Name Conflicts", {
  # Create a target that already has the column "TOTAL_E"
  wards_conflict <- wards
  wards_conflict$TOTAL_E <- 99999
  
  # Logic: The function should Warn the user, but proceed by 
  # prioritizing the Interpolated value over the original Target value.
  expect_warning(
    res_conflict <- ddbs_interpolate_aw(
      target = wards_conflict, source = race, tid = "WARD", sid = "GEOID",
      extensive = "TOTAL_E"
    ),
    "conflict detected"
  )
  
  # Check that the value is the interpolated value (~4000ish), not the dummy 99999
  # We check the first row that isn't NA
  val <- res_conflict$TOTAL_E[!is.na(res_conflict$TOTAL_E)][1]
  expect_lt(val, 10000) 
})

# -------------------------------------------------------------------------
# Advanced CRS Edge Cases (Missing CRS)
# -------------------------------------------------------------------------
test_that("ddbs_interpolate_aw handles Missing CRS inputs appropriately", {
  # Setup data with NO CRS
  wards_no_crs <- wards
  sf::st_crs(wards_no_crs) <- NA
  
  race_no_crs <- race
  sf::st_crs(race_no_crs) <- NA
  
  # Case 1: Both missing CRS, NO join_crs requested.
  # This should pass (assuming raw coordinates overlap).
  expect_no_error(
    res <- ddbs_interpolate_aw(
      target = wards_no_crs, source = race_no_crs, 
      tid = "WARD", sid = "GEOID", extensive = "TOTAL_E"
    )
  )
  expect_true(is.na(sf::st_crs(res)))
  
  # Case 2: Missing CRS, but join_crs IS requested.
  # This MUST fail because we cannot project something if we don't know what it is.
  expect_error(
    ddbs_interpolate_aw(
      target = wards_no_crs, source = race_no_crs, 
      tid = "WARD", sid = "GEOID", extensive = "TOTAL_E",
      join_crs = 5070
    ),
    "Cannot transform" # Matches: "Target CRS unknown... Cannot transform..."
  )
  
  # Case 3: One has CRS, one does not (No join_crs).
  # This should error due to mismatch.
  expect_error(
    ddbs_interpolate_aw(
      target = wards_no_crs, source = race, 
      tid = "WARD", sid = "GEOID", extensive = "TOTAL_E"
    ),
    "different" # Matches: "CRS mismatch: One input has a defined CRS..."
  )
})

## restore
ddbs_options(mode = "duckspatial")

# testthat::skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()

test_that("crs_to_sql handles various inputs correctly", {
  # Numeric input (EPSG code)
  expect_equal(crs_to_sql(4326), "'EPSG:4326'")
  expect_equal(crs_to_sql(3857.9), "'EPSG:3857'") # casts to integer

  # Character input (WKT or PROJ string)
  expect_equal(crs_to_sql("EPSG:4326"), "'EPSG:4326'")
  
  # Character input with single quotes (SQL escaping)
  wkt_quote <- "PROJCS['Test', ...]"
  expected_quote <- "'PROJCS[''Test'', ...]'"
  expect_equal(crs_to_sql(wkt_quote), expected_quote)

  # sf crs objects
  # We construct a mock CRS object to test logic independently of PROJ/GDAL versions
  
  # Case 1: CRS with EPSG code (Standard SF object structure)
  crs_epsg <- structure(list(epsg = 4326, wkt = "GEOGCS[...]"), class = "crs")
  expect_equal(crs_to_sql(crs_epsg), "'EPSG:4326'")

  # Case 2: CRS without EPSG code (WKT only)
  # Mimic an sf object where epsg is NA
  wkt_raw <- "GEOGCS['TestCRS', ...]"
  crs_wkt <- structure(list(epsg = NA, wkt = wkt_raw), class = "crs")
  
  sql_out <- crs_to_sql(crs_wkt)
  # Should return quoted string with escaped internal quotes
  expected_wkt_sql <- "'GEOGCS[''TestCRS'', ...]'"
  expect_equal(sql_out, expected_wkt_sql)

  # NULL/NA inputs
  expect_equal(crs_to_sql(NULL), "NULL")
  expect_equal(crs_to_sql(NA), "NULL")
})

test_that("assert_col_exists validates database columns", {
  skip_if_not_installed("duckdb")
  
  # Setup dummy connection and table
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn), add = TRUE)
  
  DBI::dbExecute(conn, "CREATE TABLE test_utils (id INTEGER, val_a DOUBLE, val_b DOUBLE)")

  # Success cases
  expect_no_error(assert_col_exists(conn, "test_utils", "id", "my_func"))
  expect_no_error(assert_col_exists(conn, "test_utils", c("id", "val_a"), "my_func"))

  # Failure cases
  # 1. Single missing column
  expect_error(
    assert_col_exists(conn, "test_utils", "missing_col", "test_ref"),
    "missing_col"
  )
  
  # 2. Partial missing columns
  expect_error(
    assert_col_exists(conn, "test_utils", c("id", "missing_one"), "test_ref"),
    "missing_one"
  )
})

testthat::skip_on_cran()

# Tests for connection management

test_that("ddbs_create_conn accepts only supported DuckDB file extensions", {
  for (ext in c(".duckdb", ".db", ".ddb")) {
    db_path <- tempfile(fileext = ext)
    on.exit(unlink(db_path), add = TRUE)

    # 1. Create and write using public API
    conn <- ddbs_create_conn(dbdir = db_path)
    DBI::dbExecute(conn, "CREATE TABLE extension_check AS SELECT 1 AS value")
    ddbs_stop_conn(conn)

    # 2. Reopen and verify using public API
    conn_reopened <- ddbs_create_conn(dbdir = db_path)
    expect_equal(
      DBI::dbGetQuery(conn_reopened, "SELECT value FROM extension_check")$value,
      1
    )
    ddbs_stop_conn(conn_reopened)
  }

  expect_error(
    ddbs_create_conn(tempfile(fileext = ".txt")),
    "duckdb.*db.*ddb"
  )
})

test_that("cross-connection filtering works with proper fallback strategies", {
  skip_if_not_installed("sf")

  # Setup: Two distinct connections
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })

  # Load data using internal package data
  countries_path <- system.file("spatial/countries.geojson", package = "duckspatial")
  ds1 <- ddbs_open_dataset(countries_path, conn = conn1)
  ds2 <- ddbs_open_dataset(countries_path, conn = conn2)

  # CASE 1: Direct View Import (Strategy 1)
  # This should trigger warnings about cross-connection imports
  expect_warning(
    expect_warning(
      res1 <- ddbs_filter(ds1, ds2),
      "come from different DuckDB connections"
    ),
    "Importing.*to the target connection"
  )
  res_df1 <- collect(res1)
  expect_true(nrow(res_df1) > 0)

  # CASE 2: Transformed Query Import (Strategy 3 via Collect)
  # Filter ds2 so it becomes a query, not a direct view
  # This forces fallback to Strategy 3 because dbplyr queries can't be SQL-recreated across conns easily
  ds2_mod <- ds2 |> dplyr::filter(CNTR_ID == "AR")

  expect_warning(
    expect_warning(
      expect_warning(
        res2 <- ddbs_filter(ds1, ds2_mod),
        "come from different DuckDB connections"
      ),
      "Importing.*to the target connection"
    ),
    "Imported via collection"
  )

  res_df2 <- collect(res2)
  # Argentina + neighbors should be returned
  expect_true(nrow(res_df2) >= 1)
  expect_true(nrow(res_df2) < 257) # Should be a subset
})


test_that("ddbs_crs works on character tables without CRS column using view analysis", {
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn))
  
  nc_path <- system.file("shape/nc.shp", package = "sf")
  
  # Create a view manually that mimics one without crs_duckspatial
  # (Standard ST_Read view)
  view_name <- "raw_view"
  DBI::dbExecute(conn, glue::glue("
    CREATE VIEW {view_name} AS SELECT * FROM ST_Read('{nc_path}')
  "))
  
  # Check if ddbs_crs can auto-detect it
  crs <- ddbs_crs(view_name, conn = conn)
  
  # Should find EPSG:4267
  expect_false(is.na(crs))
  expect_equal(crs$epsg, 4267)
})

test_that("ddbs_open_dataset does not leak connections on error with persistent files", {
  db_path <- tempfile(fileext = ".duckdb")
  on.exit(unlink(db_path), add = TRUE)
  
  # Create a dummy duckdb file with one table
  local({
    conn <- ddbs_create_conn(db_path)
    DBI::dbExecute(conn, "CREATE TABLE my_table AS SELECT 1 AS id")
    ddbs_stop_conn(conn)
  })
  
  # Try to open a NON-EXISTENT layer. This triggers the error inside tryCatch
  # after the connection is opened.
  expect_error(
    ddbs_open_dataset(db_path, layer = "non_existent_layer"),
    "not present in DuckDB database"
  )
  
  # If the connection leaked, the file might be locked. 
  # Attempting to delete it should succeed if closed.
  # (On Windows this is a strong check; on Unix it is less so but still good practice)
  expect_true(unlink(db_path) == 0)
  expect_false(file.exists(db_path))
})

testthat::skip_on_cran()

# Test ddbs_filter and mixed inputs
test_that("ddbs_filter works with mixed inputs", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  # Setup connections/data
  conn <- ddbs_create_conn() # separate connection for isolation
  on.exit(ddbs_stop_conn(conn))
  
  # Prepare inputs
  # 1. sf object
  p_sf <- points_sf[1:10, ]
  c_sf <- countries_sf[1:5, ]
  
  # 2. duckspatial_df (lazy) in the same connection
  # (Requires writing first)
  ddbs_write_table(conn, p_sf, "points_lazy")
  p_lazy <- as_duckspatial_df(
    x = "points_lazy",
    conn = conn,
    geom_col = "geometry", 
    crs = sf::st_crs(p_sf)
  )
    
  # TEST 1: filter(sf, sf) -> covered implicitly by other tests, but good to have
  res1 <- ddbs_filter(p_sf, c_sf, mode = "sf")
  expect_s3_class(res1, "sf")
  
  # TEST 2: filter(duckspatial_df, sf)
  # Should treat duckspatial_df as target connection origin
  res2 <- ddbs_filter(p_lazy, c_sf)
  # Default output is now duckspatial_df (lazy)
  expect_s3_class(res2, "duckspatial_df")
  expect_true(inherits(res2, "tbl_lazy"))
  
  # VERIFY CRS RETENTION (Regression Test)
  expect_false(is.na(sf::st_crs(res2)))
  expect_true(sf::st_crs(res2) == sf::st_crs(p_sf))
  
  # Test result correctness
  res2_collected <- collect(res2)
  expect_s3_class(res2_collected, "sf")
  expect_equal(nrow(res2_collected), nrow(res1))
  
  # TEST 3: filter(sf, duckspatial_df)
  # Should use duckspatial_df connection as target
  res3 <- ddbs_filter(p_sf, p_lazy) # Point in Point (just for type testing)
  expect_s3_class(res3, "duckspatial_df")
  expect_false(is.na(sf::st_crs(res3))) 
  
  # TEST 4: filter(duckspatial_df, duckspatial_df)
  ddbs_write_table(conn, c_sf, "countries_lazy")
  c_lazy <- as_duckspatial_df("countries_lazy", conn)
    
  res4 <- ddbs_filter(p_lazy, c_lazy)
  expect_s3_class(res4, "duckspatial_df")
  expect_false(is.na(sf::st_crs(res4)))
})

test_that("ddbs_join works with mixed inputs and cross-connection", {
  conn1 <- ddbs_create_conn()
  conn2 <- ddbs_create_conn()
  on.exit({
    ddbs_stop_conn(conn1)
    ddbs_stop_conn(conn2)
  })
  
  p_sf <- points_sf[1:10, ]
  c_sf <- countries_sf[1:5, ]
  
  # Setup: p_lazy in conn1
  ddbs_write_table(conn1, p_sf, "points_c1")
  # p_lazy1 <- dplyr::tbl(conn1, "points_c1") |>
  #   as_duckspatial_df(geom_col="geometry", crs=sf::st_crs(4326))
  p_lazy1 <- as_duckspatial_df(conn = conn1, "points_c1")

  # Setup: c_lazy in conn2
  ddbs_write_table(conn2, c_sf, "countries_c2")
  # c_lazy2 <- dplyr::tbl(conn2, "countries_c2") |>
  #   as_duckspatial_df(geom_col="geometry", crs=sf::st_crs(4326))
  c_lazy2 <- as_duckspatial_df(conn = conn2, "countries_c2")
  
  # Cross-connection join: p_lazy1 (conn1) LEFT JOIN c_lazy2 (conn2)
  # Should import c_lazy2 into conn1
  suppressWarnings(expect_warning(
    res <- ddbs_join(p_lazy1, c_lazy2, join = "intersects"),
    "connection are different"
  ))
  
  # Result should be lazy table in conn1
  expect_true(inherits(res, "duckspatial_df"))
  expect_identical(dbplyr::remote_con(res), conn1)
  
  # VERIFY CRS RETENTION
  expect_false(is.na(sf::st_crs(res)))
  expect_true(sf::st_crs(res) == sf::st_crs(p_sf))
  
  # Collect to verify
  res_df <- collect(res)
  expect_s3_class(res_df, "sf")
  
  # Test with function predicate (Regression check)
   suppressWarnings(expect_warning(
    res_func <- ddbs_join(p_lazy1, c_lazy2, join = "intersects"),
    "connection are different"
  ))
  expect_s3_class(res_func, "duckspatial_df")
})

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## create a random variable for quad keys
rand_sf <- sf::st_sample(argentina_sf, 100) |> sf::st_as_sf()
rand_sf["var"] <- runif(100)
rand_ddbs <- as_duckspatial_df(rand_sf)

## write data
duckspatial::ddbs_write_table(conn_test, rand_sf, "points")


# 1. ddbs_quadkey() --------------------------------------------------------

## 1.1. Expected behaviour -------------------

## expected behaviour
## - CHECK 1.1: works on all input formats
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, tbl, raster)
## - CHECK 1.3: messages work
## - CHECK 1.4: "var" argument works
## - CHECK 1.5: "fun" argument works
## - CHECK 1.6: "background" argument works
describe("ddbs_quadkey()", {
  
  ### EXPECTED BEHAVIOUR
  
  describe("expected behavior", {
    
    it("works on all formats", {
      output_ddbs <- ddbs_quadkey(rand_ddbs, level = 1)
      output_sf   <- ddbs_quadkey(rand_sf, level = 1)
      output_conn <- ddbs_quadkey("points", conn = conn_test, level = 1)
      
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })
    
    it("returns different output formats (duckspatial_df, tilexy, raster)", {
      output_ddbs   <- ddbs_quadkey(rand_ddbs, level = 5)
      output_tilexy <- ddbs_quadkey(rand_ddbs, level = 5, output = "tilexy")
      output_raster <- ddbs_quadkey(rand_ddbs, level = 5, output = "raster")
      
      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_tilexy, "tbl_df")
      expect_s4_class(output_raster, "SpatRaster")
      expect_true(all(terra::values(output_raster, mat = FALSE, na.rm = TRUE) == 1))
    })
    
    it("writes tables to the database", {
      output <- ddbs_quadkey("points", conn = conn_test, name = "quadkey2")
      expect_true(output)
    })
    
    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_quadkey(rand_ddbs, level = 1))
      expect_message(ddbs_quadkey("points", conn = conn_test, name = "quadkey"))
      expect_message(ddbs_quadkey("points", conn = conn_test, name = "quadkey", overwrite = TRUE))
      
      expect_no_message(ddbs_quadkey(rand_ddbs, quiet = TRUE))
      expect_no_message(ddbs_quadkey("points", conn = conn_test, name = "quadkey", overwrite = TRUE, quiet = TRUE))
    })
    
    it("field argument aggregates variable values correctly", {
      output_raster <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", output = "raster")
      
      expect_false(all(terra::values(output_raster, mat = FALSE, na.rm = TRUE) == 1))
    })
    
    it("fun argument controls aggregation function (min)", {
      output_raster_min <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", fun = "min", output = "raster")
      output_raster_max <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", fun = "max", output = "raster")
      
      expect_gt(terra::minmax(output_raster_max)[2], terra::minmax(output_raster_min)[2])
    })
    
    it("fun argument supports mean aggregation", {
      output_raster_mean <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", fun = "mean", output = "raster")
      
      expect_s4_class(output_raster_mean, "SpatRaster")
      expect_false(all(terra::values(output_raster_mean, mat = FALSE, na.rm = TRUE) == 1))
    })
    
    it("fun argument supports sum aggregation", {
      output_raster_sum <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", fun = "sum", output = "raster")
      
      expect_s4_class(output_raster_sum, "SpatRaster")
      expect_false(all(terra::values(output_raster_sum, mat = FALSE, na.rm = TRUE) == 1))
    })
    
    it("background argument fills empty cells with specified value", {
      output_raster_bg <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", background = 0, output = "raster")
      
      expect_equal(terra::minmax(output_raster_bg)[1], 0)
      expect_false(NA %in% terra::values(output_raster_bg, mat = FALSE))
    })
    
    it("background argument supports different values", {
      output_raster_bg_neg <- ddbs_quadkey(rand_ddbs, level = 5, field = "var", background = -999, output = "raster")
      
      expect_equal(terra::minmax(output_raster_bg_neg)[1], -999)
    })
    
    it("level argument controls quadkey zoom level", {
      output_level_1 <- ddbs_quadkey(rand_ddbs, level = 1, field = "var")
      output_level_5 <- ddbs_quadkey(rand_ddbs, level = 5, field = "var")
      
      expect_lt(nrow(ddbs_collect(output_level_1)), nrow(ddbs_collect(output_level_5)))
    })
  })
  
  ### EXPECTED ERRORS
  
  describe("errors", {
    
    it("requires connection when using table names", {
      expect_error(ddbs_quadkey(x = "999", conn = NULL))
    })
    
    it("only works with point geometries", {
      expect_error(ddbs_quadkey(argentina_ddbs))
      expect_error(ddbs_quadkey(rivers_ddbs))
    })
    
    it("validates level argument range (1-23)", {
      expect_error(ddbs_quadkey(points_ddbs, level = 0))
      expect_error(ddbs_quadkey(points_ddbs, level = 100))
    })
    
    it("validates level argument type", {
      expect_error(ddbs_quadkey(points_ddbs, level = "10"))
      expect_error(ddbs_quadkey(points_ddbs, level = FALSE))
    })
    
    it("validates field argument type", {
      expect_error(ddbs_quadkey(points_ddbs, field = 2))
      expect_error(ddbs_quadkey(points_ddbs, field = TRUE))
    })
    
    it("validates fun argument type", {
      expect_error(ddbs_quadkey(points_ddbs, fun = TRUE, output = "raster"))
      expect_error(ddbs_quadkey(points_ddbs, fun = 27, output = "raster"))
    })
    
    it("validates output argument value", {
      expect_error(ddbs_quadkey(points_ddbs, fun = "mean", output = "banana"))
    })
    
    it("validates x argument type", {
      expect_error(ddbs_quadkey(x = 999))
      expect_error(ddbs_quadkey(x = "999", conn = conn_test))
    })
    
    it("validates conn argument type", {
      expect_error(ddbs_quadkey(points_ddbs, conn = 999))
    })
    
    it("validates overwrite argument type", {
      expect_error(ddbs_quadkey(points_ddbs, overwrite = 999))
    })
    
    it("validates quiet argument type", {
      expect_error(ddbs_quadkey(points_ddbs, quiet = 999))
    })
    
    it("requires name to be single character string", {
      expect_error(ddbs_quadkey(points_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
})

## stop connection
ddbs_stop_conn(conn_test)

testthat::skip_on_cran()

test_that("ddbs_temp_conn: default in-memory connection", {
  conn <- ddbs_temp_conn()
  expect_true(DBI::dbIsValid(conn))
  # dbdir is NULL for in-memory connections in this environment
  expect_true(is.null(DBI::dbGetInfo(conn)$dbdir) || DBI::dbGetInfo(conn)$dbdir == ":memory:")
  
  # Auto-close check
  get_conn <- function() {
    c <- ddbs_temp_conn()
    c
  }
  ret_conn <- get_conn()
  expect_false(DBI::dbIsValid(ret_conn))
})

test_that("ddbs_temp_conn: file-based temporary connection", {
  conn <- ddbs_temp_conn(file = TRUE)
  expect_true(DBI::dbIsValid(conn))
  
  db_file <- attr(conn, "db_file")
  expect_true(is.character(db_file))
  expect_true(file.exists(db_file))
  
  # Cleanup check
  get_file_status <- function() {
    c <- ddbs_temp_conn(file = TRUE)
    f <- attr(c, "db_file")
    list(conn = c, file = f)
  }
  status <- get_file_status()
  expect_false(DBI::dbIsValid(status$conn))
  expect_false(file.exists(status$file))
})

test_that("ddbs_temp_conn: custom file path", {
  custom_path <- tempfile(fileext = ".myduck")
  conn <- ddbs_temp_conn(file = custom_path)
  expect_true(DBI::dbIsValid(conn))
  expect_equal(attr(conn, "db_file"), custom_path)
  expect_true(file.exists(custom_path))
  
  # Manual cleanup trigger (by exiting scope)
  # Use a separate path to avoid locking issues on Windows with the still-open 'conn'
  custom_path_cleanup <- tempfile(fileext = ".myduck_cleanup")
  test_custom_cleanup <- function(path) {
    c <- ddbs_temp_conn(file = path)
  }
  test_custom_cleanup(custom_path_cleanup)
  expect_false(file.exists(custom_path_cleanup))
  
  # Ensure we close the original connection too, safely
  if (DBI::dbIsValid(conn)) DBI::dbDisconnect(conn, shutdown = TRUE)
})

test_that("ddbs_temp_conn: custom path with cleanup = FALSE", {
  custom_path <- tempfile(fileext = ".duckdb")
  
  test_no_cleanup <- function(path) {
    c <- ddbs_temp_conn(file = path, cleanup = FALSE)
    DBI::dbExecute(c, "CREATE TABLE t(id INT)")
  }
  
  test_no_cleanup(custom_path)
  
  # File should STILL exist
  expect_true(file.exists(custom_path))
  expect_false(file.exists(paste0(custom_path, ".wal")))
  
  # Verify we can connect to it
  c2 <- ddbs_create_conn(custom_path)
  expect_true("t" %in% DBI::dbListTables(c2))
  ddbs_stop_conn(c2)
  
  # Clean up manually
  unlink(custom_path)
})

test_that("ddbs_temp_conn: read_only file connection", {
  # Create a file first
  base_path <- tempfile(fileext = ".base")
  c1 <- DBI::dbConnect(duckdb::duckdb(), dbdir = base_path)
  DBI::dbExecute(c1, "CREATE TABLE t(id INT)")
  DBI::dbDisconnect(c1, shutdown = TRUE)
  
  # Use ddbs_temp_conn as read-only
  test_ro <- function(path) {
    c <- ddbs_temp_conn(file = path, read_only = TRUE, cleanup = FALSE)
    expect_error(DBI::dbExecute(c, "INSERT INTO t VALUES (1)"), "read-only")
  }
  test_ro(base_path)
  
  unlink(base_path)
})

test_that("ddbs_temp_conn: read_only with file=TRUE creates file first (regression)", {
  # Regression test: Previously failed because DuckDB can't open non-existent
  # files in read-only mode. The fix creates the file first, then opens as read-only.
  test_tempfile_readonly <- function() {
    conn <- ddbs_temp_conn(file = TRUE, read_only = TRUE, cleanup = FALSE)
    
    # Verify connection is valid
    expect_true(DBI::dbIsValid(conn))
    
    # Verify it's actually read-only by attempting a write operation
    expect_error(
      DBI::dbExecute(conn, "CREATE TABLE should_fail (id INT)"),
      "read-only"
    )
    
    # Verify spatial extension is loaded and works
    # result <- DBI::dbGetQuery(conn, "SELECT ST_Point(0, 0) as geom;")
    result <- DBI::dbGetQuery(conn, "SELECT ST_AsWKB(ST_Point(0, 0)) as geom;")
    expect_equal(nrow(result), 1)
    
    # Cleanup
    db_file <- attr(conn, "db_file")
    DBI::dbDisconnect(conn, shutdown = TRUE)
    unlink(db_file)
  }
  
  test_tempfile_readonly()
})

testthat::skip_on_cran()

test_that("ddbs_collect supports all output formats", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn))
  
  # Setup data
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)[1:5,]
  ddbs_write_table(conn, nc_sf, "nc_data")
  
  nc_lazy <- as_duckspatial_df("nc_data", conn)
    
  # 1. Default (sf)
  res_sf <- ddbs_collect(nc_lazy)
  expect_s3_class(res_sf, "sf")
  expect_s3_class(res_sf$geometry, "sfc")
  expect_equal(nrow(res_sf), 5)
  
  # 2. Tibble (no geometry vs geometry dropped?)
  # The implementation of collect.duckspatial_df(as="tibble") drops geometry if it exists in lazy table
  res_tbl <- ddbs_collect(nc_lazy, as = "tibble")
  expect_s3_class(res_tbl, "tbl_df")
  expect_false(inherits(res_tbl, "sf"))
  expect_false("geometry" %in% names(res_tbl)) # Should be dropped by logic
  
  # 3. Raw (WKB)
  res_raw <- ddbs_collect(nc_lazy, as = "raw")
  expect_s3_class(res_raw, "tbl_df")
  expect_true("geometry" %in% names(res_raw))
  # Check content is raw list or blob
  expect_true(is.list(res_raw$geometry))
  expect_true(is.raw(res_raw$geometry[[1]]) || inherits(res_raw$geometry[[1]], "blob"))
  
  # 4. GeoArrow
  # Only if geoarrow package is available (optional dependency)
  if (requireNamespace("geoarrow", quietly = TRUE)) {
    res_ga <- ddbs_collect(nc_lazy, as = "geoarrow")
    expect_s3_class(res_ga, "tbl_df")
    expect_true("geometry" %in% names(res_ga))
    expect_true(inherits(res_ga$geometry, "geoarrow_vctr"))
  }
})

test_that("st_as_sf.duckspatial_df delegates to collect(as='sf')", {
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn))
  
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)[1:5,]
  ddbs_write_table(conn, nc_sf, "nc_as_sf")
  nc_lazy <- as_duckspatial_df("nc_as_sf", conn)
  
  res <- sf::st_as_sf(nc_lazy)
  expect_s3_class(res, "sf")
})

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## overlapping polygons in two groups:
## - group "a": 3 squares (0-3, 1-4, 2-5) -> common area is the (2,2)-(3,3) square
## - group "b": 2 squares (0-2, 1-3)      -> common area is the (1,1)-(2,2) square
sq <- function(xmin, ymin, side) {
  sf::st_polygon(list(matrix(
    c(xmin, ymin, xmin + side, ymin, xmin + side, ymin + side, xmin, ymin + side, xmin, ymin),
    ncol = 2, byrow = TRUE
  )))
}
inter_sf <- sf::st_sf(
  grp = c("a", "a", "a", "b", "b"),
  geometry = sf::st_sfc(sq(0, 0, 3), sq(1, 1, 3), sq(2, 2, 3), sq(0, 0, 2), sq(1, 1, 2), crs = 4326)
)
inter_ddbs <- as_duckspatial_df(inter_sf)
duckspatial::ddbs_write_table(conn_test, inter_sf, "inter")


# 1. ddbs_intersection_agg() ---------------------------------------------

## - CHECK 1.1: works on all input formats and they agree
## - CHECK 1.2: ddbs / sf output modes
## - CHECK 1.3: no `by` -> single common geometry (here empty: no area common to all 5)
## - CHECK 1.4: `by` -> one row per group with the correct common area
## - CHECK 1.5: messages and table writing
## - CHECK 2.x: errors
describe("ddbs_intersection_agg()", {

  describe("expected behavior", {

    it("works on all input formats and they agree", {
      out_ddbs <- ddbs_intersection_agg(inter_ddbs, by = "grp")
      out_sf   <- ddbs_intersection_agg(inter_sf, by = "grp")
      out_conn <- ddbs_intersection_agg("inter", by = "grp", conn = conn_test)

      expect_s3_class(out_ddbs, "duckspatial_df")
      expect_equal(
        ddbs_collect(out_ddbs) |> dplyr::arrange(grp),
        ddbs_collect(out_sf)   |> dplyr::arrange(grp)
      )
      expect_equal(
        ddbs_collect(out_ddbs) |> dplyr::arrange(grp),
        ddbs_collect(out_conn) |> dplyr::arrange(grp)
      )
    })

    it("returns an sf object with mode = 'sf'", {
      expect_s3_class(ddbs_intersection_agg(inter_ddbs, by = "grp", mode = "sf"), "sf")
    })

    it("intersects within each group correctly", {
      out <- ddbs_intersection_agg(inter_ddbs, by = "grp", mode = "sf") |>
        dplyr::arrange(grp)

      expect_equal(nrow(out), 2L)
      ## both groups' common area is a 1x1 square; drop CRS so st_area is planar
      planar <- sf::st_set_crs(out, NA)
      expect_equal(as.numeric(sf::st_area(planar)), c(1, 1))
    })

    it("aggregates all geometries into one when `by` is NULL", {
      ## three fully overlapping squares -> common (2,2)-(3,3) 1x1 square
      overlap_ddbs <- as_duckspatial_df(sf::st_sf(
        geometry = sf::st_sfc(sq(0, 0, 3), sq(1, 1, 3), sq(2, 2, 3), crs = 4326)
      ))
      out <- ddbs_intersection_agg(overlap_ddbs, mode = "sf")
      expect_equal(nrow(out), 1L)
      planar <- sf::st_set_crs(out, NA)
      expect_equal(as.numeric(sf::st_area(planar)), 1)
    })

    it("preserves the CRS", {
      out <- ddbs_intersection_agg(inter_ddbs, by = "grp", mode = "sf")
      expect_equal(sf::st_crs(out), sf::st_crs(4326))
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_intersection_agg(inter_ddbs, by = "grp"))
      expect_message(ddbs_intersection_agg("inter", by = "grp", conn = conn_test, name = "iagg"))
      expect_no_message(
        ddbs_intersection_agg("inter", by = "grp", conn = conn_test, name = "iagg", overwrite = TRUE, quiet = TRUE)
      )
    })

    it("writes a table when name is provided", {
      expect_true(ddbs_intersection_agg("inter", by = "grp", conn = conn_test, name = "iagg2"))
      expect_true(DBI::dbExistsTable(conn_test, "iagg2"))
    })
  })

  describe("errors", {

    it("requires connection when using table names", {
      expect_error(ddbs_intersection_agg("inter", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_intersection_agg(x = 999))
    })

    it("validates conn argument type", {
      expect_error(ddbs_intersection_agg(inter_ddbs, conn = 999))
    })

    it("validates `by` argument type", {
      expect_error(ddbs_intersection_agg(inter_ddbs, by = 1))
    })

    it("validates table name exists", {
      expect_error(ddbs_intersection_agg(x = "999", conn = conn_test))
    })
  })
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

testthat::skip_on_cran()


test_that("ddbs_create_conn accepts threads and memory_limit_gb", {
  # Test with specific values
  conn <- tryCatch(
    ddbs_create_conn(threads = 1, memory_limit_gb = 1),
    error = function(e) skip(paste("Could not create connection:", e$message))
  )
  on.exit(ddbs_stop_conn(conn), add = TRUE)
  
  settings <- ddbs_get_resources(conn)
  expect_equal(settings$threads, 1L)
  # DuckDB output should match requested GB (within small tolerance for format rounding)
  expect_true(settings$memory_limit_gb >= 0.9 && settings$memory_limit_gb <= 1.1)
})

test_that("ddbs_temp_conn accepts threads and memory_limit_gb", {
  # In-memory path
  conn <- ddbs_temp_conn(threads = 2, memory_limit_gb = 2)
  # Cleanup is handled by ddbs_temp_conn's on.exit
  
  settings <- ddbs_get_resources(conn)
  expect_equal(settings$threads, 2L)
  # DuckDB output should match requested GB (within small tolerance for format rounding)
  expect_true(settings$memory_limit_gb >= 1.9 && settings$memory_limit_gb <= 2.1)
  
  # File-based path
  conn2 <- ddbs_temp_conn(file = TRUE, threads = 1, memory_limit_gb = 1)
  settings2 <- ddbs_get_resources(conn2)
  expect_equal(settings2$threads, 1L)
  expect_true(settings2$memory_limit_gb >= 0.9 && settings2$memory_limit_gb <= 1.1)
})

test_that("ddbs_set_resources and ddbs_get_resources work", {
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn), add = TRUE)
  
  # Set new resources
  res <- ddbs_set_resources(conn, threads = 1, memory_limit_gb = 4)
  
  expect_equal(res$threads, 1L)
  # Use numeric field with tolerance to handle DuckDB format variations
  expect_true(res$memory_limit_gb >= 3.9 && res$memory_limit_gb <= 4.1)
  
  # Verify with get_resources
  res2 <- ddbs_get_resources(conn)
  expect_equal(res, res2)
  
  # Partial update
  ddbs_set_resources(conn, threads = 2)
  res3 <- ddbs_get_resources(conn)
  expect_equal(res3$threads, 2L)
  # Memory should still be ~4GB from previous set
  expect_true(res3$memory_limit_gb >= 3.9 && res3$memory_limit_gb <= 4.1)
})

test_that("assertions enforce valid inputs", {
  # ddbs_create_conn
  expect_error(ddbs_create_conn(threads = -1), "positive integer")
  expect_error(ddbs_create_conn(threads = 1.5), "positive integer")
  expect_error(ddbs_create_conn(memory_limit_gb = -1), "positive number")
  expect_error(ddbs_create_conn(memory_limit_gb = "invalid"), "positive number")
  
  # ddbs_temp_conn
  expect_error(ddbs_temp_conn(threads = 0), "positive integer")
  expect_error(ddbs_temp_conn(memory_limit_gb = 0), "positive number")
  
  # ddbs_set_resources
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn), add = TRUE)
  
  expect_error(ddbs_set_resources(conn, threads = -5), "positive integer")
  expect_error(ddbs_set_resources(conn, memory_limit_gb = -10), "positive number")
})

testthat::skip_on_cran()

test_that("collect.duckspatial_df respects global options", {
  skip_if_not_installed("sf")
  skip_if_not_installed("duckdb")
  
  conn <- ddbs_create_conn()
  on.exit(ddbs_stop_conn(conn))
  
  # Setup data
  nc_sf <- sf::st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)[1:5,]
  ddbs_write_table(conn, nc_sf, "nc_test_opts")
  
  # Create duckspatial_df
  # x <- dplyr::tbl(conn, "nc_test_opts") |> 
  #    as_duckspatial_df(geom_col = "geometry", crs = sf::st_crs(nc_sf))
  x <- as_duckspatial_df("nc_test_opts", conn)
  
  # Save original option
  op_orig <- options()
  on.exit(options(op_orig), add = TRUE)
  
  # 1. Default (sf) behavior when option is "duckspatial_df" (default)
  ddbs_options(output_type = "sf")
  res <- dplyr::collect(x)
  expect_s3_class(res, "sf")
  
  # 2. Option = "tibble"
  ddbs_options(output_type = "tibble")
  res_tbl <- dplyr::collect(x)
  expect_false(inherits(res_tbl, "sf"))
  expect_s3_class(res_tbl, "tbl_df")
  expect_false("geometry" %in% names(res_tbl))
  
  # 3. Option = "raw"
  ddbs_options(output_type = "raw")
  res_raw <- dplyr::collect(x)
  expect_true("geometry" %in% names(res_raw))
  expect_true(is.list(res_raw$geometry))
  
  # 4. Explicit override wins
  ddbs_options(output_type = "tibble")
  res_sf <- dplyr::collect(x, as = "sf")
  expect_s3_class(res_sf, "sf")
})



test_that("duckspatial.mode works as expected", {
  skip_if_not_installed("sf")
  
  # Setup data
  x <- ddbs_open_dataset(
    system.file("spatial/argentina.geojson", 
    package = "duckspatial")
  )
  
  # Save original option
  op_orig <- options()
  on.exit(options(op_orig), add = TRUE)
  
  # 1. Mode = "sf"
  ddbs_options(mode = "sf")
  res <- ddbs_centroid(x)
  expect_s3_class(res, "sf")
  
  # 2. Mode = "sf" on operations that return a vector
  res <- ddbs_area(x)
  expect_s3_class(res, "units")

  # 3. Mode = "sf" on operations that return a list
  res <- ddbs_intersects(x, x)
  expect_type(res, "list")
  
  # 4. Default mode (duckspatial)
  ddbs_options(mode = "duckspatial")
  res <- ddbs_centroid(x)
  expect_s3_class(res, "duckspatial_df")

})

# Tests for cross-connection import in ddbs_write_table
# These test the efficient import path when source is from a different connection
testthat::skip_on_cran()

test_that("ddbs_write_table imports from external duckspatial_df without collecting", {
  skip_if_not_installed("duckdbfs")
  
  # Create a "source" connection (simulating duckdbfs pattern)
  source_conn <- ddbs_create_conn(dbdir = "memory")
  ddbs_write_table(source_conn, countries_sf, "source_countries", quiet = TRUE)
  
  # Create duckspatial_df from source
  source_tbl <- ddbs_read_table(source_conn, "source_countries")
  
  # Create our target connection
  target_conn <- ddbs_create_conn(dbdir = "memory")
  on.exit({
    DBI::dbDisconnect(target_conn, shutdown = TRUE)
    DBI::dbDisconnect(source_conn, shutdown = TRUE)
  }, add = TRUE)
  
  # Import from source to target - should use efficient cross-connection path
  expect_no_error(
    ddbs_write_table(target_conn, source_tbl, "imported_countries", quiet = TRUE)
  )
  
  # Verify table was created
  tables <- ddbs_list_tables(target_conn)
  expect_true("imported_countries" %in% tables$table_name)
  
  # Verify geometry column exists and is GEOMETRY type
  desc <- DBI::dbGetQuery(target_conn, "DESCRIBE imported_countries")
  geom_types <- desc$column_type[desc$column_type == "GEOMETRY('EPSG:4326')"]
  expect_true(length(geom_types) >= 1)
  
  # Verify CRS was preserved
  imported <- ddbs_read_table(target_conn, "imported_countries")
  expect_equal(ddbs_crs(imported)$input, ddbs_crs(source_tbl)$input)
  
  # Verify data was imported correctly
  imported_sf <- st_as_sf(imported)
  expect_equal(nrow(imported_sf), nrow(countries_sf))
})

test_that("ddbs_write_table respects overwrite for cross-connection import", {
  skip_if_not_installed("duckdbfs")
  
  # Setup source
  source_conn <- ddbs_create_conn(dbdir = "memory")
  ddbs_write_table(source_conn, countries_sf, "source_countries", quiet = TRUE)
  source_tbl <- ddbs_read_table(source_conn, "source_countries")
  
  # Setup target with existing table
  target_conn <- ddbs_create_conn(dbdir = "memory")
  ddbs_write_table(target_conn, argentina_sf, "imported", quiet = TRUE)
  on.exit({
    DBI::dbDisconnect(target_conn, shutdown = TRUE)
    DBI::dbDisconnect(source_conn, shutdown = TRUE)
  }, add = TRUE)
  
  # Should fail without overwrite
  expect_error(
    ddbs_write_table(target_conn, source_tbl, "imported", quiet = TRUE),
    "already present"
  )
  
  # Should succeed with overwrite
  expect_no_error(
    ddbs_write_table(target_conn, source_tbl, "imported", overwrite = TRUE, quiet = TRUE)
  )
  
  # Verify it's the countries data now, not argentina
  result <- DBI::dbGetQuery(target_conn, "SELECT COUNT(*) as n FROM imported")
  expect_equal(result$n, nrow(countries_sf))
})

test_that("ddbs_write_table same-connection falls back to collect", {
  # When source and target are the SAME connection, should use collect path
  conn <- ddbs_create_conn(dbdir = "memory")
  on.exit(DBI::dbDisconnect(conn, shutdown = TRUE), add = TRUE)
  
  # Write initial data
  ddbs_write_table(conn, countries_sf, "original", quiet = TRUE)
  
  # Create duckspatial_df from same connection
  tbl <- ddbs_read_table(conn, "original")
  
  # Write to new table - should work (same connection doesn't need import)
  expect_no_error(
    ddbs_write_table(conn, tbl, "copy", quiet = TRUE)
  )
  
  # Verify both tables exist
  tables <- ddbs_list_tables(conn)
  expect_true(all(c("original", "copy") %in% tables$table_name))
})

# 0. Set up --------------------------------------------------------------

# skip tests on CRAN
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write data
duckspatial::ddbs_write_table(conn_test, argentina_sf, "argentina")


# 1. ddbs_generate_points() ----------------------------------------------

## - CHECK 1.1: works on all formats, n works, and seed works
## - CHECK 1.2: ddbs returns different outputs (duckspatial_df, geoarrow, sf, tbl)
## - CHECK 1.3: messages work
## - CHECK 1.4: writting a table works
## - CHECK 1.5: check different seeds
describe("ddbs_generate_points()", {

  ### EXPECTED BEHAVIOR

  describe("expected behavior", {

    it("works on all input formats, n works, and seed works", {
      output_ddbs <- ddbs_generate_points(argentina_ddbs, 50, seed = 123)
      output_sf   <- ddbs_generate_points(argentina_sf, 50, seed = 123)
      output_conn <- ddbs_generate_points("argentina", 50, conn = conn_test, seed = 123)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_equal(nrow(ddbs_collect(output_ddbs)), 50)
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_sf))
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different outputs based on mode argument", {
      output_sf_fmt <- ddbs_generate_points(argentina_ddbs, 10, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_generate_points(argentina_ddbs, 10))
      expect_message(ddbs_generate_points("argentina", 50, conn = conn_test, name = "generate_points"))
      expect_message(ddbs_generate_points("argentina", 50, conn = conn_test, name = "generate_points", overwrite = TRUE))
      expect_true(ddbs_generate_points("argentina", 50, conn = conn_test, name = "generate_points2"))

      expect_no_message(ddbs_generate_points(argentina_ddbs, 50, quiet = TRUE))
      expect_no_message(ddbs_generate_points("argentina", 50, seed = 123, conn = conn_test, name = "generate_points", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables correctly to DuckDB", {
      output_ddbs <- ddbs_generate_points(argentina_ddbs, 50, seed = 123)
      output_tbl <- ddbs_read_table(conn_test, "generate_points")
      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("produces different points with different seeds", {
      output_ddbs <- ddbs_generate_points(argentina_ddbs, 50, seed = 123)
      output_ddbs_2 <- ddbs_generate_points(argentina_ddbs, 50, seed = 678)
      expect_false(
        identical(ddbs_collect(output_ddbs), ddbs_collect(output_ddbs_2))
      )
    })

  })

  ### ERRORS ------------------------------------------------------------

  describe("errors", {

    it("requires a valid connection when using table name", {
      expect_error(ddbs_generate_points("countries", conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_generate_points(x = 999))
      expect_error(ddbs_generate_points(x = "999", conn = conn_test))
    })

    it("validates conn argument type", {
      expect_error(ddbs_generate_points(countries_ddbs, conn = 999))
    })

    it("validates new_column argument type", {
      expect_error(ddbs_generate_points(countries_ddbs, new_column = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_generate_points(countries_ddbs, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_generate_points(countries_ddbs, quiet = 999))
    })

    it("requires name to be a single character string", {
      expect_error(ddbs_generate_points(countries_ddbs, conn = conn_test, name = c('banana', 'banana')))
    })
  })
  
})


## stop connection
duckspatial::ddbs_stop_conn(conn_test)

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## 2D source data with a numeric column that will become the M value
src_sf <- sf::st_as_sf(
  data.frame(id = 1:2, m_val = c(50, 50)),
  geometry = sf::st_sfc(
    sf::st_linestring(matrix(c(0, 0, 1, 0, 2, 0), ncol = 2, byrow = TRUE)),
    sf::st_linestring(matrix(c(0, 1, 1, 1, 2, 1), ncol = 2, byrow = TRUE)),
    crs = 4326
  )
)
src_ddbs <- duckspatial::as_duckspatial_df(src_sf)

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## write 2D source to conn_test, then materialise M-geometry via ddbs_force_3d
## (writing XYM WKB directly from sf fails in DuckDB; SQL-based creation works)
duckspatial::ddbs_write_table(conn_test, src_sf, "lines_2d")
ddbs_force_3d("lines_2d", "m_val", dim = "m", conn = conn_test,
              name = "lines_m", quiet = TRUE)

## Fixtures
## - line_m_sf:   sf read back from DuckDB (M-geometry encoded by DuckDB, readable again)
## - line_m_ddbs: duckspatial_df with lazy M-geometry created by ddbs_force_3d
line_m_sf   <- duckspatial::ddbs_read_table(conn_test, "lines_m")
line_m_ddbs <- ddbs_force_3d(src_ddbs, "m_val", dim = "m")


# 1. ddbs_locate_along() -------------------------------------------------

describe("ddbs_locate_along()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_locate_along(line_m_ddbs, measure = 50)
      output_sf   <- ddbs_locate_along(line_m_sf, measure = 50)
      output_conn <- ddbs_locate_along("lines_m", measure = 50, conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_locate_along(line_m_ddbs, measure = 50, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_locate_along(line_m_ddbs, measure = 50))
      expect_message(ddbs_locate_along("lines_m", 50, conn = conn_test, name = "locate_along"))
      expect_message(ddbs_locate_along("lines_m", 50, conn = conn_test, name = "locate_along", overwrite = TRUE))
      expect_true(ddbs_locate_along("lines_m", 50, conn = conn_test, name = "locate_along2"))

      expect_no_message(ddbs_locate_along(line_m_ddbs, 50, quiet = TRUE))
      expect_no_message(ddbs_locate_along("lines_m", 50, conn = conn_test, name = "locate_along", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_locate_along(line_m_ddbs, measure = 50)
      output_tbl  <- ddbs_read_table(conn_test, "locate_along")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("returns one row per feature at the given M value", {
      output <- ddbs_locate_along(line_m_ddbs, measure = 50, mode = "sf")
      expect_equal(nrow(output), 2L)
    })

    it("filters out features with no point at the given M value", {
      output <- ddbs_locate_along(line_m_ddbs, measure = 999, mode = "sf")
      expect_equal(nrow(output), 0L)
    })

    describe("offset parameter", {

      it("works with default value (0)", {
        output <- ddbs_locate_along(line_m_ddbs, measure = 50, offset = 0)
        expect_s3_class(output, "duckspatial_df")
      })

      it("works with non-zero offset", {
        output <- ddbs_locate_along(line_m_ddbs, measure = 50, offset = 1)
        expect_s3_class(output, "duckspatial_df")
      })
    })
  })

  describe("errors", {

    it("requires numeric measure", {
      expect_error(ddbs_locate_along(line_m_ddbs, measure = "50"))
    })

    it("requires numeric offset", {
      expect_error(ddbs_locate_along(line_m_ddbs, measure = 50, offset = "1"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_locate_along("lines_m", measure = 50, conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_locate_along(x = 999, measure = 50))
    })

    it("validates conn argument type", {
      expect_error(ddbs_locate_along(line_m_ddbs, measure = 50, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_locate_along(line_m_ddbs, measure = 50, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_locate_along(line_m_ddbs, measure = 50, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_locate_along("999", measure = 50, conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_locate_along(line_m_ddbs, 50, conn = conn_test, name = c("banana", "banana")))
    })
  })
})




# 2. ddbs_locate_between() -----------------------------------------------

describe("ddbs_locate_between()", {

  describe("expected behavior", {

    it("works on all formats", {
      output_ddbs <- ddbs_locate_between(line_m_ddbs, start_measure = 25, end_measure = 75)
      output_sf   <- ddbs_locate_between(line_m_sf, start_measure = 25, end_measure = 75)
      output_conn <- ddbs_locate_between("lines_m", start_measure = 25, end_measure = 75, conn = conn_test)

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_s3_class(output_sf, "duckspatial_df")
      expect_equal(ddbs_collect(output_ddbs), ddbs_collect(output_conn))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_sf_fmt <- ddbs_locate_between(line_m_ddbs, 25, 75, mode = "sf")
      expect_s3_class(output_sf_fmt, "sf")
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_locate_between(line_m_ddbs, 25, 75))
      expect_message(ddbs_locate_between("lines_m", 25, 75, conn = conn_test, name = "locate_between"))
      expect_message(ddbs_locate_between("lines_m", 25, 75, conn = conn_test, name = "locate_between", overwrite = TRUE))
      expect_true(ddbs_locate_between("lines_m", 25, 75, conn = conn_test, name = "locate_between2"))

      expect_no_message(ddbs_locate_between(line_m_ddbs, 25, 75, quiet = TRUE))
      expect_no_message(ddbs_locate_between("lines_m", 25, 75, conn = conn_test, name = "locate_between", overwrite = TRUE, quiet = TRUE))
    })

    it("writes tables to the database", {
      output_ddbs <- ddbs_locate_between(line_m_ddbs, 25, 75)
      output_tbl  <- ddbs_read_table(conn_test, "locate_between")

      expect_equal(
        ddbs_collect(output_ddbs)$geometry,
        output_tbl$geometry
      )
    })

    it("returns one row per feature within the M range", {
      output <- ddbs_locate_between(line_m_ddbs, 25, 75, mode = "sf")
      expect_equal(nrow(output), 2L)
    })

    it("filters out features outside the M range", {
      output <- ddbs_locate_between(line_m_ddbs, 200, 300, mode = "sf")
      expect_equal(nrow(output), 0L)
    })

    describe("offset parameter", {

      it("works with default value (0)", {
        output <- ddbs_locate_between(line_m_ddbs, 25, 75, offset = 0)
        expect_s3_class(output, "duckspatial_df")
      })

      it("works with non-zero offset", {
        output <- ddbs_locate_between(line_m_ddbs, 25, 75, offset = 1)
        expect_s3_class(output, "duckspatial_df")
      })
    })
  })

  describe("errors", {

    it("requires numeric start_measure", {
      expect_error(ddbs_locate_between(line_m_ddbs, start_measure = "0", end_measure = 75))
    })

    it("requires numeric end_measure", {
      expect_error(ddbs_locate_between(line_m_ddbs, start_measure = 25, end_measure = "75"))
    })

    it("requires numeric offset", {
      expect_error(ddbs_locate_between(line_m_ddbs, 25, 75, offset = "1"))
    })

    it("requires connection when using table names", {
      expect_error(ddbs_locate_between("lines_m", 25, 75, conn = NULL))
    })

    it("validates x argument type", {
      expect_error(ddbs_locate_between(x = 999, start_measure = 25, end_measure = 75))
    })

    it("validates conn argument type", {
      expect_error(ddbs_locate_between(line_m_ddbs, 25, 75, conn = 999))
    })

    it("validates overwrite argument type", {
      expect_error(ddbs_locate_between(line_m_ddbs, 25, 75, overwrite = 999))
    })

    it("validates quiet argument type", {
      expect_error(ddbs_locate_between(line_m_ddbs, 25, 75, quiet = 999))
    })

    it("validates table name exists", {
      expect_error(ddbs_locate_between("999", 25, 75, conn = conn_test))
    })

    it("requires name to be single character string", {
      expect_error(ddbs_locate_between(line_m_ddbs, 25, 75, conn = conn_test, name = c("banana", "banana")))
    })
  })
})


## stop connection
ddbs_stop_conn(conn_test)

#' @details Spatial Join Predicates:
#'
#' A spatial predicate is really just a function that evaluates some spatial
#' relation between two geometries and returns true or false, e.g., “does a
#' contain b” or “is a within distance x of b”. Here is a quick overview of the
#' most commonly used ones, taking two geometries a and b:
#'
#' - `"ST_Intersects"`: Whether a intersects b
#' - `"ST_Contains"`: Whether a contains b
#' - `"ST_ContainsProperly"`: Whether a contains b without b touching a's boundary
#' - `"ST_Within"`: Whether a is within b
#' - `"ST_Overlaps"`: Whether a overlaps b
#' - `"ST_Touches"`: Whether a touches b
#' - `"ST_Equals"`: Whether a is equal to b
#' - `"ST_Crosses"`: Whether a crosses b
#' - `"ST_Covers"`: Whether a covers b
#' - `"ST_CoveredBy"`: Whether a is covered by b
#' - `"ST_DWithin"`: x)	Whether a is within distance x of b

#' Write spatial dataset to disk
#'
#' Writes spatial data to disk using DuckDB's `COPY` command for Parquet and
#' GDAL spatial formats, or as a native DuckDB database for `.duckdb`, `.db`,
#' and `.ddb` paths. Format is auto-detected from file extension for common
#' formats, or can be specified explicitly via `gdal_driver`.
#' 
#' Persistent DuckDB database files created by duckspatial use **Native 
#' Spatial Storage** (`storage_version = "v1.5.0"`) by default so CRS metadata is 
#' retained in native `GEOMETRY` columns. These files require DuckDB >= 1.5.0 
#' to open; use **Legacy Compatibility** (`storage_version = "v1.0.0"`) 
#' when the output must be readable by older DuckDB versions.
#'
#' @param data A `duckspatial_df`, `tbl_lazy` (DuckDB), or `sf` object.
#' @param path Path to output file.
#' @param gdal_driver GDAL driver name for writing spatial formats. If `NULL` (default),
#'   the driver is auto-detected from the file extension for common formats:
#'   - `.geojson`, `.json` → "GeoJSON"
#'   - `.shp` → "ESRI Shapefile"  
#'   - `.gpkg` → "GPKG"
#'   - `.fgb` → "FlatGeobuf"
#'   - `.kml` → "KML"
#'   - `.gpx` → "GPX"
#'   - `.gml` → "GML"
#'   - `.sqlite` → "SQLite"
#'   
#'   For **non-standard file extensions** (e.g., `.dat`, `.xyz`) or to **explicitly override** 
#'   auto-detection, specify the exact driver name as it appears in `ddbs_drivers()$short_name`. 
#'   Examples: `gdal_driver = "GeoJSON"`, `gdal_driver = "ESRI Shapefile"`.
#'   
#'   **Note**: If you specify a driver that doesn't match the file extension (e.g., 
#'   `path = "output.shp"` with `gdal_driver = "GeoJSON"`), a warning will be issued but 
#'   your explicit driver choice will be honored (creating a GeoJSON file with `.shp` extension).
#'   
#'   The function validates that the specified driver is available and writable on your 
#'   system. Note: `.parquet` and `.csv` files use native DuckDB writers and do not 
#'   require a GDAL driver.
#' @template conn_null
#' @param overwrite Logical. If `TRUE`, overwrites existing file.
#' @param crs Output CRS (e.g., "EPSG:4326"). Passed to GDAL as `SRS` option. Ignored for Parquet.
#' @param layer Table name for native DuckDB database output.
#' @param options Named list of additional options passed to `COPY`.
#' @param partitioning Character vector of columns to partition by (Parquet/CSV only).
#' @param parquet_compression Compression codec for Parquet.
#' @param parquet_row_group_size Row group size for Parquet.
#' @param layer_creation_options GDAL layer creation options.
#' @template quiet
#' @param duckdb_storage_version Storage compatibility for newly created native
#'   DuckDB database files (\code{.duckdb}, \code{.db}, \code{.ddb}). See
#'   \url{https://duckdb.org/docs/internals/storage} for more information on
#'   DuckDB storage versions and compatibility.
#'   \itemize{
#'     \item \code{"v1.5.0"} (\strong{Native Spatial Storage}, Default): Preserves
#'           CRS metadata in native DuckDB \code{GEOMETRY} columns. Requires
#'           DuckDB >= 1.5.0 to open the file.
#'     \item \code{"v1.0.0"} (\strong{Legacy Compatibility}): Creates
#'           files readable by older DuckDB versions (>= 1.0.0). Persists CRS
#'           metadata in duckspatial-managed column comments (a convention not
#'           recognized by other spatial software).
#'     \item \code{"latest"}: Use the highest storage version supported by your
#'           installed DuckDB engine.
#'   }
#'   Other major version strings like \code{"v1.4.0"}, \code{"v1.3.0"}, etc., are also supported.
#'
#' @return The `path` invisibly.
#' 
#' @seealso [ddbs_drivers()] to list all available GDAL drivers and formats.
#' 
#' @references 
#' This function is inspired by and builds upon the logic found in the 
#' \code{duckdbfs} package (\url{https://github.com/cboettig/duckdbfs}), 
#' particularly its \code{write_dataset} and \code{write_geo} functions.
#' For advanced features like cloud storage (S3) support, the 
#' \code{duckdbfs} package is highly recommended.
#' 
#' @export
#'
#' @examples
#' \dontrun{
#' library(duckspatial)
#' 
#' # Read example data
#' path <- system.file("spatial/countries.geojson", package = "duckspatial")
#' ds <- ddbs_open_dataset(path)
#' 
#' # Auto-detect format from extension
#' ddbs_write_dataset(ds, "output.geojson")
#' ddbs_write_dataset(ds, "output.gpkg")
#' ddbs_write_dataset(ds, "output.parquet")
#' 
#' # Explicit GDAL driver for non-standard extension
#' ddbs_write_dataset(ds, "mydata.dat", gdal_driver = "GeoJSON")
#' 
#' # See available drivers on your system
#' drivers <- ddbs_drivers()
#' writable <- drivers[drivers$can_create == TRUE, ]
#' head(writable)
#' 
#' # CRS override
#' ddbs_write_dataset(ds, "output_3857.geojson", crs = "EPSG:3857")
#' 
#' # Overwrite existing file
#' ddbs_write_dataset(ds, "output.gpkg", overwrite = TRUE)
#' }
ddbs_write_dataset <- function(
    data,
    path,
    gdal_driver = NULL,
    conn = NULL,
    overwrite = FALSE,
    crs = NULL,
    layer = "spatial",
    options = list(),
    partitioning = if (inherits(data, c("tbl_lazy", "duckspatial_df"))) dplyr::group_vars(data) else NULL,
    parquet_compression = NULL,
    parquet_row_group_size = NULL,
    layer_creation_options = NULL,
    quiet = FALSE,
    duckdb_storage_version = duckspatial_storage_default()
) {
  duckdb_storage_version <- match_duckdb_storage_version(duckdb_storage_version)
  
  # 1. Resolve connection
  if (is.null(conn)) {
    conn <- extract_connection(data)
    if (is.null(conn)) conn <- ddbs_default_conn()
  }
  
  # 2. Ensure DuckDB spatial extension
  ddbs_install(conn, quiet = TRUE)
  ddbs_load(conn, quiet = TRUE)
  
  # 3. Format detection and driver resolution
  ext <- tolower(tools::file_ext(path))
  
  # Determine if native format
  is_parquet <- ext == "parquet"
  is_csv <- ext == "csv"
  is_duckdb <- has_duckdb_file_extension(path)
  is_native <- is_parquet || is_csv || is_duckdb
  
  # For GDAL formats, resolve driver
  driver_name <- NULL
  if (!is_native) {
    # Step 1: Try auto-detection from extension
    driver_from_ext <- get_driver_from_extension(ext)
    
    # Step 2: Use explicit driver if provided
    if (!is.null(gdal_driver)) {
      driver_name <- gdal_driver
      
      # Warn if driver and extension mismatch
      if (!is.null(driver_from_ext) && driver_from_ext != driver_name) {
        expected_ext <- get_extension_for_driver(driver_name)
        if (!is.null(expected_ext)) {
          cli::cli_warn(c(
            "Extension/driver mismatch detected.",
            "i" = "File extension {.val .{ext}} typically maps to driver {.val {driver_from_ext}}.",
            "i" = "You specified driver {.val {driver_name}}, which typically uses {.val .{expected_ext}}.",
            "i" = "Proceeding with your explicit driver choice."
          ))
        }
      }
    } else if (!is.null(driver_from_ext)) {
      # Use auto-detected driver
      driver_name <- driver_from_ext
    } else {
      # Unknown extension, no driver provided
      cli::cli_abort(c(
        "Cannot determine GDAL driver for extension {.val .{ext}}.",
        "i" = "Please specify the driver explicitly using {.arg gdal_driver}:",
        " " = "  ddbs_write_dataset(data, path, gdal_driver = \"GeoJSON\")",
        "i" = "Run {.code ddbs_drivers()} to see all available drivers."
      ))
    }
    
    # Step 3: Validate driver exists and is writable
    available_drivers <- tryCatch({
      ddbs_drivers(conn)
    }, error = function(e) {
      cli::cli_warn("Could not query available GDAL drivers. Proceeding without validation.")
      NULL
    })
    
    if (!is.null(available_drivers)) {
      # Check if driver exists
      driver_info <- available_drivers[available_drivers$short_name == driver_name, ]
      
      if (nrow(driver_info) == 0) {
        # Driver not found - provide helpful error
        writable_drivers <- available_drivers[available_drivers$can_create == TRUE, ]
        cli::cli_abort(c(
          "GDAL driver {.val {driver_name}} is not available on this system.",
          "i" = "Run {.code ddbs_drivers()} to see all available drivers.",
          "i" = "Available writable drivers include:",
          " " = paste("  -", head(writable_drivers$short_name, 10), collapse = "\n")
        ))
      } else if (!driver_info$can_create[1]) {
        # Driver exists but is read-only
        writable_drivers <- available_drivers[available_drivers$can_create == TRUE, ]
        cli::cli_abort(c(
          "GDAL driver {.val {driver_name}} is read-only and cannot be used for writing.",
          "i" = "Run {.code ddbs_drivers()} and check the {.field can_create} column.",
          "i" = "Available writable drivers include:",
          " " = paste("  -", head(writable_drivers$short_name, 10), collapse = "\n")
        ))
      }
    }
  }
  
  # 4. Check for existing file
  if (file.exists(path)) {
    if (!overwrite) {
      cli::cli_abort("File {.path {path}} already exists. Use {.arg overwrite = TRUE} to replace.")
    } else {
      # For GDAL formats, explicit delete is safer/more consistent
      if (!is_native) {
        unlink(path)
      }
      if (is_duckdb) {
        unlink(path)
      }
      # For native (Parquet/CSV), we can use OVERWRITE_OR_IGNORE option below
    }
  }

  if (is_duckdb) {
    return(ddbs_write_duckdb_dataset(
      data = data,
      path = path,
      layer = layer,
      overwrite = TRUE,
      crs = crs,
      duckdb_storage_version = duckdb_storage_version,
      quiet = quiet
    ))
  }

  # 4b. Auto-detect CRS if not provided
  # For GDAL formats, it's crucial to pass SRS if possible so output has projection.
  # For Parquet, DuckDB might not automatically write CRS metadata unless enabled/configured,
  # but passing SRS usually doesn't hurt or is ignored by native COPY (check docs?). 
  # Actually native Parquet COPY doesn't take SRS. We only pass checks for GDAL.
  
  if (is.null(crs) && !is_native) {
      # Try to detect from data
      obj_crs <- tryCatch(ddbs_crs(data, conn = conn), error = function(e) NULL)
      if (!is.null(obj_crs) && !is.na(obj_crs)) {
          # Prefer Authority Code (EPSG:xxxx) if available, else WKT
          # sf crs object
          epsg <- obj_crs$epsg
          if (!is.null(epsg) && !is.na(epsg)) {
              crs <- paste0("EPSG:", epsg)
          } else {
              crs <- obj_crs$wkt
          }
      }
  }
  
  # 5. Logic to generate the SQL source
  # We want the 'source' to be either a table name or (subquery)
  view_name <- NULL
  sql_source <- NULL
  
  if (inherits(data, c("duckspatial_df", "tbl_lazy", "tbl_duckdb_connection"))) {
    # Efficient table name extraction or subquery construction
    r_name <- attr(data, "source_table") %||% dbplyr::remote_name(data)
    is_simple_table <- !is.null(r_name) && is.character(r_name)
    
    if (is_simple_table) {
       sql_source <- r_name
    } else {
       sql_source <- paste0("(", dbplyr::sql_render(data, con = conn), ")")
    }
    
    # Strict Remote Validation: introspect types using DESCRIBE
    # We must ensure there is a GEOMETRY column.
    dtypes <- tryCatch({
        DBI::dbGetQuery(conn, glue::glue("DESCRIBE SELECT * FROM {sql_source} LIMIT 0"))
    }, error = function(e) {
        # Fallback if describe fails (complex queries?), though usually works
        NULL
    })
    
    has_geometry_type <- FALSE
    if (!is.null(dtypes) && "column_type" %in% names(dtypes)) {
        has_geometry_type <- any(grepl("GEOMETRY", dtypes$column_type))
    }
    
    
    if (!has_geometry_type) {
         # Double check if maybe it's named 'geometry' but not typed yet (rare in valid spatial tables)
         # Using DuckDB spatial, it should be GEOMETRY.
         cli::cli_abort("Input DuckDB table/query does not contain a spatial column of type 'GEOMETRY'.")
    }

    # Sanitize column names for specific drivers (e.g. FID in GeoPackage)
    if (!is.null(driver_name) && driver_name == "GPKG" && !is.null(dtypes)) {
         cols_in_table <- dtypes$column_name
         if (any(toupper(cols_in_table) == "FID")) {
             q_cols_list <- lapply(cols_in_table, function(col) {
                 q <- DBI::dbQuoteIdentifier(conn, col)
                 if (toupper(col) == "FID") return(paste0(q, " AS FID_original"))
                 return(q)
             })
             cols_sql <- paste(unlist(q_cols_list), collapse = ", ")
             sql_source <- glue::glue("(SELECT {cols_sql} FROM {sql_source})")
             cli::cli_alert_warning("Column 'FID' renamed to 'FID_original' to avoid conflict with GeoPackage primary key.")
         }
    }
    
  } else if (inherits(data, c("sf", "data.frame"))) {
    
    # Strict Local Validation: Must be 'sf'
    if (!inherits(data, "sf")) {
       cli::cli_abort("Input local data must be an 'sf' object. Plain data.frames are not supported.")
    }
    
    view_name <- ddbs_temp_view_name()
    
    tryCatch({
       # Ensure view doesn't exist (robust check)
       tryCatch(duckdb::duckdb_unregister(conn, view_name), error = function(e) NULL)
       DBI::dbExecute(conn, glue::glue("DROP VIEW IF EXISTS {view_name}"))
       
       if (inherits(data, "sf")) {
           local_sf_crs <- sf::st_crs(data)
           local_sf_crs_text <- if (is_parquet) {
               ddbs_write_dataset_crs_text(local_sf_crs, conn)
           } else {
               NULL
           }

           # Explicitly convert to WKB to ensure consistent DuckDB typing (BLOB)
           wkb_col <- attr(data, "sf_column")
           data[[wkb_col]] <- sf::st_as_binary(data[[wkb_col]])
           
           # Register the WKB-fied data frame
           duckdb::duckdb_register(conn, view_name, data)
           
           # Cast Blob geometry to GEOMETRY type for spatial formats
           gcol <- attr(data, "sf_column")
           col_names <- setdiff(names(data), gcol)
           
           # Use proper quoting
           q_gcol <- DBI::dbQuoteIdentifier(conn, gcol)
           geom_expr <- glue::glue("ST_GeomFromWKB({q_gcol})")
           if (!is.null(local_sf_crs_text)) {
               crs_sql <- as.character(DBI::dbQuoteString(conn, local_sf_crs_text))
               geom_expr <- glue::glue("ST_GeomFromWKB({q_gcol})::GEOMETRY({crs_sql})")
           }
           
           if (length(col_names) > 0) {
               # Check for potentially conflicting columns (e.g. FID in GeoPackage)
               if (!is.null(driver_name) && driver_name == "GPKG" && any(toupper(col_names) == "FID")) {
                   # Rename conflicting FID column
                   q_cols_list <- lapply(col_names, function(col) {
                       q <- DBI::dbQuoteIdentifier(conn, col)
                       if (toupper(col) == "FID") return(paste0(q, " AS FID_original"))
                       return(q)
                   })
                   cols_sql <- paste(unlist(q_cols_list), collapse = ", ")
                   cli::cli_alert_warning("Column 'FID' renamed to 'FID_original' to avoid conflict with GeoPackage primary key.")
               } else {
                   q_cols <- DBI::dbQuoteIdentifier(conn, col_names)
                   cols_sql <- paste(q_cols, collapse = ", ")
               }
              subquery <- glue::glue("SELECT {cols_sql}, {geom_expr} AS {q_gcol} FROM {view_name}")
           } else {
              subquery <- glue::glue("SELECT {geom_expr} AS {q_gcol} FROM {view_name}")
           }
           sql_source <- paste0("(", subquery, ")")
           
       } else {
           duckdb::duckdb_register(conn, view_name, data)
           sql_source <- view_name
       }
       
    }, error = function(e) {
       cli::cli_abort("Failed to register local data frame: {e$message}")
    })
    
  } else {
    cli::cli_abort("Unsupported input type: {.cls {class(data)}}")
  }
  
  # 6. Build COPY options
  copy_ops <- list()
  
  # Format
  if (is_parquet) {
      copy_ops$FORMAT <- "PARQUET"
  } else if (is_csv) {
      copy_ops$FORMAT <- "CSV"
  } else {
      # GDAL
      copy_ops$FORMAT <- "GDAL"
      copy_ops$DRIVER <- driver_name
      
      # Default layer creation options for GDAL if not provided
      if (is.null(layer_creation_options)) {
          layer_creation_options <- "WRITE_BBOX=YES"
      }
      
      # Pass SRS if available
      if (!is.null(crs)) {
          copy_ops$SRS <- crs
      }
  }
  
  # Options: Overwrite (Native only)
  if (is_native && overwrite) {
      copy_ops$OVERWRITE_OR_IGNORE <- TRUE
  }
  
  # Options: Parquet Compression
  if (!is.null(parquet_compression)) {
      copy_ops$COMPRESSION <- parquet_compression
  }
  
  # Options: Row Group Size
  if (!is.null(parquet_row_group_size)) {
      copy_ops$ROW_GROUP_SIZE <- as.integer(parquet_row_group_size)
  }
  
  # Options: Partitioning (Native only)
  if (!is.null(partitioning) && length(partitioning) > 0) {
     p_cols <- paste(partitioning, collapse = ", ")
     copy_ops$PARTITION_BY <- paste0("(", p_cols, ")")
  }
  
  # Options: Layer Creation (GDAL only)
  if (!is.null(layer_creation_options)) {
      copy_ops$LAYER_CREATION_OPTIONS <- layer_creation_options
  }
  
  # Merge Generic Options
  if (length(options) > 0) {
      copy_ops <- utils::modifyList(copy_ops, options)
  }
  
  # 7. Execute COPY
  
  # Helper to format options
  fmt_opts <- function(ops) {
      parts <- vapply(names(ops), function(n) {
          val <- ops[[n]]
          if (is.logical(val)) {
              val_str <- ifelse(val, "TRUE", "FALSE")
          } else if (is.numeric(val)) {
              val_str <- as.character(val)
          } else {
              # String: handle single quotes escaping if needed, though rare in options
              # PARTITION_BY handled specially (starts with parens)
              if (grepl("^\\(", val)) {
                  val_str <- val 
              } else {
                  val_str <- paste0("'", val, "'")
              }
          }
          paste0(n, " ", val_str)
      }, character(1))
      paste(parts, collapse = ", ")
  }
  
  opt_str <- fmt_opts(copy_ops)
  query <- glue::glue("COPY {sql_source} TO '{path}' ({opt_str})")
  
  tryCatch({
      DBI::dbExecute(conn, query)
  }, finally = {
      # Cleanup temp view if we created one
      if (!is.null(view_name)) {
          duckdb::duckdb_unregister(conn, view_name)
      }
  })
  
  if (!quiet) {
      cli::cli_alert_success("Written to {.path {path}}")
  }
  
  invisible(path)
}

#' Write a dataset to a native DuckDB database file
#' @noRd
ddbs_write_duckdb_dataset <- function(
  data,
  path,
  layer,
  overwrite,
  crs,
  duckdb_storage_version,
  quiet
) {
  target_conn <- ddbs_create_conn(path, duckdb_storage_version = duckdb_storage_version)
  on.exit(ddbs_stop_conn(target_conn), add = TRUE)

  if (!is.null(crs)) {
    if (inherits(data, "sf")) {
      sf::st_crs(data) <- sf::st_crs(crs)
    } else if (inherits(data, c("duckspatial_df", "tbl_lazy", "tbl_duckdb_connection"))) {
      data <- as_duckspatial_df(data, crs = crs)
    }
  }

  ddbs_write_table(
    conn = target_conn,
    data = data,
    name = layer,
    overwrite = overwrite,
    quiet = TRUE
  )
  ddbs_checkpoint_if_possible(target_conn)

  if (!quiet) {
    cli::cli_alert_success("Written to {.path {path}}")
  }

  invisible(path)
}

#' Convert an sf CRS to DuckDB GEOMETRY CRS text for GeoParquet writes
#' @noRd
ddbs_write_dataset_crs_text <- function(crs, conn) {
    if (is.null(crs) || is.na(crs)) {
        return(NULL)
    }

    epsg <- crs$epsg
    if (!is.null(epsg) && length(epsg) > 0 && !is.na(epsg)) {
        return(paste0("EPSG:", as.integer(epsg)))
    }

    wkt <- crs$wkt
    if (is.null(wkt) || length(wkt) == 0 || is.na(wkt) || identical(wkt, "")) {
        cli::cli_abort(
            "DuckDB GeoParquet export requires PROJJSON CRS metadata for custom or non-EPSG CRS values, but the input CRS has no WKT representation."
        )
    }

    projjson <- tryCatch(
        {
            sf::st_as_text(crs, projjson = TRUE)
        },
        error = function(e) {
            cli::cli_abort(
                c(
                    "DuckDB GeoParquet export requires PROJJSON CRS metadata for custom or non-EPSG CRS values.",
                    "x" = "{.fun sf::st_as_text} failed to convert the input CRS to PROJJSON.",
                    "i" = conditionMessage(e)
                )
            )
        }
    )

    if (!grepl("^\\s*\\{", projjson) || !grepl("\\}\\s*$", projjson)) {
        cli::cli_abort(
            c(
                "DuckDB GeoParquet export requires PROJJSON CRS metadata for custom or non-EPSG CRS values.",
                "x" = "{.fun sf::st_as_text} did not return a PROJJSON object."
            )
        )
    }

    projjson
}

#' Helper to extract connection from object
#' @noRd
extract_connection <- function(x) {
    if (inherits(x, c("tbl_duckdb_connection", "duckspatial_df"))) {
        return(dbplyr::remote_con(x))
    }
    if (inherits(x, "tbl_lazy")) {
        # Check if src is duckdb
        con <- dbplyr::remote_con(x)
        if (inherits(con, "duckdb_connection")) return(con)
    }
    NULL
}

#' Map file extension to GDAL driver name
#' @noRd
get_driver_from_extension <- function(ext) {
  ext_map <- list(
    "geojson" = "GeoJSON",
    "json" = "GeoJSON",
    "shp" = "ESRI Shapefile",
    "gpkg" = "GPKG",
    "fgb" = "FlatGeobuf",
    "kml" = "KML",
    "gpx" = "GPX",
    "gml" = "GML",
    "sqlite" = "SQLite",
    "tab" = "MapInfo File",
    "mif" = "MapInfo File"
  )
  ext_map[[tolower(ext)]]
}

#' Get extension for a driver (inverse mapping)
#' @noRd
get_extension_for_driver <- function(driver) {
  driver_to_ext <- list(
    "GeoJSON" = "geojson",
    "ESRI Shapefile" = "shp",
    "GPKG" = "gpkg",
    "FlatGeobuf" = "fgb",
    "KML" = "kml",
    "GPX" = "gpx",
    "GML" = "gml",
    "SQLite" = "sqlite",
    "MapInfo File" = "tab"
  )
  driver_to_ext[[driver]]
}

# Helper PROJJSON strings for GeoParquet tests
# Used when sf::st_crs()$projjson returns NULL in some environments

projjson_fallback_3857 <- '{"$schema":"https://proj.org/schemas/v0.7/projjson.schema.json","type":"ProjectedCRS","name":"WGS 84 / Pseudo-Mercator","base_crs":{"type":"GeographicCRS","name":"WGS 84","datum":{"type":"GeodeticReferenceFrame","name":"World Geodetic System 1984","ellipsoid":{"name":"WGS 84","semi_major_axis":6378137,"inverse_flattening":298.257223563}},"coordinate_system":{"subtype":"ellipsoidal","axis":[{"name":"Latitude","abbreviation":"lat","direction":"north","unit":"degree"},{"name":"Longitude","abbreviation":"lon","direction":"east","unit":"degree"}]},"id":{"authority":"EPSG","code":4326}},"conversion":{"name":"unnamed","method":{"name":"Popular Visualisation Pseudo Mercator","id":{"authority":"EPSG","code":1024}},"parameters":[{"name":"Latitude of natural origin","value":0,"unit":"degree","id":{"authority":"EPSG","code":8801}},{"name":"Longitude of natural origin","value":0,"unit":"degree","id":{"authority":"EPSG","code":8802}},{"name":"False easting","value":0,"unit":"metre","id":{"authority":"EPSG","code":8806}},{"name":"False northing","value":0,"unit":"metre","id":{"authority":"EPSG","code":8807}}]},"coordinate_system":{"subtype":"Cartesian","axis":[{"name":"Easting","abbreviation":"X","direction":"east","unit":"metre"},{"name":"Northing","abbreviation":"Y","direction":"north","unit":"metre"}]},"id":{"authority":"EPSG","code":3857}}'

projjson_fallback_4326 <- '{"$schema":"https://proj.org/schemas/v0.7/projjson.schema.json","type":"GeographicCRS","name":"WGS 84","datum_ensemble":{"name":"World Geodetic System 1984 ensemble","members":[{"name":"World Geodetic System 1984 (Transit)","id":{"authority":"EPSG","code":1166}},{"name":"World Geodetic System 1984 (G730)","id":{"authority":"EPSG","code":1152}},{"name":"World Geodetic System 1984 (G873)","id":{"authority":"EPSG","code":1153}},{"name":"World Geodetic System 1984 (G1150)","id":{"authority":"EPSG","code":1154}},{"name":"World Geodetic System 1984 (G1674)","id":{"authority":"EPSG","code":1155}},{"name":"World Geodetic System 1984 (G1762)","id":{"authority":"EPSG","code":1156}},{"name":"World Geodetic System 1984 (G2139)","id":{"authority":"EPSG","code":1157}}],"ellipsoid":{"name":"WGS 84","semi_major_axis":6378137,"inverse_flattening":298.257223563},"accuracy":"2.0","id":{"authority":"EPSG","code":6326}},"coordinate_system":{"subtype":"ellipsoidal","axis":[{"name":"Geodetic latitude","abbreviation":"Lat","direction":"north","unit":"degree"},{"name":"Geodetic longitude","abbreviation":"Lon","direction":"east","unit":"degree"}]},"id":{"authority":"EPSG","code":4326}}'

#' @returns Depends on the \code{output} argument (or global preference set by \code{\link{ddbs_options}}):
#'   \itemize{
#'     \item \code{duckspatial_df} (default): A lazy spatial data frame backed by dbplyr/DuckDB.
#'     \item \code{sf}: An eagerly collected \code{sf} object in R memory.
#'     \item \code{tibble}: An eagerly collected \code{tibble} without geometry in R memory.
#'     \item \code{raw}: An eagerly collected \code{tibble} with WKB geometry (no conversion).
#'     \item \code{geoarrow}: An eagerly collected \code{tibble} with geometry converted to \code{geoarrow_vctr}.
#'   }
#'   When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).

#' Locate geometries at specific M values
#'
#' Return the geometries at a specific M values or range of M values.
#'
#' @template x
#' @param measure A numeric value specifying the M value at which to locate a
#'   point along the geometry. Used only by \code{ddbs_locate_along}.
#' @param start_measure A numeric value specifying the lower bound of the M
#'   range.
#' @param end_measure A numeric value specifying the upper bound of the M
#'   range.
#' @param offset A numeric value specifying a lateral offset to apply
#'   perpendicular to the line direction at the located point(s). Default is
#'   \code{0} (no offset).
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#' 
#' @details
#' 
#' - `ddbs_locate_along()`: returns a point or multi-point, containing the point(s) 
#'   at the geometry with the given measure
#' 
#' - `ddbs_locate_between()`: returns a geometry or geometry collection created by 
#'   filtering and interpolating vertices within a range of "M" values
#'
#'
#' @examples
#' \dontrun{
#' ## load package
#' library(duckspatial)
#'
#' ## read data (must contain M-enabled linestring geometries)
#' rivers_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/rivers.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## Calculate the length of the rivers
#' rivers_agg_ddbs <- rivers_ddbs |> 
#'   ddbs_union_agg("RIVER_NAME") |> 
#'   ddbs_length()
#' 
#' ## Add M dimension to the rivers
#' rivers_m_ddbs <- rivers_agg_ddbs |> 
#'   ddbs_force_3d("length", dim = "M")
#' 
#' ## Locate rivers with M between 10000 and 20000
#' ddbs_locate_between(rivers_m_ddbs, 10000, 20000)
#' 
#' }
#' @name ddbs_locate
#' @rdname ddbs_locate
NULL



#' @rdname ddbs_locate
#' @export
ddbs_locate_along <- function(
    x,
    measure,
    offset = 0,
    conn = NULL,
    name = NULL,
    mode = NULL,
    overwrite = FALSE,
    quiet = FALSE) {

    # 0. Validate inputs
    assert_xy(x, "x")
    assert_conn_x_name(conn, x, name)
    assert_conn_character(conn, x)
    assert_name(name)
    assert_name(mode, "mode")
    assert_logic(overwrite, "overwrite")
    assert_logic(quiet, "quiet")
    assert_numeric(measure, "measure")
    assert_numeric(offset, "offset")


    # 1. Prepare inputs
    
    ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
    ## validate character table names)
    x <- normalize_spatial_input(x, conn)

    ## 1.2. Pre-extract attributes
    crs_x    <- ddbs_crs(x, conn)
    sf_col_x <- attr(x, "sf_column")
    mode     <- get_mode(mode, name)

    ## 1.3. Resolve spatial connections and handle imports
    resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
    target_conn  <- resolve_conn$conn
    x            <- resolve_conn$x
    ## register cleanup of the connection
    on.exit(resolve_conn$cleanup(), add = TRUE)

    ## 1.4. Get list with query names for the input data
    x_list <- get_query_list(x, target_conn)
    on.exit(x_list$cleanup(), add = TRUE)


    # 2. Prepare the query

    ## 2.1. Get the geometry column name (try to extract from attributes, if not 
    ## available get it from the database)
    x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
    assert_geometry_column(x_geom, x_list)

    ## 2.2. Build the function arguments
    args <- sprintf(
        "%s, %s",
        x_geom,
        glue::glue("{measure}, {offset}")
    )

    ## 2.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
    ## We drop empty geometries. Two steps are required:
    ## (1) apply ST_LocateAlong keeping the result as raw GEOMETRY so ST_IsEmpty works,
    ## (2) filter empties, then convert to the requested output format in the outer SELECT.
    st_function <- glue::glue("ST_LocateAlong({args})")
    base.query <- glue::glue("
        WITH located AS (
          SELECT * REPLACE ({st_function} AS {x_geom})
          FROM {x_list$query_name}
        )
        SELECT * REPLACE ({build_geom_query(x_geom, name, crs_x, mode)} AS {x_geom})
        FROM located
        WHERE NOT ST_IsEmpty({x_geom});
      ")
    


    # 3. Table creation if name is provided, or 
    # create duckspatial_df or sf object if name is NULL
    if (!is.null(name)) {
        create_duckdb_table(
          conn      = target_conn,
          name      = name,
          query     = base.query,
          overwrite = overwrite,
          quiet     = quiet
        )
    } else {
      ddbs_handle_query(
        query  = base.query,
        conn   = target_conn,
        mode   = mode,
        crs    = crs_x,
        x_geom = x_geom
      )
    }

}




#' @rdname ddbs_locate
#' @export
ddbs_locate_between <- function(
  x,
  start_measure,
  end_measure,
  offset = 0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  # 0. Validate inputs
  assert_xy(x, "x")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(mode, "mode")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")
  assert_numeric(start_measure, "start_measure")
  assert_numeric(end_measure, "end_measure")
  assert_numeric(offset, "offset")


  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")
  mode     <- get_mode(mode, name)

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. Build the function arguments
  args <- sprintf(
    "%s, %s",
    x_geom,
    glue::glue("{start_measure}, {end_measure}, {offset}")
  )

  ## 2.3. Build the base query (depends on the output type - sf, duckspatial_df, table)
  ## We drop empty geometries. Two steps are required:
  ## (1) apply ST_LocateBetween keeping the result as raw GEOMETRY so ST_IsEmpty works,
  ## (2) filter empties, then convert to the requested output format in the outer SELECT.
  st_function <- glue::glue("ST_LocateBetween({args})")
  base.query <- glue::glue("
    WITH located AS (
      SELECT * REPLACE ({st_function} AS {x_geom})
      FROM {x_list$query_name}
    )
    SELECT * REPLACE ({build_geom_query(x_geom, name, crs_x, mode)} AS {x_geom})
    FROM located
    WHERE NOT ST_IsEmpty({x_geom});
  ")
  


  # 3. Table creation if name is provided, or 
  # create duckspatial_df or sf object if name is NULL
  if (!is.null(name)) {
    create_duckdb_table(
      conn      = target_conn,
      name      = name,
      query     = base.query,
      overwrite = overwrite,
      quiet     = quiet
    )
  } else {
    ddbs_handle_query(
      query  = base.query,
      conn   = target_conn,
      mode   = mode,
      crs    = crs_x,
      x_geom = x_geom
    )
  }

}

duckspatial_storage_default <- function() {
  getOption("duckspatial.duckdb_storage_version", "v1.5.0")
}

duckspatial_storage_versions <- function() {
  c(duckspatial_storage_default(), "v1.4.0", "v1.3.0", "v1.2.0", "v1.0.0", "latest")
}

match_duckdb_storage_version <- function(x) {
  if (!is.character(x) || length(x) != 1) {
    cli::cli_abort("{.arg duckdb_storage_version} must be a single string.")
  }

  choices <- duckspatial_storage_versions()

  # 1. Exact match in curated list (e.g., 'latest', 'v1.5.0', 'v1.0.0')
  if (x %in% choices) {
    return(x)
  }

  # 2. Pass-through for strings that look like DuckDB version tags (vX.Y.Z)
  # This automatically supports future DuckDB releases without package updates.
  # If the version is unknown to the DuckDB binary, it will throw a descriptive
  # error during connection.
  if (grepl("^v[0-9]+", x)) {
    return(x)
  }

  # 3. Fallback to standard match.arg for suggestions/error reporting
  match.arg(x, choices)
}

is_storage_v15_or_newer <- function(tag) {
  if (is.null(tag) || is.na(tag) || !nzchar(tag)) {
    return(FALSE)
  }

  # DuckDB tags can be "v1.5.0+", "v1.0.0 - v1.1.3", etc.
  # We extract the first semantic version string found.
  version_match <- regexpr("[0-9]+\\.[0-9]+\\.[0-9]+", tag)
  if (version_match == -1) {
    return(FALSE)
  }

  v_str <- regmatches(tag, version_match)
  utils::compareVersion(v_str, "1.5.0") >= 0
}

ddbs_assert_duckdb_crs_support <- function() {
  if (utils::packageVersion("duckdb") < "1.5.0") {
    cli::cli_abort(c(
      "duckspatial requires duckdb >= 1.5.0 to open files written by recent duckspatial versions.",
      "i" = "Installed: {utils::packageVersion('duckdb')}. Upgrade: install.packages('duckdb')."
    ))
  }

  invisible(TRUE)
}

ddbs_storage_is_legacy <- function(conn) {
  identical(attr(conn, "duckspatial_storage_mode"), "legacy")
}

ddbs_input_crs_for_write <- function(data, conn = NULL) {
  tryCatch(
    {
      if (inherits(data, "sf")) {
        return(sf::st_crs(data))
      }
      if (
        inherits(data, c("duckspatial_df", "tbl_duckdb_connection", "tbl_lazy"))
      ) {
        return(ddbs_crs(data))
      }
      if (
        is.character(data) &&
          length(data) == 1 &&
          file.exists(data) &&
          !has_duckdb_file_extension(data) &&
          !is.null(conn)
      ) {
        return(get_file_crs(data, conn))
      }
      NULL
    },
    error = function(e) NULL
  )
}

ddbs_write_legacy_crs_comment_if_needed <- function(
  conn,
  table,
  geom_col = NULL,
  crs = NULL
) {
  if (!ddbs_storage_is_legacy(conn)) {
    return(invisible(FALSE))
  }

  if (is.null(crs) || is.na(sf::st_crs(crs))) {
    return(invisible(FALSE))
  }

  geom_col <- geom_col %||%
    tryCatch(get_geom_name(conn, table), error = function(e) NULL)
  if (is.null(geom_col) || is.na(geom_col)) {
    return(invisible(FALSE))
  }

  write_crs_comment(conn, table, geom_col, crs)
}

ddbs_quote_table <- function(conn, table) {
  if (inherits(table, "SQL")) {
    return(as.character(table))
  }

  if (length(table) == 2) {
    return(as.character(DBI::dbQuoteIdentifier(
      conn,
      DBI::Id(schema = table[[1]], table = table[[2]])
    )))
  }

  table <- as.character(table)
  if (length(table) != 1) {
    cli::cli_abort("{.arg table} must identify one table.")
  }

  parts <- strsplit(table, ".", fixed = TRUE)[[1]]
  if (length(parts) == 2) {
    return(as.character(DBI::dbQuoteIdentifier(
      conn,
      DBI::Id(schema = parts[[1]], table = parts[[2]])
    )))
  }

  as.character(DBI::dbQuoteIdentifier(conn, table))
}

crs_to_duckdb_literal <- function(x) {
  if (is.null(x)) {
    return(list(literal = NA_character_, kind = "none"))
  }

  crs <- sf::st_crs(x)
  if (is.na(crs)) {
    return(list(literal = NA_character_, kind = "none"))
  }

  parsed <- sf::st_crs(crs, parameters = TRUE)
  srid <- parsed$srid
  if (!is.null(srid) && length(srid) > 0 && !is.na(srid)) {
    return(list(literal = as.character(srid), kind = "authority"))
  }

  wkt <- crs$wkt
  if (is.null(wkt) || length(wkt) == 0 || is.na(wkt) || !nzchar(wkt)) {
    return(list(literal = NA_character_, kind = "none"))
  }

  list(literal = wkt, kind = "wkt")
}

ddbs_quote_sql_string <- function(conn, x) {
  if (!is.null(conn)) {
    return(as.character(DBI::dbQuoteString(conn, x)))
  }

  paste0("'", gsub("'", "''", x, fixed = TRUE), "'")
}

duckdb_geometry_type <- function(conn, x) {
  lit <- crs_to_duckdb_literal(x)
  if (identical(lit$kind, "none")) {
    return("GEOMETRY")
  }

  paste0("GEOMETRY(", ddbs_quote_sql_string(conn, lit$literal), ")")
}

read_native_crs <- function(conn, table, geom_col) {
  q_geom <- as.character(DBI::dbQuoteIdentifier(conn, geom_col))
  q_table <- ddbs_quote_table(conn, table)
  query <- paste0(
    "SELECT ST_CRS(",
    q_geom,
    ") AS crs ",
    "FROM ",
    q_table,
    " ",
    "WHERE ",
    q_geom,
    " IS NOT NULL ",
    "LIMIT 1"
  )

  res <- tryCatch(DBI::dbGetQuery(conn, query), error = function(e) NULL)
  if (
    is.null(res) ||
      nrow(res) == 0 ||
      is.na(res$crs[[1]]) ||
      identical(res$crs[[1]], "")
  ) {
    return(sf::st_crs(NA))
  }

  sf::st_crs(res$crs[[1]])
}

make_crs_comment <- function(crs, geom_col, existing_payload = NULL) {
  parsed <- sf::st_crs(crs)
  payload <- list(
    duckspatial = list(
      version = 1L,
      geometry_column = geom_col,
      crs = list(
        input = as.character(parsed$input),
        authority = if (!is.null(parsed$authority)) parsed$authority else NULL,
        srid = if (!is.na(parsed$epsg)) as.integer(parsed$epsg) else NULL,
        wkt2_2019 = parsed$wkt
      )
    )
  )

  if (is.list(existing_payload) && !is.null(existing_payload$user_comment)) {
    payload$duckspatial$user_comment <- existing_payload$user_comment
  }

  jsonlite::toJSON(payload, auto_unbox = TRUE, null = "null", na = "null")
}

parse_crs_comment <- function(comment_text) {
  if (is.null(comment_text) || is.na(comment_text) || !nzchar(comment_text)) {
    return(NULL)
  }

  obj <- tryCatch(
    jsonlite::fromJSON(comment_text, simplifyVector = FALSE),
    error = function(e) NULL
  )
  if (is.null(obj) || is.null(obj$duckspatial)) {
    return(NULL)
  }

  obj$duckspatial
}

get_column_comment <- function(conn, table, geom_col) {
  name_list <- get_query_name(table)
  query <- paste0(
    "SELECT comment FROM duckdb_columns() ",
    "WHERE schema_name = ",
    DBI::dbQuoteString(conn, name_list$schema_name),
    " AND table_name = ",
    DBI::dbQuoteString(conn, name_list$table_name),
    " AND column_name = ",
    DBI::dbQuoteString(conn, geom_col),
    " LIMIT 1"
  )
  res <- DBI::dbGetQuery(conn, query)
  if (nrow(res) == 0) {
    return(NA_character_)
  }

  res$comment[[1]]
}

write_crs_comment <- function(conn, table, geom_col, crs) {
  parsed <- sf::st_crs(crs)
  if (is.na(parsed)) {
    return(invisible(FALSE))
  }

  cur <- get_column_comment(conn, table, geom_col)
  parsed_cur <- parse_crs_comment(cur)
  existing <- NULL
  if (is.null(parsed_cur) && !is.null(cur) && !is.na(cur) && nzchar(cur)) {
    existing <- list(user_comment = cur)
  }

  payload <- make_crs_comment(parsed, geom_col, existing_payload = existing)
  sql <- sprintf(
    "COMMENT ON COLUMN %s.%s IS %s",
    ddbs_quote_table(conn, table),
    DBI::dbQuoteIdentifier(conn, geom_col),
    DBI::dbQuoteString(conn, payload)
  )
  DBI::dbExecute(conn, sql)
  invisible(TRUE)
}

crs_from_comment_payload <- function(payload) {
  if (is.null(payload) || is.null(payload$crs)) {
    return(sf::st_crs(NA))
  }

  pref <- NULL
  if (!is.null(payload$crs$authority) && !is.null(payload$crs$srid)) {
    pref <- paste0(payload$crs$authority, ":", payload$crs$srid)
  }
  pref <- pref %||%
    payload$crs$srid %||%
    payload$crs$wkt2_2019 %||%
    payload$crs$input

  tryCatch(sf::st_crs(pref), error = function(e) sf::st_crs(NA))
}

resolve_crs <- function(
  conn,
  table,
  geom_col,
  explicit = NULL,
  warn_on_disagree = TRUE,
  quiet_unknown = FALSE
) {
  if (!is.null(explicit)) {
    return(sf::st_crs(explicit))
  }

  native <- read_native_crs(conn, table, geom_col)
  comment <- parse_crs_comment(get_column_comment(conn, table, geom_col))
  comment_crs <- crs_from_comment_payload(comment)

  has_native <- !is.na(native)
  has_comment <- !is.na(comment_crs)

  if (
    has_native &&
      has_comment &&
      warn_on_disagree &&
      !crs_equal(native, comment_crs)
  ) {
    cli::cli_warn(
      "CRS disagreement for {.val {table}}.{.val {geom_col}}: using native DuckDB CRS metadata."
    )
  }
  if (has_native) {
    return(native)
  }
  if (has_comment) {
    return(comment_crs)
  }

  if (!quiet_unknown) {
    cli::cli_warn(c(
      "CRS could not be detected for {.val {table}}.{.val {geom_col}}.",
      "i" = "This can happen when reading an older DuckDB storage file that was created before duckspatial persisted CRS metadata, or when the table has no CRS information.",
      "i" = "Use {.arg crs} when opening the dataset, or rewrite the file using {.code ddbs_create_conn(..., storage_version = \"v1.5.0\")} to persist CRS natively."
    ))
  }

  sf::st_crs(NA)
}

ddbs_duckdb_storage_tag <- function(conn) {
  tags <- tryCatch(
    DBI::dbGetQuery(
      conn,
      "SELECT tags FROM duckdb_databases() WHERE database_name = current_database()"
    )$tags[[1]],
    error = function(e) NULL
  )

  if (!is.data.frame(tags) || nrow(tags) == 0) {
    return(NA_character_)
  }

  value <- tags$value[tags$key == "storage_version"]
  if (length(value) == 0) {
    return(NA_character_)
  }

  as.character(value[[1]])
}

ddbs_open_persistent <- function(
  dbdir,
  duckdb_storage_version = duckspatial_storage_default(),
  read_only = FALSE,
  ...
) {
  duckdb_storage_version <- match_duckdb_storage_version(duckdb_storage_version)

  if (dbdir %in% c(":memory:", "memory")) {
    conn <- DBI::dbConnect(
      duckdb::duckdb(dbdir = ":memory:"),
      geometry = "wk",
      ...
    )
    attr(conn, "duckspatial_storage_mode") <- "memory"
    return(conn)
  }

  is_new <- !file.exists(dbdir)
  config <- list()
  if (is_new && !identical(duckdb_storage_version, "v1.0.0")) {
    config$storage_compatibility_version <- duckdb_storage_version
  }

  conn <- DBI::dbConnect(
    duckdb::duckdb(dbdir = dbdir, read_only = read_only, config = config),
    geometry = "wk",
    ...
  )

  actual <- ddbs_duckdb_storage_tag(conn)
  requested <- if (duckdb_storage_version %in% c("v1.0.0", "latest")) {
    NA_character_
  } else {
    paste0(duckdb_storage_version, "+")
  }

  # Fallback to legacy (comment) mode if storage is older than v1.5.0
  legacy_mode <- !is_storage_v15_or_newer(actual)

  if (
    !legacy_mode &&
      !is.na(requested) &&
      !identical(actual, requested) &&
      !isTRUE(read_only)
  ) {
    # If they requested a specific modern version but got a different modern version,
    # we still use native spatial storage but warn.
    cli::cli_warn(
      c(
        "Requested DuckDB storage {.val {duckdb_storage_version}} but file storage is {.val {actual}}.",
        "i" = "Native CRS persistence will be used."
      ),
      class = "duckspatial_storage_mismatch"
    )
  } else if (legacy_mode && !identical(duckdb_storage_version, "v1.0.0") && !isTRUE(read_only)) {
    cli::cli_warn(
      c(
        "Requested DuckDB storage {.val {duckdb_storage_version}} but file storage is {.val {actual}} (Legacy).",
        "i" = "CRS persistence will use duckspatial-managed column-comment metadata for this compatibility connection."
      ),
      class = "duckspatial_storage_mismatch"
    )
  }

  attr(conn, "db_file") <- dbdir
  attr(conn, "duckspatial_storage") <- actual
  attr(conn, "duckspatial_storage_mode") <- if (legacy_mode) {
    "legacy"
  } else {
    duckdb_storage_version
  }
  conn
}

testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb", minimum_version = "1.5.1")
testthat::skip_if_not_installed("sf")

test_that("duckdb geometry type preserves authority and custom CRS literals", {
  conn <- ddbs_temp_conn()

  expect_equal(
    duckspatial:::duckdb_geometry_type(conn, "EPSG:4326"),
    "GEOMETRY('EPSG:4326')"
  )
  expect_equal(duckspatial:::duckdb_geometry_type(conn, NA), "GEOMETRY")

  custom <- "+proj=lcc +lat_0=49 +lon_0=-95 +lat_1=49 +lat_2=77 +datum=NAD83"
  literal <- duckspatial:::crs_to_duckdb_literal(custom)
  expect_equal(literal$kind, "wkt")
  expect_match(
    duckspatial:::duckdb_geometry_type(conn, custom),
    "^GEOMETRY\\('PROJCRS",
    perl = TRUE
  )
})

test_that("ddbs_create_conn persists native CRS metadata in v1.5 storage", {
  db_path <- tempfile(fileext = ".duckdb")
  on.exit(unlink(db_path), add = TRUE)

  conn <- ddbs_create_conn(db_path)
  expect_equal(attr(conn, "duckspatial_storage_mode"), "v1.5.0")
  expect_equal(attr(conn, "duckspatial_storage"), "v1.5.0+")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(conn), silent = TRUE)),
    add = TRUE
  )
  ddbs_write_table(conn, points_sf[1:3, ], "points", quiet = TRUE)
  ddbs_stop_conn(conn)

  conn2 <- ddbs_create_conn(db_path)
  expect_equal(attr(conn2, "duckspatial_storage_mode"), "v1.5.0")
  expect_equal(attr(conn2, "duckspatial_storage"), "v1.5.0+")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(conn2), silent = TRUE)),
    add = TRUE
  )
  ds <- as_duckspatial_df("points", conn = conn2)
  geom_col <- attr(ds, "sf_column")

  expect_equal(sf::st_crs(ds)$epsg, 4326)
  expect_false(is.na(duckspatial:::read_native_crs(conn2, "points", geom_col)))
  expect_true(is.na(duckspatial:::get_column_comment(
    conn2,
    "points",
    geom_col
  )))
})

test_that("v1.0.0 storage (Legacy Compatibility) writes and reads CRS column comments", {
  db_path <- tempfile(fileext = ".duckdb")
  on.exit(unlink(db_path), add = TRUE)

  conn <- ddbs_create_conn(db_path, duckdb_storage_version = "v1.0.0")
  expect_equal(attr(conn, "duckspatial_storage_mode"), "legacy")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(conn), silent = TRUE)),
    add = TRUE
  )
  ddbs_write_table(conn, points_sf[1:3, ], "points", quiet = TRUE)
  geom_col <- get_geom_name(conn, "points")

  comment <- duckspatial:::get_column_comment(conn, "points", geom_col)
  expect_match(comment, "\"duckspatial\"", fixed = TRUE)
  expect_equal(sf::st_crs(ddbs_crs("points", conn = conn))$epsg, 4326)
  ddbs_stop_conn(conn)

  conn2 <- ddbs_create_conn(db_path, duckdb_storage_version = "v1.0.0")
  expect_equal(attr(conn2, "duckspatial_storage_mode"), "legacy")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(conn2), silent = TRUE)),
    add = TRUE
  )
  expect_equal(sf::st_crs(ddbs_crs("points", conn = conn2))$epsg, 4326)
})

test_that("ddbs_write_dataset writes native DuckDB database files", {
  db_path <- tempfile(fileext = ".duckdb")
  on.exit(unlink(db_path), add = TRUE)

  expect_no_error(
    ddbs_write_dataset(
      points_sf[1:3, ],
      db_path,
      layer = "points",
      quiet = TRUE
    )
  )

  ds <- ddbs_open_dataset(db_path, layer = "points")
  on.exit(
    suppressWarnings(try(
      ddbs_stop_conn(attr(ds, "source_conn")),
      silent = TRUE
    )),
    add = TRUE
  )

  expect_s3_class(ds, "duckspatial_df")
  expect_equal(dplyr::count(ds) |> dplyr::pull(n), 3)
  expect_equal(sf::st_crs(ds)$epsg, 4326)
})

test_that("ddbs_write_dataset v1.0.0 DuckDB output (Legacy Compatibility) writes CRS column comments", {
  db_path <- tempfile(fileext = ".duckdb")
  on.exit(unlink(db_path), add = TRUE)

  expect_no_error(
    ddbs_write_dataset(
      points_sf[1:3, ],
      db_path,
      layer = "points",
      duckdb_storage_version = "v1.0.0",
      quiet = TRUE
    )
  )

  conn <- ddbs_create_conn(db_path, duckdb_storage_version = "v1.0.0")
  expect_equal(attr(conn, "duckspatial_storage_mode"), "legacy")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(conn), silent = TRUE)),
    add = TRUE
  )

  geom_col <- get_geom_name(conn, "points")
  comment <- duckspatial:::get_column_comment(conn, "points", geom_col)

  expect_match(comment, "\"duckspatial\"", fixed = TRUE)
  expect_equal(sf::st_crs(ddbs_crs("points", conn = conn))$epsg, 4326)
  ddbs_stop_conn(conn)

  expect_no_warning({
    ds <- ddbs_open_dataset(db_path, layer = "points")
  })
  on.exit(
    suppressWarnings(try(
      ddbs_stop_conn(attr(ds, "source_conn")),
      silent = TRUE
    )),
    add = TRUE
  )
  expect_equal(sf::st_crs(ds)$epsg, 4326)
})

test_that("custom proj4 CRS round-trips through native and comment metadata", {
  custom_crs <- paste(
    "+proj=lcc +lat_0=49 +lon_0=-95 +lat_1=49 +lat_2=77",
    "+datum=NAD83 +units=m +no_defs"
  )
  custom_points <- sf::st_sf(
    id = 1:2,
    geom = sf::st_sfc(
      sf::st_point(c(0, 0)),
      sf::st_point(c(1000, 1000)),
      crs = custom_crs
    )
  )

  native_db <- tempfile(fileext = ".duckdb")
  compat_db <- tempfile(fileext = ".duckdb")
  on.exit(unlink(c(native_db, compat_db)), add = TRUE)

  native_conn <- ddbs_create_conn(native_db)
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(native_conn), silent = TRUE)),
    add = TRUE
  )
  ddbs_write_table(native_conn, custom_points, "pts", quiet = TRUE)
  ddbs_stop_conn(native_conn)

  native_conn2 <- ddbs_create_conn(native_db)
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(native_conn2), silent = TRUE)),
    add = TRUE
  )
  native_crs <- ddbs_crs("pts", conn = native_conn2)
  expect_true(duckspatial:::crs_equal(native_crs, sf::st_crs(custom_crs)))
  expect_false(is.na(duckspatial:::read_native_crs(native_conn2, "pts", "geom")))

  compat_conn <- ddbs_create_conn(compat_db, duckdb_storage_version = "v1.0.0")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(compat_conn), silent = TRUE)),
    add = TRUE
  )
  ddbs_write_table(compat_conn, custom_points, "pts", quiet = TRUE)
  ddbs_stop_conn(compat_conn)

  compat_conn2 <- ddbs_create_conn(compat_db, duckdb_storage_version = "v1.0.0")
  on.exit(
    suppressWarnings(try(ddbs_stop_conn(compat_conn2), silent = TRUE)),
    add = TRUE
  )
  compat_crs <- ddbs_crs("pts", conn = compat_conn2)
  expect_true(duckspatial:::crs_equal(compat_crs, sf::st_crs(custom_crs)))
  expect_true(is.na(duckspatial:::read_native_crs(compat_conn2, "pts", "geom")))
  expect_match(
    duckspatial:::get_column_comment(compat_conn2, "pts", "geom"),
    "\"wkt2_2019\"",
    fixed = TRUE
  )
})

#' @param x Input spatial data. Can be:
#'   \itemize{
#'     \item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
#'     \item An \code{sf} object
#'     \item A \code{tbl_lazy} from dbplyr
#'     \item A character string naming a table/view in \code{conn}
#'   }
#'   Data is returned from this object.

#' Get or set connection resources
#'
#' Configure technical system settings for a DuckDB connection, such as memory limits
#' and CPU threads.
#'
#' @template conn
#' @template threads
#' @template memory_limit_gb
#'
#' @return For \code{ddbs_set_resources()}, invisibly returns a list containing the current system settings; for \code{ddbs_get_resources()}, visibly returns the same list for direct inspection.
#' @export
#'
#' @examples
#' \dontrun{
#' # Create a connection
#' conn <- ddbs_create_conn()
#'
#' # Set resources: 1 thread and 4GB
#' ddbs_set_resources(conn, threads = 1, memory_limit_gb = 4)
#'
#' # Check current settings
#' ddbs_get_resources(conn)
#'
#' ddbs_stop_conn(conn)
#' }
ddbs_set_resources <- function(conn, threads = NULL, memory_limit_gb = NULL) {
  # 1. Checks
  dbConnCheck(conn)

  assert_threads(threads)
  assert_memory_limit_gb(memory_limit_gb)

  # 2. SETTER logic
  if (!is.null(threads)) {
    DBI::dbExecute(conn, sprintf("SET threads = %d;", as.integer(threads)))
  }

  if (!is.null(memory_limit_gb)) {
    DBI::dbExecute(conn, sprintf("SET memory_limit = '%.1fGB';", as.numeric(memory_limit_gb)))
  }

  # 3. Return current state
  invisible(ddbs_get_resources(conn))
}

#' @rdname ddbs_set_resources
#' @export
ddbs_get_resources <- function(conn) {
  # 1. Checks
  dbConnCheck(conn)

  # 2. Query settings
  settings <- DBI::dbGetQuery(conn, "
    SELECT name, value 
    FROM duckdb_settings() 
    WHERE name IN ('threads', 'memory_limit');
  ")

  # Format result
  res <- as.list(stats::setNames(settings$value, settings$name))
  
  # Parse numeric values for programmatic use
  if (!is.null(res$threads)) res$threads <- as.integer(res$threads)
  if (!is.null(res$memory_limit)) {
    res$memory_limit_gb <- parse_memory_limit_gb(res$memory_limit)
  }
  
  return(res)
}

#' Parse DuckDB memory limit string to numeric GB
#' 
#' @param mem_str Character string from DuckDB (e.g., "4GB", "3.7 GiB", "953.6 MiB")
#' @return Numeric value in GB
#' @noRd
parse_memory_limit_gb <- function(mem_str) {
  if (is.null(mem_str) || is.na(mem_str)) return(NA_real_)
  mem_str <- trimws(mem_str)
  
  # Extract numeric part and unit
  num_part <- as.numeric(sub("^([0-9.]+).*", "\\1", mem_str))
  unit_part <- toupper(sub("^[0-9.]+\\s*", "", mem_str))
  
  if (is.na(num_part)) return(NA_real_)
  
  # Conversion factors to GB (10^9 bytes)
  # DuckDB: KB/MB/GB/TB = 1000 base; KiB/MiB/GiB/TiB = 1024 base
  multipliers <- c(
    "TIB" = 1024^4 / 10^9, "TB" = 1000,
    "GIB" = 1024^3 / 10^9, "GB" = 1,
    "MIB" = 1024^2 / 10^9, "MB" = 1/1000,
    "KIB" = 1024 / 10^9,   "KB" = 1/10^6,
    "B"   = 1/10^9
  )
  
  m <- multipliers[unit_part]
  if (is.na(m)) m <- 1 # Default to GB if unit not found
  
  num_part * m
}

#' Convert point geometries to QuadKey tiles
#'
#' Transforms point geometries into QuadKey identifiers at a specified zoom level, 
#' a hierarchical spatial indexing system used by mapping services.
#' 
#' @template x
#' @param level An integer specifying the zoom level for QuadKey generation (1-23).
#' Higher values provide finer spatial resolution. Default is 10.
#' @param field Character string specifying the field name for aggregation.
#' @param fun aggregation function for when there are multiple quadkeys (e.g. "mean",
#' "min", "max", "sum").
#' @param background numeric. Default value in raster cells without values. Only used when 
#' \code{output = "raster"}
#' @template conn_null
#' @template name
#' @param output Character string specifying output format. One of:
#'   \itemize{
#'     \item \code{"polygon"} - Returns QuadKey tile boundaries as `duckspatial_df` (default)
#'     \item \code{"raster"} - Returns QuadKey values as a `SpatRaster`
#'     \item \code{"tilexy"} - Returns tile XY coordinates as a `tibble`
#' }
#' @template overwrite
#' @template quiet
#'
#' @returns Depends on the output argument
#' \itemize{
#'     \item \code{polygon} (default): A lazy spatial data frame backed by dbplyr/DuckDB.
#'     \item \code{raster}: An eagerly collected \code{SpatRaster} object in R memory.
#'     \item \code{tilexy}: An eagerly collected \code{tibble} without geometry in R memory.
#'   }
#' When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
#'
#' @details
#' QuadKeys divide the world into a hierarchical grid of tiles, where each tile
#' is subdivided into four smaller tiles at the next zoom level. This function
#' wraps DuckDB's ST_QuadKey spatial function to generate these tiles from input geometries.
#' 
#' Note that creating a table inside the connection will generate a non-spatial table, and 
#' therefore, it cannot be read with [ddbs_read_table].
#'
#' @export
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(duckspatial)
#' library(sf)
#' library(terra)
#'
#' # create a duckdb database in memory (with spatial extension)
#' conn <- ddbs_create_conn(dbdir = "memory")
#'
#' ## create random points in Argentina
#' argentina_sf <- st_read(system.file("spatial/argentina.geojson", package = "duckspatial"))
#' rand_sf <- st_sample(argentina_sf, 100) |> st_as_sf()
#' rand_sf["var"] <- runif(100)
#'
#' ## store in duckdb
#' ddbs_write_vector(conn, rand_sf, "rand_sf")
#'
#' ## generate QuadKey polygons at zoom level 8
#' qkey_ddbs <- ddbs_quadkey(conn = conn, "rand_sf", level = 8, output = "polygon")
#'
#' ## generate QuadKey raster with custom field name
#' qkey_rast <- ddbs_quadkey(conn = conn, "rand_sf", level = 6, output = "raster", field = "var")
#'
#' ## generate Quadkey XY tiles
#' qkey_tiles_tbl <- ddbs_quadkey(conn = conn, "rand_sf", level = 10, output = "tilexy")
#' }
ddbs_quadkey <- function(
  x,
  level = 10,
  field  = NULL,
  fun    = "mean",
  background = NA,
  conn = NULL,
  name = NULL,
  output = "polygon",
  overwrite = FALSE,
  quiet = FALSE
) {

  
  # 0. Validate inputs
  assert_xy(x, "x")
  assert_numeric(level, "level")
  assert_name(field, "field")
  assert_character_scalar(fun, "fun")
  assert_conn_x_name(conn, x, name)
  assert_conn_character(conn, x)
  assert_name(name)
  assert_name(output, "output")
  assert_logic(overwrite, "overwrite")
  assert_logic(quiet, "quiet")

  ## valid outputs
  if (!output %in% c("polygon", "raster", "tilexy")) cli::cli_abort("{.arg output} must be one of: {.val {c('polygon', 'raster', 'tilexy')}}")

  ## suggested packages
  rlang::check_installed(
    "quadkeyr",
    "to convert quadkeys to sf or raster."
  )

  if (output == "raster") {
    rlang::check_installed(
      "terra",
      "to convert quadkeys to raster."
    )
  }  

  # 1. Prepare inputs
  
  ## 1.1. Normalize inputs (coerce tbl_duckdb_connection to duckspatial_df, 
  ## validate character table names)
  x <- normalize_spatial_input(x, conn)

  ## 1.2. Pre-extract attributes
  crs_x    <- ddbs_crs(x, conn)
  sf_col_x <- attr(x, "sf_column")

  ## 1.3. Resolve spatial connections and handle imports
  resolve_conn <- resolve_spatial_connections(x, y = NULL, conn = conn, quiet = quiet)
  target_conn  <- resolve_conn$conn
  x            <- resolve_conn$x
  ## register cleanup of the connection
  on.exit(resolve_conn$cleanup(), add = TRUE)

  ## 1.4. Get list with query names for the input data
  x_list <- get_query_list(x, target_conn)
  on.exit(x_list$cleanup(), add = TRUE)


  # 2. Prepare the query

  ## 2.1. Get the geometry column name (try to extract from attributes, if not 
  ## available get it from the database)
  x_geom <- sf_col_x %||% get_geom_name(target_conn, x_list$query_name)
  assert_geometry_column(x_geom, x_list)

  ## 2.2. check CRS (we need EPSG:4326 for quadkeys)
  if (!crs_x$input %in% c("EPSG:4326", "WGS 84")) {
    if (!quiet) cli::cli_alert_info("Transforming {.arg x} crs to {.val EPSG:4326}")
    ## query
    tmp.query <- glue::glue("
      CREATE OR REPLACE TABLE {x_list$query_name} AS
      SELECT *
      REPLACE (ST_Transform({x_geom}, '{crs_x$input}', 'EPSG:4326') AS {x_geom}) 
      FROM {x_list$query_name};
    ")
    ## execute
    DBI::dbExecute(target_conn, tmp.query)
  }


  # 3. Table creation if name is provided
  if (!is.null(name)) {

      ## convenient names of table and/or schema.table
      name_list <- get_query_name(name)

      ## handle overwrite
      overwrite_table(name_list$query_name, target_conn, quiet, overwrite)

      ## create query with optional aggregation
      if (!is.null(field)) {
        tmp.query <- glue::glue("
          CREATE TABLE {name_list$query_name} AS
          SELECT 
            ST_QuadKey({x_geom}, {level}) as quadkey,
            {fun}({field}) as {field}
          FROM {x_list$query_name}
          GROUP BY quadkey;
        ")
      } else {
        tmp.query <- glue::glue("
          CREATE TABLE {name_list$query_name} AS
          SELECT * EXCLUDE ({x_geom}),
          ST_QuadKey({x_geom}, {level}) as quadkey 
          FROM {x_list$query_name};
        ")
      }
      ## execute intersection query
      DBI::dbExecute(target_conn, tmp.query)
      feedback_query(quiet)
      return(invisible(TRUE))
  }


  # 4. Get data frame with query-level aggregation

  ## 4.1. create query with aggregation if field is specified
  if (!is.null(field)) {
    tmp.query <- glue::glue("
      SELECT 
        ST_QuadKey({x_geom}, {level}) as quadkey,
        {fun}({field}) as {field}
      FROM {x_list$query_name}
      GROUP BY quadkey;
    ")
  } else {
    tmp.query <- glue::glue("
      SELECT * EXCLUDE ({x_geom}),
      ST_QuadKey({x_geom}, {level}) as quadkey 
      FROM {x_list$query_name};
    ")
  }
  
  ## 4.2. retrieve results from the query
  data_tbl <- DBI::dbGetQuery(target_conn, tmp.query)

  
  # 5. convert to desired output format
  if (output == "polygon") {

    ## get 1 quadkey per row (already aggregated if field was specified)
    data_sf <- quadkeyr::quadkey_df_to_polygon(data_tbl)

    ## convert to duckspatial_df
    prep_data <- as_duckspatial_df(data_sf)

  } else if (tolower(output) == "tilexy") {

    ## convert to tibble
    prep_data <- dplyr::bind_cols(
      dplyr::bind_rows(lapply(data_tbl$quadkey, quadkeyr::quadkey_to_tileXY)),
      data_tbl |> dplyr::select(-dplyr::all_of("quadkey"))
    )

  } else if (output == "raster") {

    ## get raster grid as polygons
    grid_lst <- quadkeyr::get_regular_polygon_grid(data_tbl)

    ## get field variable as SpatVector points
    pts_sf        <- dplyr::bind_rows(lapply(data_tbl$quadkey, quadkeyr::quadkey_to_latlong))
    pts_sf[field] <- as.vector(data_tbl[field])
    
    ## convert to spatvector to work with terra
    pts_vect  <- terra::vect(pts_sf)
    grid_vect <- terra::vect(grid_lst$data)

    ## create raster template
    grid_rast <- terra::rast(
      grid_vect,
      nrows = grid_lst$num_rows, 
      ncols = grid_lst$num_cols,
      crs   = terra::crs(grid_vect)
    )

    ## convert to raster
    prep_data <- terra::rasterize(
      x          = pts_vect,
      y          = grid_rast,
      field      = field,
      fun        = fun,
      background = background
    )

  }
  
  return(prep_data)

}

#' Count geometry components
#'
#' Functions to count the number of points, sub-geometries or interior rings in
#' a geometry
#'
#' @template x
#' @template new_column
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @details
#' These functions query structural properties of geometries:
#'
#' - `ddbs_get_npoints()` returns the number of points (vertices) in a geometry.
#'   For LINESTRING geometries this is the vertex count; for POLYGON types it
#'   includes all vertices of the exterior ring and any interior rings.
#'
#' - `ddbs_get_ngeometries()` returns the number of sub-geometries in a
#'   GEOMETRYCOLLECTION or MULTI* geometry (e.g. MULTIPOLYGON, MULTILINESTRING).
#'   Returns 1 for simple (non-collection) geometry types.
#'
#' - `ddbs_get_ninterior_rings()` returns the number of interior rings (holes)
#'   in a POLYGON geometry. Returns 0 for polygons without holes and `NULL` for
#'   non-polygon geometry types.
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(dplyr)
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## count points, sub-geometries and interior rings
#' ddbs_get_npoints(countries_ddbs)
#' ddbs_get_ngeometries(countries_ddbs)
#' ddbs_get_ninterior_rings(countries_ddbs)
#' }
#'
#' @name ddbs_get_npoints
#' @rdname ddbs_get_npoints
NULL



#' @rdname ddbs_get_npoints
#' @export
ddbs_get_npoints <- function(
  x,
  new_column = "npoints",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_NumPoints"
  )

}



#' @rdname ddbs_get_npoints
#' @export
ddbs_get_ngeometries <- function(
  x,
  new_column = "ngeometries",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_NumGeometries"
  )

}



#' @rdname ddbs_get_npoints
#' @export
ddbs_get_ninterior_rings <- function(
  x,
  new_column = "ninterior_rings",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_NumInteriorRings"
  )

}




#' Get the topological dimension of a geometry
#'
#' Returns the topological dimension of each geometry: 0 for points, 1 for
#' lines, 2 for polygons, and -1 for empty geometries.
#'
#' @template x
#' @template new_column
#' @template conn_null
#' @template name
#' @template mode
#' @template overwrite
#' @template quiet
#'
#' @template returns_mode
#'
#' @examples
#' \dontrun{
#' ## load packages
#' library(dplyr)
#' library(duckspatial)
#'
#' ## read data
#' countries_ddbs <- ddbs_open_dataset(
#'   system.file("spatial/countries.geojson",
#'   package = "duckspatial")
#' )
#'
#' ## get topological dimension (2 for polygons)
#' ddbs_dimension(countries_ddbs)
#' }
#'
#' @export
ddbs_dimension <- function(
  x,
  new_column = "dimension",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE) {

  template_new_column(
    x = x,
    new_column = new_column,
    conn = conn,
    name = name,
    mode = mode,
    overwrite = overwrite,
    quiet = quiet,
    fun = "ST_Dimension"
  )

}

# 0. Set up --------------------------------------------------------------

## skip tests on CRAN because they take too much time
skip_if(Sys.getenv("TEST_ONE") != "")
testthat::skip_on_cran()
testthat::skip_if_not_installed("duckdb")

## create duckdb connection
conn_test <- duckspatial::ddbs_create_conn()

## known-vertex polygon: 4 corners + ring closure = 5 points; ngeometries = 1
simple_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(
    sf::st_polygon(list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol = 2, byrow = TRUE)))
  ),
  crs = 4326
)
simple_ddbs <- as_duckspatial_df(simple_sf)

## two-part MULTIPOLYGON: ngeometries = 2
two_poly_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(sf::st_multipolygon(list(
    list(matrix(c(0,0, 1,0, 1,1, 0,1, 0,0), ncol = 2, byrow = TRUE)),
    list(matrix(c(2,2, 3,2, 3,3, 2,3, 2,2), ncol = 2, byrow = TRUE))
  ))),
  crs = 4326
)
two_poly_ddbs <- as_duckspatial_df(two_poly_sf)

## polygon with one interior ring (hole): ninterior_rings = 1
hole_poly_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(sf::st_polygon(list(
    matrix(c(0,0, 3,0, 3,3, 0,3, 0,0), ncol = 2, byrow = TRUE),
    matrix(c(1,1, 2,1, 2,2, 1,2, 1,1), ncol = 2, byrow = TRUE)
  ))),
  crs = 4326
)
hole_poly_ddbs <- as_duckspatial_df(hole_poly_sf)

## point (dimension 0) and linestring (dimension 1) for ddbs_dimension tests
point_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(sf::st_point(c(0, 0))),
  crs = 4326
)
point_ddbs <- as_duckspatial_df(point_sf)

line_sf <- sf::st_sf(
  id = 1L,
  geometry = sf::st_sfc(sf::st_linestring(matrix(c(0,0, 1,1), ncol = 2, byrow = TRUE))),
  crs = 4326
)
line_ddbs <- as_duckspatial_df(line_sf)

## write data in the database
duckspatial::ddbs_write_table(conn_test, simple_sf,   "simple")
duckspatial::ddbs_write_table(conn_test, two_poly_sf, "two_poly")
duckspatial::ddbs_write_table(conn_test, hole_poly_sf, "hole_poly")
duckspatial::ddbs_write_table(conn_test, point_sf,    "point_dim")
duckspatial::ddbs_write_table(conn_test, line_sf,     "line_dim")


# 1. ddbs_get_npoints ----------------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns duckspatial_df by default
## - CHECK 1.2: returns correct output formats
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: messages shown/suppressed correctly
## - CHECK 1.5: counts vertices correctly (5 for a closed rectangle ring)
## - CHECK 1.6: materializes data correctly
## - CHECK 1.7: vector and column outputs are identical
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns duckspatial_df by default
## - CHECK 2.2: returns numeric with mode sf
## - CHECK 2.3: duckspatial_df is written into the database
## - CHECK 2.4: messages shown/suppressed correctly
## - CHECK 2.5: warns when creating table from a different connection
## - CHECK 2.6: materializes data correctly
## - CHECK 2.7: vector and column outputs are identical
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns duckspatial_df by default
## - CHECK 3.2: returns numeric with mode sf
## - CHECK 3.3: table is written into the database
## - CHECK 3.4: messages shown/suppressed correctly
## - CHECK 3.5: counts vertices correctly
## - CHECK 3.6: conn = NULL errors
## - CHECK 3.7: materializes data correctly
## - CHECK 3.8: vector and column outputs are identical
## Check that errors work
## - CHECK 4.1: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.2: incorrect x type
## - CHECK 4.3: conn required when x is a table name
describe("ddbs_get_npoints()", {

  ### EXPECTED BEHAVIOUR - SF INPUT

  describe("expected behavior on sf input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_npoints(simple_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_npoints(simple_sf, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_get_npoints(simple_sf, mode = NULL)
      output_sf   <- ddbs_get_npoints(simple_sf, mode = "sf")

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_true(is.numeric(output_sf) || is.integer(output_sf))
    })

    it("writes tables to the database", {
      output <- ddbs_get_npoints(simple_sf, conn = conn_test, name = "np_sf_tbl", new_column = "np")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_npoints(simple_sf))
      expect_message(ddbs_get_npoints(simple_sf, conn = conn_test, name = "np_sf_tbl2", new_column = "np"))

      expect_no_message(ddbs_get_npoints(simple_sf, quiet = TRUE))
      expect_no_message(ddbs_get_npoints(simple_sf, conn = conn_test, name = "np_sf_tbl3", new_column = "np", quiet = TRUE))
    })

    it("counts vertices correctly (closed rectangle ring: 5 points)", {
      result <- ddbs_get_npoints(simple_sf, mode = "sf")
      expect_equal(as.integer(result), 5L)
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_npoints(simple_sf)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_npoints(simple_sf, mode = "sf"))
      output_column <- ddbs_get_npoints(simple_sf) |> ddbs_collect()

      expect_equal(output_vec, output_column$npoints)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT

  describe("expected behavior on duckspatial_df input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_npoints(simple_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_npoints(simple_ddbs, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_npoints(simple_ddbs, conn = conn_test, name = "np_ddbs_tbl", new_column = "np")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_npoints(simple_ddbs))
      expect_message(ddbs_get_npoints(simple_ddbs, conn = conn_test, name = "np_ddbs_tbl2", new_column = "np"))

      expect_no_message(ddbs_get_npoints(simple_ddbs, quiet = TRUE))
      expect_no_message(ddbs_get_npoints(simple_ddbs, conn = conn_test, name = "np_ddbs_tbl3", new_column = "np", quiet = TRUE))
    })

    it("warns when creating table from different connections", {
      expect_warning(ddbs_get_npoints(simple_ddbs, conn = conn_test, name = "np_ddbs_tbl4", new_column = "np"))
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_npoints(simple_ddbs)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_npoints(simple_ddbs, mode = "sf"))
      output_column <- ddbs_get_npoints(simple_ddbs) |> ddbs_collect()

      expect_equal(output_vec, output_column$npoints)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT

  describe("expected behavior on DuckDB table input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_npoints("simple", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_npoints("simple", conn = conn_test, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_npoints("simple", conn = conn_test, name = "np_tbl_tbl", new_column = "np")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_npoints("simple", conn = conn_test))
      expect_message(ddbs_get_npoints("simple", conn = conn_test, name = "np_tbl_tbl2", new_column = "np"))

      expect_no_message(ddbs_get_npoints("simple", conn = conn_test, quiet = TRUE))
      expect_no_message(ddbs_get_npoints("simple", conn = conn_test, name = "np_tbl_tbl3", new_column = "np", quiet = TRUE))
    })

    it("counts vertices correctly (closed rectangle ring: 5 points)", {
      result <- ddbs_get_npoints("simple", conn = conn_test, mode = "sf")
      expect_equal(as.integer(result), 5L)
    })

    it("requires conn when using table names", {
      expect_error(ddbs_get_npoints("simple", conn = NULL))
    })

    it("materializes data correctly (collect, ddbs_collect)", {
      output_with_column <- ddbs_get_npoints("simple", conn = conn_test)

      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_collect, output_ddbs)
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_npoints("simple", conn = conn_test, mode = "sf"))
      output_column <- ddbs_get_npoints("simple", conn = conn_test) |> ddbs_collect()

      expect_equal(output_vec, output_column$npoints)
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_get_npoints(simple_sf, conn = conn_test, name = "dup_np_tbl", new_column = "np")
      expect_error(
        ddbs_get_npoints(simple_sf, conn = conn_test, name = "dup_np_tbl", new_column = "np")
      )
    })

    it("errors on invalid x type", {
      expect_error(ddbs_get_npoints(999))
      expect_error(ddbs_get_npoints(TRUE))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_get_npoints("simple"))
    })
  })
})


# 2. ddbs_get_ngeometries ------------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns duckspatial_df by default
## - CHECK 1.2: returns correct output formats
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: messages shown/suppressed correctly
## - CHECK 1.5: returns 1 for a simple polygon
## - CHECK 1.6: returns 2 for a two-part MULTIPOLYGON
## - CHECK 1.7: materializes data correctly
## - CHECK 1.8: vector and column outputs are identical
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns duckspatial_df by default
## - CHECK 2.2: returns numeric with mode sf
## - CHECK 2.3: duckspatial_df is written into the database
## - CHECK 2.4: messages shown/suppressed correctly
## - CHECK 2.5: warns when creating table from a different connection
## - CHECK 2.6: materializes data correctly
## - CHECK 2.7: vector and column outputs are identical
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns duckspatial_df by default
## - CHECK 3.2: returns numeric with mode sf
## - CHECK 3.3: table is written into the database
## - CHECK 3.4: messages shown/suppressed correctly
## - CHECK 3.5: returns 2 for a two-part MULTIPOLYGON
## - CHECK 3.6: conn = NULL errors
## - CHECK 3.7: materializes data correctly
## - CHECK 3.8: vector and column outputs are identical
## Check that errors work
## - CHECK 4.1: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.2: incorrect x type
## - CHECK 4.3: conn required when x is a table name
describe("ddbs_get_ngeometries()", {

  ### EXPECTED BEHAVIOUR - SF INPUT

  describe("expected behavior on sf input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ngeometries(simple_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ngeometries(simple_sf, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_get_ngeometries(simple_sf, mode = NULL)
      output_sf   <- ddbs_get_ngeometries(simple_sf, mode = "sf")

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_true(is.numeric(output_sf) || is.integer(output_sf))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ngeometries(simple_sf, conn = conn_test, name = "ng_sf_tbl", new_column = "ng")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ngeometries(simple_sf))
      expect_message(ddbs_get_ngeometries(simple_sf, conn = conn_test, name = "ng_sf_tbl2", new_column = "ng"))

      expect_no_message(ddbs_get_ngeometries(simple_sf, quiet = TRUE))
      expect_no_message(ddbs_get_ngeometries(simple_sf, conn = conn_test, name = "ng_sf_tbl3", new_column = "ng", quiet = TRUE))
    })

    it("returns 1 for a simple polygon", {
      result <- ddbs_get_ngeometries(simple_sf, mode = "sf")
      expect_equal(as.integer(result), 1L)
    })

    it("returns 2 for a two-part MULTIPOLYGON", {
      result <- ddbs_get_ngeometries(two_poly_sf, mode = "sf")
      expect_equal(as.integer(result), 2L)
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ngeometries(simple_sf)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ngeometries(two_poly_sf, mode = "sf"))
      output_column <- ddbs_get_ngeometries(two_poly_sf) |> ddbs_collect()

      expect_equal(output_vec, output_column$ngeometries)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT

  describe("expected behavior on duckspatial_df input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ngeometries(simple_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ngeometries(simple_ddbs, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ngeometries(simple_ddbs, conn = conn_test, name = "ng_ddbs_tbl", new_column = "ng")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ngeometries(simple_ddbs))
      expect_message(ddbs_get_ngeometries(simple_ddbs, conn = conn_test, name = "ng_ddbs_tbl2", new_column = "ng"))

      expect_no_message(ddbs_get_ngeometries(simple_ddbs, quiet = TRUE))
      expect_no_message(ddbs_get_ngeometries(simple_ddbs, conn = conn_test, name = "ng_ddbs_tbl3", new_column = "ng", quiet = TRUE))
    })

    it("warns when creating table from different connections", {
      expect_warning(ddbs_get_ngeometries(simple_ddbs, conn = conn_test, name = "ng_ddbs_tbl4", new_column = "ng"))
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ngeometries(simple_ddbs)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ngeometries(two_poly_ddbs, mode = "sf"))
      output_column <- ddbs_get_ngeometries(two_poly_ddbs) |> ddbs_collect()

      expect_equal(output_vec, output_column$ngeometries)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT

  describe("expected behavior on DuckDB table input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ngeometries("two_poly", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ngeometries("two_poly", conn = conn_test, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ngeometries("two_poly", conn = conn_test, name = "ng_tbl_tbl", new_column = "ng")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ngeometries("two_poly", conn = conn_test))
      expect_message(ddbs_get_ngeometries("two_poly", conn = conn_test, name = "ng_tbl_tbl2", new_column = "ng"))

      expect_no_message(ddbs_get_ngeometries("two_poly", conn = conn_test, quiet = TRUE))
      expect_no_message(ddbs_get_ngeometries("two_poly", conn = conn_test, name = "ng_tbl_tbl3", new_column = "ng", quiet = TRUE))
    })

    it("returns 2 for a two-part MULTIPOLYGON", {
      result <- ddbs_get_ngeometries("two_poly", conn = conn_test, mode = "sf")
      expect_equal(as.integer(result), 2L)
    })

    it("requires conn when using table names", {
      expect_error(ddbs_get_ngeometries("two_poly", conn = NULL))
    })

    it("materializes data correctly (collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ngeometries("two_poly", conn = conn_test)

      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_collect, output_ddbs)
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ngeometries("two_poly", conn = conn_test, mode = "sf"))
      output_column <- ddbs_get_ngeometries("two_poly", conn = conn_test) |> ddbs_collect()

      expect_equal(output_vec, output_column$ngeometries)
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_get_ngeometries(simple_sf, conn = conn_test, name = "dup_ng_tbl", new_column = "ng")
      expect_error(
        ddbs_get_ngeometries(simple_sf, conn = conn_test, name = "dup_ng_tbl", new_column = "ng")
      )
    })

    it("errors on invalid x type", {
      expect_error(ddbs_get_ngeometries(999))
      expect_error(ddbs_get_ngeometries(TRUE))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_get_ngeometries("simple"))
    })
  })
})


# 3. ddbs_dimension ------------------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns duckspatial_df by default
## - CHECK 1.2: returns correct output formats
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: messages shown/suppressed correctly
## - CHECK 1.5: returns 2 for a polygon
## - CHECK 1.6: returns 1 for a linestring
## - CHECK 1.7: returns 0 for a point
## - CHECK 1.8: materializes data correctly
## - CHECK 1.9: vector and column outputs are identical
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns duckspatial_df by default
## - CHECK 2.2: returns numeric with mode sf
## - CHECK 2.3: duckspatial_df is written into the database
## - CHECK 2.4: messages shown/suppressed correctly
## - CHECK 2.5: warns when creating table from a different connection
## - CHECK 2.6: materializes data correctly
## - CHECK 2.7: vector and column outputs are identical
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns duckspatial_df by default
## - CHECK 3.2: returns numeric with mode sf
## - CHECK 3.3: table is written into the database
## - CHECK 3.4: messages shown/suppressed correctly
## - CHECK 3.5: returns 2 for a polygon
## - CHECK 3.6: conn = NULL errors
## - CHECK 3.7: materializes data correctly
## - CHECK 3.8: vector and column outputs are identical
## Check that errors work
## - CHECK 4.1: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.2: incorrect x type
## - CHECK 4.3: conn required when x is a table name
describe("ddbs_dimension()", {

  ### EXPECTED BEHAVIOUR - SF INPUT

  describe("expected behavior on sf input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_dimension(simple_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_dimension(simple_sf, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_dimension(simple_sf, mode = NULL)
      output_sf   <- ddbs_dimension(simple_sf, mode = "sf")

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_true(is.numeric(output_sf) || is.integer(output_sf))
    })

    it("writes tables to the database", {
      output <- ddbs_dimension(simple_sf, conn = conn_test, name = "dim_sf_tbl", new_column = "dim")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_dimension(simple_sf))
      expect_message(ddbs_dimension(simple_sf, conn = conn_test, name = "dim_sf_tbl2", new_column = "dim"))

      expect_no_message(ddbs_dimension(simple_sf, quiet = TRUE))
      expect_no_message(ddbs_dimension(simple_sf, conn = conn_test, name = "dim_sf_tbl3", new_column = "dim", quiet = TRUE))
    })

    it("returns 2 for a polygon", {
      result <- ddbs_dimension(simple_sf, mode = "sf")
      expect_equal(as.integer(result), 2L)
    })

    it("returns 1 for a linestring", {
      result <- ddbs_dimension(line_sf, mode = "sf")
      expect_equal(as.integer(result), 1L)
    })

    it("returns 0 for a point", {
      result <- ddbs_dimension(point_sf, mode = "sf")
      expect_equal(as.integer(result), 0L)
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_dimension(simple_sf)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_dimension(simple_sf, mode = "sf"))
      output_column <- ddbs_dimension(simple_sf) |> ddbs_collect()

      expect_equal(output_vec, output_column$dimension)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT

  describe("expected behavior on duckspatial_df input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_dimension(simple_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_dimension(simple_ddbs, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_dimension(simple_ddbs, conn = conn_test, name = "dim_ddbs_tbl", new_column = "dim")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_dimension(simple_ddbs))
      expect_message(ddbs_dimension(simple_ddbs, conn = conn_test, name = "dim_ddbs_tbl2", new_column = "dim"))

      expect_no_message(ddbs_dimension(simple_ddbs, quiet = TRUE))
      expect_no_message(ddbs_dimension(simple_ddbs, conn = conn_test, name = "dim_ddbs_tbl3", new_column = "dim", quiet = TRUE))
    })

    it("warns when creating table from different connections", {
      expect_warning(ddbs_dimension(simple_ddbs, conn = conn_test, name = "dim_ddbs_tbl4", new_column = "dim"))
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_dimension(simple_ddbs)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_dimension(simple_ddbs, mode = "sf"))
      output_column <- ddbs_dimension(simple_ddbs) |> ddbs_collect()

      expect_equal(output_vec, output_column$dimension)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT

  describe("expected behavior on DuckDB table input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_dimension("simple", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_dimension("simple", conn = conn_test, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_dimension("simple", conn = conn_test, name = "dim_tbl_tbl", new_column = "dim")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_dimension("simple", conn = conn_test))
      expect_message(ddbs_dimension("simple", conn = conn_test, name = "dim_tbl_tbl2", new_column = "dim"))

      expect_no_message(ddbs_dimension("simple", conn = conn_test, quiet = TRUE))
      expect_no_message(ddbs_dimension("simple", conn = conn_test, name = "dim_tbl_tbl3", new_column = "dim", quiet = TRUE))
    })

    it("returns 2 for a polygon table", {
      result <- ddbs_dimension("simple", conn = conn_test, mode = "sf")
      expect_equal(as.integer(result), 2L)
    })

    it("requires conn when using table names", {
      expect_error(ddbs_dimension("simple", conn = NULL))
    })

    it("materializes data correctly (collect, ddbs_collect)", {
      output_with_column <- ddbs_dimension("simple", conn = conn_test)

      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_collect, output_ddbs)
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_dimension("simple", conn = conn_test, mode = "sf"))
      output_column <- ddbs_dimension("simple", conn = conn_test) |> ddbs_collect()

      expect_equal(output_vec, output_column$dimension)
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_dimension(simple_sf, conn = conn_test, name = "dup_dim_tbl", new_column = "dim")
      expect_error(
        ddbs_dimension(simple_sf, conn = conn_test, name = "dup_dim_tbl", new_column = "dim")
      )
    })

    it("errors on invalid x type", {
      expect_error(ddbs_dimension(999))
      expect_error(ddbs_dimension(TRUE))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_dimension("simple"))
    })
  })
})


# 4. ddbs_get_ninterior_rings --------------------------------------------

## expected behaviour for inherits(x, "sf")
## - CHECK 1.1: returns duckspatial_df by default
## - CHECK 1.2: returns correct output formats
## - CHECK 1.3: sf is written into the database
## - CHECK 1.4: messages shown/suppressed correctly
## - CHECK 1.5: returns 1 for a polygon with one hole
## - CHECK 1.6: returns 0 for a polygon without holes
## - CHECK 1.7: materializes data correctly
## - CHECK 1.8: vector and column outputs are identical
## expected behaviour for inherits(x, "duckspatial_df")
## - CHECK 2.1: returns duckspatial_df by default
## - CHECK 2.2: returns numeric with mode sf
## - CHECK 2.3: duckspatial_df is written into the database
## - CHECK 2.4: messages shown/suppressed correctly
## - CHECK 2.5: warns when creating table from a different connection
## - CHECK 2.6: materializes data correctly
## - CHECK 2.7: vector and column outputs are identical
## expected behaviour for inherits(x, "character")
## - CHECK 3.1: returns duckspatial_df by default
## - CHECK 3.2: returns numeric with mode sf
## - CHECK 3.3: table is written into the database
## - CHECK 3.4: messages shown/suppressed correctly
## - CHECK 3.5: returns 1 for a polygon with one hole
## - CHECK 3.6: conn = NULL errors
## - CHECK 3.7: materializes data correctly
## - CHECK 3.8: vector and column outputs are identical
## Check that errors work
## - CHECK 4.1: if overwrite = FALSE, it won't delete an existing table
## - CHECK 4.2: incorrect x type
## - CHECK 4.3: conn required when x is a table name
describe("ddbs_get_ninterior_rings()", {

  ### EXPECTED BEHAVIOUR - SF INPUT

  describe("expected behavior on sf input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ninterior_rings(hole_poly_sf)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ninterior_rings(hole_poly_sf, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("returns different output formats (duckspatial_df, sf)", {
      output_ddbs <- ddbs_get_ninterior_rings(hole_poly_sf, mode = NULL)
      output_sf   <- ddbs_get_ninterior_rings(hole_poly_sf, mode = "sf")

      expect_s3_class(output_ddbs, "duckspatial_df")
      expect_true(is.numeric(output_sf) || is.integer(output_sf))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ninterior_rings(hole_poly_sf, conn = conn_test, name = "nir_sf_tbl", new_column = "nir")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ninterior_rings(hole_poly_sf))
      expect_message(ddbs_get_ninterior_rings(hole_poly_sf, conn = conn_test, name = "nir_sf_tbl2", new_column = "nir"))

      expect_no_message(ddbs_get_ninterior_rings(hole_poly_sf, quiet = TRUE))
      expect_no_message(ddbs_get_ninterior_rings(hole_poly_sf, conn = conn_test, name = "nir_sf_tbl3", new_column = "nir", quiet = TRUE))
    })

    it("returns 1 for a polygon with one hole", {
      result <- ddbs_get_ninterior_rings(hole_poly_sf, mode = "sf")
      expect_equal(as.integer(result), 1L)
    })

    it("returns 0 for a polygon without holes", {
      result <- ddbs_get_ninterior_rings(simple_sf, mode = "sf")
      expect_equal(as.integer(result), 0L)
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ninterior_rings(hole_poly_sf)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ninterior_rings(hole_poly_sf, mode = "sf"))
      output_column <- ddbs_get_ninterior_rings(hole_poly_sf) |> ddbs_collect()

      expect_equal(output_vec, output_column$ninterior_rings)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKSPATIAL_DF INPUT

  describe("expected behavior on duckspatial_df input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ninterior_rings(hole_poly_ddbs)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ninterior_rings(hole_poly_ddbs, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ninterior_rings(hole_poly_ddbs, conn = conn_test, name = "nir_ddbs_tbl", new_column = "nir")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ninterior_rings(hole_poly_ddbs))
      expect_message(ddbs_get_ninterior_rings(hole_poly_ddbs, conn = conn_test, name = "nir_ddbs_tbl2", new_column = "nir"))

      expect_no_message(ddbs_get_ninterior_rings(hole_poly_ddbs, quiet = TRUE))
      expect_no_message(ddbs_get_ninterior_rings(hole_poly_ddbs, conn = conn_test, name = "nir_ddbs_tbl3", new_column = "nir", quiet = TRUE))
    })

    it("warns when creating table from different connections", {
      expect_warning(ddbs_get_ninterior_rings(hole_poly_ddbs, conn = conn_test, name = "nir_ddbs_tbl4", new_column = "nir"))
    })

    it("materializes data correctly (st_as_sf, collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ninterior_rings(hole_poly_ddbs)

      output_sf      <- output_with_column |> st_as_sf()
      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_sf, output_collect)
      expect_identical(output_collect, output_ddbs)
      expect_s3_class(output_sf, "sf")
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ninterior_rings(hole_poly_ddbs, mode = "sf"))
      output_column <- ddbs_get_ninterior_rings(hole_poly_ddbs) |> ddbs_collect()

      expect_equal(output_vec, output_column$ninterior_rings)
    })
  })

  ### EXPECTED BEHAVIOUR - DUCKDB TABLE INPUT

  describe("expected behavior on DuckDB table input", {

    it("returns a duckspatial_df by default", {
      output <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test)
      expect_s3_class(output, "duckspatial_df")
    })

    it("returns a numeric vector with mode sf", {
      output <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test, mode = "sf")
      expect_true(is.numeric(output) || is.integer(output))
    })

    it("writes tables to the database", {
      output <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test, name = "nir_tbl_tbl", new_column = "nir")
      expect_true(output)
    })

    it("shows and suppresses messages correctly", {
      expect_no_message(ddbs_get_ninterior_rings("hole_poly", conn = conn_test))
      expect_message(ddbs_get_ninterior_rings("hole_poly", conn = conn_test, name = "nir_tbl_tbl2", new_column = "nir"))

      expect_no_message(ddbs_get_ninterior_rings("hole_poly", conn = conn_test, quiet = TRUE))
      expect_no_message(ddbs_get_ninterior_rings("hole_poly", conn = conn_test, name = "nir_tbl_tbl3", new_column = "nir", quiet = TRUE))
    })

    it("returns 1 for a polygon with one hole", {
      result <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test, mode = "sf")
      expect_equal(as.integer(result), 1L)
    })

    it("requires conn when using table names", {
      expect_error(ddbs_get_ninterior_rings("hole_poly", conn = NULL))
    })

    it("materializes data correctly (collect, ddbs_collect)", {
      output_with_column <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test)

      output_collect <- output_with_column |> collect()
      output_ddbs    <- output_with_column |> ddbs_collect()

      expect_identical(output_collect, output_ddbs)
    })

    it("produces identical results for vector and column outputs", {
      output_vec    <- as.integer(ddbs_get_ninterior_rings("hole_poly", conn = conn_test, mode = "sf"))
      output_column <- ddbs_get_ninterior_rings("hole_poly", conn = conn_test) |> ddbs_collect()

      expect_equal(output_vec, output_column$ninterior_rings)
    })
  })

  ### EXPECTED ERRORS

  describe("errors", {

    it("errors if overwrite = FALSE and table already exists", {
      ddbs_get_ninterior_rings(hole_poly_sf, conn = conn_test, name = "dup_nir_tbl", new_column = "nir")
      expect_error(
        ddbs_get_ninterior_rings(hole_poly_sf, conn = conn_test, name = "dup_nir_tbl", new_column = "nir")
      )
    })

    it("errors on invalid x type", {
      expect_error(ddbs_get_ninterior_rings(999))
      expect_error(ddbs_get_ninterior_rings(TRUE))
    })

    it("requires conn when x is a table name", {
      expect_error(ddbs_get_ninterior_rings("hole_poly"))
    })
  })
})


## stop connection
ddbs_stop_conn(conn_test)

---
title: "Benchmark"
date: "`r Sys.Date()`"
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Benchmark}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
execute:
  eval: false
knitr:
  opts_chunk:
    collapse: true
    comment: "#>"
    out.width: "100%"
---

```{r}
#| include: false

# Limit threads to avoid a CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)
```

```{r}
#| echo: false

## Internal packages
library(patchwork)

## Functions used in the vignette
ggplot_benchmark <- function(
  data, 
  log = FALSE,
  show.legend = TRUE,
  ...
  ) {

    ## Generate base plot
    ggplot(
      data = data, 
      aes(
        x     = factor(format(n, big.mark = ".")),
        y     = if (log) median else as.numeric(median),
        color = pkg,
        group = pkg
    )) +
      geom_line(
        linewidth = 0.7,
        linetype  = "dashed",
        alpha     = 0.6,
        show.legend = show.legend
      ) +
      geom_point(
        aes(size = as.numeric(mem_alloc)),
        alpha = 0.9,
        show.legend = show.legend
      ) +
      scale_size_binned(
        name   = "Memory allocated",
        labels = scales::label_bytes(),
        n.breaks = 4,
        range = c(1, 8)
      ) +
      scale_color_brewer(palette = "Set1") +
      labs(
        color    = "Package",
        x        = "Data size (rows)",
        y        = "Computation time",
        ...
      ) +
      theme_minimal(base_size = 13) +
      theme(
        # legend.position  = "top",
        panel.grid.minor = element_blank(),
        plot.title       = element_text(face = "bold")
      )
}

ggplot_assemble <- function(plot1, plot2, fun_name) {

  plot1 +
    plot2 +
    plot_annotation(
      title    = paste("Benchmark Comparison for", fun_name),
      subtitle = "Execution time vs. data size",
      theme    = theme(
        plot.title = element_text(face = "bold", hjust = .5),
        plot.subtitle = element_text(hjust = .5),
        text = element_text(size = rel(3.5))
      )
    )

}
```

This vignette benchmarks **{duckspatial}** against **{sf}** across several spatial operations, comparing computation time and memory usage as dataset size grows. We plan to extend it with additional operation types in future releases. Versions of packages used in this vignette:

* `duckspatial`: v1.0.0

* `duckdb`: v1.5.1

* `sf`: v1.1.0

## TL;DR

{duckspatial} is substantially faster and allocates far less memory than {sf} in almost all cases, with the advantage becoming more pronounced on larger datasets. The one exception is pairwise distance calculation, where {sf} retains a memory advantage for Euclidean distances.

## Prepare data

We are going to test the speed of a a bunch of functions using simulated data. We use the `make_points()` function to generate `n` random points within the globe, and 10,000 random rectangles.

```{r}
#| message: false
#| warning: false
#| code-fold: true
#| code-summary: "Set-up"

# Load necessary packages
library(duckspatial)
library(bench)
library(dplyr)
library(sf)
library(ggplot2)
options(scipen = 999)

# Function to generate random points
make_points <- function(n_points) {
    points_df <- data.frame(
      id = 1:n_points,
      x = runif(n_points, min = -180, max = 180),
      y = runif(n_points, min = -90, max = 90),
      value = rnorm(n_points, mean = 100, sd = 15),
      category = sample(c("A", "B", "C", "D"), n_points, replace = TRUE)
  ) |>
    sf::st_as_sf(coords = c("x", "y"), crs = 4326)
}

# Generate datasets of different sizes
withr::with_seed(27, {
  points_sf_100k <- make_points(1e5)
  points_sf_1mi  <- make_points(1e6)
  points_sf_3mi  <- make_points(3e6)
})

# Generate polygons
# Create large polygon dataset (e.g., administrative regions, zones, etc.)
n_polygons <- 10000
polygons_list <- vector("list", n_polygons)

for(i in 1:n_polygons) {
  # Random center point with buffer from edges
  center_x <- runif(1, min = -170, max = 170)
  center_y <- runif(1, min = -80, max = 80)
  
  # Create simple rectangular polygons to avoid geometry issues
  width <- runif(1, min = 0.5, max = 3)
  height <- runif(1, min = 0.5, max = 3)
  
  # Create rectangle coordinates (must be closed: first point = last point)
  x_coords <- c(
    center_x - width/2,
    center_x + width/2,
    center_x + width/2,
    center_x - width/2,
    center_x - width/2  # Close the polygon
  )
  
  y_coords <- c(
    center_y - height/2,
    center_y - height/2,
    center_y + height/2,
    center_y + height/2,
    center_y - height/2  # Close the polygon
  )
  
  # Create polygon matrix
  coords <- cbind(x_coords, y_coords)
  
  # Create polygon (wrapped in list as required by st_polygon)
  polygons_list[[i]] <- st_polygon(list(coords))
}

polygons_sf <- st_sf(
  poly_id    = 1:n_polygons,
  region     = sample(c("North", "South", "East", "West"), n_polygons, replace = TRUE),
  population = sample(1000:1000000, n_polygons, replace = TRUE),
  geometry   = st_sfc(polygons_list, crs = 4326)
)
```

# Spatial join

The spatial join aims to bring the attributes of a dataset to another dataset based in a spatial predicate. One example would be to have an `x` dataset with points that represent observations of wolves. In an `y` dataset, we could have polygons with attributes describing the geopraphical location (e.g. the name of the country, the name of the region..). So, by using a `ST_Join(x, y, "intersects")`, we would assign the attributes of `y` to `x` according to where the observation of the wolf fall.

This operation can be intensive. In the current version of {duckspatial} we see an improvement in speed and memory usage in bigger datasets, as shown in @fig-st-join.

-   Using 1 million points: {duckspatial} was about 50 times faster than {sf}, and allocated 18 times less memory.

-   Using 3 million points: {duckspatial} was about 35 times faster than {sf}, and allocated 16 times less memory.

```{r}
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_join"

# Helper to run the benchmark
run_join_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_join(points_sf, polygons_sf, join = "within"),
    sf          = st_join(points_sf, polygons_sf, join = st_within)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_join <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_join_benchmark
) |>
  dplyr::bind_rows()
```

```{r}
#| echo: false
#| warning: false

# Id to store the figures, saving the older ones
id_output <- "v1.5.1"

# Generate the plots
gg_join <- ggplot_benchmark(
  data     = df_bench_join,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_join_log <- ggplot_benchmark(
  data     = df_bench_join,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_join,
  plot2 = gg_join_log,
  fun_name = "ddbs_join()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-join-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)
```

![Benchmark that compares spatial join for duckspatial and sf](../man/figures/bench/bench-st-join-v1.5.1.png){#fig-st-join fig-align="center"}

# Spatial filter

The spatial filter aims to filter rows of `x` based in a spatial relationship with `y`. For example, let's imagine that we have an `x` dataset with observations of wolves all around the world, and we want to filter only those that are in a specific country, for instance in Spain. If the dataset has this attritbute, we can just do it with a simple `dplyr::filter()`, but, if the wolves datase doesn't include a column with the country, we can use a spatial filter. For that, we need a second dataset `y` with the boundaries of Spain, and by using `ST_Filter(x, y, "intersects")`, we would filter only those wolves observations that intersects with the Spain's polygon.

In the current version of {duckspatial} we see an improvement in speed and memory usage in bigger datasets, as shown in @fig-st-filter.

-   Using 1 million points: {duckspatial} was about 42 times faster than {sf}, and allocated 5 times less memory.

-   Using 3 million points: {duckspatial} was about 12 times faster than {sf}, and allocated 5 times less memory.

```{r}
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_filter"

# Helper to run the benchmark
run_filter_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_filter(points_sf, polygons_sf),
    sf          = st_filter(points_sf, polygons_sf)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_filter <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_filter_benchmark
) |>
  dplyr::bind_rows()
```

```{r}
#| echo: false
#| warning: false

# Generate the plots
gg_filter <- ggplot_benchmark(
  data     = df_bench_filter,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_filter_log <- ggplot_benchmark(
  data     = df_bench_filter,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_filter,
  plot2 = gg_filter_log,
  fun_name = "ddbs_filter()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-filter-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)
```

![Benchmark that compares spatial filter for duckspatial and sf](../man/figures/bench/bench-st-filter-v1.5.1.png){#fig-st-filter fig-align="center"}

# Spatial distances

The `ST_Distance(x, y)` calculates the distance between each observation in `x` against each observation in `y`. The default {duckspatial} mode will return a lazy table with three columns: (id_x) the id of the row in `x`; (id_y) the id of the row in `y`; (distance) the actual distance between those pair of observations. In the case of `mode = 'sf'`, the result will be a sparse matrix. Note that {duckspatial} will use by default the best distance for the input CRS and geometry type.

When calculating the distance between 10,000 pairs of points, {duckspatial} was about 1.5 times faster, but used much less memory (> 400 times).

```{r}
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_distance"

# Helper to run the benchmark
run_distance_benchmark <- function(n) {

  points_sf <- withr::with_seed(27, make_points(n))

  temp <- bench::mark(
    iterations  = 1,
    check       = FALSE,
    duckspatial = ddbs_distance(points_sf, points_sf),
    sf          = st_distance(points_sf, points_sf)
  )
  temp$n   <- n
  temp$pkg <- c("duckspatial", "sf")
  temp
}

df_bench_distance <- lapply(
  X   = c(1000, 5000, 10000),
  FUN = run_distance_benchmark
) |>
  dplyr::bind_rows()
```


```{r}
#| echo: false
#| warning: false

# Generate the plots
gg_distance <- ggplot_benchmark(
  data     = df_bench_distance,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_distance_log <- ggplot_benchmark(
  data     = df_bench_distance,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_distance,
  plot2 = gg_distance_log,
  fun_name = "ddbs_distance()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-distance-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)
```

![Benchmark that compares spatial distance for duckspatial and sf](../man/figures/bench/bench-st-distance-v1.5.1.png){#fig-st-distance fig-align="center"}


# Dissolving geometries

Dissolving geometries consist in merging/aggregating geometries that share a common attribute into a single geometry per group.

In the current version of {duckspatial} we see an improvement in speed and memory usage in bigger datasets, as shown in @fig-st-dissolve.

-   Using 1 million points: {duckspatial} was about 14 times faster than {sf}, and allocated 10 times less memory.

-   Using 3 million points: {duckspatial} was about 12 times faster than {sf}, and allocated 10 times less memory.

```{r}
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_union_agg"

# Helper to run the benchmark
run_union_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 3,
    check       = FALSE,
    duckspatial = ddbs_union_agg(points_sf, by = "category"),
    sf          = points_sf |> 
      group_by(category) |> 
      summarise(geometry = st_union(geometry))
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_union <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_union_benchmark
) |>
  dplyr::bind_rows()
```

```{r}
#| echo: false
#| warning: false

# Generate the plots
gg_union_agg <- ggplot_benchmark(
  data     = df_bench_union,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_union_agg_log <- ggplot_benchmark(
  data     = df_bench_union,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_union_agg,
  plot2 = gg_union_agg_log,
  fun_name = "ddbs_union_agg()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-dissolve-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)
```

![Benchmark that compares dissolving geometries by group for duckspatial and sf](../man/figures/bench/bench-st-dissolve-v1.5.1.png){#fig-st-dissolve fig-align="center"}


# Geometry predicates

Geometry predicates are boolean functions that test spatial relationships between two geometries (they return true or false). They're fundamental to spatial queries, letting you filter or join data based on how geometries relate to one another in space.
Common examples include: ST_Contains, ST_Within, ST_Touches, ST_Overlaps, ST_Crosses, ST_Disjoint, and ST_Equals.

In the following example, we compare `ST_Intersects(geometry A, geometry B)`, which returns true if the two geometries share any point in common (their interiors or boundaries overlap in any way).

-   Using 1 million points: {duckspatial} was about 28 times faster than {sf}, and allocated 3 times less memory.

-   Using 3 million points: {duckspatial} was about 23 times faster than {sf}, and allocated 3 times less memory.

```{r}
#| message: false
#| code-fold: true
#| code-summary: "Benchmark code - ddbs_intersects"

# Helper to run the benchmark
run_predicate_benchmark <- function(points_sf) {
  temp <- bench::mark(
    iterations  = 1,
    check       = FALSE,
    duckspatial = ddbs_intersects(points_sf, polygons_sf),
    sf          = st_intersects(points_sf, polygons_sf)
  )
  temp$n   <- nrow(points_sf)
  temp$pkg <- c("duckspatial", "sf")
  temp
}

# Run the benchmark
df_bench_predicate <- lapply(
  X   = list(points_sf_100k, points_sf_1mi, points_sf_3mi),
  FUN = run_predicate_benchmark
) |>
  dplyr::bind_rows()
```

```{r}
#| echo: false
#| warning: false

# Generate the plots
gg_predicate <- ggplot_benchmark(
  data     = df_bench_predicate,
  log      = FALSE,
  show.legend = T,
  subtitle = "A) Normal scale"
)

gg_predicate_log <- ggplot_benchmark(
  data     = df_bench_predicate,
  log      = TRUE,
  show.legend = F,
  subtitle = "B) Log scale"
)

# Assemble them in a single plot
ggplot_assemble(
  plot1 = gg_predicate,
  plot2 = gg_predicate_log,
  fun_name = "ddbs_intersects()"
)

# Export it
ggsave(
  filename = paste0("man/figures/bench/bench-st-intersects-", id_output, ".png"),
  height   = 15,
  width    = 30,
  units    = "cm"
)
```

![Benchmark that compares the geometry predicate ST_Intersects(x, y) for duckspatial and sf](../man/figures/bench/bench-st-intersects-v1.5.1.png){#fig-st-intersects fig-align="center"}

---
title: "Getting started"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Getting started}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
execute:
  eval: !expr identical(tolower(Sys.getenv("NOT_CRAN")), "true")
knitr:
  opts_chunk:
    collapse: true
    comment: "#>"
    out.width: "100%"
---

```{r}
#| include: false

# Limit threads to avoid a CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

# Quarto spawns a new R subprocess during R CMD build and may not pick up
# R_LIBS set by the build process; propagate it manually.
local({
  r_libs <- Sys.getenv("R_LIBS")
  if (nzchar(r_libs)) {
    paths <- unlist(strsplit(r_libs, .Platform$path.sep))
    .libPaths(unique(c(paths, .libPaths())))
  }
})
```

# Getting started

**{duckspatial}** provides fast, memory-efficient functions for analysing and manipulating large spatial vector datasets in R. It bridges [DuckDB's spatial extension](https://duckdb.org/docs/stable/core_extensions/spatial/functions) with R's spatial ecosystem — in particular **{sf}** — so you can leverage DuckDB's analytical power without leaving your familiar R workflow.

Let's start by loading the packages we need:

```{r}
#| label: setup
#| message: false
#| warning: false

library(duckdb)
library(duckspatial)
library(dplyr)
library(sf)
```

# Installation

Install the stable release from CRAN:

``` r
# install.packages("pak")
pak::pak("duckspatial")
```

Install the latest GitHub version (more features, fewer accumulated bugs):

``` r
pak::pak("Cidree/duckspatial")
```

Install the development version (may be unstable):

``` r
pak::pak("Cidree/duckspatial@dev")
```

# Reading data

{duckspatial} is built around the `duckspatial_df` S3 class: a lazy-table-like object that holds a geometry column alongside its geospatial properties, but keeps the data **outside R's memory** until you explicitly ask for it.

If you have a local file (a GeoPackage, a Shapefile, a GeoJSON, etc.) you can open it lazily with `ddbs_open_dataset()`:

```{r}
countries_ddbs <- ddbs_open_dataset(
  system.file(
    "spatial/countries.geojson",
    package = "duckspatial"
  )
)

print(countries_ddbs)
```

> **Note:** The first call to `ddbs_open_dataset()` may take a few seconds. Internally, {duckspatial} creates a default DuckDB connection, then installs and loads the Spatial extension into it. Subsequent calls reuse the same connection and are much faster. We cover this connection in more detail in the [Working in a database] section below.

Printing a `duckspatial_df` object displays the most important metadata at a glance:

| Field | Description |
|-------------------------|------------------------------------|
| **CRS** | Coordinate reference system (AUTHORITY:CODE) |
| **Geometry column** | Name of the column holding geometries |
| **Geometry type** | Type(s) present (e.g. `POLYGON`, `MULTIPOLYGON`, `POINT`) |
| **Bounding box** | Four coordinates bounding all geometries |
| **Source** | Name of the temporary view inside DuckDB |
| **Database** | DuckDB database path and version |
| **Data** | First rows of the dataset |

The table is not in R's memory; it lives inside the DuckDB connection. Every {duckspatial} operation you apply runs there, using the DuckDB engine.

Alternatively, if you already have an `sf` object in memory, you can convert it to a `duckspatial_df` with `as_duckspatial_df()`:

```{r}
## read with sf as usual
countries_sf <- read_sf(
  system.file(
    "spatial/countries.geojson",
    package = "duckspatial"
  )
)

## push into DuckDB
countries_ddbs <- as_duckspatial_df(countries_sf)

class(countries_ddbs)
```

# Processing data

Let's run a typical spatial workflow: dissolving all country polygons into a single `MULTIPOLYGON` with internal boundaries removed, using `ddbs_union()` (the {duckspatial} equivalent of `sf::st_union()`).

`ddbs_union()` requires all geometries to be valid. We can check this first with `ddbs_is_valid()` (equivalent to `sf::st_is_valid()`), which appends a logical `is_valid` column to the lazy table so the subsequent `filter()` also runs inside DuckDB:

```{r}
countries_ddbs |>
  ddbs_is_valid() |>
  filter(!is_valid)
```

Antarctica has invalid geometries (likely self-intersections). We can repair them with `ddbs_make_valid()` before computing the union, and because `duckspatial_df` objects are lazy, we can chain both steps in a single pipe:

```{r}
world_ddbs <- countries_ddbs |>
  ddbs_make_valid() |>
  ddbs_union()

print(world_ddbs)
```

The result is still a lazy `duckspatial_df`. To visualise it we need to pull the data into R as an `sf` object. Any of the following three calls do that:

``` r
# Option A
world_sf <- ddbs_collect(world_ddbs)

# Option B
world_sf <- collect(world_ddbs)

# Option C
world_sf <- st_as_sf(world_ddbs)
```

```{r}
world_sf <- world_ddbs |>
  ddbs_collect()

print(world_sf)
```

```{r}
plot(world_sf)
```

# Working in a database

So far we have been using the default, temporary DuckDB connection that {duckspatial} manages for us. For some use cases you may want to manage the connection yourself, most commonly when you need a **persistent database** that survives the R session.

There are two connection modes:

-   **Non-persistent (in-memory):** data exists only for the duration of the R session or until it is closed. As of v1.0.0, this mode is kept for backward compatibility, but working with `duckspatial_df` objects directly achieves the same goals with less boilerplate.
-   **Persistent:** data is written to a DuckDB database file on disk and survives after the session ends. {duckspatial} accepts `.duckdb`, `.db`, and `.ddb` paths for persistent DuckDB databases.

## Creating a connection

{duckspatial} provides a convenience wrapper, `ddbs_create_conn()`, that creates a DuckDB connection, installs the Spatial extension, and loads it, all in one call:

```{r}
conn <- ddbs_create_conn()
```

You can also limit the resources DuckDB is allowed to use:

```{r}
conn <- ddbs_create_conn(
  threads         = 2,
  memory_limit_gb = 8
)
```

Under the hood, `ddbs_create_conn()` is equivalent to:

``` r
conn <- dbConnect(duckdb())
ddbs_install(conn)
ddbs_load(conn)
```

## Non-persistent database

Once you have a connection, write spatial data into it with `ddbs_write_table()`. It accepts both `sf` and `duckspatial_df` objects:

```{r}
ddbs_write_table(conn, countries_sf, name = "countries")
```

Verify the table is there:

```{r}
ddbs_list_tables(conn)
```

From here the workflow mirrors the `duckspatial_df` workflow. Functions that accept a `duckspatial_df` also accept a table name + connection pair:

```{r}
ddbs_is_valid("countries", conn = conn) |>
  filter(!is_valid)
```

You can write intermediate results as named tables in the database by passing a `name` argument:

```{r}
ddbs_make_valid("countries", conn = conn, name = "countries_valid")
ddbs_union("countries_valid", conn = conn, name = "world")
```

`ddbs_read_table()` materialises a table directly as `sf` (not lazily), so the result can be passed straight to `plot()`:

```{r}
ddbs_read_table(conn, "world") |>
  plot()
```

When you are done, close the connection. Because this is an in-memory database, all tables written to it will be discarded:

```{r}
ddbs_stop_conn(conn)
```

## Persistent database

The workflow is identical to the non-persistent case. The only difference is the connection string, pass a file path to `ddbs_create_conn()`:

```{r}
#| eval: false

conn <- ddbs_create_conn("my_database.duckdb")
```

A practical pattern is to do all processing with the `duckspatial_df` workflow (which is lazily evaluated inside the default connection), and only write the final results to the persistent database:

```{r}
#| eval: false

## open persistent connection
conn <- ddbs_create_conn("my_database.duckdb")

## do all processing with duckspatial_df objects
world_ddbs <- ddbs_open_dataset(
    system.file("spatial/countries.geojson", package = "duckspatial")
  ) |>
  ddbs_make_valid() |>
  ddbs_union()

## write only the final result to the persistent database
ddbs_write_table(conn, world_ddbs, name = "world")

## close — "my_database.duckdb" will persist on disk
ddbs_stop_conn(conn)
```

### Reopening saved tables

You can easily reopen a saved DuckDB database and lazily "register" the spatial tables using either `ddbs_open_dataset()` directly on the file, or `as_duckspatial_df()` if you are manually managing connections.

**A Note on CRS Persistence:** By default, `duckspatial` ensures that CRS metadata is preserved when saving tables to persistent DuckDB files. It achieves this using two complementary strategies:

1. **Native Spatial Storage** (`duckdb_storage_version = "v1.5.0"`): When using DuckDB 1.5.0+ (the default mode in both `ddbs_create_conn()` and `ddbs_write_dataset()`), CRS information is stored directly within the database catalog.
2. **Legacy Compatibility** (`duckdb_storage_version = "v1.0.0"`): For older DuckDB versions or legacy storage modes, `duckspatial` embeds the CRS as a specialized JSON comment on the geometry column.

For more information on DuckDB storage versions and compatibility guarantees, see the official [DuckDB Storage documentation](https://duckdb.org/docs/internals/storage).

::: {.callout-note}
### Technical Note on Interoperability
The **Legacy Compatibility** strategy is a `duckspatial`-specific convention. Because DuckDB does not natively support CRS metadata in older storage versions, we "hijack" the column comment field to store this information. 

While this ensures consistent behavior within `duckspatial`, **other spatial software (like QGIS, GDAL, or other DuckDB clients) will not recognize this metadata.** If you need to share your DuckDB files with other tools while preserving CRS, we strongly recommend using the **Native Spatial Storage** (default) or exporting to **GeoParquet**.
:::

This means that in most cases, you **do not** need to provide a `crs` argument when reopening a table:

```r
# CRS is automatically detected in most cases
reopened_world <- ddbs_open_dataset("my_database.duckdb", layer = "world")
```

Manual CRS input is typically only required if:
* The original data lacked CRS information entirely.
* You are reading a DuckDB file created by another tool that does not follow these persistence conventions.

**Recommendation:** While DuckDB storage is now robust for spatial metadata, **saving your data to GeoParquet files (`.parquet`) via `ddbs_write_dataset()`** remains an excellent choice for interoperability, as GeoParquet is a standardized format designed specifically for durable spatial data storage.

---
title: "Areal Interpolation"
date: "`r Sys.Date()`"
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Areal Interpolation}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
knitr:
  opts_chunk:
    collapse: true
    comment: "#>"
    out.width: "100%"
---

```{r}
#| include: false

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

# Quarto spawns a new R subprocess during R CMD build and may not pick up
# R_LIBS set by the build process; propagate it manually.
local({
  r_libs <- Sys.getenv("R_LIBS")
  if (nzchar(r_libs)) {
    paths <- unlist(strsplit(r_libs, .Platform$path.sep))
    .libPaths(unique(c(paths, .libPaths())))
  }
})
```

This vignette demonstrates how to use `ddbs_interpolate_aw()` to perform areal-weighted interpolation. This technique is essential when you need to transfer attributes from one set of polygons (source) to another incongruent set of polygons (target) based on their area of overlap.

`ddbs_interpolate_aw()` can be 9-30x faster than `sf::st_interpolate_aw` or [`areal::aw_interpolate`](https://chris-prener.github.io/areal/reference/aw_interpolate.html) (see benchmarks in [`{ducksf}`](https://www.ekotov.pro/ducksf/){target="_blank"} where prototype of `ddbs_interpolate_aw()` was originally developed).

`duckspatial` handles these heavy geometric calculations efficiently using DuckDB. We will cover three scenarios:

1.  **Extensive vs. Intensive**: Understanding the difference between mass-preserving counts and densities.
2.  **Fast Output**: Returning a `tibble` (no geometry) for maximum speed.
3.  **Database Mode**: Performing operations on persistent database tables.

### Setup Data

We will use the North Carolina dataset from the `sf` package as our **source**, and create a generic grid as our **target**. 

*Note: We project the data to Albers Equal Area (EPSG:5070) because accurate interpolation requires an equal-area projection.*

```{r}
#| label: setup
#| message: false
#| warning: false
library(duckspatial)
library(sf)

# 1. Load Source Data (NC Counties)
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)

# 2. Transform to projected CRS (Albers) for accurate area calculations
nc <- st_transform(nc, 5070)

# 3. Create a Target Grid
grid <- st_make_grid(nc, n = c(10, 5)) |> st_as_sf()

# 4. Create Unique IDs (Required for interpolation)
nc$source_id <- 1:nrow(nc)
grid$target_id <- 1:nrow(grid)
```

## 1) Extensive vs. Intensive Interpolation

Areal interpolation works differently depending on the nature of the data.

### Case A: Extensive Variables (Counts)
Variables like population counts or total births (`BIR74`) are **spatially extensive**. If a source polygon is split in half, the count should also be split in half. We use `weight = "total"` to ensure strict mass preservation relative to the source.

```{r}
# Interpolate Total Births (Extensive)
res_extensive <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  mode = "sf"
)
```

**Verification:** The total sum of births in the result should match the original data (mass preservation).

```{r}
orig_sum <- sum(nc$BIR74)
new_sum  <- sum(res_extensive$BIR74, na.rm = TRUE)

sprintf("Original: %s | Interpolated: %s", orig_sum, round(new_sum, 1))
```

### Case B: Intensive Variables (Densities/Ratios)
Variables like population density or infection rates are **spatially intensive**. If a source polygon is split, the density remains the same in both pieces. `duckspatial` handles this by calculating the area-weighted average.

```{r}
# Interpolate 'BIR74' treating it as an intensive variable (e.g. density assumption)
res_intensive <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  intensive = "BIR74", # Treated as density here
  weight = "sum",      # Standard behavior for intensive vars
  mode = "sf"
)
```

### Visual Comparison

Notice the difference in patterns. Extensive interpolation accumulates values based on how much "stuff" falls into a grid cell, while intensive interpolation smoothes the values based on overlap.

```{r, fig.height=5, fig.width=7}
# Combine for plotting
plot_data <- res_extensive[, "BIR74"]
names(plot_data)[1] <- "Extensive_Count"
plot_data$Intensive_Value <- res_intensive$BIR74

plot(plot_data[c("Extensive_Count", "Intensive_Value")], 
     main = "Interpolation Methods Comparison",
     border = "grey90",
     key.pos = 4)
```

## 2) High Performance: Output as Tibble

If you are working with massive datasets, constructing the geometry for the result `sf` object can be slow. If you only need the interpolated numbers, collect the default `duckspatial_df` using `ddbs_collect(as = "tibble")`. This skips the geometry construction step and is significantly faster.

```{r}
# Return a standard data.frame/tibble without geometry
res_tbl <- ddbs_interpolate_aw(
  target = grid,
  source = nc,
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74"
) |> 
  ddbs_collect(as = "tibble")

head(res_tbl)
```

## 3) Database Mode: Large Data Workflows

For datasets larger than memory, or for persistent pipelines, you can perform the interpolation directly inside the DuckDB database without pulling data into R until the end.

First, let's establish a connection and load our spatial layers into tables.

```{r}
# Create connection
conn <- ddbs_create_conn()

# Write layers to DuckDB
ddbs_write_vector(conn, nc, "nc_table", overwrite = TRUE)
ddbs_write_vector(conn, grid, "grid_table", overwrite = TRUE)
```

Now we run the interpolation by referencing the table names. We can also use the `name` argument to save the result directly to a new table instead of returning it to R.

```{r}
# Run interpolation and save to new table 'nc_grid_births'
ddbs_interpolate_aw(
  conn = conn,
  target = "grid_table",
  source = "nc_table",
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  name = "nc_grid_births", # <--- Writes to DB
  overwrite = TRUE
)

# Verify the table was created
DBI::dbListTables(conn)
```

We can now query this table or read it back later.

```{r}
# Read the result back from the database
final_sf <- ddbs_read_vector(conn, "nc_grid_births")

head(final_sf)
```

If we only wanted the database table, without the geometry, we could do:

```{r}
ddbs_interpolate_aw(
  conn = conn,
  target = "grid_table",
  source = "nc_table",
  tid = "target_id",
  sid = "source_id",
  extensive = "BIR74",
  weight = "total",
  name = "nc_grid_births", # <--- Writes to DB
  overwrite = TRUE,
  mode = "tibble"
)
```

And preview this table directly in the database:

```{r}
as_duckspatial_df("nc_grid_births", conn)
```

### Cleanup

Always close the connection when finished.

```{r}
duckdb::dbDisconnect(conn)
```

---
title: "Spatial joins"
date: "`r Sys.Date()`"
code-annotations: hover
urlcolor: blue
vignette: >
  %\VignetteIndexEntry{Spatial joins}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
execute:
  eval: !expr identical(tolower(Sys.getenv("NOT_CRAN")), "true")
knitr:
  opts_chunk:
    collapse: true
    comment: "#>"
    out.width: "100%"
---
  
```{r}
#| include: false

# CRAN OMP THREAD LIMIT to avoid CRAN NOTE
Sys.setenv(OMP_THREAD_LIMIT = 2)

# Quarto spawns a new R subprocess during R CMD build and may not pick up
# R_LIBS set by the build process; propagate it manually.
local({
  r_libs <- Sys.getenv("R_LIBS")
  if (nzchar(r_libs)) {
    paths <- unlist(strsplit(r_libs, .Platform$path.sep))
    .libPaths(unique(c(paths, .libPaths())))
  }
})
```

This vignette shows how to use `ddbs_join()` to perform fast spatial join 
operations on large data sets with three different approaches:

1 **In-memory**: pass `sf` objects and get an `sf` result (DuckDB runs under the hood, no persistent DataBase).
2. **Connected**: pass table names stored in an existing DuckDB connection and get an `sf` result.
3. **Write-to-DB**: same as (2), but write the result to a new DuckDB table.

Let's see a few examples. First, let's load a few libraries and our sample data:

```{r}
#| message: false
#| warning: false
library(duckspatial)
library(sf)

# polygons
countries_sf  <- sf::st_read(
    system.file("spatial/countries.geojson",  package = "duckspatial"),
    quiet = TRUE
    )

# random points
set.seed(42)
n <- 10000
points_sf <- data.frame(
  id = 1:n,
  x  = runif(n, min = -180, max = 180),
  y  = runif(n, min =  -90, max =  90)
) |>
  sf::st_as_sf(coords = c("x","y"), crs = 4326)

```

# 1) In-memory: pass `sf`, return `sf`

The simplest way to perform fast spatial join. You simply pass two `sf` objects,
and `ddbs_join()` spins up a temporary DuckDB, runs the join, and returns an `sf.`

- **When to use:** quick analysis, prototyping, or when you don’t need to persist 
intermediate tables.

```{r, message=FALSE}
out_sf1 <- ddbs_join(
  x    = points_sf,
  y    = countries_sf,
  join = "within"
)

# quick peek
# mapview(out_sf1, zcol="NAME_ENGL")

```

# 2) Connected: pass table names in DuckDB, return `sf`

In the second and third approaches, we make use of a connection to an existing 
DuckDB database. So let's create a fresh DuckDB connection using the `ddbs_create_conn()` function, which automatically install and load DuckDB spatial extension to the connection.

```{r}
# create a fresh DuckDB connection
conn <- duckspatial::ddbs_create_conn()

```
Now, in this second approach you need first to write your layers to DuckDB, and 
perform the spatial join by referencing their table names. Like this:

```{r, message=FALSE}
# write data to DuckDB
ddbs_write_table(conn, points_sf,   "points",    overwrite = TRUE)
ddbs_write_table(conn, countries_sf, "countries", overwrite = TRUE)

# spatial join inside DuckDB; result returned as sf
out_sf2 <- ddbs_join(
  conn,
  x    = "points",
  y    = "countries",
  join = "within"
)
```
- **When to use:** iterative workflows, larger-than-memory data, or when you’ll run multiple queries on the same tables.


# 3) Write-to-DB: create a new table with the join result

The output of approaches 1 and 2 is an `sf` object loaded to your memory. In this 
third approach, `ddbs_join()` writes a new table in the DuckDB database. You 
simply need to the `name` of the new table.

```{r, message=FALSE}
ddbs_join(
    conn = conn,
    x = "points",
    y = "countries",
    join = "within",
    name = "points_in_countries",
    overwrite = TRUE
)

# use the result in SQL (or read back as sf later)
# DBI::dbReadTable(conn, "points_in_countries") |>
#     sf::st_as_sf(wkt = 'geometry') |> 
#     head()

```
- **When to use:** iterative workflows, larger-than-memory data, or when you’ll run multiple queries on the same tables.


# Spatial Join Predicates:

A spatial predicate is really just a function that evaluates some spatial 
relation between two geometries and returns true or false, e.g., “does a contain 
b” or “is a within distance x of b”. The `join` argument accepts the spatial 
predicates:

- `"ST_Intersects"`: Whether a intersects b
- `"ST_Contains"`: Whether a contains b
- `"ST_ContainsProperly"`: Whether a contains b without b touching a's boundary
- `"ST_Within"`: Whether a is within b
- `"ST_Overlaps"`: Whether a overlaps b
- `"ST_Touches"`: Whether a touches b
- `"ST_Equals"`: Whether a is equal to b
- `"ST_Crosses"`: Whether a crosses b
- `"ST_Covers"`: Whether a covers b
- `"ST_CoveredBy"`: Whether a is covered by b
- `"ST_DWithin"`: x)	Whether a is within distance x of b


# Clean up

Don’t forget to disconnect from the database.

```{r}
duckdb::dbDisconnect(conn)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_sf_methods.R
\name{ddbs_geom_col}
\alias{ddbs_geom_col}
\title{Get the geometry column name}
\usage{
ddbs_geom_col(x)
}
\arguments{
\item{x}{A duckspatial_df object}
}
\value{
Character string with geometry column name
}
\description{
Get the geometry column name
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_dimensions_ops.R
\name{ddbs_locate}
\alias{ddbs_locate}
\alias{ddbs_locate_along}
\alias{ddbs_locate_between}
\title{Locate geometries at specific M values}
\usage{
ddbs_locate_along(
  x,
  measure,
  offset = 0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_locate_between(
  x,
  start_measure,
  end_measure,
  offset = 0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{measure}{A numeric value specifying the M value at which to locate a
point along the geometry. Used only by \code{ddbs_locate_along}.}

\item{offset}{A numeric value specifying a lateral offset to apply
perpendicular to the line direction at the located point(s). Default is
\code{0} (no offset).}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{start_measure}{A numeric value specifying the lower bound of the M
range.}

\item{end_measure}{A numeric value specifying the upper bound of the M
range.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Return the geometries at a specific M values or range of M values.
}
\details{
\itemize{
\item \code{ddbs_locate_along()}: returns a point or multi-point, containing the point(s)
at the geometry with the given measure
\item \code{ddbs_locate_between()}: returns a geometry or geometry collection created by
filtering and interpolating vertices within a range of "M" values
}
}
\examples{
\dontrun{
## load package
library(duckspatial)

## read data (must contain M-enabled linestring geometries)
rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson",
  package = "duckspatial")
)

## Calculate the length of the rivers
rivers_agg_ddbs <- rivers_ddbs |> 
  ddbs_union_agg("RIVER_NAME") |> 
  ddbs_length()

## Add M dimension to the rivers
rivers_m_ddbs <- rivers_agg_ddbs |> 
  ddbs_force_3d("length", dim = "M")

## Locate rivers with M between 10000 and 20000
ddbs_locate_between(rivers_m_ddbs, 10000, 20000)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{get_nrow}
\alias{get_nrow}
\title{Get the number of rows in a table}
\usage{
get_nrow(conn, table)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{table}{name of the table}
}
\value{
number of rows in the table
}
\description{
Get the number of rows in a table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{feedback_query}
\alias{feedback_query}
\title{Feedback for query success}
\usage{
feedback_query(quiet)
}
\arguments{
\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
cli message
}
\description{
Feedback for query success
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_buffer}
\alias{ddbs_buffer}
\title{Creates a buffer around geometries}
\usage{
ddbs_buffer(
  x,
  distance,
  num_triangles = 8L,
  cap_style = "CAP_ROUND",
  join_style = "JOIN_ROUND",
  mitre_limit = 1,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{distance}{a numeric value specifying the buffer distance. Units correspond to
the coordinate system of the geometry (e.g. degrees or meters)}

\item{num_triangles}{an integer representing how many triangles will be produced to
approximate a quarter circle. The larger the number, the smoother the resulting geometry.
Default is 8.}

\item{cap_style}{a character string specifying the cap style. Must be one of
"CAP_ROUND" (default), "CAP_FLAT", or "CAP_SQUARE". Case-insensitive.}

\item{join_style}{a character string specifying the join style. Must be one of
"JOIN_ROUND" (default), "JOIN_MITRE", or "JOIN_BEVEL". Case-insensitive.}

\item{mitre_limit}{a numeric value specifying the mitre limit ratio. Only applies when
\code{join_style} is "JOIN_MITRE". It is the ratio of the distance from the corner to
the mitre point to the corner radius. Default is 1.0.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Computes a polygon that represents all locations within a specified distance from the
original geometry
}
\examples{
\dontrun{
## load package
library(duckspatial)

## create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", 
  package = "duckspatial")
)

## store in duckdb
ddbs_write_vector(conn, argentina_ddbs, "argentina")

## basic buffer
ddbs_buffer(conn = conn, "argentina", distance = 1)

## buffer with custom parameters
ddbs_buffer(conn = conn, "argentina", distance = 1, 
            num_triangles = 16, cap_style = "CAP_SQUARE")

## buffer without using a connection
ddbs_buffer(argentina_ddbs, distance = 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_sf_methods.R
\name{convert_wkb_to_sfc}
\alias{convert_wkb_to_sfc}
\title{Convert WKB data to sfc}
\usage{
convert_wkb_to_sfc(geom_data, crs)
}
\description{
Convert WKB data to sfc
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_create_conn}
\alias{ddbs_create_conn}
\title{Create a DuckDB connection with spatial extension}
\usage{
ddbs_create_conn(
  dbdir = "memory",
  threads = NULL,
  memory_limit_gb = NULL,
  upgrade = FALSE,
  ...,
  duckdb_storage_version = duckspatial_storage_default()
)
}
\arguments{
\item{dbdir}{String. Either \code{"tempdir"}, \code{"memory"}, or a DuckDB database
file path with \code{.duckdb}, \code{.db}, or \code{.ddb} extension. Defaults to \code{"memory"}.}

\item{threads}{Integer. Number of threads to use. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use all available cores it detects (warning, on some shared HPC nodes the detected number of cores might be total number of cores on the node, not the per-job allocation).}

\item{memory_limit_gb}{Numeric. Memory limit in GB. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use 80\% of available operating system memory it detects (warning, on some shared HPC nodes the detected memory might be the full node memory, not the per-job allocation).}

\item{upgrade}{if TRUE, it upgrades the DuckDB extension to the latest version}

\item{...}{Additional parameters to be passed to \code{\link[DBI]{dbConnect}}}

\item{duckdb_storage_version}{Storage compatibility for newly created persistent
native DuckDB files (\code{.duckdb}, \code{.db}, \code{.ddb}). See
\url{https://duckdb.org/docs/internals/storage} for more information on
DuckDB storage versions and compatibility.
\itemize{
\item \code{"v1.5.0"} (\strong{Native Spatial Storage}, Default): Preserves
CRS metadata in native DuckDB \code{GEOMETRY} columns. Requires
DuckDB >= 1.5.0 to open the file.
\item \code{"v1.0.0"} (\strong{Legacy Compatibility}): Creates
files readable by older DuckDB versions (>= 1.0.0). Persists CRS
metadata in duckspatial-managed column comments (a convention not
recognized by other spatial software).
\item \code{"latest"}: Use the highest storage version supported by your
installed DuckDB engine.
}
Other major version strings like \code{"v1.4.0"}, \code{"v1.3.0"}, etc., are also supported.}
}
\value{
A \code{duckdb_connection}
}
\description{
It creates a DuckDB connection, and then it installs and loads the
spatial extension
}
\examples{
\dontrun{
# load packages
library(duckspatial)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

# create a duckdb database in disk  (with spatial extension)
conn <- ddbs_create_conn(dbdir = "tempdir")

# create a connection with 1 thread and 2GB memory limit
conn <- ddbs_create_conn(threads = 1, memory_limit_gb = 2)
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_remove_repeated_points}
\alias{ddbs_remove_repeated_points}
\title{Remove repeated points from a geometry}
\usage{
ddbs_remove_repeated_points(
  x,
  tolerance = 0,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{tolerance}{A numeric value specifying the minimum distance between
consecutive vertices. Vertices closer than this threshold are considered
repeated and removed. Default is \code{0}, which removes only exact
duplicates.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Removes duplicate consecutive vertices from geometries, optionally within
a tolerance distance.
}
\examples{
\dontrun{
## load package
library(duckspatial)
library(sf)

## Create a polygon with repeated points
poly <- st_polygon(list(matrix(
  c(0, 0,
    1, 0,
    1, 0,  # repeated point
    1, 1,
    0, 1,
    0, 0),
  ncol = 2, byrow = TRUE
)))

poly_sf <- st_as_sf(st_sfc(poly))

## remove exact duplicate consecutive vertices
ddbs_remove_repeated_points(poly_sf)

## remove vertices within a tolerance of 1 unit
ddbs_remove_repeated_points(poly_sf, tolerance = 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_dimensions.R
\name{ddbs_force_dim}
\alias{ddbs_force_dim}
\alias{ddbs_force_2d}
\alias{ddbs_force_3d}
\alias{ddbs_force_4d}
\title{Force geometry dimensions}
\usage{
ddbs_force_2d(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_force_3d(
  x,
  var,
  dim = "z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_force_4d(
  x,
  var_z,
  var_m,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{var}{A numeric variable in \code{x} to be converted in the dimension specified
in the argument \code{dim}}

\item{dim}{The dimension to add: either \code{"z"} (default) for
elevation or \code{"m"} for measure values.}

\item{var_z}{A numeric variable in \code{x} to be convered in \code{Z} dimension}

\item{var_m}{A numeric variable in \code{x} to be convered in \code{M} dimension}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Functions to force geometries to have specific coordinate dimensions (X, Y, Z, M)
}
\details{
These functions modify the dimensionality of geometries:
\itemize{
\item \code{ddbs_force_2d()} removes Z and M coordinates from geometries, returning only
X and Y coordinates. This is useful for simplifying 3D or measured geometries
to 2D.
\item \code{ddbs_force_3d()} forces geometries to have three dimensions. When \code{dim = "z"}
(default), adds or retains Z coordinates (X, Y, Z). When \code{dim = "m"}, adds or
retains M coordinates (X, Y, M). Missing values are typically set to 0. If the
input geometry has a third dimension already, it will be replaced by the new one.
If the input geometry has 4 dimensions, it will drop the dimension that wasn't
specified.
\item \code{ddbs_force_4d()} forces geometries to have all four dimensions (X, Y, Z, M).
Missing Z or M values are typically set to 0.
}
}
\examples{
\dontrun{
## load packages
library(dplyr)
library(duckspatial)

## load data and add 2 numeric vars
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |> 
  dplyr::filter(ISO3_CODE != "ATA") |> 
  ddbs_area(new_column = "area") |> 
  ddbs_perimeter(new_column = "perim") 

## add a Z dimension
countries_z_ddbs <- ddbs_force_3d(countries_ddbs, "area")
ddbs_has_z(countries_z_ddbs)

## add a M dimension as 3D (removes current Z)
countries_m_ddbs <- ddbs_force_3d(countries_z_ddbs, "area", "m")
ddbs_has_z(countries_m_ddbs)
ddbs_has_m(countries_m_ddbs)

## add both Z and M
countries_zm_ddbs <- ddbs_force_4d(countries_ddbs, "area", "perim")
ddbs_has_z(countries_zm_ddbs)
ddbs_has_m(countries_zm_ddbs)

## drop both ZM
countries_drop_ddbs <- ddbs_force_2d(countries_zm_ddbs)
ddbs_has_z(countries_drop_ddbs)
ddbs_has_m(countries_drop_ddbs)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_measure.R
\name{ddbs_measure_funs}
\alias{ddbs_measure_funs}
\alias{ddbs_area}
\alias{ddbs_length}
\alias{ddbs_perimeter}
\alias{ddbs_distance}
\alias{ddbs_azimuth}
\title{Calculate geometric measurements}
\usage{
ddbs_area(
  x,
  new_column = "area",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_length(
  x,
  new_column = "length",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_perimeter(
  x,
  new_column = "perimeter",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_distance(
  x,
  y,
  dist_type = NULL,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  id_x = NULL,
  id_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_azimuth(
  x,
  y,
  unit = "radians",
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  id_x = NULL,
  id_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input geometry (sf object, duckspatial_df, or table name in DuckDB)}

\item{new_column}{Name of the new column to create on the input data. Ignored
with \code{mode = "sf"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{y}{Second input geometry for distance and azimuth calculations (sf object, duckspatial_df, or table name)}

\item{dist_type}{Character. Distance type to be calculated. By default it uses
the best option for the input CRS (see details).}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{id_x}{Character; optional name of the column in \code{x} whose values will
be used to name the list elements. If \code{NULL}, integer row numbers of \code{x} are used.}

\item{id_y}{Character; optional name of the column in \code{y} whose values will
replace the integer indices returned in each element of the list.}

\item{unit}{Character. Output unit: \code{"radians"} (default) or \code{"degrees"}.}
}
\value{
For \code{ddbs_area}, \code{ddbs_length}, and \code{ddbs_perimeter}:
\itemize{
\item \code{mode = "duckspatial"} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{mode = "sf"}: An eagerly collected vector in R memory.
\item When \code{name} is provided: writes the table in the DuckDB connection and returns \code{TRUE} (invisibly).
}

For \code{ddbs_distance}: A \code{units} matrix in meters with dimensions nrow(x), nrow(y).

For \code{ddbs_azimuth}: A numeric matrix of azimuth values (in the specified \code{unit})
with dimensions nrow(x) by nrow(y) when \code{mode = "sf"}, or a lazy
\code{tbl_duckdb_connection} with columns \code{id_x}, \code{id_y}, and \code{azimuth} otherwise.
Both inputs must contain only POINT geometries.
}
\description{
Compute area, length, perimeter, or distance of geometries with automatic
method selection based on the coordinate reference system (CRS).
}
\details{
These functions automatically select the appropriate calculation method based on the input CRS:

\strong{For EPSG:4326 (geographic coordinates):}
\itemize{
\item Uses \code{ST_*_Spheroid} functions (e.g., \code{ST_Area_Spheroid}, \code{ST_Length_Spheroid})
\item Leverages GeographicLib library for ellipsoidal earth model calculations
\item Highly accurate but slower than planar calculations
\item For \code{ddbs_distance} with POINT geometries: defaults to \code{"haversine"}
\item For \code{ddbs_distance} with other geometries: defaults to \code{"spheroid"}
}

\strong{For projected CRS (e.g., UTM, Web Mercator):}
\itemize{
\item Uses planar \code{ST_*} functions (e.g., \code{ST_Area}, \code{ST_Length})
\item Faster performance with accurate results in meters
\item For \code{ddbs_distance}: defaults to \code{"planar"}
}

\strong{Distance calculation methods} (\code{dist_type} argument):
\itemize{
\item \code{NULL} (default): Automatically selects best method for input CRS
\item \code{"planar"}: Planar distance (for projected CRS)
\item \code{"geos"}: Planar distance using GEOS library (for projected CRS)
\item \code{"haversine"}: Great circle distance (requires EPSG:4326 and POINT geometries)
\item \code{"spheroid"}: Ellipsoidal model using GeographicLib (most accurate, slowest)
}

\strong{Distance type requirements:}
\itemize{
\item \code{"planar"} and \code{"geos"}: Require projected coordinates (not degrees)
\item \code{"haversine"} and \code{"spheroid"}: Require POINT geometries and EPSG:4326
}
}
\section{Performance}{

Speed comparison (fastest to slowest):
\enumerate{
\item Planar calculations on projected CRS
\item Haversine (spherical approximation)
\item Spheroid functions (ellipsoidal model)
}
}

\examples{
\dontrun{
library(duckspatial)
library(dplyr)

# Create a DuckDB connection
conn <- ddbs_create_conn(dbdir = "memory")

# ===== AREA CALCULATIONS =====

# Load polygon data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", package = "duckspatial")
) |>
  ddbs_transform("EPSG:3857") |> 
  filter(NAME_ENGL != "Antarctica")

# Store in DuckDB
ddbs_write_table(conn, countries_ddbs, "countries")

# Calculate area (adds a new column - area by default)
ddbs_area("countries", conn)

# Calculate area with custom column name
ddbs_area("countries", conn, new_column = "area_sqm")

# Create new table with area calculations
ddbs_area("countries", conn, name = "countries_with_area", new_column = "area_sqm")

# Calculate area from sf object directly
ddbs_area(countries_ddbs)

# Calculate area using dplyr syntax
countries_ddbs |> 
  mutate(area = ddbs_area(geom))

# Calculate total area 
countries_ddbs |> 
  mutate(area = ddbs_area(geom)) |> 
  summarise(
    area = sum(area),
    geom = ddbs_union(geom)
  )

# ===== LENGTH CALCULATIONS =====

# Load line data
rivers_ddbs <- sf::read_sf(
  system.file("spatial/rivers.geojson", package = "duckspatial")
) |> 
  as_duckspatial_df()

# Store in DuckDB
ddbs_write_table(conn, rivers_ddbs, "rivers")

# Calculate length (add a new column - length by default)
ddbs_length("rivers", conn)

# Calculate length with custom column name
ddbs_length(rivers_ddbs, new_column = "length_meters")

# Calculate length by river name
rivers_ddbs |> 
  ddbs_union_agg("RIVER_NAME") |> 
  ddbs_length()

# Add length within dplyr
rivers_ddbs |> 
  mutate(length = ddbs_length(geometry))


# ===== PERIMETER CALCULATIONS =====

# Calculate perimeter (returns sf object with perimeter column)
ddbs_perimeter(countries_ddbs)

# Calculate perimeter within dplyr
countries_ddbs |> 
  mutate(perim = ddbs_perimeter(geom))


# ===== DISTANCE CALCULATIONS =====

# Create sample points in EPSG:4326
n <- 10
points_sf <- data.frame(
  id = 1:n,
  x = runif(n, min = -180, max = 180),
  y = runif(n, min = -90, max = 90)
) |>
  ddbs_as_spatial(coords = c("x", "y"), crs = "EPSG:4326")

# Option 1: Using sf objects (auto-selects haversine for EPSG:4326 points)
dist_matrix <- ddbs_distance(x = points_sf, y = points_sf)
head(dist_matrix)

# Option 2: Explicitly specify distance type
dist_matrix_harv <- ddbs_distance(
  x = points_sf,
  y = points_sf,
  dist_type = "haversine"
)

# Option 3: Using DuckDB tables
ddbs_write_table(conn, points_sf, "points", overwrite = TRUE)
dist_matrix_sph <- ddbs_distance(
  conn = conn,
  x = "points",
  y = "points",
  dist_type = "spheroid"  # Most accurate for geographic coordinates
)
head(dist_matrix_sph)

# Close connection
ddbs_stop_conn(conn)
}
\dontrun{
library(duckspatial)

# Create two sets of points in a projected CRS
origins <- sf::st_as_sf(
  data.frame(id = 1:2, x = c(0, 0), y = c(0, 0)),
  coords = c("x", "y"), crs = "EPSG:3857"
)
destinations <- sf::st_as_sf(
  data.frame(id = 1:2, x = c(0, 1), y = c(1, 0)),
  coords = c("x", "y"), crs = "EPSG:3857"
)

# Returns a numeric matrix (nrow(origins) x nrow(destinations))
ddbs_azimuth(origins, destinations, mode = "sf")

# In degrees
ddbs_azimuth(origins, destinations, unit = "degrees", mode = "sf")

# Lazy tbl with all pairs (default mode)
ddbs_azimuth(origins, destinations)
}
}
\references{
\url{https://geographiclib.sourceforge.io/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_sf_methods.R
\name{ddbs_compute}
\alias{ddbs_compute}
\title{Force computation of a lazy duckspatial_df}
\usage{
ddbs_compute(x, ..., name = NULL, temporary = TRUE)
}
\arguments{
\item{x}{A \code{duckspatial_df} object}

\item{...}{Additional arguments passed to \code{dplyr::compute}}

\item{name}{Optional name for the result table. If NULL, a unique temporary
name is generated.}

\item{temporary}{If TRUE (default), creates a temporary table that is
automatically cleaned up when the connection closes.}
}
\value{
A new \code{duckspatial_df} pointing to the materialized table
}
\description{
Executes the accumulated query and stores the result in a DuckDB temporary
table. The result remains lazy (a \code{duckspatial_df}) but points to the
materialized data, avoiding repeated computation of complex query plans.
}
\details{
This is useful when you want to:
\itemize{
\item Cache intermediate results for reuse across multiple subsequent operations
\item Simplify complex query plans before heavy operations like spatial joins
\item Force execution at a specific point without pulling data into R memory
}
}
\examples{
\dontrun{
library(duckspatial)
library(dplyr)

# Load lazy spatial data
countries <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", package = "duckspatial")
)

# Complex pipeline - ddbs_compute() caches intermediate result
cached <- countries |>
  filter(CNTR_ID \%in\% c("DE", "FR", "IT")) |>
  ddbs_compute()  # Execute and store in temp table

# Check query plan - should reference temp table
show_query(cached)

# Further operations continue from cached result
result <- cached |>
  ddbs_filter(other_layer, predicate = "intersects") |>
  st_as_sf()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary_line.R
\name{ddbs_line_interpolate}
\alias{ddbs_line_interpolate}
\title{Interpolates a point or points along a line geometry}
\usage{
ddbs_line_interpolate(
  x,
  fraction = 0.5,
  intervals = FALSE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{fraction}{a numeric value between 0 and 1. When
\code{intervals = FALSE}, specifies the position along the line to
interpolate, where \code{0} is the start and \code{1} is the end.
When \code{intervals = TRUE}, specifies the spacing between interpolated
points as a proportion of the total line length. Defaults to \code{0.5}.}

\item{intervals}{a logical value. If \code{FALSE} (default), returns a
single \code{POINT} at the position given by \code{fraction}. If
\code{TRUE}, returns a \code{MULTIPOINT} of equally-spaced points along
the line at intervals defined by \code{fraction}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Returns either a single point at a specified position along a line, or
multiple equally-spaced points along a line, depending on the value of
\code{intervals}. When \code{intervals = FALSE}, this wraps
\code{ST_LineInterpolatePoint}; when \code{intervals = TRUE}, it wraps
\code{ST_LineInterpolatePoints}.
}
\examples{
\dontrun{
## load package
library(duckspatial)

## read data
rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson",
  package = "duckspatial")
)

## return the midpoint of a line (default)
ddbs_line_interpolate(rivers_ddbs)

## return the point 25\% along the line
ddbs_line_interpolate(rivers_ddbs, fraction = 0.25)

## return equally-spaced points every 10\% of the line length
ddbs_line_interpolate(rivers_ddbs, fraction = 0.1, intervals = TRUE)

## return equally-spaced points every 50\% of the line length (i.e. midpoint and end)
ddbs_line_interpolate(rivers_ddbs, fraction = 0.5, intervals = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_interpolate_aw.R
\name{ddbs_interpolate_aw}
\alias{ddbs_interpolate_aw}
\title{Areal-Weighted Interpolation using DuckDB}
\usage{
ddbs_interpolate_aw(
  target,
  source,
  tid,
  sid,
  extensive = NULL,
  intensive = NULL,
  weight = "sum",
  mode = NULL,
  keep_NA = TRUE,
  na.rm = FALSE,
  join_crs = NULL,
  conn = NULL,
  name = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{target}{An \code{sf} object or the name of a persistent table in the DuckDB connection
representing the destination geometries.}

\item{source}{An \code{sf} object or the name of a persistent table in the DuckDB connection
containing the data to be interpolated.}

\item{tid}{Character. The name of the column in \code{target} that uniquely identifies features.}

\item{sid}{Character. The name of the column in \code{source} that uniquely identifies features.}

\item{extensive}{Character vector. Names of columns in \code{source} to be treated as
spatially extensive (e.g., population counts).}

\item{intensive}{Character vector. Names of columns in \code{source} to be treated as
spatially intensive (e.g., population density).}

\item{weight}{Character. Determines the denominator calculation for extensive variables.
Either \code{"sum"} (default) or \code{"total"}. See \strong{Mass Preservation} in Details.}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{keep_NA}{Logical. If \code{TRUE} (default), returns all features from the target,
even those that do not overlap with the source (values will be NA). If \code{FALSE},
performs an inner join, dropping non-overlapping target features.}

\item{na.rm}{Logical. If \code{TRUE}, source features with \code{NA} values in the
interpolated variables are completely removed from the calculation (area calculations
will behave as if that polygon did not exist). Defaults to \code{FALSE}.}

\item{join_crs}{Numeric or Character (optional). EPSG code or WKT for the CRS to use
for area calculations. If provided, both \code{target} and \code{source} are transformed
to this CRS within the database before interpolation.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Transfers attribute data from a source spatial layer to a target spatial layer based
on the area of overlap between their geometries. This function executes all spatial
calculations within DuckDB, enabling efficient processing of large datasets without
loading all geometries into R memory.
}
\details{
Areal-weighted interpolation is used when the source and target geometries are incongruent (they do not align). It relies on the assumption of \strong{uniform distribution}: values in the source polygons are assumed to be spread evenly across the polygon's area.

\strong{Coordinate Systems:}
Area calculations are highly sensitive to the Coordinate Reference System (CRS).
While the function can run on geographic coordinates (lon/lat), it is strongly recommended
to use a \strong{projected CRS} (e.g., EPSG:3857, UTM, or Albers) to ensure accurate area measurements.
Use the \code{join_crs} argument to project data on-the-fly during the interpolation.

\strong{Extensive vs. Intensive Variables:}
\itemize{
\item \strong{Extensive} variables are counts or absolute amounts (e.g., total population,
number of voters). When a source polygon is split, the value is divided proportionally
to the area.
\item \strong{Intensive} variables are ratios, rates, or densities (e.g., population density,
cancer rates). When a source polygon is split, the value remains constant for each piece.
}

\strong{Mass Preservation (The \code{weight} argument):}
For extensive variables, the choice of weight determines the denominator used in calculations:
\itemize{
\item \code{"sum"} (default): The denominator is the sum of all overlapping areas
for that source feature. This preserves the "mass" of the variable \emph{relative to the target's coverage}.
If the target polygons do not completely cover a source polygon, some data is technically "lost"
because it falls outside the target area. This matches \code{areal::aw_interpolate(weight="sum")}.
\item \code{"total"}: The denominator is the full geometric area of the source feature.
This assumes the source value is distributed over the entire source polygon. If the target
covers only 50\% of the source, only 50\% of the value is transferred. This is strictly
mass-preserving relative to the source. This matches \code{sf::st_interpolate_aw(extensive=TRUE)}.
}
\emph{Note:} Intensive variables are always calculated using the \code{"sum"} logic (averaging
based on intersection areas) regardless of this parameter.
}
\examples{
\donttest{
library(sf)

# 1. Prepare Data
# Load NC counties (Source) and project to Albers (EPSG:5070)
nc <- st_read(system.file("shape/nc.shp", package = "sf"), quiet = TRUE)
nc <- st_transform(nc, 5070)
nc$sid <- seq_len(nrow(nc)) # Create Source ID

# Create a target grid
g <- st_make_grid(nc, n = c(10, 5))
g_sf <- st_as_sf(g)
g_sf$tid <- seq_len(nrow(g_sf)) # Create Target ID

# 2. Extensive Interpolation (Counts)
# Use weight = "total" for strict mass preservation (e.g., total births)
res_ext <- ddbs_interpolate_aw(
  target = g_sf, source = nc,
  tid = "tid", sid = "sid",
  extensive = "BIR74",
  weight = "total",
  mode = "sf"
)

# Check mass preservation
sum(res_ext$BIR74, na.rm = TRUE) / sum(nc$BIR74) # Should be ~1

# 3. Intensive Interpolation (Density/Rates)
# Calculates area-weighted average (e.g., assumption of uniform density)
res_int <- ddbs_interpolate_aw(
  target = g_sf, source = nc,
  tid = "tid", sid = "sid",
  intensive = "BIR74",
  mode = "sf"
)

# 4. Quick Visualization
par(mfrow = c(1, 2))
plot(res_ext["BIR74"], main = "Extensive (Total Count)", border = NA)
plot(res_int["BIR74"], main = "Intensive (Weighted Avg)", border = NA)
}

}
\references{
Prener, C. and Revord, C. (2019). \emph{areal: An R package for areal weighted interpolation}.
\emph{Journal of Open Source Software}, 4(37), 1221.
Available at: \doi{10.21105/joss.01221}
}
\seealso{
\code{\link[areal:aw_interpolate]{areal::aw_interpolate()}} — reference implementation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial-package.R
\docType{package}
\name{duckspatial-package}
\alias{duckspatial}
\alias{duckspatial-package}
\title{duckspatial: R Interface to 'DuckDB' Database with Spatial Extension}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Fast and memory-efficient functions to analyze and manipulate large spatial datasets. It leverages the fast analytical capabilities of ‘DuckDB’ and its spatial extension (see \url{https://duckdb.org/docs/stable/core_extensions/spatial/overview}) while maintaining compatibility with R’s spatial data ecosystem to work with spatial vector data.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://cidree.github.io/duckspatial/}
  \item \url{https://github.com/Cidree/duckspatial}
  \item Report bugs at \url{https://github.com/Cidree/duckspatial/issues}
}

}
\author{
\strong{Maintainer}: Adrián Cidre González \email{adrian.cidre@gmail.com} (\href{https://orcid.org/0000-0002-3310-3052}{ORCID})

Authors:
\itemize{
  \item Adrián Cidre González \email{adrian.cidre@gmail.com} (\href{https://orcid.org/0000-0002-3310-3052}{ORCID})
  \item Egor Kotov \email{kotov.egor@gmail.com} (\href{https://orcid.org/0000-0001-6690-5345}{ORCID})
  \item Rafael H. M. Pereira \email{rafa.pereira.br@gmail.com} (\href{https://orcid.org/0000-0003-2125-7465}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_extension.R
\name{ddbs_load}
\alias{ddbs_load}
\title{Loads the Spatial extension}
\usage{
ddbs_load(conn, quiet = FALSE, extension = "spatial", create_macros = TRUE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{extension}{name of the extension to load, default is "spatial"}

\item{create_macros}{if TRUE (default), it creates macros that allow
some functions to be used within dplyr pipelines}
}
\value{
TRUE (invisibly) for successful installation
}
\description{
Checks if a spatial extension is installed, and loads it in a DuckDB database
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(duckdb)

## connect to in memory database
conn <- duckdb::dbConnect(duckdb::duckdb())

## install the spatial exntesion
ddbs_install(conn)
ddbs_load(conn)

## disconnect from db
duckdb::dbDisconnect(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_list_tables}
\alias{ddbs_list_tables}
\title{Check tables and schemas inside a database}
\usage{
ddbs_list_tables(conn)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}
}
\value{
\code{data.frame}
}
\description{
Check tables and schemas inside a database
}
\examples{
\dontrun{
## load packages
library(duckspatial)

## create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read some data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
)

argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", 
  package = "duckspatial")
)

## insert into the database
ddbs_write_table(conn, argentina_ddbs, "argentina")
ddbs_write_table(conn, countries_ddbs, "countries")

## list tables in the database
ddbs_list_tables(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_resources.R
\name{ddbs_set_resources}
\alias{ddbs_set_resources}
\alias{ddbs_get_resources}
\title{Get or set connection resources}
\usage{
ddbs_set_resources(conn, threads = NULL, memory_limit_gb = NULL)

ddbs_get_resources(conn)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{threads}{Integer. Number of threads to use. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use all available cores it detects (warning, on some shared HPC nodes the detected number of cores might be total number of cores on the node, not the per-job allocation).}

\item{memory_limit_gb}{Numeric. Memory limit in GB. If \code{NULL} (default), the setting is not changed, and DuckDB engine will use 80\% of available operating system memory it detects (warning, on some shared HPC nodes the detected memory might be the full node memory, not the per-job allocation).}
}
\value{
For \code{ddbs_set_resources()}, invisibly returns a list containing the current system settings; for \code{ddbs_get_resources()}, visibly returns the same list for direct inspection.
}
\description{
Configure technical system settings for a DuckDB connection, such as memory limits
and CPU threads.
}
\examples{
\dontrun{
# Create a connection
conn <- ddbs_create_conn()

# Set resources: 1 thread and 4GB
ddbs_set_resources(conn, threads = 1, memory_limit_gb = 4)

# Check current settings
ddbs_get_resources(conn)

ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_stop_conn}
\alias{ddbs_stop_conn}
\title{Close a DuckDB connection}
\usage{
ddbs_stop_conn(conn)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}
}
\value{
TRUE (invisibly) for successful disconnection
}
\description{
Close a DuckDB connection
}
\examples{
\dontrun{
## load packages
library(duckspatial)

## create an in-memory duckdb database
conn <- ddbs_create_conn(dbdir = "memory")

## close the connection
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_concave_hull}
\alias{ddbs_concave_hull}
\title{Compute the concave hull of geometries}
\usage{
ddbs_concave_hull(
  x,
  ratio = 0.5,
  allow_holes = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{ratio}{Numeric. The ratio parameter dictates the level of concavity; \code{1}
returns the convex hull, while \code{0} indicates to return the most concave
hull possible. Defaults to \code{0.5}.}

\item{allow_holes}{Boolean. If \code{TRUE} (the default), it allows the output to
contain holes.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Returns the concave hull that tightly encloses the geometry, capturing its overall
shape more closely than a convex hull.
}
\examples{
\dontrun{
## load package
library(duckspatial)
library(sf)

# create points data
n <- 5
points_ddbs <- data.frame(
  id = 1,
  x = runif(n, min = -180, max = 180),
  y = runif(n, min = -90, max = 90)
) |>
  ddbs_as_spatial(coords = c("x", "y")) |>
  ddbs_combine()

# option 1: passing ddbs or sf objects
output1 <- duckspatial::ddbs_concave_hull(points_ddbs, mode = "sf")

plot(output1)


# option 2: passing the name of a table in a duckdb db

# creates a duckdb
conn <- duckspatial::ddbs_create_conn()

# write sf to duckdb
ddbs_write_vector(conn, points_ddbs, "points_tbl")

# spatial join
output2 <- duckspatial::ddbs_concave_hull(
 conn = conn,
 x = "points_tbl",
 mode = "sf"
)

plot(output2)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_mvt.R
\name{ddbs_as_mvt_geom}
\alias{ddbs_as_mvt_geom}
\title{Transform geometries into Mapbox Vector Tile coordinate space}
\usage{
ddbs_as_mvt_geom(
  x,
  bounds,
  extent = 4096L,
  buffer = 256L,
  clip_geom = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{bounds}{The tile bounding box, in the same CRS as \code{x}. Either a
numeric vector \code{c(xmin, ymin, xmax, ymax)} or an object understood by
\code{sf::st_bbox()} (e.g. an \code{sf}, \code{sfc}, or \code{bbox}
object).}

\item{extent}{Integer. The width and height of the tile coordinate space
(the geometry is mapped into the square \code{[0, extent]}). Defaults to
\code{4096}.}

\item{buffer}{Integer. The number of pixels by which the tile bounds are
expanded before clipping, to avoid rendering artefacts at tile edges.
Defaults to \code{256}.}

\item{clip_geom}{Logical. If \code{TRUE} (default), geometries are clipped to
the (buffered) tile bounds; geometries falling entirely outside become
\code{NULL}. If \code{FALSE}, geometries are transformed but not clipped.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Prepares geometries for encoding as Mapbox Vector Tiles (MVT) by clipping
them to a tile's bounding box and transforming the coordinates into the
tile's integer pixel space. This is a low-level building block; to produce a
ready-to-serve tile pyramid use \code{\link{ddbs_write_mbtiles}}.
}
\details{
This is a wrapper around DuckDB's \code{ST_AsMVTGeom}. The output geometries
use the MVT tile coordinate system: the origin is the top-left corner and the
y-axis points downwards, so the result is no longer in the input CRS (the CRS
is dropped). Geometries that fall completely outside the tile bounds return
\code{NULL} when \code{clip_geom = TRUE}.
}
\examples{
\dontrun{
library(duckspatial)

argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", package = "duckspatial")
)

## use the layer's own bounding box as the tile bounds
ddbs_as_mvt_geom(argentina_ddbs, bounds = sf::st_bbox(sf::st_as_sf(argentina_ddbs)))

## explicit numeric bounds and a smaller tile extent
ddbs_as_mvt_geom(
  argentina_ddbs,
  bounds = c(-74, -56, -53, -21),
  extent = 256
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_union.R
\name{ddbs_intersection_agg}
\alias{ddbs_intersection_agg}
\title{Aggregate the intersection of geometries}
\usage{
ddbs_intersection_agg(
  x,
  by = NULL,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{by}{Character vector of one or more column names to group by. The
intersection is computed within each group. When \code{NULL} (default), all
geometries are intersected into a single geometry.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Computes the geometric intersection of a set of geometries — the area common
to all of them — using DuckDB's \code{ST_Intersection_Agg()} aggregate. This is the
intersection counterpart to \code{\link[=ddbs_union_agg]{ddbs_union_agg()}}.
}
\examples{
\dontrun{
## load package
library(duckspatial)

## create a connection and three overlapping polygons
conn <- ddbs_create_conn()
polys <- sf::st_as_sf(
  data.frame(grp = c("a", "a", "a")),
  geometry = sf::st_sfc(
    sf::st_polygon(list(matrix(c(0,0, 3,0, 3,3, 0,3, 0,0), ncol = 2, byrow = TRUE))),
    sf::st_polygon(list(matrix(c(1,1, 4,1, 4,4, 1,4, 1,1), ncol = 2, byrow = TRUE))),
    sf::st_polygon(list(matrix(c(2,2, 5,2, 5,5, 2,5, 2,2), ncol = 2, byrow = TRUE))),
    sf::st_polygon(list(matrix(c(3,3, 6,3, 6,6, 3,6, 3,3), ncol = 2, byrow = TRUE)))
  ),
  crs = 4326
)

## intersect all geometries into their common area
polys_inters <- ddbs_intersection_agg(polys, mode = "sf")

## intersect within groups
polys_inters_grp <- ddbs_intersection_agg(polys, by = "grp", mode = "sf")

## plot them
plot(polys["grp"], key.pos = NULL, reset = FALSE)
plot(polys_inters_grp["grp"], pal = c("red", "blue"), add = TRUE)
plot(sf::st_geometry(polys_inters), col = "black", pch = 19, cex = 2, add = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{get_query_name}
\alias{get_query_name}
\title{Get names for the query}
\usage{
get_query_name(name)
}
\arguments{
\item{name}{table name}
}
\value{
list with fixed names
}
\description{
Get names for the query
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_extension.R
\name{ddbs_extension_info}
\alias{ddbs_extension_info}
\title{Glimpse the status of a DuckDB extension}
\usage{
ddbs_extension_info(conn = NULL, extension = "spatial")
}
\arguments{
\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{extension}{name of the extension to inspect, default is "spatial"}
}
\value{
A one-row \code{tibble} with the extension's metadata (invisibly).
Called mainly for the glimpse printed as a side effect.
}
\description{
Retrieves the row from DuckDB's \code{duckdb_extensions()} catalog for a given
extension (the spatial extension by default) and prints a transposed
\code{\link[dplyr]{glimpse}} of it: whether it is installed and loaded, its
version, install path, description, and so on.
}
\examples{
\dontrun{
## load package
library(duckspatial)

# inspect the spatial extension on the default connection
ddbs_extension_info()

# or pass an explicit connection
conn <- ddbs_create_conn()
ddbs_extension_info(conn)
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read.R
\name{ddbs_read_vector}
\alias{ddbs_read_vector}
\title{Load spatial vector data from DuckDB into R}
\usage{
ddbs_read_vector(conn, name, clauses = NULL, quiet = FALSE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{clauses}{character, additional SQL code to modify the query from the
table (e.g. "WHERE ...", "ORDER BY...")}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
an sf object
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{ddbs_read_vector()} was renamed to \code{\link{ddbs_read_table}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read.R
\name{ddbs_read_table}
\alias{ddbs_read_table}
\title{Reads a vectorial table from DuckDB into R}
\usage{
ddbs_read_table(conn, name, clauses = NULL, quiet = FALSE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{clauses}{character, additional SQL code to modify the query from the
table (e.g. "WHERE ...", "ORDER BY...")}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
an sf object
}
\description{
Retrieves the data from a DuckDB table, view, or Arrow view with a geometry
column, and converts it to an R \code{sf} object. This function works with
both persistent tables created by \code{ddbs_write_table} and temporary
Arrow views created by \code{ddbs_register_table}.
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(sf)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## create random points
random_points <- data.frame(
  id = 1:5,
  x = runif(5, min = -180, max = 180),
  y = runif(5, min = -90, max = 90)
)

## convert to sf
sf_points <- st_as_sf(random_points, coords = c("x", "y"), crs = 4326)

## Example 1: Write and read persistent table
ddbs_write_vector(conn, sf_points, "points")
ddbs_read_table(conn, "points")

## Example 2: Register and read Arrow view (faster, temporary)
ddbs_register_vector(conn, sf_points, "points_view")
ddbs_read_table(conn, "points_view")

## disconnect from db
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_options.R
\name{ddbs_options}
\alias{ddbs_options}
\title{Get or set global duckspatial options}
\usage{
ddbs_options(output_type = NULL, mode = NULL, duckdb_storage_version = NULL)
}
\arguments{
\item{output_type}{Character string. Controls the default return type for \link{ddbs_collect}.
Must be one of:
\itemize{
\item \code{"sf"} (default): Eagerly collected \code{sf} object (in-memory).
\item \code{"tibble"}: Eagerly collected \code{tibble} without geometry.
\item \code{"raw"}: Eagerly collected \code{tibble} with geometry as raw WKB bytes.
\item \code{"geoarrow"}: Eagerly collected \code{tibble} with geometry as \code{geoarrow_vctr}.
}
If \code{NULL} (the default), the existing option is not changed.}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
If \code{NULL} (the default), the existing option is not changed.}

\item{duckdb_storage_version}{Character. The default DuckDB storage compatibility version
for newly created database files. See \url{https://duckdb.org/docs/internals/storage}
for details.}
}
\value{
Invisibly returns a list containing the currently set options.
}
\description{
Get or set global duckspatial options
}
\examples{
\dontrun{
# Set default mode to geoarrow
ddbs_options(mode = "geoarrow")

# Set default output to tibble
ddbs_options(output_type = "tibble")

# Check current settings
ddbs_options()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_queries.R
\name{ddbs_get_npoints}
\alias{ddbs_get_npoints}
\alias{ddbs_get_ngeometries}
\alias{ddbs_get_ninterior_rings}
\title{Count geometry components}
\usage{
ddbs_get_npoints(
  x,
  new_column = "npoints",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_get_ngeometries(
  x,
  new_column = "ngeometries",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_get_ninterior_rings(
  x,
  new_column = "ninterior_rings",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{new_column}{Name of the new column to create on the input data. Ignored
with \code{mode = "sf"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Functions to count the number of points, sub-geometries or interior rings in
a geometry
}
\details{
These functions query structural properties of geometries:
\itemize{
\item \code{ddbs_get_npoints()} returns the number of points (vertices) in a geometry.
For LINESTRING geometries this is the vertex count; for POLYGON types it
includes all vertices of the exterior ring and any interior rings.
\item \code{ddbs_get_ngeometries()} returns the number of sub-geometries in a
GEOMETRYCOLLECTION or MULTI* geometry (e.g. MULTIPOLYGON, MULTILINESTRING).
Returns 1 for simple (non-collection) geometry types.
\item \code{ddbs_get_ninterior_rings()} returns the number of interior rings (holes)
in a POLYGON geometry. Returns 0 for polygons without holes and \code{NULL} for
non-polygon geometry types.
}
}
\examples{
\dontrun{
## load packages
library(dplyr)
library(duckspatial)

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson",
  package = "duckspatial")
)

## count points, sub-geometries and interior rings
ddbs_get_npoints(countries_ddbs)
ddbs_get_ngeometries(countries_ddbs)
ddbs_get_ninterior_rings(countries_ddbs)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df.R
\name{is_duckspatial_df}
\alias{is_duckspatial_df}
\title{Check if object is a duckspatial_df}
\usage{
is_duckspatial_df(x)
}
\arguments{
\item{x}{Object to test}
}
\value{
Logical
}
\description{
Check if object is a duckspatial_df
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_write_dataset.R
\name{ddbs_write_dataset}
\alias{ddbs_write_dataset}
\title{Write spatial dataset to disk}
\usage{
ddbs_write_dataset(
  data,
  path,
  gdal_driver = NULL,
  conn = NULL,
  overwrite = FALSE,
  crs = NULL,
  layer = "spatial",
  options = list(),
  partitioning = if (inherits(data, c("tbl_lazy", "duckspatial_df")))
    dplyr::group_vars(data) else NULL,
  parquet_compression = NULL,
  parquet_row_group_size = NULL,
  layer_creation_options = NULL,
  quiet = FALSE,
  duckdb_storage_version = duckspatial_storage_default()
)
}
\arguments{
\item{data}{A \code{duckspatial_df}, \code{tbl_lazy} (DuckDB), or \code{sf} object.}

\item{path}{Path to output file.}

\item{gdal_driver}{GDAL driver name for writing spatial formats. If \code{NULL} (default),
the driver is auto-detected from the file extension for common formats:
\itemize{
\item \code{.geojson}, \code{.json} → "GeoJSON"
\item \code{.shp} → "ESRI Shapefile"
\item \code{.gpkg} → "GPKG"
\item \code{.fgb} → "FlatGeobuf"
\item \code{.kml} → "KML"
\item \code{.gpx} → "GPX"
\item \code{.gml} → "GML"
\item \code{.sqlite} → "SQLite"
}

For \strong{non-standard file extensions} (e.g., \code{.dat}, \code{.xyz}) or to \strong{explicitly override}
auto-detection, specify the exact driver name as it appears in \code{ddbs_drivers()$short_name}.
Examples: \code{gdal_driver = "GeoJSON"}, \code{gdal_driver = "ESRI Shapefile"}.

\strong{Note}: If you specify a driver that doesn't match the file extension (e.g.,
\code{path = "output.shp"} with \code{gdal_driver = "GeoJSON"}), a warning will be issued but
your explicit driver choice will be honored (creating a GeoJSON file with \code{.shp} extension).

The function validates that the specified driver is available and writable on your
system. Note: \code{.parquet} and \code{.csv} files use native DuckDB writers and do not
require a GDAL driver.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{overwrite}{Logical. If \code{TRUE}, overwrites existing file.}

\item{crs}{Output CRS (e.g., "EPSG:4326"). Passed to GDAL as \code{SRS} option. Ignored for Parquet.}

\item{layer}{Table name for native DuckDB database output.}

\item{options}{Named list of additional options passed to \code{COPY}.}

\item{partitioning}{Character vector of columns to partition by (Parquet/CSV only).}

\item{parquet_compression}{Compression codec for Parquet.}

\item{parquet_row_group_size}{Row group size for Parquet.}

\item{layer_creation_options}{GDAL layer creation options.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{duckdb_storage_version}{Storage compatibility for newly created native
DuckDB database files (\code{.duckdb}, \code{.db}, \code{.ddb}). See
\url{https://duckdb.org/docs/internals/storage} for more information on
DuckDB storage versions and compatibility.
\itemize{
\item \code{"v1.5.0"} (\strong{Native Spatial Storage}, Default): Preserves
CRS metadata in native DuckDB \code{GEOMETRY} columns. Requires
DuckDB >= 1.5.0 to open the file.
\item \code{"v1.0.0"} (\strong{Legacy Compatibility}): Creates
files readable by older DuckDB versions (>= 1.0.0). Persists CRS
metadata in duckspatial-managed column comments (a convention not
recognized by other spatial software).
\item \code{"latest"}: Use the highest storage version supported by your
installed DuckDB engine.
}
Other major version strings like \code{"v1.4.0"}, \code{"v1.3.0"}, etc., are also supported.}
}
\value{
The \code{path} invisibly.
}
\description{
Writes spatial data to disk using DuckDB's \code{COPY} command for Parquet and
GDAL spatial formats, or as a native DuckDB database for \code{.duckdb}, \code{.db},
and \code{.ddb} paths. Format is auto-detected from file extension for common
formats, or can be specified explicitly via \code{gdal_driver}.
}
\details{
Persistent DuckDB database files created by duckspatial use \strong{Native
Spatial Storage} (\code{storage_version = "v1.5.0"}) by default so CRS metadata is
retained in native \code{GEOMETRY} columns. These files require DuckDB >= 1.5.0
to open; use \strong{Legacy Compatibility} (\code{storage_version = "v1.0.0"})
when the output must be readable by older DuckDB versions.
}
\examples{
\dontrun{
library(duckspatial)

# Read example data
path <- system.file("spatial/countries.geojson", package = "duckspatial")
ds <- ddbs_open_dataset(path)

# Auto-detect format from extension
ddbs_write_dataset(ds, "output.geojson")
ddbs_write_dataset(ds, "output.gpkg")
ddbs_write_dataset(ds, "output.parquet")

# Explicit GDAL driver for non-standard extension
ddbs_write_dataset(ds, "mydata.dat", gdal_driver = "GeoJSON")

# See available drivers on your system
drivers <- ddbs_drivers()
writable <- drivers[drivers$can_create == TRUE, ]
head(writable)

# CRS override
ddbs_write_dataset(ds, "output_3857.geojson", crs = "EPSG:3857")

# Overwrite existing file
ddbs_write_dataset(ds, "output.gpkg", overwrite = TRUE)
}
}
\references{
This function is inspired by and builds upon the logic found in the
\code{duckdbfs} package (\url{https://github.com/cboettig/duckdbfs}),
particularly its \code{write_dataset} and \code{write_geo} functions.
For advanced features like cloud storage (S3) support, the
\code{duckdbfs} package is highly recommended.
}
\seealso{
\code{\link[=ddbs_drivers]{ddbs_drivers()}} to list all available GDAL drivers and formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{ddbs_default_conn}
\alias{ddbs_default_conn}
\title{Get or create default DuckDB connection with spatial extension installed and loaded}
\usage{
ddbs_default_conn(create = TRUE, ...)
}
\arguments{
\item{create}{Logical. If TRUE and no connection exists, create one.
Default is TRUE.}

\item{...}{Additional parameters to pass to \code{ddbs_create_conn()}}
}
\value{
A \code{duckdb_connection} or NULL if no connection exists and
create = FALSE
}
\description{
Get or create default DuckDB connection with spatial extension installed and loaded
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{crs_equal}
\alias{crs_equal}
\title{Compare two CRS objects for equality}
\usage{
crs_equal(crs1, crs2)
}
\arguments{
\item{crs1}{First CRS object}

\item{crs2}{Second CRS object}
}
\value{
Logical indicating if CRS are equal
}
\description{
Properly compares CRS objects, handling different representations of the same CRS.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{convert_to_sf_wkb}
\alias{convert_to_sf_wkb}
\title{Converts from data frame to sf using WKB conversion}
\usage{
convert_to_sf_wkb(data, crs, x_geom)
}
\arguments{
\item{data}{a tibble or data frame}

\item{crs}{The coordinates reference system}

\item{x_geom}{name of geometry column}
}
\value{
sf
}
\description{
Converts a table that has been read from DuckDB into an sf object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_centroid}
\alias{ddbs_centroid}
\title{Calculates the centroid of geometries}
\usage{
ddbs_centroid(
  x,
  method = "centroid",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{method}{Character string specifying the method to calculate the
centroid. Must be one of "centroid" (default) or "surface". "centroid"
calculates the default centroid, which may fall outside the geometry
for certain shapes (e.g., donuts). "surface" calculates a point guaranteed
to fall within the geometry.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Returns the geometric center (centroid) of a geometry as a point,
representing its average position.
}
\examples{
\dontrun{
## load package
library(duckspatial)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", 
  package = "duckspatial")
)

## store in duckdb
ddbs_write_vector(conn, argentina_ddbs, "argentina")

## centroid
ddbs_centroid("argentina", conn)

## centroid without using a connection
ddbs_centroid(argentina_ddbs)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_drop_geometry}
\alias{ddbs_drop_geometry}
\title{Drop geometry column from a duckspatial_df object}
\usage{
ddbs_drop_geometry(x)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}
}
\value{
A lazy tibble backed by dbplyr
}
\description{
Removes the geometry column from a \code{duckspatial_df} object, returning a
lazy tibble without spatial information.
}
\examples{
\dontrun{
## load package
library(duckspatial)

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson",
  package = "duckspatial")
)

## drop geometry column
countries_tbl <- ddbs_drop_geometry(countries_ddbs)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_coordinates.R
\name{ddbs_xy}
\alias{ddbs_xy}
\alias{ddbs_x}
\alias{ddbs_y}
\alias{ddbs_m}
\alias{ddbs_z}
\title{Extract coordinates from geometries}
\usage{
ddbs_x(
  x,
  new_column = "X",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_y(
  x,
  new_column = "Y",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_m(
  x,
  new_column = "M",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_z(
  x,
  new_column = "Z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{new_column}{Name of the new column to store the extracted coordinate.
Defaults to \code{"X"}, \code{"Y"}, \code{"M"}, or \code{"Z"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Extracts the X, Y, M, or Z coordinates from \code{POINT} geometries
}
\details{
\itemize{
\item \code{ddbs_x()}: Extracts the X coordinate (longitude).
\item \code{ddbs_y()}: Extracts the Y coordinate (latitude).
\item \code{ddbs_m()}: Extracts the M coordinate (measure).
\item \code{ddbs_z()}: Extracts the Z coordinate (elevation).
}
}
\examples{
\dontrun{
## load package
library(duckspatial)

## read data
argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson",
  package = "duckspatial")
)

## extract coordinates
ddbs_x(argentina_ddbs)
ddbs_y(argentina_ddbs)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_coordinates.R
\name{ddbs_coord_bounds}
\alias{ddbs_coord_bounds}
\alias{ddbs_xmax}
\alias{ddbs_xmin}
\alias{ddbs_ymax}
\alias{ddbs_ymin}
\alias{ddbs_zmax}
\alias{ddbs_zmin}
\alias{ddbs_mmax}
\alias{ddbs_mmin}
\title{Coordinate bounds of geometries}
\usage{
ddbs_xmax(
  x,
  new_column = "xmax",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_xmin(
  x,
  new_column = "xmin",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_ymax(
  x,
  new_column = "ymax",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_ymin(
  x,
  new_column = "ymin",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_zmax(
  x,
  new_column = "zmax",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_zmin(
  x,
  new_column = "zmin",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_mmax(
  x,
  new_column = "mmax",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_mmin(
  x,
  new_column = "mmin",
  by_feature = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{new_column}{Name of the new column. Defaults to the lowercase function
name (e.g. \code{"xmax"}, \code{"xmin"}, \code{"ymax"}, \ldots).}

\item{by_feature}{Logical. If \code{TRUE}, the geometric operation is applied
separately to each geometry. If \code{FALSE}, the geometric
operation is applied to the data as a whole.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
\itemize{
\item \code{by_feature = TRUE} and \code{mode = "duckspatial"} (default):
A \code{duckspatial_df}.
\item \code{by_feature = TRUE} and \code{mode = "sf"}: A numeric vector.
\item \code{by_feature = FALSE}: A single \code{numeric} scalar.
\item When \code{name} is provided: writes the table in DuckDB and
returns \code{TRUE} (invisibly).
}
}
\description{
Returns the minimum or maximum value of a specific coordinate axis across
all points of a geometry. When \code{by_feature = TRUE} (default), a value
is computed per row. When \code{by_feature = FALSE}, a single global
value is returned for the entire dataset.
}
\details{
\itemize{
\item \code{ddbs_xmax()} / \code{ddbs_xmin()}: maximum / minimum X coordinate.
\item \code{ddbs_ymax()} / \code{ddbs_ymin()}: maximum / minimum Y coordinate.
\item \code{ddbs_zmax()} / \code{ddbs_zmin()}: maximum / minimum Z coordinate.
\item \code{ddbs_mmax()} / \code{ddbs_mmin()}: maximum / minimum M (measure) coordinate.
}

When \code{by_feature = FALSE}, the result is always a single \code{numeric}
scalar representing the global extreme across the entire dataset.
}
\examples{
\dontrun{
library(duckspatial)

argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", package = "duckspatial")
)

## per-feature X extent (default)
ddbs_xmax(argentina_ddbs)
ddbs_xmin(argentina_ddbs)

## global bounding values
ddbs_xmax(argentina_ddbs, by_feature = FALSE)
ddbs_ymin(argentina_ddbs, by_feature = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_geom_conversion.R
\name{ddbs_geom_from}
\alias{ddbs_geom_from}
\alias{ddbs_geom_from_text}
\alias{ddbs_geom_from_wkb}
\alias{ddbs_geom_from_hexwkb}
\alias{ddbs_geom_from_hexewkb}
\alias{ddbs_geom_from_geojson}
\title{Create geometries from standard interchange formats}
\usage{
ddbs_geom_from_text(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_geom_from_wkb(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_geom_from_hexwkb(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_geom_from_hexewkb(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_geom_from_geojson(
  x,
  crs = NULL,
  ...,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{For \code{ddbs_geom_from_wkb()}, a list of raw vectors (binary WKB). For
all other functions, a character vector of serialized geometries.}

\item{crs}{Character or numeric CRS specification (e.g. \code{"EPSG:4326"} or
\code{4326}) to assign to the resulting geometries. Defaults to \code{NULL} (no CRS
assigned).}

\item{...}{Named vectors of additional attribute columns to include in the
output. Each must have the same length as \code{x}.}

\item{geom_col}{Name of the geometry column in the output. Defaults to
\code{"geometry"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Parse serialized geometries from common interchange formats into a spatial
object, using DuckDB spatial deserialization functions. These are the
inverses of the \code{\link[=ddbs_as_text]{ddbs_as_text()}} / \code{\link[=ddbs_as_wkb]{ddbs_as_wkb()}} / \code{\link[=ddbs_as_hexwkb]{ddbs_as_hexwkb()}} /
\code{\link[=ddbs_as_geojson]{ddbs_as_geojson()}} serializers.
\itemize{
\item \code{ddbs_geom_from_text()} – Parse Well-Known Text (WKT)
\item \code{ddbs_geom_from_wkb()} – Parse Well-Known Binary (WKB)
\item \code{ddbs_geom_from_hexwkb()} – Parse hexadecimal Well-Known Binary (HEXWKB)
\item \code{ddbs_geom_from_hexewkb()} – Parse hexadecimal Extended Well-Known Binary (HEXEWKB)
\item \code{ddbs_geom_from_geojson()} – Parse GeoJSON
}
}
\details{
These functions are thin wrappers around the DuckDB spatial deserialization
functions (\code{ST_GeomFromText}, \code{ST_GeomFromWKB}, \code{ST_GeomFromHEXWKB},
\code{ST_GeomFromHEXEWKB}, and \code{ST_GeomFromGeoJSON}).

They are useful for importing geometries produced by external spatial tools,
databases, and web services back into a \code{duckspatial} workflow.
}
\examples{
\dontrun{
library(duckspatial)

## from WKT
ddbs_geom_from_text(
  c("POINT (-58.38 -34.60)", "POINT (-64.18 -31.42)"),
  crs = 4326
)

## from GeoJSON, with an extra attribute column
ddbs_geom_from_geojson(
  '{"type":"Point","coordinates":[-58.38,-34.60]}',
  city = "Buenos Aires",
  crs  = 4326
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_dimensions.R
\name{ddbs_has_dim}
\alias{ddbs_has_dim}
\alias{ddbs_has_z}
\alias{ddbs_has_m}
\title{Check geometry dimensions}
\usage{
ddbs_has_z(
  x,
  by_feature = TRUE,
  new_column = "has_z",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_has_m(
  x,
  by_feature = TRUE,
  new_column = "has_m",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{by_feature}{Logical. If \code{TRUE}, the geometric operation is applied
separately to each geometry. If \code{FALSE}, the geometric
operation is applied to the data as a whole.}

\item{new_column}{Name of the new column to create on the input data. Ignored
with \code{mode = "sf"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Functions to check whether geometries have Z (elevation) or M (measure) dimensions
}
\details{
These functions check for additional coordinate dimensions beyond X and Y:
\itemize{
\item \code{ddbs_has_z()} checks if a geometry has Z coordinates (elevation/altitude values).
Geometries with Z dimension are often referred to as 3D geometries and have
coordinates in the form (X, Y, Z).
\item \code{ddbs_has_m()} checks if a geometry has M coordinates (measure values). The M
dimension typically represents a measurement along the geometry, such as distance
or time, and results in coordinates of the form (X, Y, M) or (X, Y, Z, M).
}
}
\examples{
\dontrun{
## load packages
library(dplyr)
library(duckspatial)

## create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |> 
  filter(ISO3_CODE != "ATA")

## check if it has Z or M
ddbs_has_m(countries_ddbs)
ddbs_has_z(countries_ddbs)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_write.R
\name{ddbs_write_table}
\alias{ddbs_write_table}
\title{Write an SF Object to a DuckDB Database}
\usage{
ddbs_write_table(
  conn,
  data,
  name,
  overwrite = FALSE,
  temp_view = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{data}{A \code{sf} object to write to the DuckDB database, or the path to
a local file that can be read with \code{ST_READ}}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{temp_view}{If \code{TRUE}, registers the \code{sf} object as a temporary Arrow-backed database 'view'
using \link{ddbs_register_table} instead of creating a persistent table. This is much faster but the view
will not persist. Defaults to \code{FALSE}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
TRUE (invisibly) for successful import
}
\description{
This function writes a Simple Features (SF) object into a DuckDB database as a new table.
The table is created in the specified schema of the DuckDB database.
}
\details{
\strong{CRS Persistence:} \code{duckspatial} ensures CRS metadata is retained across
sessions using two strategies: \strong{Native Spatial Storage} (for DuckDB 1.5.0+
databases) and \strong{Legacy Compatibility} (using column comments for older
database versions). For file-based spatial data interchange,
\code{\link[=ddbs_write_dataset]{ddbs_write_dataset()}} to GeoParquet (\code{.parquet}) is recommended.
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(sf)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## create random points
random_points <- data.frame(
  id = 1:5,
  x = runif(5, min = -180, max = 180),  # Random longitude values
  y = runif(5, min = -90, max = 90)     # Random latitude values
)

## convert to sf
sf_points <- st_as_sf(random_points, coords = c("x", "y"), crs = 4326)

## insert data into the database
ddbs_write_table(conn, sf_points, "points")

## read data back into R
ddbs_read_table(conn, "points")

## disconnect from db
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{get_conn_from_input}
\alias{get_conn_from_input}
\title{Get DuckDB connection from an object}
\usage{
get_conn_from_input(x)
}
\arguments{
\item{x}{A duckspatial_df, tbl_lazy, duckdb_connection, or NULL}
}
\value{
A duckdb_connection or NULL
}
\description{
Extracts the connection from duckspatial_df, tbl_lazy, or validates a direct connection.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_crs.R
\name{ddbs_set_crs}
\alias{ddbs_set_crs}
\title{Set the coordinate reference system of geometries}
\usage{
ddbs_set_crs(
  x,
  y,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Target CRS. Can be:
\itemize{
\item A character string with EPSG code (e.g., "EPSG:4326")
\item A \code{crs} object created with \link[sf:st_crs]{sf::st_crs}
}}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Assigns or replaces the coordinate reference system (CRS) of geometries
without transforming their coordinates. This is useful when the CRS is
missing or incorrectly defined.
}
\examples{
\dontrun{
## load package
library(duckspatial)
library(sf)

## read data
rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson",
  package = "duckspatial")
)

## Remove CRS
rivers_no_crs_ddbs <- ddbs_set_crs(rivers_ddbs, st_crs(NA))

## Set the CRS back
ddbs_set_crs(rivers_no_crs_ddbs, "EPSG:3035")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df.R
\name{as_duckspatial_df}
\alias{as_duckspatial_df}
\alias{as_duckspatial_df.duckspatial_df}
\alias{as_duckspatial_df.sf}
\alias{as_duckspatial_df.tbl_duckdb_connection}
\alias{as_duckspatial_df.tbl_lazy}
\alias{as_duckspatial_df.character}
\alias{as_duckspatial_df.data.frame}
\title{Convert objects to duckspatial_df}
\usage{
as_duckspatial_df(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{duckspatial_df}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{sf}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{tbl_duckdb_connection}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{tbl_lazy}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{character}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)

\method{as_duckspatial_df}{data.frame}(x, conn = NULL, crs = NULL, geom_col = NULL, ...)
}
\arguments{
\item{x}{Object to convert (sf, tbl_lazy, data.frame, or table name)}

\item{conn}{DuckDB connection (required for character table names)}

\item{crs}{CRS object or string. Auto-detected from \code{sf} objects and
persistent DuckDB tables.}

\item{geom_col}{Geometry column name (default: "geom")}

\item{...}{Additional arguments passed to methods:
\describe{
\item{\code{coords}}{Character vector of length 2 for point ingestion}
\item{\code{wkt}}{Character name of WKT column for ingestion}
\item{\code{remove}}{Logical. If TRUE (default), coordinate/WKT columns are removed}
\item{\code{na.fail}}{Logical. If TRUE (default), errors on missing values}
}}
}
\value{
A duckspatial_df object
}
\description{
\code{as_duckspatial_df()} creates a lazy spatial data frame (\code{duckspatial_df}) from
various inputs. When \code{x} is a table name (character) or an existing DuckDB
table (\code{tbl_duckdb_connection}), the function creates a zero-copy representation
of the data directly from the database without loading it into memory. This is
the canonical way to "register" or wrap existing persistent spatial tables.

\strong{CRS Persistence:} \code{duckspatial} reads native DuckDB 1.5.0+ CRS metadata
and, for compatibility with files written by older versions of \code{duckspatial},
CRS metadata stored in column comments. DuckDB files saved in pre-1.5.0
format without \code{duckspatial}-managed comments will not have CRS information
and will default to \code{NA} with a warning.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_sf_methods.R
\name{ddbs_format_crs}
\alias{ddbs_format_crs}
\title{Format a CRS object compactly for printing}
\usage{
ddbs_format_crs(crs)
}
\arguments{
\item{crs}{An sf crs object}
}
\value{
A character string
}
\description{
Format a CRS object compactly for printing
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read.R
\name{ddbs_read_meta}
\alias{ddbs_read_meta}
\title{Read metadata from a spatial file}
\usage{
ddbs_read_meta(path, conn = NULL)
}
\arguments{
\item{path}{character, path to the spatial file to inspect.}

\item{conn}{A \code{DBIConnection} object to a DuckDB database}
}
\value{
A \code{tibble} with one row per file and the
following columns:
\describe{
\item{file_name}{Path to the file.}
\item{driver_short_name}{Short name of the GDAL driver (e.g.
\code{"GPKG"}).}
\item{driver_long_name}{Full name of the GDAL driver (e.g.
\code{"GeoPackage"}).}
\item{layers}{A list-column of data frames, one per file, each
describing the layers contained in the file. Unnest with
\code{\link[tidyr]{unnest}} to access individual layer attributes
such as name, geometry type, and feature count.}
}
}
\description{
Retrieves file-level metadata from a spatial vector file (e.g. GeoPackage,
Shapefile, GeoJSON) using DuckDB's \code{ST_Read_Meta()} function. Returns
information about the file's driver and its layers as a tibble.
}
\examples{
\dontrun{
## Read metadata from a GeoPackage
meta <- ddbs_read_meta(
  system.file("spatial/rivers.geojson",
  package = "duckspatial")
)

## View file-level metadata
meta

## Inspect layer details
tidyr::unnest(meta, layers)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_glimpse}
\alias{ddbs_glimpse}
\title{Check first rows of the data}
\usage{
ddbs_glimpse(conn, name, quiet = FALSE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Invisibly \code{duckspatial_df} object
}
\description{
Prints a transposed table of the first rows of a DuckDB table, similarly
as the S3 \link[dplyr:glimpse]{dplyr::glimpse} method.
}
\examples{
\dontrun{
library(duckspatial)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
argentina_sf <- ddbs_open_dataset(system.file("spatial/argentina.geojson", package = "duckspatial"))

## store in duckdb
ddbs_write_table(conn, argentina_sf, "argentina")

## glimpse the inserted table
ddbs_glimpse(conn, "argentina")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_create_schema}
\alias{ddbs_create_schema}
\title{Check and create schema}
\usage{
ddbs_create_schema(conn, name, quiet = FALSE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{name}{A character string with the name of the schema to be created}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
TRUE (invisibly) for successful schema creation
}
\description{
Check and create schema
}
\examples{
## load packages
\dontrun{
library(duckspatial)
library(duckdb)

## connect to in memory database
conn <- ddbs_create_conn(dbdir = "memory")

## create a new schema
ddbs_create_schema(conn, "new_schema")

## check schemas
dbGetQuery(conn, "SELECT * FROM information_schema.schemata;")

## disconnect from db
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_utils.R
\name{ddbs_drivers}
\alias{ddbs_drivers}
\title{Get list of GDAL drivers and file formats}
\usage{
ddbs_drivers(conn = NULL)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database.
If not specified (\code{conn = NULL}), uses the default connection created with \code{ddbs_default_conn()} (or a temporary one).}
}
\value{
\code{data.frame}
}
\description{
Get list of GDAL drivers and file formats
}
\examples{
\dontrun{
## load package
library(duckspatial)

## database setup
conn <- ddbs_create_conn()

## check drivers
ddbs_drivers(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{ddbs_temp_table_name}
\alias{ddbs_temp_table_name}
\title{Generate unique temporary table name}
\usage{
ddbs_temp_table_name()
}
\value{
Character string with unique table name
}
\description{
Creates a unique name for temporary tables to avoid collisions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_crs.R
\name{ddbs_crs}
\alias{ddbs_crs}
\alias{ddbs_crs.duckspatial_df}
\alias{ddbs_crs.sf}
\alias{ddbs_crs.tbl_duckdb_connection}
\alias{ddbs_crs.character}
\alias{ddbs_crs.duckdb_connection}
\alias{ddbs_crs.numeric}
\alias{ddbs_crs.crs}
\alias{ddbs_crs.data.frame}
\alias{ddbs_crs.default}
\title{Check CRS of spatial objects or database tables}
\usage{
ddbs_crs(x, ...)

\method{ddbs_crs}{duckspatial_df}(x, ...)

\method{ddbs_crs}{sf}(x, ...)

\method{ddbs_crs}{tbl_duckdb_connection}(x, ...)

\method{ddbs_crs}{character}(x, conn, ...)

\method{ddbs_crs}{duckdb_connection}(x, name, ...)

\method{ddbs_crs}{numeric}(x, ...)

\method{ddbs_crs}{crs}(x, ...)

\method{ddbs_crs}{data.frame}(x, ...)

\method{ddbs_crs}{default}(x, ...)
}
\arguments{
\item{x}{An object containing spatial data. Can be:
\itemize{
\item \code{duckspatial_df}: Lazy spatial data frame (CRS from attributes)
\item \code{sf}: sf object (CRS from sf metadata)
\item \code{character}: Name of table in database (requires \code{conn})
}}

\item{...}{Additional arguments passed to methods}

\item{conn}{A DuckDB connection (required for character method)}

\item{name}{Table name (for backward compatibility when first arg is connection)}
}
\value{
CRS object from \code{sf} package
}
\description{
This is an S3 generic that extracts CRS information from various spatial objects.
}
\examples{
\dontrun{
## load packages
library(duckdb)
library(duckspatial)
library(sf)

# Method 1: duckspatial_df objects
nc <- ddbs_open_dataset(system.file("shape/nc.shp", package = "sf"))
ddbs_crs(nc)

# Method 2: sf objects
nc_sf <- st_read(system.file("shape/nc.shp", package = "sf"))
ddbs_crs(nc_sf)

# Method 3: table name in database
conn <- ddbs_create_conn(dbdir = "memory")
ddbs_write_table(conn, nc_sf, "nc_table")
ddbs_crs(conn, "nc_table")
ddbs_stop_conn(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_affine.R
\name{ddbs_scale}
\alias{ddbs_scale}
\title{Scale geometries by X and Y factors}
\usage{
ddbs_scale(
  x,
  x_scale = 1,
  y_scale = 1,
  by_feature = FALSE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{x_scale}{numeric value specifying the scaling factor in the X direction (default = 1)}

\item{y_scale}{numeric value specifying the scaling factor in the Y direction (default = 1)}

\item{by_feature}{Logical. If \code{TRUE}, the geometric operation is applied
separately to each geometry. If \code{FALSE}, the geometric
operation is applied to the data as a whole.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Resizes geometries by specified X and Y scale factors. By default, scaling is
performed relative to the centroid of all geometries; if \code{by_feature = TRUE},
each geometry is scaled relative to its own centroid.
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(dplyr)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |>
  filter(CNTR_ID \%in\% c("PT", "ES", "FR", "IT"))

## store in duckdb
ddbs_write_table(conn, countries_ddbs, "countries")

## scale to 150\% in both directions
ddbs_scale(conn = conn, "countries", x_scale = 1.5, y_scale = 1.5)

## scale to 200\% horizontally, 50\% vertically
ddbs_scale(conn = conn, "countries", x_scale = 2, y_scale = 0.5)

## scale all features together (default)
ddbs_scale(countries_ddbs, x_scale = 1.5, y_scale = 1.5, by_feature = FALSE)

## scale each feature independently
ddbs_scale(countries_ddbs, x_scale = 1.5, y_scale = 1.5, by_feature = TRUE)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_unary.R
\name{ddbs_geometry_type}
\alias{ddbs_geometry_type}
\title{Get the geometry type of features}
\usage{
ddbs_geometry_type(x, by_feature = TRUE, conn = NULL)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{by_feature}{Logical. If \code{TRUE}, the geometric operation is applied
separately to each geometry. If \code{FALSE}, the geometric
operation is applied to the data as a whole.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}
}
\value{
A factor with geometry type(s)
}
\description{
Returns the type of each geometry (e.g., POINT, LINESTRING, POLYGON) in the
input features.
}
\examples{
\dontrun{
## load package
library(duckspatial)

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
)

# option 1: passing sf objects
# Get geometry type for each feature
ddbs_geometry_type(countries_ddbs)

# Get overall geometry type
ddbs_geometry_type(countries_ddbs, by_feature = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{get_geom_name}
\alias{get_geom_name}
\title{Get column names in a DuckDB database}
\usage{
get_geom_name(conn, x, rest = FALSE, collapse = FALSE, table_id = NULL)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{x}{name of the table}

\item{rest}{whether to return geometry column name, of the rest of the columns}
}
\value{
name of the geometry column of a table
}
\description{
Get column names in a DuckDB database
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{import_view_to_connection}
\alias{import_view_to_connection}
\title{Import a view/table from one connection to another}
\usage{
import_view_to_connection(
  target_conn,
  source_conn,
  source_object,
  target_name = NULL,
  quiet = FALSE
)
}
\arguments{
\item{target_conn}{Target DuckDB connection}

\item{source_conn}{Source DuckDB connection}

\item{source_object}{duckspatial_df, tbl_lazy, or tbl_duckdb_connection from source_conn}

\item{target_name}{Name for view in target connection (auto-generated if NULL)}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
List with imported view name and cleanup function
}
\description{
Enables cross-connection operations by importing views using one of three strategies.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_extent_funs.R
\name{ddbs_make_envelope}
\alias{ddbs_make_envelope}
\title{Create a rectangular polygon from bounding coordinates}
\usage{
ddbs_make_envelope(
  xmin,
  ymin,
  xmax,
  ymax,
  crs = "EPSG:4326",
  name = NULL,
  conn = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{xmin}{A numeric value for the minimum X (longitude) coordinate.}

\item{ymin}{A numeric value for the minimum Y (latitude) coordinate.}

\item{xmax}{A numeric value for the maximum X (longitude) coordinate.}

\item{ymax}{A numeric value for the maximum Y (latitude) coordinate.}

\item{crs}{A character string specifying the coordinate reference system
of the output geometry. Default is \code{"EPSG:4326"}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Creates a rectangular POLYGON geometry from four bounding coordinates.
}
\examples{
\dontrun{
## load package
library(duckspatial)

## without storing in duckdb
finland_bbox_ddbs <- ddbs_make_envelope(
  xmin = 19.1, ymin = 59.7,
  xmax = 31.6, ymax = 70.1
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_geom_validation.R
\name{ddbs_geom_validation_funs}
\alias{ddbs_geom_validation_funs}
\alias{ddbs_is_simple}
\alias{ddbs_is_valid}
\alias{ddbs_is_closed}
\alias{ddbs_is_empty}
\alias{ddbs_is_ring}
\title{Geometry validation functions}
\usage{
ddbs_is_simple(
  x,
  new_column = "is_simple",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_is_valid(
  x,
  new_column = "is_valid",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_is_closed(
  x,
  new_column = "is_closed",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_is_empty(
  x,
  new_column = "is_empty",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_is_ring(
  x,
  new_column = "is_ring",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{new_column}{Name of the new column to create on the input data. Ignored
with \code{mode = "sf"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
\itemize{
\item \code{mode = "duckspatial"} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{mode = "sf"}: An eagerly collected vector in R memory.
\item When \code{name} is provided: writes the table in the DuckDB connection and returns \code{TRUE} (invisibly).
}
}
\description{
Functions to check various geometric properties and validity conditions of spatial
geometries using DuckDB's spatial extension.
}
\details{
These functions provide different types of geometric validation. Note that by default,
the functions add a new column as a logical vector. This behaviour allows to filter the
data within DuckDB without the need or materializating a vector in R (see details).
\itemize{
\item \code{ddbs_is_valid()} checks if a geometry is valid according to the OGC Simple Features
specification. Invalid geometries may have issues like self-intersections in polygons,
duplicate points, or incorrect ring orientations.
\item \code{ddbs_is_simple()} determines whether geometries are simple, meaning they are free of
self-intersections. For example, a linestring that crosses itself is not simple.
\item \code{ddbs_is_ring()} checks if a linestring geometry is closed (first and last points are
identical) and simple (no self-intersections), forming a valid ring.
\item \code{ddbs_is_empty()} tests whether a geometry is empty, containing no points. Empty
geometries are valid but represent the absence of spatial information.
\item \code{ddbs_is_closed()} determines if a linestring geometry is closed, meaning the first
and last coordinates are identical. Unlike \code{ddbs_is_ring()}, this does not check for
simplicity.
}
}
\examples{
\dontrun{
## load package
library(duckspatial)
library(dplyr)

## create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
)

rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson", 
  package = "duckspatial")
)

## geometry validation
ddbs_is_valid(countries_ddbs)
ddbs_is_simple(countries_ddbs)
ddbs_is_ring(rivers_ddbs)
ddbs_is_empty(countries_ddbs)
ddbs_is_closed(countries_ddbs)

## filter invalid countries
ddbs_is_valid(countries_ddbs) |> filter(!is_valid)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_dplyr_methods.R
\name{duckspatial_df_dplyr}
\alias{duckspatial_df_dplyr}
\alias{dplyr_reconstruct.duckspatial_df}
\alias{compute.duckspatial_df}
\alias{select.duckspatial_df}
\alias{filter.duckspatial_df}
\alias{arrange.duckspatial_df}
\alias{rename.duckspatial_df}
\alias{slice.duckspatial_df}
\alias{head.duckspatial_df}
\alias{glimpse.duckspatial_df}
\alias{mutate.duckspatial_df}
\alias{count.duckspatial_df}
\alias{distinct.duckspatial_df}
\alias{left_join.duckspatial_df}
\alias{inner_join.duckspatial_df}
\alias{right_join.duckspatial_df}
\alias{full_join.duckspatial_df}
\alias{group_by.duckspatial_df}
\alias{ungroup.duckspatial_df}
\alias{summarise.duckspatial_df}
\title{dplyr methods for duckspatial_df}
\usage{
\method{dplyr_reconstruct}{duckspatial_df}(data, template)

\method{compute}{duckspatial_df}(x, name = NULL, temporary = TRUE, ...)

\method{select}{duckspatial_df}(.data, ...)

\method{filter}{duckspatial_df}(.data, ...)

\method{arrange}{duckspatial_df}(.data, ...)

\method{rename}{duckspatial_df}(.data, ...)

\method{slice}{duckspatial_df}(.data, ...)

\method{head}{duckspatial_df}(x, n = 6L, ...)

\method{glimpse}{duckspatial_df}(x, width = NULL, ...)

\method{mutate}{duckspatial_df}(.data, ...)

\method{count}{duckspatial_df}(x, ..., wt = NULL, sort = FALSE, name = NULL)

\method{distinct}{duckspatial_df}(.data, ..., .keep_all = FALSE)

\method{left_join}{duckspatial_df}(x, y, by = NULL, ...)

\method{inner_join}{duckspatial_df}(x, y, by = NULL, ...)

\method{right_join}{duckspatial_df}(x, y, by = NULL, ...)

\method{full_join}{duckspatial_df}(x, y, by = NULL, ...)

\method{group_by}{duckspatial_df}(.data, ..., .add = FALSE, .drop = dplyr::group_by_drop_default(.data))

\method{ungroup}{duckspatial_df}(x, ...)

\method{summarise}{duckspatial_df}(.data, ...)
}
\arguments{
\item{x}{A \code{duckspatial_df} object}

\item{name}{Optional name for the result table. If NULL, a unique temporary
name is generated.}

\item{temporary}{If TRUE (default), creates a temporary table that is
automatically cleaned up when the connection closes.}

\item{...}{Additional arguments passed to \code{\link[dplyr:compute]{dplyr::compute()}}}
}
\value{
A new \code{duckspatial_df} pointing to the materialized table, with
spatial metadata (CRS, geometry column) preserved.
}
\description{
These methods use dplyr's extension mechanism (dplyr_reconstruct) to
properly preserve spatial metadata through operations.

Executes the accumulated query and stores the result in a DuckDB temporary
table. The result remains lazy (a \code{duckspatial_df}) but points to the
materialized data, avoiding repeated computation of complex query plans.
}
\details{
This is useful when you want to:
\itemize{
\item Cache intermediate results for reuse across multiple subsequent operations
\item Simplify complex query plans before heavy operations like spatial joins
\item Force execution at a specific point without pulling data into R memory
}
}
\examples{
\dontrun{
library(dplyr)

# Complex pipeline - compute() caches intermediate result
result <- countries |>
  filter(POP_EST > 50000000) |>
  ddbs_filter(argentina, predicate = "touches") |>
  compute() |>  # Execute and cache here
  select(NAME_ENGL, POP_EST) |>
  ddbs_join(rivers, join = "intersects")

# Check query plan - should reference the cached table
show_query(result)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df.R
\name{handle_heterogeneous_ingestion}
\alias{handle_heterogeneous_ingestion}
\title{Handle ingestion from coords or WKT}
\usage{
handle_heterogeneous_ingestion(x, conn, crs, geom_col, ...)
}
\description{
Handle ingestion from coords or WKT
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_dplyr_methods.R,
%   R/duckspatial_df_sf_methods.R
\name{collect.duckspatial_df}
\alias{collect.duckspatial_df}
\alias{ddbs_collect}
\title{Collect a duckspatial_df with flexible output formats}
\usage{
\method{collect}{duckspatial_df}(x, ..., as = NULL)

ddbs_collect(x, ..., as = c("sf", "tibble", "raw", "geoarrow"))
}
\arguments{
\item{x}{A \code{duckspatial_df} object}

\item{...}{Additional arguments passed to \code{collect}}

\item{as}{Output format. One of:
\describe{
\item{\code{"sf"}}{(Default) Returns an \code{sf} object with \code{sfc} geometry}
\item{\code{"tibble"}}{Returns a tibble with geometry column dropped (fastest)}
\item{\code{"raw"}}{Returns a tibble with geometry as raw WKB bytes}
\item{\code{"geoarrow"}}{Returns a tibble with geometry as \code{geoarrow_vctr}}
}}
}
\value{
Collected data in the specified format

Data in the specified format
}
\description{
Materializes a lazy \code{duckspatial_df} object by executing the underlying
DuckDB query. Supports multiple output formats.
}
\examples{
\dontrun{
library(duckspatial)

# Load lazy spatial data
nc <- ddbs_open_dataset(system.file("shape/nc.shp", package = "sf"))

# Perform lazy operations
result <- nc |> dplyr::filter(AREA > 0.1)

# Collect to sf (default)
result_sf <- ddbs_collect(result)
plot(result_sf["AREA"])

# Collect as tibble without geometry (fast)
result_tbl <- ddbs_collect(result, as = "tibble")

# Collect with raw WKB bytes
result_raw <- ddbs_collect(result, as = "raw")

# Collect as geoarrow for Arrow workflows
result_ga <- ddbs_collect(result, as = "geoarrow")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{dbConnCheck}
\alias{dbConnCheck}
\title{Check if a supported DuckDB connection}
\usage{
dbConnCheck(conn)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}
}
\value{
TRUE (invisibly) for successful import
}
\description{
Check if a supported DuckDB connection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_crs.R
\name{ddbs_transform}
\alias{ddbs_transform}
\title{Transform the coordinate reference system of geometries}
\usage{
ddbs_transform(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Target CRS. Can be:
\itemize{
\item A character string with EPSG code (e.g., "EPSG:4326")
\item A \code{crs} object created with \link[sf:st_crs]{sf::st_crs}
\item An \code{sf} object (uses its CRS)
\item Name of a DuckDB table (uses its CRS)
}}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Converts geometries to a different coordinate reference system (CRS), updating
their coordinates accordingly.
}
\examples{
\dontrun{
## load package
library(duckspatial)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", 
  package = "duckspatial")
)

## store in duckdb
ddbs_write_vector(conn, argentina_ddbs, "argentina")

## transform to different CRS using EPSG code
ddbs_transform("argentina", "EPSG:3857", conn)

## transform to match CRS of another object
argentina_3857_ddbs <- ddbs_transform(argentina_ddbs, "EPSG:3857")
ddbs_write_vector(conn, argentina_3857_ddbs, "argentina_3857")
ddbs_transform("argentina", argentina_3857_ddbs, conn)

## transform to match CRS of another DuckDB table
ddbs_transform("argentina", "argentina_3857", conn)

## transform without using a connection
ddbs_transform(argentina_ddbs, "EPSG:3857")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_options.R
\name{ddbs_sitrep}
\alias{ddbs_sitrep}
\title{Report duckspatial configuration status}
\usage{
ddbs_sitrep()
}
\value{
Invisibly returns a list with the current status configuration.
}
\description{
Displays useful information about the current configuration, including global options
and the status of the default DuckDB connection.
}
\examples{
ddbs_sitrep()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_register.R
\name{ddbs_register_vector}
\alias{ddbs_register_vector}
\title{Register an SF Object as an Arrow Table in DuckDB}
\usage{
ddbs_register_vector(conn, data, name, overwrite = FALSE, quiet = FALSE)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{data}{A \code{sf} object to write to the DuckDB database, or the path to
a local file that can be read with \code{ST_READ}}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
TRUE (invisibly) on successful registration.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

\code{ddbs_register_vector()} was renamed to \code{\link{ddbs_register_table}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_as_spatial.R
\name{ddbs_point}
\alias{ddbs_point}
\title{Create point geometries from coordinate vectors}
\usage{
ddbs_point(
  x,
  y,
  z = NULL,
  m = NULL,
  ...,
  crs = NULL,
  geom_col = "geometry",
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Numeric vector of X (longitude) coordinates.}

\item{y}{Numeric vector of Y (latitude) coordinates.}

\item{z}{Optional numeric vector of Z (elevation) coordinates.}

\item{m}{Optional numeric vector of M (measure) coordinates. Requires
\code{z}.}

\item{...}{Named vectors of additional attribute columns to include in the
output. Each must have the same length as \code{x}.}

\item{crs}{Character or numeric CRS specification (e.g. \code{"EPSG:4326"}
or \code{4326}). Defaults to \code{NULL} (no CRS assigned).}

\item{geom_col}{Name of the geometry column in the output. Defaults to
\code{"geometry"}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Constructs POINT geometries from numeric coordinate vectors, optionally
including Z (elevation) and M (measure) dimensions and extra attribute
columns.
}
\examples{
\dontrun{
library(duckspatial)

## 2D points
ddbs_point(
  x = c(-58.38, -64.18, -60.64),
  y = c(-34.60, -31.42, -32.95),
  crs = 4326
)

## 3D points with extra columns
ddbs_point(
  x   = c(0, 1, 2),
  y   = c(0, 1, 2),
  z   = c(10, 20, 30),
  id  = 1:3,
  crs = 4326
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_predicates.R
\name{ddbs_predicate}
\alias{ddbs_predicate}
\alias{ddbs_intersects}
\alias{ddbs_covers}
\alias{ddbs_touches}
\alias{ddbs_is_within_distance}
\alias{ddbs_disjoint}
\alias{ddbs_within}
\alias{ddbs_contains}
\alias{ddbs_overlaps}
\alias{ddbs_crosses}
\alias{ddbs_equals}
\alias{ddbs_covered_by}
\alias{ddbs_intersects_extent}
\alias{ddbs_contains_properly}
\alias{ddbs_within_properly}
\title{Evaluate spatial predicates between geometries}
\usage{
ddbs_predicate(
  x,
  y,
  predicate = "intersects",
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  id_x = NULL,
  id_y = NULL,
  sparse = TRUE,
  distance = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = TRUE
)

ddbs_intersects(x, y, ...)

ddbs_covers(x, y, ...)

ddbs_touches(x, y, ...)

ddbs_is_within_distance(x, y, distance = NULL, ...)

ddbs_disjoint(x, y, ...)

ddbs_within(x, y, ...)

ddbs_contains(x, y, ...)

ddbs_overlaps(x, y, ...)

ddbs_crosses(x, y, ...)

ddbs_equals(x, y, ...)

ddbs_covered_by(x, y, ...)

ddbs_intersects_extent(x, y, ...)

ddbs_contains_properly(x, y, ...)

ddbs_within_properly(x, y, ...)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}}

\item{predicate}{A geometry predicate function. Defaults to \code{intersects},
a wrapper of \code{ST_Intersects}. See details for other options.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{id_x}{Character; optional name of the column in \code{x} whose values will
be used to name the list elements. If \code{NULL}, integer row numbers of \code{x} are used.}

\item{id_y}{Character; optional name of the column in \code{y} whose values will
replace the integer indices returned in each element of the list.}

\item{sparse}{A logical value. If \code{TRUE}, it returns a sparse table/list. If \code{FALSE},
it returns a wide table/matrix.}

\item{distance}{a numeric value specifying the distance for ST_DWithin. Units correspond to
the coordinate system of the geometry (e.g. degrees or meters)}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{...}{Passed to \link{ddbs_predicate}}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{tbl_duckdb_connection} (lazy data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected list.
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Determines which geometries in one dataset satisfy a specified spatial
relationship with geometries in another dataset, such as intersection,
containment, or touching.
}
\details{
This function provides a unified interface to all spatial predicate operations
in DuckDB's spatial extension. It performs pairwise comparisons between all
geometries in \code{x} and \code{y} using the specified predicate.
\subsection{Available Predicates}{
\itemize{
\item \strong{intersects}: Geometries share at least one point
\item \strong{covers}: Geometry \code{x} completely covers geometry \code{y}
\item \strong{touches}: Geometries share a boundary but interiors do not intersect
\item \strong{disjoint}: Geometries have no points in common
\item \strong{within}: Geometry \code{x} is completely inside geometry \code{y}
\item \strong{dwithin}: Geometry \code{x} is completely within a distance of geometry \code{y}
\item \strong{contains}: Geometry \code{x} completely contains geometry \code{y}
\item \strong{overlaps}: Geometries share some but not all points
\item \strong{crosses}: Geometries have some interior points in common
\item \strong{equals}: Geometries are spatially equal
\item \strong{covered_by}: Geometry \code{x} is completely covered by geometry \code{y}
\item \strong{intersects_extent}: Bounding boxes of geometries intersect (faster but less precise)
\item \strong{contains_properly}: Geometry \code{x} contains geometry \code{y} without boundary contact
\item \strong{within_properly}: Geometry \code{x} is within geometry \code{y} without boundary contact
}

If \code{x} or \code{y} are not DuckDB tables, they are automatically copied into a
temporary in-memory DuckDB database (unless a connection is supplied via \code{conn}).

\code{id_x} or \code{id_y} may be used to replace the default integer indices with the
values of an identifier column in \code{x} or \code{y}, respectively.
}
}
\examples{
\dontrun{
## Load packages
library(duckspatial)
library(dplyr)

## create in-memory DuckDB database
conn <- ddbs_create_conn(dbdir = "memory")

## read countries data, and rivers
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |>
  filter(CNTR_ID \%in\% c("PT", "ES", "FR", "IT"))

rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson", 
  package = "duckspatial")
) |>
  ddbs_transform(ddbs_crs(countries_ddbs))

## Store in DuckDB
ddbs_write_vector(conn, countries_ddbs, "countries")
ddbs_write_vector(conn, rivers_ddbs, "rivers")

## Example 1: Check which rivers intersect each country
ddbs_predicate(countries_ddbs, rivers_ddbs, predicate = "intersects")
ddbs_intersects(countries_ddbs, rivers_ddbs)

## Example 2: Find neighboring countries
ddbs_predicate(
  countries_ddbs, 
  countries_ddbs, 
  predicate = "touches",
  id_x = "NAME_ENGL", 
  id_y = "NAME_ENGL"
)

ddbs_touches(
  countries_ddbs, 
  countries_ddbs, 
  id_x = "NAME_ENGL", 
  id_y = "NAME_ENGL"
)

## Example 3: Find rivers that don't intersect countries
ddbs_predicate(
  countries_ddbs, 
  rivers_ddbs, 
  predicate = "disjoint",
  id_x = "NAME_ENGL", 
  id_y = "RIVER_NAME"
)

## Example 4: Use table names inside duckdb
ddbs_predicate("countries", "rivers", predicate = "within", conn, id_x = "NAME_ENGL")
ddbs_within("countries", "rivers", conn,  id_x = "NAME_ENGL")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{get_st_predicate}
\alias{get_st_predicate}
\title{Gets predicate name}
\usage{
get_st_predicate(predicate)
}
\arguments{
\item{predicate}{A geometry predicate function. Defaults to \code{intersects},
a wrapper of \code{ST_Intersects}. See details for other options.}
}
\value{
character
}
\description{
Gets a full predicate name from the shorter version
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_not_exported.R
\name{overwrite_table}
\alias{overwrite_table}
\title{Feedback for overwrite argument}
\usage{
overwrite_table(x, conn, quiet, overwrite)
}
\arguments{
\item{x}{table name}

\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}
}
\value{
cli message
}
\description{
Feedback for overwrite argument
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_quadkey.R
\name{ddbs_quadkey}
\alias{ddbs_quadkey}
\title{Convert point geometries to QuadKey tiles}
\usage{
ddbs_quadkey(
  x,
  level = 10,
  field = NULL,
  fun = "mean",
  background = NA,
  conn = NULL,
  name = NULL,
  output = "polygon",
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{level}{An integer specifying the zoom level for QuadKey generation (1-23).
Higher values provide finer spatial resolution. Default is 10.}

\item{field}{Character string specifying the field name for aggregation.}

\item{fun}{aggregation function for when there are multiple quadkeys (e.g. "mean",
"min", "max", "sum").}

\item{background}{numeric. Default value in raster cells without values. Only used when
\code{output = "raster"}}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{output}{Character string specifying output format. One of:
\itemize{
\item \code{"polygon"} - Returns QuadKey tile boundaries as \code{duckspatial_df} (default)
\item \code{"raster"} - Returns QuadKey values as a \code{SpatRaster}
\item \code{"tilexy"} - Returns tile XY coordinates as a \code{tibble}
}}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the output argument
\itemize{
\item \code{polygon} (default): A lazy spatial data frame backed by dbplyr/DuckDB.
\item \code{raster}: An eagerly collected \code{SpatRaster} object in R memory.
\item \code{tilexy}: An eagerly collected \code{tibble} without geometry in R memory.
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Transforms point geometries into QuadKey identifiers at a specified zoom level,
a hierarchical spatial indexing system used by mapping services.
}
\details{
QuadKeys divide the world into a hierarchical grid of tiles, where each tile
is subdivided into four smaller tiles at the next zoom level. This function
wraps DuckDB's ST_QuadKey spatial function to generate these tiles from input geometries.

Note that creating a table inside the connection will generate a non-spatial table, and
therefore, it cannot be read with \link{ddbs_read_table}.
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(sf)
library(terra)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## create random points in Argentina
argentina_sf <- st_read(system.file("spatial/argentina.geojson", package = "duckspatial"))
rand_sf <- st_sample(argentina_sf, 100) |> st_as_sf()
rand_sf["var"] <- runif(100)

## store in duckdb
ddbs_write_vector(conn, rand_sf, "rand_sf")

## generate QuadKey polygons at zoom level 8
qkey_ddbs <- ddbs_quadkey(conn = conn, "rand_sf", level = 8, output = "polygon")

## generate QuadKey raster with custom field name
qkey_rast <- ddbs_quadkey(conn = conn, "rand_sf", level = 6, output = "raster", field = "var")

## generate Quadkey XY tiles
qkey_tiles_tbl <- ddbs_quadkey(conn = conn, "rand_sf", level = 10, output = "tilexy")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_geom_conversion.R
\name{ddbs_as_format}
\alias{ddbs_as_format}
\alias{ddbs_as_text}
\alias{ddbs_as_wkb}
\alias{ddbs_as_hexwkb}
\alias{ddbs_as_geojson}
\title{Convert geometries to standard interchange formats}
\usage{
ddbs_as_text(x, conn = NULL)

ddbs_as_wkb(x, conn = NULL)

ddbs_as_hexwkb(x, conn = NULL)

ddbs_as_geojson(x, conn = NULL, feature_collection = TRUE)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{feature_collection}{Logical, only used by \code{ddbs_as_geojson()}. If \code{TRUE}
(default), all rows are returned as a single GeoJSON \code{FeatureCollection}
string. If \code{FALSE}, a character vector with one \code{Feature} string per row is
returned.}
}
\value{
Depending on the function:
\itemize{
\item \code{ddbs_as_text()} returns a character vector of WKT geometries
\item \code{ddbs_as_wkb()} returns a list of raw vectors (binary WKB)
\item \code{ddbs_as_hexwkb()} returns a character vector of HEXWKB strings
\item \code{ddbs_as_geojson()} returns a single GeoJSON \code{FeatureCollection}
string (class \code{"geojson"}) when \code{feature_collection = TRUE}, or a
character vector of \code{Feature} strings (one per row) when \code{FALSE};
non-geometry columns are included as feature properties
}
}
\description{
Convert spatial geometries to common interchange formats using DuckDB spatial
serialization functions.
\itemize{
\item \code{ddbs_as_text()} – Convert geometries to Well-Known Text (WKT)
\item \code{ddbs_as_wkb()} – Convert geometries to Well-Known Binary (WKB)
\item \code{ddbs_as_hexwkb()} – Convert geometries to hexadecimal Well-Known Binary (HEXWKB)
\item \code{ddbs_as_geojson()} – Convert to GeoJSON (a \code{FeatureCollection} or one \code{Feature} per row)
}
}
\details{
These functions are thin wrappers around DuckDB spatial serialization
functions (\code{ST_AsText}, \code{ST_AsWKB}, \code{ST_AsHEXWKB}, and \code{ST_AsGeoJSON}).

They are useful for exporting geometries into widely supported formats for
interoperability with external spatial tools, databases, and web services.

Unlike the other serializers, which only encode the geometry, \code{ddbs_as_geojson()}
produces complete GeoJSON \code{Feature}s: the geometry is placed in the \code{geometry}
member and all remaining (non-geometry) columns are included as feature
\code{properties}. By default the features are wrapped in a single \code{FeatureCollection},
the canonical GeoJSON artifact for writing \code{.geojson} files and feeding web
services and mapping libraries.
}
\examples{
\dontrun{
library(duckspatial)

argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", package = "duckspatial")
)

ddbs_as_text(argentina_ddbs)
ddbs_as_wkb(argentina_ddbs)
ddbs_as_hexwkb(argentina_ddbs)

## a single FeatureCollection (default)
ddbs_as_geojson(argentina_ddbs)

## one Feature string per row
ddbs_as_geojson(argentina_ddbs, feature_collection = FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_open_dataset.R
\name{ddbs_open_dataset}
\alias{ddbs_open_dataset}
\title{Open spatial dataset lazily via DuckDB}
\usage{
ddbs_open_dataset(
  path,
  crs = NULL,
  layer = NULL,
  geom_col = NULL,
  conn = NULL,
  parquet_binary_as_string = NULL,
  parquet_file_row_number = NULL,
  parquet_filename = NULL,
  parquet_hive_partitioning = NULL,
  parquet_union_by_name = NULL,
  parquet_encryption_config = NULL,
  read_shp_mode = c("ST_ReadSHP", "GDAL"),
  read_osm_mode = c("GDAL", "ST_ReadOSM"),
  shp_encoding = NULL,
  gdal_spatial_filter = NULL,
  gdal_spatial_filter_box = NULL,
  gdal_keep_wkb = NULL,
  gdal_max_batch_size = NULL,
  gdal_sequential_layer_scan = NULL,
  gdal_sibling_files = NULL,
  gdal_allowed_drivers = NULL,
  gdal_open_options = NULL
)
}
\arguments{
\item{path}{Path to spatial file. Supports Parquet (\code{.parquet}, with optional GeoParquet metadata),
GeoJSON, GeoPackage, Shapefile, FlatGeoBuf, OSM PBF, DuckDB database files
(\code{.duckdb}, \code{.db}, and \code{.ddb}), and other GDAL-supported formats.}

\item{crs}{Coordinate reference system. Can be an EPSG code (e.g., 4326),
a CRS string, or an \code{sf} crs object. If \code{NULL} (default),
attempts to auto-detect from the file, including native DuckDB CRS metadata
and duckspatial-managed column-comment metadata for compatibility
\code{.duckdb} files.}

\item{layer}{Layer name or index to read (ST_Read). For DuckDB database
files, this is required and specifies the table name to read. Default is
NULL (first layer for ST_Read).}

\item{geom_col}{Name of the geometry column. Default is \code{NULL}, which attempts
auto-detection.}

\item{conn}{DuckDB connection to use. If NULL, uses the default connection.}

\item{parquet_binary_as_string}{Logical. (Parquet) If TRUE, load binary columns as strings.}

\item{parquet_file_row_number}{Logical. (Parquet) If TRUE, include a \code{file_row_number} column.}

\item{parquet_filename}{Logical. (Parquet) If TRUE, include a \code{filename} column.}

\item{parquet_hive_partitioning}{Logical. (Parquet) If TRUE, interpret path as Hive partitioned.}

\item{parquet_union_by_name}{Logical. (Parquet) If TRUE, unify columns by name.}

\item{parquet_encryption_config}{List/Struct. (Parquet) Encryption configuration (advanced).}

\item{read_shp_mode}{Mode for reading Shapefiles. "ST_ReadSHP" (default, fast native reader) or "GDAL" (ST_Read).}

\item{read_osm_mode}{Mode for reading OSM PBF files. "GDAL" (default, ST_Read) or "ST_ReadOSM" (fast native reader, no geometry).}

\item{shp_encoding}{Encoding for Shapefiles when using "ST_ReadSHP" (e.g., "UTF-8", "ISO-8859-1").}

\item{gdal_spatial_filter}{Optional WKB geometry (as raw vector or hex string) to filter spatially (ST_Read only).}

\item{gdal_spatial_filter_box}{Optional bounding box (as numeric vector \code{c(minx, miny, maxx, maxy)})
(ST_Read only).}

\item{gdal_keep_wkb}{Logical. If TRUE, return WKB blobs instead of GEOMETRY type (ST_Read only).}

\item{gdal_max_batch_size}{Integer. Maximum batch size for reading (ST_Read only).}

\item{gdal_sequential_layer_scan}{Logical. If TRUE, scan layers sequentially (ST_Read only).}

\item{gdal_sibling_files}{Character vector. List of sibling files (ST_Read only).}

\item{gdal_allowed_drivers}{Character vector. List of allowed GDAL drivers (ST_Read only).}

\item{gdal_open_options}{Character vector. Driver-specific open options (ST_Read only).}
}
\value{
A \code{duckspatial_df} object.
}
\description{
Reads spatial data directly from disk using DuckDB's spatial extension or
native Parquet reader, returning a \code{duckspatial_df} object for lazy processing.
}
\references{
This function is inspired by the dataset opening logic in the
\code{duckdbfs} package (\url{https://github.com/cboettig/duckdbfs}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_mvt.R
\name{ddbs_write_mbtiles}
\alias{ddbs_write_mbtiles}
\title{Write a Mapbox Vector Tile pyramid to an MBTiles file}
\usage{
ddbs_write_mbtiles(
  x,
  dsn,
  layer_name = "layer",
  zoom_levels = 0:6,
  extent = 4096L,
  buffer = 256L,
  conn = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{dsn}{Path to the output \code{.mbtiles} file.}

\item{layer_name}{Name of the vector tile layer. This is the value you pass
as \code{source_layer} when adding the layer in a web map. Defaults to
\code{"layer"}.}

\item{zoom_levels}{Integer vector of zoom levels to generate (e.g.
\code{0:8}). Defaults to \code{0:6}.}

\item{extent}{Integer tile extent (resolution). Defaults to \code{4096}.}

\item{buffer}{Integer tile buffer in pixels. Defaults to \code{256}.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{overwrite}{Logical. Overwrite \code{dsn} if it already exists. Defaults
to \code{FALSE}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
The path to \code{dsn}, invisibly.
}
\description{
Generates a pyramid of Mapbox Vector Tiles (MVT) from a spatial dataset and
stores them in an \href{https://github.com/mapbox/mbtiles-spec}{MBTiles} file
(a SQLite database). For each tile in the requested zoom range the geometries
are clipped and re-projected into tile space (\code{ST_AsMVTGeom}) and encoded
into a binary vector tile (\code{ST_AsMVT}).
}
\details{
The input is re-projected to web mercator (\code{EPSG:3857}), so \code{x} must
have a CRS. All non-geometry columns are carried into the tiles as feature
properties; columns whose type is not supported by the vector tile format are
cast to \code{DOUBLE} (numeric types) or \code{VARCHAR} (everything else), and
\code{BLOB} columns are dropped.

Writing the SQLite container relies on DuckDB's \code{sqlite} extension, which
is installed and loaded automatically (an internet connection may be required
the first time).

MBTiles can be served with a vector tile server, or converted to
\href{https://github.com/protomaps/PMTiles}{PMTiles} (e.g. with the
\code{pmtiles convert} CLI) for static hosting. The resulting tiles can then
be displayed with the \pkg{mapgl} package, for example:

\preformatted{
library(mapgl)
maplibre() |>
  add_pmtiles_source(id = "src", url = "https://example.com/argentina.pmtiles") |>
  add_fill_layer(id = "fill", source = "src", source_layer = "argentina")
}
}
\examples{
\dontrun{
library(duckspatial)

argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", package = "duckspatial")
)

ddbs_write_mbtiles(
  argentina_ddbs,
  dsn         = tempfile(fileext = ".mbtiles"),
  layer_name  = "argentina",
  zoom_levels = 0:6
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_ops_binary.R
\name{ddbs_binary_funs}
\alias{ddbs_binary_funs}
\alias{ddbs_intersection}
\alias{ddbs_difference}
\alias{ddbs_sym_difference}
\alias{ddbs_crop}
\alias{ddbs_shortest_line}
\title{Geometry binary operations}
\usage{
ddbs_intersection(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_difference(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_sym_difference(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_crop(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_shortest_line(
  x,
  y,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Perform geometric set operations between two sets of geometries.
}
\details{
These functions perform different geometric set operations:
\describe{
\item{\code{ddbs_intersection}}{Returns the geometric intersection of two sets
of geometries, producing the area, line, or point shared by both.}
\item{\code{ddbs_crop}}{Returns the geometric intersection of two sets of
geometries, using the bounding box of \code{y}, rather than its original geometry}
\item{\code{ddbs_difference}}{Returns the portion of the first geometry that
does not overlap with the second geometry.}
\item{\code{ddbs_sym_difference}}{Returns the portions of both geometries
that do not overlap with each other. Equivalent to
\verb{(A - B) UNION (B - A)}.}
\item{\code{ddbs_shortest_line}}{Returns a LINESTRING connecting the closest
points between each pair of geometries from \code{x} and \code{y}. Performs
a cross join (all combinations), so the output contains columns from both
\code{x} and \code{y} (excluding \code{y}'s geometry column). For each pair
the line has the same length as \code{ST_Distance}.}
}
}
\examples{
\dontrun{
library(duckspatial)
library(sf)

# Create two overlapping polygons for testing
poly1 <- st_polygon(list(matrix(c(
  0, 0,
  4, 0,
  4, 4,
  0, 4,
  0, 0
), ncol = 2, byrow = TRUE)))

poly2 <- st_polygon(list(matrix(c(
  2, 2,
  6, 2,
  6, 6,
  2, 6,
  2, 2
), ncol = 2, byrow = TRUE)))

x <- st_sf(id = 1, geometry = st_sfc(poly1), crs = 4326)
y <- st_sf(id = 2, geometry = st_sfc(poly2), crs = 4326)

# Visualize the input polygons
plot(st_geometry(x), col = "lightblue", main = "Input Polygons")
plot(st_geometry(y), col = "lightcoral", add = TRUE, alpha = 0.5)

# Intersection: only the overlapping area (2,2 to 4,4)
result_intersect <- ddbs_intersection(x, y)
plot(st_geometry(result_intersect), col = "purple", 
     main = "Intersection")

# Difference: part of x not in y (L-shaped area)
result_diff <- ddbs_difference(x, y)
plot(st_geometry(result_diff), col = "lightblue", 
     main = "Difference (x - y)")

# Symmetric Difference: parts of both that don't overlap
result_symdiff <- ddbs_sym_difference(x, y)
plot(st_geometry(result_symdiff), col = "orange", 
     main = "Symmetric Difference")

# Using with database connection
conn <- ddbs_create_conn(dbdir = "memory")

ddbs_write_vector(conn, x, "poly_x")
ddbs_write_vector(conn, y, "poly_y")

# Perform operations with connection
ddbs_intersection("poly_x", "poly_y", conn)
ddbs_difference("poly_x", "poly_y", conn)
ddbs_sym_difference("poly_x", "poly_y", conn)

# Save results to database table
ddbs_difference("poly_x", "poly_y", conn, name = "diff_result")
}

\dontrun{
library(duckspatial)
library(sf)

# Two separate point clouds in a projected CRS
origins <- st_as_sf(
  data.frame(id = 1:3, x = c(0, 5, 10), y = c(0, 0, 0)),
  coords = c("x", "y"), crs = "EPSG:3857"
)
targets <- st_as_sf(
  data.frame(id = 1:3, x = c(0, 5, 10), y = c(3, 4, 5)),
  coords = c("x", "y"), crs = "EPSG:3857"
)

# Returns LINESTRING geometries connecting the closest points
ddbs_shortest_line(origins, targets)
ddbs_shortest_line(origins, targets, mode = "sf")

# Works with any geometry type
ddbs_shortest_line(origins, argentina_ddbs)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nanoarrow_methods.R
\name{as_nanoarrow_array_stream.duckspatial_df}
\alias{as_nanoarrow_array_stream.duckspatial_df}
\title{Convert a duckspatial_df to a nanoarrow_array_stream}
\usage{
as_nanoarrow_array_stream.duckspatial_df(x, ..., schema = NULL, native = FALSE)
}
\arguments{
\item{x}{A \code{duckspatial_df} object}

\item{...}{Additional arguments passed to downstream methods}

\item{schema}{Optional target schema for the entire stream.}

\item{native}{If TRUE, transforms WKB to a "Native" GeoArrow layout (e.g.,
Point, Polygon) using optimized Arrow-to-Arrow kernels. This layout is
optimized for high-performance rendering in tools like Deck.GL.}
}
\value{
A \code{nanoarrow_array_stream}
}
\description{
Convert a duckspatial_df to a nanoarrow_array_stream
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_extent_funs.R
\name{ddbs_envelope}
\alias{ddbs_envelope}
\title{Get the envelope (bounding box) of geometries}
\usage{
ddbs_envelope(
  x,
  by_feature = FALSE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{by_feature}{Logical. If \code{TRUE}, the geometric operation is applied
separately to each geometry. If \code{FALSE}, the geometric
operation is applied to the data as a whole.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Returns the minimum axis-aligned rectangle that fully contains the geometry.
}
\details{
ST_Envelope returns the minimum bounding rectangle (MBR) of a geometry as a
polygon. For points and lines, this creates a rectangular polygon that
encompasses the geometry. For polygons, it returns the smallest rectangle
that contains the entire polygon.

When \code{by_feature = FALSE}, all geometries are combined and a single envelope
is returned that encompasses the entire dataset.
}
\examples{
\dontrun{
## load packages
library(duckspatial)

# read data
argentina_ddbs <- ddbs_open_dataset(
  system.file("spatial/argentina.geojson", 
  package = "duckspatial")
)

# input as sf, and output as sf
env <- ddbs_envelope(x = argentina_ddbs, by_feature = TRUE)

# create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

# store in duckdb
ddbs_write_table(conn, argentina_ddbs, "argentina")

# envelope for each feature
env <- ddbs_envelope("argentina", conn, by_feature = TRUE)

# single envelope for entire dataset
env_all <- ddbs_envelope("argentina", conn, by_feature = FALSE)

# create a new table with envelopes
ddbs_envelope("argentina", conn, name = "argentina_bbox", by_feature = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_join.R
\name{ddbs_join}
\alias{ddbs_join}
\title{Perform a spatial join of two geometries}
\usage{
ddbs_join(
  x,
  y,
  join = "intersects",
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  distance = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}}

\item{join}{A geometry predicate function. Defaults to \code{"intersects"}. See
the details for other options.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{distance}{a numeric value specifying the distance for ST_DWithin. The units
should be specified in meters}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Combines two sets of geometries based on spatial relationships, such as
intersection or containment, attaching attributes from one set to the other.
}
\details{
Spatial Join Predicates:

A spatial predicate is really just a function that evaluates some spatial
relation between two geometries and returns true or false, e.g., “does a
contain b” or “is a within distance x of b”. Here is a quick overview of the
most commonly used ones, taking two geometries a and b:
\itemize{
\item \code{"ST_Intersects"}: Whether a intersects b
\item \code{"ST_Contains"}: Whether a contains b
\item \code{"ST_ContainsProperly"}: Whether a contains b without b touching a's boundary
\item \code{"ST_Within"}: Whether a is within b
\item \code{"ST_Overlaps"}: Whether a overlaps b
\item \code{"ST_Touches"}: Whether a touches b
\item \code{"ST_Equals"}: Whether a is equal to b
\item \code{"ST_Crosses"}: Whether a crosses b
\item \code{"ST_Covers"}: Whether a covers b
\item \code{"ST_CoveredBy"}: Whether a is covered by b
\item \code{"ST_DWithin"}: x)	Whether a is within distance x of b
}
}
\examples{
\dontrun{
# RECOMMENDED: Efficient lazy workflow using ddbs_open_dataset
library(duckspatial)

# Load data directly as lazy spatial data frames (CRS auto-detected)
countries <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", package = "duckspatial")
)

# Create random points
n <- 100
points <- data.frame(
    id = 1:n,
    x = runif(n, min = -180, max = 180),
    y = runif(n, min = -90, max = 90)
) |> 
  sf::st_as_sf(coords = c("x", "y"), crs = 4326) |>
  as_duckspatial_df()

# Lazy join - computation stays in DuckDB
result <- ddbs_join(points, countries, join = "within")

# Collect to sf when needed
result_sf <- dplyr::collect(result) |> sf::st_as_sf()
plot(result_sf["CNTR_NAME"])


# Alternative: using sf objects directly (legacy compatibility)
library(sf)

countries_sf <- sf::st_read(system.file("spatial/countries.geojson", package = "duckspatial"))

output <- duckspatial::ddbs_join(
    x = points,
    y = countries_sf,
    join = "within"
)


# Alternative: using table names in a duckdb connection
conn <- duckspatial::ddbs_create_conn()

ddbs_write_table(conn, points, "points", overwrite = TRUE)
ddbs_write_table(conn, countries_sf, "countries", overwrite = TRUE)

output2 <- ddbs_join(
    conn = conn,
    x = "points",
    y = "countries",
    join = "within"
)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/db_extension.R
\name{ddbs_install}
\alias{ddbs_install}
\title{Checks and installs the Spatial extension}
\usage{
ddbs_install(
  conn,
  upgrade = FALSE,
  quiet = FALSE,
  extension = "spatial",
  repos = NULL
)
}
\arguments{
\item{conn}{A \code{DBIConnection} object to a DuckDB database}

\item{upgrade}{if TRUE, it upgrades the DuckDB extension to the latest version}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{extension}{name of the extension to install, default is "spatial"}

\item{repos}{optional character string naming the repository to install the
extension from (e.g. \code{"core"}, \code{"core_nightly"}, or
\code{"community"}); a URL or path can also be supplied. If \code{NULL}
(default), the \code{core} repository is tried first, then \code{community}.
Switching an already-installed extension to a different repository requires
\code{upgrade = TRUE}. See
\url{https://duckdb.org/docs/stable/extensions/installing_extensions}.}
}
\value{
TRUE (invisibly) for successful installation
}
\description{
Checks if a spatial extension is available, and installs it in a DuckDB database
}
\examples{
\dontrun{
## load packages
library(duckspatial)
library(duckdb)

# connect to in memory database
conn <- duckdb::dbConnect(duckdb::duckdb())

# install the spatial extension
ddbs_install(conn)

# disconnect from db
duckdb::dbDisconnect(conn)

# install the h3 community extension
conn <- duckdb::dbConnect(duckdb::duckdb())
ddbs_install(conn, extension = "h3")
duckdb::dbDisconnect(conn)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_as_spatial.R
\name{ddbs_as_points}
\alias{ddbs_as_points}
\title{Generate point geometries from coordinates}
\usage{
ddbs_as_points(
  x,
  coords = c("lon", "lat"),
  crs = "EPSG:4326",
  remove = TRUE,
  na.fail = TRUE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{coords}{Character vector of length 2 specifying the names of the
longitude and latitude columns (or X and Y coordinates). Defaults to
\code{c("lon", "lat")}.}

\item{crs}{Character or numeric. The Coordinate Reference System (CRS) of the
input coordinates. Can be specified as an EPSG code (e.g., \code{"EPSG:4326"}
or \code{4326}) or a WKT string. Defaults to \code{"EPSG:4326"} (WGS84
longitude/latitude).}

\item{remove}{Logical. If \code{TRUE} (default), the coordinate columns
specified in \code{coords} are removed from the output.}

\item{na.fail}{Logical. If \code{TRUE} (default), the function errors if
any missing values (NAs) are found in the coordinate columns.}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{...}{Additional arguments. Currently supports \code{geom_col} to
specify the name of the geometry column in the output.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Converts a data frame with coordinate columns into spatial point geometries.
}
\examples{
\dontrun{
## load packages
library(duckspatial)

## create sample data with coordinates
cities_df <- data.frame(
  city = c("Buenos Aires", "Córdoba", "Rosario"),
  lon = c(-58.3816, -64.1811, -60.6393),
  lat = c(-34.6037, -31.4201, -32.9468),
  population = c(3075000, 1391000, 1193605)
)

# option 1: convert data frame to sf object
cities_ddbs <- ddbs_as_points(cities_df)

# specify custom coordinate column names and keep them in output
cities_df2 <- data.frame(
  city = c("Mendoza", "Tucumán"),
  longitude = c(-68.8272, -65.2226),
  latitude = c(-32.8895, -26.8241)
)

ddbs_as_points(cities_df2, coords = c("longitude", "latitude"), remove = FALSE)


## option 2: convert table in duckdb to spatial table

# create a duckdb connection and write data
conn <- duckspatial::ddbs_create_conn()
DBI::dbWriteTable(conn, "cities_tbl", cities_df, overwrite = TRUE)

# convert to spatial table in database
ddbs_as_points(
    x = "cities_tbl",
    conn = conn,
    name = "cities_spatial",
    overwrite = TRUE
)

# read the spatial table
ddbs_read_table(conn, "cities_spatial")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df.R
\name{new_duckspatial_df}
\alias{new_duckspatial_df}
\title{Create a duckspatial lazy spatial data frame}
\usage{
new_duckspatial_df(
  x,
  crs = NULL,
  geom_col = NULL,
  source_table = NULL,
  source_conn = NULL,
  create_view = FALSE
)
}
\arguments{
\item{x}{Input: tbl_duckdb_connection, tbl_lazy, or similar dbplyr object}

\item{crs}{CRS object or string}

\item{geom_col}{Name of geometry column (default: "geom")}

\item{source_table}{Name of the source table if applicable}

\item{source_conn}{Name of the source connection if applicable}

\item{create_view}{Logical. If TRUE, creates a temporary view for the input query.
Otherwise it generates a temporary table.}
}
\value{
A duckspatial_df object
}
\description{
Extends tbl_duckdb_connection with spatial metadata (CRS, geometry column).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ddbs_union.R
\name{ddbs_union_funs}
\alias{ddbs_union_funs}
\alias{ddbs_union}
\alias{ddbs_combine}
\alias{ddbs_union_agg}
\title{Union and combine geometries}
\usage{
ddbs_union(
  x,
  y = NULL,
  by_feature = FALSE,
  conn = NULL,
  conn_x = NULL,
  conn_y = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_combine(
  x,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)

ddbs_union_agg(
  x,
  by,
  mem = FALSE,
  conn = NULL,
  name = NULL,
  mode = NULL,
  overwrite = FALSE,
  quiet = FALSE
)
}
\arguments{
\item{x}{Input spatial data. Can be:
\itemize{
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}
Data is returned from this object.}

\item{y}{Input spatial data. Can be:
\itemize{
\item \code{NULL} (default): performs only the union of \code{x}
\item A \code{duckspatial_df} object (lazy spatial data frame via dbplyr)
\item An \code{sf} object
\item A \code{tbl_lazy} from dbplyr
\item A character string naming a table/view in \code{conn}
}}

\item{by_feature}{Logical. When \code{y} is provided:
\itemize{
\item \code{FALSE} (default) - Union all geometries from both \code{x} and \code{y} into a single geometry
\item \code{TRUE} - Perform row-by-row union between matching features from \code{x} and \code{y} (requires same number of rows)
}}

\item{conn}{A connection object to a DuckDB database. If \code{NULL}, the function
runs on a temporary DuckDB database.}

\item{conn_x}{A \code{DBIConnection} object to a DuckDB database for the input \code{x}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{x}.}

\item{conn_y}{A \code{DBIConnection} object to a DuckDB database for the input \code{y}.
If \code{NULL} (default), it is resolved from \code{conn} or extracted from \code{y}.}

\item{name}{A character string of length one specifying the name of the table,
or a character string of length two specifying the schema and table
names. If \code{NULL} (the default), the function returns the result as an
\code{sf} object}

\item{mode}{Character. Controls the return type. Options:
\itemize{
\item \code{"duckspatial"} (default): Lazy spatial data frame backed by dbplyr/DuckDB
\item \code{"sf"}: Eagerly collected sf object (uses memory)
}
Can be set globally via \code{\link{ddbs_options}(mode = "...")} or
per-function via this argument. Per-function overrides global setting.}

\item{overwrite}{Boolean. whether to overwrite the existing table if it exists. Defaults
to \code{FALSE}. This argument is ignored when \code{name} is \code{NULL}.}

\item{quiet}{A logical value. If \code{TRUE}, suppresses any informational messages.
Defaults to \code{FALSE}.}

\item{by}{Character vector specifying one or more column names to
group by when computing unions. Geometries will be unioned within each group.
Default is \code{NULL}}

\item{mem}{Logical. If \code{TRUE}, uses \code{ST_MemUnion_Agg()} instead of
\code{ST_Union_Agg()} — slower but more memory efficient. Default is
\code{FALSE}. Only applies to \code{ddbs_union_agg()}.}
}
\value{
Depends on the \code{mode} argument (or global preference set by \code{\link{ddbs_options}}):
\itemize{
\item \code{duckspatial} (default): A \code{duckspatial_df} (lazy spatial data frame) backed by dbplyr/DuckDB.
\item \code{sf}: An eagerly collected object in R memory, that will return the same data type as the
\code{sf} equivalent (e.g. \code{sf} or \code{units} vector).
}
When \code{name} is provided, the result is also written as a table or view in DuckDB and the function returns \code{TRUE} (invisibly).
}
\description{
Perform union and combine operations on spatial geometries in DuckDB.
\itemize{
\item \code{ddbs_union()} - Union all geometries into one, or perform pairwise union between two datasets
\item \code{ddbs_union_agg()} - Union geometries grouped by one or more columns
\item \code{ddbs_combine()} - Combine geometries into a MULTI-geometry without dissolving boundaries
}
}
\details{
\subsection{ddbs_union(x, y, by_feature)}{

Performs geometric union operations that dissolve internal boundaries:
\itemize{
\item When \code{y = NULL}: Unions all geometries in \code{x} into a single geometry
\item When \code{y != NULL} and \code{by_feature = FALSE}: Unions all geometries from both \code{x} and \code{y} into a single geometry
\item When \code{y != NULL} and \code{by_feature = TRUE}: Performs row-wise union, pairing the first geometry from \code{x} with the first from \code{y}, second with second, etc.
}
}

\subsection{ddbs_union_agg(x, by)}{

Groups geometries by one or more columns, then unions geometries within each group.
Useful for dissolving boundaries between features that share common attributes.
Set \code{mem = TRUE} to use \code{ST_MemUnion_Agg()} when memory is a constraint.
}

\subsection{ddbs_combine(x)}{

Combines all geometries into a single MULTI-geometry (e.g., MULTIPOLYGON, MULTILINESTRING)
without dissolving shared boundaries. This is faster than union but preserves all
original geometry boundaries.
}
}
\examples{
\dontrun{
## load packages
library(dplyr)
library(duckspatial)

## create a duckdb database in memory (with spatial extension)
conn <- ddbs_create_conn(dbdir = "memory")

## read data
countries_ddbs <- ddbs_open_dataset(
  system.file("spatial/countries.geojson", 
  package = "duckspatial")
) |> 
  filter(ISO3_CODE != "ATA")

rivers_ddbs <- ddbs_open_dataset(
  system.file("spatial/rivers.geojson", 
  package = "duckspatial")
) |> 
 ddbs_transform("EPSG:4326")

## combine countries into a single MULTI-geometry
## (without solving boundaries)
combined_countries_ddbs <- ddbs_combine(countries_ddbs)

## combine countries into a single MULTI-geometry
## (solving boundaries)
union_countries_ddbs <- ddbs_union(countries_ddbs)

## union of geometries of two objects, into 1 geometry
union_countries_rivers_ddbs <- ddbs_union(countries_ddbs, rivers_ddbs)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duckspatial_df_sf_methods.R
\name{duckspatial_df_sf}
\alias{duckspatial_df_sf}
\alias{st_crs.duckspatial_df}
\alias{st_geometry.duckspatial_df}
\alias{st_bbox.duckspatial_df}
\alias{st_as_sf.duckspatial_df}
\alias{print.duckspatial_df}
\title{sf methods for duckspatial_df}
\usage{
\method{st_crs}{duckspatial_df}(x, ...)

\method{st_geometry}{duckspatial_df}(obj, ...)

\method{st_bbox}{duckspatial_df}(obj, ...)

\method{st_as_sf}{duckspatial_df}(x, ...)

\method{print}{duckspatial_df}(x, ..., n = 10)
}
\description{
These methods provide sf compatibility for duckspatial_df objects,
allowing them to work with sf functions like st_crs(), st_geometry(), etc.
}
\keyword{internal}
